/*  Simple Packet Sender (SPS) - a linux packet crafting network
    tool using GTK+ graphical interface.

    Copyright (C) 2011-2023  Hohlraum (h0h1r4um@gmail.com)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "main.h"

extern int flood_mode;  // Flood flag
extern int sending;  // Sending flag for sending finite number of packets
extern int tracing;  // Flag to indicate a traceroute thread is active
extern int show_dl_messages;  // Flag to show sent and received messages on stdout as files download
extern int ftp_busy;  // Flag to indicate an FTP file download thread is active
extern int ip_asn_searching;  // Flag to indicate an IP delegation search or ASN delegation search thread is active

G_LOCK_EXTERN (sending);  // MUTEX for flag indicating packets are being sent

// Someone has entered a hostname in Host-to-IP converter
int
on_entry19_changed (GtkWidget *entry19, SPSData *data) {

  const char *entry19_text;
  GtkStyleContext *context;
  GtkTextBuffer *textbuffer3, *textbuffer6;  // IPv4 textview, IPv6 textview
  GtkTextIter end;
  struct addrinfo hints, *res, *p;
  int status;
  char ipstr [INET6_ADDRSTRLEN + 2];  // Add 2 chars for \n + string termination
  void *ptr;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (entry19);

  entry19_text = gtk_entry_get_text (GTK_ENTRY (entry19));

  memset (&hints, 0, sizeof (struct addrinfo));
  hints.ai_family = AF_UNSPEC;
  hints.ai_socktype = SOCK_STREAM;
  hints.ai_flags = hints.ai_flags | AI_CANONNAME;

  textbuffer3 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview3));
  textbuffer6 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview6));

  if ((status = getaddrinfo (entry19_text, NULL, &hints, &res)) != 0) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    gtk_text_buffer_set_text (textbuffer3, "No result", -1);
    gtk_text_buffer_set_text (textbuffer6, "No result", -1);
    return (EXIT_FAILURE);
  }

  // Clear previous IP results from textviews.
  gtk_text_buffer_set_text (textbuffer3, "", -1);
  gtk_text_buffer_set_text (textbuffer6, "", -1);
 
  for (p = res; p != NULL; p = p->ai_next) {

    switch (p->ai_family) {
      case AF_INET:
        ptr = &((struct sockaddr_in *) p->ai_addr)->sin_addr;
        break;
      case AF_INET6:
        ptr = &((struct sockaddr_in6 *) p->ai_addr)->sin6_addr;
        break;
      default:
        ptr = NULL;
        break;
    }
    if (ptr == NULL) {
      gtk_style_context_add_class (context, "error");  // Apply the error style to entry text.
      freeaddrinfo (res);
      textbuffer3 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview3));
      gtk_text_buffer_set_text (textbuffer3, "No result", -1);
      textbuffer6 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview6));
      gtk_text_buffer_set_text (textbuffer6, "No result", -1);
      return (EXIT_FAILURE);
    }

    if (inet_ntop (p->ai_family, ptr, ipstr, sizeof ipstr) == NULL) {
      gtk_style_context_add_class (context, "error");  // Apply the error style to entry text.
      freeaddrinfo (res);
      textbuffer3 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview3));
      gtk_text_buffer_set_text (textbuffer3, "No result", -1);
      textbuffer6 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview6));
      gtk_text_buffer_set_text (textbuffer6, "No result", -1);
      return (EXIT_FAILURE);
    }

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    strncat (ipstr, "\n", 1 * sizeof (char));

    if (p->ai_family == AF_INET6) {
      textbuffer6 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview6));
      gtk_text_buffer_get_end_iter (textbuffer6, &end);
      gtk_text_buffer_insert (textbuffer6, &end, ipstr, -1);
      gtk_text_view_scroll_to_iter (GTK_TEXT_VIEW (data->textview6), &end, 0.0, FALSE, 0, 0);
    } else {
      textbuffer3 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview3));
      gtk_text_buffer_get_end_iter (textbuffer3, &end);
      gtk_text_buffer_insert (textbuffer3, &end, ipstr, -1);
      gtk_text_view_scroll_to_iter (GTK_TEXT_VIEW (data->textview3), &end, 0.0, FALSE, 0, 0);
    }
  }

  freeaddrinfo (res);

  return (EXIT_SUCCESS);
}

// Someone has entered an IPv4 address in IP-to-Hostname converter
int
on_entry20_changed (GtkWidget *entry20, SPSData *data) {

  const char *entry20_text;
  GtkStyleContext *context;
  GtkTextBuffer *textbuffer2;
  struct sockaddr_in sa; 
  sa.sin_family = AF_INET;
  char hostname[NI_MAXHOST];

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (entry20);

  memset (hostname, 0, NI_MAXHOST * sizeof (char));

  entry20_text = gtk_entry_get_text (GTK_ENTRY (entry20));
  textbuffer2 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview2));

  // Is it a valid IP address?
  if (inet_pton (AF_INET, entry20_text, &sa.sin_addr) < 1) {
    gtk_style_context_add_class (context, "error");  // Apply the error style to entry text.
    gtk_text_buffer_set_text (textbuffer2, "No result", -1);
    return (EXIT_FAILURE);
  }

  if (getnameinfo ((struct sockaddr*)&sa, sizeof (sa), hostname, sizeof (hostname), NULL, 0, 0) != 0) {
    gtk_style_context_add_class (context, "error");  // Apply the error style to entry text.
    gtk_text_buffer_set_text (textbuffer2, "No result", -1);
    return (EXIT_FAILURE);
  }

  // Remove the error style (if applied) to entry text.
  gtk_style_context_remove_class (context, "error");

  gtk_entry_set_text (GTK_ENTRY (data->entry20), entry20_text);

  // Show hostname
  gtk_text_buffer_set_text (textbuffer2, hostname, -1);

  return (EXIT_SUCCESS);
}

// Someone has entered an IPv6 address in IP-to-Hostname converter
int
on_entry21_changed (GtkWidget *entry21, SPSData *data) {

  const char *entry21_text;
  GtkStyleContext *context;
  GtkTextBuffer *textbuffer29;
  struct sockaddr_in6 sa;
  char hostname[NI_MAXHOST];

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (entry21);

  memset (hostname, 0, NI_MAXHOST * sizeof (char));

  memset (&sa, 0, sizeof (sa));
  sa.sin6_family = AF_INET6;

  entry21_text = gtk_entry_get_text (GTK_ENTRY (entry21));

  textbuffer29 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview29));

  // Is it a valid IP address? Valid if inet_pton returns 1.
  if (inet_pton (AF_INET6, entry21_text, &sa.sin6_addr) != 1) {

    // Invalid IP
    gtk_style_context_add_class (context, "error");  // Apply the error style to entry text.
    gtk_text_buffer_set_text (textbuffer29, "No result", -1);
    return (EXIT_FAILURE);

  // Valid IP
  } else {
    if (getnameinfo ((struct sockaddr *)&sa, sizeof (sa), hostname, sizeof (hostname), NULL, 0, 0) != 0) {
      gtk_style_context_add_class (context, "error");  // Apply the error style to entry text.
      gtk_text_buffer_set_text (textbuffer29, "No result", -1);
      return (EXIT_FAILURE);
    }

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    // Show hostname
    textbuffer29 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview29));
    gtk_text_buffer_set_text (textbuffer29, hostname, -1);

    return (EXIT_SUCCESS);
  }
}

// This host's interface name to be converted to link-layer (MAC) address.
int
on_entry139_changed (GtkWidget *entry139, SPSData *data) {

  int i, mtu;
  const char *entry139_text;
  GtkTextBuffer *textbuffer10, *textbuffer21;
  GtkStyleContext *context;
  char *text, *temp;
  uint8_t *mac;

  // Set parent window for reporting error messages.
  data->parent = data->main_window;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (entry139);

  // Array to hold various strings of text
  text = allocate_strmem (TEXT_STRINGLEN);
  temp = allocate_strmem (TEXT_STRINGLEN);

  // Array of unsigned chars
  mac = allocate_ustrmem (6);

  entry139_text = gtk_entry_get_text (GTK_ENTRY (entry139));
  textbuffer10 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview10));
  textbuffer21 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview21));

  // Obtain MAC address for interface.
  if (interface_getmac ((char *) entry139_text, mac) == EXIT_FAILURE) {
    gtk_text_buffer_set_text (textbuffer10, "", -1);
    gtk_text_buffer_set_text (textbuffer21, "", -1);
    gtk_style_context_add_class (context, "error");
    free (text);
    free (temp);
    free (mac);
    return (EXIT_FAILURE);

  } else {

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    // Put MAC address into a string.
    memset (text, 0, TEXT_STRINGLEN * sizeof (char));
    for (i=0; i<5; i++) {
      memset (temp, 0, TEXT_STRINGLEN * sizeof (char));
      sprintf (temp, "%02x:", (uint8_t) mac[i]);
      strncat (text, temp, TEXT_STRINGLEN - strnlen (text, TEXT_STRINGLEN - 1) - 1);
    }
    memset (temp, 0, TEXT_STRINGLEN * sizeof (char));
    sprintf (temp, "%02x", (uint8_t) mac[5]);
    strncat (text, temp, TEXT_STRINGLEN - strnlen (text, TEXT_STRINGLEN - 1) - 1);

    // Print MAC address in textview.
    gtk_text_buffer_set_text (textbuffer10, text, -1);

    // Obtain interface's current MTU.
    if ((mtu = interface_getmtu ((char *) entry139_text)) == 0) {
      gtk_text_buffer_set_text (textbuffer10, "", -1);
      gtk_text_buffer_set_text (textbuffer21, "", -1);
      gtk_style_context_add_class (context, "error");
      sprintf (data->warning_text, "Failed to obtain interface's MTU.");
      report_warning (data);
      free (text);
      free (temp);
      free (mac);
      return (EXIT_FAILURE);
    }

    // Print MTU in textview.
    memset (text, 0, TEXT_STRINGLEN * sizeof (char));
    sprintf (text, "%i", mtu);
    gtk_text_buffer_set_text (textbuffer21, text, -1);
  }

  // Free allocated memory.
  free (text);
  free (temp);
  free (mac);

  return (EXIT_SUCCESS);
}

// Interface name for IP-to-link-layer (MAC) address conversion.
int
on_entry140_changed (GtkWidget *entry140, SPSData *data) {

  (void) data;  // This statement suppresses compiler warning about unused parameter.

  const char *entry140_text;
  GtkStyleContext *context;
  struct sockaddr_ll device;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (entry140);

  entry140_text = gtk_entry_get_text (GTK_ENTRY (entry140));

  // We attempt to resolve interface index only to see if interface name exists.
  // The contents of entry140 will be accessed by arp() or neighbor_discovery().
  if ((device.sll_ifindex = if_nametoindex (entry140_text)) == 0) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    return (EXIT_FAILURE);
  }

  // Remove the error style (if applied) to entry text.
  gtk_style_context_remove_class (context, "error");

  return (EXIT_SUCCESS);
}

// Someone has entered a hostname or IPv4 address into IP-to-link-layer (MAC) address converter
int
on_entry137_changed (GtkWidget *entry137, SPSData *data) {

  (void) data;  // This statement suppresses compiler warning about unused parameter.
  (void) entry137;  // This statement suppresses compiler warning about unused parameter.

  return (EXIT_SUCCESS);
}

// Send an ARP request.
int
on_button146_clicked (GtkButton *button146, SPSData *data) {

  (void) button146;  // This statement suppresses compiler warning about unused parameter.

  int i;
  const char *entry137_text, *entry140_text;
  GtkTextBuffer *textbuffer11;
  uint8_t *dst_mac;
  char *text, *temp;

  // Get IPv4 address from UI.
  entry137_text = gtk_entry_get_text (GTK_ENTRY (data->entry137));

  // Get interface name from UI.
  entry140_text = gtk_entry_get_text (GTK_ENTRY (data->entry140));

  // Array to hold various strings of text
  text = allocate_strmem (TEXT_STRINGLEN);
  temp = allocate_strmem (TEXT_STRINGLEN);

  // Array of unsigned chars
  dst_mac = allocate_ustrmem (6);

  textbuffer11 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview11));

  // Do ARP request
  arp (dst_mac, entry137_text, entry140_text, data);

  // Put MAC address into a string.
  memset (text, 0, TEXT_STRINGLEN * sizeof (char));
  for (i=0; i<5; i++) {
    memset (temp, 0, TEXT_STRINGLEN * sizeof (char));
    sprintf (temp, "%02x:", (uint8_t) dst_mac[i]);
    strncat (text, temp, TEXT_STRINGLEN - strnlen (text, TEXT_STRINGLEN - 1) - 1);
  }
  memset (temp, 0, TEXT_STRINGLEN * sizeof (char));
  sprintf (temp, "%02x", (uint8_t) dst_mac[5]);
  strncat (text, temp, TEXT_STRINGLEN - strnlen (text, TEXT_STRINGLEN - 1) - 1);

  // Print MAC address in textview
  textbuffer11 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview11));
  gtk_text_buffer_set_text (textbuffer11, text, -1);

  // Free allocated memory.
  free (text);
  free (temp);
  free (dst_mac);

  return (EXIT_SUCCESS);
}

// Someone has entered hostname or IPv6 address into IP-to-link-layer (MAC) address converter
int
on_entry138_changed (GtkWidget *entry138, SPSData *data) {

  (void) data;  // This statement suppresses compiler warning about unused parameter.
  (void) entry138;  // This statement suppresses compiler warning about unused parameter.

  return (EXIT_SUCCESS);
}

// Send neighbor solicitation.
int
on_button147_clicked (GtkButton *button147, SPSData *data) {

  (void) button147;  // This statement suppresses compiler warning about unused parameter.

  int i;
  const char *entry138_text, *entry140_text;
  GtkTextBuffer *textbuffer11;
  char *text, *temp;
  uint8_t *dst_mac;

  // Get IPv6 address from UI.
  entry138_text = gtk_entry_get_text (GTK_ENTRY (data->entry138));

  // Get interface name from UI.
  entry140_text = gtk_entry_get_text (GTK_ENTRY (data->entry140));

  // Array to hold various strings of text
  text = allocate_strmem (TEXT_STRINGLEN);
  temp = allocate_strmem (TEXT_STRINGLEN);

  // Array of unsigned chars
  dst_mac = allocate_ustrmem (6);

  textbuffer11 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview11));

  // Do neighbor discovery
  neighbor_discovery (dst_mac, entry138_text, entry140_text, data);

  // Put MAC address into a string.
  memset (text, 0, TEXT_STRINGLEN * sizeof (char));
  for (i=0; i<5; i++) {
    memset (temp, 0, TEXT_STRINGLEN * sizeof (char));
    sprintf (temp, "%02x:", (uint8_t) dst_mac[i]);
    strncat (text, temp, TEXT_STRINGLEN - strnlen (text, TEXT_STRINGLEN - 1) - 1);
  }
  memset (temp, 0, TEXT_STRINGLEN * sizeof (char));
  sprintf (temp, "%02x", (uint8_t) dst_mac[5]);
  strncat (text, temp, TEXT_STRINGLEN - strnlen (text, TEXT_STRINGLEN - 1) - 1);

  // Print MAC address in textview
  textbuffer11 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview11));
  gtk_text_buffer_set_text (textbuffer11, text, -1);

  // Free allocate memory.
  free (text);
  free (temp);
  free (dst_mac);

  return (EXIT_SUCCESS);
}

// Someone has requested the About page
int
on_about_requested () {

  GtkWidget *window, *textview, *scrolled_window;
  GtkTextBuffer *buffer;
  GdkPixbuf *about;
  GtkTextIter line;

  // Create new window to display About information.
  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_position (GTK_WINDOW (window), GTK_WIN_POS_CENTER);
  gtk_window_set_title (GTK_WINDOW (window), "About");
  gtk_container_set_border_width (GTK_CONTAINER (window), 1);
  gtk_window_set_default_size ((GtkWindow *) window, 1024, 768);

  // Create scrolled window and set its policy to automatic. i.e., only appear when needed
  scrolled_window = gtk_scrolled_window_new (NULL, NULL);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_window), GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);

  // Create a textview object to put image into.
  textview = gtk_text_view_new ();
  buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (textview));
  gtk_text_view_set_cursor_visible (GTK_TEXT_VIEW (textview), FALSE);

  // Load GdkPixbuf image "about" from file.
  about = gdk_pixbuf_new_from_file ("./about.png", NULL);

  // Pack GdkPixbuf into textview.
  gtk_text_buffer_get_iter_at_line (buffer, &line, 0);
  gtk_text_buffer_insert_pixbuf (buffer, &line, about);

  // Pack textview into scrolled window.
  gtk_container_add (GTK_CONTAINER (scrolled_window), textview);

  // Pack scrolled window into new window.
  gtk_container_add (GTK_CONTAINER (window), scrolled_window);

  // Display window.
  gtk_widget_show_all (window);

  return (EXIT_SUCCESS);
}

// Someone has requested the Documentation page {
int
on_documentation_activate (SPSData *data) {

  GtkWidget *window, *scrolled_win, *textview;
  int i, n, nbytes;
  uint8_t *filedata;
  GtkTextBuffer *buffer;
  FILE *fi;

  // Create new window to display documentation
  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_position (GTK_WINDOW (window), GTK_WIN_POS_CENTER);
  gtk_window_set_resizable (GTK_WINDOW (window), TRUE);
  gtk_window_set_title (GTK_WINDOW (window), "SPS Documentation");
  gtk_container_set_border_width (GTK_CONTAINER (window), 1);
  gtk_window_set_default_size (GTK_WINDOW (window), 900, 600);

  // Create a textview object to put image in
  textview = gtk_text_view_new ();
  buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (textview));
  gtk_text_view_set_cursor_visible (GTK_TEXT_VIEW (textview), TRUE);
  gtk_text_view_set_wrap_mode (GTK_TEXT_VIEW (textview), GTK_WRAP_WORD);

  // Open documentation file
  fi = fopen ("documentation.txt", "rb");
  if (fi == NULL) {
    memset (data->error_text, 0, TEXT_STRINGLEN * sizeof (char));
    strncpy (data->error_text, "Can't open file documentation.txt.", TEXT_STRINGLEN - 1);  // Minus 1 for string termination.
    data->parent = data->menubar1;
    report_error (data);
    return (EXIT_FAILURE);
  }

  // Count number of bytes in file.
  nbytes = 0;
  while ((n = fgetc (fi)) != EOF) {
    nbytes++;
  }
  rewind (fi);

  if (nbytes < 1) {
    memset (data->error_text, 0, TEXT_STRINGLEN * sizeof (char));
    strncpy (data->error_text, "File documentation.txt has 0 bytes.", TEXT_STRINGLEN - 1);  // Minus 1 for string termination.
    data->parent = data->menubar1;
    report_error (data);
    return (EXIT_FAILURE);
  } else {
    filedata = allocate_ustrmem (nbytes + 1);
  }

  // Read file into array.
  for (i=0; i<nbytes; i++) {
    filedata[i] = (uint8_t) fgetc (fi);
  }
  filedata[nbytes] = 0u;  // Ensure string termination.

  // Close file descriptor.
  fclose (fi);

  // Examine documentation file and report any non-UTF-8 characters.
  if (!is_valid_utf8 (filedata, nbytes, data->menubar1, data)) {
    free (filedata);
    return (EXIT_FAILURE);
  }

  // Put the text into the buffer.
  gtk_text_buffer_set_text (buffer, (const char *) filedata, -1);

  // Add scrolled window and textview to the new window and display.
  scrolled_win = gtk_scrolled_window_new (NULL, NULL);
  gtk_container_add (GTK_CONTAINER (scrolled_win), textview);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_win), GTK_POLICY_NEVER, GTK_POLICY_ALWAYS);
  gtk_container_add (GTK_CONTAINER (window), scrolled_win);
  gtk_widget_show_all (window);

  // Free allocated memory.
  free (filedata);

  return (EXIT_SUCCESS);
}

// Display TCP packet contents (IPv4)
int
on_button3_clicked (GtkButton *button3, SPSData *data) {

  int i;
  char **title;

  (void) button3;  // This statement suppresses compiler warning about unused parameter.

  title = allocate_strmemp (2);
  for (i=0; i<2; i++) {
    title[i] = allocate_strmem (100);
  }
  memcpy (title[0], "TCP (IPv4) Ethernet Frame Contents (fragmented)", 47 * sizeof (char));
  memcpy (title[1], "TCP (IPv4) Ethernet Frame Contents (unfragmented)", 49 * sizeof (char));

  show_packet (0, title, data);  // type, title, data

  // Free allocated memory.
  for (i=0; i<2; i++) {
    free (title[i]);
  }
  free (title);

  return (EXIT_SUCCESS);
}

// Display ICMP packet contents (IPv4)
int
on_button2_clicked (GtkButton *button2, SPSData *data) {

  int i;
  char **title;

  (void) button2;  // This statement suppresses compiler warning about unused parameter.

  title = allocate_strmemp (2);
  for (i=0; i<2; i++) {
    title[i] = allocate_strmem (100);
  }
  memcpy (title[0], "ICMP (IPv4) Ethernet Frame Contents (fragmented)", 48 * sizeof (char));
  memcpy (title[1], "ICMP (IPv4) Ethernet Frame Contents (unfragmented)", 50 * sizeof (char));

  show_packet (1, title, data);  // type, title, data

  // Free allocated memory.
  for (i=0; i<2; i++) {
    free (title[i]);
  }
  free (title);

  return (EXIT_SUCCESS);
}

// Display UDP packet contents (IPv4)
int
on_button4_clicked (GtkButton *button4, SPSData *data) {

  int i;
  char **title;

  (void) button4;  // This statement suppresses compiler warning about unused parameter.

  title = allocate_strmemp (2);
  for (i=0; i<2; i++) {
    title[i] = allocate_strmem (100);
  }
  memcpy (title[0], "UDP (IPv4) Ethernet Frame Contents (fragmented)", 47 * sizeof (char));
  memcpy (title[1], "UDP (IPv4) Ethernet Frame Contents (unfragmented)", 49 * sizeof (char));

  show_packet (2, title, data);  // type, title, data

  // Free allocated memory.
  for (i=0; i<2; i++) {
    free (title[i]);
  }
  free (title);

  return (EXIT_SUCCESS);
}

// Display TCP packet contents (IPv6)
int
on_button6_clicked (GtkButton *button6, SPSData *data) {

  int i;
  char **title;

  (void) button6;  // This statement suppresses compiler warning about unused parameter.

  title = allocate_strmemp (2);
  for (i=0; i<2; i++) {
    title[i] = allocate_strmem (100);
  }
  if (data->sixto4_flag) {
    memcpy (title[0], "TCP (IPv6 tunneled over IPv4) Ethernet Frame Contents (fragmented)", 66 * sizeof (char));
    memcpy (title[1], "TCP (IPv6 tunneled over IPv4) Ethernet Frame Contents (unfragmented)", 68 * sizeof (char));
    show_packet (6, title, data);  // type, title, data
  } else {
    memcpy (title[0], "TCP (IPv6) Ethernet Frame Contents (fragmented)", 47 * sizeof (char));
    memcpy (title[1], "TCP (IPv6) Ethernet Frame Contents (unfragmented)", 49 * sizeof (char));
    show_packet (3, title, data);  // type, title, data
  }

  // Free allocated memory.
  for (i=0; i<2; i++) {
    free (title[i]);
  }
  free (title);

  return (EXIT_SUCCESS);
}

// Display ICMP packet contents (IPv6)
int
on_button17_clicked (GtkButton *button17, SPSData *data) {

  int i;
  char **title;

  (void) button17;  // This statement suppresses compiler warning about unused parameter.

  title = allocate_strmemp (2);
  for (i=0; i<2; i++) {
    title[i] = allocate_strmem (100);
  }
  if (data->sixto4_flag) {
    memcpy (title[0], "ICMP (IPv6 tunneled over IPv4) Ethernet Frame Contents (fragmented)", 67 * sizeof (char));
    memcpy (title[1], "ICMP (IPv6 tunneled over IPv4) Ethernet Frame Contents (unfragmented)", 69 * sizeof (char));
    show_packet (7, title, data);  // type, title, data
  } else {
    strncpy (title[0], "ICMP (IPv6) Ethernet Frame Contents (fragmented)", 80 - 1);
    strncpy (title[1], "ICMP (IPv6) Ethernet Frame Contents (unfragmented)", 80 - 1);
    show_packet (4, title, data);  // type, title, data
  }

  // Free allocated memory.
  for (i=0; i<2; i++) {
    free (title[i]);
  }
  free (title);

  return (EXIT_SUCCESS);
}

// Display UDP packet contents (IPv6)
int
on_button13_clicked (GtkButton *button13, SPSData *data) {

  int i;
  char **title;

  (void) button13;  // This statement suppresses compiler warning about unused parameter.

  title = allocate_strmemp (2);
  for (i=0; i<2; i++) {
    title[i] = allocate_strmem (100);
  }
  if (data->sixto4_flag) {
    memcpy (title[0], "UDP (IPv6 tunneled over IPv4) Ethernet Frame Contents (fragmented)", 66 * sizeof (char));
    memcpy (title[1], "UDP (IPv6 tunneled over IPv4) Ethernet Frame Contents (unfragmented)", 68 * sizeof (char));
    show_packet (8, title, data);  // type, title, data
  } else {
    memcpy (title[0], "UDP (IPv6) Ethernet Frame Contents (fragmented)", 47 * sizeof (char));
    memcpy (title[1], "UDP (IPv6) Ethernet Frame Contents (unfragmented)", 49 * sizeof (char));
    show_packet (5, title, data);  // type, title, data
  }

  // Free allocated memory.
  for (i=0; i<2; i++) {
    free (title[i]);
  }
  free (title);

  return (EXIT_SUCCESS);
}

// Display TCP traceroute packet contents (IPv4)
int
on_tr_button28_clicked (GtkButton *tr_button28, SPSData *data) {

  int i;
  char **title;

  (void) tr_button28;  // This statement suppresses compiler warning about unused parameter.

  title = allocate_strmemp (2);
  for (i=0; i<2; i++) {
    title[i] = allocate_strmem (100);
  }
  memcpy (title[0], "TCP (IPv4) Ethernet Frame Contents for Traceroute (fragmented)", 62 * sizeof (char));
  memcpy (title[1], "TCP (IPv4) Ethernet Frame Contents for Traceroute (unfragmented)", 64 * sizeof (char));

  show_packet (9, title, data);  // type, title, data

  // Free allocated memory.
  for (i=0; i<2; i++) {
    free (title[i]);
  }
  free (title);

  return (EXIT_SUCCESS);
}

// Display ICMP traceroute packet contents (IPv4)
int
on_tr_button31_clicked (GtkButton *tr_button31, SPSData *data) {

  int i;
  char **title;

  (void) tr_button31;  // This statement suppresses compiler warning about unused parameter.

  title = allocate_strmemp (2);
  for (i=0; i<2; i++) {
    title[i] = allocate_strmem (100);
  }
  memcpy (title[0], "ICMP (IPv4) Ethernet Frame Contents for Traceroute (fragmented)", 63 * sizeof (char));
  memcpy (title[1], "ICMP (IPv4) Ethernet Frame Contents for Traceroute (unfragmented)", 65 * sizeof (char));

  show_packet (10, title, data);  // type, title, data

  // Free allocated memory.
  for (i=0; i<2; i++) {
    free (title[i]);
  }
  free (title);

  return (EXIT_SUCCESS);
}

// Display UDP traceroute packet contents (IPv4)
int
on_tr_button34_clicked (GtkButton *tr_button34, SPSData *data) {

  int i;
  char **title;

  (void) tr_button34;  // This statement suppresses compiler warning about unused parameter.

  title = allocate_strmemp (2);
  for (i=0; i<2; i++) {
    title[i] = allocate_strmem (100);
  }
  memcpy (title[0], "UDP (IPv4) Ethernet Frame Contents for Traceroute (fragmented)", 62 * sizeof (char));
  memcpy (title[1], "UDP (IPv4) Ethernet Frame Contents for Traceroute (unfragmented)", 64 * sizeof (char));

  show_packet (11, title, data);  // type, title, data

  // Free allocated memory.
  for (i=0; i<2; i++) {
    free (title[i]);
  }
  free (title);

  return (EXIT_SUCCESS);
}

// Display TCP traceroute packet contents (IPv6)
int
on_tr_button37_clicked (GtkButton *tr_button37, SPSData *data) {

  int i;
  char **title;

  (void) tr_button37;  // This statement suppresses compiler warning about unused parameter.

  title = allocate_strmemp (2);
  for (i=0; i<2; i++) {
    title[i] = allocate_strmem (100);
  }
  if (data->sixto4_tr_flag) {
    memcpy (title[0], "TCP (IPv6 tunneled over IPv4) Ethernet Frame Contents for Traceroute (fragmented)", 81 * sizeof (char));
    memcpy (title[1], "TCP (IPv6 tunneled over IPv4) Ethernet Frame Contents for Traceroute (unfragmented)", 83 * sizeof (char));
    show_packet (15, title, data);  // type, title, data
  } else {
    memcpy (title[0], "TCP (IPv6) Ethernet Frame Contents for Traceroute (fragmented)", 62 * sizeof (char));
    memcpy (title[1], "TCP (IPv6) Ethernet Frame Contents for Traceroute (unfragmented)", 64 * sizeof (char));
    show_packet (12, title, data);  // type, title, data
  }

  // Free allocated memory.
  for (i=0; i<2; i++) {
    free (title[i]);
  }
  free (title);

  return (EXIT_SUCCESS);
}

// Display ICMP traceroute packet contents (IPv6)
int
on_tr_button40_clicked (GtkButton *tr_button40, SPSData *data) {

  int i;
  char **title;

  (void) tr_button40;  // This statement suppresses compiler warning about unused parameter.

  title = allocate_strmemp (2);
  for (i=0; i<2; i++) {
    title[i] = allocate_strmem (80);
  }
  if (data->sixto4_tr_flag) {
    memcpy (title[0], "ICMP (IPv6 tunneled over IPv4) Ethernet Frame Contents for Traceroute (fragmented)", 82 * sizeof (char));
    memcpy (title[1], "ICMP (IPv6 tunneled over IPv4) Ethernet Frame Contents for Traceroute (unfragmented)", 84 * sizeof (char));
    show_packet (16, title, data);  // type, title, data
  } else {
    memcpy (title[0], "ICMP (IPv6) Ethernet Frame Contents for Traceroute (fragmented)", 63 * sizeof (char));
    memcpy (title[1], "ICMP (IPv6) Ethernet Frame Contents (unfragmented)", 50 * sizeof (char));
    show_packet (13, title, data);  // type, title, data
  }

  // Free allocated memory.
  for (i=0; i<2; i++) {
    free (title[i]);
  }
  free (title);

  return (EXIT_SUCCESS);
}

// Display UDP traceroute packet contents (IPv6)
int
on_tr_button43_clicked (GtkButton *tr_button43, SPSData *data) {

  int i;
  char **title;

  (void) tr_button43;  // This statement suppresses compiler warning about unused parameter.

  title = allocate_strmemp (2);
  for (i=0; i<2; i++) {
    title[i] = allocate_strmem (100);
  }
  if (data->sixto4_tr_flag) {
    memcpy (title[0], "UDP (IPv6 tunneled over IPv4) Ethernet Frame Contents for Traceroute (fragmented)", 81* sizeof (char));
    memcpy (title[1], "UDP (IPv6 tunneled over IPv4) Ethernet Frame Contents for Traceroute (unfragmented)", 83 * sizeof (char));
    show_packet (17, title, data);  // type, title, data
  } else {
    memcpy (title[0], "UDP (IPv6) Ethernet Frame Contents for Traceroute (fragmented)", 62 * sizeof (char));
    memcpy (title[1], "UDP (IPv6) Ethernet Frame Contents for Traceroute (unfragmented)", 64 * sizeof (char));
    show_packet (14, title, data);  // type, title, data
  }

  // Free allocated memory.
  for (i=0; i<2; i++) {
    free (title[i]);
  }
  free (title);

  return (EXIT_SUCCESS);
}

// TCP page (IPv4)

// Destination link-layer (MAC) address (48 bits)
int
on_entry144_changed (GtkWidget *entry144, SPSData *data) {

  mac_entry (entry144, data->ethhdr[0].dst_mac, 0, data);

  return (EXIT_SUCCESS);
}

// Source link-layer (MAC) address (48 bits)
int
on_entry145_changed (GtkWidget *entry145, SPSData *data) {

  mac_entry (entry145, data->ethhdr[0].src_mac, 0, data);

  return (EXIT_SUCCESS);
}

// This host's interface name to be converted to link-layer (MAC) address and find MTU.
int
on_entry150_changed (GtkWidget *entry150, SPSData *data) {

  // Check for valid source interface name, check MTU, and update source MAC address.
  ifname_entry (entry150, data->entry145, 0, data);

  // Update MTU spinbutton.
  update_mtu_spinbutton (0, data);

  return (EXIT_SUCCESS);
}

// Maximum transmission unit (MTU) - Changed via spinbutton + or -.
int
on_spinbutton5_value_changed (GtkWidget *spinbutton5, SPSData *data) {

  int value, min, ifmtu;

  // Set parent window for reporting error messages.
  data->parent = data->main_window;

  // Get requested MTU value from spinbutton.
  value = (int) gtk_spin_button_get_value (GTK_SPIN_BUTTON (spinbutton5));

  // Obtain interface's MTU.
  ifmtu = interface_getmtu (data->ifname[0]);
  if (ifmtu == 0) {
    update_mtu_spinbutton (0, data);  // Return spinbutton to current MTU value.
    return (EXIT_FAILURE);
  }

  // Determine minimum size needed to accommodate packet
  // when packet is fragmented to maximum extent.
  min = min4_mtu (0, data);
  
  // Ensure min required size is less than or equal to interface's MTU.
  if (min > ifmtu) {
    sprintf (data->error_text, "The interface's MTU is too small to accommodate packet.");
    report_error (data);
    data->ifmtu[0] = ifmtu;
    update_mtu_spinbutton (0, data);
    create_ip4_frame (0, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }
  
  // Ensure requested size is greater or equal to min required size to accommodate packet.
  if (value < min) {
    data->ifmtu[0] = min;
    update_mtu_spinbutton (0, data);
    create_ip4_frame (0, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }
  
  // Ensure requested size if <= interface's MTU.
  if (value > ifmtu) {
    data->ifmtu[0] = ifmtu;
    update_mtu_spinbutton (0, data);
    create_ip4_frame (0, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }
  
  // Accept requested MTU.
  data->ifmtu[0] = value;
  create_ip4_frame (0, data);  // Update ethernet frames.

  return (EXIT_SUCCESS);
}

// Maximum transmission unit (MTU) - Changed via spinbutton text entry.
int
on_spinbutton5_change_value (GtkWidget *spinbutton5, SPSData *data) {

  int value, min, ifmtu;

  // Set parent window for reporting error messages.
  data->parent = data->main_window;

  // Get requested MTU value from spinbutton text entry.
  value = (int) gtk_spin_button_get_value (GTK_SPIN_BUTTON (spinbutton5));

  // Obtain interface's MTU.
  ifmtu = interface_getmtu (data->ifname[0]);
  if (ifmtu == 0) {
    update_mtu_spinbutton (0, data);  // Return spinbutton to current MTU value.
    return (EXIT_FAILURE);
  }

  // Determine minimum size needed to accommodate packet
  // when packet is fragmented to maximum extent.
  min = min4_mtu (0, data);

  // Ensure min required size is less than or equal to interface's MTU.
  if (min > ifmtu) {
    sprintf (data->error_text, "The interface's MTU is too small to accommodate packet.");
    report_error (data);
    data->ifmtu[0] = ifmtu;
    update_mtu_spinbutton (0, data);
    create_ip4_frame (0, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size is greater or equal to min required size to accommodate packet.
  if (value < min) {
    sprintf (data->error_text, "The requested MTU is too small to accommodate packet.\nValue set to minimum required to accommodate packet instead.");
    report_error (data);
    data->ifmtu[0] = min;
    update_mtu_spinbutton (0, data);
    create_ip4_frame (0, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size if <= interface's MTU.
  if (value > ifmtu) {
    sprintf (data->error_text, "The requested MTU is larger than the interface's MTU.\nValue set to the interface's MTU instead.");
    report_error (data);
    data->ifmtu[0] = ifmtu;
    update_mtu_spinbutton (0, data);
    create_ip4_frame (0, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Accept requested MTU.
  data->ifmtu[0] = value;
  create_ip4_frame (0, data);  // Update ethernet frames.

  return (EXIT_SUCCESS);
}

// Ethernet type code (16 bits)
// http://www.iana.org/assignments/ethernet-numbers
int
on_entry146_changed (GtkWidget *entry146, SPSData *data) {

  ethtype_entry (entry146, 0, data);

  return (EXIT_SUCCESS);
}

// IPv4 header length (4 bits)
int
on_entry29_changed (GtkWidget *entry29, SPSData *data) {

  ip4_hdrlen_entry (entry29, 0, data);

  return (EXIT_SUCCESS);
}

// IP version (4 bits)
int
on_entry30_changed (GtkWidget *entry30, SPSData *data) {

  ip4_ver_entry (entry30, 0, data);

  return (EXIT_SUCCESS);
}

// Type of service (8 bits)
int
on_entry31_changed (GtkWidget *entry31, SPSData *data) {

  ip4_tos_entry (entry31, 0, data);

  return (EXIT_SUCCESS);
}

// Total length of datagram (16 bits)
int
on_entry32_changed (GtkWidget *entry32, SPSData *data) {

  ip4_tld_entry (entry32, 0, data);

  return (EXIT_SUCCESS);
}

// ID sequence number (16 bits)
int
on_entry33_changed (GtkWidget *entry33, SPSData *data) {

  ip4_seq_entry (entry33, 0, data);

  return (EXIT_SUCCESS);
}

// Zero (1 bit)
int
on_entry122_changed (GtkWidget *entry122, SPSData *data) {

  ip4_zero_entry (entry122, data->entry37, data->entry38, data->entry39, 0, data);

  return (EXIT_SUCCESS);
}

// Do not fragment flag (1 bit)
int
on_entry37_changed (GtkWidget *entry37, SPSData *data) {

  ip4_dnf_entry (data->entry122, entry37, data->entry38, data->entry39, 0, data);

  return (EXIT_SUCCESS);
}

// More fragments following flag (1 bit)
int
on_entry38_changed (GtkWidget *entry38, SPSData *data) {

  ip4_mff_entry (data->entry122, data->entry37, entry38, data->entry39, 0, data);

  return (EXIT_SUCCESS);
}

// Fragmentation offset (13 bits)
int
on_entry39_changed (GtkWidget *entry39, SPSData *data) {

  ip4_fo_entry (data->entry122, data->entry37, data->entry38, entry39, 0, data);

  return (EXIT_SUCCESS);
}

// Time-to-Live (8 bits)
int
on_entry40_changed (GtkWidget *entry40, SPSData *data) {

  ip4_ttl_entry (entry40, 0, data);

  return (EXIT_SUCCESS);
}

// Transport layer protocol (8 bits)
int
on_entry74_changed (GtkWidget *entry74, SPSData *data) {

  ip4_tlp_entry (entry74, 0, data);

  return (EXIT_SUCCESS);
}

// Source IPv4 address (32 bits)
int
on_entry35_changed (GtkWidget *entry35, SPSData *data) {

  ip4_sip_entry (entry35, 0, data);

  if (data->ran_tcp4_sourceip) {
    // Source IP address has been entered manually. Uncheck "Randomize" checkbutton.
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton5), FALSE);
    data->ran_tcp4_sourceip = 0;
  }

  return (EXIT_SUCCESS);
}

// Randomize source IPv4 address
int
on_checkbutton5_toggled (GtkButton *checkbutton5, SPSData *data) {

  (void) checkbutton5;  // This statement suppresses compiler warning about unused parameter.

  if (data->ran_tcp4_sourceip) {
    data->ran_tcp4_sourceip = 0;
  } else {
    ip4_ransip_checkbutton (&data->ran_tcp4_sourceip, data->entry35, 0, data);
  }

  return (EXIT_SUCCESS);
}

// Destination IPv4 address (32 bits)
int
on_entry36_changed (GtkWidget *entry36, SPSData *data) {

  ip4_dip_entry (entry36, 0, data);

  return (EXIT_SUCCESS);
}

// IPv4 checksum (16 bits)
int
on_entry34_changed (GtkWidget *entry34, SPSData *data) {

  ip4_chksum_entry (entry34, 0, data);

  return (EXIT_SUCCESS);
}

// IP options for IPv4 TCP

// Decimal option data entry
int
on_radiobutton24_toggled (GtkButton *radiobutton24, SPSData *data) {

  (void) radiobutton24;  // This statement suppresses compiler warning about unused parameter.

  if (data->dec_hex_ipopt_tcp4) {
    data->dec_hex_ipopt_tcp4 = 0;
  } else {
    data->dec_hex_ipopt_tcp4 = 1;
  }

  return (EXIT_SUCCESS);
}

// Hexadecimal option data entry
int
on_radiobutton25_toggled (GtkButton *radiobutton25) {

  (void) radiobutton25;  // This statement suppresses compiler warning about unused parameter.

  // on_radiobutton24_toggled() will change value of data->dec_hex_ipopt_tcp4.

  return (EXIT_SUCCESS);
}

// IPv4 option data
int
on_entry400_changed (GtkWidget *entry400, SPSData *data) {

  data_entry (entry400, &data->ip_optlenbuf[0], MAX_IP4OPTLEN, data->ip_optionsbuf[0], data->dec_hex_ipopt_tcp4, data->main_window, data);

  return (EXIT_SUCCESS);
}

// Insert IPv4 option after another, as specified by user.
int
on_button26_clicked (GtkButton *button26, SPSData *data) {

  (void) button26;  // This statement suppresses compiler warning about unused parameter.

  ip4_option (0, 0, data->ipopt_tcp4_after, data->textview30, data);

  return (EXIT_SUCCESS);
}

// Number of option after which to insert new IPv4 option.
int
on_entry401_changed (GtkWidget *entry401, SPSData *data) {

  ip4_opt_insert_after_entry (entry401, 0, &data->ipopt_tcp4_after, data);

  return (EXIT_SUCCESS);
}

// Append an IPv4 option.
int
on_button60_clicked (GtkButton *button60, SPSData *data) {

  (void) button60;  // This statement suppresses compiler warning about unused parameter.

  ip4_option (2, 0, 0, data->textview30, data);

  return (EXIT_SUCCESS);
}

// View IPv4 header options for TCP.
int
on_button61_clicked (GtkButton *button61, SPSData *data) {

  (void) button61;  // This statement suppresses compiler warning about unused parameter.

  opt_view (0, 0, data);

  return (EXIT_SUCCESS);
}

// Number of IPv4 option to remove.
int
on_entry403_changed (GtkWidget *entry403, SPSData *data) {

  ip4_opt_remove_entry (entry403, 0, &data->ipopt_tcp4_remove, data);

  return (EXIT_SUCCESS);
}

// Remove selected IPv4 option.
int
on_button62_clicked (GtkButton *button62, SPSData *data) {

  (void) button62;  // This statement suppresses compiler warning about unused parameter.

  ip4_option (1, 0, data->ipopt_tcp4_remove, data->textview30, data);

  // Clear entry with option number to remove.
  gtk_entry_set_text (GTK_ENTRY (data->entry403), "");

  return (EXIT_SUCCESS);
}

// TCP header (IPv4)

// TCP source port (16 bits)
int
on_entry41_changed (GtkWidget *entry41, SPSData *data) {

  tcp_sport_entry (entry41, 0, data);

  if (data->ran_tcp4_sourceport) {
    // Source port has been entered manually. Uncheck "Randomize" checkbutton.
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton7), FALSE);
    data->ran_tcp4_sourceport = 0;
  }

  return (EXIT_SUCCESS);
}

// Randomize TCP source port number.
int
on_checkbutton7_toggled (GtkButton *checkbutton7, SPSData *data) {

  (void) checkbutton7;  // This statement suppresses compiler warning about unused parameter.

  if (data->ran_tcp4_sourceport) {
    data->ran_tcp4_sourceport = 0;
  } else {
    tcp_ran_sport (&data->ran_tcp4_sourceport, 0, data->entry41, data);
  }

  return (EXIT_SUCCESS);
}

// TCP destination port (16 bits)
int
on_entry42_changed (GtkWidget *entry42, SPSData *data) {

  tcp_dport_entry (entry42, 0, data);

  return (EXIT_SUCCESS);
}

// TCP sequence number (32 bits)
int
on_entry43_changed (GtkWidget *entry43, SPSData *data) {

  tcp_seqnum_entry (entry43, 0, data);

  return (EXIT_SUCCESS);
}

// TCP acknowledgement number (32 bits)
int
on_entry44_changed (GtkWidget *entry44, SPSData *data) {

  tcp_acknum_entry (entry44, 0, data);

  return (EXIT_SUCCESS);
}

// TCP reserved (4 bits)
int
on_entry45_changed (GtkWidget *entry45, SPSData *data) {

  tcp_res_entry (entry45, 0, data);

  return (EXIT_SUCCESS);
}

// TCP data offset (4 bits)
int
on_entry46_changed (GtkWidget *entry46, SPSData *data) {

  tcp_off_entry (entry46, 0, data);

  return (EXIT_SUCCESS);
}

// TCP FIN flag (1 bit)
int
on_entry47_changed (GtkWidget *entry47, SPSData *data) {

  tcp_fin_entry (entry47, 0, data);

  return (EXIT_SUCCESS);
}

// TCP SYN flag (1 bit)
int
on_entry48_changed (GtkWidget *entry48, SPSData *data) {

  tcp_syn_entry (entry48, 0, data);

  return (EXIT_SUCCESS);
}

// TCP RST flag (1 bit)
int
on_entry49_changed (GtkWidget *entry49, SPSData *data) {

  tcp_rst_entry (entry49, 0, data);

  return (EXIT_SUCCESS);
}

// TCP PSH flag (1 bit)
int
on_entry50_changed (GtkWidget *entry50, SPSData *data) {

  tcp_psh_entry (entry50, 0, data);

  return (EXIT_SUCCESS);
}

// TCP ACK flag (1 bit)
int
on_entry51_changed (GtkWidget *entry51, SPSData *data) {

  tcp_ack_entry (entry51, 0, data);

  return (EXIT_SUCCESS);
}

// TCP URG flag (1 bit)
int
on_entry52_changed (GtkWidget *entry52, SPSData *data) {

  tcp_urg_entry (entry52, 0, data);

  return (EXIT_SUCCESS);
}

// TCP ECE flag (1 bit)
int
on_entry53_changed (GtkWidget *entry53, SPSData *data) {

  tcp_ece_entry (entry53, 0, data);

  return (EXIT_SUCCESS);
}

// TCP CWR flag (1 bit)
int
on_entry54_changed (GtkWidget *entry54, SPSData *data) {

  tcp_cwr_entry (entry54, 0, data);

  return (EXIT_SUCCESS);
}

// TCP window size (16 bits)
int
on_entry55_changed (GtkWidget *entry55, SPSData *data) {

  tcp_win_entry (entry55, 0, data);

  return (EXIT_SUCCESS);
}

// TCP checksum (16 bits)
int
on_entry56_changed (GtkWidget *entry56, SPSData *data) {

  tcp_chksum_entry (entry56, 0, data);

  return (EXIT_SUCCESS);
}

// TCP urgent pointer (16 bits)
int
on_entry77_changed (GtkWidget *entry77, SPSData *data) {

  tcp_urgptr_entry (entry77, 0, data);

  return (EXIT_SUCCESS);
}

// TCP options (IPv4)

// Decimal option data entry
int
on_radiobutton41_toggled (GtkButton *radiobutton41, SPSData *data) {

  (void) radiobutton41;  // This statement suppresses compiler warning about unused parameter.

  if (data->dec_hex_tcpopt_tcp4) {
    data->dec_hex_tcpopt_tcp4 = 0;
  } else {
    data->dec_hex_tcpopt_tcp4 = 1;
  }

  return (EXIT_SUCCESS);
}

// Hexadecimal option data entry
int
on_radiobutton42_toggled (GtkButton *radiobutton42) {

  (void) radiobutton42;  // This statement suppresses compiler warning about unused parameter.

  // on_radiobutton41_toggled() will change value of data->dec_hex_tcpopt_tcp4.

  return (EXIT_SUCCESS);
}

// TCP option data (IPv4)
int
on_entry404_changed (GtkWidget *entry404, SPSData *data) {

  data_entry (entry404, &data->tcp_optlenbuf[0], MAX_TCPOPTLEN, data->tcp_optionsbuf[0], data->dec_hex_tcpopt_tcp4, data->main_window, data);

  return (EXIT_SUCCESS);
}

// Insert TCP option after another, as specified by user.
int
on_button47_clicked (GtkButton *button47, SPSData *data) {

  (void) button47;  // This statement suppresses compiler warning about unused parameter.

  tcp4_option (0, 0, data->tcpopt_tcp4_after, data->textview24, data);

  return (EXIT_SUCCESS);
}

// Number of option after which to insert new TCP option.
int
on_entry405_changed (GtkWidget *entry405, SPSData *data) {

  tcp_opt_insert_after_entry (entry405, 0, &data->tcpopt_tcp4_after, data);

  return (EXIT_SUCCESS);
}

// Append TCP option for IPv4 TCP packet.
int
on_button48_clicked (GtkButton *button48, SPSData *data) {

  (void) button48;  // This statement suppresses compiler warning about unused parameter.

  tcp4_option (2, 0, 0, data->textview24, data);

  return (EXIT_SUCCESS);
}

// View TCP options for IPv4 TCP.
int
on_button49_clicked (GtkButton *button49, SPSData *data) {

  (void) button49;  // This statement suppresses compiler warning about unused parameter.

  opt_view (0, 1, data);

  return (EXIT_SUCCESS);
}

// Number of TCP option to remove.
int
on_entry406_changed (GtkWidget *entry406, SPSData *data) {

  tcp_opt_remove_entry (entry406, 0, &data->tcpopt_tcp4_remove, data);

  return (EXIT_SUCCESS);
}

// Remove selected TCP option.
int
on_button58_clicked (GtkButton *button58, SPSData *data) {

  (void) button58;

  tcp4_option (1, 0, data->tcpopt_tcp4_remove, data->textview24, data);

  // Clear entry with option number to remove.
  gtk_entry_set_text (GTK_ENTRY (data->entry406), "");

  return (EXIT_SUCCESS);
}

// TCP payload (IPv4)

// ASCII data entry
int
on_radiobutton17_toggled (GtkButton *radiobutton17, SPSData *data) {

  (void) radiobutton17;  // This statement suppresses compiler warning about unused parameter.

  if (data->ascii_hex_tcp4) {
    data->ascii_hex_tcp4 = 0;
  } else {
    data->ascii_hex_tcp4 = 1;
  }

  return (EXIT_SUCCESS);
}

// Hexadecimal data entry
int
on_radiobutton18_toggled (GtkButton *radiobutton18) {

  (void) radiobutton18;  // This statement suppresses compiler warning about unused parameter.

  // on_radiobutton17_toggled() will change value of data->ascii_hex_tcp4.

  return (EXIT_SUCCESS);
}

// TCP data from keyboard (IPv4)
int
on_entry162_changed (GtkWidget *entry162, SPSData *data) {

  key_data (entry162, 0, data->ascii_hex_tcp4, data);

  if (data->nframes[0] > 1) {
    sprintf (data->warning_text, "Packet will be fragmented to accommodate TCP data.");
    data->parent = data->main_window;
    report_warning (data);
  }

  return (EXIT_SUCCESS);
}

// TCP data from file (IPv4)
int
on_button20_clicked (GtkButton *button20, SPSData *data) {

  (void) button20;  // This statement suppresses compiler warning about unused parameter.

  file_payload_data (data->entry162, 0, data);

  if (data->nframes[0] > 1) {
    sprintf (data->warning_text, "Packet will be fragmented to accommodate TCP data.");
    data->parent = data->main_window;
    report_warning (data);
  }

  return (EXIT_SUCCESS);
}

// Unload file data for TCP (IPv4)
int
on_button200_clicked (GtkButton *button200, SPSData *data) {

  (void) button200;  // This statement suppresses compiler warning about unused parameter.
  GtkTextBuffer *textbuffer60;

  // Remove file data.
  data->payloadlen[0] = 0;
  memset (data->payload[0], 0, IP_MAXPACKET * sizeof (uint8_t));

  // Update ethernet frame.
  create_ip4_frame (0, data);

  textbuffer60 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview60));
  gtk_text_buffer_set_text (textbuffer60, "", -1);

  return (EXIT_SUCCESS);
}

// TCP page (IPv6)

// Destination link-layer (MAC) address (48 bits)
int
on_entry131_changed (GtkWidget *entry131, SPSData *data) {

  mac_entry (entry131, data->ethhdr[3].dst_mac, 3, data);

  return (EXIT_SUCCESS);
}

// Source link-layer (MAC) address (48 bits)
int
on_entry132_changed (GtkWidget *entry132, SPSData *data) {

  mac_entry (entry132, data->ethhdr[3].src_mac, 3, data);

  return (EXIT_SUCCESS);
}

// This host's interface name to be converted to link-layer (MAC) address and find MTU.
int
on_entry154_changed (GtkWidget *entry154, SPSData *data) {

  // Check for valid source interface name, check MTU, and update source MAC address.
  ifname_entry (entry154, data->entry132, 3, data);

  // Update MTU spinbutton.
  update_mtu_spinbutton (3, data);

  return (EXIT_SUCCESS);
}

// Maximum transmission unit (MTU) - Changed via spinbutton + or -.
int
on_spinbutton8_value_changed (GtkWidget *spinbutton8, SPSData *data) {

  int value, min, ifmtu;

  // Set parent window for reporting error messages.
  data->parent = data->main_window;

  // Get requested MTU value from spinbutton.
  value = (int) gtk_spin_button_get_value (GTK_SPIN_BUTTON (spinbutton8));

  // Obtain interface's MTU.
  ifmtu = interface_getmtu (data->ifname[3]);
  if (ifmtu == 0) {
    update_mtu_spinbutton (3, data);  // Return spinbutton to current MTU value.
    return (EXIT_FAILURE);
  }

  // Determine minimum size needed to accommodate packet
  // when packet is fragmented to maximum extent.
  min = min6_mtu (3, data);

  // Ensure min required size is less than or equal to interface's MTU.
  if (min > ifmtu) {
    sprintf (data->error_text, "The interface's MTU is too small to accommodate packet.");
    report_error (data);
    data->ifmtu[3] = ifmtu;
    update_mtu_spinbutton (3, data);
    create_ip6_frame (3, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size is greater or equal to min required size to accommodate packet.
  if (value < min) {
    data->ifmtu[3] = min;
    update_mtu_spinbutton (3, data);
    create_ip6_frame (3, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size if <= interface's MTU.
  if (value > ifmtu) {
    data->ifmtu[3] = ifmtu;
    update_mtu_spinbutton (3, data);
    create_ip6_frame (3, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Accept requested MTU.
  data->ifmtu[3] = value;
  create_ip6_frame (3, data);  // Update ethernet frames.

  return (EXIT_SUCCESS);
}

// Maximum transmission unit (MTU) - Changed via spinbutton text entry.
int
on_spinbutton8_change_value (GtkWidget *spinbutton8, SPSData *data) {

  int value, min, ifmtu;

  // Set parent window for reporting error messages.
  data->parent = data->main_window;

  // Get requested MTU value from spinbutton text entry.
  value = (int) gtk_spin_button_get_value (GTK_SPIN_BUTTON (spinbutton8));

  // Obtain interface's MTU.
  ifmtu = interface_getmtu (data->ifname[3]);
  if (ifmtu == 0) {
    update_mtu_spinbutton (3, data);  // Return spinbutton to current MTU value.
    return (EXIT_FAILURE);
  }

  // Determine minimum size needed to accommodate packet
  // when packet is fragmented to maximum extent.
  min = min6_mtu (3, data);

  // Ensure min required size is less than or equal to interface's MTU.
  if (min > ifmtu) {
    sprintf (data->error_text, "The interface's MTU is too small to accommodate packet.");
    report_error (data);
    data->ifmtu[3] = ifmtu;
    update_mtu_spinbutton (3, data);
    create_ip6_frame (3, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size is greater or equal to min required size to accommodate packet.
  if (value < min) {
    sprintf (data->error_text, "The requested MTU is too small to accommodate packet.\nValue set to minimum required to accommodate packet instead.");
    report_error (data);
    data->ifmtu[3] = min;
    update_mtu_spinbutton (3, data);
    create_ip6_frame (3, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size if <= interface's MTU.
  if (value > ifmtu) {
    sprintf (data->error_text, "The requested MTU is larger than the interface's MTU.\nValue set to the interface's MTU instead.");
    report_error (data);
    data->ifmtu[3] = ifmtu;
    update_mtu_spinbutton (3, data);
    create_ip6_frame (3, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Accept requested MTU.
  data->ifmtu[3] = value;
  create_ip6_frame (3, data);  // Update ethernet frames.

  return (EXIT_SUCCESS);
}

// Ethernet type code (16 bits)
// http://www.iana.org/assignments/ethernet-numbers
int
on_entry133_changed (GtkWidget *entry133, SPSData *data) {

  ethtype_entry (entry133, 3, data);

  return (EXIT_SUCCESS);
}

// IP version (4 bits)
int
on_entry22_changed (GtkWidget *entry22, SPSData *data) {

  ip6_ver_entry (entry22, 3, data);

  return (EXIT_SUCCESS);
}

// Traffic class (8 bits)
int
on_entry23_changed (GtkWidget *entry23, SPSData *data) {

  ip6_tc_entry (entry23, 3, data);

  return (EXIT_SUCCESS);
}

// Flow label (20 bits)
int
on_entry24_changed (GtkWidget *entry24, SPSData *data) {

  ip6_fl_entry (entry24, 3, data);

  return (EXIT_SUCCESS);
}

// Payload length (16 bits)
int
on_entry25_changed (GtkWidget *entry25, SPSData *data) {

  ip6_plen_entry (entry25, 3, data);

  return (EXIT_SUCCESS);
}

// Next header (8 bits)
int
on_entry26_changed (GtkWidget *entry26, SPSData *data) {

  ip6_nxt_entry (entry26, 3, data);

  return (EXIT_SUCCESS);
}

// Hop limit (8 bits)
int
on_entry27_changed (GtkWidget *entry27, SPSData *data) {

  ip6_hops_entry (entry27, 3, data);

  return (EXIT_SUCCESS);
}

// Source IPv6 address (128 bits)
int
on_entry28_changed (GtkWidget *entry28, SPSData *data) {

  ip6_sip_entry (entry28, 3, data);

  if (data->ran_tcp6_sourceip) {
    // Source IP address has been entered manually. Uncheck "Randomize" checkbutton.
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton2), FALSE);
    data->ran_tcp6_sourceip = 0;
  }

  return (EXIT_SUCCESS);
}

// Randomize source IPv6 address
int
on_checkbutton2_toggled (GtkButton *checkbutton2, SPSData *data) {

  (void) checkbutton2;  // This statement suppresses compiler warning about unused parameter.

  if (data->ran_tcp6_sourceip) {
    data->ran_tcp6_sourceip = 0;
  } else {
    ip6_ransip_checkbutton (&data->ran_tcp6_sourceip, data->entry28, 3, data);
  }

  return (EXIT_SUCCESS);
}

// Destination IP address (128 bits)
int
on_entry78_changed (GtkWidget *entry78, SPSData *data) {

  ip6_dip_entry (entry78, 3, data);

  return (EXIT_SUCCESS);
}

// Hop-by-Hop (original/inner) extension header
int
on_button98_clicked (GtkWidget *button98, SPSData *data) {

  (void) button98;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 3;
  hop_win (data);

  return (EXIT_SUCCESS);
}

// Destination extension header
int
on_button110_clicked (GtkWidget *button110, SPSData *data) {

  (void) button110;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 3;
  dst_win (data);

  return (EXIT_SUCCESS);
}

// Routing extension header
int
on_button132_clicked (GtkWidget *button132, SPSData *data) {

  (void) button132;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 3;
  route_win (data);

  return (EXIT_SUCCESS);
}

// Authentication extension header
int
on_button116_clicked (GtkWidget *button116, SPSData *data) {

  (void) button116;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 3;
  auth_win (data);

  return (EXIT_SUCCESS);
}

// ESP extension header
int
on_button126_clicked (GtkWidget *button126, SPSData *data) {

  (void) button126;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 3;
  esp_win (data);

  return (EXIT_SUCCESS);
}

// Outer hop-by-Hop extension header
int
on_button300_clicked (GtkWidget *button300, SPSData *data) {

  (void) button300;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 3;
  hop_win_outer (data);

  return (EXIT_SUCCESS);
}

// Outer destination extension header
int
on_button301_clicked (GtkWidget *button301, SPSData *data) {

  (void) button301;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 3;
  dst_win_outer (data);

  return (EXIT_SUCCESS);
}

// Outer routing extension header
int
on_button302_clicked (GtkWidget *button302, SPSData *data) {

  (void) button302;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 3;
  route_win_outer (data);

  return (EXIT_SUCCESS);
}

// TCP header (IPv6)

// TCP source port (16 bits)
int
on_entry79_changed (GtkWidget *entry79, SPSData *data) {

  tcp_sport_entry (entry79, 3, data);

  if (data->ran_tcp6_sourceport) {
    // Source port has been entered manually. Uncheck "Randomize" checkbutton.
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton4), FALSE);
    data->ran_tcp6_sourceport = 0;
  }

  return (EXIT_SUCCESS);
}

// Randomize TCP source port number.
int
on_checkbutton4_toggled (GtkButton *checkbutton4, SPSData *data) {

  (void) checkbutton4;  // This statement suppresses compiler warning about unused parameter.

  if (data->ran_tcp6_sourceport) {
    data->ran_tcp6_sourceport = 0;
  } else {
    tcp_ran_sport (&data->ran_tcp6_sourceport, 3, data->entry79, data);
  }

  return (EXIT_SUCCESS);
}

// TCP destination port (16 bits)
int
on_entry80_changed (GtkWidget *entry80, SPSData *data) {

  tcp_dport_entry (entry80, 3, data);

  return (EXIT_SUCCESS);
}

// TCP sequence number (32 bits)
int
on_entry81_changed (GtkWidget *entry81, SPSData *data) {

  tcp_seqnum_entry (entry81, 3, data);

  return (EXIT_SUCCESS);
}

// TCP acknowledgement number (32 bits)
int
on_entry82_changed (GtkWidget *entry82, SPSData *data) {

  tcp_acknum_entry (entry82, 3, data);

  return (EXIT_SUCCESS);
}

// TCP reserved (4 bits)
int
on_entry83_changed (GtkWidget *entry83, SPSData *data) {

  tcp_res_entry (entry83, 3, data);

  return (EXIT_SUCCESS);
}

// TCP data offset (4 bits)
int
on_entry84_changed (GtkWidget *entry84, SPSData *data) {

  tcp_off_entry (entry84, 3, data);

  return (EXIT_SUCCESS);
}

// TCP FIN flag (1 bit)
int
on_entry85_changed (GtkWidget *entry85, SPSData *data) {

  tcp_fin_entry (entry85, 3, data);

  return (EXIT_SUCCESS);
}

// TCP SYN flag (1 bit)
int
on_entry86_changed (GtkWidget *entry86, SPSData *data) {

  tcp_syn_entry (entry86, 3, data);

  return (EXIT_SUCCESS);
}

// TCP RST flag (1 bit)
int
on_entry87_changed (GtkWidget *entry87, SPSData *data) {

  tcp_rst_entry (entry87, 3, data);

  return (EXIT_SUCCESS);
}

// TCP PSH flag (1 bit)
int
on_entry88_changed (GtkWidget *entry88, SPSData *data) {

  tcp_psh_entry (entry88, 3, data);

  return (EXIT_SUCCESS);
}

// TCP ACK flag (1 bit)
int
on_entry89_changed (GtkWidget *entry89, SPSData *data) {

  tcp_ack_entry (entry89, 3, data);

  return (EXIT_SUCCESS);
}

// TCP URG flag (1 bit)
int
on_entry90_changed (GtkWidget *entry90, SPSData *data) {

  tcp_urg_entry (entry90, 3, data);

  return (EXIT_SUCCESS);
}

// TCP ECE flag (1 bit)
int
on_entry91_changed (GtkWidget *entry91, SPSData *data) {

  tcp_ece_entry (entry91, 3, data);

  return (EXIT_SUCCESS);
}

// TCP CWR flag (1 bit)
int
on_entry92_changed (GtkWidget *entry92, SPSData *data) {

  tcp_cwr_entry (entry92, 3, data);

  return (EXIT_SUCCESS);
}

// TCP window size (16 bits)
int
on_entry93_changed (GtkWidget *entry93, SPSData *data) {

  tcp_win_entry (entry93, 3, data);

  return (EXIT_SUCCESS);
}

// TCP checksum (16 bits)
int
on_entry94_changed (GtkWidget *entry94, SPSData *data) {

  tcp_chksum_entry (entry94, 3, data);

  return (EXIT_SUCCESS);
}

// TCP urgent pointer (16 bits)
int
on_entry95_changed (GtkWidget *entry95, SPSData *data) {

  tcp_urgptr_entry (entry95, 3, data);

  return (EXIT_SUCCESS);
}

// TCP options (IPv6)

// Decimal option data entry
int
on_radiobutton47_toggled (GtkButton *radiobutton47, SPSData *data) {

  (void) radiobutton47;  // This statement suppresses compiler warning about unused parameter.

  if (data->dec_hex_tcpopt_tcp6) {
    data->dec_hex_tcpopt_tcp6 = 0;
  } else {
    data->dec_hex_tcpopt_tcp6 = 1;
  }

  return (EXIT_SUCCESS);
}

// Hexadecimal option data entry
int
on_radiobutton48_toggled (GtkButton *radiobutton48) {

  (void) radiobutton48;  // This statement suppresses compiler warning about unused parameter.

  // on_radiobutton47_toggled() will change value of data->dec_hex_tcpopt_tcp6.

  return (EXIT_SUCCESS);
}

// TCP option data (IPv6)
int
on_entry413_changed (GtkWidget *entry413, SPSData *data) {

  data_entry (entry413, &data->tcp_optlenbuf[3], MAX_TCPOPTLEN, data->tcp_optionsbuf[3], data->dec_hex_tcpopt_tcp6, data->main_window, data);

  return (EXIT_SUCCESS);
}

// Insert TCP option after another, as specified by user.
int
on_button70_clicked (GtkButton *button70, SPSData *data) {

  (void) button70;  // This statement suppresses compiler warning about unused parameter.

  tcp6_option (0, 3, data->tcpopt_tcp6_after, data->textview32, data);

  return (EXIT_SUCCESS);
}

// Number of option after which to insert new TCP option.
int
on_entry414_changed (GtkWidget *entry414, SPSData *data) {

  tcp_opt_insert_after_entry (entry414, 3, &data->tcpopt_tcp6_after, data);

  return (EXIT_SUCCESS);
}

// Append TCP option for IPv6 TCP packet.
int
on_button71_clicked (GtkButton *button71, SPSData *data) {

  (void) button71;  // This statement suppresses compiler warning about unused parameter.

  tcp6_option (2, 3, 0, data->textview32, data);

  return (EXIT_SUCCESS);
}

// View TCP options for IPv6 TCP.
int
on_button72_clicked (GtkButton *button72, SPSData *data) {

  (void) button72;  // This statement suppresses compiler warning about unused parameter.

  opt_view (3, 1, data);

  return (EXIT_SUCCESS);
}

// Number of TCP option to remove.
int
on_entry415_changed (GtkWidget *entry415, SPSData *data) {

  tcp_opt_remove_entry (entry415, 3, &data->tcpopt_tcp6_remove, data);

  return (EXIT_SUCCESS);
}

// Remove selected TCP option.
int
on_button73_clicked (GtkButton *button73, SPSData *data) {

  (void) button73;  // This statement suppresses compiler warning about unused parameter.

  tcp6_option (1, 3, data->tcpopt_tcp6_remove, data->textview32, data);

  // Clear entry with option number to remove.
  gtk_entry_set_text (GTK_ENTRY (data->entry415), "");

  return (EXIT_SUCCESS);
}

// TCP payload (IPv6)

// ASCII data entry
int
on_radiobutton19_toggled (GtkButton *radiobutton19, SPSData *data) {

  (void) radiobutton19;  // This statement suppresses compiler warning about unused parameter.

  if (data->ascii_hex_tcp6) {
    data->ascii_hex_tcp6 = 0;
  } else {
    data->ascii_hex_tcp6 = 1;
  }

  return (EXIT_SUCCESS);
}

// Hexadecimal data entry
int
on_radiobutton20_toggled (GtkButton *radiobutton20) {

  (void) radiobutton20;  // This statement suppresses compiler warning about unused parameter.

  // on_radiobutton19_toggled() will change value of data->ascii_hex_tcp6.

  return (EXIT_SUCCESS);
}

// TCP data from keyboard (IPv6)
int
on_entry163_changed (GtkWidget *entry163, SPSData *data) {

  key_data (entry163, 3, data->ascii_hex_tcp6, data);

  if (data->nframes[3] > 1) {
    sprintf (data->warning_text, "Packet will be fragmented to accommodate TCP data.");
    data->parent = data->main_window;
    report_warning (data);
  }

  // Update ethernet frame and extension header editing window widgets, if visible.
  ip6data_update (3, data);

  return (EXIT_SUCCESS);
}

// TCP data from file (IPv6)
int
on_button22_clicked (GtkButton *button22, SPSData *data) {

  (void) button22;  // This statement suppresses compiler warning about unused parameter.

  if (file_payload_data (data->entry163, 3, data) == EXIT_SUCCESS) {

    if (data->nframes[3] > 1) {
      sprintf (data->warning_text, "Packet will be fragmented to accommodate TCP data.");
      data->parent = data->main_window;
      report_warning (data);
    }

    // Update ethernet frame and extension header editing window widgets, if visible.
    ip6data_update (3, data);
  }

  return (EXIT_SUCCESS);
}

// Unload file data for TCP (IPv6)
int
on_button201_clicked (GtkButton *button201, SPSData *data) {

  (void) button201;  // This statement suppresses compiler warning about unused parameter.
  GtkTextBuffer *textbuffer61;

  // Remove file data.
  data->payloadlen[3] = 0;
  memset (data->payload[3], 0, IP_MAXPACKET * sizeof (uint8_t));

  // Update ethernet frames.
  create_ip6_frame (3, data);
  create_6to4_frame (6, data);

  textbuffer61 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview61));
  gtk_text_buffer_set_text (textbuffer61, "", -1);

  return (EXIT_SUCCESS);
}

// ICMP page (IPv4)

// Destination link-layer (MAC) address (48 bits)
int
on_entry141_changed (GtkWidget *entry141, SPSData *data) {

  mac_entry (entry141, data->ethhdr[1].dst_mac, 1, data);

  return (EXIT_SUCCESS);
}

// Source link-layer (MAC) address (48 bits)
int
on_entry142_changed (GtkWidget *entry142, SPSData *data) {

  mac_entry (entry142, data->ethhdr[1].src_mac, 1, data);

  return (EXIT_SUCCESS);
}

// This host's interface name to be converted to link-layer (MAC) address and find MTU.
int
on_entry151_changed (GtkWidget *entry151, SPSData *data) {

  // Check for valid source interface name, check MTU, and update source MAC address.
  ifname_entry (entry151, data->entry142, 1, data);

  // Update MTU spinbutton.
  update_mtu_spinbutton (1, data);

  return (EXIT_SUCCESS);
}

// Maximum transmission unit (MTU) - Changed via spinbutton + or -.
int
on_spinbutton6_value_changed (GtkWidget *spinbutton6, SPSData *data) {

  int value, min, ifmtu;

  // Set parent window for reporting error messages.
  data->parent = data->main_window;

  // Get requested MTU value from spinbutton.
  value = (int) gtk_spin_button_get_value (GTK_SPIN_BUTTON (spinbutton6));

  // Obtain interface's MTU.
  ifmtu = interface_getmtu (data->ifname[1]);
  if (ifmtu == 0) {
    update_mtu_spinbutton (1, data);  // Return spinbutton to current MTU value.
    return (EXIT_FAILURE);
  }

  // Determine minimum size needed to accommodate packet
  // when packet is fragmented to maximum extent.
  min = min4_mtu (1, data);

  // Ensure min required size is less than or equal to interface's MTU.
  if (min > ifmtu) {
    sprintf (data->error_text, "The interface's MTU is too small to accommodate packet.");
    report_error (data);
    data->ifmtu[1] = ifmtu;
    update_mtu_spinbutton (1, data);
    create_ip4_frame (1, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size is greater or equal to min required size to accommodate packet.
  if (value < min) {
    data->ifmtu[1] = min;
    update_mtu_spinbutton (1, data);
    create_ip4_frame (1, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size if <= interface's MTU.
  if (value > ifmtu) {
    data->ifmtu[1] = ifmtu;
    update_mtu_spinbutton (1, data);
    create_ip4_frame (1, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Accept requested MTU.
  data->ifmtu[1] = value;
  create_ip4_frame (1, data);  // Update ethernet frames.

  return (EXIT_SUCCESS);
}

// Maximum transmission unit (MTU) - Changed via spinbutton text entry.
int
on_spinbutton6_change_value (GtkWidget *spinbutton6, SPSData *data) {

  int value, min, ifmtu;

  // Set parent window for reporting error messages.
  data->parent = data->main_window;

  // Get requested MTU value from spinbutton text entry.
  value = (int) gtk_spin_button_get_value (GTK_SPIN_BUTTON (spinbutton6));

  // Obtain interface's MTU.
  ifmtu = interface_getmtu (data->ifname[1]);
  if (ifmtu == 0) {
    update_mtu_spinbutton (1, data);  // Return spinbutton to current MTU value.
    return (EXIT_FAILURE);
  }

  // Determine minimum size needed to accommodate packet
  // when packet is fragmented to maximum extent.
  min = min4_mtu (1, data);

  // Ensure min required size is less than or equal to interface's MTU.
  if (min > ifmtu) {
    sprintf (data->error_text, "The interface's MTU is too small to accommodate packet.");
    report_error (data);
    data->ifmtu[1] = ifmtu;
    update_mtu_spinbutton (1, data);
    create_ip4_frame (1, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size is greater or equal to min required size to accommodate packet.
  if (value < min) {
    sprintf (data->error_text, "The requested MTU is too small to accommodate packet.\nValue set to minimum required to accommodate packet instead.");
    report_error (data);
    data->ifmtu[1] = min;
    update_mtu_spinbutton (1, data);
    create_ip4_frame (1, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size if <= interface's MTU.
  if (value > ifmtu) {
    sprintf (data->error_text, "The requested MTU is larger than the interface's MTU.\nValue set to the interface's MTU instead.");
    report_error (data);
    data->ifmtu[1] = ifmtu;
    update_mtu_spinbutton (1, data);
    create_ip4_frame (1, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Accept requested MTU.
  data->ifmtu[1] = value;
  create_ip4_frame (1, data);  // Update ethernet frames.

  return (EXIT_SUCCESS);
}

// Ethernet type code (16 bits)
// http://www.iana.org/assignments/ethernet-numbers
int
on_entry143_changed (GtkWidget *entry143, SPSData *data) {

  ethtype_entry (entry143, 1, data);

  return (EXIT_SUCCESS);
}

// IPv4 header length (4 bits)
int
on_entry57_changed (GtkWidget *entry57, SPSData *data) {

  ip4_hdrlen_entry (entry57, 1, data);

  return (EXIT_SUCCESS);
}

// IP version (4 bits)
int
on_entry58_changed (GtkWidget *entry58, SPSData *data) {

  ip4_ver_entry (entry58, 1, data);

  return (EXIT_SUCCESS);
}

// Type of service (8 bits)
int
on_entry59_changed (GtkWidget *entry59, SPSData *data) {

  ip4_tos_entry (entry59, 1, data);

  return (EXIT_SUCCESS);
}

// Total length of datagram (16 bits)
int
on_entry60_changed (GtkWidget *entry60, SPSData *data) {

  ip4_tld_entry (entry60, 1, data);

  return (EXIT_SUCCESS);
}

// ID sequence number (16 bits)
int
on_entry61_changed (GtkWidget *entry61, SPSData *data) {

  ip4_seq_entry (entry61, 1, data);

  return (EXIT_SUCCESS);
}

// Zero (1 bit)
int
on_entry123_changed (GtkWidget *entry123, SPSData *data) {

  ip4_zero_entry (entry123, data->entry65, data->entry66, data->entry67, 1, data);

  return (EXIT_SUCCESS);
}

// Do not fragment flag (1 bit)
int
on_entry65_changed (GtkWidget *entry65, SPSData *data) {

  ip4_dnf_entry (data->entry123, entry65, data->entry66, data->entry67, 1, data);

  return (EXIT_SUCCESS);
}

// More fragments following flag (1 bit)
int
on_entry66_changed (GtkWidget *entry66, SPSData *data) {

  ip4_mff_entry (data->entry123, data->entry65, entry66, data->entry67, 1, data);

  return (EXIT_SUCCESS);
}

// Fragmentation offset (13 bits)
int
on_entry67_changed (GtkWidget *entry67, SPSData *data) {

  ip4_fo_entry (data->entry123, data->entry65, data->entry66, entry67, 1, data);

  return (EXIT_SUCCESS);
}

// Time-to-Live (8 bits)
int
on_entry68_changed (GtkWidget *entry68, SPSData *data) {

  ip4_ttl_entry (entry68, 1, data);

  return (EXIT_SUCCESS);
}

// Transport layer protocol (8 bits)
int
on_entry75_changed (GtkWidget *entry75, SPSData *data) {

  ip4_tlp_entry (entry75, 1, data);

  return (EXIT_SUCCESS);
}

// Source IPv4 address (32 bits)
int
on_entry63_changed (GtkWidget *entry63, SPSData *data) {

  ip4_sip_entry (entry63, 1, data);

  if (data->ran_icmp4_sourceip) {
    // Source IP address has been entered manually. Uncheck "Randomize" checkbutton.
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton9), FALSE);
    data->ran_icmp4_sourceip = 0;
  }

  return (EXIT_SUCCESS);
}

// Randomize source IPv4 address
int
on_checkbutton9_toggled (GtkButton *checkbutton9, SPSData *data) {

  (void) checkbutton9;  // This statement suppresses compiler warning about unused parameter.

  if (data->ran_icmp4_sourceip) {
    data->ran_icmp4_sourceip = 0;
  } else {
    ip4_ransip_checkbutton (&data->ran_icmp4_sourceip, data->entry63, 1, data);
  }

  return (EXIT_SUCCESS);
}

// Destination IPv4 address (32 bits)
int
on_entry64_changed (GtkWidget *entry64, SPSData *data) {

  ip4_dip_entry (entry64, 1, data);

  return (EXIT_SUCCESS);
}

// IPv4 checksum (16 bits)
int
on_entry62_changed (GtkWidget *entry62, SPSData *data) {

  ip4_chksum_entry (entry62, 1, data);

  return (EXIT_SUCCESS);
}

// IP options for IPv4 ICMP 

// Decimal option data entry
int
on_radiobutton43_toggled (GtkButton *radiobutton43, SPSData *data) {

  (void) radiobutton43;  // This statement suppresses compiler warning about unused parameter.

  if (data->dec_hex_ipopt_icmp4) {
    data->dec_hex_ipopt_icmp4 = 0;
  } else {
    data->dec_hex_ipopt_icmp4 = 1;
  }

  return (EXIT_SUCCESS);
}

// Hexadecimal option data entry
int
on_radiobutton44_toggled (GtkButton *radiobutton44) {

  (void) radiobutton44;  // This statement suppresses compiler warning about unused parameter.

  // on_radiobutton43_toggled() will change value of data->dec_hex_ipopt_icmp4.

  return (EXIT_SUCCESS);
}

// IPv4 option data
int
on_entry407_changed (GtkWidget *entry407, SPSData *data) {

  data_entry (entry407, &data->ip_optlenbuf[1], MAX_IP4OPTLEN, data->ip_optionsbuf[1], data->dec_hex_ipopt_icmp4, data->main_window, data);

  return (EXIT_SUCCESS);
}

// Insert IPv4 option after another, as specified by user.
int
on_button59_clicked (GtkButton *button59, SPSData *data) {

  (void) button59;  // This statement suppresses compiler warning about unused parameter.

  ip4_option (0, 1, data->ipopt_icmp4_after, data->textview26, data);

  return (EXIT_SUCCESS);
}

// Number of option after which to insert new IPv4 option.
int
on_entry408_changed (GtkWidget *entry408, SPSData *data) {

  ip4_opt_insert_after_entry (entry408, 1, &data->ipopt_icmp4_after, data);

  return (EXIT_SUCCESS);
}

// Append IP option for IPv4 ICMP packet.
int
on_button63_clicked (GtkButton *button63, SPSData *data) {

  (void) button63;  // This statement suppresses compiler warning about unused parameter.

  ip4_option (2, 1, 0, data->textview26, data);

  return (EXIT_SUCCESS);
}

// View IPv4 header options for ICMP.
int
on_button64_clicked (GtkButton *button64, SPSData *data) {

  (void) button64;  // This statement suppresses compiler warning about unused parameter.

  opt_view (1, 0, data);

  return (EXIT_SUCCESS);
}

// Number of IPv4 option to remove.
int
on_entry409_changed (GtkWidget *entry409, SPSData *data) {

  ip4_opt_remove_entry (entry409, 1, &data->ipopt_icmp4_remove, data);

  return (EXIT_SUCCESS);
}

// Remove selected IPv4 option.
int
on_button65_clicked (GtkButton *button65, SPSData *data) {

  (void) button65;  // This statement suppresses compiler warning about unused parameter.

  ip4_option (1, 1, data->ipopt_icmp4_remove, data->textview26, data);

  // Clear entry with option number to remove.
  gtk_entry_set_text (GTK_ENTRY (data->entry409), "");

  return (EXIT_SUCCESS);
}

// ICMP header (IPv4)

// Message Type (8 bits)
int
on_entry69_changed (GtkWidget *entry69, SPSData *data) {

  icmp_mt_entry (entry69, 1, data);

  return (EXIT_SUCCESS);
}

// Message Code (8 bits)
int
on_entry70_changed (GtkWidget *entry70, SPSData *data) {

  icmp_mc_entry (entry70, 1, data);

  return (EXIT_SUCCESS);
}

// ICMP checksum (16 bits)
int
on_entry71_changed (GtkWidget *entry71, SPSData *data) {

  icmp_chksum_entry (entry71, 1, data);

  return (EXIT_SUCCESS);
}

// ID (16 bits)
int
on_entry72_changed (GtkWidget *entry72, SPSData *data) {

  icmp_id_entry (entry72, 1, data);

  return (EXIT_SUCCESS);
}

// Sequence number (16 bits)
int
on_entry73_changed (GtkWidget *entry73, SPSData *data) {

  icmp_seq_entry (entry73, 1, data);

  return (EXIT_SUCCESS);
}

// ASCII data entry
int
on_radiobutton9_toggled (GtkButton *radiobutton9, SPSData *data) {

  (void) radiobutton9;  // This statement suppresses compiler warning about unused parameter.

  if (data->ascii_hex_icmp4) {
    data->ascii_hex_icmp4 = 0;
  } else {
    data->ascii_hex_icmp4 = 1;
  }

  return (EXIT_SUCCESS);
}

// Hexadecimal data entry
int
on_radiobutton10_toggled (GtkButton *radiobutton10) {

  (void) radiobutton10;  // This statement suppresses compiler warning about unused parameter.

  // on_radiobutton9_toggled() will change value of data->ascii_hex_icmp4.

  return (EXIT_SUCCESS);
}

// ICMP data from keyboard (IPv4)
int
on_entry126_changed (GtkWidget *entry126, SPSData *data) {

  key_data (entry126, 1, data->ascii_hex_icmp4, data);

  if (data->nframes[1] > 1) {
    sprintf (data->warning_text, "Packet will be fragmented to accommodate ICMP data.");
    data->parent = data->main_window;
    report_warning (data);
  }

  return (EXIT_SUCCESS);
}

// ICMP data from file (IPv4)
int
on_button7_clicked (GtkButton *button7, SPSData *data) {

  (void) button7;  // This statement suppresses compiler warning about unused parameter.

  file_payload_data (data->entry126, 1, data);

  if (data->nframes[1] > 1) {
    sprintf (data->warning_text, "Packet will be fragmented to accommodate ICMP data.");
    data->parent = data->main_window;
    report_warning (data);
  }

  return (EXIT_SUCCESS);
}

// Unload file data for ICMP (IPv4)
int
on_button202_clicked (GtkButton *button202, SPSData *data) {

  (void) button202;  // This statement suppresses compiler warning about unused parameter.
  GtkTextBuffer *textbuffer62;

  // Remove file data.
  data->payloadlen[1] = 0;
  memset (data->payload[1], 0, IP_MAXPACKET * sizeof (uint8_t));

  // Update ethernet frame.
  create_ip4_frame (1, data);

  textbuffer62 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview62));
  gtk_text_buffer_set_text (textbuffer62, "", -1);

  return (EXIT_SUCCESS);
}

// ICMP page (IPv6)

// Destination link-layer (MAC) address (48 bits)
int
on_entry128_changed (GtkWidget *entry128, SPSData *data) {

  mac_entry (entry128, data->ethhdr[4].dst_mac, 4, data);

  return (EXIT_SUCCESS);
}

// Source link-layer (MAC) address (48 bits)
int
on_entry129_changed (GtkWidget *entry129, SPSData *data) {

  mac_entry (entry129, data->ethhdr[4].src_mac, 4, data);

  return (EXIT_SUCCESS);
}

// This host's interface name to be converted to link-layer (MAC) address and find MTU.
int
on_entry153_changed (GtkWidget *entry153, SPSData *data) {

  // Check for valid source interface name, check MTU, and update source MAC address.
  ifname_entry (entry153, data->entry129, 4, data);

  // Update MTU spinbutton.
  update_mtu_spinbutton (4, data);

  return (EXIT_SUCCESS);
}

// Maximum transmission unit (MTU) - Changed via spinbutton + or -.
int
on_spinbutton9_value_changed (GtkWidget *spinbutton9, SPSData *data) {

  int value, min, ifmtu;

  // Set parent window for reporting error messages.
  data->parent = data->main_window;

  // Get requested MTU value from spinbutton.
  value = (int) gtk_spin_button_get_value (GTK_SPIN_BUTTON (spinbutton9));

  // Obtain interface's MTU.
  ifmtu = interface_getmtu (data->ifname[4]);
  if (ifmtu == 0) {
    update_mtu_spinbutton (4, data);  // Return spinbutton to current MTU value.
    return (EXIT_FAILURE);
  }

  // Determine minimum size needed to accommodate packet
  // when packet is fragmented to maximum extent.
  min = min6_mtu (4, data);

  // Ensure min required size is less than or equal to interface's MTU.
  if (min > ifmtu) {
    sprintf (data->error_text, "The interface's MTU is too small to accommodate packet.");
    report_error (data);
    data->ifmtu[4] = ifmtu;
    update_mtu_spinbutton (4, data);
    create_ip6_frame (4, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size is greater or equal to min required size to accommodate packet.
  if (value < min) {
    data->ifmtu[4] = min;
    update_mtu_spinbutton (4, data);
    create_ip6_frame (4, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size if <= interface's MTU.
  if (value > ifmtu) {
    data->ifmtu[4] = ifmtu;
    update_mtu_spinbutton (4, data);
    create_ip6_frame (4, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Accept requested MTU.
  data->ifmtu[4] = value;
  create_ip6_frame (4, data);  // Update ethernet frames.

  return (EXIT_SUCCESS);
}

// Maximum transmission unit (MTU) - Changed via spinbutton text entry.
int
on_spinbutton9_change_value (GtkWidget *spinbutton9, SPSData *data) {

  int value, min, ifmtu;

  // Set parent window for reporting error messages.
  data->parent = data->main_window;

  // Get requested MTU value from spinbutton text entry.
  value = (int) gtk_spin_button_get_value (GTK_SPIN_BUTTON (spinbutton9));

  // Obtain interface's MTU.
  ifmtu = interface_getmtu (data->ifname[4]);
  if (ifmtu == 0) {
    update_mtu_spinbutton (4, data);  // Return spinbutton to current MTU value.
    return (EXIT_FAILURE);
  }

  // Determine minimum size needed to accommodate packet
  // when packet is fragmented to maximum extent.
  min = min6_mtu (4, data);

  // Ensure min required size is less than or equal to interface's MTU.
  if (min > ifmtu) {
    sprintf (data->error_text, "The interface's MTU is too small to accommodate packet.");
    report_error (data);
    data->ifmtu[4] = ifmtu;
    update_mtu_spinbutton (4, data);
    create_ip6_frame (4, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size is greater or equal to min required size to accommodate packet.
  if (value < min) {
    sprintf (data->error_text, "The requested MTU is too small to accommodate packet.\nValue set to minimum required to accommodate packet instead.");
    report_error (data);
    data->ifmtu[4] = min;
    update_mtu_spinbutton (4, data);
    create_ip6_frame (4, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size if <= interface's MTU.
  if (value > ifmtu) {
    sprintf (data->error_text, "The requested MTU is larger than the interface's MTU.\nValue set to the interface's MTU instead.");
    report_error (data);
    data->ifmtu[4] = ifmtu;
    update_mtu_spinbutton (4, data);
    create_ip6_frame (4, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Accept requested MTU.
  data->ifmtu[4] = value;
  create_ip6_frame (4, data);  // Update ethernet frames.

  return (EXIT_SUCCESS);
}

// Ethernet type code (16 bits)
// http://www.iana.org/assignments/ethernet-numbers
int
on_entry130_changed (GtkWidget *entry130, SPSData *data) {

  ethtype_entry (entry130, 4, data);

  return (EXIT_SUCCESS);
}

// IP version (4 bits)
int
on_entry110_changed (GtkWidget *entry110, SPSData *data) {

  ip6_ver_entry (entry110, 4, data);

  return (EXIT_SUCCESS);
}

// Traffic class (8 bits)
int
on_entry111_changed (GtkWidget *entry111, SPSData *data) {

  ip6_tc_entry (entry111, 4, data);

  return (EXIT_SUCCESS);
}

// Flow label (20 bits)
int
on_entry112_changed (GtkWidget *entry112, SPSData *data) {

  ip6_fl_entry (entry112, 4, data);

  return (EXIT_SUCCESS);
}

// Payload length (16 bits)
int
on_entry113_changed (GtkWidget *entry113, SPSData *data) {

  ip6_plen_entry (entry113, 4, data);

  return (EXIT_SUCCESS);
}

// Next header (8 bits)
int
on_entry114_changed (GtkWidget *entry114, SPSData *data) {

  ip6_nxt_entry (entry114, 4, data);

  return (EXIT_SUCCESS);
}

// Hop limit (8 bits)
int
on_entry115_changed (GtkWidget *entry115, SPSData *data) {

  ip6_hops_entry (entry115, 4, data);

  return (EXIT_SUCCESS);
}

// Source IPv6 address (128 bits)
int
on_entry109_changed (GtkWidget *entry109, SPSData *data) {

  ip6_sip_entry (entry109, 4, data);

  if (data->ran_icmp6_sourceip) {
    // Source IP address has been entered manually. Uncheck "Randomize" checkbutton.
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton10), FALSE);
    data->ran_icmp6_sourceip = 0;
  }

  return (EXIT_SUCCESS);
}

// Randomize source IPv6 address
int
on_checkbutton10_toggled (GtkButton *checkbutton10, SPSData *data) {

  (void) checkbutton10;  // This statement suppresses compiler warning about unused parameter.

  if (data->ran_icmp6_sourceip) {
    data->ran_icmp6_sourceip = 0;
  } else {
    ip6_ransip_checkbutton (&data->ran_icmp6_sourceip, data->entry109, 4, data);
  }

  return (EXIT_SUCCESS);
}

// Destination IP address (128 bits)
int
on_entry116_changed (GtkWidget *entry116, SPSData *data) {

  ip6_dip_entry (entry116, 4, data);

  return (EXIT_SUCCESS);
}

// Hop-by-Hop extension header
int
on_button99_clicked (GtkWidget *button99, SPSData *data) {

  (void) button99;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 4;
  hop_win (data);

  return (EXIT_SUCCESS);
}

// Destination extension header
int
on_button111_clicked (GtkWidget *button111, SPSData *data) {

  (void) button111;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 4;
  dst_win (data);

  return (EXIT_SUCCESS);
}

// Routing extension header
int
on_button133_clicked (GtkWidget *button133, SPSData *data) {

  (void) button133;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 4;
  route_win (data);

  return (EXIT_SUCCESS);
}

// Authentication extension header
int
on_button117_clicked (GtkWidget *button117, SPSData *data) {

  (void) button117;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 4;
  auth_win (data);

  return (EXIT_SUCCESS);
}

// ESP extension header
int
on_button127_clicked (GtkWidget *button127, SPSData *data) {

  (void) button127;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 4;
  esp_win (data);

  return (EXIT_SUCCESS);
}

// Outer hop-by-Hop extension header
int
on_button303_clicked (GtkWidget *button303, SPSData *data) {

  (void) button303;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 4;
  hop_win_outer (data);

  return (EXIT_SUCCESS);
}

// Outer destination extension header
int
on_button304_clicked (GtkWidget *button304, SPSData *data) {

  (void) button304;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 4;
  dst_win_outer (data);

  return (EXIT_SUCCESS);
}

// Outer routing extension header
int
on_button305_clicked (GtkWidget *button305, SPSData *data) {

  (void) button305;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 4;
  route_win_outer (data);

  return (EXIT_SUCCESS);
}

// ICMP header (IPv6)

// Message Type (8 bits)
int
on_entry117_changed (GtkWidget *entry117, SPSData *data) {

  icmp_mt_entry (entry117, 4, data);

  return (EXIT_SUCCESS);
}

// Message Code (8 bits)
int
on_entry118_changed (GtkWidget *entry118, SPSData *data) {

  icmp_mc_entry (entry118, 4, data);

  return (EXIT_SUCCESS);
}

// ICMP checksum (16 bits)
int
on_entry119_changed (GtkWidget *entry119, SPSData *data) {

  icmp_chksum_entry (entry119, 4, data);

  return (EXIT_SUCCESS);
}

// ID (16 bits)
int
on_entry120_changed (GtkWidget *entry120, SPSData *data) {

  icmp_id_entry (entry120, 4, data);

  return (EXIT_SUCCESS);
}

// Sequence number (16 bits)
int
on_entry121_changed (GtkWidget *entry121, SPSData *data) {

  icmp_seq_entry (entry121, 4, data);

  return (EXIT_SUCCESS);
}

// ASCII data entry
int
on_radiobutton11_toggled (GtkButton *radiobutton11, SPSData *data) {

  (void) radiobutton11;  // This statement suppresses compiler warning about unused parameter.

  if (data->ascii_hex_icmp6) {
    data->ascii_hex_icmp6 = 0;
  } else {
    data->ascii_hex_icmp6 = 1;
  }

  return (EXIT_SUCCESS);
}

// Hexadecimal data entry
int
on_radiobutton12_toggled (GtkButton *radiobutton12) {

  (void) radiobutton12;  // This statement suppresses compiler warning about unused parameter.

  // on_radiobutton11_toggled() will change value of data->ascii_hex_icmp6.

  return (EXIT_SUCCESS);
}

// ICMP data from keyboard (IPv6)
int
on_entry127_changed (GtkWidget *entry127, SPSData *data) {

  key_data (entry127, 4, data->ascii_hex_icmp6, data);

  if (data->nframes[4] > 1) {
    sprintf (data->warning_text, "Packet will be fragmented to accommodate ICMP data.");
    data->parent = data->main_window;
    report_warning (data);
  }

  // Update ethernet frame and extension header editing window widgets, if visible.
  ip6data_update (4, data);

  return (EXIT_SUCCESS);
}

// ICMP data from file (IPv6)
int
on_button9_clicked (GtkButton *button9, SPSData *data) {

  (void) button9;  // This statement suppresses compiler warning about unused parameter.

  if (file_payload_data (data->entry127, 4, data) == EXIT_SUCCESS) {

    if (data->nframes[4] > 1) {
      sprintf (data->warning_text, "Packet will be fragmented to accommodate ICMP data.");
      data->parent = data->main_window;
      report_warning (data);
    }

    // Update ethernet frame and extension header editing window widgets, if visible.
    ip6data_update (4, data);
  }

  return (EXIT_SUCCESS);
}

// Unload file data for ICMP (IPv6)
int
on_button203_clicked (GtkButton *button203, SPSData *data) {

  (void) button203;  // This statement suppresses compiler warning about unused parameter.
  GtkTextBuffer *textbuffer63;

  // Remove file data.
  data->payloadlen[4] = 0;
  memset (data->payload[4], 0, IP_MAXPACKET * sizeof (uint8_t));

  // Update ethernet frames.
  create_ip6_frame (4, data);
  create_6to4_frame (7, data);

  textbuffer63 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview63));
  gtk_text_buffer_set_text (textbuffer63, "", -1);

  return (EXIT_SUCCESS);
}

// UDP page (IPv4)

// Destination link-layer (MAC) address (48 bits)
int
on_entry147_changed (GtkWidget *entry147, SPSData *data) {

  mac_entry (entry147, data->ethhdr[2].dst_mac, 2, data);

  return (EXIT_SUCCESS);
}

// Source link-layer (MAC) address (48 bits)
int
on_entry148_changed (GtkWidget *entry148, SPSData *data) {

  mac_entry (entry148, data->ethhdr[2].src_mac, 2, data);

  return (EXIT_SUCCESS);
}

// This host's interface name to be converted to link-layer (MAC) address and find MTU.
int
on_entry152_changed (GtkWidget *entry152, SPSData *data) {

  // Check for valid source interface name, check MTU, and update source MAC address.
  ifname_entry (entry152, data->entry148, 2, data);

  // Update MTU spinbutton.
  update_mtu_spinbutton (2, data);

  return (EXIT_SUCCESS);
}

// Maximum transmission unit (MTU) - Changed via spinbutton + or -.
int
on_spinbutton7_value_changed (GtkWidget *spinbutton7, SPSData *data) {

  int value, min, ifmtu;

  // Set parent window for reporting error messages.
  data->parent = data->main_window;

  // Get requested MTU value from spinbutton.
  value = (int) gtk_spin_button_get_value (GTK_SPIN_BUTTON (spinbutton7));

  // Obtain interface's MTU.
  ifmtu = interface_getmtu (data->ifname[2]);
  if (ifmtu == 0) {
    update_mtu_spinbutton (2, data);  // Return spinbutton to current MTU value.
    return (EXIT_FAILURE);
  }

  // Determine minimum size needed to accommodate packet
  // when packet is fragmented to maximum extent.
  min = min4_mtu (2, data);

  // Ensure min required size is less than or equal to interface's MTU.
  if (min > ifmtu) {
    sprintf (data->error_text, "The interface's MTU is too small to accommodate packet.");
    report_error (data);
    data->ifmtu[2] = ifmtu;
    update_mtu_spinbutton (2, data);
    create_ip4_frame (2, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size is greater or equal to min required size to accommodate packet.
  if (value < min) {
    data->ifmtu[2] = min;
    update_mtu_spinbutton (2, data);
    create_ip4_frame (2, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size if <= interface's MTU.
  if (value > ifmtu) {
    data->ifmtu[2] = ifmtu;
    update_mtu_spinbutton (2, data);
    create_ip4_frame (2, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Accept requested MTU.
  data->ifmtu[2] = value;
  create_ip4_frame (2, data);  // Update ethernet frames.

  return (EXIT_SUCCESS);
}

// Maximum transmission unit (MTU) - Changed via spinbutton text entry.
int
on_spinbutton7_change_value (GtkWidget *spinbutton7, SPSData *data) {

  int value, min, ifmtu;

  // Set parent window for reporting error messages.
  data->parent = data->main_window;

  // Get requested MTU value from spinbutton text entry.
  value = (int) gtk_spin_button_get_value (GTK_SPIN_BUTTON (spinbutton7));

  // Obtain interface's MTU.
  ifmtu = interface_getmtu (data->ifname[2]);
  if (ifmtu == 0) {
    update_mtu_spinbutton (2, data);  // Return spinbutton to current MTU value.
    return (EXIT_FAILURE);
  }

  // Determine minimum size needed to accommodate packet
  // when packet is fragmented to maximum extent.
  min = min4_mtu (2, data);

  // Ensure min required size is less than or equal to interface's MTU.
  if (min > ifmtu) {
    sprintf (data->error_text, "The interface's MTU is too small to accommodate packet.");
    report_error (data);
    data->ifmtu[2] = ifmtu;
    update_mtu_spinbutton (2, data);
    create_ip4_frame (2, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size is greater or equal to min required size to accommodate packet.
  if (value < min) {
    sprintf (data->error_text, "The requested MTU is too small to accommodate packet.\nValue set to minimum required to accommodate packet instead.");
    report_error (data);
    data->ifmtu[2] = min;
    update_mtu_spinbutton (2, data);
    create_ip4_frame (2, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size if <= interface's MTU.
  if (value > ifmtu) {
    sprintf (data->error_text, "The requested MTU is larger than the interface's MTU.\nValue set to the interface's MTU instead.");
    report_error (data);
    data->ifmtu[2] = ifmtu;
    update_mtu_spinbutton (2, data);
    create_ip4_frame (2, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Accept requested MTU.
  data->ifmtu[2] = value;
  create_ip4_frame (2, data);  // Update ethernet frames.

  return (EXIT_SUCCESS);
}

// Ethernet type code (16 bits)
// http://www.iana.org/assignments/ethernet-numbers
int
on_entry149_changed (GtkWidget *entry149, SPSData *data) {

  ethtype_entry (entry149, 2, data);

  return (EXIT_SUCCESS);
}

// IPv4 header length (4 bits)
int
on_entry1_changed (GtkWidget *entry1, SPSData *data) {

  ip4_hdrlen_entry (entry1, 2, data);

  return (EXIT_SUCCESS);
}

// IP version (4 bits)
int
on_entry2_changed (GtkWidget *entry2, SPSData *data) {

  ip4_ver_entry (entry2, 2, data);

  return (EXIT_SUCCESS);
}

// Type of service (8 bits)
int
on_entry3_changed (GtkWidget *entry3, SPSData *data) {

  ip4_tos_entry (entry3, 2, data);

  return (EXIT_SUCCESS);
}

// Total length of datagram (16 bits)
int
on_entry4_changed (GtkWidget *entry4, SPSData *data) {

  ip4_tld_entry (entry4, 2, data);

  return (EXIT_SUCCESS);
}

// ID sequence number (16 bits)
int
on_entry5_changed (GtkWidget *entry5, SPSData *data) {

  ip4_seq_entry (entry5, 2, data);

  return (EXIT_SUCCESS);
}

// Zero (1 bit)
int
on_entry124_changed (GtkWidget *entry124, SPSData *data) {

  ip4_zero_entry (entry124, data->entry9, data->entry10, data->entry11, 2, data);

  return (EXIT_SUCCESS);
}

// Do not fragment flag (1 bit)
int
on_entry9_changed (GtkWidget *entry9, SPSData *data) {

  ip4_dnf_entry (data->entry124, entry9, data->entry10, data->entry11, 2, data);

  return (EXIT_SUCCESS);
}

// More fragments following flag (1 bit)
int
on_entry10_changed (GtkWidget *entry10, SPSData *data) {

  ip4_mff_entry (data->entry124, data->entry9, entry10, data->entry11, 2, data);

  return (EXIT_SUCCESS);
}

// Fragmentation offset (13 bits)
int
on_entry11_changed (GtkWidget *entry11, SPSData *data) {

  ip4_fo_entry (data->entry124, data->entry9, data->entry10, entry11, 2, data);

  return (EXIT_SUCCESS);
}

// Time-to-Live (8 bits)
int
on_entry12_changed (GtkWidget *entry12, SPSData *data) {

  ip4_ttl_entry (entry12, 2, data);

  return (EXIT_SUCCESS);
}

// Transport layer protocol (8 bits)
int
on_entry76_changed (GtkWidget *entry76, SPSData *data) {

  ip4_tlp_entry (entry76, 2, data);

  return (EXIT_SUCCESS);
}

// Source IPv4 address (32 bits)
int
on_entry7_changed (GtkWidget *entry7, SPSData *data) {

  ip4_sip_entry (entry7, 2, data);

  if (data->ran_udp4_sourceip) {
    // Source IP address has been entered manually. Uncheck "Randomize" checkbutton.
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton1), FALSE);
    data->ran_udp4_sourceip = 0;
  }

  return (EXIT_SUCCESS);
}

// Randomize source IPv4 address
int
on_checkbutton1_toggled (GtkButton *checkbutton1, SPSData *data) {

  (void) checkbutton1;  // This statement suppresses compiler warning about unused parameter.

  if (data->ran_udp4_sourceip) {
    data->ran_udp4_sourceip = 0;
  } else {
    ip4_ransip_checkbutton (&data->ran_udp4_sourceip, data->entry7, 2, data);
  }

  return (EXIT_SUCCESS);
}

// Destination IPv4 address (32 bits)
int
on_entry8_changed (GtkWidget *entry8, SPSData *data) {

  ip4_dip_entry (entry8, 2, data);

  return (EXIT_SUCCESS);
}

// IPv4 checksum (16 bits)
int
on_entry6_changed (GtkWidget *entry6, SPSData *data) {

  ip4_chksum_entry (entry6, 2, data);

  return (EXIT_SUCCESS);
}

// IP options for IPv4 UDP

// Decimal option data entry
int
on_radiobutton45_toggled (GtkButton *radiobutton45, SPSData *data) {

  (void) radiobutton45;  // This statement suppresses compiler warning about unused parameter.

  if (data->dec_hex_ipopt_udp4) {
    data->dec_hex_ipopt_udp4 = 0;
  } else {
    data->dec_hex_ipopt_udp4 = 1;
  }

  return (EXIT_SUCCESS);
}

// Hexadecimal option data entry
int
on_radiobutton46_toggled (GtkButton *radiobutton46) {

  (void) radiobutton46;  // This statement suppresses compiler warning about unused parameter.

  // on_radiobutton45_toggled() will change value of data->dec_hex_ipopt_udp4.

  return (EXIT_SUCCESS);
}

// IPv4 option data
int
on_entry410_changed (GtkWidget *entry410, SPSData *data) {

  data_entry (entry410, &data->ip_optlenbuf[2], MAX_IP4OPTLEN, data->ip_optionsbuf[2], data->dec_hex_ipopt_udp4, data->main_window, data);

  return (EXIT_SUCCESS);
}

// Insert IPv4 option after another, as specified by user.
int
on_button66_clicked (GtkButton *button66, SPSData *data) {

  (void) button66;  // This statement suppresses compiler warning about unused parameter.

  ip4_option (0, 2, data->ipopt_udp4_after, data->textview28, data);

  return (EXIT_SUCCESS);
}

// Number of option after which to insert new IPv4 option.
int
on_entry411_changed (GtkWidget *entry411, SPSData *data) {

  ip4_opt_insert_after_entry (entry411, 2, &data->ipopt_udp4_after, data);

  return (EXIT_SUCCESS);
}

// Append IP option for IPv4 UDP packet.
int
on_button67_clicked (GtkButton *button67, SPSData *data) {

  (void) button67;  // This statement suppresses compiler warning about unused parameter.

  ip4_option (2, 2, 0, data->textview28, data);

  return (EXIT_SUCCESS);
}

// View IPv4 header options for UDP.
int
on_button68_clicked (GtkButton *button68, SPSData *data) {

  (void) button68;  // This statement suppresses compiler warning about unused parameter.

  opt_view (2, 0, data);

  return (EXIT_SUCCESS);
}

// Number of IPv4 option to remove.
int
on_entry412_changed (GtkWidget *entry412, SPSData *data) {

  ip4_opt_remove_entry (entry412, 2, &data->ipopt_tcp4_remove, data);

  return (EXIT_SUCCESS);
}

// Remove selected IPv4 option.
int
on_button69_clicked (GtkButton *button69, SPSData *data) {

  (void) button69;  // This statement suppresses compiler warning about unused parameter.

  ip4_option (1, 2, data->ipopt_udp4_remove, data->textview28, data);

  // Clear entry with option number to remove.
  gtk_entry_set_text (GTK_ENTRY (data->entry412), "");

  return (EXIT_SUCCESS);
}

// UDP header (IPv4)

// UDP source port (16 bits)
int
on_entry13_changed (GtkWidget *entry13, SPSData *data) {

  udp_sport_entry (entry13, 2, data);

  if (data->ran_udp4_sourceport) {
    // Source port has been entered manually. Uncheck "Randomize" checkbutton.
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton3), FALSE);
    data->ran_udp4_sourceport = 0;
  }

  return (EXIT_SUCCESS);
}

// Randomize UDP source port number.
int
on_checkbutton3_toggled (GtkButton *checkbutton3, SPSData *data) {

  (void) checkbutton3;  // This statement suppresses compiler warning about unused parameter.

  if (data->ran_udp4_sourceport) {
    data->ran_udp4_sourceport = 0;
  } else {
    udp_ran_sport (&data->ran_udp4_sourceport, 2, data->entry13, data);
  }

  return (EXIT_SUCCESS);
}

// UDP destination port (16 bits)
int
on_entry14_changed (GtkWidget *entry14, SPSData *data) {

  udp_dport_entry (entry14, 2, data);

  return (EXIT_SUCCESS);
}

// Length of UDP datagram (16 bits)
int
on_entry15_changed (GtkWidget *entry15, SPSData *data) {

  udp_len_entry (entry15, 2, data);

  return (EXIT_SUCCESS);
}

// UDP checksum (16 bits)
int
on_entry16_changed (GtkWidget *entry16, SPSData *data) {

  udp_chksum_entry (entry16, 2, data);

  return (EXIT_SUCCESS);
}

// ASCII data entry
int
on_radiobutton13_toggled (GtkButton *radiobutton13, SPSData *data) {

  (void) radiobutton13;  // This statement suppresses compiler warning about unused parameter.

  if (data->ascii_hex_udp4) {
    data->ascii_hex_udp4 = 0;
  } else {
    data->ascii_hex_udp4 = 1;
  }

  return (EXIT_SUCCESS);
}

// Hexadecimal data entry
int
on_radiobutton14_toggled (GtkButton *radiobutton14) {

  (void) radiobutton14;  // This statement suppresses compiler warning about unused parameter.

  // on_radiobutton13_toggled() will change value of data->ascii_hex_udp4.

  return (EXIT_SUCCESS);
}

// UDP data from keyboard (IPv4)
int
on_entry17_changed (GtkWidget *entry17, SPSData *data) {

  key_data (entry17, 2, data->ascii_hex_udp4, data);

  if (data->nframes[2] > 1) {
    sprintf (data->warning_text, "Packet will be fragmented to accommodate UDP data.");
    data->parent = data->main_window;
    report_warning (data);
  }

  return (EXIT_SUCCESS);
}

// UDP data from file (IPv4)
int
on_button8_clicked (GtkButton *button8, SPSData *data) {

  (void) button8;  // This statement suppresses compiler warning about unused parameter.

  file_payload_data (data->entry17, 2, data);

  if (data->nframes[2] > 1) {
    sprintf (data->warning_text, "Packet will be fragmented to accommodate UDP data.");
    data->parent = data->main_window;
    report_warning (data);
  }

  return (EXIT_SUCCESS);
}

// Unload file data for UDP (IPv4)
int
on_button204_clicked (GtkButton *button204, SPSData *data) {

  (void) button204;  // This statement suppresses compiler warning about unused parameter.
  GtkTextBuffer *textbuffer64;

  // Remove file data.
  data->payloadlen[2] = 0;
  memset (data->payload[2], 0, IP_MAXPACKET * sizeof (uint8_t));

  // Update ethernet frame.
  create_ip4_frame (2, data);

  textbuffer64 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview64));
  gtk_text_buffer_set_text (textbuffer64, "", -1);

  return (EXIT_SUCCESS);
}

// UDP page (IPv6)

// Destination link-layer (MAC) address (48 bits)
int
on_entry134_changed (GtkWidget *entry134, SPSData *data) {

  mac_entry (entry134, data->ethhdr[5].dst_mac, 5, data);

  return (EXIT_SUCCESS);
}

// Source link-layer (MAC) address (48 bits)
int
on_entry135_changed (GtkWidget *entry135, SPSData *data) {

  mac_entry (entry135, data->ethhdr[5].src_mac, 5, data);

  return (EXIT_SUCCESS);
}

// This host's interface name to be converted to link-layer (MAC) address and find MTU.
int
on_entry155_changed (GtkWidget *entry155, SPSData *data) {

  // Check for valid source interface name, check MTU, and update source MAC address.
  ifname_entry (entry155, data->entry135, 5, data);

  // Update MTU spinbutton.
  update_mtu_spinbutton (5, data);

  return (EXIT_SUCCESS);
}

// Maximum transmission unit (MTU) - Changed via spinbutton + or -.
int
on_spinbutton10_value_changed (GtkWidget *spinbutton10, SPSData *data) {

  int value, min, ifmtu;

  // Set parent window for reporting error messages.
  data->parent = data->main_window;

  // Get requested MTU value from spinbutton.
  value = (int) gtk_spin_button_get_value (GTK_SPIN_BUTTON (spinbutton10));

  // Obtain interface's MTU.
  ifmtu = interface_getmtu (data->ifname[5]);
  if (ifmtu == 0) {
    update_mtu_spinbutton (5, data);  // Return spinbutton to current MTU value.
    return (EXIT_FAILURE);
  }

  // Determine minimum size needed to accommodate packet
  // when packet is fragmented to maximum extent.
  min = min6_mtu (5, data);

  // Ensure min required size is less than or equal to interface's MTU.
  if (min > ifmtu) {
    sprintf (data->error_text, "The interface's MTU is too small to accommodate packet.");
    report_error (data);
    data->ifmtu[5] = ifmtu;
    update_mtu_spinbutton (5, data);
    create_ip6_frame (5, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size is greater or equal to min required size to accommodate packet.
  if (value < min) {
    data->ifmtu[5] = min;
    update_mtu_spinbutton (5, data);
    create_ip6_frame (5, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size if <= interface's MTU.
  if (value > ifmtu) {
    data->ifmtu[5] = ifmtu;
    update_mtu_spinbutton (5, data);
    create_ip6_frame (5, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Accept requested MTU.
  data->ifmtu[5] = value;
  create_ip6_frame (5, data);  // Update ethernet frames.

  return (EXIT_SUCCESS);
}

// Maximum transmission unit (MTU) - Changed via spinbutton text entry.
int
on_spinbutton10_change_value (GtkWidget *spinbutton10, SPSData *data) {

  int value, min, ifmtu;

  // Set parent window for reporting error messages.
  data->parent = data->main_window;

  // Get requested MTU value from spinbutton text entry.
  value = (int) gtk_spin_button_get_value (GTK_SPIN_BUTTON (spinbutton10));

  // Obtain interface's MTU.
  ifmtu = interface_getmtu (data->ifname[5]);
  if (ifmtu == 0) {
    update_mtu_spinbutton (5, data);  // Return spinbutton to current MTU value.
    return (EXIT_FAILURE);
  }

  // Determine minimum size needed to accommodate packet
  // when packet is fragmented to maximum extent.
  min = min6_mtu (5, data);

  // Ensure min required size is less than or equal to interface's MTU.
  if (min > ifmtu) {
    sprintf (data->error_text, "The interface's MTU is too small to accommodate packet.");
    report_error (data);
    data->ifmtu[5] = ifmtu;
    update_mtu_spinbutton (5, data);
    create_ip6_frame (5, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size is greater or equal to min required size to accommodate packet.
  if (value < min) {
    sprintf (data->error_text, "The requested MTU is too small to accommodate packet.\nValue set to minimum required to accommodate packet instead.");
    report_error (data);
    data->ifmtu[5] = min;
    update_mtu_spinbutton (5, data);
    create_ip6_frame (5, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size if <= interface's MTU.
  if (value > ifmtu) {
    sprintf (data->error_text, "The requested MTU is larger than the interface's MTU.\nValue set to the interface's MTU instead.");
    report_error (data);
    data->ifmtu[5] = ifmtu;
    update_mtu_spinbutton (5, data);
    create_ip6_frame (5, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Accept requested MTU.
  data->ifmtu[5] = value;
  create_ip6_frame (6, data);  // Update ethernet frames.

  return (EXIT_SUCCESS);
}

// Ethernet type code (16 bits)
// http://www.iana.org/assignments/ethernet-numbers
int
on_entry136_changed (GtkWidget *entry136, SPSData *data) {

  ethtype_entry (entry136, 5, data);

  return (EXIT_SUCCESS);
}

// IP version (4 bits)
int
on_entry97_changed (GtkWidget *entry97, SPSData *data) {

  ip6_ver_entry (entry97, 5, data);

  return (EXIT_SUCCESS);
}

// Traffic class (8 bits)
int
on_entry98_changed (GtkWidget *entry98, SPSData *data) {

  ip6_tc_entry (entry98, 5, data);

  return (EXIT_SUCCESS);
}

// Flow label (20 bits)
int
on_entry99_changed (GtkWidget *entry99, SPSData *data) {

  ip6_fl_entry (entry99, 5, data);

  return (EXIT_SUCCESS);
}

// Payload length (16 bits)
int
on_entry100_changed (GtkWidget *entry100, SPSData *data) {

  ip6_plen_entry (entry100, 5, data);

  return (EXIT_SUCCESS);
}

// Next header (8 bits)
int
on_entry101_changed (GtkWidget *entry101, SPSData *data) {

  ip6_nxt_entry (entry101, 5, data);

  return (EXIT_SUCCESS);
}

// Hop limit (8 bits)
int
on_entry102_changed (GtkWidget *entry102, SPSData *data) {

  ip6_hops_entry (entry102, 5, data);

  return (EXIT_SUCCESS);
}

// Source IPv6 address (128 bits)
int
on_entry96_changed (GtkWidget *entry96, SPSData *data) {

  ip6_sip_entry (entry96, 5, data);

  if (data->ran_udp6_sourceip) {
    // Source IP address has been entered manually. Uncheck "Randomize" checkbutton.
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton6), FALSE);
    data->ran_udp6_sourceip = 0;
  }

  return (EXIT_SUCCESS);
}

// Randomize source IPv6 address
int
on_checkbutton6_toggled (GtkButton *checkbutton6, SPSData *data) {

  (void) checkbutton6;  // This statement suppresses compiler warning about unused parameter.

  if (data->ran_udp6_sourceip) {
    data->ran_udp6_sourceip = 0;
  } else {
    ip6_ransip_checkbutton (&data->ran_udp6_sourceip, data->entry96, 5, data);
  }

  return (EXIT_SUCCESS);
}

// Destination IP address (128 bits)
int
on_entry103_changed (GtkWidget *entry103, SPSData *data) {

  ip6_dip_entry (entry103, 5, data);

  return (EXIT_SUCCESS);
}

// Hop-by-Hop extension header
int
on_button100_clicked (GtkWidget *button100, SPSData *data) {

  (void) button100;  // This statement suppresses compiler warning about unused parameter.
  data->exthdr_type = 5;
  hop_win (data);

  return (EXIT_SUCCESS);
}

// Destination extension header
int
on_button112_clicked (GtkWidget *button112, SPSData *data) {

  (void) button112;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 5;
  dst_win (data);

  return (EXIT_SUCCESS);
}

// Routing extension header
int
on_button134_clicked (GtkWidget *button134, SPSData *data) {

  (void) button134;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 5;
  route_win (data);

  return (EXIT_SUCCESS);
}

// Authentication extension header
int
on_button118_clicked (GtkWidget *button118, SPSData *data) {

  (void) button118;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 5;
  auth_win (data);

  return (EXIT_SUCCESS);
}

// ESP extension header
int
on_button128_clicked (GtkWidget *button128, SPSData *data) {

  (void) button128;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 5;
  esp_win (data);

  return (EXIT_SUCCESS);
}

// Outer hop-by-Hop extension header
int
on_button306_clicked (GtkWidget *button306, SPSData *data) {

  (void) button306;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 5;
  hop_win_outer (data);

  return (EXIT_SUCCESS);
}

// Outer destination extension header
int
on_button307_clicked (GtkWidget *button307, SPSData *data) {

  (void) button307;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 5;
  dst_win_outer (data);

  return (EXIT_SUCCESS);
}

// Outer routing extension header
int
on_button308_clicked (GtkWidget *button308, SPSData *data) {

  (void) button308;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 5;
  route_win_outer (data);

  return (EXIT_SUCCESS);
}

// UDP header (IPv6)

// UDP source port (16 bits)
int
on_entry104_changed (GtkWidget *entry104, SPSData *data) {

  udp_sport_entry (entry104, 5, data);

  if (data->ran_udp6_sourceport) {
    //  Source port has been entered manually. Uncheck "Randomize" checkbutton.
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton8), FALSE);
    data->ran_udp6_sourceport = 0;
  }

  return (EXIT_SUCCESS);
}

// Randomize UDP source port number.
int
on_checkbutton8_toggled (GtkButton *checkbutton8, SPSData *data) {

  (void) checkbutton8;  // This statement suppresses compiler warning about unused parameter.

  if (data->ran_udp6_sourceport) {
    data->ran_udp6_sourceport = 0;
  } else {
    udp_ran_sport (&data->ran_udp6_sourceport, 5, data->entry104, data);
  }

  return (EXIT_SUCCESS);
}

// UDP destination port (16 bits)
int
on_entry105_changed (GtkWidget *entry105, SPSData *data) {

  udp_dport_entry (entry105, 5, data);

  return (EXIT_SUCCESS);
}

// Length of UDP datagram (16 bits)
int
on_entry106_changed (GtkWidget *entry106, SPSData *data) {

  udp_len_entry (entry106, 5, data);

  return (EXIT_SUCCESS);
}

// UDP checksum (16 bits)
int
on_entry107_changed (GtkWidget *entry107, SPSData *data) {

  udp_chksum_entry (entry107, 5, data);

  return (EXIT_SUCCESS);
}

// ASCII data entry
int
on_radiobutton15_toggled (GtkButton *radiobutton15, SPSData *data) {

  (void) radiobutton15;  // This statement suppresses compiler warning about unused parameter.

  if (data->ascii_hex_udp6) {
    data->ascii_hex_udp6 = 0;
  } else {
    data->ascii_hex_udp6 = 1;
  }

  return (EXIT_SUCCESS);
}

// Hexadecimal data entry
int
on_radiobutton16_toggled (GtkButton *radiobutton16) {

  (void) radiobutton16;  // This statement suppresses compiler warning about unused parameter.

  // on_radiobutton15_toggled() will change value of data->ascii_hex_udp6.

  return (EXIT_SUCCESS);
}

// UDP data from keyboard (IPv6)
int
on_entry108_changed (GtkWidget *entry108, SPSData *data) {

  key_data (entry108, 5, data->ascii_hex_udp6, data);

  if (data->nframes[5] > 1) {
    sprintf (data->warning_text, "Packet will be fragmented to accommodate UDP data.");
    data->parent = data->main_window;
    report_warning (data);
  }

  // Update ethernet frame and extension header editing window widgets, if visible.
  ip6data_update (5, data);

  return (EXIT_SUCCESS);
}

// UDP data from file (IPv6)
int
on_button10_clicked (GtkButton *button10, SPSData *data) {

  (void) button10;  // This statement suppresses compiler warning about unused parameter.

  if (file_payload_data (data->entry108, 5, data) == EXIT_SUCCESS) {

    if (data->nframes[5] > 1) {
      sprintf (data->warning_text, "Packet will be fragmented to accommodate UDP data.");
      data->parent = data->main_window;
      report_warning (data);
    }

    // Update ethernet frame and extension header editing window widgets, if visible.
    ip6data_update (5, data);
  }

  return (EXIT_SUCCESS);
}

// Unload file data for UDP (IPv6)
int
on_button205_clicked (GtkButton *button205, SPSData *data) {

  (void) button205;  // This statement suppresses compiler warning about unused parameter.
  GtkTextBuffer *textbuffer65;

  // Remove file data.
  data->payloadlen[5] = 0;
  memset (data->payload[5], 0, IP_MAXPACKET * sizeof (uint8_t));

  // Update ethernet frames.
  create_ip6_frame (5, data);
  create_6to4_frame (8, data);

  textbuffer65 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview65));
  gtk_text_buffer_set_text (textbuffer65, "", -1);

  return (EXIT_SUCCESS);
}

// Send page entries

// TCP (IPv4) packet type
int
on_radiobutton1_toggled (GtkButton *radiobutton1, SPSData *data) {

  (void) radiobutton1;  // This statement suppresses compiler warning about unused parameter.

  data->packet_type = 0;

  return (EXIT_SUCCESS);
}

// ICMP (IPv4) packet type
int
on_radiobutton2_toggled (GtkButton *radiobutton2, SPSData *data) {

  (void) radiobutton2;  // This statement suppresses compiler warning about unused parameter.

  data->packet_type = 1;

  return (EXIT_SUCCESS);
}

// UDP (IPv4) packet type
int
on_radiobutton3_toggled (GtkButton *radiobutton3, SPSData *data) {

  (void) radiobutton3;  // This statement suppresses compiler warning about unused parameter.

  data->packet_type = 2;

  return (EXIT_SUCCESS);
}

// All (IPv4) packet types (i.e., cycle)
int
on_radiobutton4_toggled (GtkButton *radiobutton4, SPSData *data) {

  (void) radiobutton4;  // This statement suppresses compiler warning about unused parameter.

  data->packet_type = 100;

  return (EXIT_SUCCESS);
}

// TCP (IPv6) packet type
int
on_radiobutton5_toggled (GtkButton *radiobutton5, SPSData *data) {

  (void) radiobutton5;  // This statement suppresses compiler warning about unused parameter.

  data->packet_type = 3;

  return (EXIT_SUCCESS);
}

// ICMP (IPv6) packet type
int
on_radiobutton6_toggled (GtkButton *radiobutton6, SPSData *data) {

  (void) radiobutton6;  // This statement suppresses compiler warning about unused parameter.

  data->packet_type = 4;

  return (EXIT_SUCCESS);
}

// UDP (IPv6) packet type
int
on_radiobutton7_toggled (GtkButton *radiobutton7, SPSData *data) {

  (void) radiobutton7;  // This statement suppresses compiler warning about unused parameter.

  data->packet_type = 5;

  return (EXIT_SUCCESS);
}

// All (IPv6) packet types (i.e., cycle)
int
on_radiobutton8_toggled (GtkButton *radiobutton8, SPSData *data) {

  (void) radiobutton8;  // This statement suppresses compiler warning about unused parameter.

  data->packet_type = 101;

  return (EXIT_SUCCESS);
}

// 6to4 (sps.ui)

// Tunnel IPv6 over IPv4 (sps.ui)
int
on_checkbutton12_toggled (GtkButton *checkbutton12, SPSData *data) {

  (void) checkbutton12;  // This statement suppresses compiler warning about unused parameter.

  if (data->sixto4_flag) {
    data->sixto4_flag = 0;
  } else {
    data->sixto4_flag = 1;
  }

  return (EXIT_SUCCESS);
}

// Number of packets to send
int
on_entry18_changed (GtkWidget *entry18, SPSData *data) {

  const char *entry18_text;
  char *value;

  // String to hold miscellaneous calculated values
  value = allocate_strmem (TMP_STRINGLEN);

  // Get new value from text entry
  entry18_text = gtk_entry_get_text (GTK_ENTRY (entry18));

  // Is value too large or small, or are there non-numeric characters?
  if ((!is_ascii_uint (entry18_text)) || (ascii_to_int64 ((char *) entry18_text) >= (int64_t) 0x7fffffffffffffffull) ||
      (ascii_to_int64 ((char *) entry18_text) < 1u)) {
    sprintf (value, "%" PRIu64, data->npackets);
    gtk_entry_set_text (GTK_ENTRY (data->entry18), value);
    free (value);
    return (EXIT_FAILURE);

  // Value is in acceptable range.
  } else {
    data->npackets = ascii_to_int64 ((char *) entry18_text);  // Update npackets with new value
  }

  // Free allocated memory.
  free (value);

  return (EXIT_SUCCESS);
}

// Set flood mode
int
on_checkbutton11_toggled (GtkButton *checkbutton11, SPSData *data) {

  (void) checkbutton11;  // This statement suppresses compiler warning about unused parameter.
  (void) data;  // This statement suppresses compiler warning about unused parameter.

  if (flood_mode) {
    flood_mode = 0;
  } else {
    flood_mode = 1;
  }

  return (EXIT_SUCCESS);
}

// Send packet(s)
int
on_button1_clicked (GtkButton *button1, SPSData *data) {

  (void) button1;  // This statement suppresses compiler warning about unused parameter.

  int i, type[3] = {3, 4, 5};
  GThread *thread;

  // IPv4 TCP, ICMP, and UDP
  if ((data->packet_type == 0) || (data->packet_type == 1) ||
      (data->packet_type == 2) || (data->packet_type == 100)) {
    // Spawn thread to send IPv4 packet(s).
    thread = g_thread_new ("ipv4_send", (GThreadFunc) ipv4_send, data);
    if (! thread) {
      fprintf (stderr, "Error: Unable to create new thread for ipv4_send() in on_button1_clicked().\n");
      exit (EXIT_FAILURE);
    }

  // IPv6 TCP, ICMP, and UDP (and 6to4)
  } else if ((data->packet_type == 3) || (data->packet_type == 4) ||
             (data->packet_type == 5) || (data->packet_type == 101)) {

    // 6to4 TCP, ICMP, and UDP
    if (data->sixto4_flag) {

      // MTU checks for 6to4.
      // This is necessary because checks are not performed for 6to4 when IPv6 packets
      // are changed. When IPv6 packets are modified, the user may not intend
      // to use 6to4, in which case the 6to4 IPv4 interface will not be specified, and
      // without the 6to4 interface specified, the 6to4 MTU checks cannot be completed.

      // Only one protocol packet needs to be checked.
      if (data->packet_type != 101) {
        if (sixto4_mtu_checks (data->packet_type, data) == EXIT_FAILURE) {
          return (EXIT_FAILURE);
        }

      // All three protocols need to be checked.
      } else {
        for (i=0; i<3; i++) {
          if (sixto4_mtu_checks (type[i], data) == EXIT_FAILURE) {
            return (EXIT_FAILURE);
          }
        }
      }

      // Spawn thread to send 6to4 IPv6 packet(s).
      thread = g_thread_new ("sixto4_send", (GThreadFunc) sixto4_send, data);
      if (! thread) {
        fprintf (stderr, "Error: Unable to create new thread for sixto4_send() in on_button1_clicked().\n");
        exit (EXIT_FAILURE);
      }

    // IPv6 TCP, ICMP, and UDP
    } else {
      // Spawn thread to send IPv6 packet(s).
      thread = g_thread_new ("ipv6_send", (GThreadFunc) ipv6_send, data);
      if (! thread) {
        fprintf (stderr, "Error: Unable to create new thread for ipv6_send() in on_button1_clicked().\n");
        exit (EXIT_FAILURE);
      }
    }
  }

  return (EXIT_SUCCESS);
}

// Stop sending packets
int
on_button5_clicked (GtkButton *button5, SPSData *data) {

  (void) button5;  // This statement suppresses compiler warning about unused parameter.

  // By toggling the checkbutton, the on_checkbutton11_toggled callback will be called
  // and it will set flood_mode to 0 for us.
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton11), FALSE);

  // If we're not flooding but sending a finite number of packets but want to stop,
  // set sending to 0 to stop.
  G_LOCK (sending);
  sending = 0;
  G_UNLOCK (sending);

  return (EXIT_SUCCESS);
}

// Load SPS file, which contains settings selected by user when saved
int
on_button21_clicked (GtkButton *button21, SPSData *data) {

  (void) button21;  // This statement suppresses compiler warning about unused parameter.

  load_file (data);

  return (EXIT_SUCCESS);
}

// Save SPS file, which contains settings selected by user.
int
on_button14_clicked (GtkButton *button14, SPSData *data) {

  (void) button14;  // This statement suppresses compiler warning about unused parameter.

  save_file (data);

  return (EXIT_SUCCESS);
}

// Send Packet settings
int
on_checkbutton25_toggled (GtkButton *checkbutton25, SPSData *data) {

  (void) checkbutton25;  // This statement suppresses compiler warning about unused parameter.

  if (data->save_flags[0]) {
    data->save_flags[0] = 0;
  } else {
    data->save_flags[0] = 1;
  }

  return (EXIT_SUCCESS);
}

// Select to save IPv6 over IPv4 (6to4)
int
on_checkbutton15_toggled (GtkButton *checkbutton15, SPSData *data) {

  (void) checkbutton15;  // This statement suppresses compiler warning about unused parameter.

  if (data->save_flags[1]) {
    data->save_flags[1] = 0;
  } else {
    data->save_flags[1] = 1;
  }

  return (EXIT_SUCCESS);
}

// Select to save IPv4 TCP packet
int
on_checkbutton16_toggled (GtkButton *checkbutton16, SPSData *data) {

  (void) checkbutton16;  // This statement suppresses compiler warning about unused parameter.

  if (data->save_flags[2]) {
    data->save_flags[2] = 0;
  } else {
    data->save_flags[2] = 1;
  }

  return (EXIT_SUCCESS);
}

// Select to save IPv4 ICMP packet
int
on_checkbutton20_toggled (GtkButton *checkbutton20, SPSData *data) {

  (void) checkbutton20;  // This statement suppresses compiler warning about unused parameter.

  if (data->save_flags[3]) {
    data->save_flags[3] = 0;
  } else {
    data->save_flags[3] = 1;
  }

  return (EXIT_SUCCESS);
}

// Select to save IPv4 UDP packet
int
on_checkbutton21_toggled (GtkButton *checkbutton21, SPSData *data) {

  (void) checkbutton21;  // This statement suppresses compiler warning about unused parameter.

  if (data->save_flags[4]) {
    data->save_flags[4] = 0;
  } else {
    data->save_flags[4] = 1;
  }

  return (EXIT_SUCCESS);
}

// Select to save IPv6 TCP packet
int
on_checkbutton22_toggled (GtkButton *checkbutton22, SPSData *data) {

  (void) checkbutton22;  // This statement suppresses compiler warning about unused parameter.

  if (data->save_flags[5]) {
    data->save_flags[5] = 0;
  } else {
    data->save_flags[5] = 1;
  }

  return (EXIT_SUCCESS);
}

// Select to save IPv6 ICMP packet
int
on_checkbutton23_toggled (GtkButton *checkbutton23, SPSData *data) {

  (void) checkbutton23;  // This statement suppresses compiler warning about unused parameter.

  if (data->save_flags[6]) {
    data->save_flags[6] = 0;
  } else {
    data->save_flags[6] = 1;
  }

  return (EXIT_SUCCESS);
}

// Select to save IPv6 UDP packet
int
on_checkbutton24_toggled (GtkButton *checkbutton24, SPSData *data) {

  (void) checkbutton24;  // This statement suppresses compiler warning about unused parameter.

  if (data->save_flags[7]) {
    data->save_flags[7] = 0;
  } else {
    data->save_flags[7] = 1;
  }

  return (EXIT_SUCCESS);
}

// Select to save 6to4 TCP (ethernet header, IPv4 header, IPv4 options + padding)
int
on_checkbutton14_toggled (GtkButton *checkbutton14, SPSData *data) {

  (void) checkbutton14;  // This statement suppresses compiler warning about unused parameter.

  if (data->save_flags[8]) {
    data->save_flags[8] = 0;
  } else {
    data->save_flags[8] = 1;
  }

  return (EXIT_SUCCESS);
}

// Select to save 6to4 ICMP (ethernet header, IPv4 header, IPv4 options + padding)
int
on_checkbutton17_toggled (GtkButton *checkbutton17, SPSData *data) {

  (void) checkbutton17;  // This statement suppresses compiler warning about unused parameter.

  if (data->save_flags[9]) {
    data->save_flags[9] = 0;
  } else {
    data->save_flags[9] = 1;
  }

  return (EXIT_SUCCESS);
}

// Select to save 6to4 UDP (ethernet header, IPv4 header, IPv4 options + padding)
int
on_checkbutton18_toggled (GtkButton *checkbutton18, SPSData *data) {

  (void) checkbutton18;  // This statement suppresses compiler warning about unused parameter.

  if (data->save_flags[10]) {
    data->save_flags[10] = 0;
  } else {
    data->save_flags[10] = 1;
  }

  return (EXIT_SUCCESS);
}

// Select to save traceroute settings
int
on_checkbutton36_toggled (GtkButton *checkbutton36, SPSData *data) {

  (void) checkbutton36;  // This statement suppresses compiler warning about unused parameter.

  if (data->save_flags[8]) {
    data->save_flags[8] = 0;
  } else {
    data->save_flags[8] = 1;
  }

  return (EXIT_SUCCESS);
}

// Select to save IPv6 over IPv4 (6to4) for traceroute
int
on_checkbutton46_toggled (GtkButton *checkbutton46, SPSData *data) {

  (void) checkbutton46;  // This statement suppresses compiler warning about unused parameter.

  if (data->save_flags[9]) {
    data->save_flags[9] = 0;
  } else {
    data->save_flags[9] = 1;
  }

  return (EXIT_SUCCESS);
}

// Select to save IPv4 TCP packet for traceroute
int
on_checkbutton39_toggled (GtkButton *checkbutton39, SPSData *data) {

  (void) checkbutton39;  // This statement suppresses compiler warning about unused parameter.

  if (data->save_flags[10]) {
    data->save_flags[10] = 0;
  } else {
    data->save_flags[10] = 1;
  }

  return (EXIT_SUCCESS);
}

// Select to save IPv4 ICMP packet for traceroute
int
on_checkbutton41_toggled (GtkButton *checkbutton41, SPSData *data) {

  (void) checkbutton41;  // This statement suppresses compiler warning about unused parameter.

  if (data->save_flags[11]) {
    data->save_flags[11] = 0;
  } else {
    data->save_flags[11] = 1;
  }

  return (EXIT_SUCCESS);
}

// Select to save IPv4 UDP packet for traceroute
int
on_checkbutton42_toggled (GtkButton *checkbutton42, SPSData *data) {

  (void) checkbutton42;  // This statement suppresses compiler warning about unused parameter.

  if (data->save_flags[12]) {
    data->save_flags[12] = 0;
  } else {
    data->save_flags[12] = 1;
  }

  return (EXIT_SUCCESS);
}

// Select to save IPv6 TCP packet for traceroute
int
on_checkbutton43_toggled (GtkButton *checkbutton43, SPSData *data) {

  (void) checkbutton43;  // This statement suppresses compiler warning about unused parameter.

  if (data->save_flags[13]) {
    data->save_flags[13] = 0;
  } else {
    data->save_flags[13] = 1;
  }

  return (EXIT_SUCCESS);
}

// Select to save IPv6 ICMP packet for traceroute
int
on_checkbutton44_toggled (GtkButton *checkbutton44, SPSData *data) {

  (void) checkbutton44;  // This statement suppresses compiler warning about unused parameter.

  if (data->save_flags[14]) {
    data->save_flags[14] = 0;
  } else {
    data->save_flags[14] = 1;
  }

  return (EXIT_SUCCESS);
}

// Select to save IPv6 UDP packet for traceroute
int
on_checkbutton45_toggled (GtkButton *checkbutton45, SPSData *data) {

  (void) checkbutton45;  // This statement suppresses compiler warning about unused parameter.

  if (data->save_flags[15]) {
    data->save_flags[15] = 0;
  } else {
    data->save_flags[15] = 1;
  }

  return (EXIT_SUCCESS);
}

// Timeout for ARP/ND to receive reply
int
on_spinbutton1_value_changed (GtkWidget *spinbutton1, SPSData *data) {

  data->timeout = (int) gtk_spin_button_get_value (GTK_SPIN_BUTTON (spinbutton1));

  return (EXIT_SUCCESS);
}

// Call tcp4_default()
int
on_button11_clicked (GtkButton *button11, SPSData *data) {

  (void) button11;  // This statement suppresses compiler warning about unused parameter.

  tcp4_default (data);

  return (EXIT_SUCCESS);
}

// Call icmp4_default()
int
on_button12_clicked (GtkButton *button12, SPSData *data) {

  (void) button12;  // This statement suppresses compiler warning about unused parameter.

  icmp4_default (data);

  return (EXIT_SUCCESS);
}

// Call udp4_default()
int
on_button15_clicked (GtkButton *button15, SPSData *data) {

  (void) button15;  // This statement suppresses compiler warning about unused parameter.

  udp4_default (data);

  return (EXIT_SUCCESS);
}

// Call tcp6_default()
int
on_button16_clicked (GtkButton *button16, SPSData *data) {

  (void) button16;  // This statement suppresses compiler warning about unused parameter.

  tcp6_default (data);

  return (EXIT_SUCCESS);
}

// Call icmp6_default()
int
on_button18_clicked (GtkButton *button18, SPSData *data) {

  (void) button18;  // This statement suppresses compiler warning about unused parameter.

  icmp6_default (data);

  return (EXIT_SUCCESS);
}

// Call udp6_default()
int
on_button19_clicked (GtkButton *button19, SPSData *data) {

  (void) button19;  // This statement suppresses compiler warning about unused parameter.

  udp6_default (data);

  return (EXIT_SUCCESS);
}

// 6to4 page (sps.ui)

// 6to4 TCP

// 6to4 TCP Destination link-layer (MAC) address (48 bits) (sps.ui)
int
on_entry600_changed (GtkWidget *entry600, SPSData *data) {

  mac_entry (entry600, data->ethhdr[6].dst_mac, 6, data);

  return (EXIT_SUCCESS);
}

// 6to4 TCP Source link-layer (MAC) address (48 bits) (sps.ui)
int
on_entry601_changed (GtkWidget *entry601, SPSData *data) {

  mac_entry (entry601, data->ethhdr[6].src_mac, 6, data);

  return (EXIT_SUCCESS);
}

// 6to4 TCP Interface Name to be converted to link-layer (MAC) address and find MTU. (sps.ui)
int
on_entry460_changed (GtkWidget *entry460, SPSData *data) {

  // Check for valid source interface name, check MTU, and update source MAC address.
  sixto4_ifname_entry (entry460, data->entry601, 6, data);

  return (EXIT_SUCCESS);
}

// 6to4 TCP Maximum transmission unit (MTU) - Changed via spinbutton + or -. (sps.ui)
int
on_spinbutton17_value_changed (GtkWidget *spinbutton17, SPSData *data) {

  int value, min, ifmtu;

  // Set parent window for reporting error messages.
  data->parent = data->main_window;

  // Get requested MTU value from spinbutton.
  value = (int) gtk_spin_button_get_value (GTK_SPIN_BUTTON (spinbutton17));

  // Obtain interface's MTU.
  ifmtu = interface_getmtu (data->ifname[6]);
  if (ifmtu == 0) {
    update_mtu_spinbutton (6, data);  // Return spinbutton to current MTU value.
    return (EXIT_FAILURE);
  }

  // Determine minimum size needed to accommodate packet
  // when packet is fragmented to maximum extent.
  min = min6to4_mtu (6, data);

  // Ensure min required size is less than or equal to interface's MTU.
  if (min > ifmtu) {
    sprintf (data->error_text, "The interface's MTU is too small to accommodate packet.");
    report_error (data);
    data->ifmtu[6] = ifmtu;
    update_mtu_spinbutton (6, data);
    create_6to4_frame (6, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size is greater or equal to min required size to accommodate packet.
  if (value < min) {
    data->ifmtu[6] = min;
    update_mtu_spinbutton (6, data);
    create_6to4_frame (6, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size if <= interface's MTU.
  if (value > ifmtu) {
    data->ifmtu[6] = ifmtu;
    update_mtu_spinbutton (6, data);
    create_6to4_frame (6, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Accept requested MTU.
  data->ifmtu[6] = value;
  create_6to4_frame (6, data);  // Update ethernet frames.

  return (EXIT_SUCCESS);
}

// 6to4 TCP Maximum transmission unit (MTU) - Changed via spinbutton text entry. (sps.ui)
int
on_spinbutton17_change_value (GtkWidget *spinbutton17, SPSData *data) {

  int value, min, ifmtu;

  // Set parent window for reporting error messages.
  data->parent = data->main_window;

  // Get requested MTU value from spinbutton text entry.
  value = (int) gtk_spin_button_get_value (GTK_SPIN_BUTTON (spinbutton17));

  // Obtain interface's MTU.
  ifmtu = interface_getmtu (data->ifname[6]);
  if (ifmtu == 0) {
    update_mtu_spinbutton (6, data);  // Return spinbutton to current MTU value.
    return (EXIT_FAILURE);
  }

  // Determine minimum size needed to accommodate packet
  // when packet is fragmented to maximum extent.
  min = min6to4_mtu (6, data);

  // Ensure min required size is less than or equal to interface's MTU.
  if (min > ifmtu) {
    sprintf (data->error_text, "The interface's MTU is too small to accommodate packet.");
    report_error (data);
    data->ifmtu[6] = ifmtu;
    update_mtu_spinbutton (6, data);
    create_6to4_frame (6, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size is greater or equal to min required size to accommodate packet.
  if (value < min) {
    sprintf (data->error_text, "The requested MTU is too small to accommodate packet.\nValue set to minimum required to accommodate packet instead.");
    report_error (data);
    data->ifmtu[6] = min;
    update_mtu_spinbutton (6, data);
    create_6to4_frame (6, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size if <= interface's MTU.
  if (value > ifmtu) {
    sprintf (data->error_text, "The requested MTU is larger than the interface's MTU.\nValue set to the interface's MTU instead.");
    report_error (data);
    data->ifmtu[6] = ifmtu;
    update_mtu_spinbutton (6, data);
    create_6to4_frame (6, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Accept requested MTU.
  data->ifmtu[6] = value;
  create_6to4_frame (6, data);  // Update ethernet frames.

  return (EXIT_SUCCESS);
}

// 6to4 TCP Ethernet type code (16 bits) (sps.ui)
// http://www.iana.org/assignments/ethernet-numbers
int
on_entry602_changed (GtkWidget *entry602, SPSData *data) {

  ethtype_entry (entry602, 6, data);

  return (EXIT_SUCCESS);
}

// 6to4 TCP IPv4 header length (4 bits) (sps.ui)
int
on_entry609_changed (GtkWidget *entry609, SPSData *data) {

  ip4_hdrlen_entry (entry609, 6, data);

  return (EXIT_SUCCESS);
}

// 6to4 TCP IP version (4 bits) (sps.ui)
int
on_entry610_changed (GtkWidget *entry610, SPSData *data) {

  ip4_ver_entry (entry610, 6, data);

  return (EXIT_SUCCESS);
}

// 6to4 TCP Type of service (8 bits) (sps.ui)
int
on_entry611_changed (GtkWidget *entry611, SPSData *data) {

  ip4_tos_entry (entry611, 6, data);

  return (EXIT_SUCCESS);
}

// 6to4 TCP Total length of datagram (16 bits) (sps.ui)
int
on_entry612_changed (GtkWidget *entry612, SPSData *data) {

  ip4_tld_entry (entry612, 6, data);

  return (EXIT_SUCCESS);
}

// 6to4 TCP ID sequence number (16 bits) (sps.ui)
int
on_entry613_changed (GtkWidget *entry613, SPSData *data) {

  ip4_seq_entry (entry613, 6, data);

  return (EXIT_SUCCESS);
}

// 6to4 TCP Zero (1 bit) (sps.ui)
int
on_entry614_changed (GtkWidget *entry614, SPSData *data) {

  ip4_zero_entry (entry614, data->entry615, data->entry616, data->entry617, 6, data);

  return (EXIT_SUCCESS);
}

// 6to4 TCP Do not fragment flag (1 bit) (sps.ui)
int
on_entry615_changed (GtkWidget *entry615, SPSData *data) {

  ip4_dnf_entry (data->entry614, entry615, data->entry616, data->entry617, 6, data);

  return (EXIT_SUCCESS);
}

// 6to4 TCP More fragments following flag (1 bit) (sps.ui)
int
on_entry616_changed (GtkWidget *entry616, SPSData *data) {

  ip4_mff_entry (data->entry614, data->entry615, entry616, data->entry617, 6, data);

  return (EXIT_SUCCESS);
}

// 6to4 TCP Fragmentation offset (13 bits) (sps.ui)
int
on_entry617_changed (GtkWidget *entry617, SPSData *data) {

  ip4_fo_entry (data->entry614, data->entry615, data->entry616, entry617, 6, data);

  return (EXIT_SUCCESS);
}

// 6to4 TCP Time-to-Live (8 bits) (sps.ui)
int
on_entry618_changed (GtkWidget *entry618, SPSData *data) {

  ip4_ttl_entry (entry618, 6, data);

  return (EXIT_SUCCESS);
}

// 6to4 TCP Transport layer protocol (8 bits) (sps.ui)
int
on_entry619_changed (GtkWidget *entry619, SPSData *data) {

  ip4_tlp_entry (entry619, 6, data);

  return (EXIT_SUCCESS);
}

// 6to4 TCP IPv4 Source Address (sps.ui)
int
on_entry125_changed (GtkWidget *entry125, SPSData *data) {

  ip4_sip_entry (entry125, 6, data);

  if (data->ran_tcp6to4_sourceip) {
    // Source IP address has been entered manually. Uncheck "Randomize" checkbutton.
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton13), FALSE);
    data->ran_tcp6to4_sourceip = 0;
  }

  return (EXIT_SUCCESS);
}

// 6to4 TCP Randomize IPv4 source address (sps.ui)
int
on_checkbutton13_toggled (GtkButton *checkbutton13, SPSData *data) {

  (void) checkbutton13; // This statement suppresses compiler warning about unused parameter.

  if (data->ran_tcp6to4_sourceip) {
    data->ran_tcp6to4_sourceip = 0;
  } else {
    ip4_ransip_checkbutton (&data->ran_tcp6to4_sourceip, data->entry125, 6, data);
  }

  return (EXIT_SUCCESS);
}

// 6to4 TCP Destination IPv4 address (32 bits) (sps.ui)
int
on_entry620_changed (GtkWidget *entry620, SPSData *data) {

  ip4_dip_entry (entry620, 6, data);

  return (EXIT_SUCCESS);
}

// 6to4 TCP IPv4 checksum (16 bits) (sps.ui)
int
on_entry621_changed (GtkWidget *entry621, SPSData *data) {

  ip4_chksum_entry (entry621, 6, data);

  return (EXIT_SUCCESS);
}

// 6to4 TCP Restore default IPv4 values.
int
on_button600_clicked (GtkButton *button600, SPSData *data) {

  (void) button600;  // This statement suppresses compiler warning about unused parameter.

  sixto4_tcp6_default (data);

  return (EXIT_SUCCESS);
}

// 6to4 ICMP

// 6to4 ICMP Destination link-layer (MAC) address (48 bits) (sps.ui)
int
on_entry603_changed (GtkWidget *entry603, SPSData *data) {

  mac_entry (entry603, data->ethhdr[7].dst_mac, 7, data);

  return (EXIT_SUCCESS);
}

// 6to4 ICMP Source link-layer (MAC) address (48 bits) (sps.ui)
int
on_entry604_changed (GtkWidget *entry604, SPSData *data) {

  mac_entry (entry604, data->ethhdr[7].src_mac, 7, data);

  return (EXIT_SUCCESS);
}

// 6to4 ICMP Interface Name to be converted to link-layer (MAC) address and find MTU. (sps.ui)
int
on_entry461_changed (GtkWidget *entry461, SPSData *data) {

  // Check for valid source interface name, check MTU, and update source MAC address.
  sixto4_ifname_entry (entry461, data->entry604, 7, data);

  return (EXIT_SUCCESS);
}

// 6to4 ICMP Maximum transmission unit (MTU) - Changed via spinbutton + or -. (sps.ui)
int
on_spinbutton18_value_changed (GtkWidget *spinbutton18, SPSData *data) {

  int value, min, ifmtu;

  // Set parent window for reporting error messages.
  data->parent = data->main_window;

  // Get requested MTU value from spinbutton.
  value = (int) gtk_spin_button_get_value (GTK_SPIN_BUTTON (spinbutton18));

  // Obtain interface's MTU.
  ifmtu = interface_getmtu (data->ifname[7]);
  if (ifmtu == 0) {
    update_mtu_spinbutton (7, data);  // Return spinbutton to current MTU value.
    return (EXIT_FAILURE);
  }

  // Determine minimum size needed to accommodate packet
  // when packet is fragmented to maximum extent.
  min = min6to4_mtu (7, data);

  // Ensure min required size is less than or equal to interface's MTU.
  if (min > ifmtu) {
    sprintf (data->error_text, "The interface's MTU is too small to accommodate packet.");
    report_error (data);
    data->ifmtu[7] = ifmtu;
    update_mtu_spinbutton (7, data);
    create_6to4_frame (7, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size is greater or equal to min required size to accommodate packet.
  if (value < min) {
    data->ifmtu[7] = min;
    update_mtu_spinbutton (7, data);
    create_6to4_frame (7, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size if <= interface's MTU.
  if (value > ifmtu) {
    data->ifmtu[7] = ifmtu;
    update_mtu_spinbutton (7, data);
    create_6to4_frame (7, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Accept requested MTU.
  data->ifmtu[7] = value;
  create_6to4_frame (7, data);  // Update ethernet frames.

  return (EXIT_SUCCESS);
}

// 6to4 ICMP Maximum transmission unit (MTU) - Changed via spinbutton text entry. (sps.ui)
int
on_spinbutton18_change_value (GtkWidget *spinbutton18, SPSData *data) {

  int value, min, ifmtu;

  // Set parent window for reporting error messages.
  data->parent = data->main_window;

  // Get requested MTU value from spinbutton text entry.
  value = (int) gtk_spin_button_get_value (GTK_SPIN_BUTTON (spinbutton18));

  // Obtain interface's MTU.
  ifmtu = interface_getmtu (data->ifname[7]);
  if (ifmtu == 0) {
    update_mtu_spinbutton (7, data);  // Return spinbutton to current MTU value.
    return (EXIT_FAILURE);
  }

  // Determine minimum size needed to accommodate packet
  // when packet is fragmented to maximum extent.
  min = min6to4_mtu (7, data);

  // Ensure min required size is less than or equal to interface's MTU.
  if (min > ifmtu) {
    sprintf (data->error_text, "The interface's MTU is too small to accommodate packet.");
    report_error (data);
    data->ifmtu[7] = ifmtu;
    update_mtu_spinbutton (7, data);
    create_6to4_frame (7, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size is greater or equal to min required size to accommodate packet.
  if (value < min) {
    sprintf (data->error_text, "The requested MTU is too small to accommodate packet.\nValue set to minimum required to accommodate packet instead.");
    report_error (data);
    data->ifmtu[7] = min;
    update_mtu_spinbutton (7, data);
    create_6to4_frame (7, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size if <= interface's MTU.
  if (value > ifmtu) {
    sprintf (data->error_text, "The requested MTU is larger than the interface's MTU.\nValue set to the interface's MTU instead.");
    report_error (data);
    data->ifmtu[7] = ifmtu;
    update_mtu_spinbutton (7, data);
    create_6to4_frame (7, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Accept requested MTU.
  data->ifmtu[7] = value;
  create_6to4_frame (7, data);  // Update ethernet frames.

  return (EXIT_SUCCESS);
}

// 6to4 ICMP Ethernet type code (16 bits) (sps.ui)
// http://www.iana.org/assignments/ethernet-numbers
int
on_entry605_changed (GtkWidget *entry605, SPSData *data) {

  ethtype_entry (entry605, 6, data);

  return (EXIT_SUCCESS);
}

// 6to4 ICMP IPv4 header length (4 bits) (sps.ui)
int
on_entry622_changed (GtkWidget *entry622, SPSData *data) {

  ip4_hdrlen_entry (entry622, 7, data);

  return (EXIT_SUCCESS);
}

// 6to4 ICMP IP version (4 bits) (sps.ui)
int
on_entry623_changed (GtkWidget *entry623, SPSData *data) {

  ip4_ver_entry (entry623, 7, data);

  return (EXIT_SUCCESS);
}

// 6to4 ICMP Type of service (8 bits) (sps.ui)
int
on_entry624_changed (GtkWidget *entry624, SPSData *data) {

  ip4_tos_entry (entry624, 7, data);

  return (EXIT_SUCCESS);
}

// 6to4 ICMP Total length of datagram (16 bits) (sps.ui)
int
on_entry625_changed (GtkWidget *entry625, SPSData *data) {

  ip4_tld_entry (entry625, 7, data);

  return (EXIT_SUCCESS);
}

// 6to4 ICMP ID sequence number (16 bits) (sps.ui)
int
on_entry626_changed (GtkWidget *entry626, SPSData *data) {

  ip4_seq_entry (entry626, 7, data);

  return (EXIT_SUCCESS);
}

// 6to4 ICMP Zero (1 bit) (sps.ui)
int
on_entry627_changed (GtkWidget *entry627, SPSData *data) {

  ip4_zero_entry (entry627, data->entry628, data->entry629, data->entry630, 7, data);

  return (EXIT_SUCCESS);
}

// 6to4 ICMP Do not fragment flag (1 bit) (sps.ui)
int
on_entry628_changed (GtkWidget *entry628, SPSData *data) {

  ip4_dnf_entry (data->entry627, entry628, data->entry629, data->entry630, 7, data);

  return (EXIT_SUCCESS);
}

// 6to4 ICMP More fragments following flag (1 bit) (sps.ui)
int
on_entry629_changed (GtkWidget *entry629, SPSData *data) {

  ip4_mff_entry (data->entry627, data->entry628, entry629, data->entry630, 7, data);

  return (EXIT_SUCCESS);
}

// 6to4 ICMP Fragmentation offset (13 bits) (sps.ui)
int
on_entry630_changed (GtkWidget *entry630, SPSData *data) {

  ip4_fo_entry (data->entry627, data->entry628, data->entry629, entry630, 7, data);

  return (EXIT_SUCCESS);
}

// 6to4 ICMP Time-to-Live (8 bits) (sps.ui)
int
on_entry631_changed (GtkWidget *entry631, SPSData *data) {

  ip4_ttl_entry (entry631, 7, data);

  return (EXIT_SUCCESS);
}

// 6to4 ICMP Transport layer protocol (8 bits) (sps.ui)
int
on_entry632_changed (GtkWidget *entry632, SPSData *data) {

  ip4_tlp_entry (entry632, 7, data);

  return (EXIT_SUCCESS);
}

// 6to4 ICMP IPv4 Source Address (sps.ui)
int
on_entry160_changed (GtkWidget *entry160, SPSData *data) {

  ip4_sip_entry (entry160, 7, data);

  if (data->ran_icmp6to4_sourceip) {
    // Source IP address has been entered manually. Uncheck "Randomize" checkbutton.
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton26), FALSE);
    data->ran_icmp6to4_sourceip = 0;
  }

  return (EXIT_SUCCESS);
}

// 6to4 ICMP Randomize IPv4 source address (sps.ui)
int
on_checkbutton26_toggled (GtkButton *checkbutton26, SPSData *data) {

  (void) checkbutton26;  // This statement suppresses compiler warning about unused parameter.

  if (data->ran_icmp6to4_sourceip) {
    data->ran_icmp6to4_sourceip = 0;
  } else {
    ip4_ransip_checkbutton (&data->ran_icmp6to4_sourceip, data->entry160, 7, data);
  }

  return (EXIT_SUCCESS);
}

// 6to4 ICMP Destination IPv4 address (32 bits) (sps.ui)
int
on_entry633_changed (GtkWidget *entry633, SPSData *data) {

  ip4_dip_entry (entry633, 7, data);

  return (EXIT_SUCCESS);
}

// 6to4 ICMP IPv4 checksum (16 bits) (sps.ui)
int
on_entry634_changed (GtkWidget *entry634, SPSData *data) {

  ip4_chksum_entry (entry634, 7, data);

  return (EXIT_SUCCESS);
}

// 6to4 ICMP Restore default IPv4 values.
int
on_button601_clicked (GtkButton *button601, SPSData *data) {

  (void) button601;  // This statement suppresses compiler warning about unused parameter.

  sixto4_icmp6_default (data);

  return (EXIT_SUCCESS);
}

// 6to4 UDP

// 6to4 UDP Destination link-layer (MAC) address (48 bits) (sps.ui)
int
on_entry606_changed (GtkWidget *entry606, SPSData *data) {

  mac_entry (entry606, data->ethhdr[8].dst_mac, 8, data);

  return (EXIT_SUCCESS);
}

// 6to4 UDP Source link-layer (MAC) address (48 bits) (sps.ui)
int
on_entry607_changed (GtkWidget *entry607, SPSData *data) {

  mac_entry (entry607, data->ethhdr[8].src_mac, 8, data);

  return (EXIT_SUCCESS);
}

// 6to4 UDP Interface Name to be converted to link-layer (MAC) address and find MTU. (sps.ui)
int
on_entry462_changed (GtkWidget *entry462, SPSData *data) {

  // Check for valid source interface name, check MTU, and update source MAC address.
  sixto4_ifname_entry (entry462, data->entry607, 8, data);

  return (EXIT_SUCCESS);
}

// 6to4 UDP Maximum transmission unit (MTU) - Changed via spinbutton + or -. (sps.ui)
int
on_spinbutton19_value_changed (GtkWidget *spinbutton19, SPSData *data) {

  int value, min, ifmtu;

  // Set parent window for reporting error messages.
  data->parent = data->main_window;

  // Get requested MTU value from spinbutton.
  value = (int) gtk_spin_button_get_value (GTK_SPIN_BUTTON (spinbutton19));

  // Obtain interface's MTU.
  ifmtu = interface_getmtu (data->ifname[8]);
  if (ifmtu == 0) {
    update_mtu_spinbutton (8, data);  // Return spinbutton to current MTU value.
    return (EXIT_FAILURE);
  }

  // Determine minimum size needed to accommodate packet
  // when packet is fragmented to maximum extent.
  min = min6to4_mtu (8, data);

  // Ensure min required size is less than or equal to interface's MTU.
  if (min > ifmtu) {
    sprintf (data->error_text, "The interface's MTU is too small to accommodate packet.");
    report_error (data);
    data->ifmtu[8] = ifmtu;
    update_mtu_spinbutton (8, data);
    create_6to4_frame (8, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size is greater or equal to min required size to accommodate packet.
  if (value < min) {
    data->ifmtu[8] = min;
    update_mtu_spinbutton (8, data);
    create_6to4_frame (8, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size if <= interface's MTU.
  if (value > ifmtu) {
    data->ifmtu[8] = ifmtu;
    update_mtu_spinbutton (8, data);
    create_6to4_frame (8, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Accept requested MTU.
  data->ifmtu[8] = value;
  create_6to4_frame (8, data);  // Update ethernet frames.

  return (EXIT_SUCCESS);
}

// 6to4 UDP Maximum transmission unit (MTU) - Changed via spinbutton text entry. (sps.ui)
int
on_spinbutton19_change_value (GtkWidget *spinbutton19, SPSData *data) {

  int value, min, ifmtu;

  // Set parent window for reporting error messages.
  data->parent = data->main_window;

  // Get requested MTU value from spinbutton text entry.
  value = (int) gtk_spin_button_get_value (GTK_SPIN_BUTTON (spinbutton19));

  // Obtain interface's MTU.
  ifmtu = interface_getmtu (data->ifname[8]);
  if (ifmtu == 0) {
    update_mtu_spinbutton (8, data);  // Return spinbutton to current MTU value.
    return (EXIT_FAILURE);
  }

  // Determine minimum size needed to accommodate packet
  // when packet is fragmented to maximum extent.
  min = min6to4_mtu (8, data);

  // Ensure min required size is less than or equal to interface's MTU.
  if (min > ifmtu) {
    sprintf (data->error_text, "The interface's MTU is too small to accommodate packet.");
    report_error (data);
    data->ifmtu[8] = ifmtu;
    update_mtu_spinbutton (8, data);
    create_6to4_frame (8, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size is greater or equal to min required size to accommodate packet.
  if (value < min) {
    sprintf (data->error_text, "The requested MTU is too small to accommodate packet.\nValue set to minimum required to accommodate packet instead.");
    report_error (data);
    data->ifmtu[8] = min;
    update_mtu_spinbutton (8, data);
    create_6to4_frame (8, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size if <= interface's MTU.
  if (value > ifmtu) {
    sprintf (data->error_text, "The requested MTU is larger than the interface's MTU.\nValue set to the interface's MTU instead.");
    report_error (data);
    data->ifmtu[8] = ifmtu;
    update_mtu_spinbutton (8, data);
    create_6to4_frame (8, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Accept requested MTU.
  data->ifmtu[8] = value;
  create_6to4_frame (0, data);  // Update ethernet frames.

  return (EXIT_SUCCESS);
}

// 6to4 UDP Ethernet type code (16 bits) (sps.ui)
// http://www.iana.org/assignments/ethernet-numbers
int
on_entry608_changed (GtkWidget *entry608, SPSData *data) {

  ethtype_entry (entry608, 8, data);

  return (EXIT_SUCCESS);
}

// 6to4 UDP IPv4 header length (4 bits) (sps.ui)
int
on_entry635_changed (GtkWidget *entry635, SPSData *data) {

  ip4_hdrlen_entry (entry635, 8, data);

  return (EXIT_SUCCESS);
}

// 6to4 UDP IP version (4 bits) (sps.ui)
int
on_entry636_changed (GtkWidget *entry636, SPSData *data) {

  ip4_ver_entry (entry636, 8, data);

  return (EXIT_SUCCESS);
}

// 6to4 UDP Type of service (8 bits) (sps.ui)
int
on_entry637_changed (GtkWidget *entry637, SPSData *data) {

  ip4_tos_entry (entry637, 8, data);

  return (EXIT_SUCCESS);
}

// 6to4 UDP Total length of datagram (16 bits) (sps.ui)
int
on_entry638_changed (GtkWidget *entry638, SPSData *data) {

  ip4_tld_entry (entry638, 8, data);

  return (EXIT_SUCCESS);
}

// 6to4 UDP ID sequence number (16 bits) (sps.ui)
int
on_entry639_changed (GtkWidget *entry639, SPSData *data) {

  ip4_seq_entry (entry639, 8, data);

  return (EXIT_SUCCESS);
}

// 6to4 UDP Zero (1 bit) (sps.ui)
int
on_entry640_changed (GtkWidget *entry640, SPSData *data) {

  ip4_zero_entry (entry640, data->entry641, data->entry642, data->entry643, 8, data);

  return (EXIT_SUCCESS);
}

// 6to4 UDP Do not fragment flag (1 bit) (sps.ui)
int
on_entry641_changed (GtkWidget *entry641, SPSData *data) {

  ip4_dnf_entry (data->entry640, entry641, data->entry642, data->entry643, 8, data);

  return (EXIT_SUCCESS);
}

// 6to4 UDP More fragments following flag (1 bit) (sps.ui)
int
on_entry642_changed (GtkWidget *entry642, SPSData *data) {

  ip4_mff_entry (data->entry640, data->entry641, entry642, data->entry643, 8, data);

  return (EXIT_SUCCESS);
}

// 6to4 UDP Fragmentation offset (13 bits) (sps.ui)
int
on_entry643_changed (GtkWidget *entry643, SPSData *data) {

  ip4_fo_entry (data->entry640, data->entry641, data->entry642, entry643, 8, data);

  return (EXIT_SUCCESS);
}

// 6to4 UDP Time-to-Live (8 bits) (sps.ui)
int
on_entry644_changed (GtkWidget *entry644, SPSData *data) {

  ip4_ttl_entry (entry644, 8, data);

  return (EXIT_SUCCESS);
}

// 6to4 UDP Transport layer protocol (8 bits) (sps.ui)
int
on_entry645_changed (GtkWidget *entry645, SPSData *data) {

  ip4_tlp_entry (entry645, 8, data);

  return (EXIT_SUCCESS);
}

// 6to4 UDP IPv4 Source Address (sps.ui)
int
on_entry161_changed (GtkWidget *entry161, SPSData *data) {

  ip4_sip_entry (entry161, 8, data);

  if (data->ran_udp6to4_sourceip) {
    // Source IP address has been entered manually. Uncheck "Randomize" checkbutton.
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton27), FALSE);
    data->ran_udp6to4_sourceip = 0;
  }

  return (EXIT_SUCCESS);
}

// 6to4 UDP Randomize IPv4 source address (sps.ui)
int
on_checkbutton27_toggled (GtkButton *checkbutton27, SPSData *data) {

  (void) checkbutton27;  // This statement suppresses compiler warning about unused parameter.

  if (data->ran_udp6to4_sourceip) {
    data->ran_udp6to4_sourceip = 0;
  } else {
    ip4_ransip_checkbutton (&data->ran_udp6to4_sourceip, data->entry161, 8, data);
  }

  return (EXIT_SUCCESS);
}

// 6to4 UDP Destination IPv4 address (32 bits) (sps.ui)
int
on_entry646_changed (GtkWidget *entry646, SPSData *data) {

  ip4_dip_entry (entry646, 8, data);

  return (EXIT_SUCCESS);
}

// 6to4 UDP IPv4 checksum (16 bits) (sps.ui)
int
on_entry647_changed (GtkWidget *entry647, SPSData *data) {

  ip4_chksum_entry (entry647, 8, data);

  return (EXIT_SUCCESS);
}

// 6to4 UDP Restore default IPv4 values.
int
on_button602_clicked (GtkButton *button602, SPSData *data) {

  (void) button602;  // This statement suppresses compiler warning about unused parameter.

  sixto4_udp6_default (data);

  return (EXIT_SUCCESS);
}

// TCP page (IPv4) for Traceroute (traceroute.ui)

// TCP (traceroute) - Copy all data from main TCP (IPv4) page
int
on_tr_button500_clicked (GtkButton *tr_button500, SPSData *data) {

  (void) tr_button500;  // This statement suppresses compiler warning about unused parameter.

  // Copy IPv4 TCP entry data from main page to traceroute page.
  copy_tcp4_to_traceroute (data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) Destination link-layer (MAC) address (48 bits) (traceroute.ui)
int
on_tr_entry164_changed (GtkWidget *tr_entry164, SPSData *data) {

  mac_entry (tr_entry164, data->ethhdr[9].dst_mac, 9, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) Source link-layer (MAC) address (48 bits) (traceroute.ui)
int
on_tr_entry165_changed (GtkWidget *tr_entry165, SPSData *data) {

  mac_entry (tr_entry165, data->ethhdr[9].src_mac, 9, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) Interface Name to be converted to link-layer (MAC) address and find MTU. (traceroute.ui)
int
on_tr_entry166_changed (GtkWidget *tr_entry166, SPSData *data) {

  // Check for valid source interface name, check MTU, and update source MAC address.
  ifname_entry (tr_entry166, data->tr_entry165, 9, data);

  // Update MTU spinbutton.
  update_mtu_spinbutton (9, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) Maximum transmission unit (MTU) - Changed via spinbutton + or -. (traceroute.ui)
int
on_tr_spinbutton11_value_changed (GtkWidget *tr_spinbutton11, SPSData *data) {

  int value, min, ifmtu;

  // Set parent window for reporting error messages.
  data->parent = data->traceroute_window;

  // Get requested MTU value from spinbutton.
  value = (int) gtk_spin_button_get_value (GTK_SPIN_BUTTON (tr_spinbutton11));

  // Obtain interface's MTU.
  ifmtu = interface_getmtu (data->ifname[9]);
  if (ifmtu == 0) {
    update_mtu_spinbutton (9, data);  // Return spinbutton to current MTU value.
    return (EXIT_FAILURE);
  }

  // Determine minimum size needed to accommodate packet
  // when packet is fragmented to maximum extent.
  min = min4_mtu (9, data);

  // Ensure min required size is less than or equal to interface's MTU.
  if (min > ifmtu) {
    sprintf (data->error_text, "The interface's MTU is too small to accommodate packet.");
    report_error (data);
    data->ifmtu[9] = ifmtu;
    update_mtu_spinbutton (9, data);
    create_ip4_frame (9, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size is greater or equal to min required size to accommodate packet.
  if (value < min) {
    data->ifmtu[9] = min;
    update_mtu_spinbutton (9, data);
    create_ip4_frame (9, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size if <= interface's MTU.
  if (value > ifmtu) {
    data->ifmtu[9] = ifmtu;
    update_mtu_spinbutton (9, data);
    create_ip4_frame (9, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Accept requested MTU.
  data->ifmtu[9] = value;
  create_ip4_frame (9, data);  // Update ethernet frames.

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) Maximum transmission unit (MTU) - Changed via spinbutton text entry. (traceroute.ui)
int
on_tr_spinbutton11_change_value (GtkWidget *tr_spinbutton11, SPSData *data) {

  int value, min, ifmtu;

  // Set parent window for reporting error messages.
  data->parent = data->traceroute_window;

  // Get requested MTU value from spinbutton text entry.
  value = (int) gtk_spin_button_get_value (GTK_SPIN_BUTTON (tr_spinbutton11));

  // Obtain interface's MTU.
  ifmtu = interface_getmtu (data->ifname[9]);
  if (ifmtu == 0) {
    update_mtu_spinbutton (9, data);  // Return spinbutton to current MTU value.
    return (EXIT_FAILURE);
  }

  // Determine minimum size needed to accommodate packet
  // when packet is fragmented to maximum extent.
  min = min4_mtu (9, data);

  // Ensure min required size is less than or equal to interface's MTU.
  if (min > ifmtu) {
    sprintf (data->error_text, "The interface's MTU is too small to accommodate packet.");
    report_error (data);
    data->ifmtu[9] = ifmtu;
    update_mtu_spinbutton (9, data);
    create_ip4_frame (9, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size is greater or equal to min required size to accommodate packet.
  if (value < min) {
    sprintf (data->error_text, "The requested MTU is too small to accommodate packet.\nValue set to minimum required to accommodate packet instead.");
    report_error (data);
    data->ifmtu[9] = min;
    update_mtu_spinbutton (9, data);
    create_ip4_frame (9, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size if <= interface's MTU.
  if (value > ifmtu) {
    sprintf (data->error_text, "The requested MTU is larger than the interface's MTU.\nValue set to the interface's MTU instead.");
    report_error (data);
    data->ifmtu[9] = ifmtu;
    update_mtu_spinbutton (9, data);
    create_ip4_frame (9, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Accept requested MTU.
  data->ifmtu[9] = value;
  create_ip4_frame (9, data);  // Update ethernet frames.

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) Ethernet type code (16 bits) (traceroute.ui)
// http://www.iana.org/assignments/ethernet-numbers
int
on_tr_entry167_changed (GtkWidget *tr_entry167, SPSData *data) {

  ethtype_entry (tr_entry167, 9, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) IPv4 header length (4 bits) (traceroute.ui)
int
on_tr_entry168_changed (GtkWidget *tr_entry168, SPSData *data) {

  ip4_hdrlen_entry (tr_entry168, 9, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) IP version (4 bits) (traceroute.ui)
int
on_tr_entry169_changed (GtkWidget *tr_entry169, SPSData *data) {

  ip4_ver_entry (tr_entry169, 9, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) Type of service (8 bits) (traceroute.ui)
int
on_tr_entry170_changed (GtkWidget *tr_entry170, SPSData *data) {

  ip4_tos_entry (tr_entry170, 9, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) Total length of datagram (16 bits) (traceroute.ui)
int
on_tr_entry171_changed (GtkWidget *tr_entry171, SPSData *data) {

  ip4_tld_entry (tr_entry171, 9, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) ID sequence number (16 bits) (traceroute.ui)
int
on_tr_entry172_changed (GtkWidget *tr_entry172, SPSData *data) {

  ip4_seq_entry (tr_entry172, 9, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) Zero (1 bit) (traceroute.ui)
int
on_tr_entry173_changed (GtkWidget *tr_entry173, SPSData *data) {

  ip4_zero_entry (tr_entry173, data->tr_entry174, data->tr_entry175, data->tr_entry176, 9, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) Do not fragment flag (1 bit) (traceroute.ui)
int
on_tr_entry174_changed (GtkWidget *tr_entry174, SPSData *data) {

  ip4_dnf_entry (data->tr_entry173, tr_entry174, data->tr_entry175, data->tr_entry176, 9, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) More fragments following flag (1 bit) (traceroute.ui)
int
on_tr_entry175_changed (GtkWidget *tr_entry175, SPSData *data) {

  ip4_mff_entry (data->tr_entry173, data->tr_entry174, tr_entry175, data->tr_entry176, 9, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) Fragmentation offset (13 bits) (traceroute.ui)
int
on_tr_entry176_changed (GtkWidget *tr_entry176, SPSData *data) {

  ip4_fo_entry (data->tr_entry173, data->tr_entry174, data->tr_entry175, tr_entry176, 9, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) Time-to-Live (8 bits) (traceroute.ui)
int
on_tr_entry177_changed (GtkWidget *tr_entry177, SPSData *data) {

  ip4_ttl_entry (tr_entry177, 9, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute.ui) Transport layer protocol (8 bits) (traceroute.ui)
int
on_tr_entry178_changed (GtkWidget *tr_entry178, SPSData *data) {

  ip4_tlp_entry (tr_entry178, 9, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) Source IPv4 address (32 bits) (traceroute.ui)
int
on_tr_entry179_changed (GtkWidget *tr_entry179, SPSData *data) {

  ip4_sip_entry (tr_entry179, 9, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) Destination IPv4 address (32 bits) (traceroute.ui)
int
on_tr_entry180_changed (GtkWidget *tr_entry180, SPSData *data) {

  ip4_dip_entry (tr_entry180, 9, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) IPv4 checksum (16 bits) (traceroute.ui)
int
on_tr_entry181_changed (GtkWidget *tr_entry181, SPSData *data) {

  ip4_chksum_entry (tr_entry181, 9, data);

  return (EXIT_SUCCESS);
}

// IP options for IPv4 TCP for traceroute (traceroute.ui)

// TCP (IPv4 traceroute) IP option decimal data entry (traceroute.ui)
int
on_tr_radiobutton49_toggled (GtkButton *tr_radiobutton49, SPSData *data) {

  (void) tr_radiobutton49;  // This statement suppresses compiler warning about unused parameter.

  if (data->dec_hex_ipopt_tcp4_tr) {
    data->dec_hex_ipopt_tcp4_tr = 0;
  } else {
    data->dec_hex_ipopt_tcp4_tr = 1;
  }

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) IP option hexadecimal data entry (traceroute.ui)
int
on_tr_radiobutton50_toggled (GtkButton *tr_radiobutton50) {

  (void) tr_radiobutton50;  // This statement suppresses compiler warning about unused parameter.

  // on_radiobutton49_toggled() will change value of data->dec_hex_ipopt_tcp4_tr.

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) IP option data (traceroute.ui)
int
on_tr_entry416_changed (GtkWidget *tr_entry416, SPSData *data) {

  data_entry (tr_entry416, &data->ip_optlenbuf[9], MAX_IP4OPTLEN, data->ip_optionsbuf[9], data->dec_hex_ipopt_tcp4_tr, data->traceroute_window, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) Insert IPv4 option after another, as specified by user. (traceroute.ui)
int
on_tr_button74_clicked (GtkButton *tr_button74, SPSData *data) {

  (void) tr_button74;  // This statement suppresses compiler warning about unused parameter.

  ip4_option (0, 9, data->ipopt_tcp4_tr_after, data->tr_textview34, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) Number of option after which to insert new IPv4 option. (traceroute.ui)
int
on_tr_entry417_changed (GtkWidget *tr_entry417, SPSData *data) {

  ip4_opt_insert_after_entry (tr_entry417, 9, &data->ipopt_tcp4_tr_after, data);

  return (EXIT_SUCCESS);
}

// Append IP option for IPv4 TCP packet for traceroute. (traceroute.ui)
int
on_tr_button75_clicked (GtkButton *tr_button75, SPSData *data) {

  (void) tr_button75;  // This statement suppresses compiler warning about unused parameter.

  ip4_option (2, 9, 0, data->tr_textview34, data);

  return (EXIT_SUCCESS);
}

// View IPv4 header options for TCP for traceroute. (traceroute.ui)
int
on_tr_button76_clicked (GtkButton *tr_button76, SPSData *data) {

  (void) tr_button76;  // This statement suppresses compiler warning about unused parameter.

  opt_view (9, 0, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) Number of IPv4 option to remove. (traceroute.ui)
int
on_tr_entry418_changed (GtkWidget *tr_entry418, SPSData *data) {

  ip4_opt_remove_entry (tr_entry418, 9, &data->ipopt_tcp4_tr_remove, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) Remove selected IPv4 option. (traceroute.ui)
int
on_tr_button77_clicked (GtkButton *tr_button77, SPSData *data) {

  (void) tr_button77;  // This statement suppresses compiler warning about unused parameter.

  ip4_option (1, 9, data->ipopt_tcp4_tr_remove, data->tr_textview34, data);

  return (EXIT_SUCCESS);
}

// TCP header (IPv4) for traceroute (traceroute.ui)

// TCP (IPv4 traceroute) source port (16 bits) (traceroute.ui)
int
on_tr_entry182_changed (GtkWidget *tr_entry182, SPSData *data) {

  tcp_sport_entry (tr_entry182, 9, data);

  if (data->ran_tcp4_tr_sourceport) {
    // Source port has been entered manually. Uncheck "Randomize" checkbutton.
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->tr_checkbutton30), FALSE);
    data->ran_tcp4_tr_sourceport = 0;
  }

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) Randomize TCP source port number. (traceroute.ui)
int
on_tr_checkbutton30_toggled (GtkButton *tr_checkbutton30, SPSData *data) {

  (void) tr_checkbutton30;  // This statement suppresses compiler warning about unused parameter.

  if (data->ran_tcp4_tr_sourceport) {
    data->ran_tcp4_tr_sourceport = 0;
  } else {
    tcp_ran_sport (&data->ran_tcp4_tr_sourceport, 9, data->tr_entry182, data);
  }

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) destination port (16 bits) (traceroute.ui)
int
on_tr_entry183_changed (GtkWidget *tr_entry183, SPSData *data) {

  tcp_dport_entry (tr_entry183, 9, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) sequence number (32 bits) (traceroute.ui)
int
on_tr_entry184_changed (GtkWidget *tr_entry184, SPSData *data) {

  tcp_seqnum_entry (tr_entry184, 9, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) acknowledgement number (32 bits) (traceroute.ui)
int
on_tr_entry185_changed (GtkWidget *tr_entry185, SPSData *data) {

  tcp_acknum_entry (tr_entry185, 9, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) reserved (4 bits) (traceroute.ui)
int
on_tr_entry186_changed (GtkWidget *tr_entry186, SPSData *data) {

  tcp_res_entry (tr_entry186, 9, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute.ui) data offset (4 bits) (traceroute.ui)
int
on_tr_entry187_changed (GtkWidget *tr_entry187, SPSData *data) {

  tcp_off_entry (tr_entry187, 9, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) FIN flag (1 bit) (traceroute.ui)
int
on_tr_entry188_changed (GtkWidget *tr_entry188, SPSData *data) {

  tcp_fin_entry (tr_entry188, 9, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) SYN flag (1 bit) (traceroute.ui)
int
on_tr_entry189_changed (GtkWidget *tr_entry189, SPSData *data) {

  tcp_syn_entry (tr_entry189, 9, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) RST flag (1 bit) (traceroute.ui)
int
on_tr_entry190_changed (GtkWidget *tr_entry190, SPSData *data) {

  tcp_rst_entry (tr_entry190, 9, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) PSH flag (1 bit) (traceroute.ui)
int
on_tr_entry191_changed (GtkWidget *tr_entry191, SPSData *data) {

  tcp_psh_entry (tr_entry191, 9, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) ACK flag (1 bit) (traceroute.ui)
int
on_tr_entry192_changed (GtkWidget *tr_entry192, SPSData *data) {

  tcp_ack_entry (tr_entry192, 9, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) URG flag (1 bit) (traceroute.ui)
int
on_tr_entry193_changed (GtkWidget *tr_entry193, SPSData *data) {

  tcp_urg_entry (tr_entry193, 9, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) ECE flag (1 bit) (traceroute.ui)
int
on_tr_entry194_changed (GtkWidget *tr_entry194, SPSData *data) {

  tcp_ece_entry (tr_entry194, 9, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) CWR flag (1 bit) (traceroute.ui)
int
on_tr_entry195_changed (GtkWidget *tr_entry195, SPSData *data) {

  tcp_cwr_entry (tr_entry195, 9, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) window size (16 bits) (traceroute.ui)
int
on_tr_entry196_changed (GtkWidget *tr_entry196, SPSData *data) {

  tcp_win_entry (tr_entry196, 9, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) checksum (16 bits) (traceroute.ui)
int
on_tr_entry197_changed (GtkWidget *tr_entry197, SPSData *data) {

  tcp_chksum_entry (tr_entry197, 9, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) urgent pointer (16 bits) (traceroute.ui)
int
on_tr_entry198_changed (GtkWidget *tr_entry198, SPSData *data) {

  tcp_urgptr_entry (tr_entry198, 9, data);

  return (EXIT_SUCCESS);
}

// TCP options (IPv4) for traceroute (traceroute.ui)

// TCP (IPv4 traceroute) IP option decimal data entry (traceroute.ui)
int
on_tr_radiobutton51_toggled (GtkButton *tr_radiobutton51, SPSData *data) {

  (void) tr_radiobutton51;  // This statement suppresses compiler warning about unused parameter.

  if (data->dec_hex_tcpopt_tcp4_tr) {
    data->dec_hex_tcpopt_tcp4_tr = 0;
  } else {
    data->dec_hex_tcpopt_tcp4_tr = 1;
  }

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) IP option hexadecimal data entry (traceroute.ui)
int
on_tr_radiobutton52_toggled (GtkButton *tr_radiobutton52) {

  (void) tr_radiobutton52;  // This statement suppresses compiler warning about unused parameter.

  // on_radiobutton51_toggled() will change value of data->dec_hex_tcpopt_tcp4_tr.

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) option data (IPv4) for traceroute (traceroute.ui)
int
on_tr_entry419_changed (GtkWidget *tr_entry419, SPSData *data) {

  data_entry (tr_entry419, &data->tcp_optlenbuf[9], MAX_TCPOPTLEN, data->tcp_optionsbuf[9], data->dec_hex_tcpopt_tcp4_tr, data->traceroute_window, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) Insert TCP option after another, as specified by user. (traceroute.ui)
int
on_tr_button78_clicked (GtkButton *tr_button78, SPSData *data) {

  (void) tr_button78;  // This statement suppresses compiler warning about unused parameter.

  tcp4_option (0, 9, data->tcpopt_tcp4_tr_after, data->tr_textview36, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) Number of option after which to insert new TCP option. (traceroute.ui)
int
on_tr_entry420_changed (GtkWidget *tr_entry420, SPSData *data) {

  tcp_opt_insert_after_entry (tr_entry420, 9, &data->tcpopt_tcp4_tr_after, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) Append TCP option for IPv4 TCP packet for traceroute. (traceroute.ui)
int
on_tr_button79_clicked (GtkButton *tr_button79, SPSData *data) {

  (void) tr_button79;  // This statement suppresses compiler warning about unused parameter.

  tcp4_option (2, 9, 0, data->tr_textview36, data);

  return (EXIT_SUCCESS);
}

// View TCP options for IPv4 TCP for traceroute. (traceroute.ui)
int
on_tr_button80_clicked (GtkButton *tr_button80, SPSData *data) {

  (void) tr_button80;  // This statement suppresses compiler warning about unused parameter.

  opt_view (9, 1, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) Number of TCP option to remove. (traceroute.ui)
int
on_tr_entry421_changed (GtkWidget *tr_entry421, SPSData *data) {

  tcp_opt_remove_entry (tr_entry421, 9, &data->tcpopt_tcp4_tr_remove, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) Remove selected TCP option. (traceroute.ui)
int
on_tr_button81_clicked (GtkButton *tr_button81, SPSData *data) {

  (void) tr_button81;  // This statement suppresses compiler warning about unused parameter.

  tcp4_option (1, 9, data->tcpopt_tcp4_tr_remove, data->tr_textview36, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) ASCII data entry (traceroute.ui)
int
on_tr_radiobutton29_toggled (GtkButton *tr_radiobutton29, SPSData *data) {

  (void) tr_radiobutton29;  // This statement suppresses compiler warning about unused parameter.

  if (data->ascii_hex_tcp4_tr) {
    data->ascii_hex_tcp4_tr = 0;
  } else {
    data->ascii_hex_tcp4_tr = 1;
  }

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) Hexadecimal data entry (traceroute.ui)
int
on_tr_radiobutton30_toggled (GtkButton *tr_radiobutton30) {

  (void) tr_radiobutton30;  // This statement suppresses compiler warning about unused parameter.

  // on_radiobutton29_toggled() will change value of data->ascii_hex_tcp4_tr.

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) data from keyboard (traceroute.ui)
int
on_tr_entry199_changed (GtkWidget *tr_entry199, SPSData *data) {

  key_data (tr_entry199, 9, data->ascii_hex_tcp4_tr, data);

  if (data->nframes[9] > 1) {
    sprintf (data->warning_text, "Packet will be fragmented to accommodate TCP data.");
    data->parent = data->traceroute_window;
    report_warning (data);
  }

  return (EXIT_SUCCESS);
}

// TCP (IPv4 traceroute) data from file (traceroute.ui)
int
on_tr_button29_clicked (GtkButton *tr_button29, SPSData *data) {

  (void) tr_button29;  // This statement suppresses compiler warning about unused parameter.

  file_payload_data (data->tr_entry199, 9, data);

  if (data->nframes[9] > 1) {
    sprintf (data->warning_text, "Packet will be fragmented to accommodate TCP data.");
    data->parent = data->traceroute_window;
    report_warning (data);
  }

  return (EXIT_SUCCESS);
}

// Unload file data for traceroute TCP (IPv4) (traceroute.ui)
int
on_tr_button206_clicked (GtkButton *tr_button206, SPSData *data) {

  (void) tr_button206;  // This statement suppresses compiler warning about unused parameter.
  GtkTextBuffer *tr_textbuffer66;

  // Remove file data.
  data->payloadlen[9] = 0;
  memset (data->payload[9], 0, IP_MAXPACKET * sizeof (uint8_t));

  // Update ethernet frame.
  create_ip4_frame (9, data);

  tr_textbuffer66 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->tr_textview66));
  gtk_text_buffer_set_text (tr_textbuffer66, "", -1);

  return (EXIT_SUCCESS);
}

// Call tcp4_tr_default() (traceroute.ui)
int
on_tr_button30_clicked (GtkButton *tr_button30, SPSData *data) {

  (void) tr_button30;  // This statement suppresses compiler warning about unused parameter.

  tcp4_tr_default (data);
  tcp4_tr_show (data);

  // Update ethernet frame
  create_ip4_frame (9, data);

  return (EXIT_SUCCESS);
}

// ICMP page (IPv4) for traceroute

// ICMP (IPv4 traceroute) - Copy all data from main ICMP (IPv4) page (traceroute.ui)
int
on_tr_button501_clicked (GtkButton *tr_button501, SPSData *data) {

  (void) tr_button501;  // This statement suppresses compiler warning about unused parameter.

  // Copy IPv4 ICMP entry data from main page to traceroute page.
  copy_icmp4_to_traceroute (data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv4 traceroute) Destination link-layer (MAC) address (48 bits) (traceroute.ui)
int
on_tr_entry200_changed (GtkWidget *tr_entry200, SPSData *data) {

  mac_entry (tr_entry200, data->ethhdr[10].dst_mac, 10, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv4 traceroute) Source link-layer (MAC) address (48 bits) (traceroute.ui)
int
on_tr_entry201_changed (GtkWidget *tr_entry201, SPSData *data) {

  mac_entry (tr_entry201, data->ethhdr[10].src_mac, 10, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv4 traceroute) Interface name to be converted to link-layer (MAC) address and find MTU. (traceroute.ui)
int
on_tr_entry202_changed (GtkWidget *tr_entry202, SPSData *data) {

  // Check for valid source interface name, check MTU, and update source MAC address.
  ifname_entry (tr_entry202, data->tr_entry201, 10, data);

  // Update MTU spinbutton.
  update_mtu_spinbutton (10, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv4 traceroute) Maximum transmission unit (MTU) - Changed via spinbutton + or -. (traceroute.ui)
int
on_tr_spinbutton12_value_changed (GtkWidget *tr_spinbutton12, SPSData *data) {

  int value, min, ifmtu;

  // Set parent window for reporting error messages.
  data->parent = data->traceroute_window;

  // Get requested MTU value from spinbutton.
  value = (int) gtk_spin_button_get_value (GTK_SPIN_BUTTON (tr_spinbutton12));

  // Obtain interface's MTU.
  ifmtu = interface_getmtu (data->ifname[10]);
  if (ifmtu == 0) {
    update_mtu_spinbutton (10, data);  // Return spinbutton to current MTU value.
    return (EXIT_FAILURE);
  }

  // Determine minimum size needed to accommodate packet
  // when packet is fragmented to maximum extent.
  min = min4_mtu (10, data);

  // Ensure min required size is less than or equal to interface's MTU.
  if (min > ifmtu) {
    sprintf (data->error_text, "The interface's MTU is too small to accommodate packet.");
    report_error (data);
    data->ifmtu[10] = ifmtu;
    update_mtu_spinbutton (10, data);
    create_ip4_frame (10, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size is greater or equal to min required size to accommodate packet.
  if (value < min) {
    data->ifmtu[10] = min;
    update_mtu_spinbutton (10, data);
    create_ip4_frame (10, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size if <= interface's MTU.
  if (value > ifmtu) {
    data->ifmtu[10] = ifmtu;
    update_mtu_spinbutton (10, data);
    create_ip4_frame (10, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Accept requested MTU.
  data->ifmtu[10] = value;
  create_ip4_frame (10, data);  // Update ethernet frames.

  return (EXIT_SUCCESS);
}

// ICMP (IPv4 traceroute) Maximum transmission unit (MTU) - Changed via spinbutton text entry. (traceroute.ui)
int
on_tr_spinbutton12_change_value (GtkWidget *tr_spinbutton12, SPSData *data) {

  int value, min, ifmtu;

  // Set parent window for reporting error messages.
  data->parent = data->traceroute_window;

  // Get requested MTU value from spinbutton text entry.
  value = (int) gtk_spin_button_get_value (GTK_SPIN_BUTTON (tr_spinbutton12));

  // Obtain interface's MTU.
  ifmtu = interface_getmtu (data->ifname[10]);
  if (ifmtu == 0) {
    update_mtu_spinbutton (10, data);  // Return spinbutton to current MTU value.
    return (EXIT_FAILURE);
  }

  // Determine minimum size needed to accommodate packet
  // when packet is fragmented to maximum extent.
  min = min4_mtu (10, data);

  // Ensure min required size is less than or equal to interface's MTU.
  if (min > ifmtu) {
    sprintf (data->error_text, "The interface's MTU is too small to accommodate packet.");
    report_error (data);
    data->ifmtu[10] = ifmtu;
    update_mtu_spinbutton (10, data);
    create_ip4_frame (10, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size is greater or equal to min required size to accommodate packet.
  if (value < min) {
    sprintf (data->error_text, "The requested MTU is too small to accommodate packet.\nValue set to minimum required to accommodate packet instead.");
    report_error (data);
    data->ifmtu[10] = min;
    update_mtu_spinbutton (10, data);
    create_ip4_frame (10, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size if <= interface's MTU.
  if (value > ifmtu) {
    sprintf (data->error_text, "The requested MTU is larger than the interface's MTU.\nValue set to the interface's MTU instead.");
    report_error (data);
    data->ifmtu[10] = ifmtu;
    update_mtu_spinbutton (10, data);
    create_ip4_frame (10, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Accept requested MTU.
  data->ifmtu[10] = value;
  create_ip4_frame (10, data);  // Update ethernet frames.

  return (EXIT_SUCCESS);
}

// ICMP (IPv4 traceroute) Ethernet type code (16 bits) (traceroute.ui)
// http://www.iana.org/assignments/ethernet-numbers
int
on_tr_entry203_changed (GtkWidget *tr_entry203, SPSData *data) {

  ethtype_entry (tr_entry203, 10, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv4 traceroute) IP header length (4 bits) (traceroute.ui)
int
on_tr_entry204_changed (GtkWidget *tr_entry204, SPSData *data) {

  ip4_hdrlen_entry (tr_entry204, 10, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv4 traceroute) IP version (4 bits) (traceroute.ui)
int
on_tr_entry205_changed (GtkWidget *tr_entry205, SPSData *data) {

  ip4_ver_entry (tr_entry205, 10, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv4 traceroute) Type of service (8 bits) (traceroute.ui)
int
on_tr_entry206_changed (GtkWidget *tr_entry206, SPSData *data) {

  ip4_tos_entry (tr_entry206, 10, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv4 traceroute) Total length of datagram (16 bits) (traceroute.ui)
int
on_tr_entry207_changed (GtkWidget *tr_entry207, SPSData *data) {

  ip4_tld_entry (tr_entry207, 10, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv4 traceroute) ID sequence number (16 bits) (traceroute.ui)
int
on_tr_entry208_changed (GtkWidget *tr_entry208, SPSData *data) {

  ip4_seq_entry (tr_entry208, 10, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv4 traceroute) Zero (1 bit) (traceroute.ui)
int
on_tr_entry209_changed (GtkWidget *tr_entry209, SPSData *data) {

  ip4_zero_entry (tr_entry209, data->tr_entry210, data->tr_entry211, data->tr_entry212, 10, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv4 traceroute) Do not fragment flag (1 bit) (traceroute.ui)
int
on_tr_entry210_changed (GtkWidget *tr_entry210, SPSData *data) {

  ip4_dnf_entry (data->tr_entry209, tr_entry210, data->tr_entry211, data->tr_entry212, 10, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv4 traceroute) More fragments following flag (1 bit) (traceroute.ui)
int
on_tr_entry211_changed (GtkWidget *tr_entry211, SPSData *data) {

  ip4_mff_entry (data->tr_entry209, data->tr_entry210, tr_entry211, data->tr_entry212, 10, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv4 traceroute) Fragmentation offset (13 bits) ICMP (IPv4 traceroute)
int
on_tr_entry212_changed (GtkWidget *tr_entry212, SPSData *data) {

  ip4_fo_entry (data->tr_entry209, data->tr_entry210, data->tr_entry211, tr_entry212, 10, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv4 traceroute) Time-to-Live (8 bits) (traceroute.ui)
int
on_tr_entry213_changed (GtkWidget *tr_entry213, SPSData *data) {

  ip4_ttl_entry (tr_entry213, 10, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv4 traceroute) Transport layer protocol (8 bits) (traceroute.ui)
int
on_tr_entry214_changed (GtkWidget *tr_entry214, SPSData *data) {

  ip4_tlp_entry (tr_entry214, 10, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv4 traceroute) Source IPv4 address (32 bits) (traceroute.ui)
int
on_tr_entry215_changed (GtkWidget *tr_entry215, SPSData *data) {

  ip4_sip_entry (tr_entry215, 10, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv4 traceroute) Destination IPv4 address (32 bits) (traceroute.ui)
int
on_tr_entry216_changed (GtkWidget *tr_entry216, SPSData *data) {

  ip4_dip_entry (tr_entry216, 10, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv4 traceroute) IPv4 checksum (16 bits) (traceroute.ui)
int
on_tr_entry217_changed (GtkWidget *tr_entry217, SPSData *data) {

  ip4_chksum_entry (tr_entry217, 10, data);

  return (EXIT_SUCCESS);
}

// IP options for IPv4 ICMP for Traceroute

// ICMP (IPv4 traceroute) Decimal option data entry (traceroute.ui)
int
on_tr_radiobutton55_toggled (GtkButton *tr_radiobutton55, SPSData *data) {

  (void) tr_radiobutton55;  // This statement suppresses compiler warning about unused parameter.

  if (data->dec_hex_ipopt_icmp4_tr) {
    data->dec_hex_ipopt_icmp4_tr = 0;
  } else {
    data->dec_hex_ipopt_icmp4_tr = 1;
  }

  return (EXIT_SUCCESS);
}

// ICMP (IPv4 traceroute) Hexadecimal option data entry (traceroute.ui)
int
on_tr_radiobutton56_toggled (GtkButton *tr_radiobutton56) {

  (void) tr_radiobutton56;  // This statement suppresses compiler warning about unused parameter.

  // on_radiobutton55_toggled() will change value of data->dec_hex_ipopt_icmp4_tr.

  return (EXIT_SUCCESS);
}

// ICMP (IPv4 traceroute) IPv4 option data (traceroute.ui)
int
on_tr_entry425_changed (GtkWidget *tr_entry425, SPSData *data) {

  data_entry (tr_entry425, &data->ip_optlenbuf[10], MAX_IP4OPTLEN, data->ip_optionsbuf[10], data->dec_hex_ipopt_icmp4_tr, data->traceroute_window, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv4 traceroute) Insert IPv4 option after another, as specified by user. (traceroute.ui)
int
on_tr_button86_clicked (GtkButton *tr_button86, SPSData *data) {

  (void) tr_button86;  // This statement suppresses compiler warning about unused parameter.

  ip4_option (0, 10, data->ipopt_icmp4_tr_after, data->tr_textview40, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv4 traceroute) Number of option after which to insert new option. (traceroute.ui)
int
on_tr_entry426_changed (GtkWidget *tr_entry426, SPSData *data) {

  ip4_opt_insert_after_entry (tr_entry426, 10, &data->ipopt_icmp4_tr_after, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv4 traceroute) Append IP option for IPv4 ICMP packet. (traceroute.ui)
int
on_tr_button87_clicked (GtkButton *tr_button87, SPSData *data) {

  (void) tr_button87;  // This statement suppresses compiler warning about unused parameter.

  ip4_option (2, 10, 0, data->tr_textview40, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv4 traceroute) View IPv4 header options for ICMP for traceroute. (traceroute.ui)
int
on_tr_button88_clicked (GtkButton *tr_button88, SPSData *data) {

  (void) tr_button88;  // This statement suppresses compiler warning about unused parameter.

  opt_view (10, 0, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv4 traceroute) Number of IPv4 option to remove. (traceroute.ui)
int
on_tr_entry427_changed (GtkWidget *tr_entry427, SPSData *data) {

  ip4_opt_remove_entry (tr_entry427, 10, &data->ipopt_icmp4_tr_remove, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv4 traceroute) Remove selected IPv4 option. (traceroute.ui)
int
on_tr_button89_clicked (GtkButton *tr_button89, SPSData *data) {

  (void) tr_button89;  // This statement suppresses compiler warning about unused parameter.

  ip4_option (1, 10, data->ipopt_icmp4_tr_remove, data->tr_textview40, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv4 traceroute) ICMP header (IPv4) for trcaeroute (traceroute.ui)

// Message Type (8 bits)
int
on_tr_entry218_changed (GtkWidget *tr_entry218, SPSData *data) {

  icmp_mt_entry (tr_entry218, 10, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv4 traceroute) Message Code (8 bits) (traceroute.ui)
int
on_tr_entry219_changed (GtkWidget *tr_entry219, SPSData *data) {

  icmp_mc_entry (tr_entry219, 10, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv4 traceroute) ICMP checksum (16 bits) (traceroute.ui)
int
on_tr_entry220_changed (GtkWidget *tr_entry220, SPSData *data) {

  icmp_chksum_entry (tr_entry220, 10, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv4 traceroute) ID (16 bits) (traceroute.ui)
int
on_tr_entry221_changed (GtkWidget *tr_entry221, SPSData *data) {

  icmp_id_entry (tr_entry221, 10, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv4 traceroute) Sequence number (16 bits) (traceroute.ui)
int
on_tr_entry222_changed (GtkWidget *tr_entry222, SPSData *data) {

  icmp_seq_entry (tr_entry222, 10, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv4 traceroute) ASCII data entry (traceroute.ui)
int
on_tr_radiobutton31_toggled (GtkButton *tr_radiobutton31, SPSData *data) {

  (void) tr_radiobutton31;  // This statement suppresses compiler warning about unused parameter.

  if (data->ascii_hex_icmp4_tr) {
    data->ascii_hex_icmp4_tr = 0;
  } else {
    data->ascii_hex_icmp4_tr = 1;
  }

  return (EXIT_SUCCESS);
}

// ICMP (IPv4 traceroute) Hexadecimal data entry (traceroute.ui)
int
on_tr_radiobutton32_toggled (GtkButton *tr_radiobutton32) {

  (void) tr_radiobutton32;  // This statement suppresses compiler warning about unused parameter.

  // on_radiobutton31_toggled() will change value of data->ascii_hex_icmp4_tr.

  return (EXIT_SUCCESS);
}

// ICMP (IPv4 traceroute) data from keyboard (traceroute.ui)
int
on_tr_entry223_changed (GtkWidget *tr_entry223, SPSData *data) {

  key_data (tr_entry223, 10, data->ascii_hex_icmp4_tr, data);

  if (data->nframes[10] > 1) {
    sprintf (data->warning_text, "Packet will be fragmented to accommodate ICMP data.");
    data->parent = data->traceroute_window;
    report_warning (data);
  }

  return (EXIT_SUCCESS);
}

// ICMP (IPv4 traceroute) data from file (traceroute.ui)
int
on_tr_button32_clicked (GtkButton *tr_button32, SPSData *data) {

  (void) tr_button32;  // This statement suppresses compiler warning about unused parameter.

  file_payload_data (data->tr_entry223, 10, data);

  if (data->nframes[10] > 1) {
    sprintf (data->warning_text, "Packet will be fragmented to accommodate ICMP data.");
    data->parent = data->traceroute_window;
    report_warning (data);
  }

  return (EXIT_SUCCESS);
}

// Unload file data for traceroute ICMP (IPv4) (traceroute.ui)
int
on_tr_button207_clicked (GtkButton *tr_button207, SPSData *data) {

  (void) tr_button207;  // This statement suppresses compiler warning about unused parameter.
  GtkTextBuffer *tr_textbuffer67;

  // Remove file data.
  data->payloadlen[10] = 0;
  memset (data->payload[10], 0, IP_MAXPACKET * sizeof (uint8_t));

  // Update ethernet frame.
  create_ip4_frame (10, data);

  tr_textbuffer67 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->tr_textview67));
  gtk_text_buffer_set_text (tr_textbuffer67, "", -1);

  return (EXIT_SUCCESS);
}

// Call icmp4_tr_default() (traceroute.ui)
int
on_tr_button33_clicked (GtkButton *tr_button33, SPSData *data) {

  (void) tr_button33;  // This statement suppresses compiler warning about unused parameter.

  icmp4_tr_default (data);
  icmp4_tr_show (data);

  return (EXIT_SUCCESS);
}

// UDP page (IPv4) for traceroute

// UDP (IPv4 traceroute) - Copy all data from main UDP (IPv4) page (traceroute.ui)
int
on_tr_button502_clicked (GtkButton *tr_button502, SPSData *data) {

  (void) tr_button502;  // This statement suppresses compiler warning about unused parameter.

  // Copy IPv4 UDP entry data from main page to traceroute page.
  copy_udp4_to_traceroute (data);

  return (EXIT_SUCCESS);
}

// UDP (IPv4 traceroute) Destination link-layer (MAC) address (48 bits) (traceroute.ui)
int
on_tr_entry224_changed (GtkWidget *tr_entry224, SPSData *data) {

  mac_entry (tr_entry224, data->ethhdr[11].dst_mac, 11, data);

  return (EXIT_SUCCESS);
}

// UDP (IPv4 traceroute) Source link-layer (MAC) address (48 bits) (traceroute.ui)
int
on_tr_entry225_changed (GtkWidget *tr_entry225, SPSData *data) {

  mac_entry (tr_entry225, data->ethhdr[11].src_mac, 11, data);

  return (EXIT_SUCCESS);
}


// UDP (IPv4 traceroute) This host's interface name to be converted to link-layer (MAC) address and find MTU. (traceroute.ui)
int
on_tr_entry226_changed (GtkWidget *tr_entry226, SPSData *data) {

  // Check for valid source interface name, check MTU, and update source MAC address.
  ifname_entry (tr_entry226, data->tr_entry225, 11, data);

  // Update MTU spinbutton.
  update_mtu_spinbutton (11, data);

  return (EXIT_SUCCESS);
}

// UDP (IPv4 traceroute) Maximum transmission unit (MTU) - Changed via spinbutton + or -. (traceroute.ui)
int
on_tr_spinbutton13_value_changed (GtkWidget *tr_spinbutton13, SPSData *data) {

  int value, min, ifmtu;

  // Set parent window for reporting error messages.
  data->parent = data->traceroute_window;

  // Get requested MTU value from spinbutton.
  value = (int) gtk_spin_button_get_value (GTK_SPIN_BUTTON (tr_spinbutton13));

  // Obtain interface's MTU.
  ifmtu = interface_getmtu (data->ifname[11]);
  if (ifmtu == 0) {
    update_mtu_spinbutton (11, data);  // Return spinbutton to current MTU value.
    return (EXIT_FAILURE);
  }

  // Determine minimum size needed to accommodate packet
  // when packet is fragmented to maximum extent.
  min = min4_mtu (11, data);

  // Ensure min required size is less than or equal to interface's MTU.
  if (min > ifmtu) {
    sprintf (data->error_text, "The interface's MTU is too small to accommodate packet.");
    report_error (data);
    data->ifmtu[11] = ifmtu;
    update_mtu_spinbutton (11, data);
    create_ip4_frame (11, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size is greater or equal to min required size to accommodate packet.
  if (value < min) {
    data->ifmtu[11] = min;
    update_mtu_spinbutton (11, data);
    create_ip4_frame (11, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size if <= interface's MTU.
  if (value > ifmtu) {
    data->ifmtu[11] = ifmtu;
    update_mtu_spinbutton (11, data);
    create_ip4_frame (11, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Accept requested MTU.
  data->ifmtu[11] = value;
  create_ip4_frame (11, data);  // Update ethernet frames.

  return (EXIT_SUCCESS);
}

// UDP (IPv4 traceroute) Maximum transmission unit (MTU) - Changed via spinbutton text entry. (traceroute.ui)
int
on_tr_spinbutton13_change_value (GtkWidget *tr_spinbutton13, SPSData *data) {

  int value, min, ifmtu;

  // Set parent window for reporting error messages.
  data->parent = data->traceroute_window;

  // Get requested MTU value from spinbutton text entry.
  value = (int) gtk_spin_button_get_value (GTK_SPIN_BUTTON (tr_spinbutton13));

  // Obtain interface's MTU.
  ifmtu = interface_getmtu (data->ifname[11]);
  if (ifmtu == 0) {
    update_mtu_spinbutton (11, data);  // Return spinbutton to current MTU value.
    return (EXIT_FAILURE);
  }

  // Determine minimum size needed to accommodate packet
  // when packet is fragmented to maximum extent.
  min = min4_mtu (11, data);

  // Ensure min required size is less than or equal to interface's MTU.
  if (min > ifmtu) {
    sprintf (data->error_text, "The interface's MTU is too small to accommodate packet.");
    report_error (data);
    data->ifmtu[11] = ifmtu;
    update_mtu_spinbutton (11, data);
    create_ip4_frame (11, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size is greater or equal to min required size to accommodate packet.
  if (value < min) {
    sprintf (data->error_text, "The requested MTU is too small to accommodate packet.\nValue set to minimum required to accommodate packet instead.");
    report_error (data);
    data->ifmtu[11] = min;
    update_mtu_spinbutton (11, data);
    create_ip4_frame (11, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size if <= interface's MTU.
  if (value > ifmtu) {
    sprintf (data->error_text, "The requested MTU is larger than the interface's MTU.\nValue set to the interface's MTU instead.");
    report_error (data);
    data->ifmtu[11] = ifmtu;
    update_mtu_spinbutton (11, data);
    create_ip4_frame (11, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Accept requested MTU.
  data->ifmtu[11] = value;
  create_ip4_frame (11, data);  // Update ethernet frames.

  return (EXIT_SUCCESS);
}

// UDP (IPv4 traceroute) Ethernet type code (16 bits) (traceroute.ui)
// http://www.iana.org/assignments/ethernet-numbers
int
on_tr_entry227_changed (GtkWidget *tr_entry227, SPSData *data) {

  ethtype_entry (tr_entry227, 11, data);

  return (EXIT_SUCCESS);
}

// UDP (IPv4 traceroute) IP header length (4 bits) (traceroute.ui)
int
on_tr_entry228_changed (GtkWidget *tr_entry228, SPSData *data) {

  ip4_hdrlen_entry (tr_entry228, 11, data);

  return (EXIT_SUCCESS);
}

// UDP (IPv4 traceroute) IP version (4 bits) (traceroute.ui)
int
on_tr_entry229_changed (GtkWidget *tr_entry229, SPSData *data) {

  ip4_ver_entry (tr_entry229, 11, data);

  return (EXIT_SUCCESS);
}

// UDP (IPv4 traceroute) Type of service (8 bits) (traceroute.ui)
int
on_tr_entry230_changed (GtkWidget *tr_entry230, SPSData *data) {

  ip4_tos_entry (tr_entry230, 11, data);

  return (EXIT_SUCCESS);
}

// UDP (IPv4 traceroute) Total length of datagram (16 bits) (traceroute.ui)
int
on_tr_entry231_changed (GtkWidget *tr_entry231, SPSData *data) {

  ip4_tld_entry (tr_entry231, 11, data);

  return (EXIT_SUCCESS);
}

// UDP (IPv4 traceroute) ID sequence number (16 bits) (traceroute.ui)
int
on_tr_entry232_changed (GtkWidget *tr_entry232, SPSData *data) {

  ip4_seq_entry (tr_entry232, 11, data);

  return (EXIT_SUCCESS);
}

// UDP (IPv4 traceroute) Zero (1 bit) (traceroute.ui)
int
on_tr_entry233_changed (GtkWidget *tr_entry233, SPSData *data) {

  ip4_zero_entry (tr_entry233, data->tr_entry234, data->tr_entry235, data->tr_entry236, 11, data);

  return (EXIT_SUCCESS);
}

// UDP (IPv4 traceroute) Do not fragment flag (1 bit) (traceroute.ui)
int
on_tr_entry234_changed (GtkWidget *tr_entry234, SPSData *data) {

  ip4_dnf_entry (data->tr_entry233, tr_entry234, data->tr_entry235, data->tr_entry236, 11, data);

  return (EXIT_SUCCESS);
}

// UDP (IPv4 traceroute) More fragments following flag (1 bit) (traceroute.ui)
int
on_tr_entry235_changed (GtkWidget *tr_entry235, SPSData *data) {

  ip4_mff_entry (data->tr_entry233, data->tr_entry234, tr_entry235, data->tr_entry236, 11, data);

  return (EXIT_SUCCESS);
}

// UDP (IPv4 traceroute) Fragmentation offset (13 bits) (traceroute.ui)
int
on_tr_entry236_changed (GtkWidget *tr_entry236, SPSData *data) {

  ip4_fo_entry (data->tr_entry233, data->tr_entry234, data->tr_entry235, tr_entry236, 11, data);

  return (EXIT_SUCCESS);
}

// UDP (IPv4 traceroute) Time-to-Live (8 bits) (traceroute.ui)
int
on_tr_entry237_changed (GtkWidget *tr_entry237, SPSData *data) {

  ip4_ttl_entry (tr_entry237, 11, data);

  return (EXIT_SUCCESS);
}

// UDP (IPv4 traceroute) Transport layer protocol (8 bits) (traceroute.ui)
int
on_tr_entry238_changed (GtkWidget *tr_entry238, SPSData *data) {

  ip4_tlp_entry (tr_entry238, 11, data);

  return (EXIT_SUCCESS);
}

// UDP (IPv4 traceroute) Source IPv4 address (32 bits) (traceroute.ui)
int
on_tr_entry239_changed (GtkWidget *tr_entry239, SPSData *data) {

  ip4_sip_entry (tr_entry239, 11, data);

  return (EXIT_SUCCESS);
}

// UDP (IPv4 traceroute) Destination IP address (32 bits) (traceroute.ui)
int
on_tr_entry240_changed (GtkWidget *tr_entry240, SPSData *data) {

  ip4_dip_entry (tr_entry240, 11, data);

  return (EXIT_SUCCESS);
}

// UDP (IPv4 traceroute) IPv4 checksum (16 bits) (traceroute.ui)
int
on_tr_entry241_changed (GtkWidget *tr_entry241, SPSData *data) {

  ip4_chksum_entry (tr_entry241, 11, data);

  return (EXIT_SUCCESS);
}

// IP options for IPv4 UDP for Traceroute (traceroute.ui)

// UDP (IPv4 traceroute) Decimal option data entry (traceroute.ui)
int
on_tr_radiobutton57_toggled (GtkButton *tr_radiobutton57, SPSData *data) {

  (void) tr_radiobutton57;  // This statement suppresses compiler warning about unused parameter.

  if (data->dec_hex_ipopt_udp4_tr) {
    data->dec_hex_ipopt_udp4_tr = 0;
  } else {
    data->dec_hex_ipopt_udp4_tr = 1;
  }

  return (EXIT_SUCCESS);
}

// UDP (IPv4 traceroute) Hexadecimal option data entry (traceroute.ui)
int
on_tr_radiobutton58_toggled (GtkButton *tr_radiobutton58) {

  (void) tr_radiobutton58;  // This statement suppresses compiler warning about unused parameter.

  // on_radiobutton57_toggled() will change value of data->dec_hex_ipopt_udp4_tr.

  return (EXIT_SUCCESS);
}

// UDP (IPv4 traceroute) IP option data (traceroute.ui)
int
on_tr_entry428_changed (GtkWidget *tr_entry428, SPSData *data) {

  data_entry (tr_entry428, &data->ip_optlenbuf[11], MAX_IP4OPTLEN, data->ip_optionsbuf[11], data->dec_hex_ipopt_udp4_tr, data->traceroute_window, data);

  return (EXIT_SUCCESS);
}

// UDP (IPv4 traceroute) Insert IP option after another, as specified by user. (traceroute.ui)
int
on_tr_button90_clicked (GtkButton *tr_button90, SPSData *data) {

  (void) tr_button90;  // This statement suppresses compiler warning about unused parameter.

  ip4_option (0, 11, data->ipopt_udp4_tr_after, data->tr_textview42, data);

  return (EXIT_SUCCESS);
}

// UDP (IPv4 traceroute) Number of option after which to insert new option. (traceroute.ui)
int
on_tr_entry429_changed (GtkWidget *tr_entry429, SPSData *data) {

  ip4_opt_insert_after_entry (tr_entry429, 11, &data->ipopt_udp4_tr_after, data);

  return (EXIT_SUCCESS);
}

// UDP (IPv4 traceroute) Append IP option for IPv4 UDP packet. (traceroute.ui)
int
on_tr_button91_clicked (GtkButton *tr_button91, SPSData *data) {

  (void) tr_button91;  // This statement suppresses compiler warning about unused parameter.

  ip4_option (2, 11, 0, data->tr_textview42, data);

  return (EXIT_SUCCESS);
}

// UDP (IPv4 traceroute) View IPv4 header options for UDP for traceroute. (traceroute.ui)
int
on_tr_button92_clicked (GtkButton *tr_button92, SPSData *data) {

  (void) tr_button92;  // This statement suppresses compiler warning about unused parameter.

  opt_view (11, 0, data);

  return (EXIT_SUCCESS);
}

// UDP (IPv4 traceroute) Number of IPv4 option to remove. (traceroute.ui)
int
on_tr_entry430_changed (GtkWidget *tr_entry430, SPSData *data) {

  ip4_opt_remove_entry (tr_entry430, 11, &data->ipopt_udp4_tr_remove, data);

  return (EXIT_SUCCESS);
}

// UDP (IPv4 traceroute) Remove selected IP option. (traceroute.ui)
int
on_tr_button93_clicked (GtkButton *tr_button93, SPSData *data) {

  (void) tr_button93;  // This statement suppresses compiler warning about unused parameter.

  ip4_option (1, 11, data->ipopt_udp4_tr_remove, data->tr_textview42, data);

  return (EXIT_SUCCESS);
}

// UDP header (IPv4) for traceroute (traceroute.ui)

// UDP (IPv4 traceroute) source port (16 bits) (traceroute.ui)
int
on_tr_entry242_changed (GtkWidget *tr_entry242, SPSData *data) {

  udp_sport_entry (tr_entry242, 11, data);

  if (data->ran_udp4_tr_sourceport) {
    //  Source port has been entered manually. Uncheck "Randomize" checkbutton.
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->tr_checkbutton35), FALSE);
    data->ran_udp4_tr_sourceport = 0;
  }

  return (EXIT_SUCCESS);
}

// UDP (IPv4 traceroute) Randomize source port number. (traceroute.ui)
int
on_tr_checkbutton35_toggled (GtkButton *tr_checkbutton35, SPSData *data) {

  (void) tr_checkbutton35;  // This statement suppresses compiler warning about unused parameter.

  if (data->ran_udp4_tr_sourceport) {
    data->ran_udp4_tr_sourceport = 0;
  } else {
    udp_ran_sport (&data->ran_udp4_tr_sourceport, 11, data->tr_entry242, data);
  }

  return (EXIT_SUCCESS);
}

// UDP (IPv4 traceroute) destination port (16 bits) (traceroute.ui)
int
on_tr_entry243_changed (GtkWidget *tr_entry243, SPSData *data) {

  udp_dport_entry (tr_entry243, 11, data);

  return (EXIT_SUCCESS);
}

// UDP (IPv4 traceroute) Length of datagram (16 bits) (traceroute.ui)
int
on_tr_entry244_changed (GtkWidget *tr_entry244, SPSData *data) {

  udp_len_entry (tr_entry244, 11, data);

  return (EXIT_SUCCESS);
}

// UDP (IPv4 traceroute) UDP checksum (16 bits) (traceroute.ui)
int
on_tr_entry245_changed (GtkWidget *tr_entry245, SPSData *data) {

  udp_chksum_entry (tr_entry245, 11, data);

  return (EXIT_SUCCESS);
}

// UDP (IPv4 traceroute) ASCII data entry (traceroute.ui)
int
on_tr_radiobutton33_toggled (GtkButton *tr_radiobutton33, SPSData *data) {

  (void) tr_radiobutton33;  // This statement suppresses compiler warning about unused parameter.

  if (data->ascii_hex_udp4_tr) {
    data->ascii_hex_udp4_tr = 0;
  } else {
    data->ascii_hex_udp4_tr = 1;
  }

  return (EXIT_SUCCESS);
}

// UDP (IPv4 traceroute) Hexadecimal data entry (traceroute.ui)
int
on_tr_radiobutton34_toggled (GtkButton *tr_radiobutton34) {

  (void) tr_radiobutton34;  // This statement suppresses compiler warning about unused parameter.

  // on_radiobutton33_toggled() will change value of data->ascii_hex_udp4_tr.

  return (EXIT_SUCCESS);
}

// UDP (IPv4 traceroute) UDP data from keyboard (traceroute.ui)
int
on_tr_entry246_changed (GtkWidget *tr_entry246, SPSData *data) {

  key_data (tr_entry246, 11, data->ascii_hex_udp4_tr, data);

  if (data->nframes[11] > 1) {
    sprintf (data->warning_text, "Packet will be fragmented to accommodate UDP data.");
    data->parent = data->traceroute_window;
    report_warning (data);
  }

  return (EXIT_SUCCESS);
}

// UDP (IPv4 traceroute) UDP data from file (traceroute.ui)
int
on_tr_button35_clicked (GtkButton *tr_button35, SPSData *data) {

  (void) tr_button35;  // This statement suppresses compiler warning about unused parameter.

  file_payload_data (data->tr_entry246, 11, data);

  if (data->nframes[11] > 1) {
    sprintf (data->warning_text, "Packet will be fragmented to accommodate UDP data.");
    data->parent = data->traceroute_window;
    report_warning (data);
  }

  return (EXIT_SUCCESS);
}

// Unload file data for traceroute UDP (IPv4) (traceroute.ui)
int
on_tr_button208_clicked (GtkButton *tr_button208, SPSData *data) {

  (void) tr_button208;  // This statement suppresses compiler warning about unused parameter.
  GtkTextBuffer *tr_textbuffer68;

  // Remove file data.
  data->payloadlen[11] = 0;
  memset (data->payload[11], 0, IP_MAXPACKET * sizeof (uint8_t));

  // Update ethernet frame.
  create_ip4_frame (11, data);

  tr_textbuffer68 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->tr_textview68));
  gtk_text_buffer_set_text (tr_textbuffer68, "", -1);

  return (EXIT_SUCCESS);
}

// Call udp4_tr_default() (traceroute.ui)
int
on_tr_button36_clicked (GtkButton *tr_button36, SPSData *data) {

  (void) tr_button36;  // This statement suppresses compiler warning about unused parameter.

  udp4_tr_default (data);
  udp4_tr_show (data);

  return (EXIT_SUCCESS);
}

// TCP page (IPv6) for traceroute (traceroute.ui)

// TCP (IPv6 traceroute) - Copy all data from main TCP (IPv6) page
int
on_tr_button503_clicked (GtkButton *tr_button503, SPSData *data) {

  (void) tr_button503;  // This statement suppresses compiler warning about unused parameter.

  // Copy IPv6 TCP entry data from main page to traceroute page.
  copy_tcp6_to_traceroute (data);

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) Destination link-layer (MAC) address (48 bits) (traceroute.ui)
int
on_tr_entry247_changed (GtkWidget *tr_entry247, SPSData *data) {

  mac_entry (tr_entry247, data->ethhdr[12].dst_mac, 12, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) Source link-layer (MAC) address (48 bits) (traceroute.ui)
int
on_tr_entry248_changed (GtkWidget *tr_entry248, SPSData *data) {

  mac_entry (tr_entry248, data->ethhdr[12].src_mac, 12, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) This host's interface name to be converted to link-layer (MAC) address and find MTU. (traceroute.ui)
int
on_tr_entry249_changed (GtkWidget *tr_entry249, SPSData *data) {

  // Check for valid source interface name, check MTU, and update source MAC address.
  ifname_entry (tr_entry249, data->tr_entry248, 12, data);

  // Update MTU spinbutton.
  update_mtu_spinbutton (12, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) Maximum transmission unit (MTU) - Changed via spinbutton + or -. (traceroute.ui)
int
on_tr_spinbutton14_value_changed (GtkWidget *tr_spinbutton14, SPSData *data) {

  int value, min, ifmtu;

  // Set parent window for reporting error messages.
  data->parent = data->traceroute_window;

  // Get requested MTU value from spinbutton.
  value = (int) gtk_spin_button_get_value (GTK_SPIN_BUTTON (tr_spinbutton14));

  // Obtain interface's MTU.
  ifmtu = interface_getmtu (data->ifname[12]);
  if (ifmtu == 0) {
    update_mtu_spinbutton (12, data);  // Return spinbutton to current MTU value.
    return (EXIT_FAILURE);
  }

  // Determine minimum size needed to accommodate packet
  // when packet is fragmented to maximum extent.
  min = min6_mtu (12, data);

  // Ensure min required size is less than or equal to interface's MTU.
  if (min > ifmtu) {
    sprintf (data->error_text, "The interface's MTU is too small to accommodate packet.");
    report_error (data);
    data->ifmtu[12] = ifmtu;
    update_mtu_spinbutton (12, data);
    create_ip6_frame (12, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size is greater or equal to min required size to accommodate packet.
  if (value < min) {
    data->ifmtu[12] = min;
    update_mtu_spinbutton (12, data);
    create_ip6_frame (12, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size if <= interface's MTU.
  if (value > ifmtu) {
    data->ifmtu[12] = ifmtu;
    update_mtu_spinbutton (12, data);
    create_ip6_frame (12, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Accept requested MTU.
  data->ifmtu[12] = value;
  create_ip6_frame (12, data);  // Update ethernet frames.

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) Maximum transmission unit (MTU) - Changed via spinbutton text entry. (traceroute.ui)
int
on_tr_spinbutton14_change_value (GtkWidget *tr_spinbutton14, SPSData *data) {

  int value, min, ifmtu;

  // Set parent window for reporting error messages.
  data->parent = data->traceroute_window;

  // Get requested MTU value from spinbutton text entry.
  value = (int) gtk_spin_button_get_value (GTK_SPIN_BUTTON (tr_spinbutton14));

  // Obtain interface's MTU.
  ifmtu = interface_getmtu (data->ifname[12]);
  if (ifmtu == 0) {
    update_mtu_spinbutton (12, data);  // Return spinbutton to current MTU value.
    return (EXIT_FAILURE);
  }

  // Determine minimum size needed to accommodate packet
  // when packet is fragmented to maximum extent.
  min = min6_mtu (12, data);

  // Ensure min required size is less than or equal to interface's MTU.
  if (min > ifmtu) {
    sprintf (data->error_text, "The interface's MTU is too small to accommodate packet.");
    report_error (data);
    data->ifmtu[12] = ifmtu;
    update_mtu_spinbutton (12, data);
    create_ip6_frame (12, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size is greater or equal to min required size to accommodate packet.
  if (value < min) {
    sprintf (data->error_text, "The requested MTU is too small to accommodate packet.\nValue set to minimum required to accommodate packet instead.");
    report_error (data);
    data->ifmtu[12] = min;
    update_mtu_spinbutton (12, data);
    create_ip6_frame (12, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size if <= interface's MTU.
  if (value > ifmtu) {
    sprintf (data->error_text, "The requested MTU is larger than the interface's MTU.\nValue set to the interface's MTU instead.");
    report_error (data);
    data->ifmtu[12] = ifmtu;
    update_mtu_spinbutton (12, data);
    create_ip6_frame (12, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Accept requested MTU.
  data->ifmtu[12] = value;
  create_ip6_frame (12, data);  // Update ethernet frames.

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) Ethernet type code (16 bits) (traceroute.ui)
// http://www.iana.org/assignments/ethernet-numbers
int
on_tr_entry250_changed (GtkWidget *tr_entry250, SPSData *data) {

  ethtype_entry (tr_entry250, 12, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) IP version (4 bits) (traceroute.ui)
int
on_tr_entry251_changed (GtkWidget *tr_entry251, SPSData *data) {

  ip6_ver_entry (tr_entry251, 12, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) Traffic class (8 bits) (traceroute.ui)
int
on_tr_entry252_changed (GtkWidget *tr_entry252, SPSData *data) {

  ip6_tc_entry (tr_entry252, 12, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) Flow label (20 bits) (traceroute.ui)
int
on_tr_entry253_changed (GtkWidget *tr_entry253, SPSData *data) {

  ip6_fl_entry (tr_entry253, 12, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) Payload length (16 bits) (traceroute.ui)
int
on_tr_entry254_changed (GtkWidget *tr_entry254, SPSData *data) {

  ip6_plen_entry (tr_entry254, 12, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) Next header (8 bits) (traceroute.ui)
int
on_tr_entry255_changed (GtkWidget *tr_entry255, SPSData *data) {

  ip6_nxt_entry (tr_entry255, 12, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) Hop limit (8 bits) (traceroute.ui)
int
on_tr_entry256_changed (GtkWidget *tr_entry256, SPSData *data) {

  ip6_hops_entry (tr_entry256, 12, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) Source IP address (128 bits) (traceroute.ui)
int
on_tr_entry257_changed (GtkWidget *tr_entry257, SPSData *data) {

  ip6_sip_entry (tr_entry257, 12, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) Destination IP address (128 bits) (traceroute.ui)
int
on_tr_entry258_changed (GtkWidget *tr_entry258, SPSData *data) {

  ip6_dip_entry (tr_entry258, 12, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) Hop-by-Hop extension header (traceroute.ui)
int
on_tr_button101_clicked (GtkWidget *tr_button101, SPSData *data) {

  (void) tr_button101;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 12;
  hop_win (data);

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) Destination extension header (traceroute.ui)
int
on_tr_button113_clicked (GtkWidget *tr_button113, SPSData *data) {

  (void) tr_button113;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 12;
  dst_win (data);

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) Routing extension header (traceroute.ui)
int
on_tr_button135_clicked (GtkWidget *tr_button135, SPSData *data) {

  (void) tr_button135;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 12;
  route_win (data);

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) Authentication extension header (traceroute.ui)
int
on_tr_button119_clicked (GtkWidget *tr_button119, SPSData *data) {

  (void) tr_button119;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 12;
  auth_win (data);

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) ESP extension header (traceroute.ui)
int
on_tr_button129_clicked (GtkWidget *tr_button129, SPSData *data) {

  (void) tr_button129;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 12;
  esp_win (data);

  return (EXIT_SUCCESS);
}

// TCP header (IPv6) for Traceroute (traceroute.ui)

// TCP (IPv6 traceroute) source port (16 bits) (traceroute.ui)
int
on_tr_entry259_changed (GtkWidget *tr_entry259, SPSData *data) {

  tcp_sport_entry (tr_entry259, 12, data);

  if (data->ran_tcp6_tr_sourceport) {
    // Source port has been entered manually. Uncheck "Randomize" checkbutton.
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->tr_checkbutton37), FALSE);
    data->ran_tcp6_tr_sourceport = 0;
  }

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) Randomize source port number. (traceroute.ui)
int
on_tr_checkbutton37_toggled (GtkButton *tr_checkbutton37, SPSData *data) {

  (void) tr_checkbutton37;  // This statement suppresses compiler warning about unused parameter.

  if (data->ran_tcp6_tr_sourceport) {
    data->ran_tcp6_tr_sourceport = 0;
  } else {
    tcp_ran_sport (&data->ran_tcp6_tr_sourceport, 12, data->tr_entry259, data);
  }

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) destination port (16 bits) (traceroute.ui)
int
on_tr_entry260_changed (GtkWidget *tr_entry260, SPSData *data) {

  tcp_dport_entry (tr_entry260, 12, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) sequence number (32 bits) (traceroute.ui)
int
on_tr_entry261_changed (GtkWidget *tr_entry261, SPSData *data) {

  tcp_seqnum_entry (tr_entry261, 12, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) acknowledgement number (32 bits) (traceroute.ui)
int
on_tr_entry262_changed (GtkWidget *tr_entry262, SPSData *data) {

  tcp_acknum_entry (tr_entry262, 12, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) reserved (4 bits) (traceroute.ui)
int
on_tr_entry263_changed (GtkWidget *tr_entry263, SPSData *data) {

  tcp_res_entry (tr_entry263, 12, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) data offset (4 bits) (traceroute.ui)
int
on_tr_entry264_changed (GtkWidget *tr_entry264, SPSData *data) {

  tcp_off_entry (tr_entry264, 12, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) FIN flag (1 bit) (traceroute.ui)
int
on_tr_entry265_changed (GtkWidget *tr_entry265, SPSData *data) {

  tcp_fin_entry (tr_entry265, 12, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) SYN flag (1 bit) (traceroute.ui)
int
on_tr_entry266_changed (GtkWidget *tr_entry266, SPSData *data) {

  tcp_syn_entry (tr_entry266, 12, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) RST flag (1 bit) (traceroute.ui)
int
on_tr_entry267_changed (GtkWidget *tr_entry267, SPSData *data) {

  tcp_rst_entry (tr_entry267, 12, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) PSH flag (1 bit) (traceroute.ui)
int
on_tr_entry268_changed (GtkWidget *tr_entry268, SPSData *data) {

  tcp_psh_entry (tr_entry268, 12, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) ACK flag (1 bit) (traceroute.ui)
int
on_tr_entry269_changed (GtkWidget *tr_entry269, SPSData *data) {

  tcp_ack_entry (tr_entry269, 12, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) URG flag (1 bit) (traceroute.ui)
int
on_tr_entry270_changed (GtkWidget *tr_entry270, SPSData *data) {

  tcp_urg_entry (tr_entry270, 12, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) ECE flag (1 bit) (traceroute.ui)
int
on_tr_entry271_changed (GtkWidget *tr_entry271, SPSData *data) {

  tcp_ece_entry (tr_entry271, 12, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) CWR flag (1 bit) (traceroute.ui)
int
on_tr_entry272_changed (GtkWidget *tr_entry272, SPSData *data) {

  tcp_cwr_entry (tr_entry272, 12, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) window size (16 bits) (traceroute.ui)
int
on_tr_entry273_changed (GtkWidget *tr_entry273, SPSData *data) {

  tcp_win_entry (tr_entry273, 12, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) checksum (16 bits) (traceroute.ui)
int
on_tr_entry274_changed (GtkWidget *tr_entry274, SPSData *data) {

  tcp_chksum_entry (tr_entry274, 12, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) urgent pointer (16 bits) (traceroute.ui)
int
on_tr_entry275_changed (GtkWidget *tr_entry275, SPSData *data) {

  tcp_urgptr_entry (tr_entry275, 12, data);

  return (EXIT_SUCCESS);
}

// TCP options (IPv6) for traceroute (traceroute.ui)

// TCP (IPv6 traceroute) Decimal option data entry (traceroute.ui)
int
on_tr_radiobutton53_toggled (GtkButton *tr_radiobutton53, SPSData *data) {

  (void) tr_radiobutton53;  // This statement suppresses compiler warning about unused parameter.

  if (data->dec_hex_tcpopt_tcp6_tr) {
    data->dec_hex_tcpopt_tcp6_tr = 0;
  } else {
    data->dec_hex_tcpopt_tcp6_tr = 1;
  }

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) Hexadecimal option data entry (traceroute.ui)
int
on_tr_radiobutton54_toggled (GtkButton *tr_radiobutton54) {

  (void) tr_radiobutton54;  // This statement suppresses compiler warning about unused parameter.

  // on_radiobutton53_toggled() will change value of data->dec_hex_tcpopt_tcp6_tr.

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) option data for traceroute (traceroute.ui)
int
on_tr_entry422_changed (GtkWidget *tr_entry422, SPSData *data) {

  data_entry (tr_entry422, &data->tcp_optlenbuf[12], MAX_TCPOPTLEN, data->tcp_optionsbuf[12], data->dec_hex_tcpopt_tcp6_tr, data->traceroute_window, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) Insert TCP option after another, as specified by user. (traceroute.ui)
int
on_tr_button82_clicked (GtkButton *tr_button82, SPSData *data) {

  (void) tr_button82;  // This statement suppresses compiler warning about unused parameter.

  tcp6_option (0, 12, data->tcpopt_tcp6_tr_after, data->tr_textview38, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) Number of option after which to insert new TCP option. (traceroute.ui)
int
on_tr_entry423_changed (GtkWidget *tr_entry423, SPSData *data) {

  tcp_opt_insert_after_entry (tr_entry423, 12, &data->tcpopt_tcp6_tr_after, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) Append TCP option for IPv6 TCP packet for traceroute. (traceroute.ui)
int
on_tr_button83_clicked (GtkButton *tr_button83, SPSData *data) {

  (void) tr_button83;  // This statement suppresses compiler warning about unused parameter.

  tcp6_option (2, 12, 0, data->tr_textview38, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) View TCP options for IPv6 TCP for traceroute. (traceroute.ui)
int
on_tr_button84_clicked (GtkButton *tr_button84, SPSData *data) {

  (void) tr_button84;  // This statement suppresses compiler warning about unused parameter.

  opt_view (12, 1, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) Number of TCP option to remove. (traceroute.ui)
int
on_tr_entry424_changed (GtkWidget *tr_entry424, SPSData *data) {

  tcp_opt_remove_entry (tr_entry424, 12, &data->tcpopt_tcp6_tr_remove, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) Remove selected TCP option. (traceroute.ui)
int
on_tr_button85_clicked (GtkButton *tr_button85, SPSData *data) {

  (void) tr_button85;  // This statement suppresses compiler warning about unused parameter.

  tcp6_option (1, 12, data->tcpopt_tcp6_tr_remove, data->tr_textview38, data);

  return (EXIT_SUCCESS);
}

// TCP data (IPv6) for traceroute (traceroute.ui)

// TCP (IPv6 traceroute) ASCII data entry (traceroute.ui)
int
on_tr_radiobutton35_toggled (GtkButton *tr_radiobutton35, SPSData *data) {

  (void) tr_radiobutton35;  // This statement suppresses compiler warning about unused parameter.

  if (data->ascii_hex_tcp6_tr) {
    data->ascii_hex_tcp6_tr = 0;
  } else {
    data->ascii_hex_tcp6_tr = 1;
  }

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) Hexadecimal data entry (traceroute.ui)
int
on_tr_radiobutton36_toggled (GtkButton *tr_radiobutton36) {

  (void) tr_radiobutton36;  // This statement suppresses compiler warning about unused parameter.

  // on_radiobutton35_toggled() will change value of data->ascii_hex_tcp6_tr.

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) TCP data from keyboard (traceroute.ui)
int
on_tr_entry276_changed (GtkWidget *tr_entry276, SPSData *data) {

  key_data (tr_entry276, 12, data->ascii_hex_tcp6_tr, data);

  if (data->nframes[12] > 1) {
    sprintf (data->warning_text, "Packet will be fragmented to accommodate TCP data.");
    data->parent = data->traceroute_window;
    report_warning (data);
  }

  // Update ethernet frame and extension header editing window widgets, if visible.
  ip6data_update (12, data);

  return (EXIT_SUCCESS);
}

// TCP (IPv6 traceroute) TCP data from file (traceroute.ui)
int
on_tr_button38_clicked (GtkButton *tr_button38, SPSData *data) {

  (void) tr_button38;  // This statement suppresses compiler warning about unused parameter.

  if (file_payload_data (data->tr_entry276, 12, data) == EXIT_SUCCESS) {

    if (data->nframes[12] > 1) {
      sprintf (data->warning_text, "Packet will be fragmented to accommodate TCP data.");
      data->parent = data->traceroute_window;
      report_warning (data);
    }

    // Update ethernet frame and extension header editing window widgets, if visible.
    ip6data_update (12, data);
  }

  return (EXIT_SUCCESS);
}

// Unload file data for traceroute TCP (IPv6) (traceroute.ui)
int
on_tr_button209_clicked (GtkButton *tr_button209, SPSData *data) {

  (void) tr_button209;  // This statement suppresses compiler warning about unused parameter.
  GtkTextBuffer *tr_textbuffer69;

  // Remove file data.
  data->payloadlen[12] = 0;
  memset (data->payload[12], 0, IP_MAXPACKET * sizeof (uint8_t));

  // Update ethernet frames.
  create_ip6_frame (12, data);
  create_6to4_frame (15, data);

  tr_textbuffer69 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->tr_textview69));
  gtk_text_buffer_set_text (tr_textbuffer69, "", -1);

  return (EXIT_SUCCESS);
}

// Call tcp6_tr_default() (traceroute.ui)
int
on_tr_button39_clicked (GtkButton *tr_button39, SPSData *data) {

  (void) tr_button39;  // This statement suppresses compiler warning about unused parameter.

  tcp6_tr_default (data);
  tcp6_tr_show (data);

  return (EXIT_SUCCESS);
}

// Outer hop-by-Hop extension header
int
on_tr_button309_clicked (GtkWidget *tr_button309, SPSData *data) {

  (void) tr_button309;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 12;
  hop_win_outer (data);

  return (EXIT_SUCCESS);
}

// Outer destination extension header
int
on_tr_button310_clicked (GtkWidget *tr_button310, SPSData *data) {

  (void) tr_button310;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 12;
  dst_win_outer (data);

  return (EXIT_SUCCESS);
}

// Outer routing extension header
int
on_tr_button311_clicked (GtkWidget *tr_button311, SPSData *data) {

  (void) tr_button311;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 12;
  route_win_outer (data);

  return (EXIT_SUCCESS);
}

// ICMP page (IPv6) for traceroute (traceroute.ui)

// ICMP (IPv6 traceroute) - Copy all data from main ICMP (IPv6) page (traceroute.ui)
int
on_tr_button504_clicked (GtkButton *tr_button504, SPSData *data) {

  (void) tr_button504;  // This statement suppresses compiler warning about unused parameter.

  // Copy IPv6 ICMP entry data from main page to traceroute page.
  copy_icmp6_to_traceroute (data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv6 traceroute) Destination link-layer (MAC) address (48 bits) (traceroute.ui)
int
on_tr_entry277_changed (GtkWidget *tr_entry277, SPSData *data) {

  mac_entry (tr_entry277, data->ethhdr[13].dst_mac, 13, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv6 traceroute) Source link-layer (MAC) address (48 bits) (traceroute.ui)
int
on_tr_entry278_changed (GtkWidget *tr_entry278, SPSData *data) {

  mac_entry (tr_entry278, data->ethhdr[13].src_mac, 13, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv6 traceroute) This host's interface name to be converted to link-layer (MAC) address and find MTU. (traceroute.ui)
int
on_tr_entry279_changed (GtkWidget *tr_entry279, SPSData *data) {

  // Check for valid source interface name, check MTU, and update source MAC address.
  ifname_entry (tr_entry279, data->tr_entry278, 13, data);

  // Update MTU spinbutton.
  update_mtu_spinbutton (13, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv6 traceroute) Maximum transmission unit (MTU) - Changed via spinbutton + or -. (traceroute.ui)
int
on_tr_spinbutton15_value_changed (GtkWidget *tr_spinbutton15, SPSData *data) {

  int value, min, ifmtu;

  // Set parent window for reporting error messages.
  data->parent = data->traceroute_window;

  // Get requested MTU value from spinbutton.
  value = (int) gtk_spin_button_get_value (GTK_SPIN_BUTTON (tr_spinbutton15));

  // Obtain interface's MTU.
  ifmtu = interface_getmtu (data->ifname[13]);
  if (ifmtu == 0) {
    update_mtu_spinbutton (13, data);  // Return spinbutton to current MTU value.
    return (EXIT_FAILURE);
  }

  // Determine minimum size needed to accommodate packet
  // when packet is fragmented to maximum extent.
  min = min6_mtu (13, data);

  // Ensure min required size is less than or equal to interface's MTU.
  if (min > ifmtu) {
    sprintf (data->error_text, "The interface's MTU is too small to accommodate packet.");
    report_error (data);
    data->ifmtu[13] = ifmtu;
    update_mtu_spinbutton (13, data);
    create_ip6_frame (13, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size is greater or equal to min required size to accommodate packet.
  if (value < min) {
    data->ifmtu[13] = min;
    update_mtu_spinbutton (13, data);
    create_ip6_frame (13, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size if <= interface's MTU.
  if (value > ifmtu) {
    data->ifmtu[13] = ifmtu;
    update_mtu_spinbutton (13, data);
    create_ip6_frame (13, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Accept requested MTU.
  data->ifmtu[13] = value;
  create_ip6_frame (13, data);  // Update ethernet frames.

  return (EXIT_SUCCESS);
}

// ICMP (IPv6 traceroute) Maximum transmission unit (MTU) - Changed via spinbutton text entry. (traceroute.ui)
int
on_tr_spinbutton15_change_value (GtkWidget *tr_spinbutton15, SPSData *data) {

  int value, min, ifmtu;

  // Set parent window for reporting error messages.
  data->parent = data->traceroute_window;

  // Get requested MTU value from spinbutton text entry.
  value = (int) gtk_spin_button_get_value (GTK_SPIN_BUTTON (tr_spinbutton15));

  // Obtain interface's MTU.
  ifmtu = interface_getmtu (data->ifname[13]);
  if (ifmtu == 0) {
    update_mtu_spinbutton (13, data);  // Return spinbutton to current MTU value.
    return (EXIT_FAILURE);
  }

  // Determine minimum size needed to accommodate packet
  // when packet is fragmented to maximum extent.
  min = min6_mtu (13, data);

  // Ensure min required size is less than or equal to interface's MTU.
  if (min > ifmtu) {
    sprintf (data->error_text, "The interface's MTU is too small to accommodate packet.");
    report_error (data);
    data->ifmtu[13] = ifmtu;
    update_mtu_spinbutton (13, data);
    create_ip6_frame (13, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size is greater or equal to min required size to accommodate packet.
  if (value < min) {
    sprintf (data->error_text, "The requested MTU is too small to accommodate packet.\nValue set to minimum required to accommodate packet instead.");
    report_error (data);
    data->ifmtu[13] = min;
    update_mtu_spinbutton (13, data);
    create_ip6_frame (13, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size if <= interface's MTU.
  if (value > ifmtu) {
    sprintf (data->error_text, "The requested MTU is larger than the interface's MTU.\nValue set to the interface's MTU instead.");
    report_error (data);
    data->ifmtu[13] = ifmtu;
    update_mtu_spinbutton (13, data);
    create_ip6_frame (13, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Accept requested MTU.
  data->ifmtu[13] = value;
  create_ip6_frame (13, data);  // Update ethernet frames.

  return (EXIT_SUCCESS);
}

// ICMP (IPv6 traceroute) Ethernet type code (16 bits) (traceroute.ui)
// http://www.iana.org/assignments/ethernet-numbers
int
on_tr_entry280_changed (GtkWidget *tr_entry280, SPSData *data) {

  ethtype_entry (tr_entry280, 13, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv6 traceroute) IP version (4 bits) (traceroute.ui)
int
on_tr_entry281_changed (GtkWidget *tr_entry281, SPSData *data) {

  ip6_ver_entry (tr_entry281, 13, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv6 traceroute) Traffic class (8 bits) (traceroute.ui)
int
on_tr_entry282_changed (GtkWidget *tr_entry282, SPSData *data) {

  ip6_tc_entry (tr_entry282, 13, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv6 traceroute) Flow label (20 bits) (traceroute.ui)
int
on_tr_entry283_changed (GtkWidget *tr_entry283, SPSData *data) {

  ip6_fl_entry (tr_entry283, 13, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv6 traceroute) Payload length (16 bits) (traceroute.ui)
int
on_tr_entry284_changed (GtkWidget *tr_entry284, SPSData *data) {

  ip6_plen_entry (tr_entry284, 13, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv6 traceroute) Next header (8 bits) (traceroute.ui)
int
on_tr_entry285_changed (GtkWidget *tr_entry285, SPSData *data) {

  ip6_nxt_entry (tr_entry285, 13, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv6 traceroute) Hop limit (8 bits) (traceroute.ui)
int
on_tr_entry286_changed (GtkWidget *tr_entry286, SPSData *data) {

  ip6_hops_entry (tr_entry286, 13, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv6 traceroute) Source IP address (128 bits) (traceroute.ui)
int
on_tr_entry287_changed (GtkWidget *tr_entry287, SPSData *data) {

  ip6_sip_entry (tr_entry287, 13, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv6 traceroute) Destination IP address (128 bits) (traceroute.ui)
int
on_tr_entry288_changed (GtkWidget *tr_entry288, SPSData *data) {

  ip6_dip_entry (tr_entry288, 13, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv6 traceroute) Hop-by-Hop extension header (traceroute.ui)
int
on_tr_button102_clicked (GtkWidget *tr_button102, SPSData *data) {

  (void) tr_button102;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 13;
  hop_win (data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv6 traceroute) Destination extension header (traceroute.ui)
int
on_tr_button114_clicked (GtkWidget *tr_button114, SPSData *data) {

  (void) tr_button114;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 13;
  dst_win (data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv6 traceroute) Routing extension header (traceroute.ui)
int
on_tr_button136_clicked (GtkWidget *tr_button136, SPSData *data) {

  (void) tr_button136;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 13;
  route_win (data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv6 traceroute) Authentication extension header (traceroute.ui)
int
on_tr_button120_clicked (GtkWidget *tr_button120, SPSData *data) {

  (void) tr_button120;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 13;
  auth_win (data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv6 traceroute) ESP extension header (traceroute.ui)
int
on_tr_button130_clicked (GtkWidget *tr_button130, SPSData *data) {

  (void) tr_button130;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 13;
  esp_win (data);

  return (EXIT_SUCCESS);
}

// ICMP header (IPv6) for traceroute (traceroute.ui)

// ICMP (IPv6 traceroute) Message Type (8 bits) (traceroute.ui)
int
on_tr_entry289_changed (GtkWidget *tr_entry289, SPSData *data) {

  icmp_mt_entry (tr_entry289, 13, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv6 traceroute) Message Code (8 bits) (traceroute.ui)
int
on_tr_entry290_changed (GtkWidget *tr_entry290, SPSData *data) {

  icmp_mc_entry (tr_entry290, 13, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv6 traceroute) ICMP checksum (16 bits) (traceroute.ui)
int
on_tr_entry291_changed (GtkWidget *tr_entry291, SPSData *data) {

  icmp_chksum_entry (tr_entry291, 13, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv6 traceroute) ID (16 bits) (traceroute.ui)
int
on_tr_entry292_changed (GtkWidget *tr_entry292, SPSData *data) {

  icmp_id_entry (tr_entry292, 13, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv6 traceroute) Sequence number (16 bits) (traceroute.ui)
int
on_tr_entry293_changed (GtkWidget *tr_entry293, SPSData *data) {

  icmp_seq_entry (tr_entry293, 13, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv6 traceroute) ASCII data entry (traceroute.ui)
int
on_tr_radiobutton37_toggled (GtkButton *tr_radiobutton37, SPSData *data) {

  (void) tr_radiobutton37;  // This statement suppresses compiler warning about unused parameter.

  if (data->ascii_hex_icmp6_tr) {
    data->ascii_hex_icmp6_tr = 0;
  } else {
    data->ascii_hex_icmp6_tr = 1;
  }

  return (EXIT_SUCCESS);
}

// ICMP (IPv6 traceroute) Hexadecimal data entry (traceroute.ui)
int
on_tr_radiobutton38_toggled (GtkButton *tr_radiobutton38) {

  (void) tr_radiobutton38;  // This statement suppresses compiler warning about unused parameter.

  // on_radiobutton37_toggled() will change value of data->ascii_hex_icmp6_tr.

  return (EXIT_SUCCESS);
}

// ICMP (IPv6 traceroute) ICMP data from keyboard (traceroute.ui)
int
on_tr_entry294_changed (GtkWidget *tr_entry294, SPSData *data) {

  key_data (tr_entry294, 13, data->ascii_hex_icmp6_tr, data);

  if (data->nframes[13] > 1) {
    sprintf (data->warning_text, "Packet will be fragmented to accommodate ICMP data.");
    data->parent = data->traceroute_window;
    report_warning (data);
  }

  // Update ethernet frame and extension header editing window widgets, if visible.
  ip6data_update (13, data);

  return (EXIT_SUCCESS);
}

// ICMP (IPv6 traceroute) ICMP data from file (traceroute.ui)
int
on_tr_button41_clicked (GtkButton *tr_button41, SPSData *data) {

  (void) tr_button41;  // This statement suppresses compiler warning about unused parameter.

  if (file_payload_data (data->tr_entry294, 13, data) == EXIT_SUCCESS) {

    if (data->nframes[13] > 1) {
      sprintf (data->warning_text, "Packet will be fragmented to accommodate ICMP data.");
      data->parent = data->traceroute_window;
      report_warning (data);
    }

    // Update ethernet frame and extension header editing window widgets, if visible.
    ip6data_update (13, data);
  }

  return (EXIT_SUCCESS);
}

// Unload file data for traceroute ICMP (IPv6) (traceroute.ui)
int
on_tr_button210_clicked (GtkButton *tr_button210, SPSData *data) {

  (void) tr_button210;  // This statement suppresses compiler warning about unused parameter.
  GtkTextBuffer *tr_textbuffer70;

  // Remove file data.
  data->payloadlen[13] = 0;
  memset (data->payload[13], 0, IP_MAXPACKET * sizeof (uint8_t));

  // Update ethernet frames.
  create_ip6_frame (13, data);
  create_6to4_frame (16, data);

  tr_textbuffer70 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->tr_textview70));
  gtk_text_buffer_set_text (tr_textbuffer70, "", -1);

  return (EXIT_SUCCESS);
}

// Call icmp6_tr_default() (traceroute.ui)
int
on_tr_button42_clicked (GtkButton *tr_button42, SPSData *data) {

  (void) tr_button42;  // This statement suppresses compiler warning about unused parameter.

  icmp6_tr_default (data);
  icmp6_tr_show (data);

  return (EXIT_SUCCESS);
}

// Outer hop-by-Hop extension header
int
on_tr_button312_clicked (GtkWidget *tr_button312, SPSData *data) {

  (void) tr_button312;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 13;
  hop_win_outer (data);

  return (EXIT_SUCCESS);
}

// Outer destination extension header
int
on_tr_button313_clicked (GtkWidget *tr_button313, SPSData *data) {

  (void) tr_button313;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 13;
  dst_win_outer (data);

  return (EXIT_SUCCESS);
}

// Outer routing extension header
int
on_tr_button314_clicked (GtkWidget *tr_button314, SPSData *data) {

  (void) tr_button314;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 13;
  route_win_outer (data);

  return (EXIT_SUCCESS);
}

// UDP page (IPv6) for traceroute (traceroute.ui)

// UDP (IPv6 traceroute) - Copy all data from main UDP (IPv6) page (traceroute.ui)
int
on_tr_button505_clicked (GtkButton *tr_button505, SPSData *data) {

  (void) tr_button505;  // This statement suppresses compiler warning about unused parameter.

  // Copy IPv6 UDP entry data from main page to traceroute page.
  copy_udp6_to_traceroute (data);

  return (EXIT_SUCCESS);
}

// UDP (IPv6 traceroute) Destination link-layer (MAC) address (48 bits) (traceroute.ui)
int
on_tr_entry295_changed (GtkWidget *tr_entry295, SPSData *data) {

  mac_entry (tr_entry295, data->ethhdr[14].dst_mac, 14, data);

  return (EXIT_SUCCESS);
}

// UDP (IPv6 traceroute) Source link-layer (MAC) address (48 bits) (traceroute.ui)
int
on_tr_entry296_changed (GtkWidget *tr_entry296, SPSData *data) {

  mac_entry (tr_entry296, data->ethhdr[14].src_mac, 14, data);

  return (EXIT_SUCCESS);
}

// UDP (IPv6 traceroute) This host's interface name to be converted to link-layer (MAC) address and find MTU. (traceroute.ui)
int
on_tr_entry297_changed (GtkWidget *tr_entry297, SPSData *data) {

  // Check for valid source interface name, check MTU, and update source MAC address.
  ifname_entry (tr_entry297, data->tr_entry296, 14, data);

  // Update MTU spinbutton.
  update_mtu_spinbutton (14, data);

  return (EXIT_SUCCESS);
}

// UDP (IPv6 traceroute) Maximum transmission unit (MTU) - Changed via spinbutton + or -. (traceroute.ui)
int
on_tr_spinbutton16_value_changed (GtkWidget *tr_spinbutton16, SPSData *data) {

  int value, min, ifmtu;

  // Set parent window for reporting error messages.
  data->parent = data->traceroute_window;

  // Get requested MTU value from spinbutton.
  value = (int) gtk_spin_button_get_value (GTK_SPIN_BUTTON (tr_spinbutton16));

  // Obtain interface's MTU.
  ifmtu = interface_getmtu (data->ifname[14]);
  if (ifmtu == 0) {
    update_mtu_spinbutton (14, data);  // Return spinbutton to current MTU value.
    return (EXIT_FAILURE);
  }

  // Determine minimum size needed to accommodate packet
  // when packet is fragmented to maximum extent.
  min = min6_mtu (14, data);

  // Ensure min required size is less than or equal to interface's MTU.
  if (min > ifmtu) {
    sprintf (data->error_text, "The interface's MTU is too small to accommodate packet.");
    report_error (data);
    data->ifmtu[14] = ifmtu;
    update_mtu_spinbutton (14, data);
    create_ip6_frame (14, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size is greater or equal to min required size to accommodate packet.
  if (value < min) {
    data->ifmtu[14] = min;
    update_mtu_spinbutton (14, data);
    create_ip6_frame (14, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size if <= interface's MTU.
  if (value > ifmtu) {
    data->ifmtu[14] = ifmtu;
    update_mtu_spinbutton (14, data);
    create_ip6_frame (14, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Accept requested MTU.
  data->ifmtu[14] = value;
  create_ip6_frame (14, data);  // Update ethernet frames.

  return (EXIT_SUCCESS);
}

// UDP (IPv6 traceroute) Maximum transmission unit (MTU) - Changed via spinbutton text entry. (traceroute.ui)
int
on_tr_spinbutton16_change_value (GtkWidget *tr_spinbutton16, SPSData *data) {

  int value, min, ifmtu;

  // Set parent window for reporting error messages.
  data->parent = data->traceroute_window;

  // Get requested MTU value from spinbutton text entry.
  value = (int) gtk_spin_button_get_value (GTK_SPIN_BUTTON (tr_spinbutton16));

  // Obtain interface's MTU.
  ifmtu = interface_getmtu (data->ifname[14]);
  if (ifmtu == 0) {
    update_mtu_spinbutton (14, data);  // Return spinbutton to current MTU value.
    return (EXIT_FAILURE);
  }

  // Determine minimum size needed to accommodate packet
  // when packet is fragmented to maximum extent.
  min = min6_mtu (14, data);

  // Ensure min required size is less than or equal to interface's MTU.
  if (min > ifmtu) {
    sprintf (data->error_text, "The interface's MTU is too small to accommodate packet.");
    report_error (data);
    data->ifmtu[14] = ifmtu;
    update_mtu_spinbutton (14, data);
    create_ip6_frame (14, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size is greater or equal to min required size to accommodate packet.
  if (value < min) {
    sprintf (data->error_text, "The requested MTU is too small to accommodate packet.\nValue set to minimum required to accommodate packet instead.");
    report_error (data);
    data->ifmtu[14] = min;
    update_mtu_spinbutton (14, data);
    create_ip6_frame (14, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Ensure requested size if <= interface's MTU.
  if (value > ifmtu) {
    sprintf (data->error_text, "The requested MTU is larger than the interface's MTU.\nValue set to the interface's MTU instead.");
    report_error (data);
    data->ifmtu[14] = ifmtu;
    update_mtu_spinbutton (14, data);
    create_ip6_frame (14, data);  // Update ethernet frames.
    return (EXIT_FAILURE);
  }

  // Accept requested MTU.
  data->ifmtu[14] = value;
  create_ip6_frame (14, data);  // Update ethernet frames.

  return (EXIT_SUCCESS);
}

// UDP (IPv6 traceroute) Ethernet type code (16 bits) (traceroute.ui)
// http://www.iana.org/assignments/ethernet-numbers
int
on_tr_entry298_changed (GtkWidget *tr_entry298, SPSData *data) {

  ethtype_entry (tr_entry298, 14, data);

  return (EXIT_SUCCESS);
}

// UDP (IPv6 traceroute) IP version (4 bits) (traceroute.ui)
int
on_tr_entry299_changed (GtkWidget *tr_entry299, SPSData *data) {

  ip6_ver_entry (tr_entry299, 14, data);

  return (EXIT_SUCCESS);
}

// UDP (IPv6 traceroute) Traffic class (8 bits) (traceroute.ui)
int
on_tr_entry300_changed (GtkWidget *tr_entry300, SPSData *data) {

  ip6_tc_entry (tr_entry300, 14, data);

  return (EXIT_SUCCESS);
}

// UDP (IPv6 traceroute) Flow label (20 bits) (traceroute.ui)
int
on_tr_entry301_changed (GtkWidget *tr_entry301, SPSData *data) {

  ip6_fl_entry (tr_entry301, 14, data);

  return (EXIT_SUCCESS);
}

// UDP (IPv6 traceroute) Payload length (16 bits) (traceroute.ui)
int
on_tr_entry302_changed (GtkWidget *tr_entry302, SPSData *data) {

  ip6_plen_entry (tr_entry302, 14, data);

  return (EXIT_SUCCESS);
}

// UDP (IPv6 traceroute) Next header (8 bits) (traceroute.ui)
int
on_tr_entry303_changed (GtkWidget *tr_entry303, SPSData *data) {

  ip6_nxt_entry (tr_entry303, 14, data);

  return (EXIT_SUCCESS);
}

// UDP (IPv6 traceroute) Hop limit (8 bits) (traceroute.ui)
int
on_tr_entry304_changed (GtkWidget *tr_entry304, SPSData *data) {

  ip6_hops_entry (tr_entry304, 14, data);

  return (EXIT_SUCCESS);
}

// UDP (IPv6 traceroute) Source IP address (128 bits) (traceroute.ui)
int
on_tr_entry305_changed (GtkWidget *tr_entry305, SPSData *data) {

  ip6_sip_entry (tr_entry305, 14, data);

  return (EXIT_SUCCESS);
}

// UDP (IPv6 traceroute) Destination IP address (128 bits) (traceroute.ui)
int
on_tr_entry306_changed (GtkWidget *tr_entry306, SPSData *data) {

  ip6_dip_entry (tr_entry306, 14, data);

  return (EXIT_SUCCESS);
}

// UDP (IPv6 traceroute) Hop-by-Hop extension header (traceroute.ui)
int
on_tr_button103_clicked (GtkWidget *tr_button103, SPSData *data) {

  (void) tr_button103;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 14;
  hop_win (data);

  return (EXIT_SUCCESS);
}

// UDP (IPv6 traceroute) Destination extension header (traceroute.ui)
int
on_tr_button115_clicked (GtkWidget *tr_button115, SPSData *data) {

  (void) tr_button115;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 14;
  dst_win (data);

  return (EXIT_SUCCESS);
}

// UDP (IPv6 traceroute) Routing extension header (traceroute.ui)
int
on_tr_button137_clicked (GtkWidget *tr_button137, SPSData *data) {

  (void) tr_button137;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 14;
  route_win (data);

  return (EXIT_SUCCESS);
}

// UDP (IPv6 traceroute) Authentication extension header (traceroute.ui)
int
on_tr_button121_clicked (GtkWidget *tr_button121, SPSData *data) {

  (void) tr_button121;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 14;
  auth_win (data);

  return (EXIT_SUCCESS);
}

// UDP (IPv6 traceroute) ESP extension header (traceroute.ui)
int
on_tr_button131_clicked (GtkWidget *tr_button131, SPSData *data) {

  (void) tr_button131;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 14;
  esp_win (data);

  return (EXIT_SUCCESS);
}

// UDP header (IPv6) for traceroute (traceroute.ui)

// UDP (IPv6 traceroute) source port (16 bits) (traceroute.ui)
int
on_tr_entry307_changed (GtkWidget *tr_entry307, SPSData *data) {

  udp_sport_entry (tr_entry307, 14, data);

  if (data->ran_udp6_tr_sourceport) {
    //  Source port has been entered manually. Uncheck "Randomize" checkbutton.
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->tr_checkbutton40), FALSE);
    data->ran_udp6_tr_sourceport = 0;
  }

  return (EXIT_SUCCESS);
}

// UDP (IPv6 traceroute) Randomize UDP source port number. (traceroute.ui)
int
on_tr_checkbutton40_toggled (GtkButton *tr_checkbutton40, SPSData *data) {

  (void) tr_checkbutton40;  // This statement suppresses compiler warning about unused parameter.

  if (data->ran_udp6_tr_sourceport) {
    data->ran_udp6_tr_sourceport = 0;
  } else {
    udp_ran_sport (&data->ran_udp6_tr_sourceport, 14, data->tr_entry307, data);
  }

  return (EXIT_SUCCESS);
}

// UDP (IPv6 traceroute) destination port (16 bits) (traceroute.ui)
int
on_tr_entry308_changed (GtkWidget *tr_entry308, SPSData *data) {

  udp_dport_entry (tr_entry308, 14, data);

  return (EXIT_SUCCESS);
}

// UDP (IPv6 traceroute) Length of datagram (16 bits) (traceroute.ui)
int
on_tr_entry309_changed (GtkWidget *tr_entry309, SPSData *data) {

  udp_len_entry (tr_entry309, 14, data);

  return (EXIT_SUCCESS);
}

// UDP (IPv6 traceroute) checksum (16 bits) (traceroute.ui)
int
on_tr_entry310_changed (GtkWidget *tr_entry310, SPSData *data) {

  udp_chksum_entry (tr_entry310, 14, data);

  return (EXIT_SUCCESS);
}

// UDP (IPv6 traceroute) ASCII data entry (traceroute.ui)
int
on_tr_radiobutton39_toggled (GtkButton *tr_radiobutton39, SPSData *data) {

  (void) tr_radiobutton39;  // This statement suppresses compiler warning about unused parameter.

  if (data->ascii_hex_udp6_tr) {
    data->ascii_hex_udp6_tr = 0;
  } else {
    data->ascii_hex_udp6_tr = 1;
  }

  return (EXIT_SUCCESS);
}

// UDP (IPv6 traceroute) Hexadecimal data entry (traceroute.ui)
int
on_tr_radiobutton40_toggled (GtkButton *tr_radiobutton40) {

  (void) tr_radiobutton40;  // This statement suppresses compiler warning about unused parameter.

  // on_radiobutton39_toggled() will change value of data->ascii_hex_udp6_tr.

  return (EXIT_SUCCESS);
}

// UDP (IPv6 traceroute) data from keyboard (traceroute.ui)
int
on_tr_entry311_changed (GtkWidget *tr_entry311, SPSData *data) {

  key_data (tr_entry311, 14, data->ascii_hex_udp6_tr, data);

  if (data->nframes[14] > 1) {
    sprintf (data->warning_text, "Packet will be fragmented to accommodate UDP data.");
    data->parent = data->traceroute_window;
    report_warning (data);
  }

  // Update ethernet frame and extension header editing window widgets, if visible.
  ip6data_update (14, data);

  return (EXIT_SUCCESS);
}

// UDP (IPv6 traceroute) data from file (traceroute.ui)
int
on_tr_button44_clicked (GtkButton *tr_button44, SPSData *data) {

  (void) tr_button44;  // This statement suppresses compiler warning about unused parameter.

  if (file_payload_data (data->tr_entry311, 14, data) == EXIT_SUCCESS) {

    if (data->nframes[14] > 1) {
      sprintf (data->warning_text, "Packet will be fragmented to accommodate UDP data.");
      data->parent = data->traceroute_window;
      report_warning (data);
    }

    // Update ethernet frame and extension header editing window widgets, if visible.
    ip6data_update (14, data);
  }

  return (EXIT_SUCCESS);
}

// Unload file data for traceroute UDP (IPv6) (traceroute.ui)
int
on_tr_button211_clicked (GtkButton *tr_button211, SPSData *data) {

  (void) tr_button211;  // This statement suppresses compiler warning about unused parameter.
  GtkTextBuffer *tr_textbuffer71;

  // Remove file data.
  data->payloadlen[14] = 0;
  memset (data->payload[14], 0, IP_MAXPACKET * sizeof (uint8_t));

  // Update ethernet frames.
  create_ip6_frame (14, data);
  create_6to4_frame (17, data);

  tr_textbuffer71 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->tr_textview71));
  gtk_text_buffer_set_text (tr_textbuffer71, "", -1);

  return (EXIT_SUCCESS);
}

// Call udp6_tr_default() (traceroute.ui)
int
on_tr_button45_clicked (GtkButton *tr_button45, SPSData *data) {

  (void) tr_button45;  // This statement suppresses compiler warning about unused parameter.

  udp6_tr_default (data);
  udp6_tr_show (data);

  return (EXIT_SUCCESS);
}

// Outer hop-by-Hop extension header
int
on_tr_button315_clicked (GtkWidget *tr_button315, SPSData *data) {

  (void) tr_button315;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 14;
  hop_win_outer (data);

  return (EXIT_SUCCESS);
}

// Outer destination extension header
int
on_tr_button316_clicked (GtkWidget *tr_button316, SPSData *data) {

  (void) tr_button316;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 14;
  dst_win_outer (data);

  return (EXIT_SUCCESS);
}

// Outer routing extension header
int
on_tr_button317_clicked (GtkWidget *tr_button317, SPSData *data) {

  (void) tr_button317;  // This statement suppresses compiler warning about unused parameter.

  data->exthdr_type = 14;
  route_win_outer (data);

  return (EXIT_SUCCESS);
}

// Traceroute page (sps.ui)

// TCP (IPv4) packet type for traceroute (sps.ui)
int
on_radiobutton21_toggled (GtkButton *radiobutton21, SPSData *data) {

  (void) radiobutton21;  // This statement suppresses compiler warning about unused parameter.

  data->packet_type_tr = 9;

  return (EXIT_SUCCESS);
}

// ICMP (IPv4) packet type for traceroute (sps.ui)
int
on_radiobutton22_toggled (GtkButton *radiobutton22, SPSData *data) {

  (void) radiobutton22;  // This statement suppresses compiler warning about unused parameter.

  data->packet_type_tr = 10;

  return (EXIT_SUCCESS);
}

// UDP (IPv4) packet type for traceroute (sps.ui)
int
on_radiobutton23_toggled (GtkButton *radiobutton23, SPSData *data) {

  (void) radiobutton23;  // This statement suppresses compiler warning about unused parameter.

  data->packet_type_tr = 11;

  return (EXIT_SUCCESS);
}

// TCP (IPv6) packet type for traceroute (sps.ui)
int
on_radiobutton26_toggled (GtkButton *radiobutton26, SPSData *data) {

  (void) radiobutton26;  // This statement suppresses compiler warning about unused parameter.

  data->packet_type_tr = 12;

  return (EXIT_SUCCESS);
}

// ICMP (IPv6) packet type for traceroute (sps.ui)
int
on_radiobutton27_toggled (GtkButton *radiobutton27, SPSData *data) {

  (void) radiobutton27;  // This statement suppresses compiler warning about unused parameter.

  data->packet_type_tr = 13;

  return (EXIT_SUCCESS);
}

// UDP (IPv6) packet type for traceroute (sps.ui)
int
on_radiobutton28_toggled (GtkButton *radiobutton28, SPSData *data) {

  (void) radiobutton28;  // This statement suppresses compiler warning about unused parameter.

  data->packet_type_tr = 14;

  return (EXIT_SUCCESS);
}

// Tunnel IPv6 over IPv4 (sps.ui)
int
on_checkbutton29_toggled (GtkButton *checkbutton29, SPSData *data) {

  (void) checkbutton29;  // This statement suppresses compiler warning about unused parameter.

  // Change value of flag for state of Tunnel IPv6 over IPv4 checkbutton
  if (data->sixto4_tr_flag) {
    data->sixto4_tr_flag = 0;
  } else {
    data->sixto4_tr_flag = 1;
  }

  return (EXIT_SUCCESS);
}

// Number of Probes per Hop for Traceroute (sps.ui)
int
on_spinbutton2_value_changed (GtkWidget *spinbutton2, SPSData *data) {

  data->num_probes = (int) gtk_spin_button_get_value (GTK_SPIN_BUTTON (spinbutton2));

  return (EXIT_SUCCESS);
}

// Timeout for Reply for Traceroute (sps.ui)
int
on_spinbutton4_value_changed (GtkWidget *spinbutton4, SPSData *data) {


  data->timeout_tr = (int) gtk_spin_button_get_value (GTK_SPIN_BUTTON (spinbutton4));

  return (EXIT_SUCCESS);
}

// Maximum Number of Hops for Traceroute (sps.ui)
int
on_spinbutton3_value_changed (GtkWidget *spinbutton3, SPSData *data) {


  data->maxhops = (int) gtk_spin_button_get_value (GTK_SPIN_BUTTON (spinbutton3));

  return (EXIT_SUCCESS);
}

// Resolve hostnames for nodes during Traceroute (sps.ui)
int
on_checkbutton28_toggled (GtkButton *checkbutton28, SPSData *data) {

  (void) checkbutton28;  // This statement suppresses compiler warning about unused parameter.

  if (data->resolve_tr) {
    data->resolve_tr = 0;
  } else {
    data->resolve_tr = 1;
  }

  return (EXIT_SUCCESS);
}

// Launch Traceroute Packet Editing UI (traceroute.ui) from sps.ui.
int
on_button23_clicked (GtkButton *button23, SPSData *data) {

  (void) button23;  // This statement suppresses compiler warning about unused parameter.

  GtkBuilder *builder;
  GError *error = NULL;

  // If a traceroute is already running, return with no action.
  if (tracing) {
    return (EXIT_FAILURE);
  }

  if (!data->traceroute_flag) {
    data->traceroute_flag = 1;
    // Create new GtkBuilder object
    builder = gtk_builder_new();
    if (!gtk_builder_add_from_file (builder, "traceroute.ui", &error)) {
      g_warning ("%s", error->message );
      g_free (error);
      exit (EXIT_SUCCESS);
    }

    // Get objects from UI file
    data->traceroute_window = GTK_WIDGET (gtk_builder_get_object (builder, "traceroute_window"));

    // Traceroute TCP (IPv4) page (traceroute.ui)

    data->tr_button500 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button500"));  // Copy all data from main TCP (IPv4) page

    // Ethernet header
    data->tr_entry164 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry164"));  // Destination link-layer (MAC) address
    data->tr_entry165 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry165"));  // Source link-layer (MAC) address
    data->tr_entry166 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry166"));  // Source interface name
    data->tr_spinbutton11 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_spinbutton11"));  // Interface maximum transmission unit
    data->tr_entry167 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry167"));  // Ethernet type code
    // IPv4 header
    data->tr_entry168   = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry168"));  // IP header length
    data->tr_entry169   = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry169"));  // Protocol version
    data->tr_entry170   = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry170"));  // Type of service
    data->tr_entry171   = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry171"));  // Total length of datagram
    data->tr_entry172   = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry172"));  // ID sequence number
    data->tr_entry173  = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry173"));  // Zero
    data->tr_entry174   = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry174"));  // Do not fragment flag
    data->tr_entry175   = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry175"));  // More fragments following flag
    data->tr_entry176   = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry176"));  // Fragmentation offset
    data->tr_entry177   = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry177"));  // Time-to-live
    data->tr_entry178   = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry178"));  // Transport layer protocol
    data->tr_entry179   = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry179"));  // Source IP address
    data->tr_entry180   = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry180"));  // Destination IP address
    data->tr_entry181   = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry181"));  // IP checksum
    // IPv4 header options
    data->tr_radiobutton49 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "tr_radiobutton49"));  // Decimal option data entry
    data->tr_radiobutton50 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "tr_radiobutton50"));  // Hexadecimal option data entry
    data->tr_entry416 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry416"));  // IPv4 options data
    data->tr_button74 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button74"));  // Insert IP Option After...
    data->tr_entry417 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry417"));  // Option # to insert after
    data->tr_button75 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button75"));  // Append IP option
    data->tr_textview34 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_textview34"));  // Number of IP options in packet
    data->tr_button76 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button76"));  // View IP options
    data->tr_entry418 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry418"));  // Option number to remove
    data->tr_button77 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button77"));  // Remove option
    // TCP header (IPv4)
    data->tr_entry182   = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry182"));  // Source port number
    data->tr_checkbutton30 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "tr_checkbutton30"));  // Source port "Randomize" checkbutton
    data->tr_entry183   = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry183"));  // Destination port number
    data->tr_entry184   = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry184"));  // Sequence number
    data->tr_entry185   = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry185"));  // Ack number
    data->tr_entry186   = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry186"));  // Reserved
    data->tr_entry187   = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry187"));  // Data offset
    data->tr_entry188   = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry188"));  // FIN flag
    data->tr_entry189   = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry189"));  // SYN flag
    data->tr_entry190   = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry190"));  // RST flag
    data->tr_entry191   = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry191"));  // PSH flag
    data->tr_entry192   = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry192"));  // ACK flag
    data->tr_entry193   = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry193"));  // URG flag
    data->tr_entry194   = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry194"));  // ECE flag
    data->tr_entry195   = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry195"));  // CWR flag
    data->tr_entry196   = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry196"));  // Window size
    data->tr_entry197   = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry197"));  // TCP Checksum
    data->tr_entry198   = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry198"));  // Urgent pointer
    data->tr_radiobutton29 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "tr_radiobutton29"));  // ASCII data entry
    data->tr_radiobutton30 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "tr_radiobutton30"));  // Hexadecimal data entry
    data->tr_entry199  = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry199"));  // TCP data from keyboard
    data->tr_button29  = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button29"));  // TCP data from file
    data->tr_textview66 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_textview66"));  // File Loaded indication
    data->tr_button206 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button206"));  // Unload file data from packet
    // TCP options (IPv4)
    data->tr_radiobutton51 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "tr_radiobutton51"));  // Decimal option data entry
    data->tr_radiobutton52 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "tr_radiobutton52"));  // Hexadecimal option data entry
    data->tr_entry419 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry419"));  // Option data
    data->tr_button78 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button78"));  // Insert TCP Option After...
    data->tr_entry420 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry420"));  // Option # to insert after
    data->tr_button79 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button79"));  // Append TCP option
    data->tr_textview36 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_textview36"));  // Number of TCP options in packet
    data->tr_button80 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button80"));  // View TCP options
    data->tr_entry421 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry421"));  // Option number to remove
    data->tr_button81 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button81"));  // Remove option

    data->tr_button28   = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button28"));  // View packet
    data->tr_button30  = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button30"));  // Restore default values

    // Traceroute ICMP (IPv4) page (traceroute.ui)

    data->tr_button501 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button501"));  // Copy all data from main ICMP (IPv4) page

    // Ethernet header
    data->tr_entry200 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry200"));  // Destination link-layer (MAC) address
    data->tr_entry201 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry201"));  // Source link-layer (MAC) address
    data->tr_entry202 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry202"));  // Source interface name
    data->tr_spinbutton12 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_spinbutton12"));  // Interface maximum transmission unit
    data->tr_entry203 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry203"));  // Ethernet type code
    // IPv4 header
    data->tr_entry204 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry204"));  // IP header length
    data->tr_entry205 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry205"));  // Protocol version
    data->tr_entry206 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry206"));  // Type of service
    data->tr_entry207 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry207"));  // Total length of datagram
    data->tr_entry208 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry208"));  // ID sequence number
    data->tr_entry209 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry209"));  // Zero
    data->tr_entry210 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry210"));  // Do not fragment flag
    data->tr_entry211 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry211"));  // More fragments following flag
    data->tr_entry212 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry212"));  // Fragmentation offset
    data->tr_entry213 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry213"));  // Time-to-live
    data->tr_entry214 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry214"));  // Transport layer protocol
    data->tr_entry215 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry215"));  // Source IP address
    data->tr_entry216 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry216"));  // Destination IP address
    data->tr_entry217 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry217"));  // IP checksum
    // IPv4 header options
    data->tr_radiobutton55 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "tr_radiobutton55"));  // Decimal option data entry
    data->tr_radiobutton56 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "tr_radiobutton56"));  // Hexadecimal option data entry
    data->tr_entry425 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry425"));  // IPv4 options data
    data->tr_button86 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button86"));  // Insert IP Option After...
    data->tr_entry426 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry426"));  // Option # to insert after
    data->tr_button87 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button87"));  // Append IP option
    data->tr_textview40 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_textview40"));  // Number of IP options in packet
    data->tr_button88 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button88"));  // View IP options
    data->tr_entry427 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry427"));  // Option number to remove
    data->tr_button89 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button89"));  // Remove option
    // ICMP header (IPv4)
    data->tr_entry218 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry218"));  // Message type
    data->tr_entry219 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry219"));  // Message code
    data->tr_entry220 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry220"));  // ICMP checksum
    data->tr_entry221 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry221"));  // Identifier
    data->tr_entry222 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry222"));  // Sequence number
    data->tr_radiobutton31 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "tr_radiobutton31"));  // ASCII data entry
    data->tr_radiobutton32 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "tr_radiobutton32"));  // Hexadecimal data entry
    data->tr_entry223 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry223"));  // ICMP data from keyboard
    data->tr_button32 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button32"));  // ICMP data from file
    data->tr_textview67 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_textview67"));  // File Loaded indication
    data->tr_button207 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button207"));  // Unload file data from packet

    data->tr_button31 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button31"));  // View packet
    data->tr_button33 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button33"));  // Restore default values

    // Traceroute UDP (IPv4) page (traceroute.ui)

    data->tr_button502 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button502"));  // Copy all data from main UDP (IPv4) page

    // Ethernet header
    data->tr_entry224 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry224"));  // Destination link-layer (MAC) address
    data->tr_entry225 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry225"));  // Source link-layer (MAC) address
    data->tr_entry226 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry226"));  // Source interface name
    data->tr_spinbutton13 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_spinbutton13"));  // Interface maximum transmission unit
    data->tr_entry227 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry227"));  // Ethernet type code
    // IPv4 header
    data->tr_entry228 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry228"));  // IP header length
    data->tr_entry229 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry229"));  // Protocol version
    data->tr_entry230 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry230"));  // Type of service
    data->tr_entry231 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry231"));  // Total length of datagram
    data->tr_entry232 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry232"));  // ID sequence number
    data->tr_entry233 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry233"));  // Zero
    data->tr_entry234 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry234"));  // Do not fragment flag
    data->tr_entry235 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry235"));  // More fragments following flag
    data->tr_entry236 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry236"));  // Fragmentation offset
    data->tr_entry237 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry237"));  // Time-to-live
    data->tr_entry238 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry238"));  // Transport layer protocol
    data->tr_entry239 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry239"));  // Source IP address
    data->tr_entry240 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry240"));  // Destination IP address
    data->tr_entry241 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry241"));  // IP checksum
    // IPv4 header options
    data->tr_radiobutton57 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "tr_radiobutton57"));  // Decimal option data entry
    data->tr_radiobutton58 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "tr_radiobutton58"));  // Hexadecimal option data entry
    data->tr_entry428 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry428"));  // IPv4 options data
    data->tr_button90 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button90"));  // Insert IP Option After...
    data->tr_entry429 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry429"));  // Option # to insert after
    data->tr_button91 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button91"));  // Append IP option
    data->tr_textview42 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_textview42"));  // Number of IP options in packet
    data->tr_button92 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button92"));  // View IP options
    data->tr_entry430 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry430"));  // Option number to remove
    data->tr_button93 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button93"));  // Remove option
    // UDP header (IPv4)
    data->tr_entry242 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry242"));  // Source port
    data->tr_checkbutton35 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "tr_checkbutton35"));  // Source IP "Randomize" checkbutton
    data->tr_entry243 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry243"));  // Destination port
    data->tr_entry244 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry244"));  // Length of UDP datagram
    data->tr_entry245 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry245"));  // UDP checksum
    data->tr_radiobutton33 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "tr_radiobutton33"));  // ASCII data entry
    data->tr_radiobutton34 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "tr_radiobutton34"));  // Hexadecimal data entry
    data->tr_entry246 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry246"));  // UDP data from keyboard
    data->tr_button35 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button35"));  // UDP data from file
    data->tr_textview68 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_textview68"));  // File Loaded indication
    data->tr_button208 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button208"));  // Unload file data from packet

    data->tr_button34 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button34"));  // View packet
    data->tr_button36 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button36"));  // Restore default values

    // Traceroute TCP (IPv6) page (traceroute.ui)

    data->tr_button503 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button503"));  // Copy all data from main TCP (IPv6) page

    // Ethernet header
    data->tr_entry247 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry247"));  // Destination link-layer (MAC) address
    data->tr_entry248 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry248"));  // Source link-layer (MAC) address
    data->tr_entry249 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry249"));  // Source interface name
    data->tr_spinbutton14 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_spinbutton14"));  // Interface maximum transmission unit
    data->tr_entry250 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry250"));  // Ethernet type code
    // IPv6 header
    data->tr_entry251 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry251"));  // IP version
    data->tr_entry252 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry252"));  // Traffic class
    data->tr_entry253 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry253"));  // Flow label
    data->tr_entry254 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry254"));  // Payload length
    data->tr_entry255 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry255"));  // Next header
    data->tr_entry256 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry256"));  // Hop limit
    data->tr_entry257 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry257"));  // Source IP address
    data->tr_entry258 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry258"));  // Destination IP address
    // TCP header (IPv6)
    data->tr_entry259 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry259"));  // Source port number
    data->tr_checkbutton37 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "tr_checkbutton37"));  // Source port "Randomize" checkbutton
    data->tr_entry260 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry260"));  // Destination port number
    data->tr_entry261 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry261"));  // Sequence number
    data->tr_entry262 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry262"));  // Ack number
    data->tr_entry263 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry263"));  // Reserved
    data->tr_entry264 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry264"));  // Data offset
    data->tr_entry265 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry265"));  // FIN flag
    data->tr_entry266 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry266"));  // SYN flag
    data->tr_entry267 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry267"));  // RST flag
    data->tr_entry268 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry268"));  // PSH flag
    data->tr_entry269 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry269"));  // ACK flag
    data->tr_entry270 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry270"));  // URG flag
    data->tr_entry271 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry271"));  // ECE flag
    data->tr_entry272 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry272"));  // CWR flag
    data->tr_entry273 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry273"));  // Window size
    data->tr_entry274 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry274"));  // TCP Checksum
    data->tr_entry275 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry275"));  // Urgent pointer
    data->tr_radiobutton35 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "tr_radiobutton35"));  // ASCII data entry
    data->tr_radiobutton36 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "tr_radiobutton36"));  // Hexadecimal data entry
    data->tr_entry276 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry276"));  // TCP data from keyboard
    data->tr_button38 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button38"));  // TCP data from file
    data->tr_textview69 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_textview69"));  // File Loaded indication
    data->tr_button209 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button209"));  // Unload file data from packet

    // TCP options (IPv6)
    data->tr_radiobutton53 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "tr_radiobutton53"));  // Decimal option data entry
    data->tr_radiobutton54 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "tr_radiobutton54"));  // Hexadecimal option data entry
    data->tr_entry422 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry422"));  // Option data
    data->tr_button82 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button82"));  // Insert TCP Option After...
    data->tr_entry423 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry423"));  // Option # to insert after
    data->tr_button83 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button83"));  // Append TCP option
    data->tr_textview38 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_textview38"));  // Number of TCP options in packet
    data->tr_button84 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button84"));  // View TCP options
    data->tr_entry424 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry424"));  // Option number to remove
    data->tr_button85 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button85"));  // Remove option
    // Extension headers
    data->tr_button101  = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button101"));  // Hop-by-hop extension header
    data->tr_button113  = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button113"));  // Destination extension header
    data->tr_button135 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button135"));  // Routing extension header
    data->tr_button119  = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button119"));  // Authentication extension header
    data->tr_button129 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button129"));  // Encapsulating security payload (ESP) extension header

    data->tr_button37 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button37"));  // View packet
    data->tr_button39 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button39"));  // Restore default values

    // Traceroute ICMP (IPv6) page (traceroute.ui)

    data->tr_button504 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button504"));  // Copy all data from main ICMP (IPv6) page

    // Ethernet header
    data->tr_entry277 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry277"));  // Destination link-layer (MAC) address
    data->tr_entry278 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry278"));  // Source link-layer (MAC) address
    data->tr_entry279 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry279"));  // Source interface name
    data->tr_spinbutton15 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_spinbutton15"));  // Interface maximum transmission unit
    data->tr_entry280 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry280"));  // Ethernet type code
    // IPv6 header
    data->tr_entry281 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry281"));  // IP version
    data->tr_entry282 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry282"));  // Traffic class
    data->tr_entry283 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry283"));  // Flow label
    data->tr_entry284 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry284"));  // Payload length
    data->tr_entry285 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry285"));  // Next header
    data->tr_entry286 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry286"));  // Hop limit
    data->tr_entry287 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry287"));  // Source IP address
    data->tr_entry288 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry288"));  // Destination IP address
    // ICMP header (IPv6)
    data->tr_entry289 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry289"));  // Message type
    data->tr_entry290 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry290"));  // Message code
    data->tr_entry291 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry291"));  // ICMP checksum
    data->tr_entry292 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry292"));  // Identifier
    data->tr_entry293 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry293"));  // Sequence number
    data->tr_radiobutton37 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "tr_radiobutton37"));  // ASCII data entry
    data->tr_radiobutton38 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "tr_radiobutton38"));  // Hexadecimal data entry
    data->tr_entry294 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry294"));  // ICMP data from keyboard
    data->tr_button41 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button41"));  // ICMP data from file
    data->tr_textview70 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_textview70"));  // File Loaded indication
    data->tr_button210 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button210"));  // Unload file data from packet
    // Extension headers
    data->tr_button102  = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button102"));  // Hop-by-hop extension header
    data->tr_button114  = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button114"));  // Destination extension header
    data->tr_button136 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button136"));  // Routing extension header
    data->tr_button120  = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button120"));  // Authentication extension header
    data->tr_button130 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button130"));  // Encapsulating security payload (ESP) extension header

    data->tr_button40 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button40"));  // View packet
    data->tr_button42 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button42"));  // Restore default values

    // Traceroute UDP (IPv6) page (traceroute.ui)

    data->tr_button505 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button505"));  // Copy all data from main UDP (IPv6) page

    // Ethernet header
    data->tr_entry295 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry295"));  // Destination link-layer (MAC) address
    data->tr_entry296 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry296"));  // Source link-layer (MAC) address
    data->tr_entry297 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry297"));  // Source interface name
    data->tr_spinbutton16 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_spinbutton16"));  // Interface maximum transmission unit
    data->tr_entry298 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry298"));  // Ethernet type code
    // IPv6 header
    data->tr_entry299 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry299"));  // IP version
    data->tr_entry300 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry300"));  // Traffic class
    data->tr_entry301 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry301"));  // Flow label
    data->tr_entry302 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry302"));  // Payload length
    data->tr_entry303 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry303"));  // Next header
    data->tr_entry304 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry304"));  // Hop limit
    data->tr_entry305 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry305"));  // Source IP address
    data->tr_entry306 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry306"));  // Destination IP address
    // UDP header (IPv6)
    data->tr_entry307 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry307"));  // Source port
    data->tr_checkbutton40 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "tr_checkbutton40"));  // Source IP "Randomize" checkbutton
    data->tr_entry308 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry308"));  // Destination port
    data->tr_entry309 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry309"));  // Length of UDP datagram
    data->tr_entry310 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry310"));  // UDP checksum
    data->tr_radiobutton39 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "tr_radiobutton39"));  // ASCII data entry
    data->tr_radiobutton40 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "tr_radiobutton40"));  // Hexadecimal data entry
    data->tr_entry311 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_entry311"));  // UDP data from keyboard
    data->tr_button44 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button44"));  // UDP data from file
    data->tr_textview71 = GTK_WIDGET (gtk_builder_get_object (builder, "tr_textview71"));  // File Loaded indication
    data->tr_button211 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button211"));  // Unload file data from packet
    // Extension headers
    data->tr_button103  = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button103"));  // Hop-by-hop extension header
    data->tr_button115  = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button115"));  // Destination extension header
    data->tr_button137 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button137"));  // Routing extension header
    data->tr_button121  = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button121"));  // Authentication extension header
    data->tr_button131 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button131"));  // Encapsulating security payload (ESP) extension header

    data->tr_button43 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button43"));  // View packet
    data->tr_button45 = GTK_BUTTON (gtk_builder_get_object (builder, "tr_button45"));  // Restore default values

    // TCP (IPv4) for traceroute
    gtk_spin_button_set_range (GTK_SPIN_BUTTON (data->tr_spinbutton11), 1, 65536.0);
    gtk_spin_button_set_digits (GTK_SPIN_BUTTON (data->tr_spinbutton11), 0);
    gtk_spin_button_set_increments (GTK_SPIN_BUTTON (data->tr_spinbutton11), 1.0, 0.0);
    gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (data->tr_spinbutton11), TRUE);
    gtk_spin_button_set_wrap (GTK_SPIN_BUTTON (data->tr_spinbutton11), FALSE);
    gtk_spin_button_set_snap_to_ticks (GTK_SPIN_BUTTON (data->tr_spinbutton11), TRUE);
    // ICMP (IPv4) for traceroute
    gtk_spin_button_set_range (GTK_SPIN_BUTTON (data->tr_spinbutton12), 1, 65536.0);
    gtk_spin_button_set_digits (GTK_SPIN_BUTTON (data->tr_spinbutton12), 0);
    gtk_spin_button_set_increments (GTK_SPIN_BUTTON (data->tr_spinbutton12), 1.0, 0.0);
    gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (data->tr_spinbutton12), TRUE);
    gtk_spin_button_set_wrap (GTK_SPIN_BUTTON (data->tr_spinbutton12), FALSE);
    gtk_spin_button_set_snap_to_ticks (GTK_SPIN_BUTTON (data->tr_spinbutton12), TRUE);
    // UDP (IPv4) for traceroute
    gtk_spin_button_set_range (GTK_SPIN_BUTTON (data->tr_spinbutton13), 1, 65536.0);
    gtk_spin_button_set_digits (GTK_SPIN_BUTTON (data->tr_spinbutton13), 0);
    gtk_spin_button_set_increments (GTK_SPIN_BUTTON (data->tr_spinbutton13), 1.0, 0.0);
    gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (data->tr_spinbutton13), TRUE);
    gtk_spin_button_set_wrap (GTK_SPIN_BUTTON (data->tr_spinbutton13), FALSE);
    gtk_spin_button_set_snap_to_ticks (GTK_SPIN_BUTTON (data->tr_spinbutton13), TRUE);
    // TCP (IPv6) for traceroute
    gtk_spin_button_set_range (GTK_SPIN_BUTTON (data->tr_spinbutton14), 1, 65536.0);
    gtk_spin_button_set_digits (GTK_SPIN_BUTTON (data->tr_spinbutton14), 0);
    gtk_spin_button_set_increments (GTK_SPIN_BUTTON (data->tr_spinbutton14), 1.0, 0.0);
    gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (data->tr_spinbutton14), TRUE);
    gtk_spin_button_set_wrap (GTK_SPIN_BUTTON (data->tr_spinbutton14), FALSE);
    gtk_spin_button_set_snap_to_ticks (GTK_SPIN_BUTTON (data->tr_spinbutton14), TRUE);
    // ICMP (IPv6) for traceroute
    gtk_spin_button_set_range (GTK_SPIN_BUTTON (data->tr_spinbutton15), 1, 65536.0);
    gtk_spin_button_set_digits (GTK_SPIN_BUTTON (data->tr_spinbutton15), 0);
    gtk_spin_button_set_increments (GTK_SPIN_BUTTON (data->tr_spinbutton15), 1.0, 0.0);
    gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (data->tr_spinbutton15), TRUE);
    gtk_spin_button_set_wrap (GTK_SPIN_BUTTON (data->tr_spinbutton15), FALSE);
    gtk_spin_button_set_snap_to_ticks (GTK_SPIN_BUTTON (data->tr_spinbutton15), TRUE);
    // UDP (IPv6) for traceroute
    gtk_spin_button_set_range (GTK_SPIN_BUTTON (data->tr_spinbutton16), 1, 65536.0);
    gtk_spin_button_set_digits (GTK_SPIN_BUTTON (data->tr_spinbutton16), 0);
    gtk_spin_button_set_increments (GTK_SPIN_BUTTON (data->tr_spinbutton16), 1.0, 0.0);
    gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (data->tr_spinbutton16), TRUE);
    gtk_spin_button_set_wrap (GTK_SPIN_BUTTON (data->tr_spinbutton16), FALSE);
    gtk_spin_button_set_snap_to_ticks (GTK_SPIN_BUTTON (data->tr_spinbutton16), TRUE);

    // Connect signals
    gtk_builder_connect_signals (builder, data);

    // Destroy builder, since we don't need it anymore
    g_object_unref (G_OBJECT (builder));

    // Show window. All other widgets are automatically shown by GtkBuilder
    gtk_widget_show (data->traceroute_window);

    // Populate entries in traceroute.ui window.
    tcp4_tr_show (data);
    tcp6_tr_show (data);
    icmp4_tr_show (data);
    icmp6_tr_show (data);
    udp4_tr_show (data);
    udp6_tr_show (data);
  }

  return (EXIT_SUCCESS);
}

// Set traceroute_flag to zero when traceroute_window is destroyed.
int
on_traceroute_window_destroy (GtkWidget *widget, SPSData *data) {

  (void) widget;  // This statement suppresses compiler warning about unused parameter.

  data->traceroute_flag = 0;

  return (EXIT_SUCCESS);
}

// Start traceroute.
int
on_button24_clicked (GtkButton *button24, SPSData *data) {

  (void) button24;  // This statement suppresses compiler warning about unused parameter.

  GThread *thread;

  if (!tracing) {

    if ((data->packet_type_tr == 9) || (data->packet_type_tr == 10) ||
        (data->packet_type_tr == 11)) {

      // Spawn thread to perform traceroute.
      thread = g_thread_new ("ipv4_tr_send", (GThreadFunc) ipv4_tr_send, data);
      if (! thread) {
        fprintf (stderr, "Error: Unable to create new thread for IPv4 traceroute.\n");
        exit (EXIT_FAILURE);
      }

    } else if ((data->packet_type_tr == 12) || (data->packet_type_tr == 13) ||
               (data->packet_type_tr == 14)) {
      if (data->sixto4_tr_flag) {

        // Spawn thread to perform traceroute.
        thread = g_thread_new ("sixto4_tr_send", (GThreadFunc) sixto4_tr_send, data);
        if (! thread) {
          fprintf (stderr, "Error: Unable to create new thread for IPv6 over IPv4 (6to4) traceroute.\n");
          exit (EXIT_FAILURE);
        }
      } else {

        // Spawn thread to perform traceroute.
        thread = g_thread_new ("ipv6_tr_send", (GThreadFunc) ipv6_tr_send, data);
        if (! thread) {
          fprintf (stderr, "Error: Unable to create new thread for IPv6 over IPv4 (6to4) traceroute.\n");
          exit (EXIT_FAILURE);
        }
      }
    }
  }  // End If !tracing

  return (EXIT_SUCCESS);
}

// Set traceroute mode (0 = terminate traceroute)
int
on_button25_clicked (GtkButton *button25, SPSData *data) {

  (void) button25;  // This statement suppresses compiler warning about unused parameter.
  (void) data;  // This statement suppresses compiler warning about unused parameter.

  tracing = 0;

  return (EXIT_SUCCESS);
}

// Toggle whether or not to show sent and received
// messages on stdout during file downloads.
int
on_checkbutton38_toggled (GtkButton *checkbutton38, SPSData *data) {

  (void) checkbutton38;  // This statement suppresses compiler warning about unused parameter.
  (void) data;  // This statement suppresses compiler warning about unused parameter.

  if (show_dl_messages) {
    show_dl_messages = 0;
  } else {
    show_dl_messages = 1;
  }

  return (EXIT_SUCCESS);
}

// Download ARIN delegation file.
int
on_button50_clicked (GtkButton *button50, SPSData *data) {

  (void) button50;  // This statement suppresses compiler warning about unused parameter.

  GThread *thread;

  if (!ftp_busy) {

    // Select ARIN.
    data->rir_index = 0;

    // Spawn thread to download file.
    thread = g_thread_new ("ftp_rir", (GThreadFunc) ftp_rir, data);
    if (! thread) {
      fprintf (stderr, "Error: Unable to create new thread for file download.\n");
      return (EXIT_FAILURE);
    }
  }

  return (EXIT_SUCCESS);
}

// Download RIPE delegation file.
int
on_button51_clicked (GtkButton *button51, SPSData *data) {

  (void) button51;  // This statement suppresses compiler warning about unused parameter.

  GThread *thread;

  if (!ftp_busy) {

    // Select RIPE.
    data->rir_index = 1;

    // Spawn thread to download file.
    thread = g_thread_new ("ftp_rir", (GThreadFunc) ftp_rir, data);
    if (! thread) {
      fprintf (stderr, "Error: Unable to create new thread for file download.\n");
      return (EXIT_FAILURE);
    }
  }

  return (EXIT_SUCCESS);
}

// Download AFRINIC delegation file.
int
on_button52_clicked (GtkButton *button52, SPSData *data) {

  (void) button52;  // This statement suppresses compiler warning about unused parameter.

  GThread *thread;

  if (!ftp_busy) {

    // Select AFRINIC.
    data->rir_index = 2;

    // Spawn thread to download file.
    thread = g_thread_new ("ftp_rir", (GThreadFunc) ftp_rir, data);
    if (! thread) {
      fprintf (stderr, "Error: Unable to create new thread for file download.\n");
      return (EXIT_FAILURE);
    }
  }

  return (EXIT_SUCCESS);
}

// Download APNIC delegation file.
int
on_button53_clicked (GtkButton *button53, SPSData *data) {

  (void) button53;  // This statement suppresses compiler warning about unused parameter.

  GThread *thread;

  if (!ftp_busy) {

    // Select APNIC.
    data->rir_index = 3;

    // Spawn thread to download file.
    thread = g_thread_new ("ftp_rir", (GThreadFunc) ftp_rir, data);
    if (! thread) {
      fprintf (stderr, "Error: Unable to create new thread for file download.\n");
      return (EXIT_FAILURE);
    }
  }

  return (EXIT_SUCCESS);
}

// Download LACNIC delegation file.
int
on_button54_clicked (GtkButton *button54, SPSData *data) {

  (void) button54;  // This statement suppresses compiler warning about unused parameter.

  GThread *thread;

  if (!ftp_busy) {

    // Select LACNIC.
    data->rir_index = 4;

    // Spawn thread to download file.
    thread = g_thread_new ("ftp_rir", (GThreadFunc) ftp_rir, data);
    if (! thread) {
      fprintf (stderr, "Error: Unable to create new thread for file download.\n");
      return (EXIT_FAILURE);
    }
  }

  return (EXIT_SUCCESS);
}

// Download all RIR ASN / IP delegation files.
int
on_button57_clicked (GtkButton *button57, SPSData *data) {

  (void) button57;  // This statement suppresses compiler warning about unused parameter.

  GThread *thread;

  if (!ftp_busy) {

    // Spawn thread to download file.
    thread = g_thread_new ("download_all", (GThreadFunc) download_all, data);
    if (! thread) {
      fprintf (stderr, "Error: Unable to create new thread for file download.\n");
      return (EXIT_FAILURE);
    }
  }

  return (EXIT_SUCCESS);
}

// View www.iso.org's ISO 3166-1 country code list file.
int
on_button56_clicked (GtkButton *button56, SPSData *data) {

  (void) button56;  // This statement suppresses compiler warning about unused parameter.

  int i, n, nbytes;
  uint8_t *filedata;
  GtkWidget *window, *scrolled_win, *textview;
  GtkTextBuffer *textbuffer;
  FILE *fi;

  // Create new window to display file contents
  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_position (GTK_WINDOW (window), GTK_WIN_POS_CENTER);
  gtk_window_set_resizable (GTK_WINDOW (window), TRUE);
  gtk_window_set_title (GTK_WINDOW (window), "ISO 3166-1 Country Code List (www.iso.org)");
  gtk_window_set_default_size (GTK_WINDOW (window), 800, 600);
  gtk_container_set_border_width (GTK_CONTAINER (window), 1);

  // Create a textview object
  textview = gtk_text_view_new ();
  textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (textview));
  gtk_text_view_set_cursor_visible (GTK_TEXT_VIEW (textview), TRUE);

  // Open www.iso.org's country code file.
  fi = fopen (data->country_file, "rb");
  if (fi == NULL) {
    sprintf (data->error_text, "Can't open file %s in on_button56_clicked().", data->country_file);
    data->parent = data->menubar1;
    report_error (data);
    return (EXIT_FAILURE);
  }

  // Count number of bytes in file.
  nbytes = 0;
  while ((n = fgetc (fi)) != EOF) {
    nbytes++;
  }
  rewind (fi);

  if (nbytes < 1) {
    memset (data->error_text, 0, TEXT_STRINGLEN * sizeof (char));
    strncpy (data->error_text, "Country code file has 0 bytes.", TEXT_STRINGLEN - 1);  // Minus 1 for string termination.
    data->parent = data->menubar1;
    report_error (data);
    return (EXIT_FAILURE);
  } else {
    filedata = allocate_ustrmem (nbytes + 1);
  }

  // Read file into array.
  for (i=0; i<nbytes; i++) {
    filedata[i] = (uint8_t) fgetc (fi);
  }
  filedata[nbytes] = 0u;  // Ensure string termination.

  // Close file descriptor.
  fclose (fi);

  // Put the text into the buffer.
  gtk_text_buffer_set_text (textbuffer, (const char *) filedata, -1);

  // Add scrolled window and textview to new window and display.
  scrolled_win = gtk_scrolled_window_new (NULL, NULL);
  gtk_container_add (GTK_CONTAINER (scrolled_win), textview);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_win), GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);

  gtk_container_add (GTK_CONTAINER (window), scrolled_win);
  gtk_widget_show_all (window);

  free (filedata);

  return (EXIT_SUCCESS);
}

// Search country names and extract country code.
int
on_entry500_changed (GtkWidget *entry500, SPSData *data)
{
  const char *entry500_text;
  GtkStyleContext *context;
  char *temp, *ccode;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (entry500);

  // Array of chars
  ccode = allocate_strmem (TEXT_STRINGLEN);

  // Array of chars
  temp = allocate_strmem (TEXT_STRINGLEN);

  // Grab country name from ui.
  entry500_text = gtk_entry_get_text (GTK_ENTRY (entry500));

  // Examine country name and report any non-UTF-8 characters.
  if (!is_valid_utf8 ((uint8_t *) entry500_text, strnlen (entry500_text, TEXT_STRINGLEN), data->main_window, data)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    free (ccode);
    free (temp);
    return (EXIT_FAILURE);

  } else {
    // Convert any lower case UTF-8/ASCII letters to upper case.
    memset (temp, 0, TEXT_STRINGLEN * sizeof (char));
    if (!raise_case ((uint8_t *) entry500_text, strnlen (entry500_text, TEXT_STRINGLEN), (uint8_t *) temp, data->main_window, data)) {
      free (ccode);
      free (temp);
      return (EXIT_FAILURE);
    }
  }

  if (!find_country_code (temp, ccode, data)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    free (ccode);
    free (temp);
    return (EXIT_FAILURE);

  } else {

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    gtk_entry_set_text (GTK_ENTRY (data->entry500), temp);
    gtk_entry_set_text (GTK_ENTRY (data->entry501), ccode);
  }

  free (ccode);
  free (temp);

  return (EXIT_SUCCESS);
}

// Search country codes and extract country name.
int
on_entry501_changed (GtkWidget *entry501, SPSData *data) {

  const char *entry501_text;
  GtkStyleContext *context;
  char *temp, *cname;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (entry501);

  // Array of chars
  cname = allocate_strmem (TEXT_STRINGLEN);

  // Array of chars
  temp = allocate_strmem (TEXT_STRINGLEN);

  // Grab country code from ui.
  entry501_text = gtk_entry_get_text (GTK_ENTRY (entry501));

  // Examine country code and report any non-UTF-8 characters.
  if (!is_valid_utf8 ((uint8_t *) entry501_text, strnlen (entry501_text, TEXT_STRINGLEN), data->main_window, data)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    free (cname);
    free (temp);
    return (EXIT_FAILURE);

  } else {
    // Convert any lower case UTF-8/ASCII letters to upper case.
    memset (temp, 0, TEXT_STRINGLEN * sizeof (char));
    if (!raise_case ((uint8_t *) entry501_text, strnlen (entry501_text, TEXT_STRINGLEN), (uint8_t *) temp, data->main_window, data)) {
      free (cname);
      free (temp);
      return (EXIT_FAILURE);
    }
  }

  if (!find_country_name (cname, temp, data)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    free (cname);
    free (temp);
    return (EXIT_FAILURE);

  } else {

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    gtk_entry_set_text (GTK_ENTRY (data->entry500), cname);
    gtk_entry_set_text (GTK_ENTRY (data->entry501), temp);
  }

  free (cname);
  free (temp);

  return (EXIT_SUCCESS);
}

// Search for autonomous system (AS) number delegations by country name.
int
on_entry502_changed (GtkWidget *entry502, SPSData *data) {

  // If a search thread is already active, then do nothing and return (EXIT_FAILURE).
  if (ip_asn_searching) {
    return (EXIT_FAILURE);
  } else {
    ip_asn_searching = 1;
  }

  const char *entry502_text;
  GtkStyleContext *context;
  char *temp, *ccode;
  GtkTextBuffer *textbuffer;
  GThread *thread;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (entry502);

  // Array of chars
  ccode = allocate_strmem (TEXT_STRINGLEN);

  // Array of chars
  temp = allocate_strmem (TEXT_STRINGLEN);

  // Clear any previous results in the textview.
  textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview16));
  gtk_text_buffer_set_text (textbuffer, "", -1);

  // If RIR files and country code file not available, return with no action.
  if (!all_files_loaded (data)) {
    sprintf (data->error_text, "You must download the RIR files and country code file before performing searches.");
    data->parent = data->main_window;
    report_error (data);
    free (ccode);
    free (temp);
    ip_asn_searching = 0;
    return (EXIT_FAILURE);
  }

  // Grab country name from ui.
  entry502_text = gtk_entry_get_text (GTK_ENTRY (entry502));

  // Examine country name and report any non-UTF-8 characters.
  if (!is_valid_utf8 ((uint8_t *) entry502_text, strnlen (entry502_text, TEXT_STRINGLEN), data->main_window, data)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    free (ccode);
    free (temp);
    return (EXIT_FAILURE);

  } else {
    // Convert any lower case UTF-8/ASCII letters to upper case.
    memset (temp, 0, TEXT_STRINGLEN * sizeof (char));
    if (!raise_case ((uint8_t *) entry502_text, strnlen (entry502_text, TEXT_STRINGLEN), (uint8_t *) temp, data->main_window, data)) {
      free (ccode);
      free (temp);
      return (EXIT_FAILURE);
    }
  }

  if (!find_country_code (temp, ccode, data)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    free (ccode);
    free (temp);
    ip_asn_searching = 0;
    return (EXIT_FAILURE);
  }

  // Remove the error style (if applied) to entry text.
  gtk_style_context_remove_class (context, "error");

  memset (data->cname, 0, TEXT_STRINGLEN * sizeof (char));
  strncpy (data->cname, temp, TEXT_STRINGLEN - 1);  // Minus 1 for string termination.
  memset (data->ccode, 0, TEXT_STRINGLEN * sizeof (char));
  strncpy (data->ccode, ccode, 2);

  // Spawn thread to perform IPv4 address delegation search.
  thread = g_thread_new ("search_asn_by_ccode", (GThreadFunc) search_asn_by_ccode, data);
  if (! thread) {
    fprintf (stderr, "Error: Unable to create new thread to search ASN delegations by country name.\n");
    free (ccode);
    free (temp);
    ip_asn_searching = 0;
    gtk_text_buffer_set_text (textbuffer, "", -1);
    return (EXIT_FAILURE);
  }

  gtk_entry_set_text (GTK_ENTRY (data->entry502), data->cname);
  gtk_entry_set_text (GTK_ENTRY (data->entry503), data->ccode);

  free (ccode);
  free (temp);

  return (EXIT_SUCCESS);
}

// Search for autonomous system (AS) number delegations by country code.
int
on_entry503_changed (GtkWidget *entry503, SPSData *data) {

  // If a search thread is already active, then do nothing and return (EXIT_FAILURE).
  if (ip_asn_searching) {
    return (EXIT_FAILURE);
  } else {
    ip_asn_searching = 1;
  }

  const char *entry503_text;
  GtkStyleContext *context;
  char *temp, *cname;
  GtkTextBuffer *textbuffer;
  GThread *thread;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (entry503);

  // Array of chars
  cname = allocate_strmem (TEXT_STRINGLEN);

  // Array of chars
  temp = allocate_strmem (TEXT_STRINGLEN);

  // Clear any previous results in the textview.
  textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview16));
  gtk_text_buffer_set_text (textbuffer, "", -1);

  // If RIR files and country code file not available, return with no action.
  if (!all_files_loaded (data)) {
    sprintf (data->error_text, "You must download the RIR files and country code file before performing searches.");
    data->parent = data->main_window;
    report_error (data);
    free (cname);
    free (temp);
    ip_asn_searching = 0;
    return (EXIT_FAILURE);
  }

  // Grab country code from ui.
  entry503_text = gtk_entry_get_text (GTK_ENTRY (entry503));

  // Examine country code and report any non-UTF-8 characters.
  if (!is_valid_utf8 ((uint8_t *) entry503_text, strnlen (entry503_text, TEXT_STRINGLEN), data->main_window, data)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    free (cname);
    free (temp);
    return (EXIT_FAILURE);

  } else {
    // Convert any lower case UTF-8/ASCII letters to upper case.
    memset (temp, 0, TEXT_STRINGLEN * sizeof (char));
    if (!raise_case ((uint8_t *) entry503_text, strnlen (entry503_text, TEXT_STRINGLEN), (uint8_t *) temp, data->main_window, data)) {
      free (cname);
      free (temp);
      return (EXIT_FAILURE);
    }
  }

  if (!find_country_name (cname, temp, data)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    free (cname);
    free (temp);
    ip_asn_searching = 0;
    return (EXIT_FAILURE);
  }

  // Remove the error style (if applied) to entry text.
  gtk_style_context_remove_class (context, "error");

  memset (data->cname, 0, TEXT_STRINGLEN * sizeof (char));
  strncpy (data->cname, cname, TEXT_STRINGLEN - 1);  // Minus 1 for string termination.
  memset (data->ccode, 0, 3 * sizeof (char));
  strncpy (data->ccode, temp, 2);

  // Spawn thread to perform ASN delegation search.
  thread = g_thread_new ("search_asn_by_ccode", (GThreadFunc) search_asn_by_ccode, data);
  if (! thread) {
    fprintf (stderr, "Error: Unable to create new thread to search ASN delegations by country name.\n");
    free (cname);
    free (temp);
    ip_asn_searching = 0;
    gtk_text_buffer_set_text (textbuffer, "", -1);
    return (EXIT_FAILURE);
  }

  gtk_entry_set_text (GTK_ENTRY (data->entry502), data->cname);
  gtk_entry_set_text (GTK_ENTRY (data->entry503), data->ccode);

  free (cname);
  free (temp);

  return (EXIT_SUCCESS);
}

// Reverse search for country by autonomous system (AS) number delegations.
int
on_entry504_changed (GtkWidget *entry504, SPSData *data) {

  // If a search thread is already active, then do nothing and return (EXIT_FAILURE).
  if (ip_asn_searching) {
    return (EXIT_FAILURE);
  } else {
    ip_asn_searching = 1;
  }

  const char *entry504_text;
  GtkStyleContext *context;
  GtkTextBuffer *textbuffer;
  GThread *thread;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (entry504);

  // Clear any previous results in the textview.
  textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview17));
  gtk_text_buffer_set_text (textbuffer, "", -1);

  // If RIR files and country code file not available, return with no action.
  if (!all_files_loaded (data)) {
    sprintf (data->error_text, "You must download the RIR files and country code file before performing searches.");
    data->parent = data->main_window;
    report_error (data);
    ip_asn_searching = 0;
    return (EXIT_FAILURE);
  }

  // Grab ASN from ui.
  entry504_text = gtk_entry_get_text (GTK_ENTRY (entry504));

  if (!is_ascii_uint (entry504_text)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    ip_asn_searching = 0;
    return (EXIT_FAILURE);

  } else {

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    data->asn = (uint64_t) ascii_to_int64 ((char *) entry504_text);
  }

  // Spawn thread to perform ASN delegation search.
  thread = g_thread_new ("rev_search_asn", (GThreadFunc) rev_search_asn, data);
  if (! thread) {
    fprintf (stderr, "Error: Unable to create new thread to reverse-search ASN delegations by ASN.\n");
    gtk_text_buffer_set_text (textbuffer, "", -1);
    ip_asn_searching = 0;
    return (EXIT_FAILURE);
  }

  return (EXIT_SUCCESS);
}

// Search for IP address delegations by country name.
int
on_entry505_changed (GtkWidget *entry505, SPSData *data) {

  // If a search thread is already active, then do nothing and return (EXIT_FAILURE).
  if (ip_asn_searching) {
    return (EXIT_FAILURE);
  } else {
    ip_asn_searching = 1;
  }

  const char *entry505_text;
  GtkStyleContext *context;
  char *temp, *ccode;
  GtkTextBuffer *textbuffer18;
  GtkTextBuffer *textbuffer19;
  GThread *thread4, *thread6;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (entry505);

  // Array of chars
  ccode = allocate_strmem (TEXT_STRINGLEN);

  // Array of chars
  temp = allocate_strmem (TEXT_STRINGLEN);

  // Clear any text in the textviews.
  textbuffer18 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview18));  // IPv4
  gtk_text_buffer_set_text (textbuffer18, "", -1);
  textbuffer19 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview19));  // IPv6
  gtk_text_buffer_set_text (textbuffer19, "", -1);

  // If RIR files and country code file not available, return with no action.
  if (!all_files_loaded (data)) {
    sprintf (data->error_text, "You must download the RIR files and country code file before performing searches.");
    data->parent = data->main_window;
    report_error (data);
    free (ccode);
    free (temp);
    ip_asn_searching = 0;
    return (EXIT_FAILURE);
  }

  // Grab country name from ui.
  entry505_text = gtk_entry_get_text (GTK_ENTRY (entry505));

  // Examine country name and report any non-UTF-8 characters.
  if (!is_valid_utf8 ((uint8_t *) entry505_text, strnlen (entry505_text, TEXT_STRINGLEN), data->main_window, data)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    free (ccode);
    free (temp);
    return (EXIT_FAILURE);

  } else {
    // Convert any lower case UTF-8/ASCII letters to upper case.
    memset (temp, 0, TEXT_STRINGLEN * sizeof (char));
    if (!raise_case ((uint8_t *) entry505_text, strnlen (entry505_text, TEXT_STRINGLEN), (uint8_t *) temp, data->main_window, data)) {
      free (ccode);
      free (temp);
      return (EXIT_FAILURE);
    }
  }

  if (!find_country_code (temp, ccode, data)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    free (ccode);
    free (temp);
    ip_asn_searching = 0;
    return (EXIT_FAILURE);
  }

  // Remove the error style (if applied) to entry text.
  gtk_style_context_remove_class (context, "error");

  memset (data->cname, 0, TEXT_STRINGLEN * sizeof (char));
  strncpy (data->cname, temp, TEXT_STRINGLEN - 1);  // Minus 1 for string termination.
  memset (data->ccode, 0, 3 * sizeof (char));
  strncpy (data->ccode, ccode, 2);

  // Spawn thread to perform IPv4 address delegation search.
  thread4 = g_thread_new ("search_ipv4_by_ccode", (GThreadFunc) search_ipv4_by_ccode, data);
  if (! thread4) {
    fprintf (stderr, "Error: Unable to create new thread to search IPv4 address delegations by country name.\n");
    free (ccode);
    free (temp);
    ip_asn_searching = 0;  // OK to set to zero since we'd never get to spawn IPv6 search.
    gtk_text_buffer_set_text (textbuffer18, "", -1);  // IPv4
    return (EXIT_FAILURE);
  }

  // Spawn thread to perform IPv6 address delegation search.
  thread6 = g_thread_new ("search_ipv6_by_ccode", (GThreadFunc) search_ipv6_by_ccode, data);
  if (! thread6) {
    fprintf (stderr, "Error: Unable to create new thread to search IPv6 address delegations by country name.\n");
    free (ccode);
    free (temp);
    // Can't set ip_asn_searching to zero since IPv4 is searching.
    gtk_text_buffer_set_text (textbuffer19, "", -1);  // IPv6
    return (EXIT_FAILURE);
  }

  gtk_entry_set_text (GTK_ENTRY (data->entry505), data->cname);
  gtk_entry_set_text (GTK_ENTRY (data->entry506), data->ccode);

  free (ccode);
  free (temp);

  return (EXIT_SUCCESS);
}

// Search for IP address delegations by country code.
int
on_entry506_changed (GtkWidget *entry506, SPSData *data) {

  // If a search thread is already active, then do nothing and return (EXIT_FAILURE).
  if (ip_asn_searching) {
    return (EXIT_FAILURE);
  }

  const char *entry506_text;
  GtkStyleContext *context;
  char *temp, *cname;
  GtkTextBuffer *textbuffer18;
  GtkTextBuffer *textbuffer19;
  GThread *thread4, *thread6;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (entry506);

  // Array of chars
  cname = allocate_strmem (TEXT_STRINGLEN);

  // Array of chars
  temp = allocate_strmem (TEXT_STRINGLEN);

  // Clear any text in the textviews.
  textbuffer18 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview18));  // IPv4
  gtk_text_buffer_set_text (textbuffer18, "", -1);
  textbuffer19 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview19));  // IPv6
  gtk_text_buffer_set_text (textbuffer19, "", -1);

  // If RIR files and country code file not available, return with no action.
  if (!all_files_loaded (data)) {
    sprintf (data->error_text, "You must download the RIR files and country code file before performing searches.");
    data->parent = data->main_window;
    report_error (data);
    free (cname);
    free (temp);
    ip_asn_searching = 0;
    return (EXIT_FAILURE);
  }

  // Grab country code from ui.
  entry506_text = gtk_entry_get_text (GTK_ENTRY (entry506));

  // Examine country code and report any non-UTF-8 characters.
  if (!is_valid_utf8 ((uint8_t *) entry506_text, strnlen (entry506_text, TEXT_STRINGLEN), data->main_window, data)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    free (cname);
    free (temp);
    return (EXIT_FAILURE);

  } else {
    // Convert any lower case UTF-8/ASCII letters to upper case.
    memset (temp, 0, TEXT_STRINGLEN * sizeof (char));
    if (!raise_case ((uint8_t *) entry506_text, strnlen (entry506_text, TEXT_STRINGLEN), (uint8_t *) temp, data->main_window, data)) {
      free (cname);
      free (temp);
      return (EXIT_FAILURE);
    }
  }

  if (!find_country_name (cname, temp, data)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    free (cname);
    free (temp);
    ip_asn_searching = 0;
    return (EXIT_FAILURE);
  }

  // Remove the error style (if applied) to entry text.
  gtk_style_context_remove_class (context, "error");

  memset (data->cname, 0, TEXT_STRINGLEN * sizeof (char));
  strncpy (data->cname, cname, TEXT_STRINGLEN - 1);  // Minus 1 for string termination.
  memset (data->ccode, 0, 3 * sizeof (char));
  strncpy (data->ccode, temp, 2);

  // Spawn thread to perform IPv4 address delegation search.
  thread4 = g_thread_new ("search_ipv4_by_ccode", (GThreadFunc) search_ipv4_by_ccode, data);
  if (! thread4) {
    fprintf (stderr, "Error: Unable to create new thread to search IPv4 address delegations by country name.\n");
    free (cname);
    free (temp);
    ip_asn_searching = 0;  // OK to set to zero since we'd never get to spawn IPv6 search.
    gtk_text_buffer_set_text (textbuffer18, "", -1);  // IPv4
    return (EXIT_FAILURE);
  }

  // Spawn thread to perform IPv6 address delegation search.
  thread6 = g_thread_new ("search_ipv6_by_ccode", (GThreadFunc) search_ipv6_by_ccode, data);
  if (! thread6) {
    fprintf (stderr, "Error: Unable to create new thread to search IPv6 address delegations by country name.\n");
    free (cname);
    free (temp);
    // Can't set ip_asn_searching to zero since IPv4 is searching.
    gtk_text_buffer_set_text (textbuffer19, "", -1);  // IPv6
    return (EXIT_FAILURE);
  }

  gtk_entry_set_text (GTK_ENTRY (data->entry505), data->cname);
  gtk_entry_set_text (GTK_ENTRY (data->entry506), data->ccode);

  free (cname);
  free (temp);

  return (EXIT_SUCCESS);
}

// Reverse-search for IPv4 address delegations by IPv4 address.
int
on_entry507_changed (GtkWidget *entry507, SPSData *data) {

  // If a search thread is already active, then do nothing and return (EXIT_FAILURE).
  if (ip_asn_searching) {
    return (EXIT_FAILURE);
  }

  // Clear IPv6 text entry.
  gtk_entry_set_text (GTK_ENTRY (data->entry508), "");

  const char *entry507_text;
  GtkStyleContext *context;
  GtkTextBuffer *textbuffer20;
  struct sockaddr_in sa;
  GThread *thread;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (entry507);

  // Clear any previous results in the textview.
  textbuffer20 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview20));
  gtk_text_buffer_set_text (textbuffer20, "", -1);

  // If RIR files and country code file not available, return with no action.
  if (!all_files_loaded (data)) {
    sprintf (data->error_text, "You must download the RIR files and country code file before performing searches.");
    data->parent = data->main_window;
    report_error (data);
    ip_asn_searching = 0;
    return (EXIT_FAILURE);
  }

  // Grab IPv4 address from ui.
  entry507_text = gtk_entry_get_text (GTK_ENTRY (entry507));

  // Check to see if address is a valid IPv4 address.
  if (inet_pton (AF_INET, entry507_text, &sa.sin_addr) < 1) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    ip_asn_searching = 0;
    return (EXIT_FAILURE);
  }

  // Remove the error style (if applied) to entry text.
  gtk_style_context_remove_class (context, "error");

  memset (data->ip4string, 0, INET_ADDRSTRLEN * sizeof (char));
  strncpy (data->ip4string, entry507_text, INET_ADDRSTRLEN - 1);  // Minus 1 for string termination.

  // Spawn thread to perform IPv4 address delegation search.
  thread = g_thread_new ("rev_search_ipv4", (GThreadFunc) rev_search_ipv4, data);
  if (! thread) {
    fprintf (stderr, "Error: Unable to create new thread to reverse-search IPv4 address delegations by IPv4 address.\n");
    gtk_text_buffer_set_text (textbuffer20, "", -1);
    ip_asn_searching = 0;
    return (EXIT_FAILURE);
  }

  return (EXIT_SUCCESS);
}

// Reverse-search for IPv6 address delegations by IPv6 address.
int
on_entry508_changed (GtkWidget *entry508, SPSData *data) {

  // If a search thread is already active, then do nothing and return (EXIT_FAILURE).
  if (ip_asn_searching) {
    return (EXIT_FAILURE);
  }

  // Clear IPv4 text entry.
  gtk_entry_set_text (GTK_ENTRY (data->entry507), "");

  const char *entry508_text;
  GtkStyleContext *context;
  GtkTextBuffer *textbuffer20;
  struct sockaddr_in6 sa;
  GThread *thread;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (entry508);

  // Clear any previous results in the textview.
  textbuffer20 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview20));
  gtk_text_buffer_set_text (textbuffer20, "", -1);

  // If RIR files and country code file not available, return with no action.
  if (!all_files_loaded (data)) {
    sprintf (data->error_text, "You must download the RIR files and country code file before performing searches.");
    data->parent = data->main_window;
    report_error (data);
    ip_asn_searching = 0;
    return (EXIT_FAILURE);
  }

  // Grab IPv6 address from ui.
  entry508_text = gtk_entry_get_text (GTK_ENTRY (entry508));

  // Check to see if address is a valid IPv6 address.
  if (inet_pton (AF_INET6, entry508_text, &sa.sin6_addr) < 1) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    ip_asn_searching = 0;
    return (EXIT_FAILURE);
  }

  // Remove the error style (if applied) to entry text.
  gtk_style_context_remove_class (context, "error");

  memset (data->ip6string, 0, INET6_ADDRSTRLEN * sizeof (char));
  strncpy (data->ip6string, entry508_text, INET6_ADDRSTRLEN - 1);  // Minus 1 for string termination.

  // Spawn thread to perform IPv6 address delegation search.
  thread = g_thread_new ("rev_search_ipv6", (GThreadFunc) rev_search_ipv6, data);
  if (! thread) {
    fprintf (stderr, "Error: Unable to create new thread to reverse-search IPv6 address delegations by IPv6 address.\n");
    gtk_text_buffer_set_text (textbuffer20, "", -1);
    ip_asn_searching = 0;
    return (EXIT_FAILURE);
  }
  return (EXIT_SUCCESS);
}

// Pseudo-random number generator (PRNG) page

// Seed for PRNG
int
on_entry319_changed (GtkWidget *entry319, SPSData *data) {

  const char *entry_text;
  GtkStyleContext *context;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (entry319);

  // Get new value from text entry
  entry_text = gtk_entry_get_text (GTK_ENTRY (entry319));

  // Is value too large or small, or are there non-numeric characters?
  if (!is_ascii_uint (entry_text) || (ascii_to_int64 ((char *) entry_text) < 1ll)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    return (EXIT_FAILURE);

  // Value is in acceptable range.
  } else {

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    data->RN_SEED0 = (uint64_t) ascii_to_int64 ((char *) entry_text);  // Update with new value
    data->RN_ITER = 0ull;  // Start at beginning of sequence.

    // Update all randomized parameters (IPv4 and IPv6, including traceroute).
    update_ran_parms (data);
  }

  return (EXIT_SUCCESS);
}

// Stride for PRNG
int
on_entry320_changed (GtkWidget *entry320, SPSData *data) {

  const char *entry_text;
  GtkStyleContext *context;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (entry320);

  // Get new value from text entry
  entry_text = gtk_entry_get_text (GTK_ENTRY (entry320));

  // Is value too large or small, or are there non-numeric characters?
  if (!is_ascii_uint (entry_text) || (ascii_to_int64 ((char *) entry_text) < 1ll)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    return (EXIT_FAILURE);

  // Value is in acceptable range.
  } else {

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    data->RN_STRIDE = (uint64_t) ascii_to_int64 ((char *) entry_text);  // Update with new value
  }

  return (EXIT_SUCCESS);
}

// Hop-by-hop header (Original/Inner) (hop-by-hop.ui)

// Toggle: attach / detach a hop-by-hop header (hop-by-hop.ui)
int
on_hop_button142_clicked (GtkButton *hop_button142, SPSData *data) {

  (void) hop_button142;  // This statement suppresses compiler warning about unused parameter.
  GtkTextBuffer *textbuffer47;

  textbuffer47 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->hop_textview47));

  // Detach a hop-by-hop header.
  if (data->hbh_hdr_flag[data->exthdr_type][0]) {
    data->hbh_hdr_flag[data->exthdr_type][0] = 0;
    gtk_text_buffer_set_text (textbuffer47, "Not Attached", -1);

  // Attach a hop-by-hop header.
  } else {
    // Abort if there are no hop-by-hop options.
    if (!data->hbh_nopt[data->exthdr_type][0]) {
      data->hbh_hdr_flag[data->exthdr_type][0] = 0;
      gtk_text_buffer_set_text (textbuffer47, "Not Attached", -1);
      sprintf (data->error_text, "In on_button142_clicked: Cannot attach a hop-by-hop header since no options were specified.");
      data->parent = data->hop_window;
      report_error (data);
      return (EXIT_FAILURE);
    } else {
      data->hbh_hdr_flag[data->exthdr_type][0] = 1;
      gtk_text_buffer_set_text (textbuffer47, "Attached", -1);
    }
  }

  // Update ethernet frame and hop_window widgets.
  ip6data_update (data->exthdr_type, data);

  return (EXIT_SUCCESS);
}

// Set hop_flag to zero when hop_window is destroyed. (hop-by-hop.ui)
int
on_hop_window_destroy (GtkWidget *widget, SPSData *data) {

  (void) widget;  // This statement suppresses compiler warning about unused parameter.

  data->hop_flag[0] = 0;

  return (EXIT_SUCCESS);
}

// Hop-by-hop header length (hop-by-hop.ui)
int
on_hop_entry438_changed (GtkWidget *hop_entry438, SPSData *data) {

  const char *entry_text;
  GtkStyleContext *context;
  char *value;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (hop_entry438);

  // String to hold miscellaneous calculated values
  value = allocate_strmem (TMP_STRINGLEN);

  // Get new value from text entry
  entry_text = gtk_entry_get_text (GTK_ENTRY (hop_entry438));

  // Is value too large or small, or are there non-numeric characters?
  if (!is_ascii_uint (entry_text) || (ascii_to_int64 ((char *) entry_text) > 255)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    free (value);
    return (EXIT_FAILURE);

  // Value is in acceptable range.
  } else {

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    data->hophdr[data->exthdr_type][0].hdr_len = (uint8_t) ascii_to_int64 ((char *) entry_text);  // Update with new value

    // Update ethernet frame and hop_window widgets.
    ip6data_update (data->exthdr_type, data);
  }

  // Free allocated memory.
  free (value);

  return (EXIT_SUCCESS);
}

// Decimal option data entry (hop-by-hop.ui)
int
on_hop_radiobutton59_toggled (GtkButton *hop_radiobutton59, SPSData *data) {

  (void) hop_radiobutton59;  // This statement suppresses compiler warning about unused parameter.

  if (data->dec_hex_hbhopt[0]) {
    data->dec_hex_hbhopt[0] = 0;
  } else {
    data->dec_hex_hbhopt[0] = 1;
  }

  return (EXIT_SUCCESS);
}

// Hexadecimal option data entry (hop-by-hop.ui)
int
on_hop_radiobutton60_toggled (GtkButton *hop_radiobutton60) {

  (void) hop_radiobutton60;  // This statement suppresses compiler warning about unused parameter.

  // on_radiobutton59_toggled() will change value of data->dec_hex_hbhopt[0].

  return (EXIT_SUCCESS);
}

// Hop-by-hop option data (hop-by-hop.ui)
int
on_hop_entry158_changed (GtkWidget *hop_entry158, SPSData *data) {

  data_entry (hop_entry158, &data->hbh_optlenbuf[data->exthdr_type][0], MAX_HBHOPTLEN, data->hbh_optionsbuf[data->exthdr_type][0], data->dec_hex_hbhopt[0], data->hop_window, data);

  return (EXIT_SUCCESS);
}

// Hop-by-hop option alignment parameter x (of (xn + y) (hop-by-hop.ui)
int
on_hop_entry313_changed (GtkWidget *hop_entry313, SPSData *data) {

  (void) data;  // This statement suppresses compiler warning about unused parameter.

  const char *entry313_text;
  GtkStyleContext *context;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (hop_entry313);

  // Get new value from text entry
  entry313_text = gtk_entry_get_text (GTK_ENTRY (hop_entry313));

  // Is value too large or small, or are there non-numeric characters?
  if (!is_ascii_uint (entry313_text) || (ascii_to_int64 ((char *) entry313_text) > 255)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    return (EXIT_FAILURE);

  // Value is OK.
  } else {

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    // We use the entry to store this alignment value.
    // hop_option() will obtain it from the user interface.
  }

  return (EXIT_SUCCESS);
}

// Hop-by-hop option alignment parameter y (of (xn + y) (hop-by-hop.ui)
int
on_hop_entry314_changed (GtkWidget *hop_entry314, SPSData *data) {

  (void) data;  // This statement suppresses compiler warning about unused parameter.

  const char *hop_entry314_text;
  GtkStyleContext *context;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (hop_entry314);

  // Get new value from text entry
  hop_entry314_text = gtk_entry_get_text (GTK_ENTRY (hop_entry314));

  // Is value too large or small, or are there non-numeric characters?
  if (!is_ascii_uint (hop_entry314_text) || (ascii_to_int64 ((char *) hop_entry314_text) > 255u)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    return (EXIT_FAILURE);

  // Value is OK.
  } else {

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    // We use the entry to store this alignment value.
    // hop_option() will obtain it from the user interface.
  }

  return (EXIT_SUCCESS);
}

// Insert hop-by-hop option after another, as specified by user. (hop-by-hop.ui)
int
on_hop_button94_clicked (GtkButton *hop_button94, SPSData *data) {

  (void) hop_button94;  // This statement suppresses compiler warning about unused parameter.

  if (hop_option (0, 0, data->hbhopt_after[0], data) == EXIT_SUCCESS) {

    // Update hop-by-hop header length in hop_window.
    hop_show (data->exthdr_type, data);

    // Update ethernet frame if hop-by-hop header is attached.
    if (data->hbh_hdr_flag[data->exthdr_type][0]) {
      ip6data_update (data->exthdr_type, data);
    }
  }

  return (EXIT_SUCCESS);
}

// Number of option after which to insert new hop-by-hop option. (hop-by-hop.ui)
int
on_hop_entry159_changed (GtkWidget *hop_entry159, SPSData *data) {

  const char *entry_text;
  GtkStyleContext *context;

  // Get new value from text entry
  entry_text = gtk_entry_get_text (GTK_ENTRY (hop_entry159));

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (hop_entry159);

  // Is value too large or small, or are there non-numeric characters?
  if (!is_ascii_uint (entry_text) || (!ascii_to_int64 ((char *) entry_text))) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    return (EXIT_FAILURE);

  // Value is in acceptable range.
  } else {
    data->hbhopt_after[0] = (int) ascii_to_int64 ((char *) entry_text);

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");
  }

  return (EXIT_SUCCESS);
}

// Append an hop-by-hop option. (hop-by-hop.ui)
int
on_hop_button95_clicked (GtkButton *hop_button95, SPSData *data) {

  (void) hop_button95;  // This statement suppresses compiler warning about unused parameter.

  if (hop_option (0, 2, 0, data) == EXIT_SUCCESS) {

    // Update hop-by-hop header length in hop_window.
    hop_show (data->exthdr_type, data);

    // Update ethernet frame if hop-by-hop header is attached.
    if (data->hbh_hdr_flag[data->exthdr_type][0]) {
      ip6data_update (data->exthdr_type, data);
    }
  }

  return (EXIT_SUCCESS);
}

// View hop-by-hop header options. (hop-by-hop.ui)
int
on_hop_button96_clicked (GtkButton *hop_button96, SPSData *data) {

  (void) hop_button96;  // This statement suppresses compiler warning about unused parameter.

  hop_opt_view (data);

  return (EXIT_SUCCESS);
}

// Number of hop-by-hop option to remove. (hop-by-hop.ui)
int
on_hop_entry312_changed (GtkWidget *hop_entry312, SPSData *data) {

  dsthop_opt_remove_entry (hop_entry312, &data->hbhopt_remove[0]);

  return (EXIT_SUCCESS);
}

// Remove selected hop-by-hop option. (hop-by-hop.ui)
int
on_hop_button97_clicked (GtkButton *hop_button97, SPSData *data) {

  (void) hop_button97;  // This statement suppresses compiler warning about unused parameter.

  if (hop_option (0, 1, data->hbhopt_remove[0], data) == EXIT_SUCCESS) {

    // Update ethernet header if hop-by-hop header is attached.
    if (data->hbh_hdr_flag[data->exthdr_type][0]) {

      // If number of options in now zero, detach header.
      if (!data->hbh_nopt[data->exthdr_type][0]) {
        data->hbh_hdr_flag[data->exthdr_type][0] = 0;
      }

      // Update ethernet header.
      ip6data_update (data->exthdr_type, data);
    }

    // Update hop-by-hop header length and attached/detached textview in hop_window.
    hop_show (data->exthdr_type, data);
  }

  return (EXIT_SUCCESS);
}

// Hop-by-hop header (outer) (hop-by-hop-outer.ui)

// Toggle: attach / detach an outer hop-by-hop header (hop-by-hop-outer.ui)
int
on_hop_button142_outer_clicked (GtkButton *hop_button142_outer, SPSData *data) {

  (void) hop_button142_outer;  // This statement suppresses compiler warning about unused parameter.
  GtkTextBuffer *textbuffer47_outer;

  textbuffer47_outer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->hop_textview47_outer));

  // Detach an outer hop-by-hop header.
  if (data->hbh_hdr_flag[data->exthdr_type][1]) {
    data->hbh_hdr_flag[data->exthdr_type][1] = 0;
    gtk_text_buffer_set_text (textbuffer47_outer, "Not Attached", -1);

  // Attach an outer hop-by-hop header.
  } else {
    // Abort if there are no outer hop-by-hop options.
    if (!data->hbh_nopt[data->exthdr_type][1]) {
      data->hbh_hdr_flag[data->exthdr_type][1] = 0;
      gtk_text_buffer_set_text (textbuffer47_outer, "Not Attached", -1);
      sprintf (data->error_text, "In on_button142_outer_clicked: Cannot attach an outer hop-by-hop header since no options were specified.");
      data->parent = data->hop_window_outer;
      report_error (data);
      return (EXIT_FAILURE);
    } else {
      data->hbh_hdr_flag[data->exthdr_type][1] = 1;
      gtk_text_buffer_set_text (textbuffer47_outer, "Attached", -1);
    }
  }

  // Update ethernet frame and hop_window widgets.
  ip6data_update (data->exthdr_type, data);

  return (EXIT_SUCCESS);
}

// Set hop_flag to zero when hop_window_outer is destroyed. (hop-by-hop-outer.ui)
int
on_hop_window_outer_destroy (GtkWidget *widget, SPSData *data) {

  (void) widget;  // This statement suppresses compiler warning about unused parameter.

  data->hop_flag[1] = 0;

  return (EXIT_SUCCESS);
}

// Hop-by-hop (outer) header length (hop-by-hop-outer.ui)
int
on_hop_entry438_outer_changed (GtkWidget *hop_entry438_outer, SPSData *data) {

  const char *entry_text;
  GtkStyleContext *context;
  char *value;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (hop_entry438_outer);

  // String to hold miscellaneous calculated values
  value = allocate_strmem (TMP_STRINGLEN);

  // Get new value from text entry
  entry_text = gtk_entry_get_text (GTK_ENTRY (hop_entry438_outer));

  // Is value too large or small, or are there non-numeric characters?
  if (!is_ascii_uint (entry_text) || (ascii_to_int64 ((char *) entry_text) > 255)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    free (value);
    return (EXIT_FAILURE);

  // Value is in acceptable range.
  } else {

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    data->hophdr[data->exthdr_type][1].hdr_len = (uint8_t) ascii_to_int64 ((char *) entry_text);  // Update with new value

    // Update ethernet frame and hop_window widgets.
    ip6data_update (data->exthdr_type, data);
  }

  // Free allocated memory.
  free (value);

  return (EXIT_SUCCESS);
}

// Decimal option data entry (hop-by-hop-outer.ui)
int
on_hop_radiobutton59_outer_toggled (GtkButton *hop_radiobutton59_outer, SPSData *data) {

  (void) hop_radiobutton59_outer;  // This statement suppresses compiler warning about unused parameter.

  if (data->dec_hex_hbhopt[1]) {
    data->dec_hex_hbhopt[1] = 0;
  } else {
    data->dec_hex_hbhopt[1] = 1;
  }

  return (EXIT_SUCCESS);
}

// Hexadecimal option data entry (hop-by-hop-outer.ui)
int
on_hop_radiobutton60_outer_toggled (GtkButton *hop_radiobutton60_outer) {

  (void) hop_radiobutton60_outer;  // This statement suppresses compiler warning about unused parameter.

  // on_radiobutton59_outer_toggled() will change value of data->dec_hex_hbhopt[1].

  return (EXIT_SUCCESS);
}

// Hop-by-hop option data (hop-by-hop-outer.ui)
int
on_hop_entry158_outer_changed (GtkWidget *hop_entry158_outer, SPSData *data) {

  data_entry (hop_entry158_outer, &data->hbh_optlenbuf[data->exthdr_type][1], MAX_HBHOPTLEN, data->hbh_optionsbuf[data->exthdr_type][1], data->dec_hex_hbhopt[1], data->hop_window_outer, data);

  return (EXIT_SUCCESS);
}

// Hop-by-hop (outer) option alignment parameter x (of (xn + y) (hop-by-hop-outer.ui)
int
on_hop_entry313_outer_changed (GtkWidget *hop_entry313_outer, SPSData *data) {

  (void) data;  // This statement suppresses compiler warning about unused parameter.

  const char *entry313_outer_text;
  GtkStyleContext *context;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (hop_entry313_outer);

  // Get new value from text entry
  entry313_outer_text = gtk_entry_get_text (GTK_ENTRY (hop_entry313_outer));

  // Is value too large or small, or are there non-numeric characters?
  if (!is_ascii_uint (entry313_outer_text) || (ascii_to_int64 ((char *) entry313_outer_text) > 255)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    return (EXIT_FAILURE);

  // Value is OK.
  } else {

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    // We use the entry to store this alignment value.
    // hop_option() will obtain it from the user interface.
  }

  return (EXIT_SUCCESS);
}

// Hop-by-hop (outer) option alignment parameter y (of (xn + y) (hop-by-hop-outer.ui)
int
on_hop_entry314_outer_changed (GtkWidget *hop_entry314_outer, SPSData *data) {

  (void) data;  // This statement suppresses compiler warning about unused parameter.

  const char *hop_entry314_outer_text;
  GtkStyleContext *context;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (hop_entry314_outer);

  // Get new value from text entry
  hop_entry314_outer_text = gtk_entry_get_text (GTK_ENTRY (hop_entry314_outer));

  // Is value too large or small, or are there non-numeric characters?
  if (!is_ascii_uint (hop_entry314_outer_text) || (ascii_to_int64 ((char *) hop_entry314_outer_text) > 255u)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    return (EXIT_FAILURE);

  // Value is OK.
  } else {

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    // We use the entry to store this alignment value.
    // hop_option() will obtain it from the user interface.
  }

  return (EXIT_SUCCESS);
}

// Insert hop-by-hop (outer) option after another, as specified by user. (hop-by-hop-outer.ui)
int
on_hop_button94_outer_clicked (GtkButton *hop_button94_outer, SPSData *data) {

  (void) hop_button94_outer;  // This statement suppresses compiler warning about unused parameter.

  if (hop_option (1, 0, data->hbhopt_after[1], data) == EXIT_SUCCESS) {

    // Update hop-by-hop header length in hop_window.
    hop_show_outer (data->exthdr_type, data);

    // Update ethernet frame if outer hop-by-hop header is attached.
    if (data->hbh_hdr_flag[data->exthdr_type][1]) {
      ip6data_update (data->exthdr_type, data);
    }
  }

  return (EXIT_SUCCESS);
}

// Number of option after which to insert new hop-by-hop option. (hop-by-hop-outer.ui)
int
on_hop_entry159_outer_changed (GtkWidget *hop_entry159_outer, SPSData *data) {

  const char *entry_text;
  GtkStyleContext *context;

  // Get new value from text entry
  entry_text = gtk_entry_get_text (GTK_ENTRY (hop_entry159_outer));

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (hop_entry159_outer);

  // Is value too large or small, or are there non-numeric characters?
  if (!is_ascii_uint (entry_text) || (!ascii_to_int64 ((char *) entry_text))) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    return (EXIT_FAILURE);

  // Value is in acceptable range.
  } else {
    data->hbhopt_after[1] = (int) ascii_to_int64 ((char *) entry_text);

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");
  }

  return (EXIT_SUCCESS);
}

// Append an outer hop-by-hop option. (hop-by-hop-outer.ui)
int
on_hop_button95_outer_clicked (GtkButton *hop_button95_outer, SPSData *data) {

  (void) hop_button95_outer;  // This statement suppresses compiler warning about unused parameter.

  if (hop_option (1, 2, 0, data) == EXIT_SUCCESS) {

    // Update outer hop-by-hop header length in hop_window_outer.
    hop_show_outer (data->exthdr_type, data);

    // Update ethernet frame if hop-by-hop header is attached.
    if (data->hbh_hdr_flag[data->exthdr_type][1]) {
      ip6data_update (data->exthdr_type, data);
    }
  }

  return (EXIT_SUCCESS);
}

// View hop-by-hop (outer) header options. (hop-by-hop-outer.ui)
int
on_hop_button96_outer_clicked (GtkButton *hop_button96_outer, SPSData *data) {

  (void) hop_button96_outer;  // This statement suppresses compiler warning about unused parameter.

  hop_opt_view_outer (data);

  return (EXIT_SUCCESS);
}

// Number of hop-by-hop (outer) option to remove. (hop-by-hop-outer.ui)
int
on_hop_entry312_outer_changed (GtkWidget *hop_entry312_outer, SPSData *data) {

  dsthop_opt_remove_entry (hop_entry312_outer, &data->hbhopt_remove[1]);

  return (EXIT_SUCCESS);
}

// Remove selected hop-by-hop (outer) option. (hop-by-hop-outer.ui)
int
on_hop_button97_outer_clicked (GtkButton *hop_button97_outer, SPSData *data) {

  (void) hop_button97_outer;  // This statement suppresses compiler warning about unused parameter.

  if (hop_option (1, 1, data->hbhopt_remove[1], data) == EXIT_SUCCESS) {

    // Update ethernet header if hop-by-hop header is attached.
    if (data->hbh_hdr_flag[data->exthdr_type][1]) {

      // If number of options in now zero, detach header.
      if (!data->hbh_nopt[data->exthdr_type][1]) {
        data->hbh_hdr_flag[data->exthdr_type][1] = 0;
      }

      // Update ethernet header.
      ip6data_update (data->exthdr_type, data);
    }

    // Update hop-by-hop header length and attached/detached textview in hop_window.
    hop_show_outer (data->exthdr_type, data);
  }

  return (EXIT_SUCCESS);
}

// Destination (original/inner) Header (destination.ui)

// Toggle: attach / detach a "first" destination header (destination.ui)
int
on_dst_button143_clicked (GtkButton *dst_button143, SPSData *data) {

  (void) dst_button143;  // This statement suppresses compiler warning about unused parameter.
  GtkTextBuffer *textbuffer48;

  textbuffer48 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->dst_textview48));

  // Detach a "first" destination header.
  if (data->dstf_hdr_flag[data->exthdr_type][0]) {
    data->dstf_hdr_flag[data->exthdr_type][0] = 0;
    gtk_text_buffer_set_text (textbuffer48, "Not Attached", -1);

  // Attach a "first" destination header.
  } else {
    // Abort if there are no destination options.
    if (!data->dstf_nopt[data->exthdr_type][0]) {
      data->dstf_hdr_flag[data->exthdr_type][0] = 0;
      gtk_text_buffer_set_text (textbuffer48, "Not Attached", -1);
      sprintf (data->error_text, "In on_button143_clicked: Cannot attach an original/inner  \"first\" destination header since no options were specified.");
      data->parent = data->dst_window;
      report_error (data);
      return (EXIT_FAILURE);
    } else {
      data->dstf_hdr_flag[data->exthdr_type][0] = 1;
      gtk_text_buffer_set_text (textbuffer48, "Attached", -1);
    }
  }

  // Update ethernet frame and dst_window widgets.
  ip6data_update (data->exthdr_type, data);

  return (EXIT_SUCCESS);
}

// Set dst_flag to zero when dst_window is destroyed. (destination.ui)
int
on_dst_window_destroy (GtkWidget *widget, SPSData *data) {

  (void) widget;  // This statement suppresses compiler warning about unused parameter.

  data->dst_flag[0] = 0;

  return (EXIT_SUCCESS);
}

// Toggle: attach / detach a "last" destination header (destination.ui)
int
on_dst_button144_clicked (GtkButton *dst_button144, SPSData *data) {

  (void) dst_button144;  // This statement suppresses compiler warning about unused parameter.
  GtkTextBuffer *textbuffer49;

  textbuffer49 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->dst_textview49));

  // Detach a "last" destination header.
  if (data->dstl_hdr_flag[data->exthdr_type][0]) {
    data->dstl_hdr_flag[data->exthdr_type][0] = 0;
    gtk_text_buffer_set_text (textbuffer49, "Not Attached", -1);

  // Attach a "last" destination header.
  } else {
    // Abort if there are no destination options.
    if (!data->dstl_nopt[data->exthdr_type][0]) {
      data->dstl_hdr_flag[data->exthdr_type][0] = 0;
      gtk_text_buffer_set_text (textbuffer49, "Not Attached", -1);
      sprintf (data->error_text, "In on_button144_clicked: Cannot attach an original/inner \"last\" destination header since no options were specified.");
      data->parent = data->dst_window;
      report_error (data);
      return (EXIT_FAILURE);
    } else {
      data->dstl_hdr_flag[data->exthdr_type][0] = 1;
      gtk_text_buffer_set_text (textbuffer49, "Attached", -1);
    }
  }

  // Update ethernet frame and dst_window widgets.
  ip6data_update (data->exthdr_type, data);

  return (EXIT_SUCCESS);
}

// Destination (original/inner) header length (first) (destination.ui)
int
on_dst_entry439_changed (GtkWidget *dst_entry439, SPSData *data) {

  const char *entry_text;
  GtkStyleContext *context;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (dst_entry439);

  // Get new value from text entry
  entry_text = gtk_entry_get_text (GTK_ENTRY (dst_entry439));

  // Is value too large or small, or are there non-numeric characters?
  if (!is_ascii_uint (entry_text) || (ascii_to_int64 ((char *) entry_text) > 255)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    return (EXIT_FAILURE);

  // Value is in acceptable range.
  } else {

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    data->dstfhdr[data->exthdr_type][0].hdr_len = (uint8_t) ascii_to_int64 ((char *) entry_text);  // Update with new value

    // Update ethernet frame and dst_window widgets.
    ip6data_update (data->exthdr_type, data);
  }

  return (EXIT_SUCCESS);
}

// Destination (original/inner) header length (last) (destination.ui)
int
on_dst_entry440_changed (GtkWidget *dst_entry440, SPSData *data) {

  const char *entry_text;
  GtkStyleContext *context;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (dst_entry440);

  // Get new value from text entry
  entry_text = gtk_entry_get_text (GTK_ENTRY (dst_entry440));

  // Is value too large or small, or are there non-numeric characters?
  if (!is_ascii_uint (entry_text) || (ascii_to_int64 ((char *) entry_text) > 255)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    return (EXIT_FAILURE);

  // Value is in acceptable range.
  } else {

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    data->dstlhdr[data->exthdr_type][0].hdr_len = (uint8_t ) ascii_to_int64 ((char *) entry_text);  // Update with new value

    // Update ethernet frame and dst_window widgets.
    ip6data_update (data->exthdr_type, data);
  }

  return (EXIT_SUCCESS);
}

// Decimal option data entry (destination.ui)
int
on_dst_radiobutton61_toggled (GtkButton *dst_radiobutton61, SPSData *data) {

  (void) dst_radiobutton61;  // This statement suppresses compiler warning about unused parameter.

  if (data->dec_hex_dstopt[0]) {
    data->dec_hex_dstopt[0] = 0;
  } else {
    data->dec_hex_dstopt[0] = 1;
  }

  return (EXIT_SUCCESS);
}

// Hexadecimal option data entry (destination.ui)
int
on_dst_radiobutton62_toggled (GtkButton *dst_radiobutton62) {

  (void) dst_radiobutton62;  // This statement suppresses compiler warning about unused parameter.

  // Let on_dst_radiobutton61_toggle() change the flag.
  // Clicking on either one of radiobutton61 or radiobutton62 causes
  // both callbacks to executed since they are part of same group.

  return (EXIT_SUCCESS);
}

// Destination (original/inner) option data (destination.ui)
int
on_dst_entry315_changed (GtkWidget *dst_entry315, SPSData *data) {

  data_entry (dst_entry315, &data->dst_optlenbuf[data->exthdr_type][0], MAX_DSTOPTLEN, data->dst_optionsbuf[data->exthdr_type][0], data->dec_hex_dstopt[0], data->dst_window, data);

  return (EXIT_SUCCESS);
}

// Destination (original/inner) option alignment parameter x (of (xn + y) (destination.ui)
int
on_dst_entry156_changed (GtkWidget *dst_entry156, SPSData *data) {

  (void) data;  // This statement suppresses compiler warning about unused parameter.

  const char *entry156_text;
  GtkStyleContext *context;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (dst_entry156);

  // Get new value from text entry
  entry156_text = gtk_entry_get_text (GTK_ENTRY (dst_entry156));

  // Is value too large or small, or are there non-numeric characters?
  if (!is_ascii_uint (entry156_text) || (ascii_to_int64 ((char *) entry156_text) > 255)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    return (EXIT_FAILURE);

  // Value is OK.
  } else {

    gtk_style_context_remove_class (context, "error");

    // We use the entry to store this alignment value.
    // dst_option() will obtain it from the user interface.
  }

  return (EXIT_SUCCESS);
}

// Destination (original/inner) option alignment parameter y (of (xn + y) (destination.ui)
int
on_dst_entry157_changed (GtkWidget *dst_entry157, SPSData *data) {

  (void) data;  // This statement suppresses compiler warning about unused parameter.

  const char *entry157_text;
  GtkStyleContext *context;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (dst_entry157);

  // Get new value from text entry
  entry157_text = gtk_entry_get_text (GTK_ENTRY (dst_entry157));

  // Is value too large or small, or are there non-numeric characters?
  if (!is_ascii_uint (entry157_text) || (ascii_to_int64 ((char *) entry157_text) > 255)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    return (EXIT_FAILURE);

  // Value is OK.
  } else {

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    // We use the entry to store this alignment value.
    // dst_option() will obtain it from the user interface.
  }

  return (EXIT_SUCCESS);
}

// Designate as first destination (original/inner) header (destination.ui)
int
on_dst_radiobutton63_toggled (GtkButton *dst_radiobutton63, SPSData *data) {

  (void) dst_radiobutton63;  // This statement suppresses compiler warning about unused parameter.

  if (data->dst_fl_flag[0]) {
    data->dst_fl_flag[0] = 0;
  } else {
    data->dst_fl_flag[0] = 1;
  }

  return (EXIT_SUCCESS);
}

// Designate as last destination (original/inner) header (destination.ui)
int
on_dst_radiobutton64_toggled (GtkButton *dst_radiobutton64) {

  (void) dst_radiobutton64;  // This statement suppresses compiler warning about unused parameter.

  // Let on_dst_radiobutton63_toggle() change the flag.
  // Clicking on either one of radiobutton63 or radiobutton64 causes
  // both callbacks to executed since they are part of same group.

  return (EXIT_SUCCESS);
}

// Insert destination (original/inner) option after another, as specified by user. (destination.ui)
int
on_dst_button104_clicked (GtkButton *dst_button104, SPSData *data) {

  (void) dst_button104;  // This statement suppresses compiler warning about unused parameter.

  if (dst_option (0, 0, data->dstopt_after[0], data) == EXIT_SUCCESS) {

    // Update destination header length in dst_window.
    dst_show (data->exthdr_type, data);

    // Update ethernet frame if destination header is attached.
    if ((data->dstf_hdr_flag[data->exthdr_type][0] && !data->dst_fl_flag[0]) || (data->dstl_hdr_flag[data->exthdr_type][0] && data->dst_fl_flag[0])) {
      ip6data_update (data->exthdr_type, data);
    }
  }

  return (EXIT_SUCCESS);
}

// Number of option after which to insert new destination option. (destination.ui)
int
on_dst_entry431_changed (GtkWidget *dst_entry431, SPSData *data) {

  const char *entry_text;
  GtkStyleContext *context;

  // Get new value from text entry
  entry_text = gtk_entry_get_text (GTK_ENTRY (dst_entry431));

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (dst_entry431);

  // Is value too large or small, or are there non-numeric characters?
  if (!is_ascii_uint (entry_text) || (!ascii_to_int64 ((char *) entry_text))) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    return (EXIT_FAILURE);

  // Value is in acceptable range.
  } else {
    data->dstopt_after[0] = (int) ascii_to_int64 ((char *) entry_text);

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");
  }

  return (EXIT_SUCCESS);
}

// Append an destination (original/inner) option. (destination.ui)
int
on_dst_button105_clicked (GtkButton *dst_button105, SPSData *data) {

  (void) dst_button105;  // This statement suppresses compiler warning about unused parameter.

  if (dst_option (0, 2, 0, data) == EXIT_SUCCESS) {

    // Update destination header length in dst_window.
    dst_show (data->exthdr_type, data);

    // Update ethernet frame if destination header is attached.
    if ((data->dstf_hdr_flag[data->exthdr_type][0] && !data->dst_fl_flag[0]) || (data->dstl_hdr_flag[data->exthdr_type][0] && data->dst_fl_flag[0])) {
      ip6data_update (data->exthdr_type, data);
    }
  }

  return (EXIT_SUCCESS);
}

// View destination (original/inner) header (first) options. (destination.ui)
int
on_dst_button106_clicked (GtkButton *dst_button106, SPSData *data) {

  (void) dst_button106;  // This statement suppresses compiler warning about unused parameter.

  dst_opt_view (0, data);

  return (EXIT_SUCCESS);
}

// View destination header (original/inner) (last) options. (destination.ui)
int
on_dst_button109_clicked (GtkButton *dst_button109, SPSData *data) {

  (void) dst_button109;  // This statement suppresses compiler warning about unused parameter.

  dst_opt_view (1, data);

  return (EXIT_SUCCESS);
}

// Number of destination (original/inner) option to remove. (destination.ui)
int
on_dst_entry432_changed (GtkWidget *dst_entry432, SPSData *data) {

  dsthop_opt_remove_entry (dst_entry432, &data->dstopt_remove[0]);

  return (EXIT_SUCCESS);
}

// Remove selected destination (original/inner) option. (destination.ui)
int
on_dst_button107_clicked (GtkButton *dst_button107, SPSData *data) {

  (void) dst_button107;  // This statement suppresses compiler warning about unused parameter.

  if (dst_option (0, 1, data->dstopt_remove[0], data) == EXIT_SUCCESS) {

    // Update ethernet frame if destination header is attached.
    if ((data->dstf_hdr_flag[data->exthdr_type][0]&& !data->dst_fl_flag[0]) || (data->dstl_hdr_flag[data->exthdr_type][0] && data->dst_fl_flag[0])) {

      // If number of options in now zero, detach header.

      // First header
      if (!data->dst_fl_flag[0] && !data->dstf_nopt[data->exthdr_type][0]) {
        data->dstf_hdr_flag[data->exthdr_type] = 0;

      // Last header
      } else if (data->dst_fl_flag[0] && !data->dstl_nopt[data->exthdr_type][0]) {
        data->dstl_hdr_flag[data->exthdr_type] = 0;
      }

      ip6data_update (data->exthdr_type, data);
    }

    // Update destination header length in dst_window.
    dst_show (data->exthdr_type, data);
  }

  return (EXIT_SUCCESS);
}

// Destination (outer) Header (destination.ui)

// Toggle: attach / detach a "first" destination (outer) header (destination-outer.ui)
int
on_dst_button143_outer_clicked (GtkButton *dst_button143_outer, SPSData *data) {

  (void) dst_button143_outer;  // This statement suppresses compiler warning about unused parameter.
  GtkTextBuffer *textbuffer48_outer;

  textbuffer48_outer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->dst_textview48_outer));

  // Detach a "first" destination header.
  if (data->dstf_hdr_flag[data->exthdr_type][1]) {
    data->dstf_hdr_flag[data->exthdr_type][1] = 0;
    gtk_text_buffer_set_text (textbuffer48_outer, "Not Attached", -1);

  // Attach a "first" destination header.
  } else {
    // Abort if there are no destination options.
    if (!data->dstf_nopt[data->exthdr_type][1]) {
      data->dstf_hdr_flag[data->exthdr_type][1] = 0;
      gtk_text_buffer_set_text (textbuffer48_outer, "Not Attached", -1);
      sprintf (data->error_text, "In on_button143_clicked: Cannot attach an outer \"first\" destination header since no options were specified.");
      data->parent = data->dst_window_outer;
      report_error (data);
      return (EXIT_FAILURE);
    } else {
      data->dstf_hdr_flag[data->exthdr_type][1] = 1;
      gtk_text_buffer_set_text (textbuffer48_outer, "Attached", -1);
    }
  }

  // Update ethernet frame and dst_window widgets.
  ip6data_update (data->exthdr_type, data);

  return (EXIT_SUCCESS);
}

// Set dst_flag to zero when dst_window_outer is destroyed. (destination-outer.ui)
int
on_dst_window_outer_destroy (GtkWidget *widget, SPSData *data) {

  (void) widget;  // This statement suppresses compiler warning about unused parameter.

  data->dst_flag[1] = 0;

  return (EXIT_SUCCESS);
}

// Toggle: attach / detach a "last" destination (outer) header (destination-outer.ui)
int
on_dst_button144_outer_clicked (GtkButton *dst_button144_outer, SPSData *data) {

  (void) dst_button144_outer;  // This statement suppresses compiler warning about unused parameter.
  GtkTextBuffer *textbuffer49_outer;

  textbuffer49_outer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->dst_textview49_outer));

  // Detach a "last" destination header.
  if (data->dstl_hdr_flag[data->exthdr_type][1]) {
    data->dstl_hdr_flag[data->exthdr_type][1] = 0;
    gtk_text_buffer_set_text (textbuffer49_outer, "Not Attached", -1);

  // Attach a "last" destination header.
  } else {
    // Abort if there are no destination options.
    if (!data->dstl_nopt[data->exthdr_type][1]) {
      data->dstl_hdr_flag[data->exthdr_type][1] = 0;
      gtk_text_buffer_set_text (textbuffer49_outer, "Not Attached", -1);
      sprintf (data->error_text, "In on_button144_outer_clicked: Cannot attach an outer \"last\" destination header since no options were specified.");
      data->parent = data->dst_window_outer;
      report_error (data);
      return (EXIT_FAILURE);
    } else {
      data->dstl_hdr_flag[data->exthdr_type][1] = 1;
      gtk_text_buffer_set_text (textbuffer49_outer, "Attached", -1);
    }
  }

  // Update ethernet frame and dst_window widgets.
  ip6data_update (data->exthdr_type, data);

  return (EXIT_SUCCESS);
}

// Destination (outer) header length (first) (destination-outer.ui)
int
on_dst_entry439_outer_changed (GtkWidget *dst_entry439_outer, SPSData *data) {

  const char *entry_text;
  GtkStyleContext *context;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (dst_entry439_outer);

  // Get new value from text entry
  entry_text = gtk_entry_get_text (GTK_ENTRY (dst_entry439_outer));

  // Is value too large or small, or are there non-numeric characters?
  if (!is_ascii_uint (entry_text) || (ascii_to_int64 ((char *) entry_text) > 255)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    return (EXIT_FAILURE);

  // Value is in acceptable range.
  } else {

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    data->dstfhdr[data->exthdr_type][1].hdr_len = (uint8_t) ascii_to_int64 ((char *) entry_text);  // Update with new value

    // Update ethernet frame and dst_window widgets.
    ip6data_update (data->exthdr_type, data);
  }

  return (EXIT_SUCCESS);
}

// Destination (outer) header length (last) (destination-outer.ui)
int
on_dst_entry440_outer_changed (GtkWidget *dst_entry440_outer, SPSData *data) {

  const char *entry_text;
  GtkStyleContext *context;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (dst_entry440_outer);

  // Get new value from text entry
  entry_text = gtk_entry_get_text (GTK_ENTRY (dst_entry440_outer));

  // Is value too large or small, or are there non-numeric characters?
  if (!is_ascii_uint (entry_text) || (ascii_to_int64 ((char *) entry_text) > 255)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    return (EXIT_FAILURE);

  // Value is in acceptable range.
  } else {

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    data->dstlhdr[data->exthdr_type][1].hdr_len = (uint8_t ) ascii_to_int64 ((char *) entry_text);  // Update with new value

    // Update ethernet frame and dst_window widgets.
    ip6data_update (data->exthdr_type, data);
  }

  return (EXIT_SUCCESS);
}

// Decimal option data entry (destination-outer.ui)
int
on_dst_radiobutton61_outer_toggled (GtkButton *dst_radiobutton61_outer, SPSData *data) {

  (void) dst_radiobutton61_outer;  // This statement suppresses compiler warning about unused parameter.

  if (data->dec_hex_dstopt[1]) {
    data->dec_hex_dstopt[1] = 0;
  } else {
    data->dec_hex_dstopt[1] = 1;
  }

  return (EXIT_SUCCESS);
}

// Hexadecimal option data entry (destination-outer.ui)
int
on_dst_radiobutton62_outer_toggled (GtkButton *dst_radiobutton62_outer) {

  (void) dst_radiobutton62_outer;  // This statement suppresses compiler warning about unused parameter.

  // Let on_dst_radiobutton61_outer_toggle() change the flag.
  // Clicking on either one of radiobutton61 or radiobutton62 causes
  // both callbacks to executed since they are part of same group.

  return (EXIT_SUCCESS);
}

// Destination option data (destination-outer.ui)
int
on_dst_entry315_outer_changed (GtkWidget *dst_entry315_outer, SPSData *data) {

  data_entry (dst_entry315_outer, &data->dst_optlenbuf[data->exthdr_type][1], MAX_DSTOPTLEN, data->dst_optionsbuf[data->exthdr_type][1], data->dec_hex_dstopt[1], data->dst_window_outer, data);

  return (EXIT_SUCCESS);
}

// Destination (outer) option alignment parameter x (of (xn + y) (destination-outer.ui)
int
on_dst_entry156_outer_changed (GtkWidget *dst_entry156_outer, SPSData *data) {

  (void) data;  // This statement suppresses compiler warning about unused parameter.

  const char *entry_text;
  GtkStyleContext *context;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (dst_entry156_outer);

  // Get new value from text entry
  entry_text = gtk_entry_get_text (GTK_ENTRY (dst_entry156_outer));

  // Is value too large or small, or are there non-numeric characters?
  if (!is_ascii_uint (entry_text) || (ascii_to_int64 ((char *) entry_text) > 255)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    return (EXIT_FAILURE);

  // Value is OK.
  } else {

    gtk_style_context_remove_class (context, "error");

    // We use the entry to store this alignment value.
    // dst_option() will obtain it from the user interface.
  }

  return (EXIT_SUCCESS);
}

// Destination (outer) option alignment parameter y (of (xn + y) (destination-outer.ui)
int
on_dst_entry157_outer_changed (GtkWidget *dst_entry157_outer, SPSData *data) {

  (void) data;  // This statement suppresses compiler warning about unused parameter.

  const char *entry_text;
  GtkStyleContext *context;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (dst_entry157_outer);

  // Get new value from text entry
  entry_text = gtk_entry_get_text (GTK_ENTRY (dst_entry157_outer));

  // Is value too large or small, or are there non-numeric characters?
  if (!is_ascii_uint (entry_text) || (ascii_to_int64 ((char *) entry_text) > 255)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    return (EXIT_FAILURE);

  // Value is OK.
  } else {

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    // We use the entry to store this alignment value.
    // dst_option() will obtain it from the user interface.
  }

  return (EXIT_SUCCESS);
}

// Designate as first destination (outer) header (destination-outer.ui)
int
on_dst_radiobutton63_outer_toggled (GtkButton *dst_radiobutton63_outer, SPSData *data) {

  (void) dst_radiobutton63_outer;  // This statement suppresses compiler warning about unused parameter.

  if (data->dst_fl_flag[1]) {
    data->dst_fl_flag[1] = 0;
  } else {
    data->dst_fl_flag[1] = 1;
  }

  return (EXIT_SUCCESS);
}

// Designate as last destination (outer) header (destination-outer.ui)
int
on_dst_radiobutton64_outer_toggled (GtkButton *dst_radiobutton64_outer) {

  (void) dst_radiobutton64_outer;  // This statement suppresses compiler warning about unused parameter.

  // Let on_dst_radiobutton63_outer_toggle() change the flag.
  // Clicking on either one of radiobutton63 or radiobutton64 causes
  // both callbacks to executed since they are part of same group.

  return (EXIT_SUCCESS);
}

// Insert destination (outer) option after another, as specified by user. (destination-outer.ui)
int
on_dst_button104_outer_clicked (GtkButton *dst_button104_outer, SPSData *data) {

  (void) dst_button104_outer;  // This statement suppresses compiler warning about unused parameter.

  if (dst_option (1, 0, data->dstopt_after[1], data) == EXIT_SUCCESS) {

    // Update destination header length in dst_window.
    dst_show (data->exthdr_type, data);

    // Update ethernet frame if destination header is attached.
    if ((data->dstf_hdr_flag[data->exthdr_type][1] && !data->dst_fl_flag[1]) || (data->dstl_hdr_flag[data->exthdr_type][1] && data->dst_fl_flag[1])) {
      ip6data_update (data->exthdr_type, data);
    }
  }

  return (EXIT_SUCCESS);
}

// Number of option after which to insert new destination (outer) option. (destination-outer.ui)
int
on_dst_entry431_outer_changed (GtkWidget *dst_entry431_outer, SPSData *data) {

  const char *entry_text;
  GtkStyleContext *context;

  // Get new value from text entry
  entry_text = gtk_entry_get_text (GTK_ENTRY (dst_entry431_outer));

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (dst_entry431_outer);

  // Is value too large or small, or are there non-numeric characters?
  if (!is_ascii_uint (entry_text) || (!ascii_to_int64 ((char *) entry_text))) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    return (EXIT_FAILURE);

  // Value is in acceptable range.
  } else {
    data->dstopt_after[1] = (int) ascii_to_int64 ((char *) entry_text);

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");
  }

  return (EXIT_SUCCESS);
}

// Append an destination (outer) option. (destination-outer.ui)
int
on_dst_button105_outer_clicked (GtkButton *dst_button105_outer, SPSData *data) {

  (void) dst_button105_outer;  // This statement suppresses compiler warning about unused parameter.

  if (dst_option (1, 2, 0, data) == EXIT_SUCCESS) {

    // Update destination header length in dst_window.
    dst_show_outer (data->exthdr_type, data);

    // Update ethernet frame if destination header is attached.
    if ((data->dstf_hdr_flag[data->exthdr_type][1] && !data->dst_fl_flag[1]) || (data->dstl_hdr_flag[data->exthdr_type][1] && data->dst_fl_flag[1])) {
      ip6data_update (data->exthdr_type, data);
    }
  }

  return (EXIT_SUCCESS);
}

// View destination (outer) header (first) options. (destination-outer.ui)
int
on_dst_button106_outer_clicked (GtkButton *dst_button106_outer, SPSData *data) {

  (void) dst_button106_outer;  // This statement suppresses compiler warning about unused parameter.

  dst_opt_view_outer (0, data);

  return (EXIT_SUCCESS);
}

// View destination (outer) header (last) options. (destination-outer.ui)
int
on_dst_button109_outer_clicked (GtkButton *dst_button109_outer, SPSData *data) {

  (void) dst_button109_outer;  // This statement suppresses compiler warning about unused parameter.

  dst_opt_view_outer (1, data);

  return (EXIT_SUCCESS);
}

// Number of destination (outer) option to remove. (destination-outer.ui)
int
on_dst_entry432_outer_changed (GtkWidget *dst_entry432_outer, SPSData *data) {

  dsthop_opt_remove_entry (dst_entry432_outer, &data->dstopt_remove[1]);

  return (EXIT_SUCCESS);
}

// Remove selected destination (outer) option. (destination-outer.ui)
int
on_dst_button107_outer_clicked (GtkButton *dst_button107_outer, SPSData *data) {

  (void) dst_button107_outer;  // This statement suppresses compiler warning about unused parameter.

  if (dst_option (1, 1, data->dstopt_remove[1], data) == EXIT_SUCCESS) {

    // Update ethernet frame if destination header is attached.
    if ((data->dstf_hdr_flag[data->exthdr_type][1] && !data->dst_fl_flag[1]) || (data->dstl_hdr_flag[data->exthdr_type][1] && data->dst_fl_flag[1])) {

      // If number of options in now zero, detach header.

      // First header
      if (!data->dst_fl_flag[1] && !data->dstf_nopt[data->exthdr_type][1]) {
        data->dstf_hdr_flag[data->exthdr_type][1] = 0;

      // Last header
      } else if (data->dst_fl_flag[1] && !data->dstl_nopt[data->exthdr_type][1]) {
        data->dstl_hdr_flag[data->exthdr_type][1] = 0;
      }

      ip6data_update (data->exthdr_type, data);
    }

    // Update destination header length in dst_window.
    dst_show (data->exthdr_type, data);
  }

  return (EXIT_SUCCESS);
}

// Routing (original/inner) header (routing.ui)

// Set route_flag to zero when route_window is destroyed. (routing.ui)
int
on_route_window_destroy (GtkWidget *widget, SPSData *data) {

  (void) widget;  // This statement suppresses compiler warning about unused parameter.

  data->route_flag[0] = 0;

  return (EXIT_SUCCESS);
}

// Toggle: attach / detach an original/inner routing header (routing.ui)
int
on_rt_button145_clicked (GtkButton *rt_button145, SPSData *data) {

  (void) rt_button145;  // This statement suppresses compiler warning about unused parameter.
  GtkTextBuffer *textbuffer50;

  textbuffer50 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->rt_textview50));

  // Detach a routing header.
  if (data->route_hdr_flag[data->exthdr_type][0]) {
    data->route_hdr_flag[data->exthdr_type][0] = 0;
    gtk_text_buffer_set_text (textbuffer50, "Not Attached", -1);

  // Attach a routing header.
  } else {
    // Abort if there is no routing data.
    if (!data->route_datlen[data->exthdr_type][0]) {
      data->route_hdr_flag[data->exthdr_type][0] = 0;
      gtk_text_buffer_set_text (textbuffer50, "Not Attached", -1);
      sprintf (data->error_text, "In on_button145_clicked: Cannot attach an original/inner routing header since no routing data was specified.");
      data->parent = data->route_window;
      report_error (data);
      return (EXIT_FAILURE);
    } else {
      data->route_hdr_flag[data->exthdr_type][0] = 1;
      gtk_text_buffer_set_text (textbuffer50, "Attached", -1);
    }
  }

  // Update ethernet frame and route_window widgets.
  ip6data_update (data->exthdr_type, data);

  return (EXIT_SUCCESS);
}

// Routing (original/inner) header length (8 bits) (routing.ui)
int
on_rt_entry446_changed (GtkWidget *rt_entry446, SPSData *data) {

  const char *entry_text;
  GtkStyleContext *context;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (rt_entry446);

  // Get new value from text entry
  entry_text = gtk_entry_get_text (GTK_ENTRY (rt_entry446));

  // Is value too large or small, or are there non-numeric characters?
  if (!is_ascii_uint (entry_text) || (ascii_to_int64 ((char *) entry_text) > 255)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    return (EXIT_FAILURE);

  // Value is in acceptable range.
  } else {

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    data->routehdr[data->exthdr_type][0].hdr_len = (uint8_t) ascii_to_int64 ((char *) entry_text);  // Update with new value

    // Update ethernet frame and route_window widgets.
    ip6data_update (data->exthdr_type, data);
  }

  return (EXIT_SUCCESS);
}

// Routing (original/inner) type (8 bits) (routing.ui)
int
on_rt_entry447_changed (GtkWidget *rt_entry447, SPSData *data) {

  const char *entry_text;
  GtkStyleContext *context;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (rt_entry447);

  // Get new value from text entry
  entry_text = gtk_entry_get_text (GTK_ENTRY (rt_entry447));

  // Is value too large or small, or are there non-numeric characters?
  if (!is_ascii_uint (entry_text) || (ascii_to_int64 ((char *) entry_text) > 255)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    return (EXIT_FAILURE);

  // Value is in acceptable range.
  } else {

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    data->routehdr[data->exthdr_type][0].routing_type = (uint8_t) ascii_to_int64 ((char *) entry_text);  // Update with new value

    // Update ethernet frame and route_window widgets.
    ip6data_update (data->exthdr_type, data);
  }

  return (EXIT_SUCCESS);
}

// Segments left (8 bits) (routing.ui)
int
on_rt_entry448_changed (GtkWidget *rt_entry448, SPSData *data) {

  const char *entry_text;
  GtkStyleContext *context;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (rt_entry448);

  // Get new value from text entry
  entry_text = gtk_entry_get_text (GTK_ENTRY (rt_entry448));

  // Is value too large or small, or are there non-numeric characters?
  if (!is_ascii_uint (entry_text) || (ascii_to_int64 ((char *) entry_text) > 255)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    return (EXIT_FAILURE);

  // Value is in acceptable range.
  } else {

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    data->routehdr[data->exthdr_type][0].segs_left = (uint8_t) ascii_to_int64 ((char *) entry_text);  // Update with new value

    // Update ethernet frame and route_window widgets.
    ip6data_update (data->exthdr_type, data);
  }

  return (EXIT_SUCCESS);
}

// Decimal routing (original/inner) data entry (routing.ui)
int
on_rt_radiobutton73_toggled (GtkButton *rt_radiobutton73, SPSData *data) {

  (void) rt_radiobutton73;  // This statement suppresses compiler warning about unused parameter.

  if (data->dec_hex_route[0]) {
    data->dec_hex_route[0] = 0;
  } else {
    data->dec_hex_route[0] = 1;
  }

  return (EXIT_SUCCESS);
}

// Hexadecimal routing (original/inner) data entry (routing.ui)
int
on_rt_radiobutton74_toggled (GtkButton *rt_radiobutton74) {

  (void) rt_radiobutton74;  // This statement suppresses compiler warning about unused parameter.

  // on_radiobutton73_toggled() will change value of data->dec_hex_route[0].

  return (EXIT_SUCCESS);
}

// Routing (original/inner) data (variable length) (routing.ui)
int
on_rt_entry449_changed (GtkWidget *rt_entry449, SPSData *data) {

  char *value;

  // Allocate memory for string.
  value = allocate_strmem (TMP_STRINGLEN);

  if (data_entry (rt_entry449, &data->route_datlen[data->exthdr_type][0], MAX_ROUTEDATA, data->route_data[data->exthdr_type][0], data->dec_hex_route[0], data->route_window, data) == EXIT_SUCCESS) {

    // Update routing header length.
    update_route_len (data->exthdr_type, 0, data);

    // If routing header is attached, update IPv6 parameters, routing header editing window, and ethernet frames.
    data->dec_hex_route[0] = 0;  // Set to decimal entry since contents will be displayed as decimal by route_show().
    if (data->route_hdr_flag[data->exthdr_type][0]) {
      ip6data_update (data->exthdr_type, data);
    }

    // Update header length in routing header editing window.
    sprintf (value, "%" PRIu8, (uint8_t) data->routehdr[data->exthdr_type][0].hdr_len);
    gtk_entry_set_text (GTK_ENTRY (data->rt_entry446), value);
  }

  // Free allocated memory.
  free (value);

  return (EXIT_SUCCESS);
}

// Load routing (original/inner) data from file (variable length) (routing.ui)
int
on_rt_button125_clicked (GtkButton *rt_button125, SPSData *data) {

  char *value;

  // Allocate memory for string.
  value = allocate_strmem (TMP_STRINGLEN);

  (void) rt_button125;  // This statement suppresses compiler warning about unused parameter.

  if (file_data (data->rt_entry449, &data->route_datlen[data->exthdr_type][0], MAX_ROUTEDATA, data->route_data[data->exthdr_type][0], data->exthdr_type, data->route_window, data) == EXIT_SUCCESS) {

    // Update routing header length.
    update_route_len (data->exthdr_type, 0, data);

    // If routing header is attached, update IPv6 parameters, routing header editing window, and ethernet frames.
    data->dec_hex_route[0] = 0;  // Set to decimal entry since contents will be displayed as decimal by route_show().
    if (data->route_hdr_flag[data->exthdr_type][0]) {
      ip6data_update (data->exthdr_type, data);
    }

    // Update header length in routing header editing window.
    sprintf (value, "%" PRIu8, (uint8_t) data->routehdr[data->exthdr_type][0].hdr_len);
    gtk_entry_set_text (GTK_ENTRY (data->rt_entry446), value);
  }

  // Free allocated memory.
  free (value);

  return (EXIT_SUCCESS);
}

// Routing (outer) header (routing-outer.ui)

// Set route_flag to zero when route_window is destroyed. (routing-outer.ui)
int
on_route_window_outer_destroy (GtkWidget *widget, SPSData *data) {

  (void) widget;  // This statement suppresses compiler warning about unused parameter.

  data->route_flag[1] = 0;

  return (EXIT_SUCCESS);
}

// Toggle: attach / detach an outer routing header (routing-outer.ui)
int
on_rt_button145_outer_clicked (GtkButton *rt_button145_outer, SPSData *data) {

  (void) rt_button145_outer;  // This statement suppresses compiler warning about unused parameter.
  GtkTextBuffer *textbuffer50_outer;

  textbuffer50_outer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->rt_textview50_outer));

  // Detach a routing header.
  if (data->route_hdr_flag[data->exthdr_type][1]) {
    data->route_hdr_flag[data->exthdr_type][1] = 0;
    gtk_text_buffer_set_text (textbuffer50_outer, "Not Attached", -1);

  // Attach a routing header.
  } else {
    // Abort if there is no routing data.
    if (!data->route_datlen[data->exthdr_type][1]) {
      data->route_hdr_flag[data->exthdr_type][1] = 0;
      gtk_text_buffer_set_text (textbuffer50_outer, "Not Attached", -1);
      sprintf (data->error_text, "In on_button145_outer_clicked: Cannot attach an outer routing header since no routing data was specified.");
      data->parent = data->route_window_outer;
      report_error (data);
      return (EXIT_FAILURE);
    } else {
      data->route_hdr_flag[data->exthdr_type][1] = 1;
      gtk_text_buffer_set_text (textbuffer50_outer, "Attached", -1);
    }
  }

  // Update ethernet frame and route_window widgets.
  ip6data_update (data->exthdr_type, data);

  return (EXIT_SUCCESS);
}

// Routing (outer) header length (8 bits) (routing-outer.ui)
int
on_rt_entry446_outer_changed (GtkWidget *rt_entry446_outer, SPSData *data) {

  const char *entry_text;
  GtkStyleContext *context;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (rt_entry446_outer);

  // Get new value from text entry
  entry_text = gtk_entry_get_text (GTK_ENTRY (rt_entry446_outer));

  // Is value too large or small, or are there non-numeric characters?
  if (!is_ascii_uint (entry_text) || (ascii_to_int64 ((char *) entry_text) > 255)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    return (EXIT_FAILURE);

  // Value is in acceptable range.
  } else {

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    data->routehdr[data->exthdr_type][1].hdr_len = (uint8_t) ascii_to_int64 ((char *) entry_text);  // Update with new value

    // Update ethernet frame and route_window widgets.
    ip6data_update (data->exthdr_type, data);
  }

  return (EXIT_SUCCESS);
}

// Routing (outer) type (8 bits) (routing-outer.ui)
int
on_rt_entry447_outer_changed (GtkWidget *rt_entry447_outer, SPSData *data) {

  const char *entry_text;
  GtkStyleContext *context;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (rt_entry447_outer);

  // Get new value from text entry
  entry_text = gtk_entry_get_text (GTK_ENTRY (rt_entry447_outer));

  // Is value too large or small, or are there non-numeric characters?
  if (!is_ascii_uint (entry_text) || (ascii_to_int64 ((char *) entry_text) > 255)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    return (EXIT_FAILURE);

  // Value is in acceptable range.
  } else {

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    data->routehdr[data->exthdr_type][1].routing_type = (uint8_t) ascii_to_int64 ((char *) entry_text);  // Update with new value

    // Update ethernet frame and route_window widgets.
    ip6data_update (data->exthdr_type, data);
  }

  return (EXIT_SUCCESS);
}

// Segments left (8 bits) (routing-outer.ui)
int
on_rt_entry448_outer_changed (GtkWidget *rt_entry448_outer, SPSData *data) {

  const char *entry_text;
  GtkStyleContext *context;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (rt_entry448_outer);

  // Get new value from text entry
  entry_text = gtk_entry_get_text (GTK_ENTRY (rt_entry448_outer));

  // Is value too large or small, or are there non-numeric characters?
  if (!is_ascii_uint (entry_text) || (ascii_to_int64 ((char *) entry_text) > 255)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    return (EXIT_FAILURE);

  // Value is in acceptable range.
  } else {

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    data->routehdr[data->exthdr_type][1].segs_left = (uint8_t) ascii_to_int64 ((char *) entry_text);  // Update with new value

    // Update ethernet frame and route_window widgets.
    ip6data_update (data->exthdr_type, data);
  }

  return (EXIT_SUCCESS);
}

// Decimal routing data entry (routing-outer.ui)
int
on_rt_radiobutton73_outer_toggled (GtkButton *rt_radiobutton73_outer, SPSData *data) {

  (void) rt_radiobutton73_outer;  // This statement suppresses compiler warning about unused parameter.

  if (data->dec_hex_route[1]) {
    data->dec_hex_route[1] = 0;
  } else {
    data->dec_hex_route[1] = 1;
  }

  return (EXIT_SUCCESS);
}

// Hexadecimal routing (outer) data entry (routing-outer.ui)
int
on_rt_radiobutton74_outer_toggled (GtkButton *rt_radiobutton74_outer) {

  (void) rt_radiobutton74_outer;  // This statement suppresses compiler warning about unused parameter.

  // on_radiobutton73_outer_toggled() will change value of data->dec_hex_route[1].

  return (EXIT_SUCCESS);
}

// Routing (outer) data (variable length) (routing-outer.ui)
int
on_rt_entry449_outer_changed (GtkWidget *rt_entry449_outer, SPSData *data) {

  char *value;

  // Allocate memory for string.
  value = allocate_strmem (TMP_STRINGLEN);

  if (data_entry (rt_entry449_outer, &data->route_datlen[data->exthdr_type][1], MAX_ROUTEDATA, data->route_data[data->exthdr_type][1], data->dec_hex_route[1], data->route_window_outer, data) == EXIT_SUCCESS) {

    // Update routing header length.
    update_route_len (data->exthdr_type, 1, data);

    // If routing header is attached, update IPv6 parameters, routing header editing window, and ethernet frames.
    data->dec_hex_route[1] = 0;  // Set to decimal entry since contents will be displayed as decimal by route_show().
    if (data->route_hdr_flag[data->exthdr_type][1]) {
      ip6data_update (data->exthdr_type, data);
    }

    // Update header length in routing header editing window.
    sprintf (value, "%" PRIu8, (uint8_t) data->routehdr[data->exthdr_type][1].hdr_len);
    gtk_entry_set_text (GTK_ENTRY (data->rt_entry446_outer), value);
  }

  // Free allocated memory.
  free (value);

  return (EXIT_SUCCESS);
}

// Load routing (outer) data from file (variable length) (routing-outer.ui)
int
on_rt_button125_outer_clicked (GtkButton *rt_button125_outer, SPSData *data) {

  char *value;

  // Allocate memory for string.
  value = allocate_strmem (TMP_STRINGLEN);

  (void) rt_button125_outer;  // This statement suppresses compiler warning about unused parameter.

  if (file_data (data->rt_entry449_outer, &data->route_datlen[data->exthdr_type][1], MAX_ROUTEDATA, data->route_data[data->exthdr_type][1], data->exthdr_type, data->route_window_outer, data) == EXIT_SUCCESS) {

    // Update routing header length.
    update_route_len (data->exthdr_type, 1, data);

    // If routing header is attached, update IPv6 parameters, routing header editing window, and ethernet frames.
    data->dec_hex_route[1] = 0;  // Set to decimal entry since contents will be displayed as decimal by route_show().
    if (data->route_hdr_flag[data->exthdr_type][1]) {
      ip6data_update (data->exthdr_type, data);
    }

    // Update header length in routing header editing window.
    sprintf (value, "%" PRIu8, (uint8_t) data->routehdr[data->exthdr_type][1].hdr_len);
    gtk_entry_set_text (GTK_ENTRY (data->rt_entry446_outer), value);
  }

  // Free allocated memory.
  free (value);

  return (EXIT_SUCCESS);
}

// Authentication header (authentication.ui)

// Toggle: attach / detach an authentication header (authentication.ui)
int
on_auth_button138_clicked (GtkButton *auth_button138, SPSData *data) {

  (void) auth_button138;  // This statement suppresses compiler warning about unused parameter.
  GtkTextBuffer *textbuffer43, *textbuffer45;

  // *****************************************************************
  // NOTE: Here we choose not to allow both AH and ESP on same packet.
  //       ESP alone is generally considered sufficient since it
  //       provides both authentication and encryption.
  // *****************************************************************

  textbuffer43 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->auth_textview43));

  if (data->esp_flag) {  // If ESP editing window is open, update status textview.
    textbuffer45 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->esp_textview45));
    gtk_text_buffer_set_text (textbuffer45, "Not Attached", -1);
  }
  data->esp_hdr_flag[data->exthdr_type] = 0;  // Detach ESP header from packet.

  // Detach an authentication header.
  if (data->auth_hdr_flag[data->exthdr_type]) {
    data->auth_hdr_flag[data->exthdr_type] = 0;
    gtk_text_buffer_set_text (textbuffer43, "Not Attached", -1);

  // Attach an authentication header.
  } else {
    // Abort if there is no authentication data.
    if (!data->auth_len[data->exthdr_type]) {
      data->auth_hdr_flag[data->exthdr_type] = 0;
      gtk_text_buffer_set_text (textbuffer43, "Not Attached", -1);
      sprintf (data->error_text, "In on_button138_clicked: Cannot attach an authentication header since none was specified.");
      data->parent = data->auth_window;
      report_error (data);
      return (EXIT_FAILURE);
    } else {
      data->auth_hdr_flag[data->exthdr_type] = 1;
      gtk_text_buffer_set_text (textbuffer43, "Attached", -1);
    }
  }

  // Update ethernet frame and auth_window widgets.
  ip6data_update (data->exthdr_type, data);

  return (EXIT_SUCCESS);
}

// Set auth_flag to zero when auth_window is destroyed. (authentication.ui)
int
on_auth_window_destroy (GtkWidget *widget, SPSData *data) {

  (void) widget;  // This statement suppresses compiler warning about unused parameter.

  data->auth_flag = 0;

  return (EXIT_SUCCESS);
}

// Toggle: transport / tunnel mode (authentication.ui)
int
on_auth_button139_clicked (GtkButton *auth_button139, SPSData *data) {

  (void) auth_button139;  // This statement suppresses compiler warning about unused parameter.
  GtkTextBuffer *textbuffer44;

  textbuffer44 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->auth_textview44));
  if (data->auth_tr_tun_flag[data->exthdr_type]) {
    data->auth_tr_tun_flag[data->exthdr_type] = 0;
    gtk_text_buffer_set_text (textbuffer44, "Transport mode", -1);
  } else {
    data->auth_tr_tun_flag[data->exthdr_type] = 1;
    gtk_text_buffer_set_text (textbuffer44, "Tunnel mode", -1);
  }

  // If an authentication header is attached, update ethernet frames.
  if (data->auth_hdr_flag[data->exthdr_type]) {
    ip6data_update (data->exthdr_type, data);
  }

  return (EXIT_SUCCESS);
}

// Payload length (8 bits) (authentication.ui)
int
on_auth_entry433_changed (GtkWidget *auth_entry433, SPSData *data) {

  const char *entry_text;
  GtkStyleContext *context;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (auth_entry433);

  // Get new value from text entry
  entry_text = gtk_entry_get_text (GTK_ENTRY (auth_entry433));

  // Is value too large or small, or are there non-numeric characters?
  if (!is_ascii_uint (entry_text) || (ascii_to_int64 ((char *) entry_text) > 255)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    return (EXIT_FAILURE);

  // Value is in acceptable range.
  } else {

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    data->authhdr[data->exthdr_type].pay_len = (uint8_t) ascii_to_int64 ((char *) entry_text);  // Update with new value

    // Update ethernet frame and auth_window widgets.
    ip6data_update (data->exthdr_type, data);
  }

  return (EXIT_SUCCESS);
}

// Reserved (16 bits) (authentication.ui)
int
on_auth_entry434_changed (GtkWidget *auth_entry434, SPSData *data) {

  const char *entry_text;
  GtkStyleContext *context;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (auth_entry434);

  // Get new value from text entry
  entry_text = gtk_entry_get_text (GTK_ENTRY (auth_entry434));

  // Is value too large or small, or are there non-numeric characters?
  if (!is_ascii_uint (entry_text) || (ascii_to_int64 ((char *) entry_text) > 65535)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    return (EXIT_FAILURE);

  // Value is in acceptable range.
  } else {

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    data->authhdr[data->exthdr_type].reserved = htons ((uint16_t) ascii_to_int64 ((char *) entry_text));  // Update with new value

    // Update ethernet frame and auth_window widgets.
    ip6data_update (data->exthdr_type, data);
  }

  return (EXIT_SUCCESS);
}

// Security parameters index (SPI) (32 bits) (authentication.ui)
int
on_auth_entry435_changed (GtkWidget *auth_entry435, SPSData *data) {

  const char *entry_text;
  GtkStyleContext *context;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (auth_entry435);

  // Get new value from text entry
  entry_text = gtk_entry_get_text (GTK_ENTRY (auth_entry435));

  // Is value too large or small, or are there non-numeric characters?
  if (!is_ascii_uint (entry_text) || (ascii_to_int64 ((char *) entry_text) >= 0xffffffffl)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    return (EXIT_FAILURE);

  // Value is in acceptable range.
  } else {

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    data->authhdr[data->exthdr_type].spi = htonl ((uint32_t) ascii_to_int64 ((char *) entry_text));  // Update with new value

    // Update ethernet frame and auth_window widgets.
    ip6data_update (data->exthdr_type, data);
  }

  return (EXIT_SUCCESS);
}

// Sequence number (32 bits) (authentication.ui)
int
on_auth_entry436_changed (GtkWidget *auth_entry436, SPSData *data) {

  const char *entry_text;
  GtkStyleContext *context;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (auth_entry436);

  // Get new value from text entry
  entry_text = gtk_entry_get_text (GTK_ENTRY (auth_entry436));

  // Is value too large or small, or are there non-numeric characters?
  if (!is_ascii_uint (entry_text) || (ascii_to_int64 ((char *) entry_text) >= 0xffffffffl)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    return (EXIT_FAILURE);

  // Value is in acceptable range.
  } else {

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    data->authhdr[data->exthdr_type].seq = htonl ((uint32_t) ascii_to_int64 ((char *) entry_text));  // Update with new value

    // Update ethernet frame and auth_window widgets.
    ip6data_update (data->exthdr_type, data);
  }

  return (EXIT_SUCCESS);
}

// Decimal integrity check value (ICV) entry (authentication.ui)
int
on_auth_radiobutton67_toggled (GtkButton *auth_radiobutton67, SPSData *data) {

  (void) auth_radiobutton67;  // This statement suppresses compiler warning about unused parameter.

  if (data->dec_hex_auth) {
    data->dec_hex_auth = 0;
  } else {
    data->dec_hex_auth = 1;
  }

  return (EXIT_SUCCESS);
}

// Hexadecimal integrity check value (ICV) entry (authentication.ui)
int
on_auth_radiobutton68_toggled (GtkButton *auth_radiobutton68) {

  (void) auth_radiobutton68;  // This statement suppresses compiler warning about unused parameter.

  // on_radiobutton67_toggled() will change value of data->dec_hex_auth.

  return (EXIT_SUCCESS);
}

// Integrity check value (ICV) (variable length) (authentication.ui)
int
on_auth_entry437_changed (GtkWidget *auth_entry437, SPSData *data) {

  char *value;

  // Allocate memory for string.
  value = allocate_strmem (TMP_STRINGLEN);

  data_entry (auth_entry437, &data->auth_len[data->exthdr_type], MAX_AUTHICVLEN, data->auth_data[data->exthdr_type], data->dec_hex_auth, data->auth_window, data);

  // Update authentication header payload and padding length.
  update_auth_pad (data->exthdr_type, data);

  // If authentication header is attached, update IPv6 parameters and ethernet frames.
  if (data->auth_hdr_flag[data->exthdr_type]) {
    ip6data_update (data->exthdr_type, data);
  }

  // Update payload length in authentication header editing window.
  sprintf (value, "%" PRIu8, (uint8_t) data->authhdr[data->exthdr_type].pay_len);
  gtk_entry_set_text (GTK_ENTRY (data->auth_entry433), value);

  // Free allocated memory.
  free (value);

  return (EXIT_SUCCESS);
}

// Load ICV from file (authentication.ui)
int
on_auth_button108_clicked (GtkButton *auth_button108, SPSData *data) {

  char *value;

  // Allocate memory for string.
  value = allocate_strmem (TMP_STRINGLEN);

  (void) auth_button108;  // This statement suppresses compiler warning about unused parameter.

  file_data (data->auth_entry437, &data->auth_len[data->exthdr_type], MAX_AUTHICVLEN, data->auth_data[data->exthdr_type], data->exthdr_type, data->auth_window, data);

  // Update authentication header payload and padding length.
  update_auth_pad (data->exthdr_type, data);

  // If authentication header is attached, update IPv6 parameters and ethernet frames.
  if (data->auth_hdr_flag[data->exthdr_type]) {
    ip6data_update (data->exthdr_type, data);
  }

  // Update payload length in authentication header editing window.
  sprintf (value, "%" PRIu8, (uint8_t) data->authhdr[data->exthdr_type].pay_len);
  gtk_entry_set_text (GTK_ENTRY (data->auth_entry433), value);

  // Free allocated memory.
  free (value);

  return (EXIT_SUCCESS);
}

// Save data for ICV calculation (authentication.ui)
int
on_auth_button122_clicked (GtkButton *auth_button122, SPSData *data) {

  (void) auth_button122;  // This statement suppresses compiler warning about unused parameter.

  int i, j, result;
  struct ip6_hdr ip6hdr, newip6hdr;
  GtkWidget *dialog;
  GtkFileChooserAction action = GTK_FILE_CHOOSER_ACTION_SAVE;
  char *filename, *text;
  uint8_t *buffer;
  FILE *fo;

  // Array for error messages.
  text = allocate_strmem (TEXT_STRINGLEN);

  // Create filechooser dialog.
  dialog = gtk_file_chooser_dialog_new ("Save File...", (GtkWindow *) data->main_window, action,
             "_Cancel", GTK_RESPONSE_CANCEL, "_Save", GTK_RESPONSE_ACCEPT, NULL);
  gtk_file_chooser_set_do_overwrite_confirmation (GTK_FILE_CHOOSER (dialog), TRUE);

  // Run the filechooser dialog.
  result = gtk_dialog_run (GTK_DIALOG (dialog));

  // Interpret result returned from filechooser dialog.
  if (result == GTK_RESPONSE_ACCEPT)
  {
    filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
    gtk_widget_destroy (dialog);
  } else {
    free (text);
    gtk_widget_destroy (dialog);
    return (EXIT_SUCCESS);
  }

  // Open output file.
  if ((fo = fopen (filename, "wb")) == NULL) {
    sprintf (data->error_text, "In callbacks.c: fopen() failed in on_button122_clicked().");
    data->parent = data->main_window;
    report_error (data);
    free (text);
    return (EXIT_FAILURE);
  }

  // Save entire packet with mutable fields and authentication data field (ICV) set to zero.
  // Section 3.1 of RFC 4302, 4305, and Section 5.1.2.2 of RFC 4301.

  // Copy IPv6 headers so we don't corrupt originals.
  memcpy (&ip6hdr, &data->ip6hdr[data->exthdr_type], IP6_HDRLEN * sizeof (uint8_t));
  memcpy (&newip6hdr, &data->ip6hdr[data->exthdr_type+20], IP6_HDRLEN * sizeof (uint8_t));

  // If in tunnel mode, include "new" IPv6 header and place AH header immediately after it.
  if (data->auth_tr_tun_flag[data->exthdr_type]) {

    // "New" IPv6 header
    newip6hdr.ip6_flow = htonl ((ntohl (newip6hdr.ip6_flow) >> 28) << 28);  // Set mutable fields to zero.
    buffer = (uint8_t *) &newip6hdr;
    for (i=0; i<IP6_HDRLEN; i++) {
      fputc (buffer[i], fo);
    }

    // Hop-by-hop header (outer)
    if (data->hbh_hdr_flag[data->exthdr_type][1]) {

      // Hop-by-hop header (outer), excluding options and padding
      buffer = (uint8_t *) &data->hophdr[data->exthdr_type][1];
      for (i=0; i<HOP_HDRLEN; i++) {
        fputc (buffer[i], fo);
      }

      // Hop-by-hop options
      for (i=0; i<data->hbh_nopt[data->exthdr_type][1]; i++) {

        // Alignment padding
        for (j=0; j<data->hbh_optleadpadlen[data->exthdr_type][1][i]; j++) {
          fputc (data->hbh_optleadpad[data->exthdr_type][1][i][j], fo);
        }

        // Option
        for (j=0; j<data->hbh_optlen[data->exthdr_type][1][i]; j++) {
          fputc (data->hbh_options[data->exthdr_type][1][i][j], fo);
        }

        // Trailing padding
        for (j=0; j<data->hbh_opttailpadlen[data->exthdr_type][1]; j++) {
          fputc (data->hbh_opttailpad[data->exthdr_type][1][j], fo);
        }
      }
    }  // End if hop-by-hop header

    // Destination (outer) header (first)
    if (data->dstf_hdr_flag[data->exthdr_type][1]) {

      // Destination header (first), excluding options and padding
      buffer = (uint8_t *) &data->dstfhdr[data->exthdr_type][1];
      for (i=0; i<DST_HDRLEN; i++) {
        fputc (buffer[i], fo);
      }

      // Destination header (first) options
      for (i=0; i<data->dstf_nopt[data->exthdr_type][1]; i++) {

        // Alignment padding
        for (j=0; j<data->dstf_optleadpadlen[data->exthdr_type][1][i]; j++) {
          fputc (data->dstf_optleadpad[data->exthdr_type][1][i][j], fo);
        }

        // Destination option
        for (j=0; j<data->dstf_optlen[data->exthdr_type][1][i]; j++) {
          fputc (data->dstf_options[data->exthdr_type][1][i][j], fo);
        }
      }

      // Trailing padding
      for (j=0; j<data->dstf_opttailpadlen[data->exthdr_type][1]; j++) {
        fputc (data->dstf_opttailpad[data->exthdr_type][1][j], fo);
      }
    }  // End if destination header (first)

    // Routing (outer) header
    if (data->route_hdr_flag[data->exthdr_type][1]) {

      // Routing header, excluding routing data
      buffer = (uint8_t *) &data->routehdr[data->exthdr_type][1];
      for (i=0; i<RTE_HDRLEN; i++) {
        fputc (buffer[i], fo);
      }

      // Routing header data
      for (i=0; i<data->route_datlen[data->exthdr_type][1]; i++) {
        fputc (data->route_data[data->exthdr_type][1][i], fo);
      }
    }  // End if routing header

    // Authentication header, excluding authentication data
    buffer = (uint8_t *) &data->authhdr[data->exthdr_type];
    for (i=0; i<ATH_HDRLEN; i++) {
      fputc (buffer[i], fo);
    }

    // Authentication header data: set to zero for ICV calculation
    for (i=0; i<data->auth_len[data->exthdr_type]; i++) {
      fputc (0, fo);
    }

  }  // End if tunnel mode

  // Original IPv6 header

  // Set mutable fields to zero.
  ip6hdr.ip6_flow = htonl ((ntohl (ip6hdr.ip6_flow) >> 28) << 28);  // Set mutable fields to zero.
  ip6hdr.ip6_hops = 0u;

  // Copy the IPv6 header.
  buffer = (uint8_t *) &ip6hdr;
  for (i=0; i<IP6_HDRLEN; i++) {
    fputc (buffer[i], fo);
  }

  // Hop-by-hop header (original/inner)
  if (data->hbh_hdr_flag[data->exthdr_type][0]) {

    // Hop-by-hop header, excluding options and padding
    buffer = (uint8_t *) &data->hophdr[data->exthdr_type][0];
    for (i=0; i<HOP_HDRLEN; i++) {
      fputc (buffer[i], fo);
    }

    // Hop-by-hop options
    for (i=0; i<data->hbh_nopt[data->exthdr_type][0]; i++) {

      // Alignment padding
      for (j=0; j<data->hbh_optleadpadlen[data->exthdr_type][0][i]; j++) {
        fputc (data->hbh_optleadpad[data->exthdr_type][0][i][j], fo);
      }

      // Option
      for (j=0; j<data->hbh_optlen[data->exthdr_type][0][i]; j++) {
        fputc (data->hbh_options[data->exthdr_type][0][i][j], fo);
      }

      // Trailing padding
      for (j=0; j<data->hbh_opttailpadlen[data->exthdr_type][0]; j++) {
        fputc (data->hbh_opttailpad[data->exthdr_type][0][j], fo);
      }
    }
  }  // End if hop-by-hop header

  // Destination (original/inner) header (first)
  if (data->dstf_hdr_flag[data->exthdr_type][0]) {

    // Destination header (first), excluding options and padding
    buffer = (uint8_t *) &data->dstfhdr[data->exthdr_type][0];
    for (i=0; i<DST_HDRLEN; i++) {
      fputc (buffer[i], fo);
    }

    // Destination header (first) options
    for (i=0; i<data->dstf_nopt[data->exthdr_type][0]; i++) {

      // Alignment padding
      for (j=0; j<data->dstf_optleadpadlen[data->exthdr_type][0][i]; j++) {
        fputc (data->dstf_optleadpad[data->exthdr_type][0][i][j], fo);
      }

      // Destination option
      for (j=0; j<data->dstf_optlen[data->exthdr_type][0][i]; j++) {
        fputc (data->dstf_options[data->exthdr_type][0][i][j], fo);
      }
    }

    // Trailing padding
    for (j=0; j<data->dstf_opttailpadlen[data->exthdr_type][0]; j++) {
      fputc (data->dstf_opttailpad[data->exthdr_type][0][j], fo);
    }
  }  // End if destination header (first)

  // Routing (original/inner) header
  if (data->route_hdr_flag[data->exthdr_type][0]) {

    // Routing header, excluding routing data
    buffer = (uint8_t *) &data->routehdr[data->exthdr_type][0];
    for (i=0; i<RTE_HDRLEN; i++) {
      fputc (buffer[i], fo);
    }

    // Routing header data
    for (i=0; i<data->route_datlen[data->exthdr_type][0]; i++) {
      fputc (data->route_data[data->exthdr_type][0][i], fo);
    }
  }  // End if routing header

  // If not in tunnel mode, place AH header here.
  if (!data->auth_tr_tun_flag[data->exthdr_type]) {

    // Authentication header, excluding authentication data
    buffer = (uint8_t *) &data->authhdr[data->exthdr_type];
    for (i=0; i<ATH_HDRLEN; i++) {
      fputc (buffer[i], fo);
    }

    // Authentication header data: set to zero for ICV calculation
    for (i=0; i<data->auth_len[data->exthdr_type]; i++) {
      fputc (0, fo);
    }

  }  // End if tunnel mode

  // Destination (original/inner) header (last)
  if (data->dstl_hdr_flag[data->exthdr_type][0]) {

    // Destination header (last), excluding options and padding
    buffer = (uint8_t *) &data->dstlhdr[data->exthdr_type][0];
    for (i=0; i<DST_HDRLEN; i++) {
      fputc (buffer[i], fo);
    }

    // Destination header (last) options
    for (i=0; i<data->dstl_nopt[data->exthdr_type][0]; i++) {

      // Alignment padding
      for (j=0; j<data->dstl_optleadpadlen[data->exthdr_type][0][i]; j++) {
        fputc (data->dstl_optleadpad[data->exthdr_type][0][i][j], fo);
      }

      // Destination option
      for (j=0; j<data->dstl_optlen[data->exthdr_type][i][0]; j++) {
        fputc (data->dstl_options[data->exthdr_type][0][i][j], fo);
      }
    }

    // Trailing padding
    for (j=0; j<data->dstl_opttailpadlen[data->exthdr_type][0]; j++) {
      fputc (data->dstl_opttailpad[data->exthdr_type][0][j], fo);
    }
  }  // End if destination header (last)

  // TCP header, options, and padding
  if ((data->exthdr_type == 3) || (data->exthdr_type == 9)) {

    // TCP header
    buffer = (uint8_t *) &data->tcphdr[data->exthdr_type];
    for (i=0; i<TCP_HDRLEN; i++) {
      fputc (buffer[i], fo);
    }

    // TCP options
    for (i=0; i<data->tcp_nopt[data->exthdr_type]; i++) {
      for (j=0; j<data->tcp_optlen[data->exthdr_type][i]; j++) {
        fputc (data->tcp_options[data->exthdr_type][i][j], fo);
      }
    }

    // TCP options padding
    for (i=0; i<data->tcp_optpadlen[data->exthdr_type]; i++) {
      fputc (0, fo);
    }

  // ICMP header
  } else if ((data->exthdr_type == 4) || (data->exthdr_type == 10)) {
    buffer = (uint8_t *) &data->icmp6hdr[data->exthdr_type];
    for (i=0; i<ICMP_HDRLEN; i++) {
      fputc (buffer[i], fo);
    }


  // UDP header
  } else if ((data->exthdr_type == 5) || (data->exthdr_type == 1)) {
    buffer = (uint8_t *) &data->udphdr[data->exthdr_type];
    for (i=0; i<UDP_HDRLEN; i++) {
      fputc (buffer[i], fo);
    }
  }

  // Upper layer protocol payload data
  for (i=0; i<data->payloadlen[data->exthdr_type]; i++) {
    fputc (data->payload[data->exthdr_type][i], fo);
  }

  // Close file descriptor.
  fclose (fo);

  // Free allocated memory.
  free (text);

  return (EXIT_SUCCESS);
}

// Encapsulating security payload (ESP) header (esp.ui)

// Toggle: attach / detach an ESP header (esp.ui)
int
on_esp_button140_clicked (GtkButton *esp_button140, SPSData *data) {

  (void) esp_button140;  // This statement suppresses compiler warning about unused parameter.
  GtkTextBuffer *textbuffer45, *textbuffer43;

  // *****************************************************************
  // NOTE: Here we choose not to allow both AH and ESP on same packet.
  //       ESP alone is generally considered sufficient since it
  //       provides both authentication and encryption.
  // *****************************************************************

  textbuffer45 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->esp_textview45));

  if (data->auth_flag) {  // If authentication editing window is open, update status textview.
    textbuffer43 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->auth_textview43));
    gtk_text_buffer_set_text (textbuffer43, "Not Attached", -1);
  }
  data->auth_hdr_flag[data->exthdr_type] = 0;  // Detach authentication header from packet.

  // Detach an ESP header.
  if (data->esp_hdr_flag[data->exthdr_type]) {
    data->esp_hdr_flag[data->exthdr_type] = 0;
    gtk_text_buffer_set_text (textbuffer45, "Not Attached", -1);

  // Attach an ESP header.
  } else {
    // Abort if there is no ESP data.
    if (!data->esp_auth_len[data->exthdr_type]) {
      data->esp_hdr_flag[data->exthdr_type] = 0;
      gtk_text_buffer_set_text (textbuffer45, "Not Attached", -1);
      sprintf (data->error_text, "In on_button140_clicked: Cannot attach an ESP header since none was specified.");
      data->parent = data->esp_window;
      report_error (data);
      return (EXIT_FAILURE);
    } else {
      data->esp_hdr_flag[data->exthdr_type] = 1;
      gtk_text_buffer_set_text (textbuffer45, "Attached", -1);
    }
  }

  // Update ethernet frame and esp_window widgets.
  ip6data_update (data->exthdr_type, data);

  return (EXIT_SUCCESS);
}

// Set esp_flag to zero when esp_window is destroyed. (esp.ui)
int
on_esp_window_destroy (GtkWidget *widget, SPSData *data) {

  (void) widget;  // This statement suppresses compiler warning about unused parameter.

  data->esp_flag = 0;

  return (EXIT_SUCCESS);
}

// Toggle: transport / tunnel mode (esp.ui)
int
on_esp_button141_clicked (GtkButton *esp_button141, SPSData *data) {

  (void) esp_button141;  // This statement suppresses compiler warning about unused parameter.
  GtkTextBuffer *textbuffer46;

  textbuffer46 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->esp_textview46));
  if (data->esp_tr_tun_flag[data->exthdr_type]) {
    data->esp_tr_tun_flag[data->exthdr_type] = 0;
    gtk_text_buffer_set_text (textbuffer46, "Transport mode", -1);
  } else {
    data->esp_tr_tun_flag[data->exthdr_type] = 1;
    gtk_text_buffer_set_text (textbuffer46, "Tunnel mode", -1);
  }

  // If an ESP header is attached, update ethernet frames.
  if (data->esp_hdr_flag[data->exthdr_type]) {
    ip6data_update (data->exthdr_type, data);
  }

  return (EXIT_SUCCESS);
}

// Security parameters index (SPI) (32 bits) (esp.ui)
int
on_esp_entry441_changed (GtkWidget *esp_entry441, SPSData *data) {

  const char *entry_text;
  GtkStyleContext *context;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (esp_entry441);

  // Get new value from text entry
  entry_text = gtk_entry_get_text (GTK_ENTRY (esp_entry441));

  // Is value too large or small, or are there non-numeric characters?
  if (!is_ascii_uint (entry_text) || (ascii_to_int64 ((char *) entry_text) >= 0xffffffffl)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    return (EXIT_FAILURE);

  // Value is in acceptable range.
  } else {

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    data->esphdr[data->exthdr_type].spi = htonl ((uint32_t) ascii_to_int64 ((char *) entry_text));  // Update with new value

    // Update ethernet frame and esp_window widgets.
    ip6data_update (data->exthdr_type, data);
  }

  return (EXIT_SUCCESS);
}

// Sequence number (32 bits) (esp.ui)
int
on_esp_entry442_changed (GtkWidget *esp_entry442, SPSData *data) {

  const char *entry_text;
  GtkStyleContext *context;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (esp_entry442);

  // Get new value from text entry
  entry_text = gtk_entry_get_text (GTK_ENTRY (esp_entry442));

  // Is value too large or small, or are there non-numeric characters?
  if (!is_ascii_uint (entry_text) || (ascii_to_int64 ((char *) entry_text) >= 0xffffffffl)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    return (EXIT_FAILURE);

  // Value is in acceptable range.
  } else {

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    data->esphdr[data->exthdr_type].seq = htonl ((uint32_t) ascii_to_int64 ((char *) entry_text));  // Update with new value

    // Update ethernet frame and esp_window widgets.
    ip6data_update (data->exthdr_type, data);
  }

  return (EXIT_SUCCESS);
}

// Save data for ICV calculation (esp.ui)
int
on_esp_button123_clicked (GtkButton *esp_button123, SPSData *data) {

  (void) esp_button123;  // This statement suppresses compiler warning about unused parameter.

  int i, j, result;
  GtkWidget *dialog;
  GtkFileChooserAction action = GTK_FILE_CHOOSER_ACTION_SAVE;
  char *filename, *text;
  uint8_t *buffer;
  FILE *fo;

  // Array for error messages.
  text = allocate_strmem (TEXT_STRINGLEN);

  // Create filechooser dialog.
  dialog = gtk_file_chooser_dialog_new ("Save File...", (GtkWindow *) data->main_window, action,
             "_Cancel", GTK_RESPONSE_CANCEL, "_Save", GTK_RESPONSE_ACCEPT, NULL);
  gtk_file_chooser_set_do_overwrite_confirmation (GTK_FILE_CHOOSER (dialog), TRUE);

  // Run the filechooser dialog.
  result = gtk_dialog_run (GTK_DIALOG (dialog));

  // Interpret result returned from filechooser dialog.
  if (result == GTK_RESPONSE_ACCEPT)
  {
    filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
    gtk_widget_destroy (dialog);
  } else {
    free (text);
    gtk_widget_destroy (dialog);
    return (EXIT_SUCCESS);
  }

  // Open output file.
  if ((fo = fopen (filename, "wb")) == NULL) {
    sprintf (data->error_text, "In callbacks.c: fopen() failed in on_button123_clicked().");
    data->parent = data->main_window;
    report_error (data);
    free (text);
    return (EXIT_FAILURE);
  }

  // Save ESP packet minus authentication data.
  // Section 3.1 of RFC 4303, 4305, and Section 5.1.2.2 of RFC 4301.

  // ESP header, excluding payload data, ESP tail, and auth. data
  buffer = (uint8_t *) &data->esphdr[data->exthdr_type];
  for (i=0; i<ESP_HDRLEN; i++) {
    fputc (buffer[i], fo);
  }

  // If in tunnel mode, write IPv6 header and all extension headers present.
  if (data->esp_tr_tun_flag[data->exthdr_type]) {

    // Original (inner) IPv6 header
    buffer = (uint8_t *) &data->ip6hdr[data->exthdr_type];
    for (i=0; i<IP6_HDRLEN; i++) {
      fputc (buffer[i], fo);
    }

    // Hop-by-hop header (original/inner)
    if (data->hbh_hdr_flag[data->exthdr_type][0]) {

      // Hop-by-hop header, excluding options and padding
      buffer = (uint8_t *) &data->hophdr[data->exthdr_type][0];
      for (i=0; i<HOP_HDRLEN; i++) {
        fputc (buffer[i], fo);
      }

      // Hop-by-hop options
      for (i=0; i<data->hbh_nopt[data->exthdr_type][0]; i++) {

        // Alignment padding
        for (j=0; j<data->hbh_optleadpadlen[data->exthdr_type][0][i]; j++) {
          fputc (data->hbh_optleadpad[data->exthdr_type][0][i][j], fo);
        }

        // Option
        for (j=0; j<data->hbh_optlen[data->exthdr_type][0][i]; j++) {
          fputc (data->hbh_options[data->exthdr_type][0][i][j], fo);
        }

        // Trailing padding
        for (j=0; j<data->hbh_opttailpadlen[data->exthdr_type][0]; j++) {
          fputc (data->hbh_opttailpad[data->exthdr_type][0][j], fo);
        }
      }
    }  // End if hop-by-hop header

    // Destination header (first)
    if (data->dstf_hdr_flag[data->exthdr_type][0]) {

      // Destination header
      buffer = (uint8_t *) &data->dstfhdr[data->exthdr_type][0];
      for (i=0; i<DST_HDRLEN; i++) {
        fputc (buffer[i], fo);
      }

      // Destination header options
      for (i=0; i<data->dstf_nopt[data->exthdr_type][0]; i++) {

        // Alignment padding
        for (j=0; j<data->dstf_optleadpadlen[data->exthdr_type][0][i]; j++) {
          fputc (data->dstf_optleadpad[data->exthdr_type][0][i][j], fo);
        }

        // Destination option
        for (j=0; j<data->dstf_optlen[data->exthdr_type][0][i]; j++) {
          fputc (data->dstf_options[data->exthdr_type][0][i][j], fo);
        }
      }

      // Trailing padding
      for (j=0; j<data->dstf_opttailpadlen[data->exthdr_type][0]; j++) {
        fputc (data->dstf_opttailpad[data->exthdr_type][0][j], fo);
      }
    }  // End if destination header (first)

    // Routing header
    if (data->route_hdr_flag[data->exthdr_type][0]) {

      // Routing header, excluding routing data
      buffer = (uint8_t *) &data->routehdr[data->exthdr_type][0];
      for (i=0; i<RTE_HDRLEN; i++) {
        fputc (buffer[i], fo);
      }

      // Routing header data
      for (i=0; i<data->route_datlen[data->exthdr_type][0]; i++) {
        fputc (data->route_data[data->exthdr_type][0][i], fo);
      }
    }  // End if routing header

  }  // End if tunnel mode

  // Destination header (last)
  if (data->dstl_hdr_flag[data->exthdr_type][0]) {

    // Destination header
    buffer = (uint8_t *) &data->dstlhdr[data->exthdr_type][0];
    for (i=0; i<DST_HDRLEN; i++) {
      fputc (buffer[i], fo);
    }

    // Destination header options
    for (i=0; i<data->dstl_nopt[data->exthdr_type][0]; i++) {

      // Alignment padding
      for (j=0; j<data->dstl_optleadpadlen[data->exthdr_type][0][i]; j++) {
        fputc (data->dstl_optleadpad[data->exthdr_type][0][i][j], fo);
      }

      // Destination option
      for (j=0; j<data->dstl_optlen[data->exthdr_type][0][i]; j++) {
        fputc (data->dstl_options[data->exthdr_type][0][i][j], fo);
      }
    }

    // Trailing padding
    for (j=0; j<data->dstl_opttailpadlen[data->exthdr_type][0]; j++) {
      fputc (data->dstl_opttailpad[data->exthdr_type][0][j], fo);
    }
  }  // End if destination header (last)

  // TCP header, options, and padding
  if ((data->exthdr_type == 3) || (data->exthdr_type == 9)) {

    // TCP header
    buffer = (uint8_t *) &data->tcphdr[data->exthdr_type];
    for (i=0; i<TCP_HDRLEN; i++) {
      fputc (buffer[i], fo);
    }

    // TCP options
    for (i=0; i<data->tcp_nopt[data->exthdr_type]; i++) {
      for (j=0; j<data->tcp_optlen[data->exthdr_type][i]; j++) {
        fputc (data->tcp_options[data->exthdr_type][i][j], fo);
      }
    }

    // TCP options padding
    for (i=0; i<data->tcp_optpadlen[data->exthdr_type]; i++) {
      fputc (0, fo);
    }

  // ICMP header
  } else if ((data->exthdr_type == 4) || (data->exthdr_type == 10)) {
    buffer = (uint8_t *) &data->icmp6hdr[data->exthdr_type];
    for (i=0; i<ICMP_HDRLEN; i++) {
      fputc (buffer[i], fo);
    }


  // UDP header
  } else if ((data->exthdr_type == 5) || (data->exthdr_type == 1)) {
    buffer = (uint8_t *) &data->udphdr[data->exthdr_type];
    for (i=0; i<UDP_HDRLEN; i++) {
      fputc (buffer[i], fo);
    }
  }

  // Upper layer protocol payload data
  for (i=0; i<data->payloadlen[data->exthdr_type]; i++) {
    fputc (data->payload[data->exthdr_type][i], fo);
  }

  // Encapsulating security payload trailer (transport and tunnel mode)
  // Here we need to insert the ESP payload padding as well.

  // ESP payload padding
  for (i=0; i<data->esptail[data->exthdr_type].pad_len; i++) {
    fputc (data->esp_pad[data->exthdr_type][i], fo);
  }

  // ESP trailer (includes only pad_len and nxt_hdr)
  buffer = (uint8_t *) &data->esptail[data->exthdr_type];
  for (i=0; i<ESP_TAILLEN; i++) {
    fputc (buffer[i], fo);
  }

  // Close file descriptor.
  fclose (fo);

  // Free allocated memory.
  free (text);

  return (EXIT_SUCCESS);
}

// Padding length (8 bits) (esp.ui)
int
on_esp_entry443_changed (GtkWidget *esp_entry443, SPSData *data) {

  int i;
  const char *entry_text;
  GtkStyleContext *context;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (esp_entry443);

  // Get new value from text entry
  entry_text = gtk_entry_get_text (GTK_ENTRY (esp_entry443));

  // Is value too large or small, or are there non-numeric characters?
  if (!is_ascii_uint (entry_text) || (ascii_to_int64 ((char *) entry_text) > 255)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    return (EXIT_FAILURE);

  // Value is in acceptable range.
  } else {

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    data->esptail[data->exthdr_type].pad_len = (uint8_t) ascii_to_int64 ((char *) entry_text);  // Revise ESP trailer with new value
    for (i=0; i<data->esptail[data->exthdr_type].pad_len; i++) {
      data->esp_pad[data->exthdr_type][i] = (unsigned char) (i + 1u);  // Add padding (Section 2.4 of RFC 4303 and 4305).
    }

    // Update ethernet frame and esp_window widgets.
    ip6data_update (data->exthdr_type, data);
  }

  return (EXIT_SUCCESS);
}

// Next header (8 bits) (esp.ui)
int
on_esp_entry444_changed (GtkWidget *esp_entry444, SPSData *data) {

  const char *entry_text;
  GtkStyleContext *context;
  char *value;
 
  // Get current style context of the entry.
  context = gtk_widget_get_style_context (esp_entry444);

  // String to hold miscellaneous calculated values
  value = allocate_strmem (TMP_STRINGLEN);
  
  // Get new value from text entry
  entry_text = gtk_entry_get_text (GTK_ENTRY (esp_entry444));
  
  // Is value too large or small, or are there non-numeric characters?
  if (!is_ascii_uint (entry_text) || (ascii_to_int64 ((char *) entry_text) > 255)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    free (value);
    return (EXIT_FAILURE);
    
  // Value is in acceptable range.
  } else {

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    data->esptail[data->exthdr_type].nxt_hdr = (uint8_t) ascii_to_int64 ((char *) entry_text);  // Update with new value

    // Update ethernet frame and esp_window widgets.
    ip6data_update (data->exthdr_type, data);
  } 
    
  // Free allocated memory.
  free (value);

  return (EXIT_SUCCESS);
}

// Decimal integrity check value (ICV) entry (esp.ui)
int
on_esp_radiobutton71_toggled (GtkButton *esp_radiobutton71, SPSData *data) {

  (void) esp_radiobutton71;  // This statement suppresses compiler warning about unused parameter.

  if (data->dec_hex_esp) {
    data->dec_hex_esp = 0;
  } else {
    data->dec_hex_esp = 1;
  }

  return (EXIT_SUCCESS);
}

// Hexadecimal integrity check value (ICV) entry (esp.ui)
int
on_esp_radiobutton72_toggled (GtkButton *esp_radiobutton72) {

  (void) esp_radiobutton72;  // This statement suppresses compiler warning about unused parameter.

  // on_radiobutton71_toggled() will change value of data->dec_hex_esp.

  return (EXIT_SUCCESS);
}

// Integrity check value (ICV) (variable length) (esp.ui)
int
on_esp_entry445_changed (GtkWidget *esp_entry445, SPSData *data) {

  char *value;

  // Allocate array for character string.
  value = allocate_strmem (TMP_STRINGLEN);

  data_entry (esp_entry445, &data->esp_auth_len[data->exthdr_type], MAX_ESPICVLEN, data->esp_auth_data[data->exthdr_type], data->dec_hex_esp, data->esp_window, data);

  // Update ESP header payload and padding length.
  update_esp_pad (data->exthdr_type, data);

  // If ESP header is attached, update IPv6 parameters and ethernet frames.
  if (data->esp_hdr_flag[data->exthdr_type]) {
    ip6data_update (data->exthdr_type, data);
  }

  // Update padding in ESP header editing window.
  sprintf (value, "%" PRIu8, (uint8_t) data->esptail[data->exthdr_type].pad_len);
  gtk_entry_set_text (GTK_ENTRY (data->esp_entry443), value);

  // Free allocated memory.
  free (value);

  return (EXIT_SUCCESS);
}

// Load ICV from file (esp.ui)
int
on_esp_button124_clicked (GtkButton *esp_button124, SPSData *data) {

  char *value;

  // Allocate array for character string.
  value = allocate_strmem (TMP_STRINGLEN);

  (void) esp_button124;  // This statement suppresses compiler warning about unused parameter.

  file_data (data->esp_entry445, &data->esp_auth_len[data->exthdr_type], MAX_ESPICVLEN, data->esp_auth_data[data->exthdr_type], data->exthdr_type, data->esp_window, data);

  // Update ESP header payload and padding length.
  update_esp_pad (data->exthdr_type, data);

  // If ESP header is attached, update IPv6 parameters and ethernet frames.
  if (data->esp_hdr_flag[data->exthdr_type]) {
    ip6data_update (data->exthdr_type, data);
  }

  // Update padding in ESP header editing window.
  sprintf (value, "%" PRIu8, (uint8_t) data->esptail[data->exthdr_type].pad_len);
  gtk_entry_set_text (GTK_ENTRY (data->esp_entry443), value);

  // Free allocated memory.
  free (value);

  return (EXIT_SUCCESS);
}
