/*  Simple Packet Sender (SPS) - a linux packet crafting network
    tool using GTK+ graphical interface.

    Copyright (C) 2011-2023  Hohlraum (h0h1r4um@gmail.com)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "main.h"

// Number of hop-by-hop or destination option to remove.
int
dsthop_opt_remove_entry (GtkWidget *entry, int *opt_remove) {

  const char *entry_text;

  // Get new value from text entry
  entry_text = gtk_entry_get_text (GTK_ENTRY (entry));

  // Is value too large or small, or are there non-numeric characters?
  if (!is_ascii_uint (entry_text) || ((int) ascii_to_int64 ((char *) entry_text) == 0)) {
    gtk_entry_set_text (GTK_ENTRY (entry), "");
    return (EXIT_FAILURE);

  // Value is in acceptable range.
  } else {
    *opt_remove = (int) ascii_to_int64 ((char *) entry_text);
  }

  return (EXIT_SUCCESS);
}

// Provide padding as needed to achieve alignment requirements of hop-by-hop or destination option.
int
option_pad (int *indx, uint8_t *padding, int c, int x, int y) {

  // The remainder cannot be greater than or equal to the divisor.
  if (y >= x) {
    return (EXIT_FAILURE);
  }

  int needpad;

  // Find number of padding bytes needed to achieve alignment requirements for option (Section 4.2 of RFC 8200).
  // Alignment is expressed as xN + y, which means the start of the option must occur at xN + y bytes
  // from the start of the hop-by-hop or destination header, where N is integer 0, 1, 2... etc.
  needpad = 0;
  while (((*indx + needpad) % x) != y) {
    needpad++;
  }

  // If required padding = 1 byte, we use Pad1 option.
  if (needpad == 1) {
    padding[c] = 0u;  // Padding option type: Pad1
    (*indx)++;
    c++;

  // If required padding is > 1 byte, we use PadN option.
  } else if (needpad > 1) {
    padding[c] = 1u;  // Padding option type: PadN
    (*indx)++;
    c++;
    padding[c] = (uint8_t) (needpad - 2);  // PadN length: N - 2
    (*indx)++;
    c++;
    memset (padding + c, 0, (needpad - 2) * sizeof (uint8_t));
    (*indx) += needpad - 2;
    c += needpad - 2;
  }

  return (EXIT_SUCCESS);
}

// Prepare a string of text containing hop-by-hop header or destination header options.
int
compose_exoptstring (int dsthop, int type, int io, char *string_out, int nopt, SPSData *data) {

  int iopt, i;
  int *optleadpadlen, *opttailpadlen;
  uint8_t **optleadpad, *opttailpad;
  #define BUFFER_SIZE (MAX_HBHOPTIONS * MAX_HBHOPTLEN * 4)  // Arbitrarily x 4 to hold all text
  char temp[BUFFER_SIZE], temp2[BUFFER_SIZE];

  switch (dsthop) {

    // Hop-by-hop header
    case 0:
      optleadpadlen = data->hbh_optleadpadlen[type][io];  // optleadpadlen[option #] = int
      optleadpad = data->hbh_optleadpad[type][io];  // optleadpad[option #] = uint8_t *
      opttailpadlen = &data->hbh_opttailpadlen[type][io];  // (*opttailpadlen) = int
      opttailpad = data->hbh_opttailpad[type][io];  // opttailpad = uint8_t *
      break;

    // First destination header
    case 1:
      optleadpadlen = data->dstf_optleadpadlen[type][io];  // optleadpadlen[option #] = int
      optleadpad = data->dstf_optleadpad[type][io];  // optleadpad[option #] = uint8_t *
      opttailpadlen = &data->dstf_opttailpadlen[type][io];  // (*opttailpadlen) = int
      opttailpad = data->dstf_opttailpad[type][io];  // opttailpad = uint8_t *
      break;

    // Last destination header
    case 2:
      optleadpadlen = data->dstl_optleadpadlen[type][io];  // optleadpadlen[option #] = int
      optleadpad = data->dstl_optleadpad[type][io];  // optleadpad[option #] = uint8_t *
      opttailpadlen = &data->dstl_opttailpadlen[type][io];  // (*opttailpadlen) = int
      opttailpad = data->dstl_opttailpad[type][io];  // opttailpad = uint8_t *
      break;

    default:
      fprintf (stderr, "ERROR: compose_exoptstring() received an unknown value of dsthop = %i.\n", dsthop);
      exit (EXIT_FAILURE);
  }

  // Loop through all options.
  for (iopt=0; iopt<nopt; iopt++) {

    // Display option number.
    sprintf (temp, "Option %i:\n", iopt + 1);
    strncat (string_out, temp, BUFFER_SIZE - strnlen (string_out, BUFFER_SIZE - 1) - 1);

    // Display option alignment padding.
    sprintf (temp, "Alignment: ");
    strncat (string_out, temp, BUFFER_SIZE - strnlen (string_out, BUFFER_SIZE - 1) - 1);
    memset (temp, 0, BUFFER_SIZE * sizeof (char));
    for (i=0; i<optleadpadlen[iopt]; i++) {
      sprintf (temp2, "%02x ", optleadpad[iopt][i]);
      strncat (temp, temp2, BUFFER_SIZE - strnlen (temp, BUFFER_SIZE - 1) - 1);
    }
    strncat (temp, "\n", BUFFER_SIZE - strnlen (temp, BUFFER_SIZE - 1) - 1);
    strncat (string_out, temp, BUFFER_SIZE - strnlen (string_out, BUFFER_SIZE - 1) - 1);

    memset (temp, 0, BUFFER_SIZE * sizeof (char));
    switch (dsthop) {

      // Hop-by-hop header
      case 0:
        hexascii_listing (data->hbh_options[type][io][iopt], data->hbh_optlen[type][io][iopt], temp);
        break;

      // First destination header
      case 1:
        hexascii_listing (data->dstf_options[type][io][iopt], data->dstf_optlen[type][io][iopt], temp);
        break;

      // Last destination header
      case 2:
        hexascii_listing (data->dstl_options[type][io][iopt], data->dstl_optlen[type][io][iopt], temp);
        break;

      default:
        fprintf (stderr, "ERROR: compose_exoptstring() received an unknown value of dsthop = %i.\n", dsthop);
        exit (EXIT_FAILURE);
    }

    // Add option to output string.
    strncat (string_out, temp, BUFFER_SIZE - strnlen (string_out, BUFFER_SIZE - 1) - 1);

    // Add two line-feeds.
    strncat (string_out, "\n\n", BUFFER_SIZE - strnlen (string_out, BUFFER_SIZE - 1) - 1);

  }  // End loop through options

  // Display header final padding.
  sprintf (temp, "End padding: ");
  strncat (string_out, temp, BUFFER_SIZE - strnlen (string_out, BUFFER_SIZE - 1) - 1);
  memset (temp, 0, BUFFER_SIZE * sizeof (char));
  for (i=0; i<(*opttailpadlen); i++) {
    sprintf (temp2, "%02x ", opttailpad[i]);
    strncat (temp, temp2, BUFFER_SIZE - strnlen (temp, BUFFER_SIZE - 1) - 1);
  }
  strncat (temp, "\n", BUFFER_SIZE - strnlen (temp, BUFFER_SIZE - 1) - 1);
  strncat (string_out, temp, BUFFER_SIZE - strnlen (string_out, BUFFER_SIZE - 1) - 1);

  return (EXIT_SUCCESS);
}
