/*  Simple Packet Sender (SPS) - a linux packet crafting network
    tool using GTK+ graphical interface.

    Copyright (C) 2011-2023  Hohlraum (h0h1r4um@gmail.com)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "main.h"

// Encapsulating security payload (ESP) extension header editing window
int
esp_win (SPSData *data) {

  GtkBuilder *builder;
  GError *error = NULL;

  // If ESP window is already open, return with no action.
  if (data->esp_flag) {
    return (EXIT_FAILURE);
  } else {
    data->esp_flag = 1;
  }

  // Create new GtkBuilder object.
  builder = gtk_builder_new();
  if (!gtk_builder_add_from_file (builder, "esp.ui", &error)) {
    g_warning ("%s", error->message);
    g_free (error);
    exit (EXIT_FAILURE);
  }

  // Get objects from UI file
  data->esp_window = GTK_WIDGET (gtk_builder_get_object (builder, "esp_window"));

  data->esp_button140 = GTK_BUTTON (gtk_builder_get_object (builder, "esp_checkbutton140"));  // Toggle: attach / detach an ESP header
  data->esp_textview45 = GTK_WIDGET (gtk_builder_get_object (builder, "esp_textview45"));  // Status: attached / detached
  data->esp_button141 = GTK_BUTTON (gtk_builder_get_object (builder, "esp_button141"));  // Toggle: transport / tunnel mode
  data->esp_textview46 = GTK_WIDGET (gtk_builder_get_object (builder, "esp_textview46"));  // Status: transport / tunnel mode
  data->esp_entry441 = GTK_WIDGET (gtk_builder_get_object (builder, "esp_entry441"));  // Security parameters index (SPI)
  data->esp_entry442 = GTK_WIDGET (gtk_builder_get_object (builder, "esp_entry442"));  // Sequence number
  data->esp_button123 = GTK_BUTTON (gtk_builder_get_object (builder, "esp_button123"));  // Save data for ICV calculation
  data->esp_entry443 = GTK_WIDGET (gtk_builder_get_object (builder, "esp_entry443"));  // Padding length
  data->esp_entry444 = GTK_WIDGET (gtk_builder_get_object (builder, "esp_entry444"));  // Next header
  data->esp_radiobutton71 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "esp_radiobutton71"));  // Decimal integrity check value (ICV) entry
  data->esp_radiobutton72 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "esp_radiobutton72"));  // Hexadecimal integrity check value (ICV) entry
  data->esp_entry445 = GTK_WIDGET (gtk_builder_get_object (builder, "esp_entry445"));  // Integrity check value (ICV)
  data->esp_button124 = GTK_BUTTON (gtk_builder_get_object (builder, "esp_button124"));  // Load ICV from file...

  // Connect signals.
  gtk_builder_connect_signals (builder, data);

  // Destroy builder, since we don't need it anymore.
  g_object_unref (G_OBJECT (builder));

  // Show window. All other widgets are automatically shown by GtkBuilder.
  gtk_widget_show (data->esp_window);

  // Populate/set widgets in ESP header editing window.
  esp_show (data->exthdr_type, data);

  return (EXIT_SUCCESS);
}

// Populate esp_window with contents of selected ESP header.
int
esp_show (int type, SPSData *data) {

  char *value;
  GtkTextBuffer *textbuffer45, *textbuffer46;

  // Strings to hold miscellaneous calculated values
  value = allocate_strmem (IP_MAXPACKET);

  // Flag indicating an ESP header
  textbuffer45 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->esp_textview45));
  if (data->esp_hdr_flag[data->exthdr_type]) {
    gtk_text_buffer_set_text (textbuffer45, "Attached", -1);
  } else {
    gtk_text_buffer_set_text (textbuffer45, "Not Attached", -1);
  }

  // Transport or tunnel mode
  textbuffer46 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->esp_textview46));
  if (data->esp_tr_tun_flag[data->exthdr_type]) {
    gtk_text_buffer_set_text (textbuffer46, "Tunnel mode", -1);
  } else {
    gtk_text_buffer_set_text (textbuffer46, "Transport mode", -1);
  }

  // Decimal or hexadecimal ICV data entry
  // esp.ui will slways start with hexdecimal checked.
  data->dec_hex_esp = 1;

  // Next header
  sprintf (value, "%" PRIu8, (uint8_t) data->esptail[type].nxt_hdr);
  gtk_entry_set_text (GTK_ENTRY (data->esp_entry444), value);

  // Free allocated memory.
  free (value);

  return (EXIT_SUCCESS);
}

// Update encapsulating security payload (ESP) header padding length.
int
update_esp_pad (int type, SPSData *data) {

  int payloadlen;
  char *value;

  // Allocate memory for character string.
  value = allocate_strmem (TMP_STRINGLEN);

  // Calculate ESP payload length (includes ESP header, but excludes: ESP trailer header, padding, and ICV).
  payloadlen = esp_paylen (type, data);

  // Update ESP payload padding length.
  // The ESP header Pad Length and Next Header must be right-aligned to nearest 4-byte block so
  // that ICV begins at start of a 4-byte boundary.
  // See Section 2.4 of RFC 4303.

  // Keep adding one byte at a time to pad_len until ICV starts at 4-byte boundary.
  // Section 2.4 of RFC 4303 specifies that the padding should be monotonically increasing
  // values starting at 1.
  // ESP_TAILLEN (2 bytes); includes only pad_len and nxt_hdr
  data->esptail[type].pad_len = 0;  // Initialize pad_len to 0.
  while (((payloadlen + data->esptail[type].pad_len + ESP_TAILLEN) % 4) != 0) {

    // Add a padding byte (see rule above).
    data->esp_pad[type][data->esptail[type].pad_len] = (unsigned char) (data->esptail[type].pad_len + 1u);

    // Increment padding length.
    data->esptail[type].pad_len++;
  }

  // Update ESP header editing window if open.
  if (data->esp_flag) {
    sprintf (value, "%" PRIu8, (uint8_t) data->esptail[type].pad_len);
    gtk_entry_set_text (GTK_ENTRY (data->esp_entry443), value);
  }

  // Free allocated memory.
  free (value);

  return (EXIT_SUCCESS);
}

// Calculate ESP payload length.
int
esp_paylen (int type, SPSData *data) {

// AH or ESP transport, or no AH or ESP
/* Unfragmentable portion
     10 = Original IPv6 header
     11 = Original/Inner hop-by-hop header
     12 = Original/Inner destination header (first)
     13 = Original/Inner routing header

   Fragmentable portion
      5 = auth header (no esp header)
      6 = esp header (no auth header)
     14 = Original/Inner destination header (last)
     15 = Original/Inner mobility header
      9 = No next header
     16 = TCP header
     17 = ICMPV6 header
     18 = UDP header
     19 = ESP trailer header + ICV
*/

// AH or ESP tunnel
/* Unfragmentable portion
      0 = New IPv6 header (if AH or ESP tunneling)
      1 = Outer hop-by-hop header
      2 = Outer destination header (first)
      3 = Outer routing header

   Fragmentable portion
      5 = auth header (no esp header)
      6 = esp header (no auth header)
      7 = Outer destination header (last)
      8 = Outer mobility header
      9 = No next header
     10 = Original/Inner IPv6 header
     11 = Original/Inner hop-by-hop header
     12 = Original/Inner destination header (first)
     13 = Original/Inner routing header
     14 = Original/Inner destination header (last)
     15 = Original/Inner mobility header
      9 = No next header
     16 = TCP header
     17 = ICMPV6 header
     18 = UDP header
     19 = ESP trailer header + ICV
 */

  int i, len;

  len = 0;

  i = 0;  // Index of chain of headers
  while (data->order[type][i] != 99) {

    switch (data->order[type][i]) {

      // New IPv6 header (AH or ESP tunneling)
      case 0:
        // Always above ESP header. Ignore.
        break;

      // Outer hop-by-hop header
      case 1:
        // Always above ESP header. Ignore.
        break;

      // Outer destination header (first)
      case 2:
        // Always above ESP header. Ignore.
        break;

      // Outer routing header
      case 3:
        // Always above ESP header. Ignore.
        break;

      // Unused
      case 4:
        // Do nothing.
        break;

      // AH header
      case 5:
        fprintf (stderr, "ERROR: esp_paylen() encountered an AH header at index %i.\n", i);
        exit (EXIT_FAILURE);

      // ESP header; contains only spi and seq
      case 6:
        // ESP payload is defined (Section 2 of RFC 4303) to be
        // everything between seq of ESP header and padding of ESP trailer.
        // Therefore, we don't include the ESP header (spi and seq).
        break;

      // Outer destination header (last)
      case 7:
        len += DST_HDRLEN + data->dstl_opt_totlen[type][1] + data->dstl_optpadlen[type][1];
        break;

      // Outer mobility header
      case 8:
        // Not implemented here. Do nothing.
        break;

      // No next header
      case 9:
        // Do nothing.
        break;

      // Original/Inner IPv6 header
      case 10:
        // Include if in ESP tunnel mode.
        if (data->esp_tr_tun_flag[type]) {
          len += IP6_HDRLEN;
        }
        break;

      // Original/Inner hop-by-hop header
      case 11:
        // Include if in ESP tunnel mode.
        if (data->esp_tr_tun_flag[type]) {
          len += HOP_HDRLEN + data->hbh_opt_totlen[type][0] + data->hbh_optpadlen[type][0];
        }
        break;

      // Original/Inner destination header (first)
      case 12:
        // Include if in ESP tunnel mode.
        if (data->esp_tr_tun_flag[type]) {
          len += DST_HDRLEN + data->dstf_opt_totlen[type][0] + data->dstf_optpadlen[type][0];
        }
        break;

      // Original/Inner routing header
      case 13:
        // Include if in ESP tunnel mode.
        if (data->esp_tr_tun_flag[type]) {
          len += RTE_HDRLEN + data->route_datlen[type][0];
        }
        break;

      // Original/Inner destination header (last)
      case 14:
        len += DST_HDRLEN + data->dstl_opt_totlen[type][0] + data->dstl_optpadlen[type][0];
        break;

      // Original/Inner mobility header
      case 15:
        // No implemented here. Do nothing.
        break;

      // TCP header, options, and payload data
      case 16:
        len += TCP_HDRLEN + data->tcp_opt_totlen[type] + data->tcp_optpadlen[type];
        len += data->payloadlen[type];
        break;

      // ICMPv6 header andpayload data
      case 17:
        len += ICMP_HDRLEN;
        len += data->payloadlen[type];
        break;

      // UDP header and payload data
      case 18:
        len += UDP_HDRLEN;
        len += data->payloadlen[type];
        break;

      // ESP trailer (padding, padlen, nxt_hdr) and ICV
      case 19:
        // ESP payload is defined (Section 2 of RFC 4304) to be
        // everything between seq of ESP header and padding of ESP trailer.
        // We don't include because this is the header we're trying to align using len.
        break;

      default:
        fprintf (stderr, "ERROR: esp_paylen() encountered an invalid value of data->order[type][i] = %i.\n", data->order[type][i]);
        exit (EXIT_FAILURE);

    }  // End switch

    i++;  // Go to next link in chain.
  }  // End while

  return (len);
}
