/*  Simple Packet Sender (SPS) - a linux packet crafting network
    tool using GTK+ graphical interface.

    Copyright (C) 2011-2023  Hohlraum (h0h1r4um@gmail.com)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "main.h"

// Check for valid link-layer (MAC) address (48 bits) and process.
int
mac_entry (GtkWidget *entry, uint8_t *mac, int type, SPSData *data) {

  int i, bad;
  const char *entry_text;
  GtkStyleContext *context;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (entry);

  // Get new value from text entry
  entry_text = gtk_entry_get_text (GTK_ENTRY (entry));

  // Is new value in appropriate form?

  // Check length.
  bad = 0;
  if (strlen (entry_text) != 17) {
    bad = 1;
  }

  // Check for colons
  i = 2;
  while (i < 17) {
    if (entry_text[i] != ':') {
      bad = 1;
    }
    i += 3;
  }

  // Check for valid hexadecimals
  i = 0;
  while (i < 17) {
    if (!(((entry_text[i] > 47) && (entry_text[i] < 58)) ||    // 0 to 10
        ((entry_text[i] > 64) && (entry_text[i] < 71)) ||      // A to Z
        ((entry_text[i] > 96) && (entry_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i++;
    if (!(((entry_text[i] > 47) && (entry_text[i] < 58)) ||    // 0 to 10
        ((entry_text[i] > 64) && (entry_text[i] < 71)) ||      // A to Z
        ((entry_text[i] > 96) && (entry_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i += 2;
  }
  if (bad) {
    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");
  } else {
    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    // Store new value.
    mac_string2bin ((char *) entry_text, mac);

    // Update ethernet frames.
    if ((type == 0) || (type == 1) || (type == 2) ||
      (type == 9) || (type == 10) || (type == 11)) {
      create_ip4_frame (type, data);

    } else if ((type == 3) || (type == 4) || (type == 5) ||
      (type == 12) || (type == 13) || (type == 14)) {
      create_ip6_frame (type, data);
      create_6to4_frame (type + 3, data);
    }
  }

  return (EXIT_SUCCESS);
}

// Find link-layer (MAC) address for given interface name.
int
ifname_entry (GtkWidget *entry, GtkWidget *mac_entry, int type, SPSData *data) {

  int ifmtu, min;
  const char *entry_text;
  uint8_t *mac;
  GtkStyleContext *context;
  char *text, *ifname;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (entry);

  // Array for error messages.
  text = allocate_strmem (TEXT_STRINGLEN);
  ifname = allocate_strmem (TMP_STRINGLEN);
  mac = allocate_ustrmem (6);

  // Get new name from text entry.
  entry_text = gtk_entry_get_text (GTK_ENTRY (entry));

  // Set parent window for reporting error messages.
  if (type < 9) {
    data->parent = data->main_window;
  } else {
    data->parent = data->traceroute_window;
  }

  // Check for bad interface name.
  if (interface_getmac ((char *) entry_text, mac) == EXIT_FAILURE) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    free (text);
    free (ifname);
    free (mac);
    return (EXIT_FAILURE);

  // Interface name is OK.
  } else {

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");
  }

  // Obtain interface's MTU.
  // If we fail to obtain MTU from interface, don't change current MTU.
  if ((ifmtu = interface_getmtu ((char *) entry_text)) == 0) {
    sprintf (data->warning_text, "Failed to obtain interface's MTU.");
    report_warning (data);
    free (text);
    free (ifname);
    free (mac);
    return (EXIT_FAILURE);
  }

  // Maximum transmission unit (MTU) checks

  // IPv4
  if ((type == 0) || (type == 1) || (type == 2) || (type == 9) || (type == 10) || (type == 11)) {

    // Determine minimum size needed to accommmodate packet,
    // even when packet is fragmented to maximum extent.
    min = min4_mtu (type, data);

    // Is interface's MTU is too small to accommmodate IPv4 packet,
    // even when packet is fragmented to maximum extent?
    if (ifmtu < min) {
      sprintf (data->warning_text, "The interface's MTU is too small to accommodate IPv4 packet.");
      report_warning (data);
      gtk_entry_set_text (GTK_ENTRY (entry), "");  // Clear interface name from entry.
      memset (text, 0, TEXT_STRINGLEN * sizeof (char));
      mac_bin2string (data->ethhdr[type].src_mac, text);
      gtk_entry_set_text (GTK_ENTRY (mac_entry), text);  // Restore current source MAC address to entry.
      free (text);
      free (ifname);
      free (mac);
      return (EXIT_FAILURE);
    }

    // Is current MTU value too small to accommodate packet,
    // even when packet is fragmented to maximum extent?
    if (data->ifmtu[type] < min) {
      sprintf (data->warning_text, "The current MTU value is too small to accommodate IPv4 packet.\nPacket needs a minimum of %i bytes.\nValue revised to interface MTU.", min);
      report_warning (data);
      data->ifmtu[type] = ifmtu;

      // Update MTU spinbutton.
      update_mtu_spinbutton (type, data);
    }

  // IPv6
  } else if ((type == 3) || (type == 4) || (type == 5) || (type == 12) || (type == 13) || (type == 14)) {

    // Determine minimum size needed to accommmodate packet,
    // even when packet is fragmented to maximum extent.
    min = min6_mtu (type, data);

    // Is interface's MTU is too small to accommmodate IPv6 packet,
    // even when packet is fragmented to maximum extent?
    if (ifmtu < min) {
      sprintf (data->warning_text, "The interface's MTU is too small to accommodate IPv6 packet.");
      report_warning (data);
      gtk_entry_set_text (GTK_ENTRY (entry), "");  // Clear interface name from entry.
      memset (text, 0, TEXT_STRINGLEN * sizeof (char));
      mac_bin2string (data->ethhdr[type].src_mac, text);
      gtk_entry_set_text (GTK_ENTRY (mac_entry), text);  // Restore current source MAC address to entry.
      free (text);
      free (ifname);
      free (mac);
      return (EXIT_FAILURE);
    }

    // Is current MTU value too small to accommodate packet,
    // even when packet is fragmented to maximum extent?
    if (data->ifmtu[type] < min) {
      sprintf (data->warning_text, "The current MTU value is too small to accommodate IPv6 packet.\nPacket needs a minimum of %i bytes.\nValue revised to interface MTU.", min);
      report_warning (data);
      data->ifmtu[type] = ifmtu;

      // Update MTU spinbutton.
      update_mtu_spinbutton (type, data);
    }
  }

  // 6to4 is not affected by interface name & MTU on IPv6 page.
  // See sixto4_ifname_entry().

  // Use interface's MTU.
  data->ifmtu[type] = ifmtu;

  // Update interface name.
  memset (ifname, 0, TMP_STRINGLEN * sizeof (char));
  strncpy (data->ifname[type], entry_text, TMP_STRINGLEN - 1);  // Minus 1 for string termination.

  // Update source MAC address and UI.
  memcpy (data->ethhdr[type].src_mac, mac, 6 * sizeof (uint8_t));
  memset (text, 0, TEXT_STRINGLEN * sizeof (char));
  mac_bin2string (mac, text);
  gtk_entry_set_text (GTK_ENTRY (mac_entry), text);

  // Update ethernet frames.
  if ((type == 0) || (type == 1) || (type == 2) ||
    (type == 9) || (type == 10) || (type == 11)) {
    create_ip4_frame (type, data);

  } else if ((type == 3) || (type == 4) || (type == 5) ||
    (type == 12) || (type == 13) || (type == 14)) {
    create_ip6_frame (type, data);
  }

  // Free allocated memory.
  free (text);
  free (ifname);
  free (mac);

  return (EXIT_SUCCESS);
}

// (6to4 only) Find link-layer (MAC) address for given interface name.
int
sixto4_ifname_entry (GtkWidget *entry, GtkWidget *mac_entry, int type, SPSData *data) {

  int ifmtu, min;
  const char *entry_text;
  uint8_t *mac;
  GtkStyleContext *context;
  char *text, *ifname;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (entry);

  // Array for error messages.
  text = allocate_strmem (TEXT_STRINGLEN);
  ifname = allocate_strmem (TMP_STRINGLEN);
  mac = allocate_ustrmem (6);

  // Get new name from text entry.
  entry_text = gtk_entry_get_text (GTK_ENTRY (entry));

  // Set parent window for reporting error messages.
  data->parent = data->main_window;

  // Check for bad interface name.
  if (interface_getmac ((char *) entry_text, mac) == EXIT_FAILURE) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    free (text);
    free (ifname);
    free (mac);
    return (EXIT_FAILURE);

  // Interface name is OK.
  } else {

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");
  }

  // Maximum transmission unit (MTU) checks
  // MTU should be 1280 bytes; can be 1280 - 1480 bytes (RFC 4213).

  // Obtain interface's MTU.
  // If we fail to obtain MTU from interface, don't change current MTU.
  if ((ifmtu = interface_getmtu ((char *) entry_text)) == 0) {
    sprintf (data->warning_text, "Failed to obtain MTU from interface for 6to4.");
    report_warning (data);
    free (text);
    free (ifname);
    free (mac);
    return (EXIT_FAILURE);
  }

  // Determine minimum size needed to accommodate packet,
  // when packet is fragmented to maximum extent.
  min = min6to4_mtu (type, data);

  // Is interface's MTU is too small to accommodate 6to4 packet,
  // even when packet is fragmented to maximum extent?
  if (ifmtu < min) {
    sprintf (data->warning_text, "The interface's MTU is too small to accommodate 6to4 packet.");
    report_warning (data);
    gtk_entry_set_text (GTK_ENTRY (entry), "");  // Clear interface name from entry.
    memset (text, 0, TEXT_STRINGLEN * sizeof (char));
    mac_bin2string (data->ethhdr[type].src_mac, text);
    gtk_entry_set_text (GTK_ENTRY (mac_entry), text);  // Restore current source MAC address to entry.
    free (text);
    free (ifname);
    free (mac);
    return (EXIT_FAILURE);
  }

  // Is current MTU value too small to accommodate packet,
  // even when packet is fragmented to maximum extent?
  if (data->ifmtu[type] < min) {
    sprintf (data->warning_text, "The current MTU value is too small to accommodate 6to4 packet.\nPacket needs a minimum of %i bytes.\nValue revised to standard 6to4 MTU of 1280 bytes (RFC 4213).", min);
    report_warning (data);
    data->ifmtu[type] = 1280;  // RFC 4213

    // Update MTU spinbutton.
    update_mtu_spinbutton (type, data);
  }

  // Update ethernet frame.
  create_6to4_frame (type, data);

  // Update interface name.
  memset (ifname, 0, TMP_STRINGLEN * sizeof (char));
  strncpy (data->ifname[type], entry_text, TMP_STRINGLEN - 1);  // Minus 1 for string termination.

  // Update source MAC address and UI.
  memcpy (data->ethhdr[type].src_mac, mac, 6 * sizeof (unsigned char));
  memset (text, 0, TEXT_STRINGLEN * sizeof (char));
  mac_bin2string (mac, text);
  gtk_entry_set_text (GTK_ENTRY (mac_entry), text);

  // Free allocated memory.
  free (text);
  free (ifname);
  free (mac);

  return (EXIT_SUCCESS);
}

// Ethernet type code (16 bits)
// http://www.iana.org/assignments/ethernet-numbers
int
ethtype_entry (GtkWidget *entry, int type, SPSData *data) {

  const char *entry_text;
  GtkStyleContext *context;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (entry);

  // Get new value from text entry
  entry_text = gtk_entry_get_text (GTK_ENTRY (entry));

  // Is value too large or small, or are there non-numeric characters?
  if (!is_ascii_uint (entry_text) || (ascii_to_int64 ((char *) entry_text) > 65535u)) {
    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");
    return (EXIT_FAILURE);

  // Value is in acceptable range.
  } else {
    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    // Store new value.
    data->ethhdr[type].type_code = htons ((uint16_t) ascii_to_int64 ((char *) entry_text));  // Update ethernet header with new value

    // Update ethernet frames.
    if ((type == 0) || (type == 1) || (type == 2) ||
      (type == 9) || (type == 10) || (type = 11)) {
      create_ip4_frame (type, data);

    } else if ((type == 3) || (type == 4) || (type == 5) ||
      (type == 12) || (type == 13) || (type == 14)) {
      // Update ethernet type code in 6to4 ethernet header, since create_6to4_frame() doesn't copy it.
      data->ethhdr[type + 3].type_code = htons ((uint16_t) ascii_to_int64 ((char *) entry_text));
      // Update ethernet frames.
      create_ip6_frame (type, data);
      create_6to4_frame (type + 3, data);
    }
  }

  return (EXIT_SUCCESS);
}

// Convert link-layer (MAC) address from binary to colon-separated string.
int
mac_bin2string (unsigned char *mac, char *string) {

  int i;
  char *temp;

  // Allocate memory for temporary string.
  // Use smallest allocation size (TMP_STRINGLEN) for string in all other functions.
  temp = allocate_strmem (TMP_STRINGLEN);

  // NOTE: we assume string was set to zero prior to call to this function.

  for (i=0; i<5; i++) {
    memset (temp, 0, TMP_STRINGLEN * sizeof (char));
    sprintf (temp, "%02x:", mac[i]);
    strncat (string, temp, TMP_STRINGLEN - strnlen (string, TMP_STRINGLEN - 1) - 1);
  }
  memset (temp, 0, TMP_STRINGLEN * sizeof (char));
  sprintf (temp, "%02x", mac[5]);
  strncat (string, temp, TMP_STRINGLEN - strnlen (string, TMP_STRINGLEN - 1) - 1);

  // Free allocated memory.
  free (temp);

  return (EXIT_SUCCESS);
}

// Convert link-layer (MAC) address from colon-separated string to binary.
int
mac_string2bin (char *string, uint8_t *mac) {

  unsigned int a, b, c, d, e, f;

  sscanf (string, "%02x:%02x:%02x:%02x:%02x:%02x", &a, &b, &c, &d, &e, &f);

  mac[0] = a;
  mac[1] = b;
  mac[2] = c;
  mac[3] = d;
  mac[4] = e;
  mac[5] = f;

  return (EXIT_SUCCESS);
}
