/*  Simple Packet Sender (SPS) - a linux packet crafting network
    tool using GTK+ graphical interface.

    Copyright (C) 2011-2023  Hohlraum (h0h1r4um@gmail.com)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "main.h"

// Upper layer protocol data from file (IPv4 or IPv6)
int
file_payload_data (GtkWidget *entry, int type, SPSData *data) {

  int i, index, filelen, result, flag, ifmtu, min;
  char *filename;
  uint8_t *buffer;
  GtkWidget *dialog;
  GtkFileChooserAction action = GTK_FILE_CHOOSER_ACTION_OPEN;
  GtkTextBuffer *textbuffer60, *textbuffer61, *textbuffer62, *textbuffer63, *textbuffer64,
                *textbuffer65, *textbuffer66, *textbuffer67, *textbuffer68, *textbuffer69,
                *textbuffer70, *textbuffer71;
  FILE *fi;

  if (type < 9) {
    data->parent = data->main_window;
  } else {
    data->parent = data->traceroute_window;
  }

  // Check for valid interface name. Abort if not.
  if (interface_getmtu (data->ifname[type]) == 0) {
    sprintf (data->warning_text, "Need valid interface to be able to check MTU.\nWas interface specified?");
    report_warning (data);
    return (EXIT_FAILURE);
  }

  // Create filechooser dialog.
  dialog = gtk_file_chooser_dialog_new ("Load File...", (GtkWindow *) data->main_window, action,
             "_Cancel", GTK_RESPONSE_CANCEL, "_Open", GTK_RESPONSE_ACCEPT, NULL);

  // Run the filechooser dialog.
  result = gtk_dialog_run (GTK_DIALOG (dialog));

  // Interpret result returned from filechooser dialog.
  if (result == GTK_RESPONSE_ACCEPT) {
    filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
    gtk_widget_destroy (dialog);
  } else {
    gtk_widget_destroy (dialog);
    return (EXIT_SUCCESS);
  }

  // Open file.
  if ((fi = fopen (filename, "rb")) == NULL) {
    sprintf (data->error_text, "file_payload_data(): fopen() failed.");
    report_error (data);
    return (EXIT_FAILURE);
  }

  // Count total number of bytes in file.
  filelen = 0;
  while ((i = fgetc(fi)) != EOF) {
    filelen++;
  }
  rewind (fi);

  // Allocate memory for array of unsigned chars
  buffer = allocate_ustrmem (filelen);

  // Read file into buffer.
  for (index=0; index<filelen; index++) {
    if ((i = fgetc (fi)) == EOF) {
      sprintf (data->error_text, "file_payload_data(): fgetc() failed.");
      report_error (data);
      free (buffer);
      return (EXIT_FAILURE);
    }
    buffer[index] = (uint8_t) i;
  }

  // Close file descriptor.
  fclose (fi);

  // Clear current data.
  memset (data->payload[type], 0, IP_MAXPACKET * sizeof (uint8_t));

  // Clear text entry since we used a file and not the keyboard.
  gtk_entry_set_text (GTK_ENTRY (entry), "");

  // New payload length based upon file contents.
  data->payloadlen[type] = filelen;

  // Truncate upper layer protocol payload if necessary to accommodate packet size limit.
  flag = 0;
  if ((type == 0) || (type == 9) || (type == 1) || (type == 10) || (type == 2) || (type == 11)) {
    if (truncate_ip4 (type, data)) {
      flag = 1;
    }
  } else {
    if (truncate_ip6 (type, data)) {
      flag = 1;
    }
  }
  if (flag) {
    sprintf (data->warning_text, "Upper layer protocol payload data was truncated\nto accommodate maximum packet size of %i bytes.", IP_MAXPACKET);
    report_warning (data);
  }

  // Copy payload from buffer.
  memcpy (data->payload[type], buffer, data->payloadlen[type] * sizeof (uint8_t));

  // Update lengths, offsets, and checksums.
  if ((type == 0) || (type == 1) || (type == 2) ||
      (type == 9) || (type == 10) || (type == 11)) {
    ip4data_update (type, data);

  } else if ((type == 3) || (type == 4) || (type == 5) ||
      (type == 12) || (type == 13) || (type == 14)) {
    ip6data_update (type, data);
  }

  // Maximum transmission unit (MTU) checks

  // IPv4
  if ((type == 0) || (type == 1) || (type == 2) || (type == 9) || (type == 10) || (type == 11)) {

    // Obtain interface's MTU.
    // If we fail to obtain MTU from interface, don't change current MTU.
    if ((ifmtu = interface_getmtu (data->ifname[type])) == 0) {
      sprintf (data->warning_text, "Failed to obtain MTU from interface for IPv4.\nWas interface specified?");
      report_warning (data);
      free (buffer);
      return (EXIT_FAILURE);
    }

    // Determine minimum size needed to accommodate packet,
    // when packet is fragmented to maximum extent.
    min = min4_mtu (type, data);

    // Is interface's MTU is too small to accommodate IPv4 packet,
    // even when packet is fragmented to maximum extent?
    if (ifmtu < min) {
      sprintf (data->warning_text, "The interface's MTU is too small to accommodate IPv4 packet.");
      report_warning (data);
      free (buffer);
      return (EXIT_FAILURE);
    }

    // Is current MTU value too small to accommodate packet,
    // even when packet is fragmented to maximum extent?
    if (data->ifmtu[type] < min) {
      sprintf (data->warning_text, "The current MTU value is too small to accommodate IPv4 packet.\nPacket needs a minimum of %i bytes.\nValue revised to interface MTU.", min);
      report_warning (data);
      data->ifmtu[type] = ifmtu;

      // Update MTU spinbutton.
      update_mtu_spinbutton (type, data);
    }

    // Update ethernet frame.
    create_ip4_frame (type, data);

  // IPv6 (which also impacts 6to4)
  } else if ((type == 3) || (type == 4) || (type == 5) || (type == 12) || (type == 13) || (type == 14)) {

    // Obtain interface's MTU.
    // If we fail to obtain MTU from interface, don't change current MTU.
    if ((ifmtu = interface_getmtu (data->ifname[type])) == 0) {
      sprintf (data->warning_text, "Failed to obtain MTU from interface for IPv6.\nWas interface specified?");
      report_warning (data);
      free (buffer);
      return (EXIT_FAILURE);
    }

    // Determine minimum size needed to accommodate packet,
    // when packet is fragmented to maximum extent.
    min = min6_mtu (type, data);

    // Is interface's MTU is too small to accommodate packet,
    // even when packet is fragmented to maximum extent?
    if (ifmtu < min) {
      sprintf (data->warning_text, "The interface's MTU is too small to accommodate IPv6 packet.");
      report_warning (data);
      free (buffer);
      return (EXIT_FAILURE);
    }

    // Is current MTU value too small to accommodate IPv6 packet,
    // even when packet is fragmented to maximum extent?
    if (data->ifmtu[type] < min) { 
      sprintf (data->warning_text, "The current MTU value is too small to accommodate IPv6 packet.\nPacket needs a minimum of %i bytes.\nValue revised to interface MTU.", min);
      report_warning (data);
      data->ifmtu[type] = ifmtu;

      // Update MTU spinbutton.
      update_mtu_spinbutton (type, data);
    }

    // 6to4 (impacted by IPv6)

    // Here we do not fetch the sixto4 interface ifmtu because the user may
    // have no intention of doing 6to4, in which case a 6to4 interface would not be specified
    // on the 6to4 page of UI. As a result, 6to4 MTU must be checked at Send time.

    // Update ethernet frame.
    create_ip6_frame (type, data);
    create_6to4_frame (type+3, data);
  }

  // Make user aware a data file is loaded.
  // Clear manual entry text.
  switch (type) {
    case 0:  // IPv4 TCP
      textbuffer60 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview60));
      gtk_text_buffer_set_text (textbuffer60, "File Loaded", -1);
      break;
    case 1:  // IPv4 ICMP
      textbuffer62 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview62));
      gtk_text_buffer_set_text (textbuffer62, "File Loaded", -1);
      break;
    case 2:  // IPv4 UDP
      textbuffer64 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview64));
      gtk_text_buffer_set_text (textbuffer64, "File Loaded", -1);
      break;
    case 3:  // IPv6 TCP
      textbuffer61 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview61));
      gtk_text_buffer_set_text (textbuffer61, "File Loaded", -1);
      break;
    case 4:  // IPv6 ICMP
      textbuffer63 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview63));
      gtk_text_buffer_set_text (textbuffer63, "File Loaded", -1);
      break;
    case 5:  // IPv6 UDP
      textbuffer65 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview65));
      gtk_text_buffer_set_text (textbuffer65, "File Loaded", -1);
      break;
    case 9:  // IPv4 TCP for traceroute
      textbuffer66 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->tr_textview66));
      gtk_text_buffer_set_text (textbuffer66, "File Loaded", -1);
      break;
    case 10:  // IPv4 ICMP for traceroute
      textbuffer67 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->tr_textview67));
      gtk_text_buffer_set_text (textbuffer67, "File Loaded", -1);
      break;
    case 11:  // IPv4 UDP for traceroute
      textbuffer68 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->tr_textview68));
      gtk_text_buffer_set_text (textbuffer68, "File Loaded", -1);
      break;
    case 12:  // IPv6 TCP for traceroute
      textbuffer69 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->tr_textview69));
      gtk_text_buffer_set_text (textbuffer69, "File Loaded", -1);
      break;
    case 13:  // IPv6 ICMP for traceroute
      textbuffer70 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->tr_textview70));
      gtk_text_buffer_set_text (textbuffer70, "File Loaded", -1);
      break;
    case 14:  // IPv6 UDP for traceroute
      textbuffer71 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->tr_textview71));
      gtk_text_buffer_set_text (textbuffer71, "File Loaded", -1);
      break;
    default:
      fprintf (stderr, "ERROR: Invalid upper layer protocol type %i in file_data().\n", type);
      free (buffer);
      exit (EXIT_FAILURE);
  }

  // Free allocated memory.
  free (buffer);

  return (EXIT_SUCCESS);
}

// Extension header data entry from file (IPv6)
int
file_data (GtkWidget *entry, int *datlen, int maxdatlen, uint8_t *dat, int type, GtkWidget *parent, SPSData *data) {

  int i, c, index, filelen, result, ifmtu, min;
  char *filename, *temp, *value;
  uint8_t *buffer;
  GtkWidget *dialog;
  GtkFileChooserAction action = GTK_FILE_CHOOSER_ACTION_OPEN;
  FILE *fi;

  // Arrays of chars
  value = allocate_strmem (IP_MAXPACKET);
  temp = allocate_strmem (TMP_STRINGLEN);

  data->parent = parent;

  // Create filechooser dialog.
  dialog = gtk_file_chooser_dialog_new ("Load File...", (GtkWindow *) data->parent, action,
             "Cancel", GTK_RESPONSE_CANCEL, "_Open", GTK_RESPONSE_ACCEPT, NULL);

  // Run the filechooser dialog.
  result = gtk_dialog_run (GTK_DIALOG (dialog));

  // Interpret result returned from filechooser dialog.
  if (result == GTK_RESPONSE_ACCEPT) {
    filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
    gtk_widget_destroy (dialog);
  } else {
    gtk_widget_destroy (dialog);
    free (value);
    free (temp);
    return (EXIT_FAILURE);
  }

  // Open file.
  if ((fi = fopen (filename, "rb")) == NULL) {
    sprintf (data->error_text, "file_data(): fopen() failed.");
    report_error (data);
    free (value);
    free (temp);
    return (EXIT_FAILURE);
  }

  // Count total number of bytes in file.
  filelen = 0;
  while ((i = fgetc(fi)) != EOF) {
    filelen++;
  }
  rewind (fi);

  // Array of unsigned chars
  buffer = allocate_ustrmem (filelen);

  // Read file into buffer.
  for (index=0; index<filelen; index++) {
    if ((i = fgetc (fi)) == EOF) {
      sprintf (data->error_text, "file_data(): fgetc() failed.");
      report_error (data);
      free (buffer);
      free (value);
      free (temp);
      return (EXIT_FAILURE);
    }
    buffer[index] = (uint8_t) i;
  }

  // Close file descriptor.
  fclose (fi);

  // Clear current data.
  memset (dat, 0, maxdatlen * sizeof (uint8_t));

  // Clear text entry since we used a file and not the keyboard.
  gtk_entry_set_text (GTK_ENTRY (entry), "");

  // New data length based upon file contents or limit.
  if (filelen <= maxdatlen) {
    (*datlen) = filelen;
  } else {
    (*datlen) = maxdatlen;
    sprintf (data->warning_text, "Extension header data was truncated\nto accommodate maximum size of %i bytes.", maxdatlen);
    report_warning (data);
  }

  // Copy new data into array specified in file_data() argument four.
  memcpy (dat, buffer, (*datlen) * sizeof (uint8_t));

  // Display data in entry.
  // NOTE: Always displays as comma-separated decimal numbers.
  memset (value, 0, IP_MAXPACKET * sizeof (char));
  c = 0;
  for (i=0; i<(*datlen); i++) {
    sprintf (temp, "%" PRIu8, (uint8_t) dat[i]);
    strncpy (value + c, temp, IP_MAXPACKET - c - 1);  // Minus 1 for string termination.
    c += strnlen (temp, TMP_STRINGLEN);
    if (i != ((*datlen) - 1)) {  // There's more values to come.
      memset (temp, 0, TMP_STRINGLEN * sizeof (char));
      sprintf (temp, ", ");
      strncat (value, temp, IP_MAXPACKET - strnlen (value, IP_MAXPACKET - 1) - 1);  // Add a comma and space.
      c += 2;
    }
  }
  gtk_entry_set_text (GTK_ENTRY (entry), value);

  // Truncate upper layer protocol payload if necessary to accommodate packet size limit.
  if (truncate_ip6 (type, data)) {
    sprintf (data->warning_text, "Upper layer protocol payload data was truncated\nto accommodate maximum packet size of %i bytes.", IP_MAXPACKET);
    report_warning (data);
  }

  // Maximum transmission unit (MTU) checks

  // IPv6 (which also impacts 6to4)

  // Obtain interface's MTU.
  // If we fail to obtain MTU from interface, don't change current MTU.
  if ((ifmtu = interface_getmtu (data->ifname[type])) == 0) {
    sprintf (data->warning_text, "Failed to obtain MTU from interface for IPv6.\nWas interface specified?");
    report_warning (data);
    free (buffer);
    free (value);
    free (temp);
    return (EXIT_FAILURE);
  }

  // Determine minimum size needed to accommodate packet,
  // when packet is fragmented to maximum extent.
  min = min6_mtu (type, data);

  // Is interface's MTU is too small to accommodate packet,
  // even when packet is fragmented to maximum extent?
  if (ifmtu < min) {
    sprintf (data->warning_text, "The interface's MTU is too small to accommodate IPv6 packet.");
    report_warning (data);
    free (buffer);
    free (value);
    free (temp);
    return (EXIT_FAILURE);
  }

  // Is current MTU value too small to accommodate packet,
  // even when packet is fragmented to maximum extent?
  if (data->ifmtu[type] < min) {
    sprintf (data->warning_text, "The current MTU value is too small to accommodate IPv6 packet.\nPacket needs a minimum of %i bytes.\nValue revised to interface MTU.", min);
    report_warning (data);
    data->ifmtu[type] = ifmtu;

    // Update MTU spinbutton.
    update_mtu_spinbutton (type, data);
  }

  // 6to4 (impacted by IPv6)

  // Here we do not fetch the sixto4 interface ifmtu because the user may
  // have no intention of doing 6to4, in which case a 6to4 interface would not be specified
  // on the 6to4 page of UI. As a result, 6to4 MTU must be checked at Send time.

  // Update MTU spinbutton.
  update_mtu_spinbutton (type, data);

  // Update ethernet frame.
  create_ip6_frame (type, data);  
  create_6to4_frame (type+3, data);

  // Free allocated memory.
  free (buffer);
  free (value);
  free (temp);

  return (EXIT_SUCCESS);
}
