/*  Simple Packet Sender (SPS) - a linux packet crafting network
    tool using GTK+ graphical interface.

    Copyright (C) 2011-2023  Hohlraum (h0h1r4um@gmail.com)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "main.h"

// Set the Next Header field to IPPROTO_FRAGMENT in last link of unfragmentable portion (IPv6 only).
int
next_hdr_to_frag (int type, struct ip6_hdr *ip6hdr, SPSData *data) {

// Possible last links in unfragmentable portion are:

// AH or ESP transport, or no AH or ESP
/* Unfragmentable portion
     10 = Original IPv6 header
     11 = Original/Inner hop-by-hop header
     12 = Original/Inner destination header (first)
     13 = Original/Inner routing header
*/

// AH or ESP tunnel
/* Unfragmentable portion
      0 = New IPv6 header (if AH or ESP tunneling)
      1 = Outer hop-by-hop header
      2 = Outer destination header (first)
      3 = Outer routing header
*/

  switch (data->order[type][data->first_frag[type]-1]) {

    // New IPv6 header (if AH or ESP tunneling)
    case 0:
      data->ip6hdr[type+20].ip6_nxt = IPPROTO_FRAGMENT;
      break;

    // Outer hop-by-hop header
    case 1:
      data->hophdr[type][1].nxt_hdr = IPPROTO_FRAGMENT;
      break;

    // Outer destination header (first)
    case 2:
      data->dstfhdr[type][1].nxt_hdr = IPPROTO_FRAGMENT;
      break;

    // Outer routing header
    case 3:
      data->routehdr[type][1].nxt_hdr = IPPROTO_FRAGMENT;
      break;

    // Unused
    case 4:
      fprintf (stderr, "ERROR: In next_hdr_to_frag(), data->order[type][data->first_frag[type]-1] = %i is an unused index.\n", data->order[type][data->first_frag[type]-1]);
      exit (EXIT_FAILURE);
      break;

    // auth header (no esp header)
    case 5:
      fprintf (stderr, "ERROR: In next_hdr_to_frag(), data->order[type][data->first_frag[type]-1] = %i is an AH header, which is fragmentable.\n", data->order[type][data->first_frag[type]-1]);
      exit (EXIT_FAILURE);
      break;

    // esp header (no auth header)
    case 6:
      fprintf (stderr, "ERROR: In next_hdr_to_frag(), data->order[type][data->first_frag[type]-1] = %i is an ESP header, which is fragmentable.\n", data->order[type][data->first_frag[type]-1]);
      exit (EXIT_FAILURE);
      break;

    // Outer destination header (last)
    case 7:
      fprintf (stderr, "ERROR: In next_hdr_to_frag(), data->order[type][data->first_frag[type]-1] = %i is an an outer destination header (first), which is fragmentable.\n", data->order[type][data->first_frag[type]-1]);
      exit (EXIT_FAILURE);
      break;

    // Outer mobility header
    case 8:
      fprintf (stderr, "ERROR: In next_hdr_to_frag(), data->order[type][data->first_frag[type]-1] = %i is an outer mobility header, which is fragmentable.\n", data->order[type][data->first_frag[type]-1]);
      exit (EXIT_FAILURE);
      break;

    // No next header
    case 9:
      fprintf (stderr, "ERROR: In next_hdr_to_frag(), data->order[type][data->first_frag[type]-1] = %i is an No Next Header.\n", data->order[type][data->first_frag[type]-1]);
      break;

    // Original/Inner IPv6 header
    case 10:
      ip6hdr->ip6_nxt = IPPROTO_FRAGMENT;
      break;

    // Original/Inner hop-by-hop header
    case 11:
      data->hophdr[type][0].nxt_hdr = IPPROTO_FRAGMENT;
      break;

    // Original/Inner destination header (first)
    case 12:
      data->dstfhdr[type][0].nxt_hdr = IPPROTO_FRAGMENT;
      break;

    // Original/Inner routing header
    case 13:
      data->routehdr[type][0].nxt_hdr = IPPROTO_FRAGMENT;
      break;

    // Original/Inner destination header (last)
    case 14:
      fprintf (stderr, "ERROR: In next_hdr_to_frag(), data->order[type][data->first_frag[type]-1] = %i is an Original/Inner destination header (first), which is fragmentable.\n", data->order[type][data->first_frag[type]-1]);
      exit (EXIT_FAILURE);
      break;

    // Original/Inner mobility header
    case 15:
      fprintf (stderr, "ERROR: In next_hdr_to_frag(), data->order[type][data->first_frag[type]-1] = %i is an Original/Inner mobility header, which is fragmentable.\n", data->order[type][data->first_frag[type]-1]);
      exit (EXIT_FAILURE);
      break;

    // TCP header
    case 16:
      fprintf (stderr, "ERROR: In next_hdr_to_frag(), data->order[type][data->first_frag[type]-1] = %i is a TCP header, which is fragmentable.\n", data->order[type][data->first_frag[type]-1]);
      exit (EXIT_FAILURE);
      break;

    // ICMPV6 header
    case 17:
      fprintf (stderr, "ERROR: In next_hdr_to_frag(), data->order[type][data->first_frag[type]-1] = %i is a ICMPv6 header, which is fragmentable.\n", data->order[type][data->first_frag[type]-1]);
      exit (EXIT_FAILURE);
      break;

    // UDP header
    case 18:
      fprintf (stderr, "ERROR: In next_hdr_to_frag(), data->order[type][data->first_frag[type]-1] = %i is a UDP header, which is fragmentable.\n", data->order[type][data->first_frag[type]-1]);
      exit (EXIT_FAILURE);
      break;

    // ESP trailer header + ICV
    case 19:
      fprintf (stderr, "ERROR: In next_hdr_to_frag(), data->order[type][data->first_frag[type]-1] = %i is an ESP trailed header + ICV, which is fragmentable.\n", data->order[type][data->first_frag[type]-1]);
      exit (EXIT_FAILURE);
      break;

    // Unknown index value
    default:
      fprintf (stderr, "ERROR: Unknown index value %i in next_hdr_to_frag().\n", data->order[type][data->first_frag[type]-1]);
      exit (EXIT_FAILURE);
  }

  return (EXIT_SUCCESS);
}

// Find length of unfragmentable portion (IPv6 only).
int
find6_unfraglen (int type, SPSData *data) {

// AH or ESP transport, or no AH or ESP
/* Unfragmentable portion
     10 = Original IPv6 header
     11 = Original/Inner hop-by-hop header
     12 = Original/Inner destination header (first)
     13 = Original/Inner routing header

   Fragmentable portion
      5 = auth header (no esp header)
      6 = esp header (no auth header)
     14 = Original/Inner destination header (last)
     15 = Original/Inner mobility header
      9 = No next header
     16 = TCP header
     17 = ICMPV6 header
     18 = UDP header
     19 = ESP trailer header + ICV
*/

// AH or ESP tunnel
/* Unfragmentable portion
      0 = New IPv6 header (if AH or ESP tunneling)
      1 = Outer hop-by-hop header
      2 = Outer destination header (first)
      3 = Outer routing header

   Fragmentable portion
      5 = auth header (no esp header)
      6 = esp header (no auth header)
      7 = Outer destination header (last)
      8 = Outer mobility header
      9 = No next header
     10 = Original/Inner IPv6 header
     11 = Original/Inner hop-by-hop header
     12 = Original/Inner destination header (first)
     13 = Original/Inner routing header
     14 = Original/Inner destination header (last)
     15 = Original/Inner mobility header
      9 = No next header
     16 = TCP header
     17 = ICMPV6 header
     18 = UDP header
     19 = ESP trailer header + ICV
 */

  int c, len;

  // Set order of all headers, extension headers, and data.
  // This must be called in order to ensure any new items are included in chain length.
  ip6_order (type, data);  // Note: This will set data->first_frag[type]

  c = 0;  // Index of order array (IPv6 header chain)
  len = 0;  // Index of unfragmentable portion

  // Loop though all unfragmentable links in IPv6 chain.
  for (c=0; c<data->first_frag[type]; c++) {

    switch (data->order[type][c]) {

      // New IPv6 header (if AH or ESP tunneling)
      case 0:
        len += IP6_HDRLEN;
        break;

      // Outer hop-by-hop header
      case 1:
        len += HOP_HDRLEN + data->hbh_opt_totlen[type][1] + data->hbh_optpadlen[type][1];
        break;

      // Outer destination header (first)
      case 2:
        len += DST_HDRLEN + data->dstf_opt_totlen[type][1] + data->dstf_optpadlen[type][1];
        break;

      // Outer routing header
      case 3:
        len += RTE_HDRLEN + data->route_datlen[type][1];
        break;

      // Unused
      case 4:
        // Do nothing.
        break;

      // Authentication header
      case 5:
        // This header is always fragmentable. See find6_fraglen().
        break;
  
      // ESP header
      case 6:
        // This header is always fragmentable. See find6_fraglen().
        break;

      // Outer destination header (last)
      case 7:
        // This header is always fragmentable. See find6_fraglen().
        break;

      // Outer mobility header
      case 8:
        // This header is always fragmentable. See find6_fraglen().
        break;

      // Unused
      case 9:
        // Do nothing.
        break;

      // Original/Inner IPv6 header
      case 10:
        // If not AH or ESP tunneling, this header is unfragmentable, otherwise see find6_fraglen().
        if (!(data->auth_hdr_flag[type] && data->auth_tr_tun_flag[type]) &&
            !(data->esp_hdr_flag[type] && data->esp_tr_tun_flag[type])) {

          len += IP6_HDRLEN;
        }
        break;

      // Original/Inner hop-by-hop header
      case 11:
        // If not AH or ESP tunneling, this header is unfragmentable, otherwise see find6_fraglen().
        if (!(data->auth_hdr_flag[type] && data->auth_tr_tun_flag[type]) &&
            !(data->esp_hdr_flag[type] && data->esp_tr_tun_flag[type])) {

          len += HOP_HDRLEN + data->hbh_opt_totlen[type][0] + data->hbh_optpadlen[type][0];
        }
        break;

      // Original/Inner Original/Inner destination header (first)
      case 12:
        // If not AH or ESP tunneling, this header is unfragmentable, otherwise see find6_fraglen().
        if (!(data->auth_hdr_flag[type] && data->auth_tr_tun_flag[type]) &&
            !(data->esp_hdr_flag[type] && data->esp_tr_tun_flag[type])) {

          len += DST_HDRLEN + data->dstf_opt_totlen[type][0] + data->dstf_optpadlen[type][0];
        }
        break;

      // Original/Inner routing header
      case 13:
        // If not AH or ESP tunneling, this header is unfragmentable, otherwise see find6_fraglen().
        if (!(data->auth_hdr_flag[type] && data->auth_tr_tun_flag[type]) &&
            !(data->esp_hdr_flag[type] && data->esp_tr_tun_flag[type])) {

          len += RTE_HDRLEN + data->route_datlen[type][0];
        }
        break;

      // Original/Inner destination header (last)
      case 14:
        // This header is always fragmentable. See find6_fraglen().
        break;

      // Original/Inner mobility header
      case 15:
        // This header is always fragmentable. See find6_fraglen().
        break;

      // TCP header, options, and payload data
      case 16:
        // This header is always fragmentable. See find6_fraglen().
        break;

      // ICMPv6 header and payload data
      case 17:
        // This header is always fragmentable. See find6_fraglen().
        break;

      // UDP header and payload data
      case 18:
        // This header is always fragmentable. See find6_fraglen().
        break;

      // ESP trailer header + ICV
      case 19:
        // This header is always fragmentable. See find6_fraglen().
        break;

      // Unknown index value
      default:
        fprintf (stderr, "ERROR: Unknown index value %i in find6_unfraglen().\n", data->order[type][c]);
        exit (EXIT_FAILURE);

    }  // End switch
  }  // End loop through unfragmentable chain

  return (len);
}

// Returns length of fragmentable portion (IPv6 only).
int
find6_fraglen (int type, SPSData *data) {

// AH or ESP transport, or no AH or ESP
/* Unfragmentable portion
     10 = Original IPv6 header
     11 = Original/Inner hop-by-hop header
     12 = Original/Inner destination header (first)
     13 = Original/Inner routing header

   Fragmentable portion
      5 = auth header (no esp header)
      6 = esp header (no auth header)
     14 = Original/Inner destination header (last)
     15 = Original/Inner mobility header
      9 = No next header
     16 = TCP header
     17 = ICMPV6 header
     18 = UDP header
     19 = ESP trailer header + ICV
*/

// AH or ESP tunnel
/* Unfragmentable portion
      0 = New IPv6 header (if AH or ESP tunneling)
      1 = Outer hop-by-hop header
      2 = Outer destination header (first)
      3 = Outer routing header

   Fragmentable portion
      5 = auth header (no esp header)
      6 = esp header (no auth header)
      7 = Outer destination header (last)
      8 = Outer mobility header
      9 = No next header
     10 = Original/Inner IPv6 header
     11 = Original/Inner hop-by-hop header
     12 = Original/Inner destination header (first)
     13 = Original/Inner routing header
     14 = Original/Inner destination header (last)
     15 = Original/Inner mobility header
      9 = No next header
     16 = TCP header
     17 = ICMPV6 header
     18 = UDP header
     19 = ESP trailer header + ICV
 */

  int c, len;

  // Set order of all headers, extension headers, and data.
  // This must be called in order to ensure any new items are included in chain length.
  ip6_order (type, data);  // Note: This will set data->first_frag[type]

  c = data->first_frag[type];  // Index of IPv6 chain
  len = 0;  // Index of fragmentable portion

  // Loop though all fragmentable links in IPv6 chain.
  while ((data->order[type][c] != 99) && (c < MAX_IP6LINKS)) {

    switch (data->order[type][c]) {

      // New IP6 header (if tunneling AH or ESP)
      case 0:
        // This header is unfragmentable. See find6_unfraglen().
        break;

      // Outer hop-by-hop header
      case 1:
        // This header is unfragmentable. See find6_unfraglen().
        break;

      // Outer destination header (first)
      case 2:
        // This header is unfragmentable. See find6_unfraglen().
        break;

      // Outer routing header
      case 3:
        // This header is unfragmentable. See find6_unfraglen().
        break;

      // Unused
      case 4:
        // Do nothing.
        break;

      // Authentication header
      case 5:
        len += ATH_HDRLEN + data->auth_len[type];
        break;

      // ESP header (8 bytes): (contains only spi and seq; nxt_hdr is in esptail)
      case 6:
        len += ESP_HDRLEN;
        break;

      // Outer destination header (last)
      case 7:
        len += DST_HDRLEN + data->dstl_opt_totlen[type][1] + data->dstl_optpadlen[type][1];
        break;

      // Outer mobility header
      case 8:
        // Not implemented here. Do nothing.
        break;

      // Unused
      case 9:
        // Do nothing.
        break;

      // Original/Inner IPv6 header
      case 10:
        // If not AH or ESP tunneling, this header is unfragmentable. See find6_unfrag().
        if ((data->auth_hdr_flag[type] && data->auth_tr_tun_flag[type]) ||
            (data->esp_hdr_flag[type] && data->esp_tr_tun_flag[type])) {

          len += IP6_HDRLEN;
        }
        break;

      // Original/Inner hop-by-hop header
      case 11:
        // If not AH or ESP tunneling, this header is unfragmentable. See find6_unfrag().
        if ((data->auth_hdr_flag[type] && data->auth_tr_tun_flag[type]) ||
            (data->esp_hdr_flag[type] && data->esp_tr_tun_flag[type])) {

          len += HOP_HDRLEN + data->hbh_opt_totlen[type][0] + data->hbh_optpadlen[type][0];
        }
        break;

      // Original/Inner destination header (first)
      case 12:
        // If not AH or ESP tunneling, this header is unfragmentable. See find6_unfrag().
        if ((data->auth_hdr_flag[type] && data->auth_tr_tun_flag[type]) ||
            (data->esp_hdr_flag[type] && data->esp_tr_tun_flag[type])) {

          len += DST_HDRLEN + data->dstf_opt_totlen[type][0] + data->dstf_optpadlen[type][0];
        }
        break;

      // Original/Inner routing header
      case 13:
        // If not AH or ESP tunneling, this header is unfragmentable. See find6_unfrag().
        if ((data->auth_hdr_flag[type] && data->auth_tr_tun_flag[type]) ||
            (data->esp_hdr_flag[type] && data->esp_tr_tun_flag[type])) {

          len += RTE_HDRLEN + data->route_datlen[type][0];
        }
        break;

      // Original/Inner destination header (last)
      case 14:
        len += DST_HDRLEN + data->dstl_opt_totlen[type][0] + data->dstl_optpadlen[type][0];
        break;

      // Original/Inner mobility header
      case 15:
        // Not implemented here. Do nothing.
        break;

      // TCP header and options
      case 16:
        len += TCP_HDRLEN + data->tcp_opt_totlen[type] + data->tcp_optpadlen[type];

        // TCP payload data
        len += data->payloadlen[type];
        break;

      // ICMPv6 header
      case 17:
        len += ICMP_HDRLEN;

        // ICMP payload data
        len += data->payloadlen[type];
        break;

      // UDP header
      case 18:
        len += UDP_HDRLEN;

        // UDP payload data
        len += data->payloadlen[type];
        break;

      // ESP trailer header + ICV
      // Add padding (Section 2.4 of RFC 4303 and 4305).
      case 19:
        len += data->esptail[type].pad_len;
        len += ESP_TAILLEN;
        len += data->esp_auth_len[type];
      break;

      // Unknown index value
      default:
        fprintf (stderr, "ERROR: Unknown index value %i in find6_fraglen().\n", data->order[type][c]);
        exit (EXIT_FAILURE);

    }  // End switch

    c++;  // Increment chain index.

  }  // Go to next link in chain.

  return (len);
}

// Build unfragmentable portion of ethernet frame (IPv6 only).
int
build_unfrag (int type, struct ip6_hdr *ip6hdr, int *i, uint8_t *unfragbuffer, SPSData *data) {

// AH or ESP transport, or no AH or ESP
/* Unfragmentable portion
     10 = Original IPv6 header
     11 = Original/Inner hop-by-hop header
     12 = Original/Inner destination header (first)
     13 = Original/Inner routing header

   Fragmentable portion
      5 = auth header (no esp header)
      6 = esp header (no auth header)
     14 = Original/Inner destination header (last)
     15 = Original/Inner mobility header
      9 = No next header
     16 = TCP header
     17 = ICMPV6 header
     18 = UDP header
     19 = ESP trailer header + ICV
*/

// AH or ESP tunnel
/* Unfragmentable portion
      0 = New IPv6 header (if AH or ESP tunneling)
      1 = Outer hop-by-hop header
      2 = Outer destination header (first)
      3 = Outer routing header

   Fragmentable portion
      5 = auth header (no esp header)
      6 = esp header (no auth header)
      7 = Outer destination header (last)
      8 = Outer mobility header
      9 = No next header
     10 = Original/Inner IPv6 header
     11 = Original/Inner hop-by-hop header
     12 = Original/Inner destination header (first)
     13 = Original/Inner routing header
     14 = Original/Inner destination header (last)
     15 = Original/Inner mobility header
      9 = No next header
     16 = TCP header
     17 = ICMPV6 header
     18 = UDP header
     19 = ESP trailer header + ICV
 */

  int j, c;

  // i = Index of unfragbuffer

  // Loop though all unfragmentable links in IPv6 chain of headers, options, and payload etc.
  for (c=0; c<data->first_frag[type]; c++) {

    switch (data->order[type][c]) {

      // New IPv6 header (if AH or ESP tunneling)
      case 0:
        memcpy (unfragbuffer + (*i), &data->ip6hdr[type+20], IP6_HDRLEN * sizeof (uint8_t));
        (*i) += IP6_HDRLEN;
        break;

      // Outer hop-by-hop header
      case 1:
        // Copy hop-by-hop extension header (without options) to unfragbuffer.
        memcpy (unfragbuffer + (*i), &data->hophdr[type][1], HOP_HDRLEN * sizeof (uint8_t));
        (*i) += HOP_HDRLEN;

        // Copy hop-by_hop extension header options to unfragbuffer.
        for (j=0; j<data->hbh_nopt[type][1]; j++) {

          // Alignment padding
          memcpy (unfragbuffer + (*i), data->hbh_optleadpad[type][1][j], data->hbh_optleadpadlen[type][1][j] * sizeof (uint8_t));
          (*i) += data->hbh_optleadpadlen[type][1][j];

          // Copy hop-by-hop option to unfragbuffer.
          memcpy (unfragbuffer + (*i), data->hbh_options[type][1][j], data->hbh_optlen[type][1][j] * sizeof (uint8_t));
          (*i) += data->hbh_optlen[type][1][j];
        }

        // Trailing padding
        memcpy (unfragbuffer + (*i), data->hbh_opttailpad[type][1], data->hbh_opttailpadlen[type][1] * sizeof (uint8_t));
        (*i) += data->hbh_opttailpadlen[type][1];
        break;

      // Outer destination header (first)
      case 2:
        // Copy destination extension header (without options) to unfragbuffer.
        memcpy (unfragbuffer + (*i), &data->dstfhdr[type][1], DST_HDRLEN * sizeof (uint8_t));
        (*i) += DST_HDRLEN;

        // Copy destination extension header options to unfragbuffer.
        for (j=0; j<data->dstf_nopt[type][1]; j++) {

          // Alignment padding
          memcpy (unfragbuffer + (*i), data->dstf_optleadpad[type][1][j], data->dstf_optleadpadlen[type][1][j] * sizeof (uint8_t));
          (*i) += data->dstf_optleadpadlen[type][1][j];

          // Copy destination option to unfragbuffer.
          memcpy (unfragbuffer + (*i), data->dstf_options[type][1][j], data->dstf_optlen[type][1][j] * sizeof (uint8_t));
          (*i) += data->dstf_optlen[type][1][j];
        }

        // Trailing padding
        memcpy (unfragbuffer + (*i), data->dstf_opttailpad[type][1], data->dstf_opttailpadlen[type][1] * sizeof (uint8_t));
        (*i) += data->dstf_opttailpadlen[type][1];
        break;

      // Outer routing header
      case 3:
        // Copy routing extension header (without data) to unfragbuffer.
        memcpy (unfragbuffer + (*i), &data->routehdr[type][1], RTE_HDRLEN * sizeof (uint8_t));
        (*i) += RTE_HDRLEN;

        // Copy routing header data to unfragbuffer.
        memcpy (unfragbuffer + (*i), data->route_data[type][1], data->route_datlen[type][1] * sizeof (uint8_t));
        (*i) += data->route_datlen[type][1];
        break;

      // Unused
      case 4:
        // Do nothing.
        break;

      // Authentication header
      case 5:
        // This header is fragmentable. See build_frag().
        break;

      // ESP header
      case 6:
        // This header is fragmentable. See build_frag().
        break;

      // Outer destination header (last)
      case 7:
        // This header is fragmentable. See build_frag().
        break;

      // Outer mobility header
      case 8:
        // This header is fragmentable. See build_frag().
        break;

      // No next header
      case 9:
        // Do nothing.
        break;

      // Original/Inner IPv6 header
      case 10:
        // If not AH or ESP tunneling, this header is unfragmentable, othewrwise see build_frag().
        if (!(data->auth_hdr_flag[type] && data->auth_tr_tun_flag[type]) &&
            !(data->esp_hdr_flag[type] && data->esp_tr_tun_flag[type])) {

          memcpy (unfragbuffer + (*i), ip6hdr, IP6_HDRLEN * sizeof (uint8_t));
          (*i) += IP6_HDRLEN;
        }
        break;

      // Original/Inner hop-by-hop header
      case 11:
        // If not AH or ESP tunneling, this header is unfragmentable, otherwise see build_unfrag().
        if (!(data->auth_hdr_flag[type] && data->auth_tr_tun_flag[type]) &&
            !(data->esp_hdr_flag[type] && data->esp_tr_tun_flag[type])) {

          // Copy hop-by-hop extension header (without options) to unfragbuffer.
          memcpy (unfragbuffer + (*i), &data->hophdr[type][0], HOP_HDRLEN * sizeof (uint8_t));
          (*i) += HOP_HDRLEN;

          // Copy hop-by_hop extension header options to unfragbuffer.
          for (j=0; j<data->hbh_nopt[type][0]; j++) {

            // Alignment padding
            memcpy (unfragbuffer + (*i), data->hbh_optleadpad[type][0][j], data->hbh_optleadpadlen[type][0][j] * sizeof (uint8_t));
            (*i) += data->hbh_optleadpadlen[type][0][j];

            // Copy hop-by-hop option to unfragbuffer.
            memcpy (unfragbuffer + (*i), data->hbh_options[type][0][j], data->hbh_optlen[type][0][j] * sizeof (uint8_t));
            (*i) += data->hbh_optlen[type][0][j];
          }

          // Trailing padding
          memcpy (unfragbuffer + (*i), data->hbh_opttailpad[type][0], data->hbh_opttailpadlen[type][0] * sizeof (uint8_t));
          (*i) += data->hbh_opttailpadlen[type][0];
        }
        break;

      // Original/Inner destination header (first)
      case 12:
        // If not AH or ESP tunneling, this header is unfragmentable, otherwise see build_unfrag().
        if (!(data->auth_hdr_flag[type] && data->auth_tr_tun_flag[type]) &&
            !(data->esp_hdr_flag[type] && data->esp_tr_tun_flag[type])) {

          // Copy destination extension header (without options) to unfragbuffer.
          memcpy (unfragbuffer + (*i), &data->dstfhdr[type][0], DST_HDRLEN * sizeof (uint8_t));
          (*i) += DST_HDRLEN;

          // Copy destination extension header options to unfragbuffer.
          for (j=0; j<data->dstf_nopt[type][0]; j++) {

            // Alignment padding
            memcpy (unfragbuffer + (*i), data->dstf_optleadpad[type][0][j], data->dstf_optleadpadlen[type][0][j] * sizeof (uint8_t));
            (*i) += data->dstf_optleadpadlen[type][0][j];

            // Copy destination option to unfragbuffer.
            memcpy (unfragbuffer + (*i), data->dstf_options[type][0][j], data->dstf_optlen[type][0][j] * sizeof (uint8_t));
            (*i) += data->dstf_optlen[type][0][j];
          }

          // Trailing padding
          memcpy (unfragbuffer + (*i), data->dstf_opttailpad[type][0], data->dstf_opttailpadlen[type][0] * sizeof (uint8_t));
          (*i) += data->dstf_opttailpadlen[type][0];
        }
        break;

      // Original/Inner routing header
      case 13:
        // If not AH or ESP tunneling, this header is unfragmentable, otherwise see build_unfrag().
        if (!(data->auth_hdr_flag[type] && data->auth_tr_tun_flag[type]) &&
            !(data->esp_hdr_flag[type] && data->esp_tr_tun_flag[type])) {

          // Copy routing extension header (without data) to unfragbuffer.
          memcpy (unfragbuffer + (*i), &data->routehdr[type][0], RTE_HDRLEN * sizeof (uint8_t));
          (*i) += RTE_HDRLEN;

          // Copy routing header data to unfragbuffer.
          memcpy (unfragbuffer + (*i), data->route_data[type][0], data->route_datlen[type][0] * sizeof (uint8_t));
          (*i) += data->route_datlen[type][0];
        }
        break;

      // Original/Inner destination header (last)
      case 14:
        // This header is fragmentable. See build_frag().
        break;

      // Original/Inner mobility header
      case 15:
        // This header is fragmentable. See build_frag().
        break;

      // TCP header
      case 16:
        // This header is fragmentable. See build_frag().
        break;

      // ICMPv6 header
      case 17:
        // This header is fragmentable. See build_frag().
        break;

      // UDP header
      case 18:
        // This header is fragmentable. See build_frag().
        break;

      // ESP trailer header + ICV
      case 19:
        // This header is fragmentable. See build_frag().
        break;

      // Unknown index value
      default:
        fprintf (stderr, "ERROR: Unknown index value %i in build_unfrag().\n", data->order[type][c]);
        exit (EXIT_FAILURE);
    }
  }  // End loop through unfragmentable chain

  return (EXIT_SUCCESS);
}

// Build fragmentable portion of ethernet frame (IPv6 only).
int
build_frag (int type, struct ip6_hdr *ip6hdr, uint8_t *fragbuffer, SPSData *data) {

// AH or ESP transport, or no AH or ESP
/* Unfragmentable portion
     10 = Original IPv6 header
     11 = Original/Inner hop-by-hop header
     12 = Original/Inner destination header (first)
     13 = Original/Inner routing header

   Fragmentable portion
      5 = auth header (no esp header)
      6 = esp header (no auth header)
     14 = Original/Inner destination header (last)
     15 = Original/Inner mobility header
      9 = No next header
     16 = TCP header
     17 = ICMPV6 header
     18 = UDP header
     19 = ESP trailer header + ICV
*/

// AH or ESP tunnel
/* Unfragmentable portion
      0 = New IPv6 header (if AH or ESP tunneling)
      1 = Outer hop-by-hop header
      2 = Outer destination header (first)
      3 = Outer routing header

   Fragmentable portion
      5 = auth header (no esp header)
      6 = esp header (no auth header)
      7 = Outer destination header (last)
      8 = Outer mobility header
      9 = No next header
     10 = Original/Inner IPv6 header
     11 = Original/Inner hop-by-hop header
     12 = Original/Inner destination header (first)
     13 = Original/Inner routing header
     14 = Original/Inner destination header (last)
     15 = Original/Inner mobility header
      9 = No next header
     16 = TCP header
     17 = ICMPV6 header
     18 = UDP header
     19 = ESP trailer header + ICV
 */

  int i, j, c;

  c = data->first_frag[type];  // Index of IPv6 chain; set to first element of fragmentable portion.
  i = 0;  // Index of fragbuffer.

  // Loop though all fragmentable links in IPv6 chain of headers, options, and payload etc.
  while ((data->order[type][c] != 99) && (c < MAX_IP6LINKS)) {

    switch (data->order[type][c]) {

      // New IPv6 header (if AH or ESP tunneling)
      case 0:
        // This header is unfragmentable. See build_unfrag().
        break;

      // Outer hop-by-hop header
      case 1:
        // This header is unfragmentable. See build_unfrag().
        break;

      // Outer destination header (first)
      case 2:
        // This header is unfragmentable. See build_unfrag().
        break;

      // Outer routing header
      case 3:
        // This header is unfragmentable. See build_unfrag().
        break;

      // Unused
      case 4:
        // Do nothing.
        break;

      // Authentication header (no ESP header)
      case 5:
        memcpy (fragbuffer + i, &data->authhdr[type], ATH_HDRLEN * sizeof (uint8_t));  // Authentication header, excluding authentication data
        i += ATH_HDRLEN;
        memcpy (fragbuffer + i, data->auth_data[type], data->auth_len[type] * sizeof (uint8_t));  // Authentication data
        i += data->auth_len[type];
        break;

      // ESP header (no AH header); includes only spi and seq (8 bytes)
      case 6:
        memcpy (fragbuffer + i, &data->esphdr[type], ESP_HDRLEN * sizeof (uint8_t));
        i += ESP_HDRLEN;
        break;

      // Outer destination header (last)
      case 7:
        // Copy destination header (without options) to fragbuffer.
        memcpy (fragbuffer + i, &data->dstlhdr[type][1], DST_HDRLEN * sizeof (uint8_t));
        i += DST_HDRLEN;

        // Copy destination header options to fragbuffer.
        for (j=0; j<data->dstl_nopt[type][1]; j++) {

          // Alignment padding
          memcpy (fragbuffer + i, data->dstl_optleadpad[type][1][j], data->dstl_optleadpadlen[type][1][j] * sizeof (uint8_t));
          i += data->dstl_optleadpadlen[type][1][j];

          // Copy destination option to fragbuffer.
          memcpy (fragbuffer + i, data->dstl_options[type][1][j], data->dstl_optlen[type][1][j] * sizeof (uint8_t));
          i += data->dstl_optlen[type][1][j];
        }

        // Trailing padding
        memcpy (fragbuffer + i, data->dstl_opttailpad[type][1], data->dstl_opttailpadlen[type][1] * sizeof (uint8_t));
        i += data->dstl_opttailpadlen[type][1];
        break;

      // Outer mobility header
      case 8:
        // Not implemented here. Do nothing.
        break;

      // No next header
      case 9:
        // Do nothing.
        break;

      // Original/Inner IPv6 header
      case 10:
        memcpy (fragbuffer + i, ip6hdr, IP6_HDRLEN * sizeof (uint8_t));
        i += IP6_HDRLEN;
        break;

      // Original/Inner hop-by-hop header
      case 11:
        // Unless AH or ESP tunneling, this header is unfragmentable. See build_unfrag().
        if ((data->auth_hdr_flag[type] && data->auth_tr_tun_flag[type]) ||
              (data->esp_hdr_flag[type] && data->esp_tr_tun_flag[type])) {

          // Copy hop-by-hop extension header (without options) to fragbuffer.
          memcpy (fragbuffer + i, &data->hophdr[type][0], HOP_HDRLEN * sizeof (uint8_t));
          i += HOP_HDRLEN;

          // Copy hop-by_hop extension header options to fragbuffer.
          for (j=0; j<data->hbh_nopt[type][0]; j++) {

            // Alignment padding
            memcpy (fragbuffer + i, data->hbh_optleadpad[type][0][j], data->hbh_optleadpadlen[type][0][j] * sizeof (uint8_t));
            i += data->hbh_optleadpadlen[type][0][j];

            // Copy hop-by-hop option to fragbuffer.
            memcpy (fragbuffer + i, data->hbh_options[type][0][j], data->hbh_optlen[type][0][j] * sizeof (uint8_t));
            i += data->hbh_optlen[type][0][j];
          }

          // Trailing padding
          memcpy (fragbuffer + i, data->hbh_opttailpad[type][0], data->hbh_opttailpadlen[type][0] * sizeof (uint8_t));
          i += data->hbh_opttailpadlen[type][0];
        }
        break;

      // Original/Inner destination header (first)
      case 12:
        // Unless AH or ESP tunneling, this header is unfragmentable. See build_unfrag().
        if ((data->auth_hdr_flag[type] && data->auth_tr_tun_flag[type]) ||
              (data->esp_hdr_flag[type] && data->esp_tr_tun_flag[type])) {

          // Copy destination extension header (without options) to fragbuffer.
          memcpy (fragbuffer + i, &data->dstfhdr[type][0], DST_HDRLEN * sizeof (uint8_t));
          i += DST_HDRLEN;

          // Copy destination extension header options to fragbuffer.
          for (j=0; j<data->dstf_nopt[type][0]; j++) {

            // Alignment padding
            memcpy (fragbuffer + i, data->dstf_optleadpad[type][0][j], data->dstf_optleadpadlen[type][0][j] * sizeof (uint8_t));
            i += data->dstf_optleadpadlen[type][0][j];

            // Copy destination option to fragbuffer.
            memcpy (fragbuffer + i, data->dstf_options[type][0][j], data->dstf_optlen[type][0][j] * sizeof (uint8_t));
            i += data->dstf_optlen[type][0][j];
          }

          // Trailing padding
          memcpy (fragbuffer + i, data->dstf_opttailpad[type][0], data->dstf_opttailpadlen[type][0] * sizeof (uint8_t));
          i += data->dstf_opttailpadlen[type][0];
        }
        break;

      // Original/Inner routing header
      case 13:
        // Unless AH or ESP tunneling, this header is unfragmentable. See build_unfrag().
        if ((data->auth_hdr_flag[type] && data->auth_tr_tun_flag[type]) ||
              (data->esp_hdr_flag[type] && data->esp_tr_tun_flag[type])) {

          // Copy routing extension header (without data) to fragbuffer.
          memcpy (fragbuffer + i, &data->routehdr[type][0], RTE_HDRLEN * sizeof (uint8_t));
          i += RTE_HDRLEN;

          // Copy routing header data to fragbuffer.
          memcpy (fragbuffer + i, data->route_data[type][0], data->route_datlen[type][0] * sizeof (uint8_t));
          i += data->route_datlen[type][0];
        }
        break;

      // Original/Inner destination header (last)
      case 14:
        // Copy destination header (without options) to fragbuffer.
        memcpy (fragbuffer + i, &data->dstlhdr[type][0], DST_HDRLEN * sizeof (uint8_t));
        i += DST_HDRLEN;

        // Copy destination header options to fragbuffer.
        for (j=0; j<data->dstl_nopt[type][0]; j++) {

          // Alignment padding
          memcpy (fragbuffer + i, data->dstl_optleadpad[type][0][j], data->dstl_optleadpadlen[type][0][j] * sizeof (uint8_t));
          i += data->dstl_optleadpadlen[type][0][j];

          // Copy destination option to fragbuffer.
          memcpy (fragbuffer + i, data->dstl_options[type][0][j], data->dstl_optlen[type][0][j] * sizeof (uint8_t));
          i += data->dstl_optlen[type][0][j];
        }

        // Trailing padding
        memcpy (fragbuffer + i, data->dstl_opttailpad[type][0], data->dstl_opttailpadlen[type][0] * sizeof (uint8_t));
        i += data->dstl_opttailpadlen[type][0];
        break;

      // Original/Inner mobility header
      case 15:
        // Not implemented here. Do nothing.
        break;

      // TCP header and options
      case 16:
        memcpy (fragbuffer + i, &data->tcphdr[type], TCP_HDRLEN * sizeof (uint8_t));
        i += TCP_HDRLEN;
        for (j=0; j<data->tcp_nopt[type]; j++) {
          memcpy (fragbuffer + i, data->tcp_options[type][j], data->tcp_optlen[type][j] * sizeof (uint8_t));
          i += data->tcp_optlen[type][j];
        }
        i += data->tcp_optpadlen[type];  // TCP options padding

        // TCP payload data
        memcpy (fragbuffer + i, data->payload[type], data->payloadlen[type] * sizeof (uint8_t));
        i += data->payloadlen[type];

        break;

      // ICMP header
      case 17:
        memcpy (fragbuffer + i, &data->icmp6hdr[type], ICMP_HDRLEN * sizeof (uint8_t));
        i += ICMP_HDRLEN;

        // ICMP payload data
        memcpy (fragbuffer + i, data->payload[type], data->payloadlen[type] * sizeof (uint8_t));
        i += data->payloadlen[type];

        break;

      // UDP header
      case 18:
        memcpy (fragbuffer + i, &data->udphdr[type], UDP_HDRLEN * sizeof (uint8_t));
        i += UDP_HDRLEN;

        // UDP payload data
        memcpy (fragbuffer + i, data->payload[type], data->payloadlen[type] * sizeof (uint8_t));
        i += data->payloadlen[type];
        break;

      // ESP trailer header + ICV
      case 19:
        memcpy (fragbuffer + i, &data->esp_pad[type], data->esptail[type].pad_len * sizeof (uint8_t));  // ESP payload padding
        i += data->esptail[type].pad_len;
        memcpy (fragbuffer + i, &data->esptail[type], ESP_TAILLEN * sizeof (uint8_t));  // ESP trailer
        i += ESP_TAILLEN;
        memcpy (fragbuffer + i, data->esp_auth_data[type], data->esp_auth_len[type] * sizeof (uint8_t));  // Authentication data (ICV)
        i += data->esp_auth_len[type];
        break;

      // Unknown index value
      default:
        fprintf (stderr, "ERROR: Unknown index value %i in build_frag().\n", data->order[type][c]);
        exit (EXIT_FAILURE);
    }

    // Go to next link in chain.
    c++;

  }  // End loop through fragmentable chain

  return (EXIT_SUCCESS);
}

// Setup chain of header, extension headers, and data by setting Next Header fields (IPv6 only).
// Assume no fragmentation here.
int
ip6_chain (int type, struct ip6_hdr *ip6hdr, SPSData *data) {

  // See case 10 below regarding use of ip6hdr versus data->ip6hdr.

  int c;

  /*  0 = New IPv6 header
      1 = Outer hop-by-hop header
      2 = Outer destination header (first)
      3 = Outer routing header
      4 = unused
      5 = auth header
      6 = esp header
      7 = Outer destination header (last)
      8 = Outer mobility header (RFC 3775: Next Header should be IPPROTO_NONE)
      9 = No next header
     10 = Original/Inner IPv6 header
     11 = Original/Inner hop-by-hop header
     12 = Original/Inner destination header (first)
     13 = Original/Inner routing header
     14 = Original/Inner destination header (last)
     15 = Original/Inner mobility header (RFC 3775: Next Header should be IPPROTO_NONE)
     16 = TCP header
     17 = ICMPV6 header
     18 = UDP header
     19 = ESP trailer header + ICV
 */

  c = 0;  // Index of order array (IPv6 header chain)
  while (data->order[type][c] != 99) {

    switch (data->order[type][c]) {

      // New IPv6 header (AH or ESP tunneling)
      case 0:
        data->ip6hdr[type+20].ip6_nxt = proto_val (data->order[type][c+1]);
        break;

      // Outer hop-by-hop header
      case 1:
        data->hophdr[type][1].nxt_hdr = proto_val (data->order[type][c+1]);
        break;

      // Outer destination header (first)
      case 2:
        data->dstfhdr[type][1].nxt_hdr = proto_val (data->order[type][c+1]);
        break;

      // Outer routing header
      case 3:
        data->routehdr[type][1].nxt_hdr = proto_val (data->order[type][c+1]);
        break;

      // Unused
      case 4:
        // Do nothing.
        break;

      // AH header
      case 5:
        data->authhdr[type].nxt_hdr = proto_val (data->order[type][c+1]);
        break;

      // ESP header; contains only spi and seq; nxt_hdr is in esptail
      case 6:
        // Do nothing since contains only spi and seq; nxt_hdr is in esptail.
        // There is no hxt_hdr to set.
        break;

      // Outer destination header (last)
      case 7:
        data->dstlhdr[type][1].nxt_hdr = proto_val (data->order[type][c+1]);
        break;

      // Outer mobility header
      case 8:
        // Not implemented here. Do nothing.
        break;

      // No next header
      case 9:
        // Do nothing.
        break;

      // Original/Inner IPv6 header
      // We don't change data->ip6hdr[type].ip6_nxt because that will make the upper layer protocol checksum incorrect.
      case 10:
        ip6hdr->ip6_nxt = proto_val (data->order[type][c+1]);
        break;

      // Original/Inner hop-by-hop header
      case 11:
        data->hophdr[type][0].nxt_hdr = proto_val (data->order[type][c+1]);
        break;

      // Original/Inner destination header (first)
      case 12:
        data->dstfhdr[type][0].nxt_hdr = proto_val (data->order[type][c+1]);
        break;

      // Original/Inner routing header
      case 13:
        data->routehdr[type][0].nxt_hdr = proto_val (data->order[type][c+1]);
        break;

      // Original/Inner destination header (last)
      case 14:
        data->dstlhdr[type][0].nxt_hdr = proto_val (data->order[type][c+1]);
        break;

      // Original/Inner mobility header
      case 15:
        // Not implemented. Do nothing.
        break;

      // TCP header
      // Upper level protocol does not link to further extension header; has no nxt_hdr field.
      case 16:
        // Do nothing. Note that an ESP tail + ICV could follow.
        break;

      // ICMP header
      // Upper level protocol does not link to further extension header; has no nxt_hdr field.
      case 17:
        // Do nothing. Note that an ESP tail + ICV could follow.
        break;

      // UDP header
      // Upper level protocol does not link to further extension header; has no nxt_hdr field.
      case 18:
        // Do nothing. Note that an ESP tail + ICV could follow.
        break;

      // ESP header tail (esphdr only contains spi and seq esptail contains pad_len and nxt_hdr)
      // In transport mode, nxt_hdr is last destination header (if present), or upper layer protocol.
      case 19:
        // Transport mode
        if (!data->esp_tr_tun_flag[type]) {
          if (data->dstl_hdr_flag[type][0]) {
            data->esptail[type].nxt_hdr = IPPROTO_DSTOPTS;
          } else if ((type == 3) || (type == 12)) {
            data->esptail[type].nxt_hdr = IPPROTO_TCP;
          } else if ((type == 4) || (type == 13)) {
            data->esptail[type].nxt_hdr = IPPROTO_ICMPV6;
          } else if ((type == 4) || (type == 13)) {
            data->esptail[type].nxt_hdr = IPPROTO_UDP;
          } else {
            fprintf (stderr, "ERROR: ip6_chain() has unknown packet type of %i.\n", type);
            exit (EXIT_FAILURE);
          }

        // Tunnel mode
        // In tunnel mode, hxt_hdr is IPPROTO_IPV6.
        } else {
          data->esptail[type].nxt_hdr = IPPROTO_IPV6;
        }
        break;

      default:
        fprintf (stderr, "ERROR: ip6_chain() has received invalid header value of %i.\n", data->order[type][c]);
        exit (EXIT_FAILURE);

    }  // End switch

    c++;  // Go to next link in chain.

  }  // End while

  return (EXIT_SUCCESS);
}

// Set order of all headers, extension headers, and data (IPv6 only).
// Determine where in chain to separate unfragmentable from fragmentable.
int
ip6_order (int type, SPSData *data) {

  int c;

  // Clear order array.
  memset (data->order[type], 0, MAX_IP6LINKS * sizeof (int));

  /*  0 = New IPv6 header
      1 = Outer hop-by-hop header
      2 = Outer destination header (first)
      3 = Outer routing header
      4 = unused
      5 = auth header
      6 = esp header
      7 = Outer destination header (last)
      8 = Outer mobility header (RFC 3775: Next Header should be IPPROTO_NONE)
      9 = No next header
     10 = Original/Inner IPv6 header
     11 = Original/Inner hop-by-hop header
     12 = Original/Inner destination header (first)
     13 = Original/Inner routing header
     14 = Original/Inner destination header (last)
     15 = Original/Inner mobility header (RFC 3775: Next Header should be IPPROTO_NONE)
     16 = TCP header
     17 = ICMPV6 header
     18 = UDP header
     19 = ESP trailer header + ICV
 */

  // *****************************************************************
  // NOTE: Here we choose not to allow both AH and ESP on same packet.
  //       ESP is generally considered the standard, since it provides
  //       both authentication and encryption.
  // *****************************************************************

  c = 0;  // Index of order array (IPv6 header chain)

  // No AH or ESP header, or AH or ESP transport mode
  if ((!data->auth_tr_tun_flag[type]) && (!data->esp_tr_tun_flag[type])) {

    // Original IPv6 header (mandatory)
    data->order[type][c] = 10;  // Original IPv6 header
    c++;

    // Original/Inner hop-by-hop header
    if (data->hbh_hdr_flag[type][0]) {
      data->order[type][c] = 11;  // Original/Inner hop-by-hop header
      c++;
    }

    // Original/Inner destination header (first)
    if (data->dstf_hdr_flag[type][0]) {
      data->order[type][c] = 12;  // Original/Inner destination header (first)
      c++;
    }

    // Original/Inner routing header
    if (data->route_hdr_flag[type][0]) {
      data->order[type][c] = 13;  // Original/Inner routing header
      c++;
    }

    // Take note of the start of the fragmentable portion.
    data->first_frag[type] = c;

    // Authentication header (no ESP header)
    if (data->auth_hdr_flag[type]) {
      data->order[type][c] = 5;  // Authentication header
      c++;

    // ESP header (no AH header)
    } else if (data->esp_hdr_flag[type]) {
      data->order[type][c] = 6;  // ESP header
      c++;
    }

    // Original/Inner destination header (last)
    if (data->dstl_hdr_flag[type][0]) {
      data->order[type][c] = 14;  // Original/Inner destination header (last)
      c++;
    }

    // Original/Inner mobility header
    // Not implemented here. Do nothing.
//  if (data->mob_hdr_flag[0]) {
//    data->order[type][c] = 15;  // Original/Inner mobility header
//    c++;
//  }

    // TCP header, options, and payload data
    if ((type == 3) || (type == 12)) {
      data->order[type][c] = 16;  // TCP header + payload data (if any)
      c++;

    // ICMPv6 header and payload data
    } else if ((type == 4) || (type == 13)) {
      data->order[type][c] = 17;  // ICMPv6 header + payload data (if any)
      c++;

    // UDP header and payload data
    } else if ((type == 5) || (type == 14)) {
      data->order[type][c] = 18;  // UDP header + payload data (if any)
      c++;
    }

    // ESP trailer + ICV
    if (data->esp_hdr_flag[type]) {
      data->order[type][c] = 19;  // ESP trailer header + ICV
      c++;
    }

  // AH or ESP tunnel mode
  } else if ((data->auth_hdr_flag[type] && data->auth_tr_tun_flag[type]) ||
      (data->esp_hdr_flag[type] && data->esp_tr_tun_flag[type])) {

    // New IPv6 header (mandatory)
    data->order[type][c] = 0;  // New IPv6 header
    c++;

    // Outer hop-by-hop header
    if (data->hbh_hdr_flag[type][1]) {
      data->order[type][c] = 1;  // Outer hop-by-hop header
      c++;
    }

    // Outer destination header (first)
    if (data->dstf_hdr_flag[type][1]) {
      data->order[type][c] = 2;  // Outer destination header (first)
      c++;
    }

    // Outer routing header
    if (data->route_hdr_flag[type][1]) {
      data->order[type][c] = 3;  // Outer routing header
      c++;
    }

    // Take note of the start of the fragmentable portion.
    data->first_frag[type] = c;

    // Authentication header (no ESP header)
    if (data->auth_hdr_flag[type]) {
      data->order[type][c] = 5;  // Authentication header
      c++;

    // ESP header (no AH header)
    } else if (data->esp_hdr_flag[type]) {
      data->order[type][c] = 6;  // ESP header
      c++;
    }

    // Outer destination header (last)
    if (data->dstl_hdr_flag[type][1]) {
      data->order[type][c] = 7;  // Outer destination header (last)
      c++;
    }

    // Outer mobility header
    // Not implemented. Do nothing.
//  if (data->mob_hdr_flag[1]) {
//    data->order[type][c] = 8;  // Outer mobility header
//    c++;
//  }

    // Original/Inner IPv6 header (mandatory)
    data->order[type][c] = 10;  // Original/Inner IPv6 header
    c++;

    // Original/Inner hop-by-hop header
    if (data->hbh_hdr_flag[type][0]) {
      data->order[type][c] = 11;  // Original/Inner hop-by-hop header
      c++;
    }

    // Original/Inner destination header (first)
    if (data->dstf_hdr_flag[type][0]) {
      data->order[type][c] = 12;  // Original/Inner destination header (first)
      c++;
    }

    // Original/Inner routing header
    if (data->route_hdr_flag[type][0]) {
      data->order[type][c] = 13;  // Original/Inner routing header
      c++;
    }

    // Original/Inner destination header (last)
    if (data->dstl_hdr_flag[type][0]) {
      data->order[type][c] = 14;  // Original/Inner destination header (last)
      c++;
    }

    // Original/Inner mobility header
    // Not implemented. Do nothing.
//  if (data->mob_hdr_flag[0]) {
//    data->order[type][c] = 15;  // Original/Inner mobility header
//    c++;
//  }

    // TCP header, options, and payload data
    if ((type == 3) || (type == 12)) {
      data->order[type][c] = 16;  // TCP header + payload data (if any)
      c++;

    // ICMPv6 header and payload data
    } else if ((type == 4) || (type == 13)) {
      data->order[type][c] = 17;  // ICMPv6 header + payload data (if any)
      c++;

    // UDP header and payload data
    } else if ((type == 5) || (type == 14)) {
      data->order[type][c] = 18;  // UDP header + payload data (if any)
      c++;
    }

    // ESP trailer + ICV
    if (data->esp_hdr_flag[type]) {
      data->order[type][c] = 19;  // ESP trailer header + ICV
      c++;
    }

  }  // End if transport or tunneling

  // End of chain
  data->order[type][c] = 99;  // End of chain marker

  return (EXIT_SUCCESS);
}


// Return appropriate protocol type value (IPPROTO), depending upon
// SPS header type index (IPv6 only). See SPS header type index values in table below.
int
proto_val (int val)
{
  // SPS header type index values (IPv6 only)
  /*  0 = New IPv6 header
      1 = Outer hop-by-hop header
      2 = Outer destination header (first)
      3 = Outer routing header
      4 = unused
      5 = auth header
      6 = esp header
      7 = Outer destination header (last)
      8 = Outer mobility header (RFC 3775: Next Header should be IPPROTO_NONE)
      9 = No next header
     10 = Original/Inner IPv6 header
     11 = Original/Inner hop-by-hop header
     12 = Original/Inner destination header (first)
     13 = Original/Inner routing header
     14 = Original/Inner destination header (last)
     15 = Original/Inner mobility header (RFC 3775: Next Header should be IPPROTO_NONE)
     16 = TCP header
     17 = ICMPV6 header
     18 = UDP header
     19 = ESP trailer header + ICV
 */

  switch (val) {
    case 0:
      return (IPPROTO_IPV6);  // 41
    case 1:
      return (IPPROTO_HOPOPTS);  // 0
    case 2:
      return (IPPROTO_DSTOPTS);  // 60
    case 3:
      return (IPPROTO_ROUTING);  // 43
    case 4:
      fprintf (stderr, "ERROR: proto_val() has received a value of %i, which is unused.\n", val);
      exit (EXIT_FAILURE);
    case 5:
      return (IPPROTO_AH);  // 51
    case 6:
      return (IPPROTO_ESP);  // 50
    case 7:
      return (IPPROTO_DSTOPTS);  // 60
    case 8:
      return (IPPROTO_MH);  // 135
    case 9:
      return (IPPROTO_NONE);  // 59
    case 10:
      return (IPPROTO_IPV6);  // 41
    case 11:
      return (IPPROTO_HOPOPTS);  // 0
    case 12:
      return (IPPROTO_DSTOPTS);  // 60
    case 13:
      return (IPPROTO_ROUTING);  // 43
    case 14:
      return (IPPROTO_DSTOPTS);  // 60
    case 15:
      return (IPPROTO_MH);  // 135
    case 16:
      return (IPPROTO_TCP);  // 6
    case 17:
      return (IPPROTO_ICMPV6);  // 58
    case 18:
      return (IPPROTO_UDP);  // 17
    case 19:
      fprintf (stderr, "ERROR: proto_val() has received a value of %i (ESP trailer). This should not happen.\n", val);
      exit (EXIT_FAILURE);
    default:
      fprintf (stderr, "ERROR: Unknown index value %i in proto_val().\n", val);
      exit (EXIT_FAILURE);
  }
}

// Determine required number of ethernet frames, frame lengths, and data offsets (IPv4 or IPv6).
int
find_nframes (int type, int mtu, int unfraglen, int fraglen, int *nframes, int *len, int *offset) {

  int i, c, totlen, frag_hdrlen;

  // Total length of ethernet frame (excludes ethernet header).
  totlen = unfraglen + fraglen;

  // Clear all fragment lengths and offsets.
  memset (len, 0, MAX_FRAGS * sizeof (int));
  memset (offset, 0, MAX_FRAGS * sizeof (int));

  // No fragmentation required if total length is less than or equal to MTU.
  if (totlen <= mtu) {
    *nframes = 1;
    offset[0] = 0;
    len[0] = fraglen;
    return (EXIT_SUCCESS);
  }

  // Fragmentation required.
  // If IPv6, we need to allow for a fragmentation extension header.
  if ((type == 3) || (type ==4) || (type == 5) ||
      (type == 12) || (type == 13) || (type == 14)) {
    frag_hdrlen = FRG_HDRLEN;
  } else {
    frag_hdrlen = 0;
  }

  // Determine number of frames.
  i = 0;  // i is frame number.
  c = 0;  // c is index of fragmentable portion of packet.
  // len[i] is amount of fragmentable part we can include in frame i.
  while (c < fraglen) {

    // Do we still need to fragment remainder of fragmentable portion?

    // Yes, need to fragment more.
    if ((fraglen - c) > (mtu - unfraglen - frag_hdrlen)) {

      if (i > MAX_FRAGS) {
        fprintf (stderr, "ERROR: In find_nframes(), the number of frames i has now exceeded the maximum number of allowed fragments MAX_FRAGS (%i)\n", MAX_FRAGS);
        fprintf (stderr, "       mtu: %i, unfraglen: %i, fraglen: %i, frag_hdrlen: %i\n", mtu, unfraglen, fraglen, frag_hdrlen);
        exit (EXIT_FAILURE);
      }

      len[i] = mtu - unfraglen - frag_hdrlen;

      if (len[i] <= 0) {
        fprintf (stderr, "ERROR: In find_nframes(), len[i] <= 0\n");
        fprintf (stderr, "       len[i] = mtu (%i) - unfraglen (%i) - frag_hdrlen (%i) = %i\n", mtu, unfraglen, frag_hdrlen, len[i]);
        exit (EXIT_FAILURE);
      }

    // No, don't need to fragment any more.
    } else {
      len[i] = fraglen - c;
    }
    c += len[i];

    // If not last fragment, make sure we have an even number of 8-byte blocks.
    // Reduce length as necessary. Section Section 3.2 of RFC 791, and Section 4.5 of RFC 8200.
    if (c < (fraglen - 1)) {
      while ((len[i] % 8) > 0) {
        len[i]--;
        c--;
      }
    }

    i++;  // Increment number of frames.
    offset[i] = (len[i-1] / 8) + offset[i-1];  // Offset of this fragment
  }

  // Total number of ethernet frames.
  *nframes = i;

  return (EXIT_SUCCESS);
}

// Find length of unfragmentable portion (IPv4 only).
int
find4_unfraglen (int type, SPSData *data) {

  int len;

  len = IP4_HDRLEN + data->ip_opt_totlen[type] + data->ip_optpadlen[type];

  return (len);
}

// Find length of fragmentable portion (IPv4 only).
int
find4_fraglen (int type, SPSData *data) {

  int len;

  // Upper layer protocol header length.
  if ((type == 0) || (type == 9)) {
    len = TCP_HDRLEN;

  } else if ((type == 1) || (type == 10)) {
    len = ICMP_HDRLEN;

  } else if ((type == 2) || (type == 11)) {
    len = UDP_HDRLEN;
  }

  // Add TCP options (if appropriate).
  if ((type == 0) || (type == 9)) {
    len += data->tcp_opt_totlen[type] + data->tcp_optpadlen[type];
  }

  // Upper protocol payload data
  len += data->payloadlen[type];

  return (len);
}
