/*  Simple Packet Sender (SPS) - a linux packet crafting network
    tool using GTK+ graphical interface.

    Copyright (C) 2011-2023  Hohlraum (h0h1r4um@gmail.com)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "main.h"

// IP version (4 bits)
int
ip6_ver_entry (GtkWidget *entry, int type, SPSData *data) {

  const char *entry_text;
  uint32_t old;
  GtkStyleContext *context;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (entry);

  // Get new value from text entry.
  entry_text = gtk_entry_get_text (GTK_ENTRY (entry));

  // Save value currently in header.
  old = ntohl (data->ip6hdr[type].ip6_flow);

  // Is value too large or small, or are there non-numeric characters?
  if (!is_ascii_uint (entry_text) || (ascii_to_int64 ((char *) entry_text) > 15l)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    return (EXIT_FAILURE);

  // Value is in acceptable range.
  } else {

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    data->ip6hdr[type].ip6_flow = htonl ((((uint32_t) ascii_to_int64 ((char *) entry_text)) << 28) + (old & 0xffffffful));  // Update IPv6 header with new value

    // Copy IPv6 header to 6to4 IPv6 header, since create_6to4_frame() doesn't copy it.
    memcpy (&data->ip6hdr[type + 3], &data->ip6hdr[type], IP6_HDRLEN * sizeof (uint8_t));

    // Update ethernet frames.
    create_ip6_frame (type, data);
    create_6to4_frame (type + 3, data);
  }

  return (EXIT_SUCCESS);
}

// Traffic class (8 bits)
int
ip6_tc_entry (GtkWidget *entry, int type, SPSData *data) {

  const char *entry_text;
  uint32_t old;
  GtkStyleContext *context;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (entry);

  // Get new value from text entry.
  entry_text = gtk_entry_get_text (GTK_ENTRY (entry));

  // Save value currently in header.
  old = ntohl (data->ip6hdr[type].ip6_flow);

  // Is value too large or small, or are there non-numeric characters?
  if (!is_ascii_uint (entry_text) || (ascii_to_int64 ((char *) entry_text) > 255l)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    return (EXIT_FAILURE);

  // Value is in acceptable range.
  } else {

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    data->ip6hdr[type].ip6_flow = htonl ((((uint32_t) ascii_to_int64 ((char *) entry_text)) << 20) + (old & 0xf00ffffful));  // Update IPv6 header with new value

    // Copy IPv6 header to 6to4 IPv6 header, since create_6to4_frame() doesn't copy it.
    memcpy (&data->ip6hdr[type + 3], &data->ip6hdr[type], IP6_HDRLEN * sizeof (uint8_t));

    // Update ethernet frames.
    create_ip6_frame (type, data);
    create_6to4_frame (type + 3, data);
  }

  return (EXIT_SUCCESS);
}

// Flow label (20 bits)
int
ip6_fl_entry (GtkWidget *entry, int type, SPSData *data) {

  const char *entry_text;
  uint32_t old;
  GtkStyleContext *context;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (entry);

  // Get new value from text entry.
  entry_text = gtk_entry_get_text (GTK_ENTRY (entry));

  // Save value currently in header.
  old = ntohl (data->ip6hdr[type].ip6_flow);

  // Is value too large or small, or are there non-numeric characters?
  if (!is_ascii_uint (entry_text) || (ascii_to_int64 ((char *) entry_text) > 0xfffffl)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    return (EXIT_FAILURE);

  // Value is in acceptable range.
  } else {

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    data->ip6hdr[type].ip6_flow = htonl (((uint32_t) ascii_to_int64 ((char *) entry_text)) + (old & 0xfff00000ul));  // Update IPv6 header with new value

    // Copy IPv6 header to 6to4 IPv6 header, since create_6to4_frame() doesn't copy it.
    memcpy (&data->ip6hdr[type + 3], &data->ip6hdr[type], IP6_HDRLEN * sizeof (uint8_t));

    // Update ethernet frames.
    create_ip6_frame (type, data);
    create_6to4_frame (type + 3, data);
  }

  return (EXIT_SUCCESS);
}

// Payload length (16 bits)
int
ip6_plen_entry (GtkWidget *entry, int type, SPSData *data) {

  const char *entry_text;
  GtkStyleContext *context;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (entry);

  // Get new value from text entry.
  entry_text = gtk_entry_get_text (GTK_ENTRY (entry));

  // Is value too large or small, or are there non-numeric characters?
  if (!is_ascii_uint (entry_text) || (ascii_to_int64 ((char *) entry_text) > 65535)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    return (EXIT_FAILURE);

  // Value is in acceptable range.
  } else {

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    data->ip6hdr[type].ip6_plen = htons ((uint16_t) ascii_to_int64 ((char *) entry_text));  // Update IPv6 header with new value

    // Copy IPv6 header to 6to4 IPv6 header, since create_6to4_frame() doesn't copy it.
    memcpy (&data->ip6hdr[type + 3], &data->ip6hdr[type], IP6_HDRLEN * sizeof (uint8_t));

    // Update ethernet frames.
    create_ip6_frame (type, data);
    create_6to4_frame (type + 3, data);
  }

  return (EXIT_SUCCESS);
}

// Next header (8 bits)
int
ip6_nxt_entry (GtkWidget *entry, int type, SPSData *data) {

  const char *entry_text;
  GtkStyleContext *context;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (entry);

  // Get new value from text entry.
  entry_text = gtk_entry_get_text (GTK_ENTRY (entry));

  // Is value too large or small, or are there non-numeric characters?
  if (!is_ascii_uint (entry_text) || (ascii_to_int64 ((char *) entry_text) > 255)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    return (EXIT_FAILURE);

  // Value is in acceptable range.
  } else {

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    data->ip6hdr[type].ip6_nxt = (uint8_t) ascii_to_int64 ((char *) entry_text);  // Update IPv6 header with new value

    // Copy IPv6 header to 6to4 IPv6 header, since create_6to4_frame() doesn't copy it.
    memcpy (&data->ip6hdr[type + 3], &data->ip6hdr[type], IP6_HDRLEN * sizeof (uint8_t));

    // Update ethernet frames.
    create_ip6_frame (type, data);
    create_6to4_frame (type + 3, data);
  }

  return (EXIT_SUCCESS);
}

// Hop limit (8 bits)
int
ip6_hops_entry (GtkWidget *entry, int type, SPSData *data) {

  const char *entry_text;
  GtkStyleContext *context;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (entry);

  // Get new value from text entry.
  entry_text = gtk_entry_get_text (GTK_ENTRY (entry));

  // Is value too large or small, or are there non-numeric characters?
  if (!is_ascii_uint (entry_text) || (ascii_to_int64 ((char *) entry_text) > 255)) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    return (EXIT_FAILURE);

  // Value is in acceptable range.
  } else {

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    data->ip6hdr[type].ip6_hops = (uint8_t) ascii_to_int64 ((char *) entry_text);  // Update IPv6 header with new value

    // Copy IPv6 header to 6to4 IPv6 header, since create_6to4_frame() doesn't copy it.
    memcpy (&data->ip6hdr[type + 3], &data->ip6hdr[type], IP6_HDRLEN * sizeof (uint8_t));

    // Update ethernet frames.
    create_ip6_frame (type, data);
    create_6to4_frame (type + 3, data);
  }

  return (EXIT_SUCCESS);
}

// Source IP address (128 bits)
int
ip6_sip_entry (GtkWidget *entry, int type, SPSData *data) {

  const char *entry_text;
  GtkStyleContext *context;
  struct sockaddr_in6 sa;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (entry);

  if (type > 8) {
    data->parent = data->traceroute_window;
  } else {
    data->parent = data->main_window;
  }

  // Get new value from text entry.
  entry_text = gtk_entry_get_text (GTK_ENTRY (entry));

  // Is it a valid IP address? Valid if inet_pton returns 1.
  if (inet_pton (AF_INET6, entry_text, &sa.sin6_addr) != 1) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    return (EXIT_FAILURE);

  // Valid IP
  } else {

    // Try to place new address in sockaddr_in.
    if (inet_pton (AF_INET6, entry_text, &(data->ip6hdr[type].ip6_src)) != 1) {

      // Apply the error style to entry text.
      gtk_style_context_add_class (context, "error");

      return (EXIT_FAILURE);
    }

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    // Copy IPv6 header to 6to4 IPv6 header, since create_6to4_frame() doesn't copy it.
    memcpy (&data->ip6hdr[type + 3], &data->ip6hdr[type], IP6_HDRLEN * sizeof (uint8_t));

    // Update upper layer protocol checksum, which protects source and destination IP addresses.
    switch (type) {
      case 3:
        tcp_chksum (type, data);
        break;
      case 4:
        icmp_chksum (type, data);
        break;
      case 5:
        udp_chksum (type, data);
        break;
      case 12:
        tcp_chksum (type, data);
        break;
      case 13:
        icmp_chksum (type, data);
        break;
      case 14:
        udp_chksum (type, data);
        break;
      default:
        fprintf (stderr, "ERROR: Invalid upper layer protocol type %i in ip6_sip_entry().\n", type);
        exit (EXIT_FAILURE);
    }
  }

  // Update ethernet frames.
  create_ip6_frame (type, data);
  create_6to4_frame (type + 3, data);

  return (EXIT_SUCCESS);
}

// Randomize source IP address
int
ip6_ransip_checkbutton (int *ran_sourceip, GtkWidget *entry, int type, SPSData *data) {

  int status;
  char *ipaddress;
  struct sockaddr_in6 sa;

  // Array of chars to hold an IPv6 address + '\0'
  ipaddress = allocate_strmem (INET6_ADDRSTRLEN);

  // Generate a valid IPv6 address.
  if ((status = inet_pton (AF_INET6, ran6_addr (ipaddress, data), &sa.sin6_addr)) != 1) {
    sprintf (data->error_text, "ip6_ransip_checkbutton(): inet_pton() failed.\nError message: %s", strerror (status));
    if (type > 8) {
      data->parent = data->traceroute_window;
    } else {
      data->parent = data->main_window;
    }
    report_error (data);
    free (ipaddress);
    return (EXIT_FAILURE);

  // Valid IP
  } else {
    if ((status = inet_pton (AF_INET6, ipaddress, &data->ip6hdr[type].ip6_src)) != 1) {
      sprintf (data->error_text, "ip6_ransip_checkbutton(): inet_pton() failed.\nError message: %s", strerror (status));
      if (type > 8) {
        data->parent = data->traceroute_window;
      } else {
        data->parent = data->main_window;
      }
      report_error (data);
      free (ipaddress);
      return (EXIT_FAILURE);
    }

    // Update entry with random IP address.
    gtk_entry_set_text (GTK_ENTRY (entry), ipaddress);

    // Now we can set the randomize flag.
    *ran_sourceip = 1;

    // Copy IPv6 header to 6to4 IPv6 header, since create_6to4_frame() doesn't copy it.
    memcpy (&data->ip6hdr[type + 3], &data->ip6hdr[type], IP6_HDRLEN * sizeof (uint8_t));

    // Update upper layer protocol checksum, which protects source and destination IP addresses.
    switch (type) {
      case 3:  // IPv6 TCP
        gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton2), TRUE);  // Set the randomize checkbutton to active.
        tcp_chksum (type, data);
        break;
      case 4:  // IPv6 ICMP
        gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton10), TRUE);  // Set the randomize checkbutton to active.
        icmp_chksum (type, data);
        break;
      case 5:  // IPv6 UDP
        gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton6), TRUE);  // Set the randomize checkbutton to active.
        udp_chksum (type, data);
        break;
      default:
        fprintf (stderr, "ERROR: Invalid upper layer protocol type %i in ip6_ransip_checkbutton().\n", type);
        free (ipaddress);
        exit (EXIT_FAILURE);
    }

    // Update ethernet frames.
    create_ip6_frame (type, data);
    create_6to4_frame (type + 3, data);
  }

  // Free allocated memory.
  free (ipaddress);

  return (EXIT_SUCCESS);
}

// Destination IP address (128 bits)
int
ip6_dip_entry (GtkWidget *entry, int type, SPSData *data) {

  const char *entry_text;
  GtkStyleContext *context;
  struct sockaddr_in6 sa;

  // Get current style context of the entry.
  context = gtk_widget_get_style_context (entry);

  if (type > 8) {
    data->parent = data->traceroute_window;
  } else {
    data->parent = data->main_window;
  }

  // Get new value from text entry.
  entry_text = gtk_entry_get_text (GTK_ENTRY (entry));

  // Is it a valid IP address? Valid if inet_pton returns 1.
  if (inet_pton (AF_INET6, entry_text, &sa.sin6_addr) != 1) {

    // Apply the error style to entry text.
    gtk_style_context_add_class (context, "error");

    return (EXIT_FAILURE);

  // Valid IP
  } else {

    // Try to place new address in sockaddr_in.
    if (inet_pton (AF_INET6, entry_text, &(data->ip6hdr[type].ip6_dst)) != 1) {

      // Apply the error style to entry text.
      gtk_style_context_add_class (context, "error");

      return (EXIT_FAILURE);
    }

    // Remove the error style (if applied) to entry text.
    gtk_style_context_remove_class (context, "error");

    // Copy IPv6 header to 6to4 IPv6 header, since create_6to4_frame() doesn't copy it.
    memcpy (&data->ip6hdr[type + 3], &data->ip6hdr[type], IP6_HDRLEN * sizeof (uint8_t));

    // Update upper layer protocol checksum, which protects source and destination IP addresses.
    switch (type) {
      case 3:
        tcp_chksum (type, data);
        break;
      case 4:
        icmp_chksum (type, data);
        break;
      case 5:
        udp_chksum (type, data);
        break;
      case 12:
        tcp_chksum (type, data);
        break;
      case 13:
        icmp_chksum (type, data);
        break;
      case 14:
        udp_chksum (type, data);
        break;
      default:
        fprintf (stderr, "ERROR: Invalid upper layer protocol type %i in ip6_dip_entry().\n", type);
        exit (EXIT_FAILURE);
    }
  }

  // Update ethernet frames.
  create_ip6_frame (type, data);
  create_6to4_frame (type + 3, data);

  return (EXIT_SUCCESS);
}

// Truncate IPv6 payload length as necessary to accommodate the
// variable lengths of extension headers and TCP options.
int
truncate_ip6 (int type, SPSData *data) {

  int unfraglen, fraglen_nopayload;

  // Find length of unfragmentable portion. Note allowance for IPv4 header for 6to4.
  unfraglen = find6_unfraglen (type, data) + IP4_HDRLEN;

  // Find length of fragmentable portion, excluding payload data.
  fraglen_nopayload = find6_fraglen (type, data) - data->payloadlen[type];

  if ((unfraglen + fraglen_nopayload + data->payloadlen[type]) > IP_MAXPACKET) {
    data->payloadlen[type] = IP_MAXPACKET - unfraglen - fraglen_nopayload;
    return (1);  // Truncation occurred.
  }

  return (0);  // Truncation did not occur.
}

// Calculate IPv6 payload length for IPv6 page or IPv6 traceroute page.
// See ip6_new_paylen() for payload length calculation for "new" IPv6 associated with AH or ESP tunneling.
uint16_t
ip6_paylen (int type, SPSData *data) {

// AH or ESP transport, or no AH or ESP
/* Unfragmentable portion
     10 = Original IPv6 header
     11 = Original/Inner hop-by-hop header
     12 = Original/Inner destination header (first)
     13 = Original/Inner routing header

   Fragmentable portion
      5 = auth header (no esp header)
      6 = esp header (no auth header)
     14 = Original/Inner destination header (last)
     15 = Original/Inner mobility header
      9 = No next header
     16 = TCP header
     17 = ICMPV6 header
     18 = UDP header
     19 = ESP trailer header + ICV
*/

// AH or ESP tunnel
/* Unfragmentable portion
      0 = New IPv6 header (if AH or ESP tunneling)
      1 = Outer hop-by-hop header
      2 = Outer destination header (first)
      3 = Outer routing header

   Fragmentable portion
      5 = auth header (no esp header)
      6 = esp header (no auth header)
      7 = Outer destination header (last)
      8 = Outer mobility header
      9 = No next header
     10 = Original/Inner IPv6 header
     11 = Original/Inner hop-by-hop header
     12 = Original/Inner destination header (first)
     13 = Original/Inner routing header
     14 = Original/Inner destination header (last)
     15 = Original/Inner mobility header
      9 = No next header
     16 = TCP header
     17 = ICMPV6 header
     18 = UDP header
     19 = ESP trailer header + ICV
 */

  int i, flag, len;

  // Set order of all headers, extension headers, and data.
  // This must be called in order to ensure any new items are included in chain length.
  ip6_order (type, data);  // Note: This will set data->first_frag[type], which is the first link in chain that will need to be fragmented.

  // Find index i of Original/Inner IPv6 header in header chain.
  i = 0;  // Index of chain of headers
  flag = 0;  // Flag indicating we found the original/inner IPv6 header
  while ((data->order[type][i] != 99) && (i < MAX_IP6LINKS)) {
    if (data->order[type][i] == 10) {
      flag = 1;  // Found it.
      break;
    }
    i++;
  }  // Next link
  if (flag == 0) {
    fprintf (stderr, "ERROR: ip6_paylen() couldn't find original/inner IPv6 header in chain of headers.\n");
    exit (EXIT_FAILURE);
  }

  // Calculate payload length by adding up all contributions.
  len = 0;  // Initialize payload length to 0.

  // Loop through each header in chain to add all contributions to payload length of original/inner IPv6 header.
  // Here we only consider those elements contributing to the original/inner IPv6 header's paylen.
  i++;  // Go to header immediately following the Original/Inner IPv6 header.
  while ((data->order[type][i] != 99) && (i < MAX_IP6LINKS)) {

    switch (data->order[type][i]) {

      // New IPv6 header (if tunneling AH or ESP)
      case 0:
        fprintf (stderr, "ERROR: ip6_paylen(): new IPv6 header shouldn't occur here at index %i.\n", i);
        exit (EXIT_FAILURE);
        break;

      // Outer hop-by-hop header
      case 1:
        // Always above Original/Inner IPv6 header. Ignore.
        break;

      // Outer destination header (first)
      case 2:
        // Always above Original/Inner IPv6 header. Ignore.
        break;

      // Outer routing header
      case 3:
        // Always above Original/Inner IPv6 header. Ignore.
        break;

      // Unused
      case 4:
        // Do nothing.
        break;

      // Authentication header (no ESP header)
      case 5:
        // Only under Original/Inner IPv6 header if NOT tunneling AH.
        // Should not exist if using ESP; protect here anyway.
        if (!(data->auth_hdr_flag[type] && data->auth_tr_tun_flag[type]) &&
            !(data->esp_hdr_flag[type] && data->esp_tr_tun_flag[type])) {
          len += ATH_HDRLEN + data->auth_len[type];
        }
        break;

      // ESP header (8 bytes) (no AH header); contains only spi and seq
      case 6:
        // Only under Original/Inner IPv6 header if NOT tunneling AH.
        // Should not exist if using ESP; protect here anyway.
        if (!(data->auth_hdr_flag[type] && data->auth_tr_tun_flag[type]) &&
            !(data->esp_hdr_flag[type] && data->esp_tr_tun_flag[type])) {
          len += ESP_HDRLEN;
        }
        break;

      // Outer destination header (last)
      case 7:
        // Always above Original/Inner IPv6 header. Ignore.
        break;

      // Outer mobility header
      case 8:
        // Always above Original/Inner IPv6 header. Ignore.
        break;

      // No next header
      case 9:
        // Do nothing.
        break;

      // IPv6 header; This is the original IPv6 header.
      case 10:
        fprintf (stderr, "ERROR: ip6_paylen(): Original/Inner IPv6 header shouldn't occur here at index %i.\n", i);
        exit (EXIT_FAILURE);
        break;

      // Original/Inner hop-by-hop header
      case 11:
        len += HOP_HDRLEN + data->hbh_opt_totlen[type][0] + data->hbh_optpadlen[type][0];
        break;

      // Original/Inner destination header (first)
      case 12:
        len += DST_HDRLEN + data->dstf_opt_totlen[type][0] + data->dstf_optpadlen[type][0];
        break;

      // Original/Inner routing header
      case 13:
        len += RTE_HDRLEN + data->route_datlen[type][0];
        break;

      // Original/Inner destination header (last)
      case 14:
        len += DST_HDRLEN + data->dstl_opt_totlen[type][0] + data->dstl_optpadlen[type][0];
        break;

      // Original/Inner mobility header
      case 15:
        // Not implemented here. Do nothing.
        break;

      // TCP header, options, and payload data
      case 16:
        len += TCP_HDRLEN + data->tcp_opt_totlen[type] + data->tcp_optpadlen[type];
        len += data->payloadlen[type];
        break;

      // ICMP header and payload data
      case 17:
        len += ICMP_HDRLEN;
        len += data->payloadlen[type];
        break;

      // UDP header and payload data
      case 18:
        len += UDP_HDRLEN;
        len += data->payloadlen[type];
        break;

      // ESP trailer header + ICV
      // Add padding (Section 2.4 of RFC 4303 and 4305).
      case 19:
        len += data->esptail[type].pad_len;
        len += ESP_TAILLEN;
        len += data->esp_auth_len[type];
        break;

      // Unknown index value
      default:
        fprintf (stderr, "ERROR: Unknown index value %i in ip6_paylen().\n", data->order[type][i]);
        exit (EXIT_FAILURE);

    }  // End switch statement

    i++;  // Go to next item in chain
  }  // End loop through IPv6 chain

  return (len);
}

// Calculate IPv6 payload length for "new" IPv6 header when tunneling AH or ESP.
// See ip6_paylen() for payload length calculation for Original/Inner IPv6 ignoring AH or ESP tunneling.
// Return length or -1 if not tunneling.
uint16_t
ip6_new_paylen (int type, SPSData *data) {

// AH or ESP tunnel
/* Unfragmentable portion
      0 = New IPv6 header (if AH or ESP tunneling)
      1 = Outer hop-by-hop header
      2 = Outer destination header (first)
      3 = Outer routing header

   Fragmentable portion
      5 = auth header (no esp header)
      6 = esp header (no auth header)
      7 = Outer destination header (last)
      8 = Outer mobility header
      9 = No next header
     10 = Original/Inner IPv6 header
     11 = Original/Inner hop-by-hop header
     12 = Original/Inner destination header (first)
     13 = Original/Inner routing header
     14 = Original/Inner destination header (last)
     15 = Original/Inner mobility header
      9 = No next header
     16 = TCP header
     17 = ICMPV6 header
     18 = UDP header
     19 = ESP trailer header + ICV
 */

  // First check to make sure we are actually tunneling AH or ESP.
  if (!(data->auth_hdr_flag[type] && data->auth_tr_tun_flag[type]) &&
      !(data->esp_hdr_flag[type] && data->esp_tr_tun_flag[type])) {
    return (-1);
  }

  int i, len;

  len = 0;  // Initialize payload length to 0.

  // Loop through each header in chain to add all contributions to payload length of "new" IPv6 header.
  i = 1;  // Skip new IPv6 header (doesn't include itself in paylen).
  while ((data->order[type][i] != 99) && (i < MAX_IP6LINKS)) {

    switch (data->order[type][i]) {

      // New IPv6 header (if tunneling AH or ESP)
      // This case should never be encountered since we start at i = 1.
      case 0:
        fprintf (stderr, "ERROR: ip6_new_paylen(): new IPv6 header (case %i) shouldn't occur here.\n", data->order[type][i]);
        exit (EXIT_FAILURE);
        break;

      // Outer hop-by-hop header
      case 1:
        len += HOP_HDRLEN + data->hbh_opt_totlen[type][1] + data->hbh_optpadlen[type][1];
        break;

      // Outer destination header (first)
      case 2:
        len += DST_HDRLEN + data->dstf_opt_totlen[type][1] + data->dstf_optpadlen[type][1];
        break;

      // Outer routing header
      case 3:
        len += RTE_HDRLEN + data->route_datlen[type][1];
        break;

      // Unused
      case 4:
        // Do nothing.
        break;

      // Authentication header (no ESP header)
      case 5:
        len += ATH_HDRLEN + data->auth_len[type];
        break;

      // ESP header (8 bytes) (no AH header): includes only spi and seq
      case 6:
        len += ESP_HDRLEN;
        break;

      // Outer destination header (last)
      case 7:
        len += DST_HDRLEN + data->dstl_opt_totlen[type][1] + data->dstl_optpadlen[type][1];
        break;

      // Outer mobility header
      case 8:
        // Not implemented here. Do nothing.
        break;

      // No next header
      case 9:
        // Do nothing.
        break;

      // Original/Inner IPv6 header
      case 10:
        len += IP6_HDRLEN;
        break;

      // Original/Inner hop-by-hop header
      case 11:
        len += HOP_HDRLEN + data->hbh_opt_totlen[type][0] + data->hbh_optpadlen[type][0];
        break;

      // Original/Inner destination header (first)
      case 12:
        len += DST_HDRLEN + data->dstf_opt_totlen[type][0] + data->dstf_optpadlen[type][0];
        break;

      // Original/Inner routing header
      case 13:
        len += RTE_HDRLEN + data->route_datlen[type][0];
        break;

      // Original/Inner destination header (last)
      case 14:
        len += DST_HDRLEN + data->dstl_opt_totlen[type][0] + data->dstl_optpadlen[type][0];
        break;

      // Original/Inner mobility header
      case 15:
        // Not implemented here. Do nothing.
        break;

      // TCP header, options, and payload data
      case 16:
        len += TCP_HDRLEN + data->tcp_opt_totlen[type] + data->tcp_optpadlen[type];
        len += data->payloadlen[type];
        break;

      // ICMP header and payload data
      case 17:
        len += ICMP_HDRLEN;
        len += data->payloadlen[type];
        break;

      // UDP header and payload data
      case 18:
        len += UDP_HDRLEN;
        len += data->payloadlen[type];
        break;

      // ESP trailer + ICV
      // Add padding (Section 2.4 of RFC 4303 and 4305).
      case 19:
        len += data->esptail[type].pad_len;
        len += ESP_TAILLEN;
        len += data->esp_auth_len[type];
        break;

      // Unknown index value
      default:
        fprintf (stderr, "ERROR: Unknown index value %i in ip6_new_paylen().\n", data->order[type][i]);
        exit (EXIT_FAILURE);

    }  // End switch statement

    i++;  // Go to next item in chain
  }  // End loop through IPv6 chain

  return (len);
}

// Update payload length entry for an IPv6 page or IPv6 traceroute page.
int
update_plen_entry (int type, uint16_t len, SPSData *data) {

  char *value;
  GtkWidget *plen_entry;

  // Allocate memory for character array.
  value = allocate_strmem (TMP_STRINGLEN);

  // Create string version of len.
  sprintf (value, "%i", len);

  // Set plen_entry to appropriate entry of interface.
  switch (type) {
    case 3:  // IPv6 TCP
      plen_entry = data->entry25;
      break;
    case 4:  // IPv6 ICMP
      plen_entry = data->entry113;
      break;
    case 5:  // IPv6 UDP
      plen_entry = data->entry100;
      break;
    case 12:  // IPv6 TCP for traceroute
      plen_entry = data->tr_entry254;
      break;
    case 13:  // IPv6 ICMP for traceroute
      plen_entry = data->tr_entry284;
      break;
    case 14:  // IPv6 UDP for traceroute
      plen_entry = data->tr_entry302;
      break;
    default:
      fprintf (stderr, "ERROR: Invalid upper layer protocol type %i in update_plen_entry().\n", type);
      free (value);
      exit (EXIT_FAILURE);
  }
  gtk_entry_set_text (GTK_ENTRY (plen_entry), value);  // Update entry with value.

  // Free allocated memory.
  free (value);

  return (EXIT_SUCCESS);
}

// Update (IPv6): lengths, offsets, checksums, and regenerate ethernet frames.
int
ip6data_update (int type, SPSData *data) {

  uint16_t len;
  int ifmtu, min;

  // Set parent window for error reporting.
  if (type < 9) {
    data->parent = data->main_window;
  } else {
    data->parent = data->traceroute_window;
  }

  // Truncate upper layer protocol payload if necessary to accommodate packet size limit.
  if (truncate_ip6 (type, data)) {
    sprintf (data->warning_text, "Upper layer protocol payload data was truncated\nto accommodate maximum packet size of %i bytes.", IP_MAXPACKET);
    report_warning (data);
  }

  // Original/Inner IPv6 header payload length
  len = ip6_paylen (type, data);

  // Copy len into original/inner IPv6 header.
  data->ip6hdr[type].ip6_plen = htons (len);

  // Also copy revised IPv6 header to 6to4 header.
  memcpy (&data->ip6hdr[type+3], &data->ip6hdr[type], IP6_HDRLEN * sizeof (uint8_t));

  // Update IPv6 header payload length entry in user interface.
  update_plen_entry (type, len, data);

  // "New" IPv6 header payload length
  len = ip6_new_paylen (type, data);

  // Copy len into "New" IPv6 header.
  data->ip6hdr[type+20].ip6_plen = htons (len);

  // "New" IPv6 header source IPv6 address (128 bits): copy from encapsulated IPv6 header
  memcpy (&data->ip6hdr[type+20].ip6_src, &data->ip6hdr[type].ip6_src, 16 * sizeof (uint8_t));

  // "New" IPv6 header destination IPv6 address (128 bits): copy from encapsulated IPv6 header
  memcpy (&data->ip6hdr[type+20].ip6_dst, &data->ip6hdr[type].ip6_dst, 16 * sizeof (uint8_t));

  // Update TCP stuff, if appropriate.
  if ((type == 3) || (type == 12)) {

    // TCP data offset (4 bits)
    tcp_dataoffset (type, data);

    // TCP header checksum (16 bits)
    tcp_chksum (type, data);
  }

  // Update ICMP stuff, if appropriate.
  if ((type == 4) || (type == 13)) {
    // ICMP header checksum (16 bits)
    icmp_chksum (type, data);
  }

  // Update UDP stuff, if appropriate.
  if ((type == 5) || (type == 14)) {
    // Length of UDP datagram (16 bits)
    udp_len (type, data);

    // UDP header checksum (16 bits)
    udp_chksum (type, data);
  }

  // Maximum transmission unit (MTU) checks

  // Obtain interface's MTU.
  // If we fail to obtain MTU from interface, don't change current MTU.
  if ((ifmtu = interface_getmtu (data->ifname[type])) == 0) {
    sprintf (data->warning_text, "Failed to obtain MTU from interface for IPv6.\nWas interface specified?");
    report_warning (data);
    return (EXIT_FAILURE);
  }

  // Determine minimum size needed to accommodate packet,
  // when packet is fragmented to maximum extent.
  min = min6_mtu (type, data);

  // Is interface's MTU is too small to accommodate packet
  // even when packet is fragmented to maximum extent?
  if (ifmtu < min) {
    sprintf (data->warning_text, "The interface's MTU is too small to accommodate IPv6 packet.");
    report_warning (data);
    return (EXIT_FAILURE);
  }

  // Is current MTU value too small to accommodate packet,
  // even when packet is fragmented to maximum extent?
  if (data->ifmtu[type] < min) {
    sprintf (data->warning_text, "The current MTU value is too small to accommodate IPv6 packet.\nPacket needs a minimum of %i bytes.\nValue revised to interface MTU.", min);
    report_warning (data);
    data->ifmtu[type] = ifmtu;

    // Update MTU spinbutton.
    update_mtu_spinbutton (type, data);
  }

  // 6to4 (impacted by IPv6)

  // Here we do not fetch the sixto4 interface ifmtu because the user may
  // have no intention of doing 6to4, in which case a 6to4 interface would not be specified
  // on the 6to4 page of UI. As a result, 6to4 MTU must be checked at Send time.

  // Update ethernet frames.
  create_ip6_frame (type, data);
  create_6to4_frame (type+3, data);

  return (EXIT_SUCCESS);
}
