/*  Simple Packet Sender (SPS) - a linux packet crafting network
    tool using GTK+ graphical interface.

    Copyright (C) 2011-2023  Hohlraum (h0h1r4um@gmail.com)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <gtk/gtk.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>           // close(), sleep(), usleep(), gethostname(), write()
#include <string.h>           // strncpy(), strncmp(), strnlen(), strncat(), memset(), and memcpy()
#include <time.h>             // localtime(), time()
#include <inttypes.h>         // uint8_t, uint16_t, uint32_t, uint64_t, int32_t, int64_t, PRIu8, PRIu16, PRIu32, PRIu64

#include <netdb.h>            // struct addrinfo
#include <sys/types.h>        // needed for socket()
#include <sys/socket.h>       // needed for socket()
#include <netinet/in_systm.h>
#include <netinet/in.h>       // IPPROTO_IP, IPPROTO_ICMP, IPPROTO_TCP, IPPROTO_UDP, IPPROTO_IPV6, IPPROTO_ICMPV6, INET_ADDRSTRLEN, INET6_ADDRSTRLEN
#include <netinet/ip.h>       // struct ip and IP_MAXPACKET (which is 65535), IPPROTO_IPV6, IPPROTO_ICMPV6
#include <netinet/ip6.h>      // struct ip6_hdr
#define __FAVOR_BSD           // Use BSD format of TCP and UDP header. Must appear here BEFORE #include <netinet/tcp.h>.
#include <netinet/tcp.h>      // struct tcphdr
#include <netinet/ip_icmp.h>  // struct icmp, ICMP_ECHO
#include <netinet/icmp6.h>    // struct icmp6_hdr, ICMP6_ECHO_REQUEST, ICMP6_TIME_EXCEEDED, struct nd_neighbor_solicit, ND_NEIGHBOR_SOLICIT
#include <netinet/udp.h>      // struct udphdr
#include <fcntl.h>            // fcntl()
#include <sys/select.h>       // select()
#include <arpa/inet.h>        // inet_pton(), inet_ntop(), htons(), ntohs(), htonl(), ntohl()
#include <sys/ioctl.h>        // macro ioctl is defined
#include <bits/ioctls.h>      // defines values for argument "request" of ioctl. Here, we need SIOCGIFHWADDR
#include <net/if.h>           // struct ifreq
#include <ifaddrs.h>          // ifaddrs()
#include <bits/socket.h>      // structs msghdr and cmsghdr
#include <linux/if_ether.h>   // ETH_P_IP = 0x0800, ETH_P_IPV6 = 0x86DDD
#include <linux/if_packet.h>  // struct sockaddr_ll (see man 7 packet)
#include <net/ethernet.h>
#include <sys/time.h>         // gettimeofday()

#include <errno.h>            // errno, perror(), EAGAIN, EINTR, ECONNREFUSED

// Define some symbolic constants
#define UTF8_FILENAME "UTF-8.csv"  // Small-capital UTF-8 cross-reference filename
#define ETH_HDRLEN 14      // Ethernet header length (bytes)
#define IP4_HDRLEN 20      // IPv4 header length (bytes)
#define IP6_HDRLEN 40      // IPv6 header length (bytes)
#define MAX_IP6LINKS 1000  // Maximum number of links in an IPv6 chain of headers, options, payloads, etc.
#define TCP_HDRLEN 20      // TCP header length, excludes options data (bytes)
#define ICMP_HDRLEN 8      // ICMP header length for echo request, excludes data (bytes)
#define UDP_HDRLEN 8       // UDP header length, excludes data (bytes)
#define FRG_HDRLEN 8       // IPv6 fragment header (bytes)
#define HOP_HDRLEN 2       // Hop-by-hop header length (bytes), excluding options
#define DST_HDRLEN 2       // Destination header length (bytes), excluding options
#define RTE_HDRLEN 4       // Routing header length, excluding data
#define ATH_HDRLEN 12      // Authentication header length (bytes), excludes authentication data
#define ESP_HDRLEN 8       // ESP header length (bytes); includes only spi and seq
#define ESP_TAILLEN 2      // ESP tail length (bytes); includes only pad_len and nxt_hdr
#define ARP_HDRLEN 28      // ARP header length (bytes)
#define ARPOP_REQUEST 1    // Taken from <linux/if_arp.h>
#define ARPOP_REPLY 2      // Taken from <linux/if_arp.h>
#define FTP_TIMEOUT 10     // FTP listening timeout
#define FTP_PORT "21"      // FTP port number
#define MAX_FRAGS 65536    // Maximum number of fragments for one packet (some big number)
#define MAX_IP4OPTLEN 40   // Maximum length (bytes) of IPv4 options
#define MAX_IP4OPTIONS 20  // Maximum number of IPv4 options (60 - IP4_HDRLEN) / 2 bytes)
#define MAX_TCPOPTLEN 40   // Maximum length (bytes) of TCP options
#define MAX_TCPOPTIONS 20  // Maximum number of TCP options (40 / 2 bytes)
#define MAX_HBHOPTIONS 20  // Maximum number of hop-by-hop extension header options
#define MAX_HBHOPTLEN 256  // Maximum length (bytes) of a hop-by-hop option (some large value)
#define MAX_DSTOPTIONS 20  // Maximum number of destination extension header options
#define MAX_DSTOPTLEN 256  // Maximum length (bytes) of a destination option (some large value)
#define MAX_ROUTEDATA 2044  // Maximum length (bytes) of routing header data (255 * 8 bytes) + 4
#define MAX_AUTHICVLEN 1024  // Maximum length (bytes) of authentication header ICV
#define MAX_ESPICVLEN 1024  // Maximum length (bytes) of ESP header ICV
#define MAX_RSVD_ADDRS 1024  // Maximum number of reserved IP address ranges (applies to IPv4 and 6, separately)
#define TMP_STRINGLEN 80  // Length of array to hold temporary values as strings
#define TEXT_STRINGLEN 1024  // Length of array to hold various strings of text
#define RIR_MEM_BLOCK 1000000  // Size of each memory chunk allocated for IP/ASN RIR search results
#define SMALL 0  // Value indicating lower case UTF-8 letter
#define CAPITAL 1  // Value indicating upper case UTF-8 letter

// Definition of pktinfo6 created from definition of in6_pktinfo in netinet/in.h.
// This should remove "redefinition of in6_pktinfo" errors in some linux variants.
typedef struct _pktinfo6 Pktinfo6;
struct _pktinfo6 {
  struct in6_addr ipi6_addr;
  int ipi6_ifindex;
};

// Struct to hold ethernet header values
typedef struct _ethhdr Ethhdr;
struct _ethhdr {
  uint8_t dst_mac[6];
  uint8_t src_mac[6];
  uint16_t type_code;
};

// Define a struct for hop-by-hop and destination headers, excluding options.
typedef struct _hopdst_hdr Hopdst_hdr;
struct _hopdst_hdr {
  uint8_t nxt_hdr;
  uint8_t hdr_len;
};

// Define a struct for routing header, excluding data.
typedef struct _route_hdr Route_hdr;
struct _route_hdr {
  uint8_t nxt_hdr;
  uint8_t hdr_len;
  uint8_t routing_type;
  uint8_t segs_left;
};

// Define a struct for authentication header, excluding authentication data.
typedef struct _auth_hdr Auth_hdr;
struct _auth_hdr {
  uint8_t nxt_hdr;
  uint8_t pay_len;
  uint16_t reserved;
  uint32_t spi;
  uint32_t seq;
};

// Define a struct for head of ESP header, excluding payload and authentication data.
typedef struct _esp_hdr Esp_hdr;
struct _esp_hdr {
  uint32_t spi;
  uint32_t seq;
};

// Define a struct for tail of ESP header, excluding payload and authentication data.
typedef struct _esp_tail Esp_tail;
struct _esp_tail {
  uint8_t pad_len;
  uint8_t nxt_hdr;
};

// Define a struct for ARP header.
typedef struct _arp_hdr Arp_hdr;
struct _arp_hdr {
  uint16_t htype;
  uint16_t ptype;
  uint8_t hlen;
  uint8_t plen;
  uint16_t opcode;
  uint8_t sender_mac[6];
  uint8_t sender_ip[4];
  uint8_t target_mac[6];
  uint8_t target_ip[4];
};

// Define a struct to contain pointers to a textview and message for post_message().
typedef struct _msgdata Msgdata;
struct _msgdata {
  char *message;
  GtkWidget *textview;
};

// Define a struct to represent a uint128_t, composed of two uint64_t values.
// Currently, inttypes.h doesn't contain uint128_t.
typedef struct {
    uint64_t h;
    uint64_t l;
} u128_t;

// Main data structure definition
typedef struct _SPSData SPSData;
struct _SPSData {
  GtkWidget *main_window;
  GtkWidget *menubar1;
  GtkWidget *view;
  GtkWidget *parent;  // Pointer to parent that spawned an error. e.g., data->menubar1, data->main_window, or data->traceroute_window

  char *error_text;  // String to hold error messages.
  char *warning_text;  // String to hold warning messages.

  // UTF-8 small-capital cross-reference array
  uint8_t ***utf8;  // Array holding contents of file UTF8_FILENAME: utf8[record][SMALL/CAP] = uint8_t * of size 6
  int nutf8;  // Number of lines in file UTF8_FILENAME

  // Pseudo-random number parameters
  // LA-UR-07-7961
  uint64_t RN_MULT;
  uint64_t RN_STRIDE;
  uint64_t RN_SEED;
  uint64_t RN_SEED0;
  uint64_t RN_ADD;
  uint64_t RN_MOD;
  double RN_NORM;
  uint64_t RN_PERIOD;
  uint64_t RN_MASK;
  int RN_BITS;
  uint64_t RN_ITER;

  /* The following are the definitions for variable 'type':
     Type: 0 = IPv4 TCP, 1 = IPv4 ICMP, 2 = IPv4 UDP
     Type: 3 = IPv6 TCP, 4 = IPv6 ICMP, 5 = IPv6 UDP
     6to4:
     Type: 6 = IPv6 TCP, 7 = IPv6 ICMP, 8 = IPv6 UDP

     Traceroute:
     Type: 9 = IPv4 TCP, 10 = IPv4 ICMP, 11 = IPv4 UDP
     Type: 12 = IPv6 TCP, 13 = IPv6 ICMP, 14 = IPv6 UDP
     Traceroute (6to4):
     Type: 15 = IPv6 TCP, 16 = IPv6 ICMP, 17 = IPv6 UDP

     New IPv6 headers for Auth. and ESP tunnel modes
     Type: 23 = IPv6 TCP, 24 = IPv6 ICMP, 25 = IPv6 UDP
     Type: 32 = IPv6 TCP, 33 = IPv6 ICMP, 34 = IPv6 UDP

     Special type for sending all (cycling through) IPv4 or IPv6 packets.
     Used only for data->packet_type, which is for the ipv4_send() and ipv6_send() functions.
     Type: 100 = all IPv4, 101 = all IPv6
  */

  int *protocol_type;  // Table of type values from above matched to protocol type values: protocol_type[type] = int

  // Number of packet types, as listed above
  int ntypes;  // Number of types of packet - currently 18

  // Ethernet header and interface
  Ethhdr *ethhdr;  // Struct to hold ethernet header values: ethhdr[type] = Ethhdr
  char **ifname;  // Interface name: ifname[type] = char *
  int *ifmtu;  // Interface MTU: ifmtu[type] = int;

  // IP and upper layer protocol Headers
  struct ip *ip4hdr;  // IPv4 header: ip4hdr[type] = struct ip
  struct ip6_hdr *ip6hdr;  // IPv6 header: ip6hdr[type] = struct ip6_hdr
  struct tcphdr *tcphdr;  // TCP header: tcphdr[type] = struct tcphdr
  struct icmp *icmp4hdr;  // IPv4 ICMP header: icmp4hdr[type] = struct icmp
  struct icmp6_hdr *icmp6hdr;  // IPv6 ICMP header: icmp6hdr[type] = struct icmp6_hdr
  struct udphdr *udphdr;  // UDP header: udphdr[type] = struct udphdr

  // Ethernet frames (can be a fragment). This is what gets sent to socket.
  // packet holds unfragmented ethernet frames. Used only for displaying packet.
  uint8_t ***ether_frame;  // Ethernet frame buffers: ether_frame[type][frame #] = uint8_t *
  uint8_t **packet;  // Entire unfragmented packet: packet[type] = uint8_t *
  int **frame_length;  // Length of ethernet frame: frame_length[type][frame] = int
  int *packet_length;  // Length of unfragmented packet: packet_length[type] = int
  int *nframes;  // Number of frames: nframes[type] = int

  // IPv4 options. Type definitions as above.
  int *ip_nopt;  // Number of IP options: ip_nopt[type] = int
  int *ip_opt_totlen;  // Total length of IP options: ip_opt_totlen[type] = int
  int **ip_optlen;  // IPv4 option length: ip_optlen[type][option #] = int
  uint8_t ***ip_options;  // IPv4 options data: ip_options[type][option #] = uint8_t *
  int *ip_optpadlen;  // IPv4 options padding length: ip_optpad[type] = int
  int *ip_optlenbuf;  // IPv4 option length buffer: ip_optlenbuf[type] = int
  uint8_t **ip_optionsbuf;  // IPv4 options data buffer: ip_optionsbuf[type] = uint8_t *

  // Order array for chain of IPv6 headers, extension headers, payloads, etc.
  int **order;  // Array giving order of links in IPv6 chain: order[type][link #] = int
  int *first_frag;  // Link in chain at which fragmentable portion begins: first_frag[type] = int

  // IPv6 Extension Headers

  // Packet type associated with extension header to be edited
  // 3: IPv6 TCP
  // 4: IPv6 ICMP
  // 5: IPv6 UDP
  // 12: IPv6 TCP for traceroute
  // 13: IPv6 ICMP for traceroute
  // 14: IPv6 UDP for traceroute
  int exthdr_type;  // Type of packet containing extension header to be edited

  // Hop-by-Hop header options. Type definitions as above.
  int *hop_flag;  // Flag indicating hop-by-hop header editing window is activated: hop_flag[inner/outer] = int
  int *dec_hex_hbhopt;  // Flag to indicate decimal (0) or hexadecimal (1) data entry: dec_hex_hbhopt[inner/outer]
  int *hbhopt_after;  // Option number after which to insert new option: hbhopt_after[inner/outer] = int
  int *hbhopt_remove;  // Option number to remove: hbhopt_remove[inner/outer]
  int **hbh_hdr_flag;  // Flag indicating hop-by-hop header: hbh_hdr_flag[type][inner/outer] = int
  Hopdst_hdr **hophdr;  // Hop-by-hop header (Next header and Length): hophdr[type][inner/outer] = struct Hopdst_hdr
  int **hbh_nopt;  // Number of hop-by-hop options: hbh_nopt[type][inner/outer] = int
  int ***hbh_x;  // Alignment factor x (of xN + y): hbh_x[type][inner/outer][option #] = int
  int ***hbh_y;  // Alignment factor y (of xN + y): hbh_y[type][inner/outer][option #] = int
  int **hbh_opt_totlen;  // Total length of hop-by-hop options: hbh_opt_totlen[type][inner/outer] = int
  int ***hbh_optlen;  // Hop-by-hop option length: hbh_optlen[type][inner/outer][option #] = int
  uint8_t ****hbh_options;  // Hop-by-hop options data: hbh_options[type][inner/outer][option #] = uint8_t *
  int **hbh_optpadlen;  // Hop-by-hop options padding length: hbh_optpad[type][inner/outer] = int
  int ***hbh_optleadpadlen;  // Hop-by-hop options alignment padding length: hbh_optleadpadlen[type][inner/outer][option #] = int
  uint8_t ****hbh_optleadpad;  // Hop-by-hop options alignment padding: hbh_optleadpad[type][inner/outer][option #] = uint8_t *
  int **hbh_opttailpadlen;  // Hop-by-hop options trailing padding length: hbh_opttailpadlen[type][inner/outer] = int
  uint8_t ***hbh_opttailpad;  // Hop-by-hop options trailing padding: hbh_opttailpad[type][inner/outer] = uint8_t *
  int **hbh_optlenbuf;  // Hop-by-hop option length buffer: hbh_optlenbuf[type][inner/outer] = int
  uint8_t ***hbh_optionsbuf;  // Hop-by-hop options data buffer: hbh_optionsbuf[type][inner/outer] = uint8_t *

  // Destination headers
  int *dst_flag;  // Flag indicating destination header editing window is activated: dst_flag[inner/outer] = int
  int *dec_hex_dstopt;  // Flag to indicate decimal (0) or hexadecimal (1) data entry: dec_hex_dstopt[inner/outer]
  int **dst_optlenbuf;  // Destination header (last) option length buffer: dstl_optlenbuf[type][inner/outer] = int
  uint8_t ***dst_optionsbuf;  // Destination header (last) options data buffer: dstl_optionsbuf[type][inner/outer] = uint8_t *
  int *dstopt_after;  // Option number after which to insert new option: dstopt_after[inner/outer] = int
  int *dstopt_remove;  // Option number to remove: dstopt_remove[inner/outer] = int
  int *dst_fl_flag;  // Flag indicating first or last destination header: dst_fl_flag[inner/outer] = int, 0 = first, 1 = last

  // Destination header (first) options. Type definitions as above.
  int **dstf_hdr_flag;  // Flag indicating destination header (first): dstf_hdr_flag[type][inner/outer] = int
  Hopdst_hdr **dstfhdr;  // Destination header (first) (Next header and Length): dstfhdr[type][inner/outer] = struct Hopdst_hdr
  int **dstf_nopt;  // Number of "first" destination options: dstf_nopt[type][inner/outer] = int
  int ***dstf_x;  // Alignment factor x (of xN + y): dstf_x[type][inner/outer][option #] = int
  int ***dstf_y;  // Alignment factor y (of xN + y): dstf_y[type][inner/outer][option #] = int
  int **dstf_opt_totlen;  // Total length of destination header (first) options: dstf_opt_totlen[type][inner/outer] = int
  int ***dstf_optlen;  // Destination header (first) option length: dstf_optlen[type][inner/outer][option #] = int
  uint8_t ****dstf_options;  // Destination header (first) options data: dstf_options[type][inner/outer][option #] = uint8_t *
  int **dstf_optpadlen;  // Destination header (first) options padding length: dstf_optpad[type][inner/outer] = int
  int ***dstf_optleadpadlen;  // Destination options alignment padding length: dstf_optleadpadlen[type][inner/outer][option #] = int
  uint8_t ****dstf_optleadpad;  // Destination options alignment padding: dstf_optleadpad[type][inner/outer][option #] = uint8_t *
  int **dstf_opttailpadlen;  // Destination options trailing padding length: dstf_opttailpadlen[type][inner/outer] = int
  uint8_t ***dstf_opttailpad;  // Destination options trailing padding: dstf_opttailpad[type][inner/outer] = uint8_t *

  // Destination header (last) options. Type definitions as above.
  int **dstl_hdr_flag;  // Flag indicating destination header (last): dstl_hdr_flag[type][inner/outer] = int
  Hopdst_hdr **dstlhdr;  // Destination header (last) (Next header and Length): dstlhdr[type][inner/outer] = struct Hopdst_hdr
  int **dstl_nopt;  // Number of "last" destination options: dstl_nopt[type][inner/outer] = int
  int ***dstl_x;  // Alignment factor x (of xN + y): dstl_x[type][inner/outer][option #] = int
  int ***dstl_y;  // Alignment factor y (of xN + y): dstl_y[type][inner/outer][option #] = int
  int **dstl_opt_totlen;  // Total length of destination header (last) options: dstl_opt_totlen[type][inner/outer] = int
  int ***dstl_optlen;  // Destination header (last) option length: dstl_optlen[type][inner/outer][option #] = int
  uint8_t ****dstl_options;  // Destination header (last) options data: dstl_options[type][inner/outer][option #] = uint8_t *
  int **dstl_optpadlen;  // Destination header (last) options padding length: dstl_optpad[type][inner/outer] = int
  int ***dstl_optleadpadlen;  // Destination options alignment padding length: dstf_optleadpadlen[type][inner/outer][option #] = int
  uint8_t ****dstl_optleadpad;  // Destination options alignment padding: dstf_optleadpad[type][inner/outer][option #] = uint8_t *
  int **dstl_opttailpadlen;  // Destination options trailing padding length: dstf_opttailpadlen[type][inner/outer] = int
  uint8_t ***dstl_opttailpad;  // Destination options trailing padding: dstf_opttailpad[type][inner/outer] = uint8_t *

  // Routing header. Type definitions as above.
  int *route_flag;  // Flag indicating routing header editing window is activated: route_flag[inner/outer] = int
  int *dec_hex_route;  // Flag to indicate decimal (0) or hexadecimal (1) routing data entry: dec_hex_route[inner/outer]
  int **route_hdr_flag;  // Flag indicating routing header: route_hdr_flag[type][inner/outer] = int
  Route_hdr **routehdr;  // Routing header, excluding data: routehdr[type][inner/outer] = struct Route_hdr
  int **route_datlen;  // Routing header data length: route_datlen[type][inner/outer] = int
  uint8_t ***route_data;  // Routing header data: route_data[type][inner/outer] = uint8_t *

  // Authentication header. Type definitions as above.
  int auth_flag;  // Flag indicating authentication header editing window is activated
  int dec_hex_auth;  // Flag to indicate decimal (0) or hexadecimal (1) ICV entry
  int *auth_hdr_flag;  // Flag indicating authentication header: auth_hdr_flag[type] = int
  int *auth_tr_tun_flag;  // Flag indicating transport or tunnel mode: auth_tr_tun_flag[type] = int
  Auth_hdr *authhdr;  // Authentication header: authhdr[type] = struct Auth_hdr
  int *auth_len;  // Authentication data length: auth_len[type] = int
  uint8_t **auth_data;  // Authentication data (integrity check value (ICV)): auth_data[type] = uint8_t *

  // Encapsulating security payload (ESP) header. Type definitions as above.
  int esp_flag;  // Flag indicating ESP header editing window is activated
  int dec_hex_esp;  // Flag to indicate decimal (0) or hexadecimal (1) ICV entry
  int *esp_hdr_flag;  // Flag indicating ESP header: esp_hdr_flag[type] = int
  int *esp_tr_tun_flag;  // Flag indicating transport or tunnel mode: esp_tr_tun_flag[type] = int
  Esp_hdr *esphdr;  // Encapsulating security payload header: esphdr[type] = struct Esp_hdr
  uint8_t **esp_pad;  // ESP payload padding: esp_pad[type] = uint8_t *
  Esp_tail *esptail;  // Encapsulating security payload trailer: esptail[type] = struct Esp_tail
  int *esp_auth_len;  // ESP authentication header data length: esp_auth_len[type] = int
  uint8_t **esp_auth_data;  // ESP auth. data (integrity check value (ICV)): esp_auth_data[type] = uint8_t *

  // TCP options for IPv4 and IPv6. Type definitions as above.
  int *tcp_nopt;  // Number of TCP options: tcp_nopt[type] = int
  int *tcp_opt_totlen;  // Total length of TCP options: tcp_opt_totlen[type] = int
  int **tcp_optlen;  // TCP option length: tcp_optlen[type][option #] = int
  uint8_t ***tcp_options;  // TCP options data: tcp_options[type][option #] = uint8_t *
  int *tcp_optpadlen;  // TCP options padding length: tcp_optpad[type] = int
  int *tcp_optlenbuf;  // TCP option length buffer: tcp_optlenbuf[type] = int
  uint8_t **tcp_optionsbuf;  // TCP option data buffer: tcp_optionsbuf[type] = uint8_t *

  // Upper layer protocol data for IPv4 and IPv6. Type definitions as above.
  uint8_t **payload;  // Upper layer protocol data: payload[type] = uint8_t *
  int *payloadlen;  // Upper layer protocol data length: payloadlen[type] = int

  // Regional Internet Registry (RIR) names, URLs, directories, and filenames.
  // RIR index values: 0 = arin, 1 = ripe, 2 = afrinic, 3 = apnic, 4 = lacnic
  char **rir_name;  // Regional Internet Registry (RIR) names: rir_name[RIR index] = char *
  char **rir_www;  // Regional Internet Registry (RIR) HTTP URLs: rir_www[RIR index] = char *
  char **rir_ftp;  // Regional Internet Registry (RIR) FTP URLs: rir_ftp[RIR index] = char *
  char **rir_dir;  // RIR directory location of IP delegation files: rir_dir[RIR index] = char *
  char **rir_file;  // RIR IP delegation filenames: rir_file[RIR index] = char *

  // International Organization for Standardization (ISO)'s country code filename.
  char *country_file;  // ISO's filename for country code file: country_file = char *

  // Reserved IP addresses lists
  char **ipv4_rsvd;  // Reserved IPv4 addresses
  char **ipv4_rsvd_names;  // Reserved IPv4 address names
  uint32_t *ipv4_rsvd_nbits;  // Reserved IPv4 address number of prefix mask bits
  int ipv4_rsvd_n;  // Number of IPv4 reserved addresses.
  char **ipv6_rsvd;  // Reserved IPv6 addresses with slash portion removeed
  char **ipv6_rsvd_names;  // Reserved IPv6 address names
  uint32_t *ipv6_rsvd_nbits;  // Reserved IPv6 address number of prefix mask bits
  int ipv6_rsvd_n;  // Number of IPv6 reserved addresses.

// Send page (sps.ui)

  // Select Packet Type
  GtkToggleButton *radiobutton1;  // TCP (IPv4) packet type, packet_type = 1
  GtkToggleButton *radiobutton2;  // ICMP (IPv4) packet type, packet_type = 2
  GtkToggleButton *radiobutton3;  // UDP (IPv4) packet type, packet_type = 3
  GtkToggleButton *radiobutton4;  // All (IPv4) packet types (i.e., cycle), packet_type = 4
  GtkToggleButton *radiobutton5;  // TCP (IPv6) packet type, packet_type = 5
  GtkToggleButton *radiobutton6;  // ICMP (IPv6) packet type, packet_type = 6
  GtkToggleButton *radiobutton7;  // UDP (IPv6) packet type, packet_type = 7
  GtkToggleButton *radiobutton8;  // All (IPv6) packet types (i.e., cycle), packet_type = 8
  int packet_type;  // Packet type

  // Number of Packets to Send
  GtkWidget *entry18;  // Number of packets to send
  int64_t npackets;  // Number of packets to send
  GtkToggleButton *checkbutton11;  // Flood

  // Activity Log
  GtkWidget *textview1;  // Activity log

  // IPv6 over IPv4 (6to4)
  GtkToggleButton *checkbutton12;  // Tunnel IPv6 over IPv4
  int sixto4_flag;  // Flag for state of Tunnel IPv6 over IPv4 checkbutton

  // Send / Stop
  GtkButton *button1;  // Send
  GtkButton *button5;  // Stop Flood

// TCP page (IPv4) (sps.ui)

  // Ethernet header for IPv4 TCP
  GtkWidget *entry144;  // Destination link-layer (MAC) address
  GtkWidget *entry145;  // Source link-layer (MAC) address
  GtkWidget *entry150;  // Source interface name
  GtkWidget *spinbutton5;  // Interface maximum transmission unit
  GtkWidget *entry146;  // Ethernet type code;

  // IPv4 header for IPv4 TCP
  GtkWidget *entry29;  // IP header length
  GtkWidget *entry30;  // Protocol version
  GtkWidget *entry31;  // Type of service
  GtkWidget *entry32;  // Total length of datagram
  GtkWidget *entry33;  // ID sequence number
  GtkWidget *entry122;  // Zero
  GtkWidget *entry37;  // Do not fragment flag
  GtkWidget *entry38;  // More fragments following flag
  GtkWidget *entry39;  // Fragmentation offset
  GtkWidget *entry40;  // Time-to-live
  GtkWidget *entry74;  // Transport layer protocol
  GtkWidget *entry35;  // Source IP address
  GtkToggleButton *checkbutton5;  // Source IP "Randomize"
  int ran_tcp4_sourceip;  // Flag for state of source IPv4 "Randomize" checkbutton
  GtkWidget *entry36;  // Destination IP address
  GtkWidget *entry34;  // IP checksum

  // IPv4 header options for IPv4 TCP
  GtkToggleButton *radiobutton24;  // Decimal option data entry
  GtkToggleButton *radiobutton25;  // Hexadecimal option data entry
  int dec_hex_ipopt_tcp4;  // Flag to indicate decimal (0) or hexadecimal (1) data entry
  GtkWidget *entry400;  // Option data
  GtkButton *button26;  // Insert IP option after...
  GtkWidget *entry401;  // Option number after which to insert new option
  int ipopt_tcp4_after;  // Option number after which to insert new option
  GtkButton *button60;  // Append IP option
  GtkWidget *textview30;  // Number of IP options in packet
  GtkButton *button61;  // View IP options
  GtkWidget *entry403;  // Option number to remove
  int ipopt_tcp4_remove;  // Option number to remove
  GtkButton *button62;  // Remove option

  // TCP header for IPv4 TCP
  GtkWidget *entry41;  // Source port number
  GtkToggleButton *checkbutton7;  // Source port "Randomize"
  int ran_tcp4_sourceport;  // Flag for state of source port "Randomize" checkbutton (IPv4)
  GtkWidget *entry42;  // Destination port number
  GtkWidget *entry43;  // Sequence number
  GtkWidget *entry44;  // Ack number
  GtkWidget *entry45;  // Reserved
  GtkWidget *entry46;  // Data offset
  GtkWidget *entry47;  // FIN flag
  GtkWidget *entry48;  // SYN flag
  GtkWidget *entry49;  // RST flag
  GtkWidget *entry50;  // PSH flag
  GtkWidget *entry51;  // ACK flag
  GtkWidget *entry52;  // URG flag
  GtkWidget *entry53;  // ECE flag
  GtkWidget *entry54;  // CWR flag
  GtkWidget *entry55;  // Window size
  GtkWidget *entry56;  // TCP checksum
  GtkWidget *entry77;  // Urgent pointer
  GtkToggleButton *radiobutton17;  // ASCII data entry
  GtkToggleButton *radiobutton18;  // Hexadecimal data entry
  int ascii_hex_tcp4;  // Flag to indicate ASCII (0) or hexadecimal (1) data entry
  GtkWidget *entry162;  // TCP payload data from keyboard
  GtkButton *button20;  // TCP payload data from file
  GtkWidget *textview60;  // File Loaded indication
  GtkButton *button200;  // Unload file data from packet

  // TCP header options for IPv4 TCP
  GtkToggleButton *radiobutton41;  // Decimal option data entry
  GtkToggleButton *radiobutton42;  // Hexadecimal option data entry
  int dec_hex_tcpopt_tcp4;  // Flag to indicate decimal (0) or hexadecimal (1) data entry
  GtkWidget *entry404;  // Option data
  GtkButton *button47;  // Insert IP option after...
  GtkWidget *entry405;  // Option number after which to insert new option
  int tcpopt_tcp4_after;  // Option number after which to insert new option
  GtkButton *button48;  // Append TCP option
  GtkWidget *textview24;  // Number of TCP options in packet
  GtkButton *button49;  // View TCP options
  GtkWidget *entry406;  // Option number to remove
  int tcpopt_tcp4_remove;  // Option number to remove
  GtkButton *button58;  // Remove option

  GtkButton *button3;  // View Packet
  GtkButton *button11;  // Restore default settings

// ICMP page (IPv4) (sps.ui)

  // Ethernet header for IPv4 ICMP
  GtkWidget *entry141;  // Destination link-layer (MAC) address
  GtkWidget *entry142;  // Source link-layer (MAC) address
  GtkWidget *entry151;  // Source interface name
  GtkWidget *spinbutton6;  // Interface maximum transmission unit
  GtkWidget *entry143;  // Ethernet type code;

  // IPv4 header for IPv4 ICMP
  GtkWidget *entry57;  // IP header length
  GtkWidget *entry58;  // Protocol version
  GtkWidget *entry59;  // Type of service
  GtkWidget *entry60;  // Total length of datagram
  GtkWidget *entry61;  // ID sequence number
  GtkWidget *entry123;  // Zero
  GtkWidget *entry65;  // Do not fragment flag
  GtkWidget *entry66;  // More fragments following flag
  GtkWidget *entry67;  // Fragmentation offset
  GtkWidget *entry68;  // Time-to-live
  GtkWidget *entry75;  // Transport layer protocol
  GtkWidget *entry63;  // Source IP address
  GtkToggleButton *checkbutton9;  // Source IP "Randomize"
  int ran_icmp4_sourceip;  // Flag for state of source IPv4 "Randomize" checkbutton
  GtkWidget *entry64;  // Destination IP address
  GtkWidget *entry62;  // IP checksum

  // IPv4 header options for IPv4 ICMP
  GtkToggleButton *radiobutton43;  // Decimal option data entry
  GtkToggleButton *radiobutton44;  // Hexadecimal option data entry
  int dec_hex_ipopt_icmp4;  // Flag to indicate decimal (0) or hexadecimal (1) data entry
  GtkWidget *entry407;  // Option data
  GtkButton *button59;  // Insert IP option after...
  GtkWidget *entry408;  // Option number after after which to insert new option
  int ipopt_icmp4_after;  // Option number after which to insert new option
  GtkButton *button63;  // Append IP option
  GtkWidget *textview26;  // Number of IP options in packet
  GtkButton *button64;  // View IP options
  GtkWidget *entry409;  // Option number to remove
  int ipopt_icmp4_remove;  // Option number to remove
  GtkButton *button65;  // Remove option

  // ICMP header for IPv4 ICMP
  GtkWidget *entry69;  // Message type
  GtkWidget *entry70;  // Message code
  GtkWidget *entry71;  // ICMP checksum
  GtkWidget *entry72;  // Identifier
  GtkWidget *entry73;  // Sequence number
  GtkToggleButton *radiobutton9;  // ASCII data entry
  GtkToggleButton *radiobutton10;  // Hexadecimal data entry
  int ascii_hex_icmp4;  // Flag to indicate ASCII (0) or hexadecimal (1) data entry
  GtkWidget *entry126;  // ICMP payload data from keyboard
  GtkButton *button7;  // ICMP payload data from file
  GtkWidget *textview62;  // File Loaded indication
  GtkButton *button202;  // Unload file data from packet

  GtkButton *button2;  // View Packet
  GtkButton *button12;  // Restore default settings

// UDP page (IPv4) (sps.ui)

  // Ethernet header for IPv4 UDP
  GtkWidget *entry147;  // Destination link-layer (MAC) address
  GtkWidget *entry148;  // Source link-layer (MAC) address
  GtkWidget *entry152;  // Source interface name
  GtkWidget *spinbutton7;  // Interface maximum transmission unit
  GtkWidget *entry149;  // Ethernet type code;

  // IPv4 header for IPv4 UDP
  GtkWidget *entry1;  // IP header length
  GtkWidget *entry2;  // Protocol version
  GtkWidget *entry3;  // Type of service
  GtkWidget *entry4;  // Total length of datagram
  GtkWidget *entry5;  // ID sequence number
  GtkWidget *entry124;  // Zero
  GtkWidget *entry9;  // Do not fragment flag
  GtkWidget *entry10;  // More fragments following flag
  GtkWidget *entry11;  // Fragmentation offset
  GtkWidget *entry12;  // Time-to-live
  GtkWidget *entry76;  // Transport layer protocol
  GtkWidget *entry7;  // Source IP address
  GtkToggleButton *checkbutton1;  // Source IP "Randomize"
  int ran_udp4_sourceip;  // Flag for state of source IPv4 "Randomize" checkbutton
  GtkWidget *entry8;  // Destination IP address
  GtkWidget *entry6;  // IP checksum

  // IPv4 header options for IPv4 UDP
  GtkToggleButton *radiobutton45;  // Decimal option data entry
  GtkToggleButton *radiobutton46;  // Hexadecimal option data entry
  int dec_hex_ipopt_udp4;  // Flag to indicate decimal (0) or hexadecimal (1) data entry
  GtkWidget *entry410;  // Option data
  GtkButton *button66;  // Insert IP option after...
  GtkWidget *entry411;  // Option number after which to insert new option
  int ipopt_udp4_after;  // Option number after which to insert new option
  GtkButton *button67;  // Append IP option
  GtkWidget *textview28;  // Number of IP options in packet
  GtkButton *button68;  // View IP options
  GtkWidget *entry412;  // Option number to remove
  int ipopt_udp4_remove;  // Option number to remove
  GtkButton *button69;  // Remove option

  // UDP header for IPv4 UDP
  GtkWidget *entry13;  // Source port
  GtkToggleButton *checkbutton3;  // Source port "Randomize"
  int ran_udp4_sourceport;  // Flag for state of source port "Randomize" checkbutton (IPv4)
  GtkWidget *entry14;  // Destination port
  GtkWidget *entry15;  // Length of UDP datagram
  GtkWidget *entry16;  // UDP checksum
  GtkToggleButton *radiobutton13;  // ASCII data entry
  GtkToggleButton *radiobutton14;  // Hexadecimal data entry
  int ascii_hex_udp4;  // Flag to indicate ASCII (0) or hexadecimal (1) data entry
  GtkWidget *entry17;  // UDP payload data from keyboard
  GtkButton *button8;  // UDP payload data from file
  GtkWidget *textview64;  // File Loaded indication
  GtkButton *button204;  // Unload file data from packet

  GtkButton *button4;  // View Packet
  GtkButton *button15;  // Restore default settings

// TCP page (IPv6) (sps.ui)

  // Ethernet header for IPv6 TCP
  GtkWidget *entry131;  // Destination link-layer (MAC) address
  GtkWidget *entry132;  // Source link-layer (MAC) address
  GtkWidget *entry154;  // Source interface name
  GtkWidget *spinbutton8;  // Interface maximum transmission unit
  GtkWidget *entry133;  // Ethernet type code;

  // IPv6 header for IPv6 TCP
  GtkWidget *entry22;  // IP version
  GtkWidget *entry23;  // Traffic class
  GtkWidget *entry24;  // Flow label
  GtkWidget *entry25;  // Payload length
  GtkWidget *entry26;  // Next header
  GtkWidget *entry27;  // Hop limit
  GtkWidget *entry28;  // Source IP address
  GtkToggleButton *checkbutton2;  // Source IPv6 address "Randomize"
  int ran_tcp6_sourceip;  // Flag for state of source IPv6 "Randomize" checkbutton
  GtkWidget *entry78;  // Destination IP address

  // TCP header for IPv6 TCP
  GtkWidget *entry79;  // Source port number
  GtkToggleButton *checkbutton4;  // Source port "Randomize"
  int ran_tcp6_sourceport;  // Flag for state of source port "Randomize" checkbutton (IPv6)
  GtkWidget *entry80;  // Destination port number
  GtkWidget *entry81;  // Sequence number
  GtkWidget *entry82;  // Ack number
  GtkWidget *entry83;  // Reserved
  GtkWidget *entry84;  // Data offset
  GtkWidget *entry85;  // FIN flag
  GtkWidget *entry86;  // SYN flag
  GtkWidget *entry87;  // RST flag
  GtkWidget *entry88;  // PSH flag
  GtkWidget *entry89;  // ACK flag
  GtkWidget *entry90;  // URG flag
  GtkWidget *entry91;  // ECE flag
  GtkWidget *entry92;  // CWR flag
  GtkWidget *entry93;  // Window size
  GtkWidget *entry94;  // TCP checksum
  GtkWidget *entry95;  // Urgent pointer
  GtkToggleButton *radiobutton19;  // ASCII data entry
  GtkToggleButton *radiobutton20;  // Hexadecimal data entry
  int ascii_hex_tcp6;  // Flag to indicate ASCII (0) or hexadecimal (1) data entry
  GtkWidget *entry163;  // TCP payload data from keyboard
  GtkButton *button22;  // TCP payload data from file
  GtkWidget *textview61;  // File Loaded indication
  GtkButton *button201;  // Unload file data from packet

  // TCP options for IPv6 TCP
  GtkToggleButton *radiobutton47;  // Decimal option data entry
  GtkToggleButton *radiobutton48;  // Hexadecimal option data entry
  int dec_hex_tcpopt_tcp6;  // Flag to indicate decimal (0) or hexadecimal (1) data entry
  GtkWidget *entry413;  // Option data
  GtkButton *button70;  // Insert IP option after...
  GtkWidget *entry414;  // Option number after which to insert new option
  int tcpopt_tcp6_after;  // Option number after which to insert new option
  GtkButton *button71;  // Append TCP option
  GtkWidget *textview32;  // Number of TCP options in packet
  GtkButton *button72;  // View TCP options
  GtkWidget *entry415;  // Option number to remove
  int tcpopt_tcp6_remove;  // Option number to remove
  GtkButton *button73;  // Remove option

  // Extension headers
  GtkButton *button98;  // Hop-by-hop extension header
  GtkButton *button110;  // Destination extension header
  GtkButton *button132;  // Routing extension header
  GtkButton *button116;  // Authentication extension header
  GtkButton *button126;  // Encapsulating security payload (ESP) extension header

  // Outer extension headers for AH or ESP tunneling
  GtkButton *button300;  // Hop-by-hop extension header
  GtkButton *button301;  // Destination extension header
  GtkButton *button302;  // Routing extension header

  GtkButton *button6;  // View Packet
  GtkButton *button16;  // Restore default settings

// ICMP page (IPv6) (sps.ui)

  // Ethernet header for IPv6 ICMP
  GtkWidget *entry128;  // Destination link-layer (MAC) address
  GtkWidget *entry129;  // Source link-layer (MAC) address
  GtkWidget *entry153;  // Source interface name
  GtkWidget *spinbutton9;  // Interface maximum transmission unit
  GtkWidget *entry130;  // Ethernet type code

  // IPv6 header for IPv6 ICMP
  GtkWidget *entry110;  // IP version
  GtkWidget *entry111;  // Traffic class
  GtkWidget *entry112;  // Flow label
  GtkWidget *entry113;  // Payload length
  GtkWidget *entry114;  // Next header
  GtkWidget *entry115;  // Hop limit
  GtkWidget *entry109;  // Source IP address
  GtkToggleButton *checkbutton10;  // Source IPv6 address "Randomize"
  int ran_icmp6_sourceip;  // Flag for state of source IPv6 "Randomize" checkbutton
  GtkWidget *entry116;  // Destination IP address

  // ICMP header for IPv6 ICMP
  GtkWidget *entry117;  // Message type
  GtkWidget *entry118;  // Message code
  GtkWidget *entry119;  // ICMP checksum
  GtkWidget *entry120;  // Identifier
  GtkWidget *entry121;  // Sequence number
  GtkToggleButton *radiobutton11;  // ASCII data entry
  GtkToggleButton *radiobutton12;  // Hexadecimal data entry
  int ascii_hex_icmp6;  // Flag to indicate ASCII (0) or hexadecimal (1) data entry
  GtkWidget *entry127;  // ICMP payload data from keyboard
  GtkButton *button9;  // ICMP payload data from file
  GtkWidget *textview63;  // File Loaded indication
  GtkButton *button203;  // Unload file data from packet

  // Extension headers
  GtkButton *button99;  // Hop-by-hop extension header
  GtkButton *button111;  // Destination extension header
  GtkButton *button133;  // Routing extension header
  GtkButton *button117;  // Authentication extension header
  GtkButton *button127;  // Encapsulating security payload (ESP) extension header

  // Outer extension headers for AH or ESP tunneling
  GtkButton *button303;  // Hop-by-hop extension header
  GtkButton *button304;  // Destination extension header
  GtkButton *button305;  // Routing extension header

  GtkButton *button17;  // View Packet
  GtkButton *button18;  // Restore default settings

// UDP page (IPv6) (sps.ui)

  // Ethernet header for IPv6 UDP
  GtkWidget *entry134;  // Destination link-layer (MAC) address
  GtkWidget *entry135;  // Source link-layer (MAC) address
  GtkWidget *entry155;  // Source interface name
  GtkWidget *spinbutton10;  // Interface maximum transmission unit
  GtkWidget *entry136;  // Ethernet type code;

  // IPv6 header for IPv6 UDP
  GtkWidget *entry97;  // IP version
  GtkWidget *entry98;  // Traffic class
  GtkWidget *entry99;  // Flow label
  GtkWidget *entry100;  // Payload length
  GtkWidget *entry101;  // Next header
  GtkWidget *entry102;  // Hop limit
  GtkWidget *entry96;  // Source IP address
  GtkToggleButton *checkbutton6;  // Source IPv6 address "Randomize"
  int ran_udp6_sourceip;  // Flag for state of source IPv6 "Randomize" checkbutton
  GtkWidget *entry103;  // Destination IP address

  // UDP header for IPv6 UDP
  GtkWidget *entry104;  // Source port
  GtkToggleButton *checkbutton8;  // Source port "Randomize"
  int ran_udp6_sourceport;  // Flag for state of source port "Randomize" checkbutton (IPv6)
  GtkWidget *entry105;  // Destination port
  GtkWidget *entry106;  // Length of UDP datagram
  GtkWidget *entry107;  // UDP checksum
  GtkToggleButton *radiobutton15;  // ASCII data entry
  GtkToggleButton *radiobutton16;  // Hexadecimal data entry
  int ascii_hex_udp6;  // Flag to indicate ASCII (0) or hexadecimal (1) data entry
  GtkWidget *entry108;  // UDP payload data from keyboard
  GtkButton *button10;  // UDP payload data from file
  GtkWidget *textview65;  // File Loaded indication
  GtkButton *button205;  // Unload file data from packet

  // Extension headers
  GtkButton *button100;  // Hop-by-hop extension header
  GtkButton *button112;  // Destination extension header
  GtkButton *button134;  // Routing extension header
  GtkButton *button118;  // Authentication extension header
  GtkButton *button128;  // Encapsulating security payload (ESP) extension header

  // Outer extension headers for AH or ESP tunneling
  GtkButton *button306;  // Hop-by-hop extension header
  GtkButton *button307;  // Destination extension header
  GtkButton *button308;  // Routing extension header

  GtkButton *button13;  // View Packet
  GtkButton *button19;  // Restore default settings

// 6to4 page (sps.ui)

// 6to4 TCP (IPv4) (sps.ui)

  // Ethernet header for IPv4 TCP

  GtkWidget *entry600;  // Destination link-layer (MAC) address
  GtkWidget *entry601;  // Source link-layer (MAC) address
  GtkWidget *entry460;  // TCP interface name
  GtkWidget *spinbutton17;  // TCP Interface maximum transmission unit
  GtkWidget *entry602;  // Ethernet type code

  // IPv4 header for IPv4 TCP
  GtkWidget *entry609;  // IP header length
  GtkWidget *entry610;  // Protocol version
  GtkWidget *entry611;  // Type of service
  GtkWidget *entry612;  // Total length of datagram
  GtkWidget *entry613;  // ID sequence number
  GtkWidget *entry614;  // Zero
  GtkWidget *entry615;  // Do not fragment flag
  GtkWidget *entry616;  // More fragments following flag
  GtkWidget *entry617;  // Fragmentation offset
  GtkWidget *entry618;  // Time-to-live
  GtkWidget *entry619;  // Transport layer protocol
  GtkWidget *entry125;  // Source IP address
  GtkToggleButton *checkbutton13;  // Source IP "Randomize"
  int ran_tcp6to4_sourceip;  // Flag for state of TCP source IPv4 "Randomize" for 6to4 packets.
  GtkWidget *entry620;  // Destination IP address
  GtkWidget *entry621;  // IP checksum

// 6to4 ICMP (IPv4) (sps.ui)

  // Ethernet header for IPv4 ICMP
  GtkWidget *entry603;  // Destination link-layer (MAC) address
  GtkWidget *entry604;  // Source link-layer (MAC) address
  GtkWidget *entry461;  // ICMP interface name
  GtkWidget *spinbutton18;  // ICMP Interface maximum transmission unit
  GtkWidget *entry605;  // Ethernet type code

  // IPv4 header for IPv4 ICMP
  GtkWidget *entry622;  // IP header length
  GtkWidget *entry623;  // Protocol version
  GtkWidget *entry624;  // Type of service
  GtkWidget *entry625;  // Total length of datagram
  GtkWidget *entry626;  // ID sequence number
  GtkWidget *entry627;  // Zero
  GtkWidget *entry628;  // Do not fragment flag
  GtkWidget *entry629;  // More fragments following flag
  GtkWidget *entry630;  // Fragmentation offset
  GtkWidget *entry631;  // Time-to-live
  GtkWidget *entry632;  // Transport layer protocol
  GtkWidget *entry160;  // Source IP address
  GtkToggleButton *checkbutton26;  // Source IP "Randomize"
  int ran_icmp6to4_sourceip;  // Flag for state of ICMP source IPv4 "Randomize" for 6to4 packets
  GtkWidget *entry633;  // Destination IP address
  GtkWidget *entry634;  // IP checksum

// 6to4 UDP (IPv4) (sps.ui)

  // Ethernet header for IPv4 UDP
  GtkWidget *entry606;  // Destination link-layer (MAC) address
  GtkWidget *entry607;  // Source link-layer (MAC) address
  GtkWidget *entry462;  // UDP interface name
  GtkWidget *spinbutton19;  // UDP Interface maximum transmission unit
  GtkWidget *entry608;  // Ethernet type code

  // IPv4 header for IPv4 UDP
  GtkWidget *entry635;  // IP header length
  GtkWidget *entry636;  // Protocol version
  GtkWidget *entry637;  // Type of service
  GtkWidget *entry638;  // Total length of datagram
  GtkWidget *entry639;  // ID sequence number
  GtkWidget *entry640;  // Zero
  GtkWidget *entry641;  // Do not fragment flag
  GtkWidget *entry642;  // More fragments following flag
  GtkWidget *entry643;  // Fragmentation offset
  GtkWidget *entry644;  // Time-to-live
  GtkWidget *entry645;  // Transport layer protocol
  GtkWidget *entry161;  // UDP source IPv4 address for 6to4 packets
  GtkToggleButton *checkbutton27;  // UDP source IPv4 "Randomize" for 6to4 packets
  int ran_udp6to4_sourceip;  // Flag for state of UDP source IPv4 "Randomize" for 6to4 packets
  GtkWidget *entry646;  // Destination IP address
  GtkWidget *entry647;  // IP checksum

// Host/IP page (sps.ui)
  GtkWidget *entry19;  // Host-to-IP: hostname
  GtkWidget *textview3;  // Host-to-IPv4 results
  GtkWidget *textview6;  // Host-to-IPv6 results

  GtkWidget *entry20;  // IP-to-Host IPv4 address
  GtkWidget *entry21;  // IP-to-Host IPv6 address
  GtkWidget *textview2;  // IPv4-to-Host hostname result
  GtkWidget *textview29;  // IPv6-to-Host hostname result

  GtkWidget *textview4;  // This Host hostname
  GtkWidget *textview23;  // This Host interface names and IPv4 addresses
  GtkWidget *textview5;  // This Host interface names and IPv6 addresses

  GtkWidget *entry139;  // This host interface name
  GtkWidget *textview10;  // This host interface MAC address
  GtkWidget *textview21;  // This host interface current MTU

  GtkWidget *entry140;  // IP-to-MAC source interface name
  GtkWidget *spinbutton1;  // Timeout for ARP/Neighbor discovery
  int timeout;  // Timeout value (s) for ARP/Neighbor discovery
  GtkWidget *entry137;  // IP-to-MAC IPv4 address
  GtkWidget *entry138;  // IP-to-MAC IPv6 address
  GtkButton *button146;  // Send ARP request.
  GtkButton *button147;  // Send neighbor solicitation.
  GtkWidget *textview11;  // IP-to-MAC results

// Traceroute page (sps.ui)

  GtkWidget *traceroute_window;  // Traceroute edit packet window
  int traceroute_flag;  // Flag indicating traceroute packet-editing window is activated

  // Probe Protocol
  GtkToggleButton *radiobutton21;  // TCP (IPv4) probe packet
  GtkToggleButton *radiobutton22;  // ICMP (IPv4) probe packet
  GtkToggleButton *radiobutton23;  // UDP (IPv4) probe packet
  GtkToggleButton *radiobutton26;  // TCP (IPv6) probe packet
  GtkToggleButton *radiobutton27;  // ICMP (IPv6) probe packet
  GtkToggleButton *radiobutton28;  // UDP (IPv6) probe packet
  int packet_type_tr;  // Packet type

  // Traceroute 6to4
  GtkToggleButton *checkbutton29;  // Tunnel IPv6 over IPv4
  int sixto4_tr_flag;  // Flag for state of Tunnel IPv6 over IPv4 checkbutton

  // Number of Probes per Hop
  GtkWidget *spinbutton2;  // Number of probes per hop
  int num_probes;  // Number of probes per hop

  // Timeout for Reply
  GtkWidget *spinbutton4;  // Timeout for reply
  int timeout_tr;  // Timeout for reply

  // Maximum number of hops
  GtkWidget *spinbutton3;  // Maximum number of hops
  int maxhops;  // Maximum number of hops

  // Resolve Node Hostnames
  GtkToggleButton *checkbutton28;
  int resolve_tr;

  // Edit, Start, Stop buttons
  GtkButton *button23;  // Edit packets
  GtkButton *button24;  // Start
  GtkButton *button25;  // Stop

  // Activity Log
  GtkWidget *textview7;  // Traceroute activity log
  char *tr_message;  // String to contain traceroute results

// IP / ASN delegation page (sps.ui)
  int *rir_file_loaded;  // Flags indicating which RIR files are loaded.
  GtkWidget *textview8;  // Status of ARIN delegation file.
  GtkButton *button50;  // Download ARIN delegation file.
  GtkWidget *textview9;  // Status of RIPE delegation file.
  GtkButton *button51;  // Download RIPE delegation file.
  GtkWidget *textview12;  // Status of AFRINIC delegation file.
  GtkButton *button52;  // Download AFRINIC delegation file.
  GtkWidget *textview13;  // Status of APNIC delegation file.
  GtkButton *button53;  // Download APNIC delegation file.
  GtkWidget *textview14;  // Status of LACNIC delegation file.
  GtkButton *button54;  // Download LACNIC delegation file.
  GtkButton *button56;  // View ISO 3166-1 country code file.
  GtkButton *button57;  // Download all RIR files.
  GtkToggleButton *checkbutton38;  // Flag to show sent and received messages during file downloads
  int ftp_family;  // Address family for FTP connection to RIRs.
  int sd_data;  // Socket descriptor for FTP data connection.
  char *ftp_data;  // Array to hold FTP data received from remote host.
  int rir_index;  // Index to strings containing RIR names, URLs, directories, and file names.
  int rir_done;  // Index of RIR delegation that just completed attempted download.

  int *nrecords;  // Number of records in each RIR's delegation file: nrecords[rir] = int
  int ncountries;  // Number of countries in www.iso.org's country code list.
  uint8_t ****array;  // Array containing all RIR delegation files: array[rir][record][field] = uint8_t *
  uint8_t ***cc;  // Array containing www.iso.org's country code list: cc[record][field] = char *
  GtkWidget *entry500;  // Country name for country name/code search.
  GtkWidget *entry501;  // Country code for country name/code search.
  GtkWidget *entry502;  // Country name for ASN search.
  GtkWidget *entry503;  // Country code for ASN search.
  GtkWidget *textview16;  // ASN delegation results for ASN delegation search.
  char *asn_results;  // ASN delegation results for ASN delegation search.
  GtkWidget *entry504;  // ASN for reverse-search for delegation of ASN.
  uint64_t asn;  // ASN for reverse-search for delegation of ASN.
  GtkWidget *textview17;  // Results of reverse-search for delegation of ASN.
  GtkWidget *entry505;  // Country name for IP delegation search.
  char *cname;  // Country name for IP delegation search.
  GtkWidget *entry506;  // Country code for IP delegation search.
  char *ccode;  // Country code for IP delegation search.
  GtkWidget *textview18;  // IPv4 address delegation results for IP delegation search.
  char *ip4_results;  // IPv4 address delegation results for IP delegation search.
  GtkWidget *textview19;  // IPv6 address delegation results for IP delegation search.
  char *ip6_results;  // IPv6 address delegation results for IP delegation search.
  GtkWidget *entry507;  // IPv4 address for IP delegation reverse-search.
  char *ip4string;  // IPv4 address for IP delegation reverse_search.
  GtkWidget *entry508;  // IPv6 address for IP delegation reverse-search.
  char *ip6string;  // IPv6 address for IP delegation reverse_search.
  GtkWidget *textview20;  // IPv4 or IPv6  address delegation results for IP delegation reverse-search.

// Load/Save page

  // Load file
  GtkButton *button21;  // Load file...

  // Save

  uint8_t *save_flags;  // Flags to indicate which items to save to SPS file
  int save_nitems;  // Number of possible items to save

  // Send Packet settings
  GtkToggleButton *checkbutton25;  // Packet type, Number of packets to send, Flood mode

  // IPv6 over IPv4 (6to4)
  GtkToggleButton *checkbutton15;  // IPv4 Ethernet header and IPv4 header

  // IPv4
  GtkToggleButton *checkbutton16;  // TCP packet (ethernet header, IP header, and TCP header)
  GtkToggleButton *checkbutton20;  // ICMP packet (ethernet header, IP header, ICMP header, and ICMP data)
  GtkToggleButton *checkbutton21;  // UDP packet (ethernet header, IP header, UDP header, and UDP data)

  // IPv6
  GtkToggleButton *checkbutton22;  // TCP packet (ethernet header, IP header, and TCP header)
  GtkToggleButton *checkbutton23;  // ICMP packet (ethernet header, IP header, ICMP header, and ICMP data)
  GtkToggleButton *checkbutton24;  // UDP packet (ethernet header, IP header, UDP header, and UDP data)

  // 6to4
  GtkToggleButton *checkbutton14;  // 6to4 TCP (ethernet header, IPv4 header, IPv4 options + padding)
  GtkToggleButton *checkbutton17;  // 6to4 ICMP (ethernet header, IPv4 header, IPv4 options + padding)
  GtkToggleButton *checkbutton18;  // 6to4 UDP (ethernet header, IPv4 header, IPv4 options + padding)

  // Traceroute settings
  GtkToggleButton *checkbutton36;  // Traceroute settings

  // IPv6 over IPv4 (6to4) for traceroute
  GtkToggleButton *checkbutton46;  // IPv4 Ethernet header and IPv4 header for traceroute

  // IPv4 traceroute
  GtkToggleButton *checkbutton39;  // TCP packet for traceroute
  GtkToggleButton *checkbutton41;  // ICMP packet for traceroute
  GtkToggleButton *checkbutton42;  // UDP packet for traceroute

  // IPv6 Traceroute
  GtkToggleButton *checkbutton43;  // TCP packet for traceroute
  GtkToggleButton *checkbutton44;  // ICMP packet for traceroute
  GtkToggleButton *checkbutton45;  // UDP packet for traceroute

  // Save file
  GtkButton *button14;  // Save SPS file, which contains settings selected by user

// Pseudo-random number generator page (sps.ui)
  GtkWidget *entry319;  // Seed for PRNG
  GtkWidget *entry320;  // Stride for PRNG

// Traceroute TCP page (IPv4) (traceroute.ui)

  GtkButton *tr_button500;  // Copy all data from main TCP (IPv4) page

  // Ethernet header for IPv4 TCP for traceroute
  GtkWidget *tr_entry164;  // Destination link-layer (MAC) address
  GtkWidget *tr_entry165;  // Source link-layer (MAC) address
  GtkWidget *tr_entry166;  // Source interface name
  GtkWidget *tr_spinbutton11;  // Interface maximum transmission unit
  GtkWidget *tr_entry167;  // Ethernet type code;

  // IPv4 header for IPv4 TCP for traceroute
  GtkWidget *tr_entry168;  // IP header length
  GtkWidget *tr_entry169;  // Protocol version
  GtkWidget *tr_entry170;  // Type of service
  GtkWidget *tr_entry171;  // Total length of datagram
  GtkWidget *tr_entry172;  // ID sequence number
  GtkWidget *tr_entry173;  // Zero
  GtkWidget *tr_entry174;  // Do not fragment flag
  GtkWidget *tr_entry175;  // More fragments following flag
  GtkWidget *tr_entry176;  // Fragmentation offset
  GtkWidget *tr_entry177;  // Time-to-live
  GtkWidget *tr_entry178;  // Transport layer protocol
  GtkWidget *tr_entry179;  // Source IP address
  GtkWidget *tr_entry180;  // Destination IP address
  GtkWidget *tr_entry181;  // IP checksum

  // IPv4 header options for IPv4 TCP traceroute
  GtkToggleButton *tr_radiobutton49;  // Decimal option data entry
  GtkToggleButton *tr_radiobutton50;  // Hexadecimal option data entry
  int dec_hex_ipopt_tcp4_tr;  // Flag to indicate decimal (0) or hexadecimal (1) data entry
  GtkWidget *tr_entry416;  // Option data
  GtkButton *tr_button74;  // Insert IP option after...
  GtkWidget *tr_entry417;  // Option number after which to insert new option
  int ipopt_tcp4_tr_after;  // Option number after which to insert new option
  GtkButton *tr_button75;  // Append IP option
  GtkWidget *tr_textview34;  // Number of IP options in packet
  GtkButton *tr_button76;  // View IP options
  GtkWidget *tr_entry418;  // Option number to remove
  int ipopt_tcp4_tr_remove;  // Option number to remove
  GtkButton *tr_button77;  // Remove option

  // TCP header for IPv4 TCP for traceroute
  GtkWidget *tr_entry182;  // Source port number
  GtkToggleButton *tr_checkbutton30;  // Source port "Randomize"
  int ran_tcp4_tr_sourceport;  // Flag for state of source port "Randomize" checkbutton (IPv4)
  GtkWidget *tr_entry183;  // Destination port number
  GtkWidget *tr_entry184;  // Sequence number
  GtkWidget *tr_entry185;  // Ack number
  GtkWidget *tr_entry186;  // Reserved
  GtkWidget *tr_entry187;  // Data offset
  GtkWidget *tr_entry188;  // FIN flag
  GtkWidget *tr_entry189;  // SYN flag
  GtkWidget *tr_entry190;  // RST flag
  GtkWidget *tr_entry191;  // PSH flag
  GtkWidget *tr_entry192;  // ACK flag
  GtkWidget *tr_entry193;  // URG flag
  GtkWidget *tr_entry194;  // ECE flag
  GtkWidget *tr_entry195;  // CWR flag
  GtkWidget *tr_entry196;  // Window size
  GtkWidget *tr_entry197;  // TCP checksum
  GtkWidget *tr_entry198;  // Urgent pointer
  GtkToggleButton *tr_radiobutton29;  // ASCII data entry
  GtkToggleButton *tr_radiobutton30;  // Hexadecimal data entry
  int ascii_hex_tcp4_tr;  // Flag to indicate ASCII (0) or hexadecimal (1) data entry
  GtkWidget *tr_entry199;  // TCP payload data from keyboard
  GtkButton *tr_button29;  // TCP payload data from file
  GtkWidget *tr_textview66;  // File Loaded indication
  GtkButton *tr_button206;  // Unload file data from packet

  // TCP options for IPv4 TCP for traceroute
  GtkToggleButton *tr_radiobutton51;  // Decimal option data entry
  GtkToggleButton *tr_radiobutton52;  // Hexadecimal option data entry
  int dec_hex_tcpopt_tcp4_tr;  // Flag to indicate decimal (0) or hexadecimal (1) data entry
  GtkWidget *tr_entry419;  // Option data
  GtkButton *tr_button78;  // Insert IP option after...
  GtkWidget *tr_entry420;  // Option number after which to insert new option
  int tcpopt_tcp4_tr_after;  // Option number after which to insert new option
  GtkButton *tr_button79;  // Append TCP option
  GtkWidget *tr_textview36;  // Number of TCP options in packet
  GtkButton *tr_button80;  // View TCP options
  GtkWidget *tr_entry421;  // Option number to remove
  int tcpopt_tcp4_tr_remove;  // Option number to remove
  GtkButton *tr_button81;  // Remove option

  GtkButton *tr_button28;  // View Packet
  GtkButton *tr_button30;  // Restore default settings

// Traceroute ICMP page (IPv4) (traceroute.ui)

  GtkButton *tr_button501;  // Copy all data from main ICMP (IPv4) page

  // Ethernet header for IPv4 ICMP for traceroute
  GtkWidget *tr_entry200;  // Destination link-layer (MAC) address
  GtkWidget *tr_entry201;  // Source link-layer (MAC) address
  GtkWidget *tr_entry202;  // Source interface name
  GtkWidget *tr_spinbutton12;  // Interface maximum transmission unit
  GtkWidget *tr_entry203;  // Ethernet type code;

  // IPv4 header for IPv4 ICMP for traceroute
  GtkWidget *tr_entry204;  // IP header length
  GtkWidget *tr_entry205;  // Protocol version
  GtkWidget *tr_entry206;  // Type of service
  GtkWidget *tr_entry207;  // Total length of datagram
  GtkWidget *tr_entry208;  // ID sequence number
  GtkWidget *tr_entry209;  // Zero
  GtkWidget *tr_entry210;  // Do not fragment flag
  GtkWidget *tr_entry211;  // More fragments following flag
  GtkWidget *tr_entry212;  // Fragmentation offset
  GtkWidget *tr_entry213;  // Time-to-live
  GtkWidget *tr_entry214;  // Transport layer protocol
  GtkWidget *tr_entry215;  // Source IP address
  GtkWidget *tr_entry216;  // Destination IP address
  GtkWidget *tr_entry217;  // IP checksum

  // IPv4 header options for IPv4 ICMP for traceroute
  GtkToggleButton *tr_radiobutton55;  // Decimal option data entry
  GtkToggleButton *tr_radiobutton56;  // Hexadecimal option data entry
  int dec_hex_ipopt_icmp4_tr;  // Flag to indicate decimal (0) or hexadecimal (1) data entry
  GtkWidget *tr_entry425;  // Option data
  GtkButton *tr_button86;  // Insert IP option after...
  GtkWidget *tr_entry426;  // Option number after which to insert new option
  int ipopt_icmp4_tr_after;  // Option number after which to insert new option
  GtkButton *tr_button87;  // Append IP option
  GtkWidget *tr_textview40;  // Number of IP options in packet
  GtkButton *tr_button88;  // View IP options
  GtkWidget *tr_entry427;  // Option number to remove
  int ipopt_icmp4_tr_remove;  // Option number to remove
  GtkButton *tr_button89;  // Remove option

  // ICMP header for IPv4 ICMP for traceroute
  GtkWidget *tr_entry218;  // Message type
  GtkWidget *tr_entry219;  // Message code
  GtkWidget *tr_entry220;  // ICMP checksum
  GtkWidget *tr_entry221;  // Identifier
  GtkWidget *tr_entry222;  // Sequence number
  GtkToggleButton *tr_radiobutton31;  // ASCII data entry
  GtkToggleButton *tr_radiobutton32;  // Hexadecimal data entry
  int ascii_hex_icmp4_tr;  // Flag to indicate ASCII (0) or hexadecimal (1) data entry
  GtkWidget *tr_entry223;  // ICMP payload data from keyboard
  GtkButton *tr_button32;  // ICMP payload data from file
  GtkWidget *tr_textview67;  // File Loaded indication
  GtkButton *tr_button207;  // Unload file data from packet

  GtkButton *tr_button31;  // View Packet
  GtkButton *tr_button33;  // Restore default settings

// Traceroute UDP page (IPv4) (traceroute.ui)

  GtkButton *tr_button502;  // Copy all data from main UDP (IPv4) page

  // Ethernet header for IPv4 UDP for traceroute
  GtkWidget *tr_entry224;  // Destination link-layer (MAC) address
  GtkWidget *tr_entry225;  // Source link-layer (MAC) address
  GtkWidget *tr_entry226;  // Source interface name
  GtkWidget *tr_spinbutton13;  // Interface maximum transmission unit
  GtkWidget *tr_entry227;  // Ethernet type code;

  // IPv4 header for IPv4 UDP for traceroute
  GtkWidget *tr_entry228;  // IP header length
  GtkWidget *tr_entry229;  // Protocol version
  GtkWidget *tr_entry230;  // Type of service
  GtkWidget *tr_entry231;  // Total length of datagram
  GtkWidget *tr_entry232;  // ID sequence number
  GtkWidget *tr_entry233;  // Zero
  GtkWidget *tr_entry234;  // Do not fragment flag
  GtkWidget *tr_entry235;  // More fragments following flag
  GtkWidget *tr_entry236;  // Fragmentation offset
  GtkWidget *tr_entry237;  // Time-to-live
  GtkWidget *tr_entry238;  // Transport layer protocol
  GtkWidget *tr_entry239;  // Source IP address
  GtkWidget *tr_entry240;  // Destination IP address
  GtkWidget *tr_entry241;  // IP checksum

  // IPv4 header options for IPv4 UDP for traceroute
  GtkToggleButton *tr_radiobutton57;  // Decimal option data entry
  GtkToggleButton *tr_radiobutton58;  // Hexadecimal option data entry
  int dec_hex_ipopt_udp4_tr;  // Flag to indicate decimal (0) or hexadecimal (1) data entry
  GtkWidget *tr_entry428;  // Option data
  GtkButton *tr_button90;  // Insert IP option after...
  GtkWidget *tr_entry429;  // Option number after which to insert new option
  int ipopt_udp4_tr_after;  // Option number after which to insert new option
  GtkButton *tr_button91;  // Append IP option
  GtkWidget *tr_textview42;  // Number of IP options in packet
  GtkButton *tr_button92;  // View IP options
  GtkWidget *tr_entry430;  // Option number to remove
  int ipopt_udp4_tr_remove;  // Option number to remove
  GtkButton *tr_button93;  // Remove option

  // UDP header for IPv4 UDP for traceroute
  GtkWidget *tr_entry242;  // Source port
  GtkToggleButton *tr_checkbutton35;  // Source port "Randomize"
  int ran_udp4_tr_sourceport;  // Flag for state of source port "Randomize" checkbutton (IPv4)
  GtkWidget *tr_entry243;  // Destination port
  GtkWidget *tr_entry244;  // Length of UDP datagram
  GtkWidget *tr_entry245;  // UDP checksum
  GtkToggleButton *tr_radiobutton33;  // ASCII data entry
  GtkToggleButton *tr_radiobutton34;  // Hexadecimal data entry
  int ascii_hex_udp4_tr;  // Flag to indicate ASCII (0) or hexadecimal (1) data entry
  GtkWidget *tr_entry246;  // UDP payload data from keyboard
  GtkButton *tr_button35;  // UDP payload data from file
  GtkWidget *tr_textview68;  // File Loaded indication
  GtkButton *tr_button208;  // Unload file data from packet

  GtkButton *tr_button34;  // View Packet
  GtkButton *tr_button36;  // Restore default settings

// Traceroute TCP page (IPv6) (traceroute.ui)

  GtkButton *tr_button503;  // Copy all data from main TCP (IPv6) page

  // Ethernet header for IPv6 TCP for traceroute
  GtkWidget *tr_entry247;  // Destination link-layer (MAC) address
  GtkWidget *tr_entry248;  // Source link-layer (MAC) address
  GtkWidget *tr_entry249;  // Source interface name
  GtkWidget *tr_spinbutton14;  // Interface maximum transmission unit
  GtkWidget *tr_entry250;  // Ethernet type code;

  // IPv6 header for IPv6 TCP for traceroute
  GtkWidget *tr_entry251;  // IP version
  GtkWidget *tr_entry252;  // Traffic class
  GtkWidget *tr_entry253;  // Flow label
  GtkWidget *tr_entry254;  // Payload length
  GtkWidget *tr_entry255;  // Next header
  GtkWidget *tr_entry256;  // Hop limit
  GtkWidget *tr_entry257;  // Source IP address
  GtkWidget *tr_entry258;  // Destination IP address

  // TCP header for IPv6 TCP for traceroute
  GtkWidget *tr_entry259;  // Source port number
  GtkToggleButton *tr_checkbutton37;  // Source port "Randomize"
  int ran_tcp6_tr_sourceport;  // Flag for state of source port "Randomize" checkbutton (IPv6)
  GtkWidget *tr_entry260;  // Destination port number
  GtkWidget *tr_entry261;  // Sequence number
  GtkWidget *tr_entry262;  // Ack number
  GtkWidget *tr_entry263;  // Reserved
  GtkWidget *tr_entry264;  // Data offset
  GtkWidget *tr_entry265;  // FIN flag
  GtkWidget *tr_entry266;  // SYN flag
  GtkWidget *tr_entry267;  // RST flag
  GtkWidget *tr_entry268;  // PSH flag
  GtkWidget *tr_entry269;  // ACK flag
  GtkWidget *tr_entry270;  // URG flag
  GtkWidget *tr_entry271;  // ECE flag
  GtkWidget *tr_entry272;  // CWR flag
  GtkWidget *tr_entry273;  // Window size
  GtkWidget *tr_entry274;  // TCP checksum
  GtkWidget *tr_entry275;  // Urgent pointer
  GtkToggleButton *tr_radiobutton35;  // ASCII data entry
  GtkToggleButton *tr_radiobutton36;  // Hexadecimal data entry
  int ascii_hex_tcp6_tr;  // Flag to indicate ASCII (0) or hexadecimal (1) data entry
  GtkWidget *tr_entry276;  // TCP payload data from keyboard
  GtkButton *tr_button38;  // TCP payload data from file
  GtkWidget *tr_textview69;  // File Loaded indication
  GtkButton *tr_button209;  // Unload file data from packet

  // TCP options for IPv6 TCP for traceroute
  GtkToggleButton *tr_radiobutton53;  // Decimal option data entry
  GtkToggleButton *tr_radiobutton54;  // Hexadecimal option data entry
  int dec_hex_tcpopt_tcp6_tr;  // Flag to indicate decimal (0) or hexadecimal (1) data entry
  GtkWidget *tr_entry422;  // Option data
  GtkButton *tr_button82;  // Insert IP option after...
  GtkWidget *tr_entry423;  // Option number after which to insert new option
  int tcpopt_tcp6_tr_after;  // Option number after which to insert new option
  GtkButton *tr_button83;  // Append TCP option
  GtkWidget *tr_textview38;  // Number of TCP options in packet
  GtkButton *tr_button84;  // View TCP options
  GtkWidget *tr_entry424;  // Option number to remove
  int tcpopt_tcp6_tr_remove;  // Option number to remove
  GtkButton *tr_button85;  // Remove option

  // Extension headers
  GtkButton *tr_button101;  // Hop-by-hop extension header
  GtkButton *tr_button113;  // Destination extension header
  GtkButton *tr_button135;  // Routing extension header
  GtkButton *tr_button119;  // Authentication extension header
  GtkButton *tr_button129;  // Encapsulating security payload (ESP) extension header

  GtkButton *tr_button37;  // View Packet
  GtkButton *tr_button39;  // Restore default settings

// Traceroute ICMP page (IPv6) (traceroute.ui)

  GtkButton *tr_button504;  // Copy all data from main ICMP (IPv6) page

  // Ethernet header for IPv6 ICMP for traceroute
  GtkWidget *tr_entry277;  // Destination link-layer (MAC) address
  GtkWidget *tr_entry278;  // Source link-layer (MAC) address
  GtkWidget *tr_entry279;  // Source interface name
  GtkWidget *tr_spinbutton15;  // Interface maximum transmission unit
  GtkWidget *tr_entry280;  // Ethernet type code;

  // IPv6 header for IPv6 ICMP for traceroute
  GtkWidget *tr_entry281;  // IP version
  GtkWidget *tr_entry282;  // Traffic class
  GtkWidget *tr_entry283;  // Flow label
  GtkWidget *tr_entry284;  // Payload length
  GtkWidget *tr_entry285;  // Next header
  GtkWidget *tr_entry286;  // Hop limit
  GtkWidget *tr_entry287;  // Source IP address
  GtkWidget *tr_entry288;  // Destination IP address

  // ICMP header for IPv6 ICMP for traceroute
  GtkWidget *tr_entry289;  // Message type
  GtkWidget *tr_entry290;  // Message code
  GtkWidget *tr_entry291;  // ICMP checksum
  GtkWidget *tr_entry292;  // Identifier
  GtkWidget *tr_entry293;  // Sequence number
  GtkToggleButton *tr_radiobutton37;  // ASCII data entry
  GtkToggleButton *tr_radiobutton38;  // Hexadecimal data entry
  int ascii_hex_icmp6_tr;  // Flag to indicate ASCII (0) or hexadecimal (1) data entry
  GtkWidget *tr_entry294;  // ICMP payload data from keyboard
  GtkButton *tr_button41;  // ICMP payload data from file
  GtkWidget *tr_textview70;  // File Loaded indication
  GtkButton *tr_button210;  // Unload file data from packet

  // Extension headers
  GtkButton *tr_button102;  // Hop-by-hop extension header
  GtkButton *tr_button114;  // Destination extension header
  GtkButton *tr_button136;  // Routing extension header
  GtkButton *tr_button120;  // Authentication extension header
  GtkButton *tr_button130;  // Encapsulating security payload (ESP) extension header

  GtkButton *tr_button40;  // View Packet
  GtkButton *tr_button42;  // Restore default settings

// Traceroute UDP page (IPv6) (traceroute.ui)

  GtkButton *tr_button505;  // Copy all data from main UDP (IPv6) page

  // Ethernet header for IPv6 UDP for traceroute
  GtkWidget *tr_entry295;  // Destination link-layer (MAC) address
  GtkWidget *tr_entry296;  // Source link-layer (MAC) address
  GtkWidget *tr_entry297;  // Source interface name
  GtkWidget *tr_spinbutton16;  // Interface maximum transmission unit
  GtkWidget *tr_entry298;  // Ethernet type code;

  // IPv6 header for IPv6 UDP for traceroute
  GtkWidget *tr_entry299;  // IP version
  GtkWidget *tr_entry300;  // Traffic class
  GtkWidget *tr_entry301;  // Flow label
  GtkWidget *tr_entry302;  // Payload length
  GtkWidget *tr_entry303;  // Next header
  GtkWidget *tr_entry304;  // Hop limit
  GtkWidget *tr_entry305;  // Source IP address
  GtkWidget *tr_entry306;  // Destination IP address

  // UDP header for IPv6 UDP for traceroute
  GtkWidget *tr_entry307;  // Source port
  GtkToggleButton *tr_checkbutton40;  // Source port "Randomize"
  int ran_udp6_tr_sourceport;  // Flag for state of source port "Randomize" checkbutton (IPv6)
  GtkWidget *tr_entry308;  // Destination port
  GtkWidget *tr_entry309;  // Length of UDP datagram
  GtkWidget *tr_entry310;  // UDP checksum
  GtkToggleButton *tr_radiobutton39;  // ASCII data entry
  GtkToggleButton *tr_radiobutton40;  // Hexadecimal data entry
  int ascii_hex_udp6_tr;  // Flag to indicate ASCII (0) or hexadecimal (1) data entry
  GtkWidget *tr_entry311;  // UDP payload data from keyboard
  GtkButton *tr_button44;  // UDP payload data from file
  GtkWidget *tr_textview71;  // File Loaded indication
  GtkButton *tr_button211;  // Unload file data from packet

  // Extension headers
  GtkButton *tr_button103;  // Hop-by-hop extension header
  GtkButton *tr_button115;  // Destination extension header
  GtkButton *tr_button137;  // Routing extension header
  GtkButton *tr_button121;  // Authentication extension header
  GtkButton *tr_button131;  // Encapsulating security payload (ESP) extension header

  GtkButton *tr_button43;  // View Packet
  GtkButton *tr_button45;  // Restore default settings

// IPv6 Extension Header Window Widgets

  // Hop-by-hop (Original/Inner) extension header page (hop-by-hop.ui)
  GtkWidget *hop_window;  // Hop-by-hop extension header editing window
  GtkButton *hop_button142;  // Toggle: attach / detach a hop-by-hop header
  GtkWidget *hop_textview47;  // Status: attached / detached
  GtkWidget *hop_entry438;  // Hop-by-hop header length
  GtkToggleButton *hop_radiobutton59;  // Decimal data entry
  GtkToggleButton *hop_radiobutton60;  // Hexadecimal data entry
  GtkWidget *hop_entry158;  // Hop-by-hop option data
  GtkWidget *hop_entry313;  // Option alignment parameter x (of xN + y)
  GtkWidget *hop_entry314;  // Option alignment parameter y (of xN + y)
  GtkButton *hop_button94;  // Insert Option After...
  GtkWidget *hop_entry159;  // Option # to insert after
  GtkButton *hop_button95;  // Append option
  GtkWidget *hop_textview22;  // Number of options in hop-by-hop header
  GtkButton *hop_button96;  // View hop-by-hop options
  GtkWidget *hop_entry312;  // Option number to remove
  GtkButton *hop_button97;  // Remove option

  // Hop-by-hop (Outer) extension header page (hop-by-hop-outer.ui)
  GtkWidget *hop_window_outer;  // Hop-by-hop extension header editing window
  GtkButton *hop_button142_outer;  // Toggle: attach / detach a hop-by-hop header
  GtkWidget *hop_textview47_outer;  // Status: attached / detached
  GtkWidget *hop_entry438_outer;  // Hop-by-hop header length
  GtkToggleButton *hop_radiobutton59_outer;  // Decimal data entry
  GtkToggleButton *hop_radiobutton60_outer;  // Hexadecimal data entry
  GtkWidget *hop_entry158_outer;  // Hop-by-hop option data
  GtkWidget *hop_entry313_outer;  // Option alignment parameter x (of xN + y)
  GtkWidget *hop_entry314_outer;  // Option alignment parameter y (of xN + y)
  GtkButton *hop_button94_outer;  // Insert Option After...
  GtkWidget *hop_entry159_outer;  // Option # to insert after
  GtkButton *hop_button95_outer;  // Append option
  GtkWidget *hop_textview22_outer;  // Number of options in hop-by-hop header
  GtkButton *hop_button96_outer;  // View hop-by-hop options
  GtkWidget *hop_entry312_outer;  // Option number to remove
  GtkButton *hop_button97_outer;  // Remove option

  // Destination (original/inner) extension header page (destination.ui)
  GtkWidget *dst_window;  // Destination extension header editing window
  GtkButton *dst_button143;  // Toggle: attach / detach a "first" destination header
  GtkWidget *dst_textview48;  // Status: attached / detached
  GtkButton *dst_button144;  // Toggle: attach / detach a "last" destination header
  GtkWidget *dst_textview49;  // Status: attached / detached
  GtkWidget *dst_entry439;  // Destination header length (first)
  GtkWidget *dst_entry440;  // Destination header length (last)
  GtkToggleButton *dst_radiobutton61;  // Decimal data entry
  GtkToggleButton *dst_radiobutton62;  // Hexadecimal data entry
  GtkWidget *dst_entry315;  // Destination option data
  GtkWidget *dst_entry156;  // Option alignment parameter x (of xN + y)
  GtkWidget *dst_entry157;  // Option alignment parameter y (of xN + y)
  GtkToggleButton *dst_radiobutton63;  // Designate as first destination header
  GtkToggleButton *dst_radiobutton64;  // Designate as last destination header
  GtkButton *dst_button104;  // Insert Option After...
  GtkWidget *dst_entry431;  // Option # to insert after
  GtkButton *dst_button105;  // Append option
  GtkWidget *dst_textview25;  // Number of options in first destination header
  GtkWidget *dst_textview27;  // Number of options in last destination header
  GtkButton *dst_button106;  // View destination options (first)
  GtkButton *dst_button109;  // View destination options (last)
  GtkWidget *dst_entry432;  // Option number to remove
  GtkButton *dst_button107;  // Remove option

  // Destination (outer) extension header page (destination-outer.ui)
  GtkWidget *dst_window_outer;  // Destination extension header editing window
  GtkButton *dst_button143_outer;  // Toggle: attach / detach a "first" destination header
  GtkWidget *dst_textview48_outer;  // Status: attached / detached
  GtkButton *dst_button144_outer;  // Toggle: attach / detach a "last" destination header
  GtkWidget *dst_textview49_outer;  // Status: attached / detached
  GtkWidget *dst_entry439_outer;  // Destination header length (first)
  GtkWidget *dst_entry440_outer;  // Destination header length (last)
  GtkToggleButton *dst_radiobutton61_outer;  // Decimal data entry
  GtkToggleButton *dst_radiobutton62_outer;  // Hexadecimal data entry
  GtkWidget *dst_entry315_outer;  // Destination option data
  GtkWidget *dst_entry156_outer;  // Option alignment parameter x (of xN + y)
  GtkWidget *dst_entry157_outer;  // Option alignment parameter y (of xN + y)
  GtkToggleButton *dst_radiobutton63_outer;  // Designate as first destination header
  GtkToggleButton *dst_radiobutton64_outer;  // Designate as last destination header
  GtkButton *dst_button104_outer;  // Insert Option After...
  GtkWidget *dst_entry431_outer;  // Option # to insert after
  GtkButton *dst_button105_outer;  // Append option
  GtkWidget *dst_textview25_outer;  // Number of options in first destination header
  GtkWidget *dst_textview27_outer;  // Number of options in last destination header
  GtkButton *dst_button106_outer;  // View destination options (first)
  GtkButton *dst_button109_outer;  // View destination options (last)
  GtkWidget *dst_entry432_outer;  // Option number to remove
  GtkButton *dst_button107_outer;  // Remove option

  // Routing (original/inner) extension header page (routing.ui)
  GtkWidget *route_window;  // Routing extension header editing window
  GtkButton *rt_button145;  // Toggle: attach / detach a routing header
  GtkWidget *rt_textview50;  // Status: attached / detached
  GtkWidget *rt_entry446;  // Header length
  GtkWidget *rt_entry447;  // Routing type
  GtkWidget *rt_entry448;  // Segments left
  GtkToggleButton *rt_radiobutton73;  // Decimal routing data entry
  GtkToggleButton *rt_radiobutton74;  // Hexadecimal routing data entry
  GtkWidget *rt_entry449;  // Routing data
  GtkButton *rt_button125;  // Load routing data from file...

  // Routing (outer) extension header page (routing-outer.ui)
  GtkWidget *route_window_outer;  // Routing extension header editing window
  GtkButton *rt_button145_outer;  // Toggle: attach / detach a routing header
  GtkWidget *rt_textview50_outer;  // Status: attached / detached
  GtkWidget *rt_entry446_outer;  // Header length
  GtkWidget *rt_entry447_outer;  // Routing type
  GtkWidget *rt_entry448_outer;  // Segments left
  GtkToggleButton *rt_radiobutton73_outer;  // Decimal routing data entry
  GtkToggleButton *rt_radiobutton74_outer;  // Hexadecimal routing data entry
  GtkWidget *rt_entry449_outer;  // Routing data
  GtkButton *rt_button125_outer;  // Load routing data from file...

  // Authentication extension header page (authentication.ui)
  GtkWidget *auth_window;  // Authentication extension header editing window
  GtkButton *auth_button138;  // Toggle: attach / detach an authentication header
  GtkWidget *auth_textview43;  // Status: attached / detached
  GtkButton *auth_button139;  // Toggle: transport / tunnel mode
  GtkWidget *auth_textview44;  // Status: transport / tunnel mode
  GtkWidget *auth_entry433;  // Payload length
  GtkWidget *auth_entry434;  // Reserved
  GtkWidget *auth_entry435;  // Security parameters index (SPI)
  GtkWidget *auth_entry436;  // Sequence number
  GtkToggleButton *auth_radiobutton67;  // Decimal integrity check value (ICV) entry
  GtkToggleButton *auth_radiobutton68;  // Hexadecimal integrity check value (ICV) entry
  GtkWidget *auth_entry437;  // Integrity check value (ICV)
  GtkButton *auth_button108;  // Load ICV from file ...
  GtkButton *auth_button122;  // Save data for ICV calculation

  // Encapsulating security payload (ESP) extension header page (esp.ui)
  GtkWidget *esp_window;  // ESP extension header editing window
  GtkButton *esp_button140;  // Toggle: attach / detach an ESP header
  GtkWidget *esp_textview45;  // Status: attached / detached
  GtkButton *esp_button141;  // Toggle: transport / tunnel mode
  GtkWidget *esp_textview46;  // Status: transport / tunnel mode
  GtkWidget *esp_entry441;  // Security parameters index (SPI)
  GtkWidget *esp_entry442;  // Sequence number
  GtkButton *esp_button123;  // Save payload data for ICV calculation
  GtkWidget *esp_entry443;  // Padding length
  GtkWidget *esp_entry444;  // Next header
  GtkToggleButton *esp_radiobutton71;  // Decimal integrity check value (ICV) entry
  GtkToggleButton *esp_radiobutton72;  // Hexadecimal integrity check value (ICV) entry
  GtkWidget *esp_entry445;  // Integrity check value (ICV)
  GtkButton *esp_button124;  // Load ICV from file ...

};

// ********** CALLBACKS FUNCTION PROTOTYPES **********
//                   callbacks.c

// Menubar
int on_about_requested (void);
int on_documentation_activate (SPSData *);

// Send page (sps.ui)

// Select Packet Type
int on_radiobutton1_toggled (GtkButton *, SPSData *);  // TCP (IPv4) packet type
int on_radiobutton2_toggled (GtkButton *, SPSData *);  // ICMP (IPv4) packet type
int on_radiobutton3_toggled (GtkButton *, SPSData *);  // UDP (IPv4) packet type
int on_radiobutton4_toggled (GtkButton *, SPSData *);  // All (IPv4) packet types (cycle)
int on_radiobutton5_toggled (GtkButton *, SPSData *);  // TCP (IPv6) packet type
int on_radiobutton6_toggled (GtkButton *, SPSData *);  // ICMP (IPv6) packet type
int on_radiobutton7_toggled (GtkButton *, SPSData *);  // UDP (IPv6) packet type
int on_radiobutton8_toggled (GtkButton *, SPSData *);  // All (IPv6) packet types (cycle)

// Number of Packets to Send
int on_entry18_changed (GtkWidget *, SPSData *);  // Number of packets to send
int on_checkbutton11_toggled (GtkButton *, SPSData *);  // Flood mode

// IPv6 over IPv4 (6to4)
int on_checkbutton12_toggled (GtkButton *, SPSData *);  // Tunnel IPv6 over IPv4

// Send / Stop
int on_button1_clicked (GtkButton *, SPSData *);  // Send packet(s)
int on_button5_clicked (GtkButton *, SPSData *);  // Stop flood

// TCP page (IPv4) (sps.ui)

// Ethernet header
int on_entry144_changed (GtkWidget *, SPSData *);  // Destination link-layer (MAC) address
int on_entry145_changed (GtkWidget *, SPSData *);  // Source link-layer (MAC) address
int on_entry150_changed (GtkWidget *, SPSData *);  // Source interface name
int on_spinbutton5_value_changed (GtkWidget *, SPSData *);  // Maximum transmission unit (MTU) - Changed via spinbutton + or -.
int on_spinbutton5_change_value (GtkWidget *, SPSData *);  // Maximum transmission unit (MTU) - Changed via spinbutton text entry.
int on_entry146_changed (GtkWidget *, SPSData *);  // Ethernet type code
// IPv4 header
int on_entry29_changed (GtkWidget *, SPSData *);  // IPv4 header length
int on_entry30_changed (GtkWidget *, SPSData *);  // Protocol version
int on_entry31_changed (GtkWidget *, SPSData *);  // Type-of-service
int on_entry32_changed (GtkWidget *, SPSData *);  // Total length of datagram
int on_entry33_changed (GtkWidget *, SPSData *);  // ID sequence number
int on_entry122_changed (GtkWidget *, SPSData *);  // Zero
int on_entry37_changed (GtkWidget *, SPSData *);  // Do not fragment flag
int on_entry38_changed (GtkWidget *, SPSData *);  // More fragments following flag
int on_entry39_changed (GtkWidget *, SPSData *);  // Fragmentation offset
int on_entry40_changed (GtkWidget *, SPSData *);  // Time-to-Live
int on_entry74_changed (GtkWidget *, SPSData *);  // Transport layer protocol
int on_entry35_changed (GtkWidget *, SPSData *);  // Source IPv4 address
int on_checkbutton5_toggled (GtkButton *, SPSData *);  // Randomize source IPv4 address
int on_entry36_changed (GtkWidget *, SPSData *);  // Destination IPv4 address
int on_entry34_changed (GtkWidget *, SPSData *);  // IPv4 Checksum
// IPv4 header options
int on_radiobutton24_toggled (GtkButton *, SPSData *);  // Decimal option data entry
int on_radiobutton25_toggled (GtkButton *);  // Hexadecimal option data entry
int on_entry400_changed (GtkWidget *, SPSData *);  // Options data
int on_button26_clicked (GtkButton *, SPSData *);  // Insert IP option after...
int on_entry401_changed (GtkWidget *, SPSData *);  // Option number after which to insert new option
int on_button60_clicked (GtkButton *, SPSData *);  // Append IP option
int on_button61_clicked (GtkButton *, SPSData *);  // View IP options
int on_entry403_changed (GtkWidget *, SPSData *);  // Option number to remove
int on_button62_clicked (GtkButton *, SPSData *);  // Remove option

// TCP header (IPv4)
int on_entry41_changed (GtkWidget *, SPSData *);  // Source Port
int on_checkbutton7_toggled (GtkButton *, SPSData *);  // Randomize source port
int on_entry42_changed (GtkWidget *, SPSData *);  // Destination port
int on_entry43_changed (GtkWidget *, SPSData *);  // Sequence number
int on_entry44_changed (GtkWidget *, SPSData *);  // Acknowledgement number
int on_entry45_changed (GtkWidget *, SPSData *);  // Reserved 4 bits
int on_entry46_changed (GtkWidget *, SPSData *);  // Data offset
int on_entry47_changed (GtkWidget *, SPSData *);  // FIN flag
int on_entry48_changed (GtkWidget *, SPSData *);  // SYN flag
int on_entry49_changed (GtkWidget *, SPSData *);  // RST flag
int on_entry50_changed (GtkWidget *, SPSData *);  // PSH flag
int on_entry51_changed (GtkWidget *, SPSData *);  // ACK flag
int on_entry52_changed (GtkWidget *, SPSData *);  // URG flag
int on_entry53_changed (GtkWidget *, SPSData *);  // ECE flag
int on_entry54_changed (GtkWidget *, SPSData *);  // CWR flag
int on_entry55_changed (GtkWidget *, SPSData *);  // Window size
int on_entry56_changed (GtkWidget *, SPSData *);  // TCP checksum
int on_entry77_changed (GtkWidget *, SPSData *);  // Urgent pointer
int on_radiobutton17_toggled (GtkButton *, SPSData *);  // ASCII data entry
int on_radiobutton18_toggled (GtkButton *);  // Hexadecimal data entry
int on_entry162_changed (GtkWidget *, SPSData *);  // TCP data from keyboard
int on_button20_clicked (GtkButton *, SPSData *);  // TCP data from file
int on_button200_clicked (GtkButton *, SPSData *);  // Unload file data from packet
// TCP options (IPv4)
int on_radiobutton41_toggled (GtkButton *, SPSData *);  // Decimal option data entry
int on_radiobutton42_toggled (GtkButton *);  // Hexadecimal option data entry
int on_entry404_changed (GtkWidget *, SPSData *);  // Option data
int on_button47_clicked (GtkButton *, SPSData *);  // Insert IP option after...
int on_entry405_changed (GtkWidget *, SPSData *);  // Option number after which to insert new option
int on_button48_clicked (GtkButton *, SPSData *);  // Append TCP option
int on_button49_clicked (GtkButton *, SPSData *);  // View TCP options
int on_entry406_changed (GtkWidget *, SPSData *);  // Option number to remove
int on_button58_clicked (GtkButton *, SPSData *);  // Remove option

int on_button3_clicked (GtkButton *, SPSData *);  // View IPv4 TCP packet
int on_button11_clicked (GtkButton *, SPSData *);  // Restore IPv4 TCP default values

// ICMP page (IPv4) (sps.ui)

// Ethernet header
int on_entry141_changed (GtkWidget *, SPSData *);  // Destination link-layer (MAC) address
int on_entry142_changed (GtkWidget *, SPSData *);  // Source link-layer (MAC) address
int on_entry151_changed (GtkWidget *, SPSData *);  // Source interface name
int on_spinbutton6_value_changed (GtkWidget *, SPSData *);  // Maximum transmission unit (MTU) - Changed via spinbutton + or -.
int on_spinbutton6_change_value (GtkWidget *, SPSData *);  // Maximum transmission unit (MTU) - Changed via spinbutton text entry.
int on_entry143_changed (GtkWidget *, SPSData *);  // Ethernet type code
// IPv4 header
int on_entry57_changed (GtkWidget *, SPSData *);  // IPv4 Header Length
int on_entry58_changed (GtkWidget *, SPSData *);  // Protocol Version
int on_entry59_changed (GtkWidget *, SPSData *);  // Type of Service
int on_entry60_changed (GtkWidget *, SPSData *);  // Total Length of Datagram
int on_entry61_changed (GtkWidget *, SPSData *);  // ID Sequence Number
int on_entry123_changed (GtkWidget *, SPSData *);  // Zero
int on_entry65_changed (GtkWidget *, SPSData *);  // Do Not Fragment Flag
int on_entry66_changed (GtkWidget *, SPSData *);  // More Fragments Following
int on_entry67_changed (GtkWidget *, SPSData *);  // Fragmentation Offset
int on_entry68_changed (GtkWidget *, SPSData *);  // Time-to-Live
int on_entry75_changed (GtkWidget *, SPSData *);  // Transport Layer Protocol
int on_entry63_changed (GtkWidget *, SPSData *);  // Source IPv4 Address
int on_checkbutton9_toggled (GtkButton *, SPSData *);  // Randomize Source IPv4 Address
int on_entry64_changed (GtkWidget *, SPSData *);  // Destination IPv4 Address
int on_entry62_changed (GtkWidget *, SPSData *);  // IPv4 Checksum
// IPv4 header options
int on_radiobutton43_toggled (GtkButton *, SPSData *);  // Decimal option data entry
int on_radiobutton44_toggled (GtkButton *);  // Hexadecimal option data entry
int on_entry407_changed (GtkWidget *, SPSData *);  // Options data
int on_button59_clicked (GtkButton *, SPSData *);  // Insert IP option after...
int on_entry408_changed (GtkWidget *, SPSData *);  // Option number after which to insert new option
int on_button63_clicked (GtkButton *, SPSData *);  // Append IP option
int on_button64_clicked (GtkButton *, SPSData *);  // View IP options
int on_entry409_changed (GtkWidget *, SPSData *);  // Option number to remove
int on_button65_clicked (GtkButton *, SPSData *);  // Remove option
// ICMP message (IPv4)
int on_entry69_changed (GtkWidget *, SPSData *);  // Message type
int on_entry70_changed (GtkWidget *, SPSData *);  // Message code
int on_entry71_changed (GtkWidget *, SPSData *);  // ICMP checksum
int on_entry72_changed (GtkWidget *, SPSData *);  // ID
int on_entry73_changed (GtkWidget *, SPSData *);  // Sequence number
int on_radiobutton9_toggled (GtkButton *, SPSData *);  // ASCII data entry
int on_radiobutton10_toggled (GtkButton *);  // Hexadecimal data entry
int on_entry126_changed (GtkWidget *, SPSData *);  // ICMP data from keyboard
int on_button7_clicked (GtkButton *, SPSData *);  // ICMP data from file
int on_button202_clicked (GtkButton *, SPSData *);  // Unload file data from packet

int on_button2_clicked (GtkButton *, SPSData *);  // View IPv4 ICMP packet
int on_button12_clicked (GtkButton *, SPSData *);  // Restore IPv4 ICMP default values

// UDP page (IPv4) (sps.ui)

// Ethernet header
int on_entry147_changed (GtkWidget *, SPSData *);  // Destination link-layer (MAC) address
int on_entry148_changed (GtkWidget *, SPSData *);  // Source link-layer (MAC) address
int on_entry152_changed (GtkWidget *, SPSData *);  // Source interface name
int on_spinbutton7_value_changed (GtkWidget *, SPSData *);  // Maximum transmission unit (MTU) - Changed via spinbutton + or -.
int on_spinbutton7_change_value (GtkWidget *, SPSData *);  // Maximum transmission unit (MTU) - Changed via spinbutton text entry.
int on_entry149_changed (GtkWidget *, SPSData *);  // Ethernet type code
// IPv4 header
int on_entry1_changed (GtkWidget *, SPSData *);  // IPv4 Header Length
int on_entry2_changed (GtkWidget *, SPSData *);  // Protocol Version
int on_entry3_changed (GtkWidget *, SPSData *);  // Type of Service
int on_entry4_changed (GtkWidget *, SPSData *);  // Total Length of Datagram
int on_entry5_changed (GtkWidget *, SPSData *);  // ID Sequence Number
int on_entry124_changed (GtkWidget *, SPSData *);  // Zero
int on_entry9_changed (GtkWidget *, SPSData *);  // Do Not Fragment Flag
int on_entry10_changed (GtkWidget *, SPSData *);  // More Fragments Following
int on_entry11_changed (GtkWidget *, SPSData *);  // Fragmentation Offset
int on_entry12_changed (GtkWidget *, SPSData *);  // Time-to-Live
int on_entry76_changed (GtkWidget *, SPSData *);  // Transport Layer Protocol
int on_entry7_changed (GtkWidget *, SPSData *);  // Source IPv4 Address
int on_checkbutton1_toggled (GtkButton *, SPSData *);  // Randomize Source IPv4 Address
int on_entry8_changed (GtkWidget *, SPSData *);  // Destination IPv4 Address
int on_entry6_changed (GtkWidget *, SPSData *);  // IPv4 Checksum
// IPv4 header options
int on_radiobutton45_toggled (GtkButton *, SPSData *);  // Decimal option data entry
int on_radiobutton46_toggled (GtkButton *);  // Hexadecimal option data entry
int on_entry410_changed (GtkWidget *, SPSData *);  // Options data
int on_button66_clicked (GtkButton *, SPSData *);  // Insert IP option after...
int on_entry411_changed (GtkWidget *, SPSData *);  // Option number after which to insert new option
int on_button67_clicked (GtkButton *, SPSData *);  // Append IP option
int on_button68_clicked (GtkButton *, SPSData *);  // View IP options
int on_entry412_changed (GtkWidget *, SPSData *);  // Option number to remove
int on_button69_clicked (GtkButton *, SPSData *);  // Remove option
// UDP header (IPv4)
int on_entry13_changed (GtkWidget *, SPSData *);  // Source Port
int on_checkbutton3_toggled (GtkButton *, SPSData *);  // Randomize Source Port
int on_entry14_changed (GtkWidget *, SPSData *);  // Destination Port
int on_entry15_changed (GtkWidget *, SPSData *);  // Length of UDP Datagram
int on_entry16_changed (GtkWidget *, SPSData *);  // UDP Checksum
int on_radiobutton13_toggled (GtkButton *, SPSData *);  // ASCII data entry
int on_radiobutton14_toggled (GtkButton *);  // Hexadecimal data entry
int on_entry17_changed (GtkWidget *, SPSData *);  // UDP Data from keyboard
int on_button8_clicked (GtkButton *, SPSData *);  // UDP data from file
int on_button204_clicked (GtkButton *, SPSData *);  // Unload file data from packet

int on_button4_clicked (GtkButton *, SPSData *);  // View IPv4 UDP packet
int on_button15_clicked (GtkButton *, SPSData *);  // Restore IPv4 UDP default values

// TCP page (IPv6) (sps.ui)

// Ethernet header
int on_entry131_changed (GtkWidget *, SPSData *);  // Destination link-layer (MAC) address
int on_entry132_changed (GtkWidget *, SPSData *);  // Source link-layer (MAC) address
int on_entry154_changed (GtkWidget *, SPSData *);  // Source interface name
int on_spinbutton8_value_changed (GtkWidget *, SPSData *);  // Maximum transmission unit (MTU) - Changed via spinbutton + or -.
int on_spinbutton8_change_value (GtkWidget *, SPSData *);  // Maximum transmission unit (MTU) - Changed via spinbutton text entry.
int on_entry133_changed (GtkWidget *, SPSData *);  // Ethernet type code
// IPv6 header
int on_entry22_changed (GtkWidget *, SPSData *);  // IP version
int on_entry23_changed (GtkWidget *, SPSData *);  // Traffic class
int on_entry24_changed (GtkWidget *, SPSData *);  // Flow label
int on_entry25_changed (GtkWidget *, SPSData *);  // Payload length
int on_entry26_changed (GtkWidget *, SPSData *);  // Next header
int on_entry27_changed (GtkWidget *, SPSData *);  // Hop limit
int on_entry28_changed (GtkWidget *, SPSData *);  // Source IPv6 address
int on_checkbutton2_toggled (GtkButton *, SPSData *);  // Randomize Source IPv6 Address
int on_entry78_changed (GtkWidget *, SPSData *);  // Destination IPv6 address
// TCP header (IPv6)
int on_entry79_changed (GtkWidget *, SPSData *);  // Source Port
int on_checkbutton4_toggled (GtkButton *, SPSData *);  // Randomize source port
int on_entry80_changed (GtkWidget *, SPSData *);  // Destination port
int on_entry81_changed (GtkWidget *, SPSData *);  // Sequence number
int on_entry82_changed (GtkWidget *, SPSData *);  // Acknowledgement number
int on_entry83_changed (GtkWidget *, SPSData *);  // Reserved 4 bits
int on_entry84_changed (GtkWidget *, SPSData *);  // Data offset
int on_entry85_changed (GtkWidget *, SPSData *);  // FIN flag
int on_entry86_changed (GtkWidget *, SPSData *);  // SYN flag
int on_entry87_changed (GtkWidget *, SPSData *);  // RST flag
int on_entry88_changed (GtkWidget *, SPSData *);  // PSH flag
int on_entry89_changed (GtkWidget *, SPSData *);  // ACK flag
int on_entry90_changed (GtkWidget *, SPSData *);  // URG flag
int on_entry91_changed (GtkWidget *, SPSData *);  // ECE flag
int on_entry92_changed (GtkWidget *, SPSData *);  // CWR flag
int on_entry93_changed (GtkWidget *, SPSData *);  // Window size
int on_entry94_changed (GtkWidget *, SPSData *);  // TCP checksum
int on_entry95_changed (GtkWidget *, SPSData *);  // Urgent pointer
int on_radiobutton19_toggled (GtkButton *, SPSData *);  // ASCII data entry
int on_radiobutton20_toggled (GtkButton *);  // Hexadecimal data entry
int on_entry163_changed (GtkWidget *, SPSData *);  // TCP data from keyboard
int on_button22_clicked (GtkButton *, SPSData *);  // TCP data from file
int on_button201_clicked (GtkButton *, SPSData *);  // Unload file data from packet
// TCP options (IPv6)
int on_radiobutton47_toggled (GtkButton *, SPSData *);  // Decimal option data entry
int on_radiobutton48_toggled (GtkButton *);  // Hexadecimal option data entry
int on_entry413_changed (GtkWidget *, SPSData *);  // Option data
int on_button70_clicked (GtkButton *, SPSData *);  // Insert IP option after...
int on_entry414_changed (GtkWidget *, SPSData *);  // Option number after which to insert new option
int on_button71_clicked (GtkButton *, SPSData *);  // Append TCP option
int on_button72_clicked (GtkButton *, SPSData *);  // View TCP options
int on_entry415_changed (GtkWidget *, SPSData *);  // Option number to remove
int on_button73_clicked (GtkButton *, SPSData *);  // Remove option
// Extension headers
int on_button98_clicked (GtkWidget *, SPSData *);  // Hop-by-hop extension header
int on_button110_clicked (GtkWidget *, SPSData *);  // Destination extension header
int on_button132_clicked (GtkWidget *, SPSData *);  // Routing extension header
int on_button116_clicked (GtkWidget *, SPSData *);  // Authentication extension header
int on_button126_clicked (GtkWidget *, SPSData *);  // Encapsulating security payload (ESP) extension header
// Outer extension headers for AH or ESP tunneling
int on_button300_clicked (GtkWidget *, SPSData *);  // Hop-by-hop extension header
int on_button301_clicked (GtkWidget *, SPSData *);  // Destination extension header
int on_button302_clicked (GtkWidget *, SPSData *);  // Routing extension header

int on_button6_clicked (GtkButton *, SPSData *);  // View IPv6 TCP packet
int on_button16_clicked (GtkButton *, SPSData *);  // Restore IPv6 TCP default values

// ICMP page (IPv6) (sps.ui)

// Ethernet header
int on_entry128_changed (GtkWidget *, SPSData *);  // Destination link-layer (MAC) address
int on_entry129_changed (GtkWidget *, SPSData *);  // Source link-layer (MAC) address
int on_entry153_changed (GtkWidget *, SPSData *);  // Source interface name
int on_spinbutton9_value_changed (GtkWidget *, SPSData *);  // Maximum transmission unit (MTU) - Changed via spinbutton + or -.
int on_spinbutton9_change_value (GtkWidget *, SPSData *);  // Maximum transmission unit (MTU) - Changed via spinbutton text entry.
int on_entry130_changed (GtkWidget *, SPSData *);  // Ethernet type code
// IPv6 header
int on_entry110_changed (GtkWidget *, SPSData *);  // IP version
int on_entry111_changed (GtkWidget *, SPSData *);  // Traffic class
int on_entry112_changed (GtkWidget *, SPSData *);  // Flow label
int on_entry113_changed (GtkWidget *, SPSData *);  // Payload length
int on_entry114_changed (GtkWidget *, SPSData *);  // Next header
int on_entry115_changed (GtkWidget *, SPSData *);  // Hop limit
int on_entry109_changed (GtkWidget *, SPSData *);  // Source IPv6 address
int on_checkbutton10_toggled (GtkButton *, SPSData *);  // Randomize Source IPv6 Address
int on_entry116_changed (GtkWidget *, SPSData *);  // Destination IPv6 address
// ICMP message (IPv6)
int on_entry117_changed (GtkWidget *, SPSData *);  // Message type
int on_entry118_changed (GtkWidget *, SPSData *);  // Message code
int on_entry119_changed (GtkWidget *, SPSData *);  // ICMP checksum
int on_entry120_changed (GtkWidget *, SPSData *);  // ID
int on_entry121_changed (GtkWidget *, SPSData *);  // Sequence number
int on_radiobutton11_toggled (GtkButton *, SPSData *);  // ASCII data entry
int on_radiobutton12_toggled (GtkButton *);  // Hexadecimal data entry
int on_entry127_changed (GtkWidget *, SPSData *);  // ICMP data from keyboard
int on_button9_clicked (GtkButton *, SPSData *);  // ICMP data from file
int on_button203_clicked (GtkButton *, SPSData *);  // Unload file data from packet
// Extension headers
int on_button99_clicked (GtkWidget *, SPSData *);  // Hop-by-hop extension header
int on_button111_clicked (GtkWidget *, SPSData *);  // Destination extension header
int on_button133_clicked (GtkWidget *, SPSData *);  // Routing extension header
int on_button117_clicked (GtkWidget *, SPSData *);  // Authentication extension header
int on_button127_clicked (GtkWidget *, SPSData *);  // Encapsulating security payload (ESP) extension header
// Outer extension headers for AH or ESP tunneling
int on_button303_clicked (GtkWidget *, SPSData *);  // Hop-by-hop extension header
int on_button304_clicked (GtkWidget *, SPSData *);  // Destination extension header
int on_button305_clicked (GtkWidget *, SPSData *);  // Routing extension header

int on_button17_clicked (GtkButton *, SPSData *);  // View IPv6 ICMP packet
int on_button18_clicked (GtkButton *, SPSData *);  // Restore IPv6 ICMP default values

// UDP page (IPv6) (sps.ui)

// Ethernet header
int on_entry134_changed (GtkWidget *, SPSData *);  // Destination link-layer (MAC) address
int on_entry135_changed (GtkWidget *, SPSData *);  // Source link-layer (MAC) address
int on_entry155_changed (GtkWidget *, SPSData *);  // Source interface name
int on_spinbutton10_value_changed (GtkWidget *, SPSData *);  // Maximum transmission unit (MTU) - Changed via spinbutton + or -.
int on_spinbutton10_change_value (GtkWidget *, SPSData *);  // Maximum transmission unit (MTU) - Changed via spinbutton text entry.
int on_entry136_changed (GtkWidget *, SPSData *);  // Ethernet type code
// IPv6 header
int on_entry97_changed (GtkWidget *, SPSData *);  // IP version
int on_entry98_changed (GtkWidget *, SPSData *);  // Traffic class
int on_entry99_changed (GtkWidget *, SPSData *);  // Flow label
int on_entry100_changed (GtkWidget *, SPSData *);  // Payload length
int on_entry101_changed (GtkWidget *, SPSData *);  // Next header
int on_entry102_changed (GtkWidget *, SPSData *);  // Hop limit
int on_entry96_changed (GtkWidget *, SPSData *);  // Source IPv6 address
int on_checkbutton6_toggled (GtkButton *, SPSData *);  // Randomize Source IPv6 Address
int on_entry103_changed (GtkWidget *, SPSData *);  // Destination IPv6 address
// UDP header (IPv6)
int on_entry104_changed (GtkWidget *, SPSData *);  // Source Port
int on_checkbutton8_toggled (GtkButton *, SPSData *);  // Randomize Source Port
int on_entry105_changed (GtkWidget *, SPSData *);  // Destination Port
int on_entry106_changed (GtkWidget *, SPSData *);  // Length of UDP Datagram
int on_entry107_changed (GtkWidget *, SPSData *);  // UDP Checksum
int on_radiobutton15_toggled (GtkButton *, SPSData *);  // ASCII data entry
int on_radiobutton16_toggled (GtkButton *);  // Hexadecimal data entry
int on_entry108_changed (GtkWidget *, SPSData *);  // UDP Data from keyboard
int on_button10_clicked (GtkButton *, SPSData *);  // UDP data from file
int on_button205_clicked (GtkButton *, SPSData *);  // Unload file data from packet
// Extension headers
int on_button100_clicked (GtkWidget *, SPSData *);  // Hop-by-hop extension header
int on_button112_clicked (GtkWidget *, SPSData *);  // Destination extension header
int on_button134_clicked (GtkWidget *, SPSData *);  // Routing extension header
int on_button118_clicked (GtkWidget *, SPSData *);  // Authentication extension header
int on_button128_clicked (GtkWidget *, SPSData *);  // Encapsulating security payload (ESP) extension header
// Outer extension headers for AH or ESP tunneling
int on_button306_clicked (GtkWidget *, SPSData *);  // Hop-by-hop extension header
int on_button307_clicked (GtkWidget *, SPSData *);  // Destination extension header
int on_button308_clicked (GtkWidget *, SPSData *);  // Routing extension header

int on_button13_clicked (GtkButton *, SPSData *);  // View IPv6 UDP packet
int on_button19_clicked (GtkButton *, SPSData *);  // Restore IPv6 UDP default values

// 6to4 page (sps.ui)
// 6to4 TCP (IPv4) (sps.ui)

// Ethernet header
int on_entry600_changed (GtkWidget *, SPSData *);  // Destination link-layer (MAC) address
int on_entry601_changed (GtkWidget *, SPSData *);  // Source link-layer (MAC) address
int on_entry460_changed (GtkWidget *, SPSData *);  // TCP interface name
int on_spinbutton17_value_changed (GtkWidget *, SPSData *);  // TCP Maximum transmission unit (MTU) - Changed via spinbutton + or -.
int on_spinbutton17_change_value (GtkWidget *, SPSData *);  // TCP Maximum transmission unit (MTU) - Changed via spinbutton text entry.
int on_entry602_changed (GtkWidget *, SPSData *);  // Ethernet type code

// IPv4 header
int on_entry609_changed (GtkWidget *, SPSData *);  // IPv4 header length
int on_entry610_changed (GtkWidget *, SPSData *);  // Protocol version
int on_entry611_changed (GtkWidget *, SPSData *);  // Type-of-service
int on_entry612_changed (GtkWidget *, SPSData *);  // Total length of datagram
int on_entry613_changed (GtkWidget *, SPSData *);  // ID sequence number
int on_entry614_changed (GtkWidget *, SPSData *);  // Zero
int on_entry615_changed (GtkWidget *, SPSData *);  // Do not fragment flag
int on_entry616_changed (GtkWidget *, SPSData *);  // More fragments following flag
int on_entry617_changed (GtkWidget *, SPSData *);  // Fragmentation offset
int on_entry618_changed (GtkWidget *, SPSData *);  // Time-to-Live
int on_entry619_changed (GtkWidget *, SPSData *);  // Transport layer protocol
int on_entry125_changed (GtkWidget *, SPSData *);  // Source IPv4 address
int on_checkbutton13_toggled (GtkButton *, SPSData *);  // Randomize source IPv4 address
int on_entry620_changed (GtkWidget *, SPSData *);  // Destination IPv4 address
int on_entry621_changed (GtkWidget *, SPSData *);  // IPv4 Checksum

int on_button600_clicked (GtkButton *, SPSData *);  // Restore default values

// 6to4 ICMP (IPv4) (sps.ui)

// Ethernet header
int on_entry603_changed (GtkWidget *, SPSData *);  // Destination link-layer (MAC) address
int on_entry604_changed (GtkWidget *, SPSData *);  // Source link-layer (MAC) address
int on_entry461_changed (GtkWidget *, SPSData *);  // ICMP interface name
int on_spinbutton18_value_changed (GtkWidget *, SPSData *);  // ICMP Maximum transmission unit (MTU) - Changed via spinbutton + or -.
int on_spinbutton18_change_value (GtkWidget *, SPSData *);  // ICMP Maximum transmission unit (MTU) - Changed via spinbutton text entry.
int on_entry605_changed (GtkWidget *, SPSData *);  // Ethernet type code

// IPv4 header
int on_entry622_changed (GtkWidget *, SPSData *);  // IPv4 Header Length
int on_entry623_changed (GtkWidget *, SPSData *);  // Protocol Version
int on_entry624_changed (GtkWidget *, SPSData *);  // Type of Service
int on_entry625_changed (GtkWidget *, SPSData *);  // Total Length of Datagram
int on_entry626_changed (GtkWidget *, SPSData *);  // ID Sequence Number
int on_entry627_changed (GtkWidget *, SPSData *);  // Zero
int on_entry628_changed (GtkWidget *, SPSData *);  // Do Not Fragment Flag
int on_entry629_changed (GtkWidget *, SPSData *);  // More Fragments Following
int on_entry630_changed (GtkWidget *, SPSData *);  // Fragmentation Offset
int on_entry631_changed (GtkWidget *, SPSData *);  // Time-to-Live
int on_entry632_changed (GtkWidget *, SPSData *);  // Transport Layer Protocol
int on_entry160_changed (GtkWidget *, SPSData *);  // Source IPv4 Address
int on_checkbutton26_toggled (GtkButton *, SPSData *);  // Randomize Source IPv4 Address
int on_entry633_changed (GtkWidget *, SPSData *);  // Destination IPv4 Address
int on_entry634_changed (GtkWidget *, SPSData *);  // IPv4 Checksum

int on_button601_clicked (GtkButton *, SPSData *);  // Restore default values

// 6to4 UDP (IPv4) (sps.ui)

// Ethernet header
int on_entry606_changed (GtkWidget *, SPSData *);  // Destination link-layer (MAC) address
int on_entry607_changed (GtkWidget *, SPSData *);  // Source link-layer (MAC) address
int on_entry462_changed (GtkWidget *, SPSData *);  // UDP interface name
int on_spinbutton19_value_changed (GtkWidget *, SPSData *);  // UDP Maximum transmission unit (MTU) - Changed via spinbutton + or -.
int on_spinbutton19_change_value (GtkWidget *, SPSData *);  // UDP Maximum transmission unit (MTU) - Changed via spinbutton text entry.
int on_entry608_changed (GtkWidget *, SPSData *);  // Ethernet type code

// IPv4 header
int on_entry635_changed (GtkWidget *, SPSData *);  // IPv4 Header Length
int on_entry636_changed (GtkWidget *, SPSData *);  // Protocol Version
int on_entry637_changed (GtkWidget *, SPSData *);  // Type of Service
int on_entry638_changed (GtkWidget *, SPSData *);  // Total Length of Datagram
int on_entry639_changed (GtkWidget *, SPSData *);  // ID Sequence Number
int on_entry640_changed (GtkWidget *, SPSData *);  // Zero
int on_entry641_changed (GtkWidget *, SPSData *);  // Do Not Fragment Flag
int on_entry642_changed (GtkWidget *, SPSData *);  // More Fragments Following
int on_entry643_changed (GtkWidget *, SPSData *);  // Fragmentation Offset
int on_entry644_changed (GtkWidget *, SPSData *);  // Time-to-Live
int on_entry645_changed (GtkWidget *, SPSData *);  // Transport Layer Protocol
int on_entry161_changed (GtkWidget *, SPSData *);  // Source IPv4 Address
int on_checkbutton27_toggled (GtkButton *, SPSData *);  // Randomize Source IPv4 Address
int on_entry646_changed (GtkWidget *, SPSData *);  // Destination IPv4 Address
int on_entry647_changed (GtkWidget *, SPSData *);  // IPv4 Checksum

int on_button602_clicked (GtkButton *, SPSData *);  // Restore default values

// Host/IP page (sps.ui)
int on_entry19_changed (GtkWidget *, SPSData *);  // Hostname to IP Address
int on_entry20_changed (GtkWidget *, SPSData *);  // IPv4 Address to Hostname
int on_entry21_changed (GtkWidget *, SPSData *);  // IPv6 Address to Hostname
int on_entry139_changed (GtkWidget *, SPSData *);  // This host's interface name to be converted to link-layer (MAC) address.
int on_entry140_changed (GtkWidget *, SPSData *);  // Interface name for IP to link-layer conversion
int on_spinbutton1_value_changed (GtkWidget *, SPSData *);  // Timeout for ARP/Neighbor discovery
int on_entry137_changed (GtkWidget *, SPSData *);  // IPv4 Address to link-layer (MAC) address
int on_entry138_changed (GtkWidget *, SPSData *);  // IPv6 Address to link-layer (MAC) address
int on_button146_clicked (GtkButton *, SPSData *);  // Send an ARP request.
int on_button147_clicked (GtkButton *, SPSData *);  // Send a neighbor solicitation.

// Traceroute page (sps.ui)

// Probe Protocol
int on_radiobutton21_toggled (GtkButton *, SPSData *);  // TCP (IPv4) probe packet
int on_radiobutton22_toggled (GtkButton *, SPSData *);  // ICMP (IPv4) probe packet
int on_radiobutton23_toggled (GtkButton *, SPSData *);  // UDP (IPv4) probe packet
int on_radiobutton26_toggled (GtkButton *, SPSData *);  // TCP (IPv6) probe packet
int on_radiobutton27_toggled (GtkButton *, SPSData *);  // ICMP (IPv6) probe packet
int on_radiobutton28_toggled (GtkButton *, SPSData *);  // UDP (IPv6) probe packet
int on_checkbutton29_toggled (GtkButton *, SPSData *);  // Tunnel IPv6 over IPv4

int on_spinbutton2_value_changed (GtkWidget *, SPSData *);  // Number of probes per hop
int on_spinbutton4_value_changed (GtkWidget *, SPSData *);  // Timeout for reply
int on_spinbutton3_value_changed (GtkWidget *, SPSData *);  // Maximum number of hops
// Resolve Node Hostnames
int on_checkbutton28_toggled (GtkButton *, SPSData *);  // Yes
// Edit, Start, Stop buttons
int on_button23_clicked (GtkButton *, SPSData *);  // Edit packets
int on_traceroute_window_destroy (GtkWidget *, SPSData *);  // Set traceroute_flag to zero when traceroute_window is destroyed
int on_button24_clicked (GtkButton *, SPSData *);  // Start
int on_button25_clicked (GtkButton *, SPSData *);  // Stop

// IP / ASN delegation page (sps.ui)
int on_checkbutton38_toggled (GtkButton *button, SPSData *);  // Change state of flag to show sent and received messages during file downloads
int on_button50_clicked (GtkButton *, SPSData *);  // Download ARIN delegation file.
int on_button51_clicked (GtkButton *, SPSData *);  // Download RIPE delegation file.
int on_button52_clicked (GtkButton *, SPSData *);  // Download AFRINIC delegation file.
int on_button53_clicked (GtkButton *, SPSData *);  // Download APNIC delegation file.
int on_button54_clicked (GtkButton *, SPSData *);  // Download LACNIC delegation file.
int on_button56_clicked (GtkButton *, SPSData *);  // View www.iso.org's ISO 3166-1 country code list.
int on_button57_clicked (GtkButton *, SPSData *);  // Download all RIR files.
int on_entry500_changed (GtkWidget *, SPSData *);  // Country name (for country name/code search).
int on_entry501_changed (GtkWidget *, SPSData *);  // Country code (for country name/code search).
int on_entry502_changed (GtkWidget *, SPSData *);  // Search for autonomous system (AS) number delegations by country name.
int on_entry503_changed (GtkWidget *, SPSData *);  // Search for autonomous system (AS) number delegations by country code.
int on_entry504_changed (GtkWidget *, SPSData *);  // Reverse search for country by autonomous system (AS) number delegations.
int on_entry505_changed (GtkWidget *, SPSData *);  // Search for IP delegations by country name.
int on_entry506_changed (GtkWidget *, SPSData *);  // Search for IP delegations by country code.
int on_entry507_changed (GtkWidget *, SPSData *);  // Reverse-search for IPv4 address delegations by IPv4 address.
int on_entry508_changed (GtkWidget *, SPSData *);  // Reverse-search for IPv6 address delegations by IPv6 address.

// Load/Save page (sps.ui)

// Load SPS file, which contains settings selected by user when saved
int on_button21_clicked (GtkButton *, SPSData *);

// Save
// Send Packet settings
int on_checkbutton25_toggled (GtkButton *, SPSData *);  // Packet type, Number of packets to send, Flood mode
// IPv6 over IPv4 (6to4)
int on_checkbutton15_toggled (GtkButton *, SPSData *);  // IPv4 Ethernet header and IPv4 header
// IPv4
int on_checkbutton16_toggled (GtkButton *, SPSData *);  // TCP packet
int on_checkbutton20_toggled (GtkButton *, SPSData *);  // ICMP packet
int on_checkbutton21_toggled (GtkButton *, SPSData *);  // UDP packet
// IPv6
int on_checkbutton22_toggled (GtkButton *, SPSData *);  // TCP packet
int on_checkbutton23_toggled (GtkButton *, SPSData *);  // ICMP packet
int on_checkbutton24_toggled (GtkButton *, SPSData *);  // UDP packet
// 6to4
int on_checkbutton14_toggled (GtkButton *, SPSData *);  // 6to4 TCP (ethernet header, IPv4 header, IPv4 options + padding)
int on_checkbutton17_toggled (GtkButton *, SPSData *);  // 6to4 ICMP (ethernet header, IPv4 header, IPv4 options + padding)
int on_checkbutton18_toggled (GtkButton *, SPSData *);  // 6to4 UDP (ethernet header, IPv4 header, IPv4 options + padding)
// Traceroute settings
int on_checkbutton36_toggled (GtkButton *, SPSData *);  // Traceroute settings
// IPv6 over IPv4 (6to4) for traceroute
int on_checkbutton46_toggled (GtkButton *, SPSData *);  // IPv4 Ethernet header and IPv4 header for traceroute
// IPv4 traceroute
int on_checkbutton39_toggled (GtkButton *, SPSData *);  // TCP packet for traceroute
int on_checkbutton41_toggled (GtkButton *, SPSData *);  // ICMP packet for traceroute
int on_checkbutton42_toggled (GtkButton *, SPSData *);  // UDP packet for traceroute
// IPv6 traceroute
int on_checkbutton43_toggled (GtkButton *, SPSData *);  // TCP packet for traceroute
int on_checkbutton44_toggled (GtkButton *, SPSData *);  // ICMP packet for traceroute
int on_checkbutton45_toggled (GtkButton *, SPSData *);  // UDP packet for traceroute
// Save SPS file, which contains settings selected by user
int on_button14_clicked (GtkButton *, SPSData *);

// Pseudo-random number generator page (sps.ui)
int on_entry319_changed (GtkWidget *, SPSData *);  // Seed for PRNG
int on_entry320_changed (GtkWidget *, SPSData *);  // Stride for PRNG

// Traceroute (traceroute.ui)
// TCP page (IPv4) (traceroute.ui)

int on_tr_button500_clicked (GtkButton *, SPSData *);  // Copy all data from main TCP (IPv4) page

// Ethernet header
int on_tr_entry164_changed (GtkWidget *, SPSData *);  // Destination link-layer (MAC) address
int on_tr_entry165_changed (GtkWidget *, SPSData *);  // Source link-layer (MAC) address
int on_tr_entry166_changed (GtkWidget *, SPSData *);  // Source interface name
int on_tr_spinbutton11_value_changed (GtkWidget *, SPSData *);  // Maximum transmission unit (MTU) - Changed via spinbutton + or -.
int on_tr_spinbutton11_change_value (GtkWidget *, SPSData *);  // Maximum transmission unit (MTU) - Changed via spinbutton text entry.
int on_tr_entry167_changed (GtkWidget *, SPSData *);  // Ethernet type code

// IPv4 header
int on_tr_entry168_changed (GtkWidget *, SPSData *);  // IPv4 header length
int on_tr_entry169_changed (GtkWidget *, SPSData *);  // Protocol version
int on_tr_entry170_changed (GtkWidget *, SPSData *);  // Type-of-service
int on_tr_entry171_changed (GtkWidget *, SPSData *);  // Total length of datagram
int on_tr_entry172_changed (GtkWidget *, SPSData *);  // ID sequence number
int on_tr_entry173_changed (GtkWidget *, SPSData *);  // Zero
int on_tr_entry174_changed (GtkWidget *, SPSData *);  // Do not fragment flag
int on_tr_entry175_changed (GtkWidget *, SPSData *);  // More fragments following flag
int on_tr_entry176_changed (GtkWidget *, SPSData *);  // Fragmentation offset
int on_tr_entry177_changed (GtkWidget *, SPSData *);  // Time-to-Live
int on_tr_entry178_changed (GtkWidget *, SPSData *);  // Transport layer protocol
int on_tr_entry179_changed (GtkWidget *, SPSData *);  // Source IPv4 address
int on_tr_entry180_changed (GtkWidget *, SPSData *);  // Destination IPv4 address
int on_tr_entry181_changed (GtkWidget *, SPSData *);  // IPv4 Checksum
// IPv4 header options
int on_tr_radiobutton49_toggled (GtkButton *, SPSData *);  // Decimal option data entry
int on_tr_radiobutton50_toggled (GtkButton *);  // Hexadecimal option data entry
int on_tr_entry416_changed (GtkWidget *, SPSData *);  // Options data
int on_tr_button74_clicked (GtkButton *, SPSData *);  // Insert IP option after...
int on_tr_entry417_changed (GtkWidget *, SPSData *);  // Option number after which to insert new option
int on_tr_button75_clicked (GtkButton *, SPSData *);  // Append IP option
int on_tr_button76_clicked (GtkButton *, SPSData *);  // View IP options
int on_tr_entry418_changed (GtkWidget *, SPSData *);  // Option number to remove
int on_tr_button77_clicked (GtkButton *, SPSData *);  // Remove option

// TCP header (IPv4)
int on_tr_entry182_changed (GtkWidget *, SPSData *);  // Source Port
int on_tr_checkbutton30_toggled (GtkButton *, SPSData *);  // Randomize source port
int on_tr_entry183_changed (GtkWidget *, SPSData *);  // Destination port
int on_tr_entry184_changed (GtkWidget *, SPSData *);  // Sequence number
int on_tr_entry185_changed (GtkWidget *, SPSData *);  // Acknowledgement number
int on_tr_entry186_changed (GtkWidget *, SPSData *);  // Reserved 4 bits
int on_tr_entry187_changed (GtkWidget *, SPSData *);  // Data offset
int on_tr_entry188_changed (GtkWidget *, SPSData *);  // FIN flag
int on_tr_entry189_changed (GtkWidget *, SPSData *);  // SYN flag
int on_tr_entry190_changed (GtkWidget *, SPSData *);  // RST flag
int on_tr_entry191_changed (GtkWidget *, SPSData *);  // PSH flag
int on_tr_entry192_changed (GtkWidget *, SPSData *);  // ACK flag
int on_tr_entry193_changed (GtkWidget *, SPSData *);  // URG flag
int on_tr_entry194_changed (GtkWidget *, SPSData *);  // ECE flag
int on_tr_entry195_changed (GtkWidget *, SPSData *);  // CWR flag
int on_tr_entry196_changed (GtkWidget *, SPSData *);  // Window size
int on_tr_entry197_changed (GtkWidget *, SPSData *);  // TCP checksum
int on_tr_entry198_changed (GtkWidget *, SPSData *);  // Urgent pointer
int on_tr_radiobutton29_toggled (GtkButton *, SPSData *);  // ASCII data entry
int on_tr_radiobutton30_toggled (GtkButton *);  // Hexadecimal data entry
int on_tr_entry199_changed (GtkWidget *, SPSData *);  // TCP data from keyboard
int on_tr_button29_clicked (GtkButton *, SPSData *);  // TCP data from file
int on_tr_button206_clicked (GtkButton *, SPSData *);  // Unload file data from packet
// TCP options (IPv4)
int on_tr_radiobutton51_toggled (GtkButton *, SPSData *);  // Decimal option data entry
int on_tr_radiobutton52_toggled (GtkButton *);  // Hexadecimal option data entry
int on_tr_entry419_changed (GtkWidget *, SPSData *);  // Option data
int on_tr_button78_clicked (GtkButton *, SPSData *);  // Insert IP option after...
int on_tr_entry420_changed (GtkWidget *, SPSData *);  // Option number after which to insert new option
int on_tr_button79_clicked (GtkButton *, SPSData *);  // Append TCP option
int on_tr_button80_clicked (GtkButton *, SPSData *);  // View TCP options
int on_tr_entry421_changed (GtkWidget *, SPSData *);  // Option number to remove
int on_tr_button81_clicked (GtkButton *, SPSData *);  // Remove option

int on_tr_button28_clicked (GtkButton *, SPSData *);  // View IPv4 TCP packet
int on_tr_button30_clicked (GtkButton *, SPSData *);  // Restore IPv4 TCP default values

// ICMP page (IPv4) (traceroute.ui)

int on_tr_button501_clicked (GtkButton *, SPSData *);  // Copy all data from main ICMP (IPv4) page

// Ethernet header
int on_tr_entry200_changed (GtkWidget *, SPSData *);  // Destination link-layer (MAC) address
int on_tr_entry201_changed (GtkWidget *, SPSData *);  // Source link-layer (MAC) address
int on_tr_entry202_changed (GtkWidget *, SPSData *);  // Source interface name
int on_tr_spinbutton12_value_changed (GtkWidget *, SPSData *);  // Maximum transmission unit (MTU) - Changed via spinbutton + or -.
int on_tr_spinbutton12_change_value (GtkWidget *, SPSData *);  // Maximum transmission unit (MTU) - Changed via spinbutton text entry.
int on_tr_entry203_changed (GtkWidget *, SPSData *);  // Ethernet type code

// IPv4 header
int on_tr_entry204_changed (GtkWidget *, SPSData *);  // IPv4 Header Length
int on_tr_entry205_changed (GtkWidget *, SPSData *);  // Protocol Version
int on_tr_entry206_changed (GtkWidget *, SPSData *);  // Type of Service
int on_tr_entry207_changed (GtkWidget *, SPSData *);  // Total Length of Datagram
int on_tr_entry208_changed (GtkWidget *, SPSData *);  // ID Sequence Number
int on_tr_entry209_changed (GtkWidget *, SPSData *);  // Zero
int on_tr_entry210_changed (GtkWidget *, SPSData *);  // Do Not Fragment Flag
int on_tr_entry211_changed (GtkWidget *, SPSData *);  // More Fragments Following
int on_tr_entry212_changed (GtkWidget *, SPSData *);  // Fragmentation Offset
int on_tr_entry213_changed (GtkWidget *, SPSData *);  // Time-to-Live
int on_tr_entry214_changed (GtkWidget *, SPSData *);  // Transport Layer Protocol
int on_tr_entry215_changed (GtkWidget *, SPSData *);  // Source IPv4 Address
int on_tr_entry216_changed (GtkWidget *, SPSData *);  // Destination IPv4 Address
int on_tr_entry217_changed (GtkWidget *, SPSData *);  // IPv4 Checksum
// IPv4 header options
int on_tr_radiobutton55_toggled (GtkButton *, SPSData *);  // Decimal option data entry
int on_tr_radiobutton56_toggled (GtkButton *);  // Hexadecimal option data entry
int on_tr_entry425_changed (GtkWidget *, SPSData *);  // Options data
int on_tr_button86_clicked (GtkButton *, SPSData *);  // Insert IP option after...
int on_tr_entry426_changed (GtkWidget *, SPSData *);  // Option number after which to insert new option
int on_tr_button87_clicked (GtkButton *, SPSData *);  // Append IP option
int on_tr_button88_clicked (GtkButton *, SPSData *);  // View IP options
int on_tr_entry427_changed (GtkWidget *, SPSData *);  // Option number to remove
int on_tr_button89_clicked (GtkButton *, SPSData *);  // Remove option

// ICMP message (IPv4)
int on_tr_entry218_changed (GtkWidget *, SPSData *);  // Message type
int on_tr_entry219_changed (GtkWidget *, SPSData *);  // Message code
int on_tr_entry220_changed (GtkWidget *, SPSData *);  // ICMP checksum
int on_tr_entry221_changed (GtkWidget *, SPSData *);  // ID
int on_tr_entry222_changed (GtkWidget *, SPSData *);  // Sequence number
int on_tr_radiobutton31_toggled (GtkButton *, SPSData *);  // ASCII data entry
int on_tr_radiobutton32_toggled (GtkButton *);  // Hexadecimal data entry
int on_tr_entry223_changed (GtkWidget *, SPSData *);  // ICMP data from keyboard
int on_tr_button32_clicked (GtkButton *, SPSData *);  // ICMP data from file
int on_tr_button207_clicked (GtkButton *, SPSData *);  // Unload file data from packet

int on_tr_button31_clicked (GtkButton *, SPSData *);  // View IPv4 ICMP packet
int on_tr_button33_clicked (GtkButton *, SPSData *);  // Restore IPv4 ICMP default values

// UDP page (IPv4) (traceroute.ui)

int on_tr_button502_clicked (GtkButton *, SPSData *);  // Copy all data from main UDP (IPv4) page

// Ethernet header
int on_tr_entry224_changed (GtkWidget *, SPSData *);  // Destination link-layer (MAC) address
int on_tr_entry225_changed (GtkWidget *, SPSData *);  // Source link-layer (MAC) address
int on_tr_entry226_changed (GtkWidget *, SPSData *);  // Source interface name
int on_tr_spinbutton13_value_changed (GtkWidget *, SPSData *);  // Maximum transmission unit (MTU) - Changed via spinbutton + or -.
int on_tr_spinbutton13_change_value (GtkWidget *, SPSData *);  // Maximum transmission unit (MTU) - Changed via spinbutton text entry.
int on_tr_entry227_changed (GtkWidget *, SPSData *);  // Ethernet type code

// IPv4 header
int on_tr_entry228_changed (GtkWidget *, SPSData *);  // IPv4 Header Length
int on_tr_entry229_changed (GtkWidget *, SPSData *);  // Protocol Version
int on_tr_entry230_changed (GtkWidget *, SPSData *);  // Type of Service
int on_tr_entry231_changed (GtkWidget *, SPSData *);  // Total Length of Datagram
int on_tr_entry232_changed (GtkWidget *, SPSData *);  // ID Sequence Number
int on_tr_entry233_changed (GtkWidget *, SPSData *);  // Zero
int on_tr_entry234_changed (GtkWidget *, SPSData *);  // Do Not Fragment Flag
int on_tr_entry235_changed (GtkWidget *, SPSData *);  // More Fragments Following
int on_tr_entry236_changed (GtkWidget *, SPSData *);  // Fragmentation Offset
int on_tr_entry237_changed (GtkWidget *, SPSData *);  // Time-to-Live
int on_tr_entry238_changed (GtkWidget *, SPSData *);  // Transport Layer Protocol
int on_tr_entry239_changed (GtkWidget *, SPSData *);  // Source IPv4 Address
int on_tr_entry240_changed (GtkWidget *, SPSData *);  // Destination IPv4 Address
int on_tr_entry241_changed (GtkWidget *, SPSData *);  // IPv4 Checksum
// IPv4 header options
int on_tr_radiobutton57_toggled (GtkButton *, SPSData *);  // Decimal option data entry
int on_tr_radiobutton58_toggled (GtkButton *);  // Hexadecimal option data entry
int on_tr_entry428_changed (GtkWidget *, SPSData *);  // Options data
int on_tr_button90_clicked (GtkButton *, SPSData *);  // Insert IP option after...
int on_tr_entry429_changed (GtkWidget *, SPSData *);  // Option number after which to insert new option
int on_tr_button91_clicked (GtkButton *, SPSData *);  // Append IP option
int on_tr_button92_clicked (GtkButton *, SPSData *);  // View IP options
int on_tr_entry430_changed (GtkWidget *, SPSData *);  // Option number to remove
int on_tr_button93_clicked (GtkButton *, SPSData *);  // Remove option

// UDP header (IPv4)
int on_tr_entry242_changed (GtkWidget *, SPSData *);  // Source Port
int on_tr_checkbutton35_toggled (GtkButton *, SPSData *);  // Randomize Source Port
int on_tr_entry243_changed (GtkWidget *, SPSData *);  // Destination Port
int on_tr_entry244_changed (GtkWidget *, SPSData *);  // Length of UDP Datagram
int on_tr_entry245_changed (GtkWidget *, SPSData *);  // UDP Checksum
int on_tr_radiobutton33_toggled (GtkButton *, SPSData *);  // ASCII data entry
int on_tr_radiobutton34_toggled (GtkButton *);  // Hexadecimal data entry
int on_tr_entry246_changed (GtkWidget *, SPSData *);  // UDP Data from keyboard
int on_tr_button35_clicked (GtkButton *, SPSData *);  // UDP data from file
int on_tr_button208_clicked (GtkButton *, SPSData *);  // Unload file data from packet

int on_tr_button34_clicked (GtkButton *, SPSData *);  // View IPv4 UDP packet
int on_tr_button36_clicked (GtkButton *, SPSData *);  // Restore IPv4 UDP default values

// TCP page (IPv6) (traceroute.ui)

int on_tr_button503_clicked (GtkButton *, SPSData *);  // Copy all data from main TCP (IPv6) page

// Ethernet header
int on_tr_entry247_changed (GtkWidget *, SPSData *);  // Destination link-layer (MAC) address
int on_tr_entry248_changed (GtkWidget *, SPSData *);  // Source link-layer (MAC) address
int on_tr_entry249_changed (GtkWidget *, SPSData *);  // Source interface name
int on_tr_spinbutton14_value_changed (GtkWidget *, SPSData *);  // Maximum transmission unit (MTU) - Changed via spinbutton + or -.
int on_tr_spinbutton14_change_value (GtkWidget *, SPSData *);  // Maximum transmission unit (MTU) - Changed via spinbutton text entry.
int on_tr_entry250_changed (GtkWidget *, SPSData *);  // Ethernet type code
// IPv6 header
int on_tr_entry251_changed (GtkWidget *, SPSData *);  // IP version
int on_tr_entry252_changed (GtkWidget *, SPSData *);  // Traffic class
int on_tr_entry253_changed (GtkWidget *, SPSData *);  // Flow label
int on_tr_entry254_changed (GtkWidget *, SPSData *);  // Payload length
int on_tr_entry255_changed (GtkWidget *, SPSData *);  // Next header
int on_tr_entry256_changed (GtkWidget *, SPSData *);  // Hop limit
int on_tr_entry257_changed (GtkWidget *, SPSData *);  // Source IPv6 address
int on_tr_entry258_changed (GtkWidget *, SPSData *);  // Destination IPv6 address
// TCP header (IPv6)
int on_tr_entry259_changed (GtkWidget *, SPSData *);  // Source Port
int on_tr_checkbutton37_toggled (GtkButton *, SPSData *);  // Randomize source port
int on_tr_entry260_changed (GtkWidget *, SPSData *);  // Destination port
int on_tr_entry261_changed (GtkWidget *, SPSData *);  // Sequence number
int on_tr_entry262_changed (GtkWidget *, SPSData *);  // Acknowledgement number
int on_tr_entry263_changed (GtkWidget *, SPSData *);  // Reserved 4 bits
int on_tr_entry264_changed (GtkWidget *, SPSData *);  // Data offset
int on_tr_entry265_changed (GtkWidget *, SPSData *);  // FIN flag
int on_tr_entry266_changed (GtkWidget *, SPSData *);  // SYN flag
int on_tr_entry267_changed (GtkWidget *, SPSData *);  // RST flag
int on_tr_entry268_changed (GtkWidget *, SPSData *);  // PSH flag
int on_tr_entry269_changed (GtkWidget *, SPSData *);  // ACK flag
int on_tr_entry270_changed (GtkWidget *, SPSData *);  // URG flag
int on_tr_entry271_changed (GtkWidget *, SPSData *);  // ECE flag
int on_tr_entry272_changed (GtkWidget *, SPSData *);  // CWR flag
int on_tr_entry273_changed (GtkWidget *, SPSData *);  // Window size
int on_tr_entry274_changed (GtkWidget *, SPSData *);  // TCP checksum
int on_tr_entry275_changed (GtkWidget *, SPSData *);  // Urgent pointer
int on_tr_radiobutton35_toggled (GtkButton *, SPSData *);  // ASCII data entry
int on_tr_radiobutton36_toggled (GtkButton *);  // Hexadecimal data entry
int on_tr_entry276_changed (GtkWidget *, SPSData *);  // TCP data from keyboard
int on_tr_button38_clicked (GtkButton *, SPSData *);  // TCP data from file
int on_tr_button209_clicked (GtkButton *, SPSData *);  // Unload file data from packet
// TCP options (IPv6)
int on_tr_radiobutton53_toggled (GtkButton *, SPSData *);  // Decimal option data entry
int on_tr_radiobutton54_toggled (GtkButton *);  // Hexadecimal option data entry
int on_tr_entry422_changed (GtkWidget *, SPSData *);  // Option data
int on_tr_button82_clicked (GtkButton *, SPSData *);  // Insert IP option after...
int on_tr_entry423_changed (GtkWidget *, SPSData *);  // Option number after which to insert new option
int on_tr_button83_clicked (GtkButton *, SPSData *);  // Append TCP option
int on_tr_button84_clicked (GtkButton *, SPSData *);  // View TCP options
int on_tr_entry424_changed (GtkWidget *, SPSData *);  // Option number to remove
int on_tr_button85_clicked (GtkButton *, SPSData *);  // Remove option
// Extension headers
int on_tr_button101_clicked (GtkWidget *, SPSData *);  // Hop-by-hop extension header
int on_tr_button113_clicked (GtkWidget *, SPSData *);  // Destination extension header
int on_tr_button135_clicked (GtkWidget *, SPSData *);  // Routing extension header
int on_tr_button119_clicked (GtkWidget *, SPSData *);  // Authentication extension header
int on_tr_button129_clicked (GtkWidget *, SPSData *);  // Encapsulating security payload (ESP) extension header
// Outer extension headers for AH or ESP tunneling
int on_tr_button309_clicked (GtkWidget *, SPSData *);  // Hop-by-hop extension header
int on_tr_button310_clicked (GtkWidget *, SPSData *);  // Destination extension header
int on_tr_button311_clicked (GtkWidget *, SPSData *);  // Routing extension header

int on_tr_button37_clicked (GtkButton *, SPSData *);  // View IPv6 TCP packet
int on_tr_button39_clicked (GtkButton *, SPSData *);  // Restore IPv6 TCP default values

// ICMP page (IPv6) (traceroute.ui)

int on_tr_button504_clicked (GtkButton *, SPSData *);  // Copy all data from main ICMP (IPv6) page

// Ethernet header
int on_tr_entry277_changed (GtkWidget *, SPSData *);  // Destination link-layer (MAC) address
int on_tr_entry278_changed (GtkWidget *, SPSData *);  // Source link-layer (MAC) address
int on_tr_entry279_changed (GtkWidget *, SPSData *);  // Source interface name
int on_tr_spinbutton15_value_changed (GtkWidget *, SPSData *);  // Maximum transmission unit (MTU) - Changed via spinbutton + or -.
int on_tr_spinbutton15_change_value (GtkWidget *, SPSData *);  // Maximum transmission unit (MTU) - Changed via spinbutton text entry.
int on_tr_entry280_changed (GtkWidget *, SPSData *);  // Ethernet type code
// IPv6 header
int on_tr_entry281_changed (GtkWidget *, SPSData *);  // IP version
int on_tr_entry282_changed (GtkWidget *, SPSData *);  // Traffic class
int on_tr_entry283_changed (GtkWidget *, SPSData *);  // Flow label
int on_tr_entry284_changed (GtkWidget *, SPSData *);  // Payload length
int on_tr_entry285_changed (GtkWidget *, SPSData *);  // Next header
int on_tr_entry286_changed (GtkWidget *, SPSData *);  // Hop limit
int on_tr_entry287_changed (GtkWidget *, SPSData *);  // Source IPv6 address
int on_tr_entry288_changed (GtkWidget *, SPSData *);  // Destination IPv6 address
// ICMP message (IPv6)
int on_tr_entry289_changed (GtkWidget *, SPSData *);  // Message type
int on_tr_entry290_changed (GtkWidget *, SPSData *);  // Message code
int on_tr_entry291_changed (GtkWidget *, SPSData *);  // ICMP checksum
int on_tr_entry292_changed (GtkWidget *, SPSData *);  // ID
int on_tr_entry293_changed (GtkWidget *, SPSData *);  // Sequence number
int on_tr_radiobutton37_toggled (GtkButton *, SPSData *);  // ASCII data entry
int on_tr_radiobutton38_toggled (GtkButton *);  // Hexadecimal data entry
int on_tr_entry294_changed (GtkWidget *, SPSData *);  // ICMP data from keyboard
int on_tr_button41_clicked (GtkButton *, SPSData *);  // ICMP data from file
int on_tr_button210_clicked (GtkButton *, SPSData *);  // Unload file data from packet
// Extension headers
int on_tr_button102_clicked (GtkWidget *, SPSData *);  // Hop-by-hop extension header
int on_tr_button114_clicked (GtkWidget *, SPSData *);  // Destination extension header
int on_tr_button136_clicked (GtkWidget *, SPSData *);  // Routing extension header
int on_tr_button120_clicked (GtkWidget *, SPSData *);  // Authentication extension header
int on_tr_button130_clicked (GtkWidget *, SPSData *);  // Encapsulating security payload (ESP) extension header
// Outer extension headers for AH or ESP tunneling
int on_tr_button312_clicked (GtkWidget *, SPSData *);  // Hop-by-hop extension header
int on_tr_button313_clicked (GtkWidget *, SPSData *);  // Destination extension header
int on_tr_button314_clicked (GtkWidget *, SPSData *);  // Routing extension header

int on_tr_button40_clicked (GtkButton *, SPSData *);  // View IPv6 ICMP packet
int on_tr_button42_clicked (GtkButton *, SPSData *);  // Restore IPv6 ICMP default values

// UDP page (IPv6) (traceroute.ui)

int on_tr_button505_clicked (GtkButton *, SPSData *);  // Copy all data from main UDP (IPv6) page

// Ethernet header
int on_tr_entry295_changed (GtkWidget *, SPSData *);  // Destination link-layer (MAC) address
int on_tr_entry296_changed (GtkWidget *, SPSData *);  // Source link-layer (MAC) address
int on_tr_entry297_changed (GtkWidget *, SPSData *);  // Source interface name
int on_tr_spinbutton16_value_changed (GtkWidget *, SPSData *);  // Maximum transmission unit (MTU) - Changed via spinbutton + or -.
int on_tr_spinbutton16_change_value (GtkWidget *, SPSData *);  // Maximum transmission unit (MTU) - Changed via spinbutton text entry.
int on_tr_entry298_changed (GtkWidget *, SPSData *);  // Ethernet type code
// IPv6 header
int on_tr_entry299_changed (GtkWidget *, SPSData *);  // IP version
int on_tr_entry300_changed (GtkWidget *, SPSData *);  // Traffic class
int on_tr_entry301_changed (GtkWidget *, SPSData *);  // Flow label
int on_tr_entry302_changed (GtkWidget *, SPSData *);  // Payload length
int on_tr_entry303_changed (GtkWidget *, SPSData *);  // Next header
int on_tr_entry304_changed (GtkWidget *, SPSData *);  // Hop limit
int on_tr_entry305_changed (GtkWidget *, SPSData *);  // Source IPv6 address
int on_tr_entry306_changed (GtkWidget *, SPSData *);  // Destination IPv6 address
// UDP header (IPv6)
int on_tr_entry307_changed (GtkWidget *, SPSData *);  // Source Port
int on_tr_checkbutton40_toggled (GtkButton *, SPSData *);  // Randomize Source Port
int on_tr_entry308_changed (GtkWidget *, SPSData *);  // Destination Port
int on_tr_entry309_changed (GtkWidget *, SPSData *);  // Length of UDP Datagram
int on_tr_entry310_changed (GtkWidget *, SPSData *);  // UDP Checksum
int on_tr_radiobutton39_toggled (GtkButton *, SPSData *);  // ASCII data entry
int on_tr_radiobutton40_toggled (GtkButton *);  // Hexadecimal data entry
int on_tr_entry311_changed (GtkWidget *, SPSData *);  // UDP Data from keyboard
int on_tr_button44_clicked (GtkButton *, SPSData *);  // UDP data from file
int on_tr_button211_clicked (GtkButton *, SPSData *);  // Unload file data from packet
// Extension headers
int on_tr_button103_clicked (GtkWidget *, SPSData *);  // Hop-by-hop extension header
int on_tr_button115_clicked (GtkWidget *, SPSData *);  // Destination extension header
int on_tr_button137_clicked (GtkWidget *, SPSData *);  // Routing extension header
int on_tr_button121_clicked (GtkWidget *, SPSData *);  // Authentication extension header
int on_tr_button131_clicked (GtkWidget *, SPSData *);  // Encapsulating security payload (ESP) extension header
// Outer extension headers for AH or ESP tunneling
int on_tr_button315_clicked (GtkWidget *, SPSData *);  // Hop-by-hop extension header
int on_tr_button316_clicked (GtkWidget *, SPSData *);  // Destination extension header
int on_tr_button317_clicked (GtkWidget *, SPSData *);  // Routing extension header

int on_tr_button43_clicked (GtkButton *, SPSData *);  // View IPv6 UDP packet
int on_tr_button45_clicked (GtkButton *, SPSData *);  // Restore IPv6 UDP default values

// Extension header pages

// Hop-by-hop extension header callback functions (hop-by-hop.ui)
int on_hop_window_destroy (GtkWidget *, SPSData *);  // Set hop_flag to zero when hop_window is destroyed
int on_hop_button142_clicked (GtkButton *, SPSData *);  // Toggle: attach / detach a hop-by-hop header
int on_hop_entry438_changed (GtkWidget *, SPSData *);  // Hop-by-hop header length
int on_hop_radiobutton59_toggled (GtkButton *, SPSData *);  // Decimal option data entry
int on_hop_radiobutton60_toggled (GtkButton *);  // Hexadecimal option data entry
int on_hop_entry158_changed (GtkWidget *, SPSData *);  // Hop-by-hop option data
int on_hop_entry313_changed (GtkWidget *, SPSData *);  // Option alignment parameter x (of xN + y)
int on_hop_entry314_changed (GtkWidget *, SPSData *);  // Option alignment parameter y (of xN + y)
int on_hop_button94_clicked (GtkButton *, SPSData *);  // Insert hop-by-hop option after...
int on_hop_entry159_changed (GtkWidget *, SPSData *);  // Option number after which to insert new option
int on_hop_button95_clicked (GtkButton *, SPSData *);  // Append hop-by-hop option
int on_hop_button96_clicked (GtkButton *, SPSData *);  // View hop-by-hop options
int on_hop_entry312_changed (GtkWidget *, SPSData *);  // Option number to remove
int on_hop_button97_clicked (GtkButton *, SPSData *);  // Remove option

// Hop-by-hop (outer) extension header callback functions (hop-by-hop-outer.ui)
int on_hop_window_outer_destroy (GtkWidget *, SPSData *);  // Set hop_flag to zero when hop_window is destroyed
int on_hop_button142_outer_clicked (GtkButton *, SPSData *);  // Toggle: attach / detach a hop-by-hop header
int on_hop_entry438_outer_changed (GtkWidget *, SPSData *);  // Hop-by-hop header length
int on_hop_radiobutton59_outer_toggled (GtkButton *, SPSData *);  // Decimal option data entry
int on_hop_radiobutton60_outer_toggled (GtkButton *);  // Hexadecimal option data entry
int on_hop_entry158_outer_changed (GtkWidget *, SPSData *);  // Hop-by-hop option data
int on_hop_entry313_outer_changed (GtkWidget *, SPSData *);  // Option alignment parameter x (of xN + y)
int on_hop_entry314_outer_changed (GtkWidget *, SPSData *);  // Option alignment parameter y (of xN + y)
int on_hop_button94_outer_clicked (GtkButton *, SPSData *);  // Insert hop-by-hop option after...
int on_hop_entry159_outer_changed (GtkWidget *, SPSData *);  // Option number after which to insert new option
int on_hop_button95_outer_clicked (GtkButton *, SPSData *);  // Append hop-by-hop option
int on_hop_button96_outer_clicked (GtkButton *, SPSData *);  // View hop-by-hop options
int on_hop_entry312_outer_changed (GtkWidget *, SPSData *);  // Option number to remove
int on_hop_button97_outer_clicked (GtkButton *, SPSData *);  // Remove option

// Destination extension header callback functions (destination.ui)
int on_dst_window_destroy (GtkWidget *, SPSData *);  // Set dst_flag to zero when dst_window is destroyed
int on_dst_button143_clicked (GtkButton *, SPSData *);  // Toggle: attach / detach a "first" destination header
int on_dst_button144_clicked (GtkButton *, SPSData *);  // Toggle: attach / detach a "last" destination header
int on_dst_entry439_changed (GtkWidget *, SPSData *);  // Destination header length (first)
int on_dst_entry440_changed (GtkWidget *, SPSData *);  // Destination header length (last)
int on_dst_radiobutton61_toggled (GtkButton *, SPSData *);  // Decimal option data entry
int on_dst_radiobutton62_toggled (GtkButton *);  // Hexadecimal option data entry
int on_dst_entry315_changed (GtkWidget *, SPSData *);  // Destination option data
int on_dst_entry156_changed (GtkWidget *, SPSData *);  // Option alignment parameter x (of xN + y)
int on_dst_entry157_changed (GtkWidget *, SPSData *);  // Option alignment parameter y (of xN + y)
int on_dst_radiobutton63_toggled (GtkButton *, SPSData *);  // Designate as first destination header
int on_dst_radiobutton64_toggled (GtkButton *);  // Designate as last destination header
int on_dst_button104_clicked (GtkButton *, SPSData *);  // Insert destination option after...
int on_dst_entry431_changed (GtkWidget *, SPSData *);  // Option number after which to insert new option
int on_dst_button105_clicked (GtkButton *, SPSData *);  // Append destination option
int on_dst_button106_clicked (GtkButton *, SPSData *);  // View destination options (first)
int on_dst_button109_clicked (GtkButton *, SPSData *);  // View destination options (last)
int on_dst_entry432_changed (GtkWidget *, SPSData *);  // Option number to remove
int on_dst_button107_clicked (GtkButton *, SPSData *);  // Remove option

// Destination (outer) extension header callback functions (destination-outer.ui)
int on_dst_window_outer_destroy (GtkWidget *, SPSData *);  // Set dst_flag to zero when dst_window is destroyed
int on_dst_button143_outer_clicked (GtkButton *, SPSData *);  // Toggle: attach / detach a "first" destination header
int on_dst_button144_outer_clicked (GtkButton *, SPSData *);  // Toggle: attach / detach a "last" destination header
int on_dst_entry439_outer_changed (GtkWidget *, SPSData *);  // Destination header length (first)
int on_dst_entry440_outer_changed (GtkWidget *, SPSData *);  // Destination header length (last)
int on_dst_radiobutton61_outer_toggled (GtkButton *, SPSData *);  // Decimal option data entry
int on_dst_radiobutton62_outer_toggled (GtkButton *);  // Hexadecimal option data entry
int on_dst_entry315_outer_changed (GtkWidget *, SPSData *);  // Destination option data
int on_dst_entry156_outer_changed (GtkWidget *, SPSData *);  // Option alignment parameter x (of xN + y)
int on_dst_entry157_outer_changed (GtkWidget *, SPSData *);  // Option alignment parameter y (of xN + y)
int on_dst_radiobutton63_outer_toggled (GtkButton *, SPSData *);  // Designate as first destination header
int on_dst_radiobutton64_outer_toggled (GtkButton *);  // Designate as last destination header
int on_dst_button104_outer_clicked (GtkButton *, SPSData *);  // Insert destination option after...
int on_dst_entry431_outer_changed (GtkWidget *, SPSData *);  // Option number after which to insert new option
int on_dst_button105_outer_clicked (GtkButton *, SPSData *);  // Append destination option
int on_dst_button106_outer_clicked (GtkButton *, SPSData *);  // View destination options (first)
int on_dst_button109_outer_clicked (GtkButton *, SPSData *);  // View destination options (last)
int on_dst_entry432_outer_changed (GtkWidget *, SPSData *);  // Option number to remove
int on_dst_button107_outer_clicked (GtkButton *, SPSData *);  // Remove option

// Routing (original/inner) extension header callback functions (routing.ui)
int on_route_window_destroy (GtkWidget *, SPSData *);  // Set route_flag to zero when route_window is destroyed
int on_rt_button145_clicked (GtkButton *, SPSData *);  // Toggle: attach / detach a routing header
int on_rt_entry446_changed (GtkWidget *, SPSData *);  // Header length
int on_rt_entry447_changed (GtkWidget *, SPSData *);  // Routing type
int on_rt_entry448_changed (GtkWidget *, SPSData *);  // Segments left
int on_rt_radiobutton73_toggled (GtkButton *, SPSData *);  // Decimal routing data entry
int on_rt_radiobutton74_toggled (GtkButton *);  // Hexadecimal routing data entry
int on_rt_entry449_changed (GtkWidget *, SPSData *);  // Routing data
int on_rt_button125_clicked (GtkButton *, SPSData *);  // Load routing data from file...

// Routing (outer) extension header callback functions (routing-outer.ui)
int on_route_window_outer_destroy (GtkWidget *, SPSData *);  // Set route_flag to zero when route_window is destroyed
int on_rt_button145_outer_clicked (GtkButton *, SPSData *);  // Toggle: attach / detach a routing header
int on_rt_entry446_outer_changed (GtkWidget *, SPSData *);  // Header length
int on_rt_entry447_outer_changed (GtkWidget *, SPSData *);  // Routing type
int on_rt_entry448_outer_changed (GtkWidget *, SPSData *);  // Segments left
int on_rt_radiobutton73_outer_toggled (GtkButton *, SPSData *);  // Decimal routing data entry
int on_rt_radiobutton74_outer_toggled (GtkButton *);  // Hexadecimal routing data entry
int on_rt_entry449_outer_changed (GtkWidget *, SPSData *);  // Routing data
int on_rt_button125_outer_clicked (GtkButton *, SPSData *);  // Load routing data from file...

// Authentication extension header callback functions (authentication.ui)
int on_auth_window_destroy (GtkWidget *, SPSData *);  // Set auth_flag to zero when auth_window is destroyed
int on_auth_button138_clicked (GtkButton *, SPSData *);  // Toggle: attach / detach an authentication header
int on_auth_button139_clicked (GtkButton *, SPSData *);  // Toggle: transport / tunnel mode
int on_auth_entry433_changed (GtkWidget *, SPSData *);  // Payload length
int on_auth_entry434_changed (GtkWidget *, SPSData *);  // Reserved
int on_auth_entry435_changed (GtkWidget *, SPSData *);  // Security parameters index (SPI)
int on_auth_entry436_changed (GtkWidget *, SPSData *);  // Sequence number
int on_auth_radiobutton67_toggled (GtkButton *, SPSData *);  // Decimal integrity check value (ICV) entry
int on_auth_radiobutton68_toggled (GtkButton *);  // Hexadecimal integrity check value (ICV) entry
int on_auth_entry437_changed (GtkWidget *, SPSData *);  // Integrity check value (ICV)
int on_auth_button108_clicked (GtkButton *, SPSData *);  // Load ICV from file ...
int on_auth_button122_clicked (GtkButton *, SPSData *);  // Save data for ICV calculation

// Encapsulating security payload (ESP) extension header callback functions (esp.ui)
int on_esp_window_destroy (GtkWidget *, SPSData *);  // Set esp_flag to zero when esp_window is destroyed
int on_esp_button140_clicked (GtkButton *, SPSData *);  // Toggle: attach / detach an ESP header
int on_esp_button141_clicked (GtkButton *, SPSData *);  // Toggle: transport / tunnel mode
int on_esp_entry441_changed (GtkWidget *, SPSData *);  // Security parameters index (SPI)
int on_esp_entry442_changed (GtkWidget *, SPSData *);  // Sequence number
int on_esp_button123_clicked (GtkButton *, SPSData *);  // Save data for ICV calculation
int on_esp_entry443_changed (GtkWidget *, SPSData *);  // Padding length
int on_esp_entry444_changed (GtkWidget *, SPSData *);  // Next header
int on_esp_radiobutton71_toggled (GtkButton *, SPSData *);  // Decimal integrity check value (ICV) entry
int on_esp_radiobutton72_toggled (GtkButton *);  // Hexadecimal integrity check value (ICV) entry
int on_esp_entry445_changed (GtkWidget *, SPSData *);  // Integrity check value (ICV)
int on_esp_button124_clicked (GtkButton *, SPSData *);  // Load ICV from file ...

// ********** NON-CALLBACK FUNCTION PROTOTYPES **********

// Send packet functions, with respective filenames.
void *ipv4_send (void *);  // Send IPv4 packet(s) (ipv4_send.c). Type is consistent with GThreadFunc().
int update_ip4_sources (SPSData *);  // (idle function) Update source IPv4 addresses and source ports, which may have changed if randomized.
void *ipv6_send (void *);  // Send IPv6 packet(s) (ipv6_send.c). Type is consistent with GThreadFunc().
int update_ip6_sources (SPSData *);  // (idle function) Update source IPv6 addresses and source ports, which may have changed if randomized.
void *sixto4_send (void *);  // Send IPv6 packet(s) over IPv4 (6to4) (sixto4_send.c). Type is consistent with GThreadFunc().
int update_6to4_sources (SPSData *);  // (idle function) Update source 6to4 IPv4 addresses, which may have changed if randomized.

// Maximum transmission unit (MTU) functions (mtu.c)
int interface_getmtu (char *);  // Obtain interface's current MTU setting.
int min4_mtu (int, SPSData *);  // Find minimum MTU for IPv4 packet.
int min6_mtu (int, SPSData *);  // Find minimum MTU for IPv6 packet.
int min6to4_mtu (int, SPSData *);  // Find minimum MTU for 6to4 packet.
int sixto4_mtu_checks (int, SPSData *);  // Perform MTU checks for 6to4.
int update_mtu_spinbutton (int, SPSData *);  // Update MTU spinbutton entry value.

// Ethernet header functions (ethernet_header.c)
int mac_entry (GtkWidget *, uint8_t *, int, SPSData *);  // Process link-layer (MAC) address entry.
int ifname_entry (GtkWidget *, GtkWidget *, int, SPSData *);  // Find link-layer (MAC) address for given interface name entry.
int sixto4_ifname_entry (GtkWidget *, GtkWidget *, int, SPSData *);  // (6to4 only) Find link-layer (MAC) address for given interface name entry.
int ethtype_entry (GtkWidget *, int, SPSData *);  // Process ethernet type code entry.
int mac_bin2string (uint8_t *, char *);  // Convert link-layer (MAC) address from binary to colon-separated string.
int mac_string2bin (char *, uint8_t *);  // Convert link-layer (MAC) address from colon-separated string to binary.

// IPv4 header functions (ipv4_header.c)
int ip4_hdrlen_entry (GtkWidget *, int, SPSData *);  // Process IPv4 header length entry.
int ip4_ver_entry (GtkWidget *, int, SPSData *);  // Process IPv4 header IP version entry.
int ip4_tos_entry (GtkWidget *, int, SPSData *);  // Process IPv4 type of service entry.
int ip4_tld_entry (GtkWidget *, int, SPSData *);  // Process IPv4 total length of datagram entry.
int ip4_seq_entry (GtkWidget *, int, SPSData *);  // Process IPv4 ID sequence number entry.
int ip4_zero_entry (GtkWidget *, GtkWidget *, GtkWidget *, GtkWidget *, int, SPSData *);  // Process IPv4 zero entry.
int ip4_dnf_entry (GtkWidget *, GtkWidget *, GtkWidget *, GtkWidget *, int, SPSData *);  // Process IPv4 do not fragment entry.
int ip4_mff_entry (GtkWidget *, GtkWidget *, GtkWidget *, GtkWidget *, int, SPSData *);  // Process IPv4 more fragments following entry.
int ip4_fo_entry (GtkWidget *, GtkWidget *, GtkWidget *, GtkWidget *, int, SPSData *);  // Process IPv4 fragmentation offset entry.
int ip4_ttl_entry (GtkWidget *, int, SPSData *);  // Process IPv4 time-to-live entry.
int ip4_tlp_entry (GtkWidget *, int, SPSData *);  // Process IPv4 transport layer protocol entry.
int ip4_sip_entry (GtkWidget *, int, SPSData *);  // Process IPv4 source IP address entry.
int ip4_ransip_checkbutton (int *, GtkWidget *, int, SPSData *);  // Process toggled checkbutton for randomizing IPv4 source IP address.
int ip4_dip_entry (GtkWidget *, int, SPSData *);  // Process IPv4 destination IP address entry.
int ip4_chksum_entry (GtkWidget *, int, SPSData *);  // Process IPv4 checksum entry.
int ip4_hlen (int, SPSData *);  // Re-calculate IPv4 header length and update text entry.
int ip4_tld (int, SPSData *);  // Re-calculate total length of IPv4 datagram and update text entry.
int truncate_ip4 (int, SPSData *);  // Truncate IPv4 payload length as necessary to accomodate the variable lengths of IPv4 header options and TCP options.
int ip4_chksum (int, SPSData *);  // Re-calculate IPv4 checksum and update text entry.
int ip4data_update (int, SPSData *);  // Update IPv4: total length of datagram and checksum, Upper layer protocol: length, offset, and checksum.

// IPv4 and TCP options functions (ipv4_and_tcp_options.c)
int compose_optstring (int, char *, int, int **, uint8_t ***, int);  // Prepare a string of text containing IPv4 or TCP options.
int ip4_option (int, int, int, GtkWidget *, SPSData *);  // Insert, remove, or append an IPv4 option.
int ip4_opt_insert_after_entry (GtkWidget *, int, int *, SPSData *);  // Number of option after which to insert new IPv4 option.
int opt_view (int, int, SPSData *);  // View IPv4 header options or TCP header (IPv4 or IPv6) options.
int ip4_opt_remove_entry (GtkWidget *, int, int *, SPSData *);  // Number of IPv4 option to remove.

// Data entry for IPv4 header or TCP header options, hop-by-hop header or destination header options, or other extension header data (data_entry.c).
int data_entry (GtkWidget *, int *, int, uint8_t *, int, GtkWidget *, SPSData *);

// IPv6 header functions (ipv6_header.c)
int ip6_ver_entry (GtkWidget *, int, SPSData *);  // Process IPv6 header IP version entry.
int ip6_tc_entry (GtkWidget *, int, SPSData *);  // Process IPv6 traffic class entry.
int ip6_fl_entry (GtkWidget *, int, SPSData *);  // Process IPv6 flow label entry.
int ip6_plen_entry (GtkWidget *, int, SPSData *);  // Process IPv6 payload length entry.
int ip6_nxt_entry (GtkWidget *, int, SPSData *);  // Process IPv6 next header entry.
int ip6_hops_entry (GtkWidget *, int, SPSData *);  // Process IPv6 hop limit entry.
int ip6_sip_entry (GtkWidget *, int, SPSData *);  // Process IPv6 source IP address entry.
int ip6_ransip_checkbutton (int *, GtkWidget *, int, SPSData *);  // Process toggled checkbutton for randomizing IPv6 source IP address.
int ip6_dip_entry (GtkWidget *, int, SPSData *);  // Process IPv6 destination IP address entry.
int truncate_ip6 (int, SPSData *);  // Truncate IPv6 payload length as necessary to accomodate the variable lengths of extension headers and TCP options.
uint16_t ip6_paylen (int, SPSData *);  // Calculate IPv6 payload length for IPv6 page or IPv6 traceroute page.
uint16_t ip6_new_paylen (int, SPSData *);  // Calculate IPv6 payload length for "new" IPv6 header when tunneling AH or ESP.
int update_plen_entry (int, uint16_t, SPSData *);  // Update payload length entry for an IPv6 page or IPv6 traceroute page.
int ip6data_update (int, SPSData *);  // Update IPv6: payload lengths, Upper layer protocol: lengths, offsets, and checksums.

// Hop-by-hop (original/inner) extension header functions (hop_header.c)
int hop_win (SPSData *);  // Create hop-by-hop extension header editing window.
int hop_show (int, SPSData *);  // Display contents of selected hop-by-hop header.
int hop_opt_view (SPSData *);  // View hop-by-hop header options.

// Hop-by-hop (outer) extension header functions (hop_header.c)
int hop_win_outer (SPSData *);  // Create hop-by-hop extension header editing window.
int hop_show_outer (int, SPSData *);  // Display contents of selected hop-by-hop header.
int hop_opt_view_outer (SPSData *);  // View hop-by-hop header options.

// Hop-by-hop header common function (hop_header.c)
int hop_option (int, int, int, SPSData *);  // Insert, remove, or append an inner or outer hop-by-hop or destination option.

// Destination (original/inner) extension header functions (dst_header.c)
int dst_win (SPSData *);  // Create destination extension header editing window.
int dst_show (int, SPSData *);  // Display contents of selected destination header.
int dst_opt_view (int, SPSData *);  // View hop-by-hop or destination header options.

// Destination (outer) extension header functions (dst_header.c)
int dst_win_outer (SPSData *);  // Create destination extension header editing window.
int dst_show_outer (int, SPSData *);  // Display contents of selected destination header.
int dst_opt_view_outer (int, SPSData *);  // View hop-by-hop or destination header options.

// Destination header common function (dst_header.c)
int dst_option (int, int, int, SPSData *);  // Insert, remove, or append a hop-by-hop or destination option.

// Hop-by-hop and destination header common functions (dst_hop_common.c)
int dsthop_opt_remove_entry (GtkWidget *, int *);  // Number of hop-by-hop or destination option to remove.
int option_pad (int *, uint8_t *, int, int, int);  // Provide padding as needed to achieve alignment requirements of hop-by-hop or destination option.
int compose_exoptstring (int, int, int, char *, int, SPSData *);  // Prepare a string of text containing hop-by-hop or destination header options.
int displ_options_window (int, char *, char *);  // Display IPv4 header options, hop-by-hop options, or destination options window.

// Routing (original/inner) extension header functions (routing_header.c)
int route_win (SPSData *);  // Create routing extension header editing window.
int route_show (SPSData *);  // Display contents of selected routing header.

// Routing (outer) extension header functions (routing_header.c)
int route_win_outer (SPSData *);  // Create routing extension header editing window.
int route_show_outer (SPSData *);  // Display contents of selected routing header.

// Routing inner/outer common function
int update_route_len (int, int, SPSData *);  // Update routing header length.

// Authentication extension header functions (auth_header.c)
int auth_win (SPSData *);  // Create authentication extenstion header editing window.
int auth_show (SPSData *);  // Display contents of selected authentication header.
int update_auth_pad (int, SPSData *);  // Update authentication header payload and padding length.

// Encapsulating security payload (ESP) extension header functions (esp_header.c)
int esp_win (SPSData *);  // Create ESP extenstion header editing window.
int esp_show (int, SPSData *);  // Display contents of selected ESP header.
int update_esp_pad (int, SPSData *);  // Update encapsulating security payload (ESP) header padding length.
int esp_paylen (int, SPSData *);  // Calculate ESP payload length.

// TCP header functions (tcp_header.c)
int tcp_sport_entry (GtkWidget *, int, SPSData *);  // Process TCP source port entry.
int tcp_ran_sport (int *, int, GtkWidget *, SPSData *);  // Create a random TCP source port number (16 bits).
int tcp_dport_entry (GtkWidget *, int, SPSData *);  // Process TCP destination port entry.
int tcp_seqnum_entry (GtkWidget *, int, SPSData *);  // Process TCP sequence number entry.
int tcp_acknum_entry (GtkWidget *, int, SPSData *);  // Process TCP acknowledgement number entry.
int tcp_res_entry (GtkWidget *, int, SPSData *);  // Process TCP reserved entry.
int tcp_off_entry (GtkWidget *, int, SPSData *);  // Process TCP data offset entry.
int tcp_fin_entry (GtkWidget *, int, SPSData *);  // Process TCP FIN flag entry.
int tcp_syn_entry (GtkWidget *, int, SPSData *);  // Process TCP SYN flag entry.
int tcp_rst_entry (GtkWidget *, int, SPSData *);  // Process TCP RST flag entry.
int tcp_psh_entry (GtkWidget *, int, SPSData *);  // Process TCP PSH flag entry.
int tcp_ack_entry (GtkWidget *, int, SPSData *);  // Process TCP ACK flag entry.
int tcp_urg_entry (GtkWidget *, int, SPSData *);  // Process TCP URG flag entry.
int tcp_ece_entry (GtkWidget *, int, SPSData *);  // Process TCP ECE flag entry.
int tcp_cwr_entry (GtkWidget *, int, SPSData *);  // Process TCP CWR flag entry.
int tcp_win_entry (GtkWidget *, int, SPSData *);  // Process TCP window size entry.
int tcp_chksum_entry (GtkWidget *, int, SPSData *);  // Process TCP checksum entry.
int tcp_urgptr_entry (GtkWidget *, int, SPSData *);  // Process TCP urgent pointer entry.
int tcp4_option (int, int, int, GtkWidget *, SPSData *);  // Insert, remove, or append an IPv4 TCP option.
int tcp6_option (int, int, int, GtkWidget *, SPSData *);  // Insert, remove, or append an IPv6 TCP option.
int tcp_opt_insert_after_entry (GtkWidget *, int, int *, SPSData *);  // Number of option after which to insert new TCP option.
int tcp_opt_remove_entry (GtkWidget *, int, int *, SPSData *);  // Number of TCP option to remove.
int tcp_dataoffset (int, SPSData *);  // Re-calculate TCP data offset and update text entry (IPv4 and IPv6).
int tcp_chksum (int, SPSData *);  // Re-calculate TCP checksum and update text entry (IPv4 or IPv6).

// ICMP header functions (icmp_header.c)
int icmp_mt_entry (GtkWidget *, int, SPSData *);  // Process ICMP message type entry.
int icmp_mc_entry (GtkWidget *, int, SPSData *);  // Process ICMP message code entry.
int icmp_chksum_entry (GtkWidget *, int, SPSData *);  // Process ICMP checksum entry.
int icmp_id_entry (GtkWidget *, int, SPSData *);  // Process ICMP ID number entry.
int icmp_seq_entry (GtkWidget *, int, SPSData *);  // Process ICMP sequence number entry.
int icmp_chksum (int, SPSData *);  // Re-calculate ICMP checksum and update text entry (IPv4 or IPv6).

// UDP header functions (udp_header.c)
int udp_sport_entry (GtkWidget *, int, SPSData *);  // Process UDP source port entry.
int udp_ran_sport (int *, int, GtkWidget *, SPSData *);  // Create a random UDP source port number (16 bits).
int udp_dport_entry (GtkWidget *, int, SPSData *);  // Process UDP destination port entry.
int udp_len_entry (GtkWidget *, int, SPSData *);  // Process length of UDP datagram entry.
int udp_chksum_entry (GtkWidget *, int, SPSData *);  // Process UDP checksum entry.
int udp_len (int, SPSData *);  // Re-calculate length of UDP datagram and update text entry (IPv4 or IPv6).
int udp_chksum (int, SPSData *);  // Re-calculate UDP checksum and update text entry (IPv4 or IPv6).

// Default settings functions for IPv4 packets (ipv4_defaults.c)
int tcp4_default (SPSData *);  // IPv4 default settings.
int icmp4_default (SPSData *);  // IPv4 default settings.
int udp4_default (SPSData *);  // IPv4 default settings.

// Default settings functions for IPv6 packets (ipv6_defaults.c)
int tcp6_default (SPSData *);  // IPv6 default settings.
int icmp6_default (SPSData *);  // IPv6 default settings.
int udp6_default (SPSData *);  // IPv6 default settings.

// Default settings functions for 6to4 packets (sixto4_defaults.c)
int sixto4_tcp6_default (SPSData *);  // 6to4 default settings.
int sixto4_icmp6_default (SPSData *);  // 6to4 default settings.
int sixto4_udp6_default (SPSData *);  // 6to4 default settings.

// Functions to link IPv6 headers into appropriate order, and perform IPv4 and IPv6 fragmentation
// (header_linking_and_fragmentation.c)
int next_hdr_to_frag (int, struct ip6_hdr *, SPSData *);  // Set the Next Header field to IPPROTO_FRAGMENT in last link of unfragmentable portion.
int find4_unfraglen (int, SPSData *);  // Returns length of unfragmentable portion (IPv4).
int find6_unfraglen (int, SPSData *);  // Returns length of unfragmentable portion (IPv6).
int find4_fraglen (int, SPSData *);  // Returns length of fragmentable portion (IPv4).
int find6_fraglen (int, SPSData *);  // Returns length of fragmentable portion (IPv6).
int build_unfrag (int, struct ip6_hdr *, int *, uint8_t *, SPSData *);  // Build unfragmentable portion of ethernet frame (IPv6).
int build_frag (int, struct ip6_hdr *, uint8_t *, SPSData *);  // Build fragmentable portion of ethernet frame (IPv6).
int ip6_chain (int, struct ip6_hdr *, SPSData *);  // Setup chain of header, extension headers, and data by setting Next Header fields (IPv6).
int ip6_order (int, SPSData *);  // Set order of all headers, extension headers, and data. Determine where in chain to separate unfragmentable from fragmentable (IPv6).
int proto_val (int val);  // Return appropriate protocol type value (IPPROTO), depending upon SPS header type index (IPv6).
int find_nframes (int, int, int, int, int *, int *, int *);  // Determine required number of IPv6 ethernet frames, frame lengths, and data offsets (IPv4 or IPv6).

// Function to process keyboard entry of upper layer protocol data (IPv4 or IPv6) (keyboard_data.c).
int key_data (GtkWidget *, int, int, SPSData *);

// Functions for loading upper layer protocol data from file (file_data.c)
int file_payload_data (GtkWidget *, int, SPSData *);  // Entry from file of upper layer protocol data (IPv4 or IPv6).
int file_data (GtkWidget *, int *, int, uint8_t *, int, GtkWidget *, SPSData *);  // Extension header data entry from file (ipv6).

// Create an IPv4 ethernet frame (create_ip4_frame.c).
int create_ip4_frame (int, SPSData *);

// Create an IPv6 ethernet frame (create_ip6_frame.c).
int create_ip6_frame (int, SPSData *);

// Create an IPv6 over IPv4 (6to4) ethernet frame (create_6to4_frame.c).
int create_6to4_frame (int, SPSData *);

// Functions to show traceroute settings on Traceroute page of sps.ui (show_traceroute_settings.c).
int tr_settings_show (SPSData *);

// IPv4 traceroute functions (ipv4_traceroute.c)
void *ipv4_tr_send (void *);  // Perform IPv4 traceroute. Type is consistent with GThreadFunc().
int free_ipv4_tr_mem (char *, uint8_t *, char *, char *, char *, char *);  // Free allocated memory used in ipv4_tr_send().
int update_tr_ip4_sources (SPSData *);  // (idle function) Update IPv4 traceroute source ports, which may have changed if randomized.
int tcp4_tr_show (SPSData *);  // Populate entries in traceroute.ui window.
int icmp4_tr_show (SPSData *);  // Populate entries in traceroute.ui window.
int udp4_tr_show (SPSData *);  // Populate entries in traceroute.ui window.

// IPv6 traceroute functions (ipv6_traceroute.c)
void *ipv6_tr_send (void *);  // Perform IPv6 traceroute. Type is consistent with GThreadFunc().
int free_ipv6_tr_mem (char *, uint8_t *, char *, char *, char *, char *);  // Free allocated memory used in ipv6_tr_send().
int update_tr_ip6_sources (SPSData *);  // (idle function) Update IPv6 traceroute source ports, which may have changed if randomized.
int tcp6_tr_show (SPSData *);  // Populate entries in traceroute.ui window.
int icmp6_tr_show (SPSData *);  // Populate entries in traceroute.ui window.
int udp6_tr_show (SPSData *);  // Populate entries in traceroute.ui window.

// 6to4 traceroute functions (sixto4_traceroute.c)
void *sixto4_tr_send (void *);  // Perform IPv6 over IPv4 (6to4) traceroute. Type is consistent with GThreadFunc().
int free_sixto4_tr_mem (char *, uint8_t *, char *, char *, char *, char *, char *);  // Free allocated memory used in sixto4_tr_send()..

// IPv4 traceroute default settings functions (ipv4_traceroute_defaults.c)
int tcp4_tr_default (SPSData *);  // IPv4 traceroute default settings.
int icmp4_tr_default (SPSData *);  // IPv4 traceroute default settings.
int udp4_tr_default (SPSData *);  // IPv4 traceroute default settings.

// IPv6 traceroute default settings functions (ipv6_traceroute_defaults.c)
int tcp6_tr_default (SPSData *);  // IPv6 traceroute default settings.
int icmp6_tr_default (SPSData *);  // IPv6 traceroute default settings.
int udp6_tr_default (SPSData *);  // IPv6 traceroute default settings.

// Copy TCP/ICMP/UDP entry data from IPv4/6 to corresponding traceroute.
int copy_tcp4_to_traceroute (SPSData *);  // Copy IPv4 TCP entry data to traceroute
int copy_icmp4_to_traceroute (SPSData *);  // Copy IPv4 ICMP entry data to traceroute
int copy_udp4_to_traceroute (SPSData *);  // Copy IPv4 UDP entry data to traceroute
int copy_tcp6_to_traceroute (SPSData *);  // Copy IPv6 TCP entry data to traceroute
int copy_icmp6_to_traceroute (SPSData *);  // Copy IPv6 ICMP entry data to traceroute
int copy_udp6_to_traceroute (SPSData *);  // Copy IPv6 UDP entry data to traceroute

// RIR IP / ASN delegation FTP functions (ftp.c)
void *ftp_rir (void *);  // FTP function to retrieve RIR delegation files. Type is consistent with GThreadFunc().
int free_rir_mem (char *, char *, char *, SPSData *);  // Free allocated memory used in ftp_rir().
void *download_all (void *);  // Download all RIR delegation files. Type is consistent with GThreadFunc().
int cntl_sock (char *, char *, SPSData *);  // Request a socket descriptor for FTP control socket.
int data_sock (char *, int, SPSData *);  // Request a socket descriptor for FTP data socket.
int cntl_send (int, char *, SPSData *);  // Send a command to FTP control socket.
int cntl_recv (int, SPSData *);  // Receive FTP command responses from remote host on the control socket.
int parse_ip4_port (char *, int *, SPSData *);  // Extract IPv4 address and port number for creating FTP data socket in passive mode.
int parse_ip6_port (int *, SPSData *);  // Extract port number for creating IPv6 FTP data socket in passive mode.
int data_recv (SPSData *);  // Receive and save data on the FTP data socket.
int rir_loading (SPSData *);  // (idle function) Change RIR delegation file status to "Loading".
int rir_not_available (SPSData *);  // (idle function) Change RIR delegation file status to "Not Available" because an error occurred while downloading.
int rir_available (SPSData *);  // (idle function) Update RIR delegation file status.
GtkTextBuffer *rir_textview (int, SPSData *);  // Return pointer to RIR file status textbuffer based on RIR index.

// Functions for reading and searching ASN / IP delegation files and www.iso.org's ISO 3166-1 country code list
// (rir_utils.c)
int all_files_loaded (SPSData *);  // Check to see if all RIR ASN/IP files and country code file are loaded (called prior to any delegation search).
int file_exists (char *);  // Check if a file exists. Return 1 if it exists..
int read_delegation (int, SPSData *);  // Read one Regional Internet Registry ASN / IP delegation file and place into arrays in memory..
int read_countries (SPSData *);  // Read ISO 3166-1 country code list file into an array in memory..
int find_country_code (char *, char *, SPSData *);  // Search www.iso.org's country name/code list for a country name.
int find_country_name (char *, char *, SPSData *);  // Search www.iso.org's country name/code list for a country code.
void *search_asn_by_ccode (void *);  // Search for ASN delegations by country code. Type is consistent with GThreadFunc().
void *rev_search_asn (void *);  // Search delegation files for asn. Return country code/name. Type is consistent with GThreadFunc().
void *search_ipv4_by_ccode (void *);  // Search for IPv4 address delegations by country code. Type is consistent with GThreadFunc().
void *rev_search_ipv4 (void *);  // Search delegation files for IPv4 address. Return country code/name. Type is consistent with GThreadFunc().
void *search_ipv6_by_ccode (void *);  // Search for IPv6 address delegations by country code. Type is consistent with GThreadFunc().
void *rev_search_ipv6 (void *);  // Search delegation files for IPv6 address. Return country code/name. Type is consistent with GThreadFunc().

// Function to update all randomized parameters (update_ran_parms.c).
int update_ran_parms (SPSData *);  // Called only if PRNG seed is changed.

// Miscellaneous functions (misc_utils.c)
int hexascii_listing (uint8_t *, int, char *);  // Prepare a string of text containing a hexadecimal and ASCII listing of data.
int alphanum_and_symbols (uint8_t *, int, char *);  // Copy data to ascii array but only allow alphanumerics and symbols.
int bin8_to_dec (int *);  // Convert 8-bit binary number to decimal.
int is_valid_hexdata (const char *, SPSData *);  // Check string for valid comma-separated hexadecimal values.
int is_valid_hex (char);  // Check for valid hexadecimal character.
int is_valid_decdata (const char *, SPSData *);  // Check string for valid comma-separated decimal values.
int is_valid_dec (char);  // Check for valid decimal. Called from is_valid_decdata().
int is_ascii_uint (const char *);  // Check for non-numeric ASCII characters.
uint32_t ascii_to_uint32 (char *);  // Convert an ASCII representation of a 32-bit integer into an unsigned 32-bit integer.
int64_t ascii_to_int64 (char *);  // Convert an ASCII representation of a 64-bit integer into a 64-bit integer.
void *get_in_addr (struct sockaddr *);  // Get sockaddr (IPv4 or IPv6). - from Brian “Beej Jorgensen” Hall (beej@beej.us)
int carry (int, int, int *, int *);  // Recursive carry function for numerical subtraction.
int parse_ipv4 (char *, uint8_t *);  // Parse character string into 4-byte IPv4 address.
int show_ipv4_range (uint32_t, uint8_t *, char *);  // Show IPv4 address range based on start and number of hosts.
int parse_ipv6 (char *, uint8_t *);  // Parse character string into 16-byte IPv6 address.
int interval_test_ipv4_nhosts (char *, char *, unsigned int);  // Test if an IPv4 address is within a range of IPv4 addresses using number of hosts.
int interval_test_ipv4_slash (char *, char *);  // Test if an IPv4 address is within a range of IPv4 addresses using slash notation.
int interval_test_ipv6 (char *, char *, uint32_t);  // Test if an IPv6 address is within a range of IPv6 addresses.
int skip_type (int, SPSData *);  // Cross-reference between "save type" index and packet type index.
int read_line (FILE *, char *, int);  // Read a single line of text from a text file.
int clear_textview (GtkWidget *);  // Clear a textview.
uint32_t s4_to_u32 (uint8_t *);  // Convert network form of an IPv4 address to a uint32_t value.
int u32_to_s4 (uint32_t, uint8_t *);  // Convert a uint32_t value to network form of an IPv4 address.
int nbits_to_u32_mask (int, uint32_t *);  // Convert number of mask bits to uint32_t mask.
u128_t u128_and (u128_t, u128_t);  // Perform logical AND of two u128_t values.
u128_t u128_or (u128_t, u128_t);  // Perform logical OR of two u128_t values.
u128_t u128_neg (u128_t);  // Perform logical NOT (i.e., invert all bits) on a u128_t value.
int u128_cmp (u128_t, u128_t);  // Compare two u128_t values.
u128_t s6_to_u128 (uint8_t *);  // Convert network form of an IPv6 address to a u128_t value.
int u128_to_s6 (u128_t, uint8_t *);  // Convert a u128_t value to network form of an IPv6 address.
int nbits_to_u128_mask (int, u128_t *);  // Convert number of mask bits to u128_t mask.
int readline (FILE *, char *, int);  // Read a single line of text from a text file.

// UTF-8 functions (utf-8.c)
int is_valid_utf8 (uint8_t *, int, GtkWidget *, SPSData *);  // Examine an array and report any non-UTF-8 characters.
int raise_case (uint8_t *, int, uint8_t *, GtkWidget *, SPSData *);  // Convert any lower case UTF-8 letters to upper case.
int is_small_utf8_letter (uint8_t *, SPSData *);  // Determine if a character is an lower case UTF-8 letter.
int find_val (uint8_t *, int *, SPSData *);  // Find a UTF-8 character's byte sequence in cross-reference array.
uint8_t *utf8_value (uint8_t *, uint8_t *);  // Convert ascii representation of UTF-8 byte sequence to uint8_t byte sequence.
int one_byte (int, int, uint8_t *);  // Determine if input is a 1-byte UTF-8 character.
int two_byte (int, int, uint8_t *);  // Determine if input is a 2-byte UTF-8 character.
int three_byte (int, int, uint8_t *);  // Determine if input is a 3-byte UTF-8 character.
int four_byte (int, int, uint8_t *);  // Determine if input is a 4-byte UTF-8 character.
int five_byte (int, int, uint8_t *);  // Determine if input is a 5-byte UTF-8 character.
int six_byte (int, int, uint8_t *);  // Determine if input is a 6-byte UTF-8 character.
int load_utf8 (SPSData *);  // Load the UTF-8 cross-reference file UTF8_FILENAME into an array.

// Pseudo-random number generation and associated functions.
uint16_t ran16_0to65535 (SPSData *);  // Generate pseudo-random numbers from 0 to 65535, inclusive.
char *ran4_addr (char *, SPSData *);  // Generate a pseudo-random IPv4 address.
char *ran6_addr (char *, SPSData *);  // Generate pseudo-random IPv6 address.
double prng (SPSData *);  // Generate a pseudo-random number.
void prng_init (SPSData *);  // Initialize random number parameters for iterate "iter".
uint64_t RN_skip_ahead (int64_t *, SPSData *);  // Skip ahead n pseudo-random numbers.

// Functions for local host's MAC and IP addresses (this_host.c)
int this_host (SPSData *);  // Find this hostname and IP addresses.
char *thishost_ipv4 (char *);  // Find this host's IPv4 address - only used in arp().
int interface_getmac (char *, uint8_t *);  // Obtain MAC address from interface name.

// Functions to display packet contents (display_packet.c)
int show_packet (int, char **, SPSData *);  // Display packet contents in a pop-up window.

// Perform ARP to obtain MAC address of a node in link-local network (arp.c).
int arp (uint8_t *, const char *, const char *, SPSData *);

// Perform ND to obtain MAC address of a node in link-local network (neighbor_discovery.c).
int neighbor_discovery (uint8_t *, const char *, const char *, SPSData *);

// Warnings and errors functions (errors_and_warnings.c)
int thread_listen (SPSData *);  // (timeout function) Create dialog pop-up for messages/errors from file download or traceroute threads.
int report_warning (SPSData *);  // Report a warning in a pop-up dialog.
int report_error (SPSData *);  // Report an error in a pop-up dialog.
void send_error (void);  // Alert thread_listen() a send error message is available, and reset sending flag.

// Save SPS file, which contains settings selected by user (save_file.c).
int save_file (SPSData *);

// Load SPS file, which contains settings selected by user when saved (load_file.c).
int load_file (SPSData *);

// Textview messaging functions (post_message.c)
Msgdata *packit (Msgdata *, char *);  // Pack Msgdata struct with pointers to a textview and a message string.
int post_message (Msgdata *);  // (idle function) Place message in a textview.

// Checksum functions with respective filenames
uint16_t checksum (uint16_t *, int);  // checksum.c
uint16_t ip4_checksum (struct ip, int, int *, uint8_t **, int);  // ip4_checksum.c
uint16_t tcp4_checksum (struct ip, struct tcphdr, int, int, int *, uint8_t **, int, uint8_t *, int);  // tcp4_checksum.c
uint16_t tcp6_checksum (struct ip6_hdr, struct tcphdr, int, int, int *, uint8_t **, int, uint8_t *, int);  // tcp6_checksum.c
uint16_t tcp6_checksum_routing (struct ip6_hdr, struct tcphdr, int, uint8_t *, int, int, int *, uint8_t **, int, uint8_t *, int); // tcp6_checksum_routing.c
uint16_t icmp4_checksum (struct icmp, uint8_t *, int);  // icmp4_checksum.c
uint16_t icmp6_checksum (struct ip6_hdr, struct icmp6_hdr, uint8_t *, int);  // icmp6_checksum.c
uint16_t icmp6_checksum_routing (struct ip6_hdr, struct icmp6_hdr, int, uint8_t *, uint8_t *, int);  // icmp6_checksum_routing.c
uint16_t udp4_checksum (struct ip, struct udphdr, uint8_t *, int);  // udp4_checksum.c
uint16_t udp6_checksum (struct ip6_hdr, struct udphdr, uint8_t *, int);  // udp6_checksum.c
uint16_t udp6_checksum_routing (struct ip6_hdr, struct udphdr, int, uint8_t *, uint8_t *, int);  // udp6_checksum_routing.c

// Obtain current local host date and time function (date_and_time.c).
char *date_and_time (char *, int);

// Memory allocation functions (allocate.c)
char *allocate_strmem (int);  // Allocate memory for an array chars.
char **allocate_strmemp (int);  // Allocate memory for an array of pointers to arrays of chars.
char *reallocate_strmem (char *, int);  // Reallocate memory for an array of chars. Do not clear contents.
uint8_t *allocate_ustrmem (int);  // Allocate memory for an array of uint8_t.
uint8_t **allocate_ustrmemp (int);  // Allocate memory for an array pointers to arrays of uint8_t.
uint8_t ***allocate_ustrmempp (int);  // Allocate memory for an array pointers to arrays of pointers to arrays of uint8_t.
uint8_t ****allocate_ustrmemppp (int);  // Allocate memory for an array pointers to arrays of pointers to arrays of pointers to arrays of uint8_t.
uint16_t *allocate_u2strmem (int);  // Allocate memory for an array of uint16_t.
int *allocate_intmem (int);  // Allocate memory for an array of ints.
int **allocate_intmemp (int);  // Allocate memory for an array of pointers to arrays of ints.
int ***allocate_intmempp (int);  // Allocate memory for an array of pointers to arrays of pointers to arrays of ints.
uint32_t *allocate_uint32mem (int);  // Allocate memory for array of uint32_t.
int nframes_change (int, int, SPSData *);  // Check if number of frames has changed and re-allocate memory as necessary.
Ethhdr *allocate_ethhdr (int);  // Allocate memory for an array of ethernet headers.
struct ip *allocate_ip4hdr (int);  // Allocate memory for an array of IPv4 headers.
struct ip6_hdr *allocate_ip6hdr (int);  // Allocate memory for an array of IPv6 headers.
Hopdst_hdr *allocate_hopdsthdr (int);  // Allocate memory for an array of hop-by-hop or destination extension headers, excluding options.
Hopdst_hdr **allocate_hopdsthdrp (int);  // Allocate memory for an array of pointers to arrays of hop-by-hop or destination extension headers, excluding options.
Route_hdr *allocate_routehdr (int);  // Allocate memory for an array of routing extension headers.
Route_hdr **allocate_routehdrp (int);  // Allocate memory for an array of pointers to arrays of routing extension headers.
Auth_hdr *allocate_authhdr (int);  // Allocate memory for an array of authentication extension headers.
Esp_hdr *allocate_esphdr (int);  // Allocate memory for an array of encapsulating security payload (ESP) extension headers.
Esp_tail *allocate_esptail (int);  // Allocate memory for an array of encapsulating security payload (ESP) trailers.
struct tcphdr *allocate_tcphdr (int);  // Allocate memory for an array of TCP headers.
struct icmp *allocate_icmp4hdr (int);  // Allocate memory for an array of IPv4 ICMP headers.
struct icmp6_hdr *allocate_icmp6hdr (int);  // Allocate memory for an array of IPv6 ICMP headers.
struct udphdr *allocate_udphdr (int);  // Allocate memory for an array of UDP headers.
Msgdata *allocate_msgdata (int);  // Allocate memory for an array of Msgdata structs.
