/*  Simple Packet Sender (SPS) - a linux packet crafting network
    tool using GTK+ graphical interface.

    Copyright (C) 2011-2023  Hohlraum (h0h1r4um@gmail.com)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "main.h"

// Prepare a string of text containing a hexadecimal and ASCII listing of data.
// Returns length of string_out.
int
hexascii_listing (uint8_t *data, int len, char *string_out) {

  int i, j, c, d, nchunks, remaindr;
  char *ascii;

  // Allocate memory for ASCII array.

  ascii = allocate_strmem (len);

  // Populate ASCII array with only alpha-numerics and symbols. i.e., no control characters
  alphanum_and_symbols (data, len, ascii);

  // Number of full lines nchunks (16 bytes per line), and partial lines remaindr.
  nchunks = (int) (len / 16);
  remaindr = len % 16;

  c = 0;  // Index of data
  d = 0;  // Index of string_out
  i = 0;  // Line number (each line is a 16-byte "chunk")

  // List data, except any partial line.
  for (  ; i<nchunks; i++) {
    sprintf (string_out + d, "%08x :", (i * 16)); d += 10;
    for (j=0; j<8; j++) { sprintf (string_out + d, " %02x", data[j+c]); d += 3; }
    string_out[d] = ' '; d++;
    for (j=8; j<16; j++) { sprintf (string_out + d, " %02x", data[j+c]); d += 3; }
    string_out[d] = ' '; d++;
    for (j=0; j<8; j++) { sprintf (string_out + d, " %c", ascii[j+c]); d += 2; }
    string_out[d] = ' '; d++;
    for (j=8; j<16; j++) { sprintf (string_out + d, " %c", ascii[j+c]); d += 2; }
    string_out[d] = '\n'; d++;
    c += 16;
  }

  // Print address for last line
  sprintf (string_out + d, "%08x : ", (nchunks * 16)); d += 11;

  // Write the final (partial) line of hexadecimals
  if (remaindr <= 8) {
    for (i=0; i<remaindr; i++) { sprintf (string_out + d, "%02x ", data[(nchunks * 16) + i]); d += 3; }
    for (i=0; i<(8-remaindr); i++) { sprintf (string_out + d, "   "); d += 3; }
    string_out[d] = ' '; d++;
    for (i=0; i<8; i++) { sprintf (string_out + d, "   "); d += 3; }
  } else if (remaindr > 8) {
    for (i=0; i<8; i++) { sprintf (string_out + d, "%02x ", data[(nchunks * 16) + i]); d += 3; }
    string_out[d] = ' '; d++;
    for (i=8; i<remaindr; i++) { sprintf (string_out + d, "%02x ", data[(nchunks * 16) + i]); d += 3; }
    for (i=0; i<(16-remaindr); i++) { sprintf (string_out + d, "   "); d += 3; }
  }
  string_out[d] = ' '; d++;

  // Write the final (partial) line of ASCII characters.
  if (remaindr <= 8) {
    for (i=0; i<remaindr; i++) { sprintf (string_out + d, "%c ", ascii[(nchunks * 16) + i]); d += 2; }
  } else if (remaindr > 8) {
    for (i=0; i<8; i++) { sprintf (string_out + d, "%c ", ascii[(nchunks * 16) + i]); d += 2; }
    string_out[d] = ' '; d++;
    for (i=8; i<remaindr; i++) { sprintf (string_out + d, "%c ", ascii[(nchunks * 16) + i]); d += 2; }
  }

  // Free allocated memory.
  free (ascii);

  return (EXIT_SUCCESS);
}

// Copy data to ascii array but only allow alphanumerics and symbols.
int
alphanum_and_symbols (uint8_t *data, int len, char *ascii) {

  int i;

  // Replace any characters outside of range ' ' to '~' into ' '.
  // 32 = ' ', 126 = '~'
  for (i=0; i<len; i++) {
    if ((data[i] < 32) || (data[i] > 126)) {
      ascii[i] = 32;
    } else {
      ascii[i] = data[i];
    }
  }

  return (EXIT_SUCCESS);
}

// Function to convert 8 bit binary to decimal
int
bin8_to_dec (int *bin) {

  int i;
  int dec;

  dec = 0;
  for (i=0; i<8; i++) {
    dec += (bin[i] << i);
  }

  return (dec);
}

// Check string for valid comma-separated hexadecimal values.
int
is_valid_hexdata (const char *string, SPSData *data) {

  int i, j, len, result;
  char temp[TMP_STRINGLEN];
  char **endptr;

  len = strnlen (string, IP_MAXPACKET);

  // Check for a character which is neither hexadecimal, a comma, nor a space.
  for (i=0; i<len; i++) {
    if (!(is_valid_hex (string[i])) && (string[i] != ',') && (string[i] != ' ')) {
      return (0);  // Character is neither hexadecimal, a comma, nor a space.
    }
  }

  // Check for data composed of only spaces (bad).
  j = 0;  // If j = 1 it means we found a non-space character (good).
  for (i=0; i<len; i++) {
    if (string[i] != ' ') {
      j = 1;
      break;
    }
  }
  if (!j) {
    return (0);  // Nothing but spaces were found in string.
  }

  // Check if first character is a comma (which is bad).
  i = 0;
  while (i < len) {
    if (string[i] == ',') {
      return (0);  // First character is a comma.
    } else if (string[i] == ' ') {
      i++;
      continue;  // This character is a space so ignore it and keep looping.
    } else {
      break;  // First character must be valid hexadecimal, so we're done.
    }
  }

  // Check if last character is a comma (which is bad).
  i = len - 1;
  while (i >= 0) {
    if (string[i] == ',') {
      return (0);  // First character is a comma.
    } else if (string[i] == ' ') {
      i--;
      continue;  // This character is a space so ignore it and keep looping.
    } else {
      break;  // Last character must be valid hexadecimal, so we're done.
    }
  }

  // Check for adjacent commas (which is bad.).
  for (i=0; i<(len-1); i++) {
    if ((string[i] == ',') && (string[i+1] == ',')) {
      return (0);  // Found two adjacent commas.
    }
  }

  // We need to limit values to 0xff.
  i = 0;
  while (i < len) {

    // Grab next hexadecimal value.
    memset (temp, 0, TMP_STRINGLEN * sizeof (char));
    j = 0;
    while ((string[i+j] != ' ') && (string[i+j] != ',') && ((i + j) < len)) {
      temp[j] = string[i+j];
      j++;
    }
    i += j;

    // Clobber leading zeros.
    for (j=0; j<(int) strnlen (temp, TMP_STRINGLEN); j++) {
      if (temp[j] == '0') {
        temp[j] = ' ';  // replace zero with space
      } else if ((temp[j] != ' ') && (temp[j] != '0')) {
        break;  // valid digit was found; we're done looking for leading zeros.
      }
    }

    // Convert to base-10 chars and store as data.
    endptr = NULL;
    result = strtol (temp, endptr, 16);
    if (endptr != NULL) {
      sprintf (data->error_text, "is_valid_hex_data(): strtol() failed.");
      data->parent = data->main_window;
      report_error (data);
      return (0);  // strtol() failed.
    }
    if ((result < 0) || (result > 0xff)) {
      return (0);  // Value is less than 0 or greater than 0xff.
    }

    // Increment i until we find more hexadecimal values or hit end of string.
    while (((string[i] == ' ') || (string[i] == ',')) && (i < len)) {
      i++;
    }
  }

  return (1);  //  String appears to be valid comma-separated hexadecimal data.
}

// Check for valid hexadecimal character.
int
is_valid_hex (char byte) {

  // Within ranges 0 to 9, a to f, or A to F is ok.
  if (!(((byte > 64) && (byte < 71)) || ((byte > 96) && (byte < 103)) ||
    ((byte > 47) && (byte < 58)))) {
    return (0);  // Found invalid hexadecimal character.
  }

  return (1);  // Appears to be valid hexadecimal character.
}

// Check string for valid comma-separated decimal values.
int
is_valid_decdata (const char *string, SPSData *data) {

  int i, j, len, result;
  char temp[TMP_STRINGLEN];
  char **endptr;

  len = strnlen (string, IP_MAXPACKET);

  // Check for a character which is neither decimal, a comma, nor a space.
  for (i=0; i<len; i++) {
    if (!(is_valid_dec (string[i])) && (string[i] != ',') && (string[i] != ' ')) {
      return (0);  // Character is neither decimal, a comma, nor a space.
    }
  }

  // Check for data composed of only spaces (bad).
  j = 0;  // If j = 1 it means we found a non-space character (good).
  for (i=0; i<len; i++) {
    if (string[i] != ' ') {
      j = 1;
      break;
    }
  }
  if (!j) {
    return (0);  // Nothing but spaces were found in string.
  }

  // Check if first character is a comma (which is bad).
  i = 0;
  while (i < len) {
    if (string[i] == ',') {
      return (0);  // First character is a comma.
    } else if (string[i] == ' ') {
      i++;
      continue;  // This character is a space so ignore it and keep looping.
    } else {
      break;  // First character must be valid hexadecimal, so we're done.
    }
  }

  // Check if last character is a comma (which is bad).
  i = len - 1;
  while (i >= 0) {
    if (string[i] == ',') {
      return (0);  // First character is a comma.
    } else if (string[i] == ' ') {
      i--;
      continue;  // This character is a space so ignore it and keep looping.
    } else {
      break;  // Last character must be valid hexadecimal, so we're done.
    }
  }

  // Check for adjacent commas (which is bad.).
  for (i=0; i<(len-1); i++) {
    if ((string[i] == ',') && (string[i+1] == ',')) {
      return (0);  // Found two adjacent commas.
    }
  }

  // We need to limit values to 255.
  i = 0;
  while (i < len) {

    // Grab next decimal value.
    memset (temp, 0, TMP_STRINGLEN * sizeof (char));
    j = 0;
    while ((string[i+j] != ' ') && (string[i+j] != ',') && ((i + j) < len)) {
      temp[j] = string[i+j];
      j++;
    }
    i += j;

    // Clobber leading zeros.
    for (j=0; j<(int) strnlen (temp, TMP_STRINGLEN); j++) {
      if (temp[j] == '0') {
        temp[j] = ' ';  // replace zero with space
      } else if ((temp[j] != ' ') && (temp[j] != '0')) {
        break;  // valid digit was found; we're done looking for leading zeros.
      }
    }

    // Store as data.
    endptr = NULL;
    result = strtol (temp, endptr, 10);
    if (endptr != NULL) {
      sprintf (data->error_text, "is_valid_dec_data(): strtol() failed.");
      data->parent = data->main_window;
      report_error (data);
      return (0);  // strtol() failed.
    }
    if ((result < 0) || (result > 255)) {
      return (0);  // Value is less than 0 or greater than 255.
    }

    // Increment i until we find more decimal values or hit end of string.
    while (((string[i] == ' ') || (string[i] == ',')) && (i < len)) {
      i++;
    }
  }

  return (1);
}

// Check for valid decimal digit.
int
is_valid_dec (char byte) {

  // '0' = 48, '9' = 57
  if (!((byte > 47) && (byte < 58))) {
    return (0);  // Character is an invalid decimal digit.
  }

  return (1);  // Character is a valid decimal digit.
}

// Check for ASCII representation of unsigned integer numbers.
int
is_ascii_uint (const char *string) {

  int i;

  // Only 0 to 9 are accepted. Negative signs and decimals will return failure.
  // '0' = 48, '9' = 57
  for (i=0; i<(int) strnlen (string, IP_MAXPACKET); i++) {
    if ((string[i] < 48) || (string[i] > 57)) {
      return (0);  // Found invalid ASCII representation of unsigned integer number.
    }
  }

  return (1);  // Appears to be valid ASCII representation of unsigned integer numbers.
}

// Convert an ASCII representation of a 32-bit integer into an unsigned 32-bit integer.
uint32_t
ascii_to_uint32 (char *string) {

  uint32_t result;
  char **endptr;

  endptr = NULL;
  errno = 0;
  result = (uint32_t) strtoll (string, endptr, 10);
  if ((errno != 0) || (endptr != NULL)) {
    return (-1);  // Error condition
  }

  return (result);
}

// Convert an ASCII representation of a 64-bit integer into a 64-bit integer.
int64_t
ascii_to_int64 (char *string) {

  int64_t result;
  char **endptr;

  endptr = NULL;
  errno = 0;
  result = (int64_t) strtoll (string, endptr, 10);
  if ((errno != 0) || (endptr != NULL)) {
    return (-1);  // Error condition
  }

  return (result);
}

// Get sockaddr (IPv4 or IPv6).
// This excellent bit of code is taken from
// Brian “Beej Jorgensen” Hall (beej@beej.us).
void *
get_in_addr (struct sockaddr *sa) {

  if (sa->sa_family == AF_INET) {
    return &(((struct sockaddr_in*)sa)->sin_addr);
  }

  return &(((struct sockaddr_in6*)sa)->sin6_addr);
}

// Recursive carry for subtracting arbitrarily long numbers.
int
carry (int n, int i, int *t, int *s) {

  // For IPv4 (4 bytes) n = 4. For IPv6 (16 bytes) n = 16.
  if (i == n) return 0;  // Can't carry from any higher order bytes.

  if (t[i+1] != 0) {
    t[i+1]--;
  } else {
    t[i+1] = 255;
    if (carry (n, i+1, t, s) == 0) return 0;  // Can't carry from any higher order bytes.
  }

  return (1);
}

// Parse character string into 4-byte IPv4 address in network s4 form.
int
parse_ipv4 (char *ip4string, uint8_t *ip4n) {

  if (inet_pton (AF_INET, ip4string, ip4n) != 1) {
    fprintf (stderr, "ERROR: Invalid IPv4 address in parse_ipv4().\n");
    exit (EXIT_FAILURE);
  }

  return (EXIT_SUCCESS);
}

// For IPv4, address range based upon start IP address
// and number of hosts given in RIR file.
int
show_ipv4_range (uint32_t nhosts, uint8_t *ip4n, char *answer) {

  int i;
  uint8_t endn[4];
  uint32_t start, end;
  char *temp;

  // Allocate memory for temporary string.
  // The string answer was allocated as size TEXT_STRINGLEN in another function.
  temp = allocate_strmem (TEXT_STRINGLEN);

  memset (answer, 0, TEXT_STRINGLEN * sizeof (char));
  for (i=0; i<3; i++) {
    memset (temp, 0, TEXT_STRINGLEN * sizeof (char));
    sprintf (temp, "%i.", ip4n[i]);  // Starting address
    strncat (answer, temp, TEXT_STRINGLEN - strnlen (answer, TEXT_STRINGLEN - 1) - 1);
  }
  memset (temp, 0, TEXT_STRINGLEN * sizeof (char));
  sprintf (temp, "%i", ip4n[3]);
  strncat (answer, temp, TEXT_STRINGLEN - strnlen (answer, TEXT_STRINGLEN - 1) - 1);

  // Convert starting address from network s4 form to uint32_t.
  start = s4_to_u32 (ip4n);

  // For IPv4, RIR files use number of hosts, starting from start.
  // i.e., Not like IPv6, for which number of bits to mask (slah notation) is used..
  end = start + nhosts;

  // Convert ending address from uint32_t to network s4 format.
  u32_to_s4 (end, endn);

  memset (temp, 0, TEXT_STRINGLEN * sizeof (char));
  sprintf (temp, " to %i.%i.%i.%i", endn[0], endn[1], endn[2], endn[3]);
  strncat (answer, temp, TEXT_STRINGLEN - strnlen (answer, TEXT_STRINGLEN - 1) - 1);

  return (EXIT_SUCCESS);
}

// Parse character string into 16-byte IPv6 address in network s6 form.
int
parse_ipv6 (char *ip6string, uint8_t *ip6n) {

  if (inet_pton (AF_INET6, ip6string, ip6n) != 1) {
    fprintf (stderr, "ERROR: Invalid IPv6 internet address in parse_ipv6().\n");
    exit (EXIT_FAILURE);
  }

  return (EXIT_SUCCESS);
}

// Test if an IPv4 address is within a range of IPv4 addresses.
// Uses number of hosts for range calculation.
// Return 1 if within range, 0 if outside of range.
int
interval_test_ipv4_nhosts (char *teststring, char *startstring, uint32_t nhosts) {

  int inrange;
  uint8_t testn[4], startn[4];
  uint32_t test, start, end;

  // Convert IPv4 test address to network format.
  parse_ipv4 (teststring, testn);

  // Convert test from network s4 form to uint32_t.
  test = s4_to_u32 (testn);

  // Convert IPv4 start address string to network s4 format.
  parse_ipv4 (startstring, startn);

  // Convert start from network s4 form to uint32_t.
  start = s4_to_u32 (startn);

  // For IPv4, val is number of hosts, starting from start.
  // i.e., Not like IPv6, for which val is number of bits to mask.
  end = start + nhosts;

  // Check if test is within range of start and end.

  // Initially assume test address is within range.
  inrange = 1;

  // If test is below start or above end, then it's outside range.
  if ((test < start) || (test > end)) {
    inrange = 0;
  }

  return (inrange);
}

// Test if an IPv4 address is within a range of IPv4 addresses.
// Uses number of slash notation for range calculation.
// Return 1 if within range, 0 if outside of range.
int
interval_test_ipv4_slash (char *teststring, char *cidrs) {

  int i, nbits, inrange;
  char temp[TEXT_STRINGLEN];
  uint8_t cidrn[4], startn[4], testn[4];
  uint32_t cidr, mask, notmask, start, end, test;

  // Initialize values to zero.
  memset (temp, 0, TEXT_STRINGLEN * sizeof (char));
  memset (cidrn, 0, 4 * sizeof (uint8_t));
  memset (startn, 0, 4 * sizeof (uint8_t));
  memset (testn, 0, 4 * sizeof (uint8_t));
  mask = 0;
  start = 0;
  notmask = 0;
  test = 0;

  // Copy CIDR without slash and number of prefix mask bits.
  i = 0;
  while ((i < (int) strnlen (cidrs, TEXT_STRINGLEN)) && (cidrs[i] != '/')) i++;
  memcpy (temp, cidrs, i * sizeof (char));

  // Extract number of prefix mask bits from CIDR.
  i++;  // Skip slash
  if (i >= (int) strnlen (cidrs, TEXT_STRINGLEN)) {
    nbits = 32;  // There is no slash, therefore mask is 32 bits.
  } else {
    nbits = ascii_to_uint32 (&cidrs[i]);
  }

  // Convert CIDR address from string to network s4 format.
  if (inet_pton (AF_INET, temp, cidrn) != 1) {
    fprintf (stderr, "ERROR: Invalid IPv4 internet address in interval_test_ipv4_slash().\n");
    exit (EXIT_FAILURE);
  }

  // Convert number of mask bits to a u32_t mask.
  nbits_to_u32_mask (nbits, &mask);

  // Convert cidr from network s4 form to uint32_t.
  cidr = s4_to_u32 (cidrn);

  // Starting address SHOULD be just the cidr,
  // but we'll trust the mask bits.
  // Perform logical AND of cidr and mask to get starting address.
  start = cidr & mask;

  // Convert start from uint32_t to network s4 format.
  u32_to_s4 (start, startn);

  // Create an inverted mask.
  notmask = ~mask;

  // Calculate ending address by setting all non-masked bits of cidr address to 1.
  // Perform logical OR of cidr with notmask.
  end = cidr | notmask;

  // Convert IPv4 test address to network format.
  parse_ipv4 (teststring, testn);

  // Convert test from network s4 form to uint32_t.
  test = s4_to_u32 (testn);

  // Check if test is within range of start and end.

  // Initially assume test address is within range.
  inrange = 1;

  // If test is below start or above end, then it's outside range.
  if ((test < start) || (test > end)) {
    inrange = 0;
  }

  return (inrange);
}

// Test if an IPv6 address is within a range of IPv6 addresses.
// nbits is number of mask bits specified in CIDR slash notation.
// Return 1 if within range, 0 if outside of range.
int
interval_test_ipv6 (char *teststring, char *cidrstring, uint32_t nbits) {

  int inrange;
  uint8_t cidrn[16], startn[16], endn[16], testn[16];
  u128_t cidr, mask, start, end, notmask, test;

  // Initialize values to zero.
  memset (cidrn, 0, 16 * sizeof (uint8_t));
  memset (startn, 0, 16 * sizeof (uint8_t));
  memset (endn, 0, 16 * sizeof (uint8_t));
  memset (testn, 0, 16 * sizeof (uint8_t));
  mask.l = 0;
  mask.h = 0;
  start.l = 0;
  start.h = 0;
  notmask.l = 0;
  notmask.h = 0;
  test.l = 0;
  test.h = 0;

  // Convert IPv6 test address string to network s6 format.
  parse_ipv6 (teststring, testn);

  // Convert test from network s6 form to u128_t.
  test = s6_to_u128 (testn);

  // Convert IPv6 CIDR address string to network s6 format.
  parse_ipv6 (cidrstring, cidrn);

  // Convert cidr from network s6 form to u128_t.
  cidr = s6_to_u128 (cidrn);

  // Convert number of mask bits to a u128_t mask.
  nbits_to_u128_mask (nbits, &mask);

  // Starting address SHOULD be just the cidr but
  // but might not be, so we'll trust the mask bits.
  // Perform logical AND of cidr and mask to get starting address.
  start = u128_and (cidr, mask);

  // Create an inverted mask.
  notmask = u128_neg (mask);

  // Calculate ending address by setting all non-masked bits of cidr address to 1.
  // Perform logical OR of cidr with notmask.
  end.l = cidr.l;
  end.h = cidr.h;
  end = u128_or (cidr, notmask);

  // Check if test is in range of cidr. i.e., between start and end

  // Initially assume test address is within range.
  inrange = 1;

  // If test is below start or above end, it's outside range.
  if ((u128_cmp (test, start) == -1) || (u128_cmp (test, end) == 1)) {
    inrange = 0;
  }

  return (inrange);
}

// Cross-reference between "save type" index and packet type index.
// Used by load_file() and save_file().
int
skip_type (int i, SPSData *data) {

  int ip6_type[6]={3,4,5,12,13,14};

  // If not selected by user, then skip it by returning 0.
  if ((!data->save_flags[5] && (ip6_type[i] == 3)) ||
      (!data->save_flags[6] && (ip6_type[i] == 4)) ||
      (!data->save_flags[7] && (ip6_type[i] == 5)) ||
      (!data->save_flags[13] && (ip6_type[i] == 12)) ||
      (!data->save_flags[14] && (ip6_type[i] == 13)) ||
      (!data->save_flags[15] && (ip6_type[i] == 14))) {
    return (0);  // Packet type not selected by user; skip it.
  }

  return (1);  // Don't skip this packet type.
}

// Read a single line of text from a text file.
// Returns -1 when EOF is encountered.
int
read_line (FILE *fi, char *line, int limit) {

  int i, n;

  i = 0;  // i is pointer to byte in line.
  for (;;) {

    // Check for line index out of bounds.
    if (i >= limit) {
      fprintf (stderr, "ERROR: Line exceeds maximum line length of %i.\n", limit);
      line[limit - 1] = 0;  // Force string termination.
      fprintf (stderr, "       Line contains: %s\n", line);
      exit (EXIT_FAILURE);
    }

    // Grab next byte from file.
    n = fgetc (fi);

    // End of file reached.
    // Tell calling function, by returning -1, that we're at end of file, so it won't call read_line() again.
    if (n == EOF) {
      return (-1);
    }

    // Found a carriage return. Ignore it.
    if (n == '\r') {
      continue;
    }

    // Found a newline. Change to 0 for string termination.
    // Break out of loop since this is the end of the current line.
    if (n == '\n') {
      line[i] = 0;  // Replace with 0 for string termination.
      break;
    }

    // Seems to be a valid character. Keep it.
    line[i] = n;
    i++;
  }

  return (0);  // Finished line, but not at end of file.
}

// Pack Msgdata struct with pointers to a textview and a message string.
Msgdata *
packit (Msgdata *msgdata, char *message) {

  msgdata->message = message;

  return (msgdata);
}

// Clear textview.
int
clear_textview (GtkWidget *textview) {

  GtkTextBuffer *textbuffer;

  // Clear any text in the textview.
  textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (textview));  // IPv4
  gtk_text_buffer_set_text (textbuffer, "", -1);

  return (EXIT_SUCCESS);
}

// Convert network form of an IPv4 address to a uint32_t value.
uint32_t s4_to_u32 (uint8_t *s4) {

  uint32_t rv;

  rv = (((uint32_t) s4[0]) << 24) |
       (((uint32_t) s4[1]) << 16) |
       (((uint32_t) s4[2]) <<  8) |
       (((uint32_t) s4[3]) <<  0);

  return rv;
}

// Convert a uint32_t value to network form of an IPv4 address.
int
u32_to_s4 (uint32_t v, uint8_t *s4) {

  s4[0] = 0xff & (v >> 24);
  s4[1] = 0xff & (v >> 16);
  s4[2] = 0xff & (v >>  8);
  s4[3] = 0xff & (v >>  0);

  return (EXIT_SUCCESS);
}

// Convert number of mask bits to uint32_t mask.
int
nbits_to_u32_mask (int nbits, uint32_t *mask) {

  int i;

  // Initialize mask to zero.
  *(mask) = 0;

  i = 31;
  while (nbits > 0) {
    *(mask) |= (UINT32_C(1) << i);
    nbits--;
    i--;
  }

  return (EXIT_SUCCESS);
}

// Perform logical AND of two u128_t values.
u128_t
u128_and (u128_t x, u128_t y) {

  u128_t rv;

  rv.h = x.h & y.h;
  rv.l = x.l & y.l;

  return rv;
}

// Perform logical OR of two u128_t values.
u128_t
u128_or (u128_t x, u128_t y) {

  u128_t rv;

  rv.h = x.h | y.h;
  rv.l = x.l | y.l;

  return rv;
}

// Perform logical NOT (i.e., invert all bits) on u128_t.
u128_t
u128_neg (u128_t v) {

  u128_t rv;

  rv.h = ~v.h;
  rv.l = ~v.l;

  return rv;
}

// Compare two u128_t values. Return:
// -1 if x < y
//  0 if x = y
//  1 if x > y
int
u128_cmp (u128_t x, u128_t y) {

  if (x.h < y.h)
    return -1;

  if (x.h > y.h)
    return 1;

  if (x.l < y.l)
    return -1;

  if (x.l > y.l)
    return 1;

  return 0;
}

// Convert network form of an IPv6 address to a u128_t value.
u128_t
s6_to_u128 (uint8_t *s6) {

  u128_t rv;

  rv.h = (((uint64_t) s6[0])  << 56) |
         (((uint64_t) s6[1])  << 48) |
         (((uint64_t) s6[2])  << 40) |
         (((uint64_t) s6[3])  << 32) |
         (((uint64_t) s6[4])  << 24) |
         (((uint64_t) s6[5])  << 16) |
         (((uint64_t) s6[6])  <<  8) |
         (((uint64_t) s6[7])  <<  0);

  rv.l = (((uint64_t) s6[8])  << 56) |
         (((uint64_t) s6[9])  << 48) |
         (((uint64_t) s6[10]) << 40) |
         (((uint64_t) s6[11]) << 32) |
         (((uint64_t) s6[12]) << 24) |
         (((uint64_t) s6[13]) << 16) |
         (((uint64_t) s6[14]) <<  8) |
         (((uint64_t) s6[15]) <<  0);

  return rv;
}

// Convert a u128_t value to network form of an IPv6 address.
int
u128_to_s6 (u128_t v, uint8_t *s6) {

  s6[0]  = 0xff & (v.h >> 56);
  s6[1]  = 0xff & (v.h >> 48);
  s6[2]  = 0xff & (v.h >> 40);
  s6[3]  = 0xff & (v.h >> 32);
  s6[4]  = 0xff & (v.h >> 24);
  s6[5]  = 0xff & (v.h >> 16);
  s6[6]  = 0xff & (v.h >>  8);
  s6[7]  = 0xff & (v.h >>  0);

  s6[8]  = 0xff & (v.l >> 56);
  s6[9]  = 0xff & (v.l >> 48);
  s6[10] = 0xff & (v.l >> 40);
  s6[11] = 0xff & (v.l >> 32);
  s6[12] = 0xff & (v.l >> 24);
  s6[13] = 0xff & (v.l >> 16);
  s6[14] = 0xff & (v.l >>  8);
  s6[15] = 0xff & (v.l >>  0);

  return (EXIT_SUCCESS);
}

// Convert number of mask bits to u128_t mask.
int
nbits_to_u128_mask (int nbits, u128_t *mask) {

  int i, lbits;

  lbits = 0;  // Start by assuming no mask bits are set for lower 64-bits.

  // Initialize mask to zero.
  mask->l = 0;
  mask->h = 0;

  // If number of mask bits is > 64, then all high bits are 1.
  if (nbits > 64)  {

    // Set all upper 64 bits to 1.
    for (i=63; i>=0; i--) {
      mask->h |= (UINT64_C(1) << i);
    }

    // Lower 64 bits of mask.
    lbits = nbits - 64;  // Mask bits for lower 64-bit part.
    i = 63;
    while (lbits > 0) {
      mask->l |= (UINT64_C(1) << i);
      lbits--;
      i--;
    }

  // Number of mask bits is <= 64.
  } else {
    i = 63;
    while (nbits > 0) {
      mask->h |= (UINT64_C(1) << i);
      nbits--;
      i--;
    }
  }

  return (EXIT_SUCCESS);
}

// Read a single line of text from a text file.
// Returns -1 if EOF is encountered.
int
readline (FILE *fi, char *line, int limit)
{
  int i, n;

  i = 0;  // i is pointer to byte in line.
  for (i=0; i<limit; i++) {

    // Grab next byte from file.
    n = fgetc (fi);

    // End of file reached.
    // Tell calling function, by returning -1, that we're at end of file, so it won't call read_line() again.
    if (n == EOF) {
      return (-1);
    }

    // Found a carriage return. Ignore it.
    if (n == '\r') {
      continue;
    }

    // Found a tab. Replace it with space.
    if (n == '\t') {
      n = ' ';
    }

    // Found a newline. Change to 0 for string termination.
    // Break out of loop since this is the end of the current line.
    if (n == '\n') {
      line[i] = 0;  // Replace with 0 for string termination.
      return (0);
    }

    // Seems to be a valid character. Keep it.
    line[i] = n;
  }

  // Advance to next line.
  n = 0;
  while ((n != '\n') && (n != EOF)) {
    n = fgetc (fi);
  }

  return (0);
}

// Display IPv4 header options, hop-by-hop options, or destination options window.
int
displ_options_window (int packet_type, char *title, char *string) {

  char *temp;
  GtkWidget *window, *scrolled_win, *textview;
  GtkTextBuffer *buffer;
  GtkCssProvider *provider;
  GtkStyleContext *context;

  // Allocate memory for various arrays.
  temp = allocate_strmem (TEXT_STRINGLEN);

  // Create new window to display options.
  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_resizable (GTK_WINDOW (window), TRUE);
  gtk_window_set_position (GTK_WINDOW (window), GTK_WIN_POS_NONE);
  memset (temp, 0, TEXT_STRINGLEN * sizeof (char));
  switch (packet_type) {
    case 0:
      sprintf (temp, "Options for IPv4 TCP Packet");
      break;
    case 1:
      sprintf (temp, "Options for IPv4 ICMP Packet");
      break;
    case 2:
      sprintf (temp, "Options for IPv4 UDP Packet");
      break;
    case 3:
      sprintf (temp, "Options for IPv6 TCP Packet");
      break;
    case 4:
      sprintf (temp, "Options for IPv6 ICMP Packet");
      break;
    case 5:
      sprintf (temp, "Options for IPv6 UDP Packet");
      break;
    case 6:
      sprintf (temp, "Options for IPv4 header of 6to4 TCP Packet");
      break;
    case 7:
      sprintf (temp, "Options for IPv4 header of 6to4 ICMP Packet");
      break;
    case 8:
      sprintf (temp, "Options for IPv4 header of 6to4 UDP Packet");
      break;
    case 9:
      sprintf (temp, "Options for IPv4 TCP Packet for Traceroute");
      break;
    case 10:
      sprintf (temp, "Options for IPv4 ICMP Packet for Traceroute");
      break;
    case 11:
      sprintf (temp, "Options for IPv4 UDP Packet for Traceroute");
      break;
    case 12:
      sprintf (temp, "Options for IPv6 TCP Packet for Traceroute");
      break;
    case 13:
      sprintf (temp, "Options for IPv6 ICMP Packet for Traceroute");
      break;
    case 14:
      sprintf (temp, "Options for IPv6 UDP Packet for Traceroute");
      break;
    default:
      fprintf (stderr, "ERROR: Invalid upper layer protocol type %i in dspl_options_window().\n", packet_type);
      free (temp);
      exit (EXIT_FAILURE);
  }  // End switch

  strncat (title, temp, TEXT_STRINGLEN - strnlen (title, TEXT_STRINGLEN - 1) - 1);
  gtk_window_set_title (GTK_WINDOW (window), title);
  gtk_window_set_default_size (GTK_WINDOW (window), 1000, 200);
  gtk_container_set_border_width (GTK_CONTAINER (window), 1);

  // Create a textview object to display options.
  textview = gtk_text_view_new ();
  buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (textview));
  gtk_text_view_set_cursor_visible (GTK_TEXT_VIEW (textview), TRUE);
  provider = gtk_css_provider_new ();
  gtk_css_provider_load_from_path (provider, "./packet.css", NULL);
  context = gtk_widget_get_style_context (textview);
  gtk_style_context_add_provider (context, GTK_STYLE_PROVIDER (provider), GTK_STYLE_PROVIDER_PRIORITY_APPLICATION);

  // Put the text into the buffer.
  gtk_text_buffer_set_text (buffer, string, -1);

  // Add scrolled window and textview to new window and display options.
  scrolled_win = gtk_scrolled_window_new (NULL, NULL);
  gtk_container_add (GTK_CONTAINER (scrolled_win), textview);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_win), GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);

  gtk_container_add (GTK_CONTAINER (window), scrolled_win);
  gtk_widget_show_all (window);

  // Free allocated memory.
  free (temp);

  return (EXIT_SUCCESS);
}
