/*  Simple Packet Sender (SPS) - a linux packet crafting network
    tool using GTK+ graphical interface.

    Copyright (C) 2011-2023  Hohlraum (h0h1r4um@gmail.com)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "main.h"

// Obtain interface current MTU setting.
int
interface_getmtu (char *interface) {

  int sd;
  struct ifreq ifr;

  // Submit request for a socket descriptor to use ioctl().
  if ((sd = socket (PF_PACKET, SOCK_RAW, htons (ETH_P_ALL))) < 0) {
    return (0);  // Failed to obtain socket decriptor.
  }

  memset (&ifr, 0, sizeof (ifr));
  strncpy (ifr.ifr_name, interface, IFNAMSIZ - 1);  // 16 - 1 for string termination.

  // Get MTU using SIOCGIFMTU.
  if (ioctl (sd, SIOCGIFMTU, &ifr) < 0) {
    return (0);  // Failed to obtain interface's current MTU.
  }

  // Close file descriptor.
  close (sd);

  return (ifr.ifr_mtu);
}

// Find minimum MTU for IPv4 packet
// when packet is fragmented to maximum extent.
int
min4_mtu (int type, SPSData *data) {

  int unfraglen, fraglen, min;

  // Find length of unfragmentable portion.
  unfraglen = find4_unfraglen (type, data);

  // Find length of fragmentable portion.
  fraglen = find4_fraglen (type, data);

  // Is fragmentation required?

  // Yes
  if ((data->ifmtu[type] - unfraglen - fraglen) < 0) {
    // Minimum fragment is 8 bytes. See Section 2.3 of RFC 791.
    min = unfraglen + 8;

  // No
  } else {
    min = unfraglen + fraglen;
  }

  return (min);
}

// Find minimum MTU for IPv6 packet
// when packet is fragmented to maximum extent.
int
min6_mtu (int type, SPSData *data) {

  int unfraglen, fraglen, min;

  // Find length of unfragmentable portion.
  unfraglen = find6_unfraglen (type, data);

  // Find length of fragmentable portion.
  fraglen = find6_fraglen (type, data);

  // Is fragmentation required?

  // Yes
  if ((data->ifmtu[type] - unfraglen - fraglen) < 0) {
    // Minimum fragment is 8 bytes (<-- this is NOT the fragmentation header). See Section 4.5 of RFC 8200.
    // Fragmentation header will also be needed.
    min = unfraglen + FRG_HDRLEN + 8;

  // No
  } else {
    min = unfraglen + fraglen;  // Don't need to fragment, so no fragmentation header.
  }

  return (min);
}

// Find minimum MTU for 6to4 packet
// when packet is fragmented to maximum extent.
int
min6to4_mtu (int type, SPSData *data) {

  int unfraglen, fraglen, min;

  // Find length of unfragmentable portion.
  unfraglen = find6_unfraglen (type-3, data);

  // Find length of fragmentable portion.
  fraglen = find6_fraglen (type-3, data);

  // Is fragmentation required?

  // Yes, fragmentation is required.
  if ((data->ifmtu[type] - IP4_HDRLEN - unfraglen - fraglen) < 0) {
    // Minimum fragment is 8 bytes (<-- this is NOT the fragmentation header). See Section 4.5 of RFC 8200.
    min = IP4_HDRLEN + unfraglen + FRG_HDRLEN + 8;  // Fragmentation header will be needed.

  // No, fragmentation is not required.
  } else {
    min = IP4_HDRLEN + unfraglen + fraglen;  // Don't need to fragment, so no fragmentation header.
  }

  return (min);
}

// Perform MTU checks for 6to4. Revise if necessary.
// This function is called by on_button1_clicked() (i.e., Send).
// This is necessary because checks are not performed for 6to4 when IPv6 packets
// are changed. When IPv6 packets are modified, the user may not intend
// to use 6to4, in which case the 6to4 IPv4 interface will not be specified, and
// without the 6to4 interface specified, the 6to4 MTU checks cannot be completed.
int
sixto4_mtu_checks (int type, SPSData *data) {

  int ifmtu, min;

  // Not traceroute
  if ((type < 9) || (type == 101)) {
    data->parent = data->main_window;
  } else {
    data->parent = data->traceroute_window;
  }

  // Note that type will be 3, 4, or 5 with data->sixto4_flag set.
  // This function is only called by on_button1_clicked (i.e., Send).

  // Obtain interface's MTU.
  // If we fail to obtain MTU from interface, don't change current MTU.
  if ((ifmtu = interface_getmtu (data->ifname[type+3])) == 0) {
    sprintf (data->warning_text, "Failed to obtain MTU from interface for 6to4.\nWas interface specified?");
    report_warning (data);
    return (EXIT_FAILURE);
  }

  // Determine minimum size needed to accommodate packet,
  // when packet is fragmented to maximum extent.
  min = min6to4_mtu (type+3, data);

  // Is interface's MTU is too small to accommodate packet,
  // even when packet is fragmented to maximum extent?
  if (ifmtu < min) {
    sprintf (data->warning_text, "The interface's MTU is too small to accommodate 6to4 packet.");
    report_warning (data);
    return (EXIT_FAILURE);
  }

  // Is current MTU value too small to accommodate IPv6 packet,
  // even when packet is fragmented to maximum extent?
  if (data->ifmtu[type+3] < min) {
    sprintf (data->warning_text, "The current MTU value is too small to accommodate 6to4 packet.\nPacket needs a minimum of %i bytes.\nValue revised to 1280 bytes (RFC 4213).", min);
    report_warning (data);
    data->ifmtu[type+3] = 1280;  // RFC 4213

    // Update MTU spinbutton.
    update_mtu_spinbutton (type+3, data);
  }

  // Update ethernet frame.
  create_6to4_frame (type+3, data);

  return (EXIT_SUCCESS);
}

// Update MTU spinbutton entry value.
int
update_mtu_spinbutton (int type, SPSData *data) {

  switch (type) {
    case 0:  // IPv4 TCP
      gtk_spin_button_set_value (GTK_SPIN_BUTTON (data->spinbutton5), data->ifmtu[type]);
      break;
    case 1:  // IPv4 ICMP
      gtk_spin_button_set_value (GTK_SPIN_BUTTON (data->spinbutton6), data->ifmtu[type]);
      break;
    case 2:  // IPv4 UDP
      gtk_spin_button_set_value (GTK_SPIN_BUTTON (data->spinbutton7), data->ifmtu[type]);
      break;

    case 3:  // IPv6 TCP
      gtk_spin_button_set_value (GTK_SPIN_BUTTON (data->spinbutton8), data->ifmtu[type]);
      break;
    case 4:  // IPv6 ICMP
      gtk_spin_button_set_value (GTK_SPIN_BUTTON (data->spinbutton9), data->ifmtu[type]);
      break;
    case 5:  // IPv6 UDP
      gtk_spin_button_set_value (GTK_SPIN_BUTTON (data->spinbutton10), data->ifmtu[type]);
      break;

    case 6:  // 6to4 TCP
      gtk_spin_button_set_value (GTK_SPIN_BUTTON (data->spinbutton17), data->ifmtu[type]);
      break;
    case 7:  // 6to4 ICMP
      gtk_spin_button_set_value (GTK_SPIN_BUTTON (data->spinbutton18), data->ifmtu[type]);
      break;
    case 8:  // 6to4 UDP
      gtk_spin_button_set_value (GTK_SPIN_BUTTON (data->spinbutton19), data->ifmtu[type]);
      break;

    case 9:  // Traceroute IPv4 TCP
      gtk_spin_button_set_value (GTK_SPIN_BUTTON (data->tr_spinbutton11), data->ifmtu[type]);
      break;
    case 10:  // Traceroute IPv4 ICMP
      gtk_spin_button_set_value (GTK_SPIN_BUTTON (data->tr_spinbutton12), data->ifmtu[type]);
      break;
    case 11:  // Traceroute IPv4 UDP
      gtk_spin_button_set_value (GTK_SPIN_BUTTON (data->tr_spinbutton13), data->ifmtu[type]);
      break;

    case 12:  // Traceroute IPv6 TCP
      gtk_spin_button_set_value (GTK_SPIN_BUTTON (data->tr_spinbutton14), data->ifmtu[type]);
      break;
    case 13:  // Traceroute IPv6 ICMP
      gtk_spin_button_set_value (GTK_SPIN_BUTTON (data->tr_spinbutton15), data->ifmtu[type]);
      break;
    case 14:  // Traceroute IPv6 UDP
      gtk_spin_button_set_value (GTK_SPIN_BUTTON (data->tr_spinbutton16), data->ifmtu[type]);
      break;
    default:
      fprintf (stderr, "ERROR: Invalid upper layer protocol type %i in update_mtu_spinbutton().\n", type);
      exit (EXIT_FAILURE);
  }

  return (EXIT_SUCCESS);
}
