/*  Simple Packet Sender (SPS) - a linux packet crafting network
    tool using GTK+ graphical interface.

    Copyright (C) 2011-2023  Hohlraum (h0h1r4um@gmail.com)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "main.h"

// Apply defaults for IPv6 TCP over IPv4 (6to4).
int
sixto4_tcp6_default (SPSData *data) {

  int status, *ip4_flags;
  char *value, *ipaddress;

  // IPv4 header fragmentation flags and offset
  ip4_flags = allocate_intmem (4);

  // Array to hold an IPv4 or IPv6 address + '\0'
  ipaddress = allocate_strmem (INET6_ADDRSTRLEN);

  // String to hold miscellaneous calculated values
  value = allocate_strmem (TMP_STRINGLEN);

  // Number of ethernet frames
  data->nframes[6] = 1;  // 6to4 IPv6 TCP
  data->nframes[15] = 1;  // 6to4 IPv6 TCP traceroute

  // Ethernet header
  
  // Destination link-layer (MAC) address (48 bits): default to zero; user needs to fill out
  memset (data->ethhdr[6].dst_mac, 0, 6 * sizeof (uint8_t));
  memset (data->ethhdr[15].dst_mac, 0, 6 * sizeof (uint8_t));
  memset (value, 0, TMP_STRINGLEN * sizeof (char));
  mac_bin2string (data->ethhdr[6].dst_mac, value);
  gtk_entry_set_text (GTK_ENTRY (data->entry600), value);

  // Source Interface Name: defaults to blank
  memset (data->ifname[6], 0, TMP_STRINGLEN * sizeof (char));
  memset (data->ifname[15], 0, TMP_STRINGLEN * sizeof (char));
  gtk_entry_set_text (GTK_ENTRY (data->entry460), "");

  // Default interface MTU (RFC 4213)
  data->ifmtu[6] = 1280;  // 6to4 IPv6 TCP
  data->ifmtu[15] = 1280;  // 6to4 IPv6 TCP traceroute

  // Set spinbutton value to interface MTU.
  update_mtu_spinbutton (6, data);

  // Source link-layer (MAC) address (48 bits): default to zero; user needs to fill out
  memset (data->ethhdr[6].src_mac, 0, 6 * sizeof (uint8_t));
  memset (data->ethhdr[15].src_mac, 0, 6 * sizeof (uint8_t));
  memset (value, 0, TMP_STRINGLEN * sizeof (char));
  mac_bin2string (data->ethhdr[6].src_mac, value);
  gtk_entry_set_text (GTK_ENTRY (data->entry601), value);

  // Ethernet type code (16 bits): IPv4
  // http://www.iana.org/assignments/ethernet-numbers
  data->ethhdr[6].type_code = htons (ETH_P_IP);
  data->ethhdr[15].type_code = htons (ETH_P_IP);
  sprintf (value, "%" PRIu16, (uint16_t) ntohs (data->ethhdr[6].type_code));
  gtk_entry_set_text (GTK_ENTRY (data->entry602), value);

  // According to Section 3.5 of RFC 4213, IPv4 header options are not permitted.

  // IPv4 header for tunneling IPv6 over IPv4 (6to4) (Section 3.5 of RFC 4213)

  // IPv4 header length (4 bits): Number of 32-bit words in header = 5
  data->ip4hdr[6].ip_hl = IP4_HDRLEN / sizeof (uint32_t);
  data->ip4hdr[15].ip_hl = IP4_HDRLEN / sizeof (uint32_t);
  sprintf (value, "%" PRIu8, (uint8_t) data->ip4hdr[6].ip_hl);
  gtk_entry_set_text (GTK_ENTRY (data->entry609), value);

  // Internet Protocol version (4 bits): IPv4
  data->ip4hdr[6].ip_v = 4u;
  data->ip4hdr[15].ip_v = 4u;
  sprintf (value, "%" PRIu8, (uint8_t) data->ip4hdr[6].ip_v);
  gtk_entry_set_text (GTK_ENTRY (data->entry610), value);

  // Type of service (8 bits)
  data->ip4hdr[6].ip_tos = 0;
  data->ip4hdr[15].ip_tos = 0;
  sprintf (value, "%" PRIu8, (uint8_t) data->ip4hdr[6].ip_tos);
  gtk_entry_set_text (GTK_ENTRY (data->entry611), value);

  // Total length of datagram (16 bits): IPv4 header + IPv6 header + IPv6 payload length
  // Assumes neither IPv4 header options nor TCP options.
  data->ip4hdr[6].ip_len = htons (IP4_HDRLEN + IP6_HDRLEN + TCP_HDRLEN + data->payloadlen[3]);
  data->ip4hdr[15].ip_len = htons (IP4_HDRLEN + IP6_HDRLEN + TCP_HDRLEN + data->payloadlen[12]);
  sprintf (value, "%" PRIu16, (uint16_t) ntohs (data->ip4hdr[6].ip_len));
  gtk_entry_set_text (GTK_ENTRY (data->entry612), value);

  // ID sequence number (16 bits): unused, since single datagram
  data->ip4hdr[6].ip_id = htons (ran16_0to65535 (data));
  data->ip4hdr[15].ip_id = htons (ran16_0to65535 (data));
  sprintf (value, "%" PRIu16, (uint16_t) ntohs (data->ip4hdr[6].ip_id));
  gtk_entry_set_text (GTK_ENTRY (data->entry613), value);

  // Zero (1 bit)
  ip4_flags[0] = 0;
  sprintf (value, "%i", ip4_flags[0]);
  gtk_entry_set_text (GTK_ENTRY (data->entry614), value);

  // Do not fragment flag (1 bit): 0 (see Section 3.2.1 of RFC 4213)
  ip4_flags[1] = 0;
  sprintf (value, "%i", ip4_flags[1]);
  gtk_entry_set_text (GTK_ENTRY (data->entry615), value);

  // More fragments following flag (1 bit)
  ip4_flags[2] = 0;
  sprintf (value, "%i", ip4_flags[2]);
  gtk_entry_set_text (GTK_ENTRY (data->entry616), value);

  // Fragmentation offset (13 bits)
  ip4_flags[3] = 0;
  sprintf (value, "%i", ip4_flags[3]);
  gtk_entry_set_text (GTK_ENTRY (data->entry617), value);

  // Flags, and Fragmentation offset (3, 13 bits): 0 since single datagram
  data->ip4hdr[6].ip_off = htons ((ip4_flags[0] << 15) + (ip4_flags[1] << 14) +
    (ip4_flags[2] << 13) + ip4_flags[3]);
  data->ip4hdr[15].ip_off = htons ((ip4_flags[0] << 15) + (ip4_flags[1] << 14) +
    (ip4_flags[2] << 13) + ip4_flags[3]);

  // Time-to-Live (8 bits): use maximum value
  data->ip4hdr[6].ip_ttl = 255u;
  data->ip4hdr[15].ip_ttl = 255u;
  sprintf (value, "%" PRIu8, (uint8_t) data->ip4hdr[6].ip_ttl);
  gtk_entry_set_text (GTK_ENTRY (data->entry618), value);

  // Transport layer protocol (8 bits): 41 for IPv6 (Section 3.5 of RFC 4213)
  data->ip4hdr[6].ip_p = IPPROTO_IPV6;
  data->ip4hdr[15].ip_p = IPPROTO_IPV6;
  sprintf (value, "%" PRIu8, (uint8_t) data->ip4hdr[6].ip_p);
  gtk_entry_set_text (GTK_ENTRY (data->entry619), value);

  // Source IPv4 address (32 bits): default to random
  // For non-traceroute packets.
  data->ran_tcp6to4_sourceip = 1;
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton13), TRUE);
  if ((status = inet_pton (AF_INET, ran4_addr (ipaddress, data), &(data->ip4hdr[6].ip_src))) != 1) {
    sprintf (data->error_text, "sixto4_tcp6_default(): inet_pton() failed for source IPv4 address.\nError message: %s", strerror (status));
    data->parent = data->main_window;
    report_error (data);
    free (ip4_flags);
    free (ipaddress);
    free (value);
    return (EXIT_FAILURE);
  }
  gtk_entry_set_text (GTK_ENTRY (data->entry125), ipaddress);

  // Source IPv4 address (32 bits): default to loopback
  // For traceroute packets.
  if ((status = inet_pton (AF_INET, "127.0.0.1", &(data->ip4hdr[15].ip_src))) != 1) {
    sprintf (data->error_text, "sixto4_tcp6_default(): inet_pton() failed for source IPv4 address.\nError message: %s", strerror (status));
    data->parent = data->main_window;
    report_error (data);
    free (ip4_flags);
    return (EXIT_FAILURE);
  }

  // Destination IPv4 address (32 bits): 6to4 anycast address
  // For non-traceroute packets.
  memset (ipaddress, 0, INET6_ADDRSTRLEN * sizeof (char));
  strncpy (ipaddress, "192.88.99.1", INET6_ADDRSTRLEN - 1);  // Minus 1 for string termination.
  if ((status = inet_pton (AF_INET, ipaddress, &(data->ip4hdr[6].ip_dst))) != 1) {
    sprintf (data->error_text, "sixto4_tcp6_default(): inet_pton() failed for destination IPv4 address.\nError message: %s", strerror (status));
    data->parent = data->main_window;
    report_error (data);
    free (ip4_flags);
    free (ipaddress);
    free (value);
    return (EXIT_FAILURE);
  }
  gtk_entry_set_text (GTK_ENTRY (data->entry620), ipaddress);

  // Destination IPv4 address (32 bits): 6to4 anycast address
  // For traceroute packets.
  if ((status = inet_pton (AF_INET, ipaddress, &(data->ip4hdr[15].ip_dst))) != 1) {
    sprintf (data->error_text, "sixto4_tcp6_default(): inet_pton() failed for destination IPv4 address.\nError message: %s", strerror (status));
    data->parent = data->main_window;
    report_error (data);
    free (ip4_flags);
    return (EXIT_FAILURE);
  }

  // IPv4 header checksum (16 bits): set to 0 when calculating checksum
  data->ip4hdr[6].ip_sum = 0;
  data->ip4hdr[6].ip_sum = checksum ((uint16_t *) &data->ip4hdr[6], IP4_HDRLEN);
  data->ip4hdr[15].ip_sum = 0;
  data->ip4hdr[15].ip_sum = checksum ((uint16_t *) &data->ip4hdr[15], IP4_HDRLEN);
  sprintf (value, "%" PRIu16, (uint16_t) ntohs (data->ip4hdr[6].ip_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry621), value);

  // Free allocated memory.
  free (ip4_flags);
  free (ipaddress);
  free (value);

  // Copy IPv6 header to 6to4 IPv6 header, since create_6to4_frame() doesn't copy it.
  memcpy (&data->ip6hdr[6], &data->ip6hdr[3], IP6_HDRLEN * sizeof (uint8_t));
  memcpy (&data->ip6hdr[15], &data->ip6hdr[12], IP6_HDRLEN * sizeof (uint8_t));

  // Update ethernet frames.
  create_6to4_frame (6, data);
  create_6to4_frame (15, data);

  return (EXIT_SUCCESS);
}

// Apply defaults for IPv6 ICMP over IPv4 (6to4).
int
sixto4_icmp6_default (SPSData *data) {

  int status, *ip4_flags;
  char *value, *ipaddress;

  // IPv4 header fragmentation flags and offset
  ip4_flags = allocate_intmem (4);

  // Array to hold an IPv4 or IPv6 address + '\0'
  ipaddress = allocate_strmem (INET6_ADDRSTRLEN);

  // String to hold miscellaneous calculated values
  value = allocate_strmem (TMP_STRINGLEN);

  // Number of ethernet frames
  data->nframes[7] = 1;  // 6to4 IPv6 ICMP
  data->nframes[16] = 1;  // 6to4 IPv6 ICMP traceroute

  // Ethernet header
  
  // Destination link-layer (MAC) address (48 bits): default to zero; user needs to fill out
  memset (data->ethhdr[7].dst_mac, 0, 6 * sizeof (uint8_t));
  memset (data->ethhdr[16].dst_mac, 0, 6 * sizeof (uint8_t));
  memset (value, 0, TMP_STRINGLEN * sizeof (char));
  mac_bin2string (data->ethhdr[7].dst_mac, value);
  gtk_entry_set_text (GTK_ENTRY (data->entry603), value);

  // Source Interface Name: defaults to blank
  memset (data->ifname[7], 0, TMP_STRINGLEN * sizeof (char));
  memset (data->ifname[16], 0, TMP_STRINGLEN * sizeof (char));
  gtk_entry_set_text (GTK_ENTRY (data->entry461), "");

  // Default interface MTU (RFC 4213)
  data->ifmtu[7] = 1280;  // 6to4 IPv6 ICMP
  data->ifmtu[16] = 1280;  // 6to4 IPv6 ICMP traceroute

  // Set spinbutton value to interface MTU.
  update_mtu_spinbutton (7, data);

  // Source link-layer (MAC) address (48 bits): default to zero; user needs to fill out
  memset (data->ethhdr[7].src_mac, 0, 6 * sizeof (uint8_t));
  memset (data->ethhdr[16].src_mac, 0, 6 * sizeof (uint8_t));
  memset (value, 0, TMP_STRINGLEN * sizeof (char));
  mac_bin2string (data->ethhdr[7].src_mac, value);
  gtk_entry_set_text (GTK_ENTRY (data->entry604), value);

  // Ethernet type code (16 bits): IPv4
  // http://www.iana.org/assignments/ethernet-numbers
  data->ethhdr[7].type_code = htons (ETH_P_IP);
  data->ethhdr[16].type_code = htons (ETH_P_IP);
  sprintf (value, "%" PRIu16, (uint16_t) ntohs (data->ethhdr[7].type_code));
  gtk_entry_set_text (GTK_ENTRY (data->entry605), value);

  // According to Section 3.5 of RFC 4213, IPv4 header options are not permitted.

  // IPv4 header for tunneling IPv6 over IPv4 (6to4) (Section 3.5 of RFC 4213)

  // IPv4 header length (4 bits): Number of 32-bit words in header = 5
  data->ip4hdr[7].ip_hl = IP4_HDRLEN / sizeof (uint32_t);
  data->ip4hdr[16].ip_hl = IP4_HDRLEN / sizeof (uint32_t);
  sprintf (value, "%" PRIu8, (uint8_t) data->ip4hdr[7].ip_hl);
  gtk_entry_set_text (GTK_ENTRY (data->entry622), value);

  // Internet Protocol version (4 bits): IPv4
  data->ip4hdr[7].ip_v = 4u;
  data->ip4hdr[16].ip_v = 4u;
  sprintf (value, "%" PRIu8, (uint8_t) data->ip4hdr[7].ip_v);
  gtk_entry_set_text (GTK_ENTRY (data->entry623), value);

  // Type of service (8 bits)
  data->ip4hdr[7].ip_tos = 0;
  data->ip4hdr[16].ip_tos = 0;
  sprintf (value, "%" PRIu8, (uint8_t) data->ip4hdr[7].ip_tos);
  gtk_entry_set_text (GTK_ENTRY (data->entry624), value);

  // Total length of datagram (16 bits): IPv4 header + IPv6 header + IPv6 payload length
  // Assumes n0 IPv4 header options.
  data->ip4hdr[7].ip_len = htons (IP4_HDRLEN + IP6_HDRLEN + ICMP_HDRLEN + data->payloadlen[4]);
  data->ip4hdr[16].ip_len = htons (IP4_HDRLEN + IP6_HDRLEN + ICMP_HDRLEN + data->payloadlen[13]);
  sprintf (value, "%" PRIu16, (uint16_t) ntohs (data->ip4hdr[7].ip_len));
  gtk_entry_set_text (GTK_ENTRY (data->entry625), value);

  // ID sequence number (16 bits): unused, since single datagram
  data->ip4hdr[7].ip_id = htons (ran16_0to65535 (data));
  data->ip4hdr[16].ip_id = htons (ran16_0to65535 (data));
  sprintf (value, "%" PRIu16, (uint16_t) ntohs (data->ip4hdr[7].ip_id));
  gtk_entry_set_text (GTK_ENTRY (data->entry626), value);

  // Zero (1 bit)
  ip4_flags[0] = 0;
  sprintf (value, "%i", ip4_flags[0]);
  gtk_entry_set_text (GTK_ENTRY (data->entry627), value);

  // Do not fragment flag (1 bit): 0 (see Section 3.2.1 of RFC 4213)
  ip4_flags[1] = 0;
  sprintf (value, "%i", ip4_flags[1]);
  gtk_entry_set_text (GTK_ENTRY (data->entry628), value);

  // More fragments following flag (1 bit)
  ip4_flags[2] = 0;
  sprintf (value, "%i", ip4_flags[2]);
  gtk_entry_set_text (GTK_ENTRY (data->entry629), value);

  // Fragmentation offset (13 bits)
  ip4_flags[3] = 0;
  sprintf (value, "%i", ip4_flags[3]);
  gtk_entry_set_text (GTK_ENTRY (data->entry630), value);

  // Flags, and Fragmentation offset (3, 13 bits): 0 since single datagram
  data->ip4hdr[7].ip_off = htons ((ip4_flags[0] << 15) + (ip4_flags[1] << 14) +
    (ip4_flags[2] << 13) + ip4_flags[3]);
  data->ip4hdr[16].ip_off = htons ((ip4_flags[0] << 15) + (ip4_flags[1] << 14) +
    (ip4_flags[2] << 13) + ip4_flags[3]);

  // Time-to-Live (8 bits): use maximum value
  data->ip4hdr[7].ip_ttl = 255u;
  data->ip4hdr[16].ip_ttl = 255u;
  sprintf (value, "%" PRIu8, (uint8_t) data->ip4hdr[7].ip_ttl);
  gtk_entry_set_text (GTK_ENTRY (data->entry631), value);

  // Transport layer protocol (8 bits): 41 for IPv6 (Section 3.5 of RFC 4213)
  data->ip4hdr[7].ip_p = IPPROTO_IPV6;
  data->ip4hdr[16].ip_p = IPPROTO_IPV6;
  sprintf (value, "%" PRIu8, (uint8_t) data->ip4hdr[7].ip_p);
  gtk_entry_set_text (GTK_ENTRY (data->entry632), value);

  // Source IPv4 address (32 bits): default to random
  // For non-traceroute packets.
  data->ran_icmp6to4_sourceip = 1;
  if ((status = inet_pton (AF_INET, ran4_addr (ipaddress, data), &(data->ip4hdr[7].ip_src))) != 1) {
    sprintf (data->error_text, "sixto4_icmp6_default(): inet_pton() failed for source IPv4 address.\nError message: %s", strerror (status));
    data->parent = data->main_window;
    report_error (data);
    free (ip4_flags);
    free (ipaddress);
    free (value);
    return (EXIT_FAILURE);
  }
  gtk_entry_set_text (GTK_ENTRY (data->entry160), ipaddress);

  // Source IPv4 address (32 bits): default to loopback
  // For traceroute packets.
  if ((status = inet_pton (AF_INET, "127.0.0.1", &(data->ip4hdr[16].ip_src))) != 1) {
    sprintf (data->error_text, "sixto4_icmp6_default(): inet_pton() failed for source IPv4 address.\nError message: %s", strerror (status));
    data->parent = data->main_window;
    report_error (data);
    free (ip4_flags);
    return (EXIT_FAILURE);
  }

  // Destination IPv4 address (32 bits): 6to4 anycast address
  // For non-traceroute packets.
  memset (ipaddress, 0, INET6_ADDRSTRLEN * sizeof (char));
  strncpy (ipaddress, "192.88.99.1", INET6_ADDRSTRLEN - 1);  // Minus 1 for string termination.
  if ((status = inet_pton (AF_INET, ipaddress, &(data->ip4hdr[7].ip_dst))) != 1) {
    sprintf (data->error_text, "sixto4_icmp6_default(): inet_pton() failed for destination IPv4 address.\nError message: %s", strerror (status));
    data->parent = data->main_window;
    report_error (data);
    free (ip4_flags);
    free (ipaddress);
    free (value);
    return (EXIT_FAILURE);
  }
  gtk_entry_set_text (GTK_ENTRY (data->entry633), ipaddress);

  // Destination IPv4 address (32 bits): 6to4 anycast address
  // For traceroute packets.
  if ((status = inet_pton (AF_INET, ipaddress, &(data->ip4hdr[16].ip_dst))) != 1) {
    sprintf (data->error_text, "sixto4_icmp6_default(): inet_pton() failed for destination IPv4 address.\nError message: %s", strerror (status));
    data->parent = data->main_window;
    report_error (data);
    free (ip4_flags);
    return (EXIT_FAILURE);
  }

  // IPv4 header checksum (16 bits): set to 0 when calculating checksum
  data->ip4hdr[7].ip_sum = 0;
  data->ip4hdr[7].ip_sum = checksum ((uint16_t *) &data->ip4hdr[7], IP4_HDRLEN);
  data->ip4hdr[16].ip_sum = 0;
  data->ip4hdr[16].ip_sum = checksum ((uint16_t *) &data->ip4hdr[16], IP4_HDRLEN);
  sprintf (value, "%" PRIu16, (uint16_t) ntohs (data->ip4hdr[7].ip_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry634), value);

  // Free allocated memory.
  free (ip4_flags);
  free (ipaddress);
  free (value);

  // Copy IPv6 header to 6to4 IPv6 header, since create_6to4_frame() doesn't copy it.
  memcpy (&data->ip6hdr[7], &data->ip6hdr[4], IP6_HDRLEN * sizeof (uint8_t));
  memcpy (&data->ip6hdr[16], &data->ip6hdr[13], IP6_HDRLEN * sizeof (uint8_t));

  // Update ethernet frames.
  create_6to4_frame (7, data);
  create_6to4_frame (16, data);

  return (EXIT_SUCCESS);
}

// Apply defaults for IPv6 UDP over IPv4 (6to4).
int
sixto4_udp6_default (SPSData *data) {

  int status, *ip4_flags;
  char *value, *ipaddress;

  // IPv4 header fragmentation flags and offset
  ip4_flags = allocate_intmem (4);

  // Array to hold an IPv4 or IPv6 address + '\0'
  ipaddress = allocate_strmem (INET6_ADDRSTRLEN);

  // String to hold miscellaneous calculated values
  value = allocate_strmem (TMP_STRINGLEN);

  // Number of ethernet frames
  data->nframes[8] = 1;  // 6to4 IPv6 UDP
  data->nframes[17] = 1;  // 6to4 IPv6 UDP traceroute

  // Ethernet header
  
  // Destination link-layer (MAC) address (48 bits): default to zero; user needs to fill out
  memset (data->ethhdr[8].dst_mac, 0, 6 * sizeof (uint8_t));
  memset (data->ethhdr[17].dst_mac, 0, 6 * sizeof (uint8_t));
  memset (value, 0, TMP_STRINGLEN * sizeof (char));
  mac_bin2string (data->ethhdr[8].dst_mac, value);
  gtk_entry_set_text (GTK_ENTRY (data->entry606), value);

  // Source Interface Name: defaults to blank
  memset (data->ifname[8], 0, TMP_STRINGLEN * sizeof (char));
  memset (data->ifname[17], 0, TMP_STRINGLEN * sizeof (char));
  gtk_entry_set_text (GTK_ENTRY (data->entry462), "");

  // Default interface MTU (RFC 4213)
  data->ifmtu[8] = 1280;  // 6to4 IPv6 UDP
  data->ifmtu[17] = 1280;  // 6to4 IPv6 UDP traceroute

  // Set spinbutton value to interface MTU.
  update_mtu_spinbutton (8, data);

  // Source link-layer (MAC) address (48 bits): default to zero; user needs to fill out
  memset (data->ethhdr[8].src_mac, 0, 6 * sizeof (uint8_t));
  memset (data->ethhdr[17].src_mac, 0, 6 * sizeof (uint8_t));
  memset (value, 0, TMP_STRINGLEN * sizeof (char));
  mac_bin2string (data->ethhdr[8].src_mac, value);
  gtk_entry_set_text (GTK_ENTRY (data->entry607), value);

  // Ethernet type code (16 bits): IPv4
  // http://www.iana.org/assignments/ethernet-numbers
  data->ethhdr[8].type_code = htons (ETH_P_IP);
  data->ethhdr[17].type_code = htons (ETH_P_IP);
  sprintf (value, "%" PRIu16, (uint16_t) ntohs (data->ethhdr[8].type_code));
  gtk_entry_set_text (GTK_ENTRY (data->entry608), value);

  // According to Section 3.5 of RFC 4213, IPv4 header options are not permitted.

  // IPv4 header for tunneling IPv6 over IPv4 (6to4) (Section 3.5 of RFC 4213)

  // IPv4 header length (4 bits): Number of 32-bit words in header = 5
  data->ip4hdr[8].ip_hl = IP4_HDRLEN / sizeof (uint32_t);
  data->ip4hdr[17].ip_hl = IP4_HDRLEN / sizeof (uint32_t);
  sprintf (value, "%" PRIu8, (uint8_t) data->ip4hdr[8].ip_hl);
  gtk_entry_set_text (GTK_ENTRY (data->entry635), value);

  // Internet Protocol version (4 bits): IPv4
  data->ip4hdr[8].ip_v = 4u;
  data->ip4hdr[17].ip_v = 4u;
  sprintf (value, "%" PRIu8, (uint8_t) data->ip4hdr[8].ip_v);
  gtk_entry_set_text (GTK_ENTRY (data->entry636), value);

  // Type of service (8 bits)
  data->ip4hdr[8].ip_tos = 0;
  data->ip4hdr[17].ip_tos = 0;
  sprintf (value, "%" PRIu8, (uint8_t) data->ip4hdr[8].ip_tos);
  gtk_entry_set_text (GTK_ENTRY (data->entry637), value);

  // Total length of datagram (16 bits): IPv4 header + IPv6 header + IPv6 payload length
  // Assumes no IPv4 header options.
  data->ip4hdr[8].ip_len = htons (IP4_HDRLEN + IP6_HDRLEN + UDP_HDRLEN + data->payloadlen[5]);
  data->ip4hdr[17].ip_len = htons (IP4_HDRLEN + IP6_HDRLEN + UDP_HDRLEN + data->payloadlen[14]);
  sprintf (value, "%" PRIu16, (uint16_t) ntohs (data->ip4hdr[8].ip_len));
  gtk_entry_set_text (GTK_ENTRY (data->entry638), value);

  // ID sequence number (16 bits): unused, since single datagram
  data->ip4hdr[8].ip_id = htons (ran16_0to65535 (data));
  data->ip4hdr[17].ip_id = htons (ran16_0to65535 (data));
  sprintf (value, "%" PRIu16, (uint16_t) ntohs (data->ip4hdr[8].ip_id));
  gtk_entry_set_text (GTK_ENTRY (data->entry639), value);

  // Zero (1 bit)
  ip4_flags[0] = 0;
  sprintf (value, "%i", ip4_flags[0]);
  gtk_entry_set_text (GTK_ENTRY (data->entry640), value);

  // Do not fragment flag (1 bit): 0 (see Section 3.2.1 of RFC 4213)
  ip4_flags[1] = 0;
  sprintf (value, "%i", ip4_flags[1]);
  gtk_entry_set_text (GTK_ENTRY (data->entry641), value);

  // More fragments following flag (1 bit)
  ip4_flags[2] = 0;
  sprintf (value, "%i", ip4_flags[2]);
  gtk_entry_set_text (GTK_ENTRY (data->entry642), value);

  // Fragmentation offset (13 bits)
  ip4_flags[3] = 0;
  sprintf (value, "%i", ip4_flags[3]);
  gtk_entry_set_text (GTK_ENTRY (data->entry643), value);

  // Flags, and Fragmentation offset (3, 13 bits): 0 since single datagram
  data->ip4hdr[8].ip_off = htons ((ip4_flags[0] << 15) + (ip4_flags[1] << 14) +
    (ip4_flags[2] << 13) + ip4_flags[3]);
  data->ip4hdr[17].ip_off = htons ((ip4_flags[0] << 15) + (ip4_flags[1] << 14) +
    (ip4_flags[2] << 13) + ip4_flags[3]);

  // Time-to-Live (8 bits): use maximum value
  data->ip4hdr[8].ip_ttl = 255u;
  data->ip4hdr[17].ip_ttl = 255u;
  sprintf (value, "%" PRIu8, (uint8_t) data->ip4hdr[8].ip_ttl);
  gtk_entry_set_text (GTK_ENTRY (data->entry644), value);

  // Transport layer protocol (8 bits): 41 for IPv6 (Section 3.5 of RFC 4213)
  data->ip4hdr[8].ip_p = IPPROTO_IPV6;
  data->ip4hdr[17].ip_p = IPPROTO_IPV6;
  sprintf (value, "%" PRIu8, (uint8_t) data->ip4hdr[8].ip_p);
  gtk_entry_set_text (GTK_ENTRY (data->entry645), value);

  // Source IPv4 address (32 bits): default to random
  // For non-traceroute packets.
  data->ran_udp6to4_sourceip = 1;
  if ((status = inet_pton (AF_INET, ran4_addr (ipaddress, data), &(data->ip4hdr[8].ip_src))) != 1) {
    sprintf (data->error_text, "sixto4_udp6_default(): inet_pton() failed for source IPv4 address.\nError message: %s", strerror (status));
    data->parent = data->main_window;
    report_error (data);
    free (ip4_flags);
    free (ipaddress);
    free (value);
    return (EXIT_FAILURE);
  }
  gtk_entry_set_text (GTK_ENTRY (data->entry161), ipaddress);

  // Source IPv4 address (32 bits): default to loopback
  // For traceroute packets.
  if ((status = inet_pton (AF_INET, "127.0.0.1", &(data->ip4hdr[17].ip_src))) != 1) {
    sprintf (data->error_text, "sixto4_udp6_default(): inet_pton() failed for source IPv4 address.\nError message: %s", strerror (status));
    data->parent = data->main_window;
    report_error (data);
    free (ip4_flags);
    return (EXIT_FAILURE);
  }

  // Destination IPv4 address (32 bits): 6to4 anycast address
  // For non-traceroute packets.
  memset (ipaddress, 0, INET6_ADDRSTRLEN * sizeof (char));
  strncpy (ipaddress, "192.88.99.1", INET6_ADDRSTRLEN - 1);  // Minus 1 for string termination.
  if ((status = inet_pton (AF_INET, ipaddress, &(data->ip4hdr[8].ip_dst))) != 1) {
    sprintf (data->error_text, "sixto4_udp6_default(): inet_pton() failed for destination IPv4 address.\nError message: %s", strerror (status));
    data->parent = data->main_window;
    report_error (data);
    free (ip4_flags);
    free (ipaddress);
    free (value);
    return (EXIT_FAILURE);
  }
  gtk_entry_set_text (GTK_ENTRY (data->entry646), ipaddress);

  // Destination IPv4 address (32 bits): 6to4 anycast address
  // For traceroute packets.
  if ((status = inet_pton (AF_INET, ipaddress, &(data->ip4hdr[17].ip_dst))) != 1) {
    sprintf (data->error_text, "sixto4_udp6_default(): inet_pton() failed for destination IPv4 address.\nError message: %s", strerror (status));
    data->parent = data->main_window;
    report_error (data);
    free (ip4_flags);
    return (EXIT_FAILURE);
  }

  // IPv4 header checksum (16 bits): set to 0 when calculating checksum
  data->ip4hdr[8].ip_sum = 0;
  data->ip4hdr[8].ip_sum = checksum ((uint16_t *) &data->ip4hdr[8], IP4_HDRLEN);
  data->ip4hdr[17].ip_sum = 0;
  data->ip4hdr[17].ip_sum = checksum ((uint16_t *) &data->ip4hdr[17], IP4_HDRLEN);
  sprintf (value, "%" PRIu16, (uint16_t) ntohs (data->ip4hdr[8].ip_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry647), value);

  // Free allocated memory.
  free (ip4_flags);
  free (ipaddress);
  free (value);

  // Copy IPv6 header to 6to4 IPv6 header, since create_6to4_frame() doesn't copy it.
  memcpy (&data->ip6hdr[8], &data->ip6hdr[5], IP6_HDRLEN * sizeof (uint8_t));
  memcpy (&data->ip6hdr[17], &data->ip6hdr[14], IP6_HDRLEN * sizeof (uint8_t));

  // Update ethernet frames.
  create_6to4_frame (8, data);
  create_6to4_frame (17, data);

  return (EXIT_SUCCESS);
}
